#ifndef CHARGEQUALITYPROCESSOR_H
#define CHARGEQUALITYPROCESSOR_H

// C++
#include <string>
#include <fstream>

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "lccd/ConditionsMap.hh"
//GEAR
#include "gear/TPCParameters.h"

//ROOT
#include "TH1D.h"
#include "TNtuple.h"

// Aida
#ifdef MARLIN_USE_AIDA
//AIDA
#include <marlin/AIDAProcessor.h>
#endif

namespace marlintpc
{

/**
 *  A processor to calculate the mean90 of the hit charge spectrum.
 *  Should be run with the AidaProcessor in the chain to save the histogram
 *
 *  <h4>Input - Prerequisites</h4>
 *  a collection of Tracks
 *
 *  <h4>Output</h4>
 *  a mean90 and its error stored in tuple in aida file
 *
 * @param InputCollectionName Name of the input collection (default: ClupatraTracks)
 *
 * @author Astrid Muennich, DESY
 */

class ChargeQualityProcessor : public marlin::Processor
{

    public:

        /* This method will be called by the marlin package
         * It returns a processor of the currend type 
         */
        virtual Processor*  newProcessor()
        {
            return new ChargeQualityProcessor ;
        }

        /* the default constructor
         * here the processor parameters are registered to the marlin package
         * other initialisation should be placed in the init method
         */
        ChargeQualityProcessor() ;

        /* Called at the beginning of the job before anything is read.
         * Use to initialize the processor, e.g. book histograms
         */
        virtual void init() ;

        /* Called for every run.
         * in this processor it is not used
         */
        virtual void processRunHeader(lcio::LCRunHeader* run) ;

        /* Called for every event - the working horse.
         * Here the real conversion take place
         */
        virtual void processEvent(lcio::LCEvent * evt) ;

        /* This method is only called if the check flag is set (default) in the main processor
         * (where the input files are given)
         * It should be used to create check plots which are not needed in a complete analysis
         * (where normaly the check flag is deactivated)
         */
        virtual void check(lcio::LCEvent * evt) ;


        /* Called after data processing for clean up.
         * e.g. saving control histogramms, delete created objects
         * (which are not stored in a collection/event), etc.
         */
        virtual void end() ;




    protected:

    std::string _inputColName ;  ///<- the name of the input collection
	std::map< std::pair<int,int>, TH1*> _hitChargeRow;

    private:
    TH1F* _histoQ;
	TNtuple* _tuple;
	///upper range of the expected hit charge for the histogram
	int _Qmax;
	int _meanMethod; ///< Defines the method of calculating the mean of the charge-histograms. (default 1:RMS90, 2:Gaussian fit within range of 3 sigma, 3:rms after 5 sigma iterative cut of tails)
	const gear::TPCParameters* _gearParameters;
} ;

}

#endif



