#ifndef CollectDelayedHitsProcessor_h
#define CollectDelayedHitsProcessor_h 1


// C++
#include <string>
#include <cmath>

// LCIO
#include "lcio.h"
#include <EVENT/TrackerHit.h>
#include "EVENT/LCIO.h"
#include "EVENT/LCCollection.h"
#include <IMPL/LCCollectionVec.h>
#include "IMPL/TrackImpl.h"
#include <IMPL/LCFlagImpl.h>
#include "IntersectionCalculator.h"

// Aida
#ifdef MARLIN_USE_AIDA

#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>

//marlin
#include <marlin/ConditionsProcessor.h>
#include "TrackFitterFactory.h"
#include "TrackFitterBase.h"
#include "LinearRegressionProcessor.h"
#include "LinearTrackRegression.h"
#include "marlin/Processor.h"
#include "TimePixEventDisplayCreator.h"

//root
#include "TH1F.h"

#endif

//root
//#include "TH2F.h"
//#include "TF1.h"

/*
namespace AIDA{
  class IHistogram1D;
  class IHistogram2D;
}
*/

namespace marlintpc{

  class TrackFitterFactory;

    /* the following comment is tranfered to doxygen documentation
     * which begins with a second asterix, or in case of on line comment an extra third slash 
     */
    /**  
     *   This processor searches for the so called "delayed hits" on tracks
     *   It creates two output collections: one containes only tracks with delayed hits,
     *   the other only tracks with "good" hits.
     *   
     *	 The algorithm calculates the mean residuals for all hits and
     *   the mean residuals to a N-1 trackfit (as the LinearGeometricMeanProcessor does)
     *   If the mean residuals differ to much it is assumed that the hit under
     *   investigation is a delayed hit
     *  
     *  
     * 
     * @param InputCollectionName: Name of input collection, TrackerRawData
     * @param OutputCollectionName1: OutputCollectionNameWithoutDelayedHits
     * @param OutputCollectionName2: OutputCollectionNameWithDelayedHits
     * @param SetOutputTransient: if not 0 the output hits collection is set transient (default: 0)
     * @param intrinsicResolution: intrinsic detector resolution in drift direction
     * @param longDiff: longitudinal Diffusion during drift [mu/sqrt(cm)] - cut is applied at (cutlevel x sigma)
     * @param CutLevel: cut is applied at (cutlevel x sigma)
     *
     * @author Christoph Brezina
     */
    
  
    class CollectDelayedHitsProcessor: public marlin::Processor {
	
    public:
	
	/* This method will be called by the marlin package
	 * It returns a processor of the currend type (here CollectDelayedHitsProcessor)
	 */
	virtual Processor*  newProcessor() { return new CollectDelayedHitsProcessor ; }
	
	/* the default constructor
	 * here the processor parameters are registered to the marlin package
	 * other initialisation should be placed in the init method
	 */
	CollectDelayedHitsProcessor() ;

	~CollectDelayedHitsProcessor() ;
	
	/* Called at the beginning of the job before anything is read.
	 * Use to initialize the processor, e.g. book histograms
	 */
	virtual void init() ;
	
	/* Called for every run.
	 * in this processor it is not used
	 */
	virtual void processRunHeader( lcio::LCRunHeader* run ) ;
	
	/* Called for every event - the working horse.
	 * Here the real conversion take place
	 */
	virtual void processEvent( lcio::LCEvent * evt ) ; 
	
	/* This method is only called if the check flag is set (default) in the main processor
	 * (where the input files are given)
	 * It should be used to create check plots which are not needed in a complete analysis
	 * (where normaly the check flag is deactivated)
	 */
	virtual void check( lcio::LCEvent * evt ) ; 
	
	
	/* Called after data processing for clean up.
	 * e.g. saving control histogramms, delete created objects
	 * (which are not stored in a collection/event), etc.
	 */
	virtual void end() ;

	
    protected:
	/* normally the doxygen documentation is palced before the method/variable
	 * but for one line comments in case of variables it may be more convenied for you
	 * to place it after the variable. In this case you use an arrow to tell doxygen
	 * to which variable the commet belongs to.
	 */
	
	std::string _inputColName ;  ///<- the name of the input collection
       	std::string _outputColNameWithoutDelayedHits; ///<- name of the output collection only with delayed hits
	std::string _outputColNameWithDelayedHits;    ///<- name of the output collection only with "good" hits
	
	int _outputIsTransient;
	
	int _histoBin;       ///<- binning of histograms
	float _histoRange;   ///<- range of histograms
	
	float _CutLevel;     ///<- cut is applied at (cutlevel x sigma)
	float _longDiff;     ///<- longitudinal Diffusion during drift  - cut is applied at (cutlevel x sigma)
	float _intrinsicRes; ///<- intrinsic detector resolution in drift direction
		
	TH1F *_OffendingZResiduals;
	TH1F *_meanZofTracksWithOffendingZResiduals;
	TimePixEventDisplayCreator *_eventDisplayCreator;

	
private:


    } ;
    
}

#endif
    
    

