#ifndef HIT_DISTANCES_PROCESSOR_H
#define HIT_DISTANCES_PROCESSOR_H 1

#define N_1D_HISTOS_HIT_DISTANCES_PROCESSOR 1
#define N_2D_HISTOS_HIT_DISTANCES_PROCESSOR 0

#include <CLHEP/Vector/ThreeVector.h>

//C++
#include <iostream>
#include <cmath>
#include <algorithm>

//LCIO
#include "marlin/Processor.h"
#include <vector>

#include <EVENT/TrackerHit.h>
#include <lcio.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>

#include "ZBinTemplateProcessor.h"

// Aida
//#ifdef MARLIN_USE_AIDA // the whole processor does not make sense without aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
// endif

//marlin
#include <marlin/Global.h>


namespace marlintpc
{

    /** Measures distances between two hits in a event.
     *  Fill the reults into AIDA histograms.
     *
     *  This Processor is derrived from ZBinTemplateProcessor to have the number of hits per track
     *  binned in z. See also processor parameters of ZBinTemplateProcessor.
     *
     *  <h4>Input - Prerequisites</h4>
     *  \li A collection of reconstructed tracks <br>
     *      To have the hits per track LCIO::TRBIT_HITS has to be set, so the hits are stored.
     *  \li A collection of reconstructed hits
     *
     *  <h4>Output</h4>
     *  \li Histograms with number of hits per track
     *  \li Histograms with number of hits per event
     *  \li Histograms with number of tracks per event
     *  \li Histograms with number of hits per track length
     *
     *  @param InputTracks Name: of input Tracks collection (default: TPCTracks)
     *  @param InputHits Name: of input TrackerHits collection (default: TPCHits)
     *  @param maxDistance: Maximum value in Distance  histogram (default: 2000 )
     *
     *  @author Christoph Brezina, University of Bonn
     */

    class HitDistancesProcessor : public ZBinTemplateProcessor<N_1D_HISTOS_HIT_DISTANCES_PROCESSOR, N_2D_HISTOS_HIT_DISTANCES_PROCESSOR>{
			      
    public:
    
	virtual Processor*  newProcessor() { return new HitDistancesProcessor ; }
    
	HitDistancesProcessor();

	virtual ~HitDistancesProcessor();

	virtual void init();

	virtual void processRunHeader(lcio::LCRunHeader* run ) ;

	virtual void processEvent(lcio::LCEvent * evt ) ; 
	
    
    protected:

     std::string _tracksColName ;    ///< the name of the input tracks collection
     std::string _hitsColName ;      ///< the name of the input hits collection
     
     /** The index of the template parameter for the hitsPerTrack histos
      *  We want to have these values divided into z bins
      */
     static const unsigned int _distanceIndex;

//     /// The index of the template parameter for the tracksPerEvent histos
//     static const unsigned int _stracksPerEventHistosIndex;

     /** For the following histograms a z binning does not make sense.
      *  They are handled in the usual way without using the ZBinTemplateProcessor.
      */
     AIDA::IHistogram1D * _distanceHisto; ///< number of hits per event

  
      double _maxDistance; ///<  Maximum value in _distanceHisto 
    };
}

#endif // HIT_DISTANCES_PROCESSOR_H
