#ifndef PAD_RESPONSE_PROCESSOR_H
#define PAD_RESPONSE_PROCESSOR_H 1

//LCIO
#include <lcio.h>

//MARLIN
#include <marlin/Processor.h>

#include <list>

namespace AIDA{
  class IHistogram1D;
  class IHistogram2D;
}


namespace marlintpc
{

    class TrackFitterFactory;
//    class TrackFitterBase;

    /** Fills residulas of hits in 2D histograms against three different values:
     *  \li Position on the pad (pad response)
     *  \li Variance of the charge distribution
     *  \li Skewness of the charge distribution
     *
     *  In case of a gaussian charge distribution a linear dependency of the residuals on
     *  the skewness is expected. The end() method  calculates a and b of the relation 
     *  residual = a * skewness + b<br>
     *  As all the output files are already closed (or not accessible) at this point 
     *  the result is dumped on the console.
     *
     *  <h4>Input - Prerequisites</h4>
     *  \li A collection of reconstructed tracks (LCIO::Track)<br>
     *  Remember that LCIO::TRBIT_HITS has to be set, so the hits are stored.
     *  \li An LCRelation collection to MCTruth tracks (optional)<br>
     *  If the relation to tracks with true parameters exists the track (helix)
     *  parameters to calculate the residuals are taken from there instead 
     *  from the reconstructed track. Also works e. g. for reference tracks
     *  from a hodoscope.
     *
     *  @param InputTrackes Name of input Tracks collection (default: TPCTracks)
     *  @param MCTruthRelationColName (optional) Name of the collection with relations
     *         to the MC truth
     *  @param ResidualMinValue Minimum residual value histogram (default: -0.3 mm)
     *  @param ResidualMaxValue Maximal residual value histogram (default: 0.3 mm)
     *  @param SkewnessMinValue Minimum skewness value histogram (default: -1 )
     *  @param SkewnessMaxValue Maximum skewness value histogram (default: 1 )
     *  @param VarianceMaxValue Maxnimum variance value histogram (default: 0.5 )
     *  @param NPadRowHistos Number of pad rows for which an individual histo
     *    should be created (default: 0 )
     *
     *  @author Martin Killenberg, CERN
     */
    class PadResponseProcessor : public marlin::Processor{
			      
    public:
    
	virtual Processor*  newProcessor() { return new PadResponseProcessor ; }
    
	PadResponseProcessor();

	~PadResponseProcessor();

	virtual void init();

	virtual void processRunHeader(lcio::LCRunHeader* run ) ;

	virtual void processEvent(lcio::LCEvent * evt ) ; 
	virtual void end( ) ; 
    
    protected:

     std::string _inputColName ;    ///< the name of the input collection
     std::string _mcTruthRelationsColName ;    ///< the name of the collection to the MC truth
     
     AIDA::IHistogram2D * _residuals_vs_position_on_pad; ///< the residuals versus the distance to pad centre (in units of pad width)
     AIDA::IHistogram2D * _residuals_vs_variance; ///< the residuals versus the variance of the charge distribution along the pad row
     AIDA::IHistogram2D * _residuals_vs_skewness; ///< the residuals versus the distance to pad centre (in units of pad width)

     std::list< std::pair< double, double  > > _skewnessResidualsList;

     float  _residualMinValue; ///< Minimum residual value in histogram
     float  _residualMaxValue; ///< Maximum residual value in histogram
     float  _skewnessMinValue; ///< Minimum skewness value in histogram 
     float  _skewnessMaxValue; ///< Maximum skewness value in histogram
     float  _varianceMaxValue; ///< Maximum variance value in histogram

     int _nPadRowHistos;
     std::vector<AIDA::IHistogram2D *> _padRowHistos;

    };
}

#endif // PAD_RESPONSE_PROCESSOR_H
