#ifndef TRACK_PARAMETERS_DISTRIBUTION_PROCESSOR_H
#define TRACK_PARAMETERS_DISTRIBUTION_PROCESSOR_H 1

//LCIO
#include <lcio.h>

//MARLIN
#include "marlin/Processor.h"

//Root
#include "TH1D.h"
#include "TH2D.h"

namespace marlintpc
{

    class TrackFitterFactory;
//    class TrackFitterBase;

    /** Fill AIDA histograms with all track parameters:
     *  phi, lambda, tan(lambda), d0, z0.
     *
     *  <h4>Input - Prerequisites</h4>
     *  \li A Collection of reconstructed tracks 
     *
     *  <h4>Output</h4>
     *  \li Histograms of the phi and lambda distributions
     *
     *  @param InputTrackes Name of input Tracks collection (default: TPCTracks)
     *  @param HistogramMinValuePhi  Minimum value in phi histogram (default: -50. mrad)
     *  @param HistogramMaxValuePhi  Maximal value in phi histogram (default: 50. mrad)
     *  @param HistogramMinValueLambda  Minimum value in lambda histogram (default: -50. mrad)
     *  @param HistogramMaxValueLambda  Maximal value in lamnda histogram (default: 50. mrad)
     *  @param HistogramMinValueTanLambda  Minimum value in tan(lambda) histogram (default: -20. )
     *  @param HistogramMaxValueTanLambda  Maximal value in tan(lamnda) histogram (default: 20. )
     *  @param HistogramMinValueD0  Minimum value in d0 histogram (default: -50. mm)
     *  @param HistogramMaxValueD0  Maximal value in d0 histogram (default: 50. mm)
     *  @param HistogramMinValueZ0  Minimum value in z0 histogram (default: -10. mm)
     *  @param HistogramMaxValueZ0  Maximal value in z0 histogram (default: 300. mm)
     *  @param HistogramMinValueOmega  Minimum value in Omega histogram (default: -0.01 mm^-1)
     *  @param HistogramMaxValueOmega  Maximal value in Omega histogram (default: 0.01. mm^-1)
     *
     *  @author Martin Killenberg, University of Bonn
     *  edited Felix Mueller, DESY
     */

    class TrackParametersDistributionProcessor : public marlin::Processor{
			      
    public:
    
	virtual Processor*  newProcessor() { return new TrackParametersDistributionProcessor ; }
    
	TrackParametersDistributionProcessor();

	~TrackParametersDistributionProcessor();

	virtual void init();

	virtual void processRunHeader(lcio::LCRunHeader* run ) ;

	virtual void processEvent(lcio::LCEvent * evt ) ; 
    
    protected:

     std::string _inputColName ;    ///< the name of the input collection
     
     TH1D* _distributionPhiHisto; ///< The histogram of the phi values
     TH1D* _distributionLambdaHisto; ///< The histogram of the lambda values
     TH1D* _distributionTanLambdaHisto; ///< The histogram of the tan(lambda) values
     TH1D* _distributionD0Histo; ///< The histogram of the d0 values
     TH1D* _distributionZ0Histo; ///< The histogram of the z0 values
     TH1D* _distributionOmegaHisto; ///< The histogram of the omega values
     TH1D* _distributionMomentumHisto; ///< The histogram of the momentum values
     TH1D* _distributionChi2Histo; ///< The histogram of the Chi^2 values
     TH1D* _distributionNdfHisto; ///< The histogram of the number of degrees of freedom values
     TH1D* _distributionDEdxHisto; ///< The histogram of the dEdx values
     TH1D* _distributionNHitsHisto; ///< The histogram of the number of hits values
     TH2D* _distributionDEdxVsOmegaHisto; ///< The histogram of the dEdx values vs omega values

     float  _bFieldStrength; ///< strength of the B field in T
     float  _histogramMinValuePhi; ///< Minimum value in phi histogram
     float  _histogramMaxValuePhi; ///< Maximum value in phi histogram
     float  _histogramMinValueLambda; ///< Minimum value in lambda histogram
     float  _histogramMaxValueLambda; ///< Maximum value in lambda histogram
     float  _histogramMinValueTanLambda; ///< Minimum value in lambda histogram
     float  _histogramMaxValueTanLambda; ///< Maximum value in lambda histogram
     float  _histogramMinValueD0; ///< Minimum value in lambda histogram
     float  _histogramMaxValueD0; ///< Maximum value in lambda histogram
     float  _histogramMinValueZ0; ///< Minimum value in lambda histogram
     float  _histogramMaxValueZ0; ///< Maximum value in lambda histogram
     float  _histogramMinValueOmega; ///< Minimum value in Omega histogram
     float  _histogramMaxValueOmega; ///< Maximum value in Omega histogram
     float  _histogramMinValueMomentum; ///< Minimum value in momentum histogram
     float  _histogramMaxValueMomentum; ///< Maximum value in momentum histogram
     float  _histogramMinValueChi2; ///< Minimum value in Chi^2 histogram
     float  _histogramMaxValueChi2; ///< Maximum value in Chi^2 histogram
     float  _histogramMinValueNdf; ///< Minimum value in number of degrees of freedom histogram
     float  _histogramMaxValueNdf; ///< Maximum value in number of degrees of freedom histogram
     float  _histogramMinValueDEdx; ///< Minimum value in dEdx histogram
     float  _histogramMaxValueDEdx; ///< Maximum value in dEdx histogram
     float  _histogramMinValueNHits; ///< Minimum value in number of hits histogram
     float  _histogramMaxValueNHits; ///< Maximum value in number of hits histogram
    };
}

#endif // PHI_LAMBDA_DISTRIBUTION_PROCESSOR_H
