#ifndef DEDXPROCESSOR_H
#define DEDXPROCESSOR_H

// C++
#include <string>
#include <fstream>
#include <set>


// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "lccd/ConditionsMap.hh"
//ROOT
#include "TH1F.h"
#include "TFile.h"
#include "TNtuple.h"

#include "gear/GEAR.h"
#include "HistogramWidthExtractionTools.h"

namespace marlintpc
{

/**
 *  A processor to calculate hitcharge per length unit 
 *  taking into account the track length over the measurement row of the hit
 *  to evalutae the dEdx performance
 * 
 *  <h4>Input - Prerequisites</h4>
 *  a collection of Tracks
 *
 * Aida is used for the output of histograms
 * @param InputCollectionName Name of the input collection (default: ClupatraTracks)
 * @param SkipRow: A pair of integers, separated by spaces identifying the module and row to select.
 * A negative number indicates any row and/or any module. For example -1 -1 signals to remove all rows from all modules
 * The user can use this parameter an arbitrary number of times in each processor. Default: -1 -1
*
 * @author Astrid Muennich, DESY
 */

class dEdxProcessor : public marlin::Processor
{

    public:

        /* This method will be called by the marlin package
         * It returns a processor of the currend type 
         */
        virtual Processor*  newProcessor()
        {
            return new dEdxProcessor ;
        }

        /* the default constructor
         * here the processor parameters are registered to the marlin package
         * other initialisation should be placed in the init method
         */
        dEdxProcessor() ;

        /* Called at the beginning of the job before anything is read.
         * Use to initialize the processor, e.g. book histograms
         */
        virtual void init() ;

        /* Called for every run.
         * in this processor it is not used
         */
        virtual void processRunHeader(lcio::LCRunHeader* run) ;

        /* Called for every event - the working horse.
         * Here the real conversion take place
         */
        virtual void processEvent(lcio::LCEvent * evt) ;

        /* This method is only called if the check flag is set (default) in the main processor
         * (where the input files are given)
         * It should be used to create check plots which are not needed in a complete analysis
         * (where normaly the check flag is deactivated)
         */
        virtual void check(lcio::LCEvent * evt) ;


        /* Called after data processing for clean up.
         * e.g. saving control histogramms, delete created objects
         * (which are not stored in a collection/event), etc.
         */
        virtual void end() ;




    protected:

        std::string _inputColName ;  ///<- the name of the input collection
   
        int _nRun ; ///<- counts the processed runs
	gear::IntVec _rows2skip;///< Vector of rows to skip
	std::set<std::pair<int,int> > _skipList; ///< The list of the rows to skip
	static const std::pair<int, int> ALLROWS; ///< The pair that identifies the default value of the SkipRow parameter and the special combination meaning "all rows and modules"


    private:

	TNtuple *_tuple;
	TH1F* _histodEdx;
	HistogramWidthExtractionTools *_myTool;
} ;

}

#endif



