#include <vector>
#include <iostream>
#include "TLatex.h"
#include "TCanvas.h"
#include "TH1.h"
#include "TColor.h"
#include "TFile.h"
#include "TNtuple.h"
#include "TStyle.h"
#include "TROOT.h"
#include "THStack.h"
#include "sstream"
#include <utility>
#include <cstdarg>

/*How to call this:
.L $MARLINTPC/analysis/rootscripts/AnalysisCutSelector.C++
AnalysisCutSelector("path for output","path to the aida file", number of runs in the following list, list of the runs)
e.g AnalysisCutSelector("../myplots","../AidaFiles/",3,18850,18851,18860)
 */
void AnalysisCutSelector(char * pathToOutput, char * pathToAida, int nRuns, ...){

	//loading the style
	gROOT->ProcessLine(".L ${MARLINTPC}/analysis/rootscripts/LCTPCStyle.C");
	gROOT->ProcessLine("LCTPCStyle()");

	//open a root file to store the graphs in
	TFile *outfile=new TFile(Form("%s%s",pathToOutput,"AnalysisCutSelector.root"),"RECREATE");

	//initialize the input list
	va_list argList;
	va_start(argList,nRuns);

	int runNumber;
	// will contain the results from the processors
	std::vector<const char*> eventCutNames;
	eventCutNames.push_back("all events");
	eventCutNames.push_back("no. of tracks");

	std::vector<const char*> trackCutNames;
	trackCutNames.push_back("all tracks");
	trackCutNames.push_back("# hits");
	trackCutNames.push_back("#varphi");
	trackCutNames.push_back("tan#lambda");
	trackCutNames.push_back("d_{0}");
	trackCutNames.push_back("z_{0}");
	trackCutNames.push_back("#Omega");
	trackCutNames.push_back("#frac{dE}{dx}");
	trackCutNames.push_back("fit worked");
	trackCutNames.push_back("Chi^2");


	THStack * eventStack=new THStack();
	THStack * trackStack=new THStack();

	//create colors
	const int BINNING=nRuns;
	Int_t colors[BINNING];

	const Int_t NRGBs = 5;

	Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
	Double_t red[NRGBs]   = { 0.00, 0.00, 0.87, 1.00, 0.51 };
	Double_t green[NRGBs] = { 0.00, 0.81, 1.00, 0.20, 0.00 };
	Double_t blue[NRGBs]  = { 0.51, 1.00, 0.12, 0.00, 0.00 };
	Int_t FI = TColor::CreateGradientColorTable(NRGBs, stops, red, green, blue, BINNING);

	for (int i=0;i<BINNING;i++)
		colors[i] = FI+i;

	TH1::AddDirectory(kFALSE);
	//loop over the given runs
	for (int i=0;i<nRuns;i++)
	{
		//get run number from input
		runNumber=va_arg(argList,int);

		//create stringstream containing the aida file name
		std::stringstream runNumberString;
		std::cout<<"run number: "<<runNumber<<std::endl;
		runNumberString <<pathToAida<<"aida_run" <<runNumber<< ".root";

		//open the aida file
		TFile *file  = TFile::Open(runNumberString.str().c_str(),"Read");
		//check that the file exists
		if (file==0)
		{
			std::cout<<"This run will be skipped. The aida file does not exist. Please check this!"<<std::endl;
			continue;
		}

		TNtuple * zPosTuple=(TNtuple*)file->Get("MyAnalysisCutSelectorProcessor/RunZPositions");

		if (zPosTuple==0 )
		{
			std::cout<<"This run will be skipped. One of the necessary root tuples does not exist. Please check that you ran the corresponding processors and that the tuple is in the aida file!"<<std::endl;
			continue;
		}

		//get the graphs from the file
		TH1D *eventCutFlow=(TH1D*)file->Get("MyAnalysisCutSelectorProcessor/cutFlowEvents");
		TH1D *trackCutFlow=(TH1D*)file->Get("MyAnalysisCutSelectorProcessor/cutFlowTracks");
		//check that the tuples exists
		if(eventCutFlow==0 || trackCutFlow==0 )
		{
			std::cout<<"This run will be skipped. One of the necessary root graphs does not exist. Please check that you ran the corresponding processors and that the graph is in the aida file!"<<std::endl;
			continue;
		}

		//fill the vectors with the values from the tuples ([0]=mean,[2]=rms)
		//567.6-hitposition is the conversion into drift distance
		zPosTuple->GetEntry(0);
		int zpos= int(567.6-zPosTuple->GetArgs()[0]) ;
		std::ostringstream legEntry;
		legEntry << "z_{d} = "<<zpos;
		//assign color
		eventCutFlow->SetTitle(legEntry.str().c_str());
		trackCutFlow->SetTitle(legEntry.str().c_str());
		eventCutFlow->SetLineColor(colors[i]);
		eventStack->Add(eventCutFlow);
		trackCutFlow->SetLineColor(colors[i]);
		trackStack->Add(trackCutFlow);

		file->Close();
		delete file;
	}

	va_end(argList);

	//save the results
	outfile->cd();
	gStyle->SetPadRightMargin(0.18);

	TCanvas *c1 = new TCanvas("eventCutFlow","eventCutFlow");
	eventStack->Draw("nostack");
	eventStack->GetXaxis()->LabelsOption("v");
	for(int i=0;i<eventStack->GetXaxis()->GetNbins();i++)
	{
		eventStack->GetXaxis()->SetBinLabel(i+1,eventCutNames[i]);
	}
	gPad->BuildLegend(0.85,0.2,0.99,0.99,"[mm]");
	eventStack->GetYaxis()->SetTitle("# events");
	c1->Modified();
	c1->SaveAs(Form("%s%s",pathToOutput,"EventCutFlow.png"));
	c1->SaveAs(Form("%s%s",pathToOutput,"EventCutFlow.pdf"));
	c1->Write();

	TCanvas *c2 = new TCanvas("trackCutFlow","trackCutFlow");
	trackStack->Draw("nostack");
	trackStack->GetXaxis()->LabelsOption("v");
	for(int i=0;i<trackStack->GetXaxis()->GetNbins();i++)
	{
		trackStack->GetXaxis()->SetBinLabel(i+1,trackCutNames[i]);
	}
	gPad->BuildLegend(0.85,0.2,0.99,0.99,"[mm]");
	trackStack->GetYaxis()->SetTitle("# tracks");
	c2->Modified();
	c2->SaveAs(Form("%s%s",pathToOutput,"TrackCutFlow.png"));
	c2->SaveAs(Form("%s%s",pathToOutput,"TrackCutFlow.pdf"));
	c2->Write();
	  
	outfile->Write();
	outfile->Close();
	gStyle->SetPadRightMargin(0.04);
	delete c1;
	delete c2;

}
