#include <vector>
#include <iostream>
#include "TLatex.h"
#include "TCanvas.h"
#include "TH1.h"
#include "TGraphErrors.h"
#include "TFile.h"
#include "TNtuple.h"
#include "TStyle.h"
#include "TROOT.h"
#include "sstream"
#include <cstdarg>

/*How to call this:
.L $MarlinTPC/analysis/rootscripts/HitQuality.C++
HitQuality("path for output","path to the aida file", number of runs in the following list, list of the runs)
e.g HitQuality("../myplots","../AidaFiles/",3,18850,18851,18860)
*/
void HitQuality(char * pathToOutput, char * pathToAida, int nRuns, ...){

//loading the style
gROOT->ProcessLine(".L ${MARLINTPC}/analysis/rootscripts/LCTPCStyle.C");
gROOT->ProcessLine("LCTPCStyle()");

//open a root file to store the graphs in
 TFile *outfile=new TFile(Form("%s%s",pathToOutput,"HitQuality.root"),"RECREATE");

//initialize the input list
 va_list argList;
 va_start(argList,nRuns);


int runNumber;

// will contain the results from the processors
std::vector<double> runZPositions;
std::vector<double> runZPositionsError;
std::vector<double> runHitOverRange;
std::vector<double> runHitOverRangeError;
std::vector<double> runHitDeadChannel;
std::vector<double> runHitDeadChannelError;

//loop over the given runs
for (int i=0;i<nRuns;i++)
{
	//get run number from input
	runNumber=va_arg(argList,int);

	//create stringstream containing the aida file name
	std::stringstream runNumberString;
	std::cout<<"run number: "<<runNumber<<std::endl;
	runNumberString <<pathToAida<<"aida_run" <<runNumber<< ".root";

	//open the aida file
	TFile *file  = TFile::Open(runNumberString.str().c_str(),"Read");
	//check that the file exists
	if (file==0)
	{
		std::cout<<"This run will be skipped. The aida file does not exist. Please check this!"<<std::endl;
		continue;
	}

	//get the tuples from the file
	TNtuple * zPosTuple=(TNtuple*)file->Get("MyAnalysisCutSelectorProcessor/RunZPositions");
	TNtuple * overRangeTuple=(TNtuple*)file->Get("MyHitQualityProcessor/OverRangeHits");
	TNtuple * deadChannelTuple=(TNtuple*)file->Get("MyHitQualityProcessor/DeadChannelsHits");
	//check that the tuples exists
	if ( overRangeTuple==0||deadChannelTuple==0||zPosTuple==0 )
	{
		std::cout<<"This run will be skipped. One of the necessary root tuples does not exist. Please check that you ran the corresponding processors and that the tuple is in the aida file!"<<std::endl;
		continue;

	}

	//open the first and only entry of the tuples
	zPosTuple->GetEntry(0);
	overRangeTuple->GetEntry(0);
	deadChannelTuple->GetEntry(0);

	//fill the vectors with the values from the tuples ([0]=mean,[2]=rms)
	//567.6-hitposition is the conversion into drift distance
	runZPositions.push_back( 567.6-zPosTuple->GetArgs()[0] );
	runZPositionsError.push_back( zPosTuple->GetArgs()[2] );
	runHitOverRange.push_back( overRangeTuple->GetArgs()[0] );
	runHitOverRangeError.push_back( overRangeTuple->GetArgs()[2] );
	runHitDeadChannel.push_back( deadChannelTuple->GetArgs()[0] );
	runHitDeadChannelError.push_back( deadChannelTuple->GetArgs()[2] );

	file->Close();
	delete file;
}

va_end(argList);

//create Graphs and fill them with the data
TGraphErrors *hitOverRangeGraph = new TGraphErrors( runZPositions.size(), &runZPositions[0], &runHitOverRange[0], &runZPositionsError[0], &runHitOverRangeError[0] );
hitOverRangeGraph->SetTitle("");
hitOverRangeGraph->GetXaxis()->SetTitle("drift distance [mm]");
hitOverRangeGraph->GetYaxis()->SetTitle("hits in overrange [%]");
TGraphErrors *hitDeadChannelGraph = new TGraphErrors( runZPositions.size(), &runZPositions[0], &runHitDeadChannel[0], &runZPositionsError[0], &runHitDeadChannelError[0] );
hitDeadChannelGraph->SetTitle("");
hitDeadChannelGraph->GetXaxis()->SetTitle("drift distance [mm]");
hitDeadChannelGraph->GetYaxis()->SetTitle("hits with dead channels [%]");

//print the results
outfile->cd();
TCanvas *c1 = new TCanvas("deadChannels","dead channels");
hitDeadChannelGraph->Draw("AP");
c1->SaveAs(Form("%s%s",pathToOutput,"DeadChannelsVsZ.png"));
c1->SaveAs(Form("%s%s",pathToOutput,"DeadChannelsVsZ.pdf"));
hitDeadChannelGraph->SetName("hitDeadChannel");
hitDeadChannelGraph->Write();

TCanvas *c2 = new TCanvas("overRange","over range hits");
hitOverRangeGraph->Draw("AP");
c2->SaveAs(Form("%s%s",pathToOutput,"OverRangeHitsVsZ.png"));
c2->SaveAs(Form("%s%s",pathToOutput,"OverRangeHitsVsZ.pdf"));
hitOverRangeGraph->SetName("hitOverRange");
hitOverRangeGraph->Write();

outfile->Write();
outfile->Close();

delete c1;
delete c2;

}
