#include <vector>
#include <iostream>
#include "TLatex.h"
#include "TCanvas.h"
#include "TH1.h"
#include "TF1.h"
#include "TGraphErrors.h"
#include "TMultiGraph.h"
#include "TFile.h"
#include "TNtuple.h"
#include "TStyle.h"
#include "TROOT.h"
#include "TFrame.h"
#include "sstream"
#include <utility>
#include <cstdarg>
#include "TColor.h"
#include "TMath.h"

/*How to call this:
.L $MARLINTPC/analysis/rootscripts/PhiResolution.C++
PhiResolution("path for output","path to the aida file", number of runs in the following list, list of the runs)
e.g PhiResolution("../myplots","../AidaFiles/",3,18850,18851,18860)
 */
void PhiResolution(char * pathToOutput, char * pathToAida, int nRuns, ...){

	//loading the style
	gROOT->ProcessLine(".L ${MARLINTPC}/analysis/rootscripts/LCTPCStyle.C");
	gROOT->ProcessLine("LCTPCStyle()");

	//open a root file to store the graphs in
	TFile *outfile=new TFile(Form("%s%s",pathToOutput,"PhiResolution.root"),"RECREATE");

	//initialize the input list
	va_list argList;
	va_start(argList,nRuns);

	int runNumber;
	int nRows=0;
	std::vector<double> runSingleZResMean;
	std::vector<double> runSingleZResMeanError;
	std::vector<double> runSingleXYResMean;
	std::vector<double> runSingleXYResMeanError;

	std::vector<double> phis;
	std::vector<double> phisErr;


	//create colors
	const int BINNING=nRuns;
	Int_t colors[BINNING];

	const Int_t NRGBs = 5;

	Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
	Double_t red[NRGBs]   = { 0.00, 0.00, 0.87, 1.00, 0.51 };
	Double_t green[NRGBs] = { 0.00, 0.81, 1.00, 0.20, 0.00 };
	Double_t blue[NRGBs]  = { 0.51, 1.00, 0.12, 0.00, 0.00 };
	Int_t FI = TColor::CreateGradientColorTable(NRGBs, stops, red, green, blue, BINNING);

	for (int i=0;i<BINNING;i++)
		colors[i] = FI+i;

	//loop over the given runs
	for (int i=0;i<nRuns;i++)
	{
		//get run number from input
		runNumber=va_arg(argList,int);

		//create stringstream containing the aida file name
		std::stringstream runNumberString;
		std::cout<<"run number: "<<runNumber<<std::endl;
		runNumberString <<pathToAida<<"aida_run" <<runNumber<< ".root";

		//open the aida file
		TFile *file  = TFile::Open(runNumberString.str().c_str(),"Read");

		if (file==0)
		{
			std::cout<<"one input file does not exist. exit"<<std::endl;
			// 			std::exit(1);
			continue;
		}

		TNtuple * singleRowTuple=(TNtuple*)file->Get("MyResolutionPerformanceProcessor/singleRowsResolution");

		if (singleRowTuple==0)
		{
			std::cout<<"This run will be skipped. One of the necessary root tuples does not exist. Please check that you ran the corresponding processors and that the tuple is in the aida file!"<< runNumberString.str().c_str()<<std::endl;

			// 			std::exit(1);
			continue;
		}
		nRows = singleRowTuple->GetEntries();
		std::vector<double> xy,z,wxy,wz;
		double sumwxy=0;
		double sumwz=0;
		for (int RowIter=0;RowIter<nRows;RowIter++)
		{

			singleRowTuple->GetEntry(RowIter);
			if(singleRowTuple->GetArgs()[0]==5 && singleRowTuple->GetArgs()[1]==10)
			{
				wxy.push_back(1/(singleRowTuple->GetArgs()[3]*singleRowTuple->GetArgs()[3]));
				wz.push_back(1/(singleRowTuple->GetArgs()[5]*singleRowTuple->GetArgs()[5]));
				xy.push_back(singleRowTuple->GetArgs()[2]);
				z.push_back(singleRowTuple->GetArgs()[4]);

				sumwxy+=wxy[RowIter];
				sumwz+=wz[RowIter];
			}
		}

		TNtuple * hitPhiBestRowTuple=(TNtuple*)file->Get("MyAnalysisCutSelectorProcessor/HitPhiBestRow");
		//check that the tuples exists
		if ( hitPhiBestRowTuple==0 )
		{
			std::cout<<"This run will be skipped. One of the necessary root tuples does not exist. Please check that you ran the corresponding processors and that the tuple is in the aida file!"<<std::endl;
			continue;

		}

		//open the first and only entry of the tuples
		hitPhiBestRowTuple->GetEntry(0);


		runSingleXYResMean.push_back(TMath::Mean(xy.size(),&xy[0],&wxy[0]));
		runSingleZResMean.push_back(TMath::Mean(xy.size(),&z[0],&wz[0]));

		runSingleXYResMeanError.push_back(sqrt(0.00001));
		runSingleZResMeanError.push_back(sqrt(0.00001));

		//TH1* phi = (TH1*)file->Get("MyAnalysisCutSelectorProcessor/hitPhiBestRow");
		//phis.push_back(phi->GetFunction("fit_range")->GetParameter(1)*180/M_PI);
		//phisErr.push_back(phi->GetFunction("fit_range")->GetParameter(2)*180/M_PI);
		phis.push_back(hitPhiBestRowTuple->GetArgs()[2]*180/M_PI);
		phisErr.push_back(hitPhiBestRowTuple->GetArgs()[4]*180/M_PI);

		file->Close();
		delete file;

	}

	TGraphErrors  *xyResolution=new TGraphErrors(phis.size(),&phis[0], &runSingleXYResMean[0],&phisErr[0], &runSingleXYResMeanError[0]);
	TGraphErrors  *zResolution=new TGraphErrors(phis.size(),&phis[0], &runSingleZResMean[0],&phisErr[0], &runSingleZResMeanError[0]);

	TF1* tan2 = new TF1("tan2","sqrt([0]*[0]+5.85*5.85/(12*[1]/cos((x-[2])*TMath::Pi()/180))*pow(tan((x-[2])*TMath::Pi()/180),2))");
	TF1* sqrtcos = new TF1("sqrtcos","sqrt([0]*[0]+[1]*[1]*[2]/10/[3]*cos((x-[4])*TMath::Pi()/180))");
	sqrtcos->SetLineColor(kRed);
	sqrtcos->SetParameters(0.07,0.104,100,30,0.);
	sqrtcos->FixParameter(4,0.);
	tan2->SetParameters(0.3,6,0);
	tan2->FixParameter(2,0);
	tan2->SetLineColor(kRed);

	xyResolution->GetYaxis()->SetTitle("#sigma_{r#varphi} [mm]");
	xyResolution->GetXaxis()->SetTitle("#alpha [#circ]");
	TCanvas *c1 = new TCanvas("PhiXYResolution","PhiXYResolution");
	xyResolution->SetMarkerStyle(22);
	xyResolution->SetFillColor(0);
	xyResolution->Draw("AP");
	c1->SaveAs(Form("%s%s",pathToOutput,"PhiXYResolution.png"));
	c1->SaveAs(Form("%s%s",pathToOutput,"PhiXYResolution.pdf"));

	zResolution->GetYaxis()->SetTitle("#sigma_{z} [mm]");
	zResolution->GetXaxis()->SetTitle("#alpha [#circ]");
	TCanvas *c2 = new TCanvas("PhiZResolution","PhiZResolution");
	zResolution->SetMarkerStyle(22);
	zResolution->SetFillColor(0);
	zResolution->Draw("AP");
	c2->SaveAs(Form("%s%s",pathToOutput,"PhiZResolution.png"));
	c2->SaveAs(Form("%s%s",pathToOutput,"PhiZResolution.pdf"));
}
