#include <vector>
#include <iostream>
#include <fstream>
#include "sstream"
#include <utility>
#include <cstdarg>
#include <cmath>

#include "TLatex.h"
#include "TCanvas.h"
#include "TAxis.h"
#include "THStack.h"
#include "TGraphErrors.h"
#include "TMultiGraph.h"
#include "TFile.h"
#include "TNtuple.h"
#include "TStyle.h"
#include "TROOT.h"
#include "TFrame.h"
#include "TColor.h"
#include "TMath.h"

/*How to call this:
.L $MARLINTPC/analysis/rootscripts/PulsePerHit.C++
PulsePerHit("path for output","path to the aida file", number of runs in the
following list, list of the runs)
e.g PulsePerHit("../myplots","../AidaFiles/",3,18850,18851,18860)
 */
void PulsePerHit(char* pathToOutput, char* pathToAida, int nRuns, ...)
{

    // loading the style
    gROOT->ProcessLine(".L ${MARLINTPC}/analysis/rootscripts/LCTPCStyle.C");
    gROOT->ProcessLine("LCTPCStyle()");

    // open a root file to store the graphs in
    TFile* outfile
        = new TFile(Form("%s%s", pathToOutput, "PulsePerHit.root"), "RECREATE");

    // initialize the input list
    va_list argList;
    va_start(argList, nRuns);

    int runNumber;
    const int MAXPULSES = 10;
    const int MINPULSES = 2;
    // will contain the results from the processors
    TMultiGraph* runPulsesVSRow = new TMultiGraph();
    runPulsesVSRow->SetTitle("Pulses per Hit;r_{row} [mm];pulses per hit");
    THStack* pulsesPerHitDist = new THStack();
    pulsesPerHitDist->SetTitle("Pulses per Hit;# pulses per hit;# entries");
    TMultiGraph* pulseFractionsVSDrift = new TMultiGraph();
    TMultiGraph* pulseFractionsVSDrift4Plus = new TMultiGraph();

    std::vector<double> pulseFractions[MAXPULSES];
    std::vector<double> pulseFractions4Plus;

    std::vector<double> runZPositions;
    std::vector<double> runZPositionsError;
    std::vector<double> runZPositionsSingleRow;
    std::vector<double> runZPositionsSingleRowError;

    std::vector<double> runPulsePerHitMean;
    std::vector<double> runPulsePerHitMeanErr;
    std::vector<double> runPulsePerHitRMS;
    std::vector<double> runPulsePerHitRMSErr;
    std::vector<double> runPulsePerHitRowMean;
    std::vector<double> runPulsePerHitRowMeanErr;
    std::vector<double> runPulsePerHitRowRMS;
    std::vector<double> runPulsePerHitRowRMSErr;

    // create colors
    const int BINNING = nRuns;
    Int_t* colors = new Int_t[BINNING];
    Int_t FractionColors[MAXPULSES];

    const Int_t NRGBs = 5;

    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs] = { 0.00, 0.00, 0.87, 1.00, 0.51 };
    Double_t green[NRGBs] = { 0.00, 0.81, 1.00, 0.20, 0.00 };
    Double_t blue[NRGBs] = { 0.51, 1.00, 0.12, 0.00, 0.00 };
    Int_t FI = TColor::CreateGradientColorTable(NRGBs, stops, red, green, blue,
                                                BINNING);

    for(int i = 0; i < BINNING; i++)
        colors[i] = FI + i;

    Int_t fc = TColor::CreateGradientColorTable(NRGBs, stops, red, green, blue,
                                                MAXPULSES - MINPULSES + 1);

    for(int i = MINPULSES; i <= MAXPULSES; i++)
        FractionColors[i - 1] = fc + i - MINPULSES;

    // loop over the given runs
    for(int i = 0; i < nRuns; i++)
    {
        // get run number from input
        runNumber = va_arg(argList, int);

        // create stringstream containing the aida file name
        std::stringstream runNumberString;
        std::cout << "run number: " << runNumber << std::endl;
        runNumberString << pathToAida << "aida_run" << runNumber << ".root";

        // open the aida file
        TFile* file = TFile::Open(runNumberString.str().c_str(), "Read");
        // check that the file exists
        if(file == 0)
        {
            std::cout << "This run will be skipped. The aida file does not "
                         "exist. Please check this!" << std::endl;
            continue;
        }

        // cd out of the file to prevent root from automatically deleting the
        // following, when closing the file
        gROOT->cd();

        // get the stuff from the file, check that everything exist
        TNtuple* zPosTuple = (TNtuple*)file->Get(
            "MyAnalysisCutSelectorProcessor/RunZPositions");
        TNtuple* pulsesPerHitTuple
            = (TNtuple*)file->Get("MyPulsePerHitProcessor/PulsesPerHitStats");
        TNtuple* pulsesPerHitRowTuple = (TNtuple*)file->Get(
            "MyPulsePerHitProcessor/PulsesPerHitRowStats");

        if(zPosTuple == 0 || pulsesPerHitTuple == 0
           || pulsesPerHitRowTuple == 0)
        {
            std::cout << "This run will be skipped. One of the necessary root "
                         "tuples does not exist. Please check that you ran the "
                         "corresponding processors and that the tuple is in "
                         "the aida file!" << std::endl;
            continue;
        }

        if(file->Get("MyPulsePerHitProcessor/PulsesPerHitVSRowGraph") == 0)
        {
            std::cout << "This run will be skipped. "
                         "MyPulsePerHitProcessor/PulsesPerHitVSRowGraph does "
                         "not exist. Please check that you ran the "
                         "corresponding processors and that the graph is in "
                         "the aida file!" << std::endl;
            continue;
        }
        TGraphErrors* pulsesPerHitGraph
            = (TGraphErrors*)
                  file->Get("MyPulsePerHitProcessor/PulsesPerHitVSRowGraph")
                      ->Clone();

        if(file->Get("MyPulsePerHitProcessor/PulsesPerHitHisto") == 0)
        {
            std::cout << "This run will be skipped. "
                         "MyPulsePerHitProcessor/PulsesPerHitHisto does not "
                         "exist. Please check that you ran the corresponding "
                         "processors and that the graph is in the aida file!"
                      << std::endl;
            continue;
        }
        TH1D* pulsesPerHitHisto
            = (TH1D*)file->Get("MyPulsePerHitProcessor/PulsesPerHitHisto")
                  ->Clone();

        // open the first entry of the tuples
        zPosTuple->GetEntry(0);
        pulsesPerHitTuple->GetEntry(0);

        // fill the vectors with the values from the tuples ([0]=mean,[2]=rms)
        runZPositions.push_back(567.6 - zPosTuple->GetArgs()[0]);
        runZPositionsError.push_back(zPosTuple->GetArgs()[2]);

        runPulsePerHitMean.push_back(pulsesPerHitTuple->GetArgs()[0]);
        runPulsePerHitMeanErr.push_back(pulsesPerHitTuple->GetArgs()[1]);
        runPulsePerHitRMS.push_back(pulsesPerHitTuple->GetArgs()[2]);
        runPulsePerHitRMSErr.push_back(pulsesPerHitTuple->GetArgs()[3]);

        int nRows = pulsesPerHitRowTuple->GetEntries();
        for(int RowIter = 0; RowIter < nRows; RowIter++)
        {
            runZPositionsSingleRow.push_back(567.6 - zPosTuple->GetArgs()[0]);
            runZPositionsSingleRowError.push_back(zPosTuple->GetArgs()[2]);
            pulsesPerHitRowTuple->GetEntry(RowIter);
            runPulsePerHitRowMean.push_back(pulsesPerHitRowTuple->GetArgs()[2]);
            runPulsePerHitRowMeanErr.push_back(
                pulsesPerHitRowTuple->GetArgs()[3]);
            runPulsePerHitRowRMS.push_back(pulsesPerHitRowTuple->GetArgs()[4]);
            runPulsePerHitRowRMSErr.push_back(
                pulsesPerHitRowTuple->GetArgs()[5]);
        }

        // 567.6-hitposition is the conversion into drift distance
        int zpos = int(567.6 - zPosTuple->GetArgs()[0]);
        std::ostringstream legEntry;
        legEntry << "z_{d} = " << zpos;

        std::ostringstream legEntryHist;
        legEntryHist << "z_{d} = " << zpos;

        // assign color and add to MultiGraph / THStack
        pulsesPerHitGraph->SetLineColor(colors[i]);
        pulsesPerHitGraph->SetMarkerColor(colors[i]);
        pulsesPerHitGraph->SetTitle(legEntry.str().c_str());

        runPulsesVSRow->Add(pulsesPerHitGraph, "P");

        pulsesPerHitHisto->SetLineColor(colors[i]);
        pulsesPerHitHisto->SetMarkerColor(colors[i]);
        pulsesPerHitHisto->SetTitle(legEntryHist.str().c_str());
        // pulsesPerHitHisto->GetFunction("fit")->SetLineColor(colors[i]);
        // pulsesPerHitHisto->GetFunction("fit")->SetNpx(1000);
        if(pulsesPerHitHisto->GetFunction("fit"))
            pulsesPerHitHisto->GetFunction("fit")->Delete();

        pulsesPerHitDist->Add(pulsesPerHitHisto);

        double frac4Plus = 0;
        for(int nPulses = MINPULSES; nPulses <= pulsesPerHitHisto->GetNbinsX();
            ++nPulses)
        {
            double fraction = pulsesPerHitHisto->GetBinContent(nPulses + 1)
                              / pulsesPerHitHisto->GetEntries();
            if(nPulses <= MAXPULSES)
                pulseFractions[nPulses - 1].push_back(fraction);
            if(nPulses >= 4) frac4Plus += fraction;
        }
        pulseFractions4Plus.push_back(frac4Plus);

        delete zPosTuple;
        delete pulsesPerHitTuple;

        file->Close();
        delete file;
    }

    va_end(argList);

    // make pulse count fraction plots and add to Multigraph
    for(int nPulses = MINPULSES; nPulses <= MAXPULSES; ++nPulses)
    {
        TGraph* pulseFractionGraph
            = new TGraph(runZPositions.size(), &runZPositions[0],
                         &pulseFractions[nPulses - 1][0]);
        pulseFractionGraph->SetNameTitle(
            Form("pulseFractionGraphGraph%d", nPulses),
            Form("%d pads", nPulses));
        pulseFractionGraph->GetXaxis()->SetTitle("drift distance [mm]");
        pulseFractionGraph->GetYaxis()->SetTitle("fraction of hits");
        pulseFractionGraph->SetLineColor(FractionColors[nPulses - 1]);
        pulseFractionGraph->SetMarkerColor(FractionColors[nPulses - 1]);
        pulseFractionGraph->SetFillColor(kWhite);
        pulseFractionGraph->SetMarkerStyle(20 + nPulses - MINPULSES);
        pulseFractionGraph->SetLineStyle(nPulses - MINPULSES + 1);
        pulseFractionsVSDrift->Add(pulseFractionGraph, "PL");
        if(nPulses < 4)
            pulseFractionsVSDrift4Plus->Add(pulseFractionGraph, "PL");
    }
    TGraph* pulseFractionGraph4Plus = new TGraph(
        runZPositions.size(), &runZPositions[0], &pulseFractions4Plus[0]);
    pulseFractionGraph4Plus->SetNameTitle("pulseFractionGraph4Plus", "4+ pads");
    pulseFractionGraph4Plus->GetXaxis()->SetTitle("drift distance [mm]");
    pulseFractionGraph4Plus->GetYaxis()->SetTitle("fraction of hits");
    pulseFractionGraph4Plus->SetLineColor(FractionColors[3]);
    pulseFractionGraph4Plus->SetMarkerColor(FractionColors[3]);
    pulseFractionGraph4Plus->SetFillColor(kWhite);
    pulseFractionGraph4Plus->SetMarkerStyle(24 - MINPULSES);
    pulseFractionGraph4Plus->SetLineStyle(4 - MINPULSES + 1);
    pulseFractionsVSDrift4Plus->Add(pulseFractionGraph4Plus, "PL");

    // save the results
    outfile->cd();

    TCanvas* c2
        = new TCanvas("Pulses per Hit all Rows", "Pulses per Hit all Rows");
    TGraphErrors* runPulsePerHitGraph = new TGraphErrors(
        runZPositions.size(), &runZPositions[0], &runPulsePerHitMean[0],
        &runZPositionsError[0], &runPulsePerHitMeanErr[0]);
    runPulsePerHitGraph->SetNameTitle("PulsePerHitGraph", "Pulses per Hit");
    runPulsePerHitGraph->GetXaxis()->SetTitle("drift distance [mm]");
    runPulsePerHitGraph->GetYaxis()->SetTitle("mean #pulses per hit");
    runPulsePerHitGraph->Draw("AP");
    gPad->Update();
    c2->SaveAs(Form("%s%s", pathToOutput, "PulsePerHit.png"));
    c2->SaveAs(Form("%s%s", pathToOutput, "PulsePerHit.pdf"));
    runPulsePerHitGraph->Write();

    TCanvas* c3 = new TCanvas("Pulses per Hit RMS all Rows",
                              "Pulses per Hit RMS all Rows");
    TGraphErrors* runPulsePerHitRMSGraph = new TGraphErrors(
        runZPositions.size(), &runZPositions[0], &runPulsePerHitRMS[0],
        &runZPositionsError[0], &runPulsePerHitRMSErr[0]);
    runPulsePerHitRMSGraph->SetNameTitle("PulsePerHitRMSGraph",
                                         "Pulses per Hit RMS");
    runPulsePerHitRMSGraph->GetXaxis()->SetTitle("drift distance [mm]");
    runPulsePerHitRMSGraph->GetYaxis()->SetTitle("sigma #pulses per hit");
    runPulsePerHitRMSGraph->Draw("AP");
    gPad->Update();
    c3->SaveAs(Form("%s%s", pathToOutput, "PulsePerHitRMS.png"));
    c3->SaveAs(Form("%s%s", pathToOutput, "PulsePerHitRMS.pdf"));
    runPulsePerHitRMSGraph->Write();

    TCanvas* c2_2 = new TCanvas("Pulses per Hit single Rows",
                                "Pulses per Hit single Rows");
    TGraphErrors* runPulsePerHitRowGraph = new TGraphErrors(
        runZPositionsSingleRow.size(), &runZPositionsSingleRow[0],
        &runPulsePerHitRowMean[0], &runZPositionsSingleRowError[0],
        &runPulsePerHitRowMeanErr[0]);
    runPulsePerHitRowGraph->SetNameTitle("PulsePerHitRowGraph",
                                         "Pulses per Hit");
    runPulsePerHitRowGraph->GetXaxis()->SetTitle("drift distance [mm]");
    runPulsePerHitRowGraph->GetYaxis()->SetTitle("mean #pulses per hit");
    runPulsePerHitRowGraph->Draw("AP");
    gPad->Update();
    c2_2->SaveAs(Form("%s%s", pathToOutput, "PulsePerHitRow.png"));
    c2_2->SaveAs(Form("%s%s", pathToOutput, "PulsePerHitRow.pdf"));
    runPulsePerHitRowGraph->Write();

    TCanvas* c3_2 = new TCanvas("Pulses per Hit RMS single Rows",
                                "Pulses per Hit RMS single Rows");
    TGraphErrors* runPulsePerHitRowRMSGraph = new TGraphErrors(
        runZPositionsSingleRow.size(), &runZPositionsSingleRow[0],
        &runPulsePerHitRowRMS[0], &runZPositionsSingleRowError[0],
        &runPulsePerHitRowRMSErr[0]);
    runPulsePerHitRowRMSGraph->SetNameTitle("PulsePerHitRowRMSGraph",
                                            "Pulses per Hit RMS");
    runPulsePerHitRowRMSGraph->GetXaxis()->SetTitle("drift distance [mm]");
    runPulsePerHitRowRMSGraph->GetYaxis()->SetTitle("sigma #pulses per hit");
    runPulsePerHitRowRMSGraph->Draw("AP");
    gPad->Update();
    c3_2->SaveAs(Form("%s%s", pathToOutput, "PulsePerHitRowRMS.png"));
    c3_2->SaveAs(Form("%s%s", pathToOutput, "PulsePerHitRowRMS.pdf"));
    runPulsePerHitRowRMSGraph->Write();

    gStyle->SetPadRightMargin(0.18);

    TCanvas* c0 = new TCanvas("Pulse Fractions", "Pulse Fractions");
    pulseFractionsVSDrift->Draw("A");
    pulseFractionsVSDrift->GetXaxis()->SetTitle("drift distance [mm]");
    pulseFractionsVSDrift->GetYaxis()->SetTitle("fraction of hits");
    gPad->BuildLegend(0.83, 0.2, 0.99, 0.9, "pads in hit");
    c0->SaveAs(Form("%s%s", pathToOutput, "PulseFractions.png"));
    c0->SaveAs(Form("%s%s", pathToOutput, "PulseFractions.pdf"));

    TCanvas* c5 = new TCanvas("Pulse Fractions 4+ combined",
                              "Pulse Fractions 4+ combined");
    pulseFractionsVSDrift4Plus->Draw("A");
    pulseFractionsVSDrift4Plus->GetXaxis()->SetTitle("drift distance [mm]");
    pulseFractionsVSDrift4Plus->GetYaxis()->SetTitle("fraction of hits");
    gPad->BuildLegend(0.83, 0.2, 0.99, 0.9, "pads in hit");
    c5->SaveAs(Form("%s%s", pathToOutput, "PulseFractions4Plus.png"));
    c5->SaveAs(Form("%s%s", pathToOutput, "PulseFractions4Plus.pdf"));

    TCanvas* c1 = new TCanvas("Pulses vs Row", "Pulses per Hit vs Row");
    runPulsesVSRow->Draw("A");
    gPad->BuildLegend(0.85, 0.2, 0.99, 0.99, "[mm]");
    c1->SaveAs(Form("%s%s", pathToOutput, "PulsePerHitVSRow.png"));
    c1->SaveAs(Form("%s%s", pathToOutput, "PulsePerHitVSRow.pdf"));

    TCanvas* c4 = new TCanvas("Pulses per Hit Dist", "Pulses per Hit Dist");
    pulsesPerHitDist->Draw("nostack");
    gPad->BuildLegend(0.85, 0.2, 0.99, 0.99, "[mm]");
    c4->SaveAs(Form("%s%s", pathToOutput, "PulsePerHitDist.png"));
    c4->SaveAs(Form("%s%s", pathToOutput, "PulsePerHitDist.pdf"));

    // Saving these to save plot details for future use.  --Peter
    c0->Write();
    c1->Write();
    c2->Write();
    c3->Write();
    c4->Write();
    c5->Write();

    outfile->Write();
    outfile->Close();
    gStyle->SetPadRightMargin(0.04);
}
