#include <vector>
#include <iostream>
#include "TLatex.h"
#include "TCanvas.h"
#include "TH1.h"
#include "TF1.h"
#include "TGraphErrors.h"
#include "TMultiGraph.h"
#include "TFile.h"
#include "TNtuple.h"
#include "TStyle.h"
#include "TROOT.h"
#include "TFrame.h"
#include "sstream"
#include <utility>
#include <cstdarg>
#include "TColor.h"
#include "TMath.h"

/*How to call this:
.L $MARLINTPC/analysis/rootscripts/ThetaResolutionPerformance.C++
ThetaResolutionPerformance("path for output","path to the aida file", number of runs in the following list, list of the runs)
e.g ThetaResolutionPerformance("../myplots","../AidaFiles/",3,18850,18851,18860)
*/
void ThetaResolutionPerformance(char * pathToOutput,char * pathToAida, int nRuns, ...)
{

	//loading the style
	gROOT->ProcessLine(".L ${MARLINTPC}/analysis/rootscripts/LCTPCStyle.C");
	gROOT->ProcessLine("LCTPCStyle()");

	//open a root file to store the graphs in
	TFile *outfile=new TFile(Form("%s%s",pathToOutput,"ThetaResolutionPerformance.root"),"RECREATE");

	//initialize the input list
	va_list argList;
	va_start(argList,nRuns);

	int runNumber;
	//create colors
	const int BINNING=nRuns;
	Int_t colors[BINNING];

	const Int_t NRGBs = 5;

	Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
	Double_t red[NRGBs]   = { 0.00, 0.00, 0.87, 1.00, 0.51 };
	Double_t green[NRGBs] = { 0.00, 0.81, 1.00, 0.20, 0.00 };
	Double_t blue[NRGBs]  = { 0.51, 1.00, 0.12, 0.00, 0.00 };
	Int_t FI = TColor::CreateGradientColorTable(NRGBs, stops, red, green, blue, BINNING);

	for (int i=0;i<BINNING;i++)
		colors[i] = FI+i;

	// will contain the results from the processors
	THStack *runStack=new THStack();
	runStack->SetTitle("number hits on row;row;# entries");
	std::vector<double> runSingleZRes;
	std::vector<double> runSingleZResError;
	std::vector<double> runSingleXYRes;
	std::vector<double> runSingleXYResError;
	std::vector<double> theta;
	std::vector<double> thetaErr;

	//loop over the given runs
	for (int i=0;i<nRuns;i++)
	{
		//get run number from input
		runNumber=va_arg(argList,int);

		//create stringstream containing the aida file name
		std::stringstream runNumberString;
		std::cout<<"run number: "<<runNumber<<std::endl;
		runNumberString <<pathToAida<<"aida_run" <<runNumber<< ".root";

		//open the aida file
		TFile *file  = TFile::Open(runNumberString.str().c_str(),"Read");
		//check that the file exists
		if (file==0)
		{
			std::cout<<"This run will be skipped. The aida file does not exist. Please check this!"<<std::endl;
			continue;
		}

		//get the tuples from the file
		TNtuple * zPosTuple=(TNtuple*)file->Get("MyAnalysisCutSelectorProcessor/RunZPositions");
		TNtuple * bestRowTuple=(TNtuple*)file->Get("MyThetaResolutionPerformanceProcessor/bestRowResolution");
		//check that the tuples exists
		if ( zPosTuple==0 || bestRowTuple==0 )
		{
			std::cout<<"This run will be skipped. One of the necessary root tuples does not exist. Please check that you ran the corresponding processors and that the tuple is in the aida file!"<<std::endl;
			continue;

		}

		//open the first and only entry of the tuples
		zPosTuple->GetEntry(0);


		if(file->Get("MyThetaResolutionPerformanceProcessor/htemp")==0)
		{
			std::cout << "This run will be skipped. MyThetaResolutionPerformanceProcessor/htemp does not exist. Please check that you ran the corresponding processors and that the graph is in the aida file!" << std::endl;
			continue;
		}
		TH1D* rowmodule = (TH1D*)(file->Get("MyThetaResolutionPerformanceProcessor/htemp")->Clone());

		rowmodule->Draw("AL");
		int zpos= int(567.6-zPosTuple->GetArgs()[0]) ;
		std::ostringstream legEntry;
		legEntry << "z = "<<zpos;
		rowmodule->SetLineColor(colors[i]);
		rowmodule->SetMarkerColor(colors[i]);
		rowmodule->SetMarkerStyle(22);
		rowmodule->SetTitle(legEntry.str().c_str());
		runStack->Add(rowmodule);
		bestRowTuple->GetEntry(0);
		runSingleXYRes.push_back(bestRowTuple->GetArgs()[4]);
		runSingleXYResError.push_back(bestRowTuple->GetArgs()[5]);


		runSingleZRes.push_back(bestRowTuple->GetArgs()[6]);
		runSingleZResError.push_back(bestRowTuple->GetArgs()[7]);


		if (bestRowTuple->GetArgs()[2]>=0)
			theta.push_back(90-atan(1/bestRowTuple->GetArgs()[2])*180/M_PI);
		else
			theta.push_back(-atan(1/bestRowTuple->GetArgs()[2])*180/M_PI-90);
		thetaErr.push_back(bestRowTuple->GetArgs()[3]*180/M_PI/(1+bestRowTuple->GetArgs()[2]*bestRowTuple->GetArgs()[2]));

		//file->Close();
		//delete file;
	}

	va_end(argList);


	//save the results
	TCanvas *c1 = new TCanvas("ModuleRow","ModuleRow");

	runStack->Draw("nostack");
	gPad->BuildLegend(0.85,0.2,0.99,0.99,"[mm]");
	c1->SaveAs(Form("%s%s",pathToOutput,"ModuleRow.png"));
	c1->SaveAs(Form("%s%s",pathToOutput,"ModuleRow.pdf"));
	c1->SaveAs(Form("%s%s",pathToOutput,"ModuleRow.tex"));

	TGraphErrors  *xyResolution=new TGraphErrors(theta.size(),&theta[0], &runSingleXYRes[0],&thetaErr[0], &runSingleXYResError[0]);
	TGraphErrors  *zResolution=new TGraphErrors(theta.size(),&theta[0], &runSingleZRes[0],&thetaErr[0], &runSingleZResError[0]);

	xyResolution->GetYaxis()->SetTitle("#sigma_{r#varphi} [mm]");
	xyResolution->GetXaxis()->SetTitle("#theta [°]");

	zResolution->GetYaxis()->SetTitle("#sigma_{z} [mm]");
	zResolution->GetXaxis()->SetTitle("#theta [°]");

	TF1* sqrtcos = new TF1("sqrtcos","sqrt([0]*[0]+[1]*[1]*[2]/10/[3]*cos((x-[4])*TMath::Pi()/180))");
	TF1* tan2 = new TF1("tan2","sqrt([0]*[0]+5.85*5.85/(12*[1])*pow(tan((x-[2])*TMath::Pi()/180),2))");
	sqrtcos->SetLineColor(kRed);
	sqrtcos->SetParameters(0.07,0.104,100,30,0.);
	tan2->SetParameters(0.3,6,0);
	tan2->FixParameter(2,0);
	tan2->SetLineColor(kRed);

	TCanvas *c2 = new TCanvas("ThetaXYResolution","ThetaXYResolution");
	xyResolution->Fit("sqrtcos","Q");
	xyResolution->Fit("sqrtcos");
	xyResolution->SetMarkerStyle(22);
	xyResolution->SetFillColor(0);
	xyResolution->Draw("AP");
	c2->SaveAs(Form("%s%s",pathToOutput,"ThetaXYResolution.png"));
	c2->SaveAs(Form("%s%s",pathToOutput,"ThetaXYResolution.pdf"));
	c2->SaveAs(Form("%s%s",pathToOutput,"ThetaXYResolution.tex"));
	TCanvas *c3 = new TCanvas("ThetaZResolution","ThetaZResolution");
	zResolution->Fit("tan2","Q");
	zResolution->Fit("tan2");
	zResolution->SetMarkerStyle(22);
	zResolution->SetFillColor(0);
	zResolution->Draw("AP");
	c3->SaveAs(Form("%s%s",pathToOutput,"ModuleRow.png"));
	c3->SaveAs(Form("%s%s",pathToOutput,"ModuleRow.pdf"));
	c3->SaveAs(Form("%s%s",pathToOutput,"ModuleRow.tex"));
}
