#include "GeometricMeanResidualsTupleProcessor.h"
#include <iostream>

//Marlin
#include "marlin/VerbosityLevels.h"
#include <marlin/Exceptions.h>
#include <marlin/Global.h>

//LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackImpl.h>
#include <IMPL/TrackerHitImpl.h>

// GEAR
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>

//Other
#include "TrackFitterFactory.h"
#include "TrackFitterBase.h"
#include "TrackFitterKalman.h"
#include <CLHEP/Vector/ThreeVector.h>
#include "IntersectionCalculator.h"


using namespace lcio ;
using namespace marlin ;


namespace marlintpc
{

  GeometricMeanResidualsTupleProcessor aGeometricMeanResidualsTupleProcessor ;

  GeometricMeanResidualsTupleProcessor::GeometricMeanResidualsTupleProcessor() : Processor("GeometricMeanResidualsTupleProcessor"), 
										 _trackCollectionParameters(0) {



    // modify processor description
    _description = std::string("GeometricMeanResidualsTupleProcessor calculates the inclusive and exclusive residuals of a track, and outputs them ") +
				   std::string("to a Root file, along with useful information about the hits and the track.") ;


		_input_track_collection_names.push_back("TPCTracks");

    // register steering parameters: name, description, class-variable, default value
    registerInputCollections( LCIO::TRACK,
            "TrackCollections" , 
            "Names of the input track collections. N.B. All tracks should use the same Track Fitter. (DEFAULT: TPCTracks)"  ,
            _input_track_collection_names,
            _input_track_collection_names
    );
    registerOptionalParameter( "TrackType" , 
            "Type of track to fit. Straight = 0, Helix = 1. (DEFAULT: 0)"  ,
            _track_type,
            0
    );
    registerOptionalParameter( "TrackFitterType" , 
            std::string("Type of track fitter to use. Options: SimpleChiSquare = 0, SimpleChiSquarePads = 1, SimpleMinimizer = 2,") +
						std::string(" Kalman = 3, LinearRegression = 4, SimpleHelix = 5. (DEFAULT 0) ")  ,
            _track_fitter_type,
            0
    );

    registerProcessorParameter("NumberOfRows" ,
			       "A output Ntuple needs number of rows.",
			       _kNlayers,
			       int(84));
    
}

GeometricMeanResidualsTupleProcessor::~GeometricMeanResidualsTupleProcessor()
{
  delete _trackCollectionParameters;
}


void GeometricMeanResidualsTupleProcessor::init() { 

    streamlog_out(DEBUG) << "   init called  " << std::endl ;

    // usually a good idea to
    printParameters() ;

  // initialise the collection parameters
    _trackCollectionParameters = new IMPL::LCParametersImpl();
    
  // set the pamameters for this fitter
    _trackCollectionParameters->setValue("TrackFitterType",
					 "KALMAN");
    _trackCollectionParameters->setValue("kNlayers",
					 _kNlayers);
    
    _nRun = 0 ;
    _nEvt = 0 ;

	if(_track_type < 0 || _track_type > 1)
	{
		streamlog_out(ERROR) << "Error: Invalid track type selected. Please choose between 0 (straight track) and 1 (helical track). " << std::endl;
		throw StopProcessingException(this);
	}

	if (_track_type == 0)
	{
		_output_tuple = AIDAProcessor::tupleFactory(this)->create("TrackTuple",
				"TrackTuple",
				"float D0, Phi, Z0, TanLambda, Chi2, int ndf, double DxIn, DxEx, DPhiIn, DPhiEx, DzIn, DzEx, HitX, HitPhi, TrackX, LocalPhi, HitZ, int Module, Row, Column, NTracks, NHits, NPulses");
	}
	else
	{
		_output_tuple = AIDAProcessor::tupleFactory(this)->create("TrackTuple",
				"TrackTuple",
				"float D0, Phi, Omega, Z0, TanLambda, Chi2, int ndf, double DxIn, DxEx, DPhiIn, DPhiEx, DzIn, DzEx, HitX, HitPhi, TrackX, LocalPhi, HitZ, TrackZ, int Module, Row, Column, NTracks, NHits, NPulses, EventNum");
	}

}


void GeometricMeanResidualsTupleProcessor::processRunHeader( LCRunHeader*  /*run*/) { 

    _nRun++ ;

} 


void GeometricMeanResidualsTupleProcessor::processEvent( LCEvent * evt ) { 

	//Get the track collection, if possible
  IMPL::LCCollectionVec* inputCol = NULL;

	for(lcio::StringVec::iterator nameIt = _input_track_collection_names.begin(); nameIt != _input_track_collection_names.end(); ++nameIt)
	{
		try
		{
			inputCol = dynamic_cast<IMPL::LCCollectionVec*>(evt->getCollection( *nameIt )) ;
		} 
		catch(DataNotAvailableException)
		{
			streamlog_out(WARNING4) << "Input collection " << *nameIt << " cannot be found in event "
				<< evt->getEventNumber() << std::endl;
			continue;
		}

		if(!inputCol)
		{
			streamlog_out(ERROR) << "Error: Failed to cast the input collection as an LCCollectionVec. " << std::endl;
			throw StopProcessingException(this);
		}

		//Get the requested type of track fitter, if possible
		marlintpc::TrackFitterBase * trackFitter = NULL;  

		switch(_track_fitter_type)
		{

			case 0 :
				trackFitter = marlintpc::TrackFitterFactory::getTrackFitter( 
						marlintpc::TrackFitterBase::FitterTypes::SIMPLECHISQUARE,0);
				break;
			case 1 :
				trackFitter = marlintpc::TrackFitterFactory::getTrackFitter( 
						marlintpc::TrackFitterBase::FitterTypes::SIMPLECHISQUAREPADS,0);
				break;
			case 2 :
				trackFitter = marlintpc::TrackFitterFactory::getTrackFitter( 
						marlintpc::TrackFitterBase::FitterTypes::SIMPLEMINIMIZER,0);
				break;
			case 3 :
//				trackFitter = marlintpc::TrackFitterFactory::getTrackFitter( 
//						marlintpc::TrackFitterBase::FitterTypes::KALMAN,0);
                               trackFitter = static_cast<TrackFitterKalman * >( marlintpc::TrackFitterFactory::getTrackFitter( 
                                                marlintpc::TrackFitterBase::FitterTypes::KALMAN, _trackCollectionParameters));

				break;
			case 4 :
				trackFitter = marlintpc::TrackFitterFactory::getTrackFitter( 
						marlintpc::TrackFitterBase::FitterTypes::LINEARREGRESSION,0);
				break;
			case 5 :
				trackFitter = marlintpc::TrackFitterFactory::getTrackFitter( 
						marlintpc::TrackFitterBase::FitterTypes::SIMPLEHELIXFIT,0);
				break;
			default:
				streamlog_out(ERROR) << "Error: Invalid track fitter selected. Please choose from: " << std::endl;
				streamlog_out(ERROR) << "SimpleChiSquare = 0, SimpleChiSquarePads = 1, SimpleMinimizer = 2" << std::endl;
				streamlog_out(ERROR) << "Kalman          = 3, LinearRegression    = 4, SimpleHelix     = 5." << std::endl;
				throw StopProcessingException(this);
		}

		if(!trackFitter)
		{
			streamlog_out(ERROR) << "Error: The track fitter failed to instantiate! Please investigate!" << std::endl;
			throw StopProcessingException(this);
		}

		//Useful for later
		const gear::TPCParameters& tpcParameters = Global::GEAR->getTPCParameters();

		//Now we calculate the residuals.

		for(IMPL::LCCollectionVec::iterator trackIt = inputCol->begin(); trackIt != inputCol->end(); ++trackIt)
		{

			//Attempt to get the track
			EVENT::Track* track = dynamic_cast<EVENT::Track*>( *trackIt ) ;

			//If this case fails, they should all fail, so give up right here.
			if(!track)
			{
				streamlog_out (ERROR) << "Error: The given collection " << *nameIt << " appears not to contain EVENT::Track objects! " << std::endl;
				throw StopProcessingException(this);
			}

			float d0    = track->getD0();
			float phi   = track->getPhi();
			float omega = track->getOmega();
			float z0    = track->getZ0();
			float tanl  = track->getTanLambda();
			float chi2  = track->getChi2();
			int   ndf   = track->getNdf();
			const float *ref = track->getReferencePoint();

			//I am assuming the hits have been stored with the track, since I don't know how to check for that.
			const EVENT::TrackerHitVec & trackHits = track->getTrackerHits();

			//Loop over the hits
			for(EVENT::TrackerHitVec::const_iterator hitIt = trackHits.begin(); hitIt != trackHits.end(); ++hitIt)
			{
				const EVENT::TrackerHit * aHit = dynamic_cast<const EVENT::TrackerHit*>(*hitIt);

				//Again, if it fails here, it should always fail here.
				if(!aHit)
				{
					streamlog_out (ERROR) << "Error: Failed to cast the track\'s hits to EVENT::TrackerHit objects." << std::endl;
					throw StopProcessingException(this);
				}

				const gear::TPCModule& theModule = tpcParameters.getNearestModule(aHit->getPosition()[0],aHit->getPosition()[1]);
				const int moduleID    = theModule.getModuleID();
				int padID 					 	= theModule.getNearestPad(aHit->getPosition()[0], aHit->getPosition()[1]);
				int rowID            	= theModule.getRowNumber(padID);
				int colID            	= theModule.getPadNumber(padID);


				gear::Vector2D local  = theModule.globalToLocal(aHit->getPosition()[0], aHit->getPosition()[1]); 

				//I am also assuming polar geometry here. If anyone needs an alternative geometry, they should be 
				//able to implement it here without much trouble.
				double hitX = local[0]*local[1];

				double trackX = 0;

				////////////////////////////MODULEDISTORTIONCALCULATOR
				//distinguish between straight track and helix track
				std::pair<double,double> phi_intersec(0,0); 
				double localPhi = 0.; 
				if(_track_type == 0)
				{
					try{
						//calculate two points on the track and transform them into local padplane coordinates to obtain a track parametrisation
						// in local coordionates
						//point1=point of closest approach, should inlcude referencepoint in case it is not 0

						double pca_x = ( -1.0 * track->getD0() ) * std::sin(track->getPhi()) + ref[0];
						double pca_y = ( track->getD0() ) * std::cos(track->getPhi()) + ref[1];
						//point 2
						double p2x=pca_x+std::cos(track->getPhi());
						double p2y=pca_y+std::sin(track->getPhi());
						//transform both points to local coordinate system
						//this gives back r and phi, not x and y for polar padlayout!
						gear::Vector2D pca_local = theModule.globalToLocal(pca_x,pca_y);
						gear::Vector2D p2_local = theModule.globalToLocal(p2x,p2y);
						double pca_local_x=pca_local[0]*cos(pca_local[1]);
						double pca_local_y=pca_local[0]*sin(pca_local[1]);
						double p2_local_x=p2_local[0]*cos(p2_local[1]);
						double p2_local_y=p2_local[0]*sin(p2_local[1]);
						double phi=atan2(p2_local_y-pca_local_y,p2_local_x-pca_local_x);
						phi_intersec= marlintpc::IntersectionCalculator::circle_straightLine(local[0], pca_local_x,pca_local_y,phi);


					}
					catch(std::exception &e) {
						std::cout<<"NO INTERSECTION: " <<local[0]<<" "
							<<track->getOmega()<<" "<< track->getD0()<<" "<<track->getPhi()<<std::endl;
						continue;
					}
				}
				else
				{
					//center of helix in global coordinates
					double xm, ym;
					xm = ref[0] + (1. / track->getOmega() - track->getD0()) * sin(track->getPhi());
					ym = ref[1] - (1. / track->getOmega()  - track->getD0()) * cos(track->getPhi());
					//transform helix center into local coordinate system
					gear::Vector2D helixCenterLocal = theModule.globalToLocal(xm,ym);
					//local comes on in polar, go back to cartesian
					double xm_l=helixCenterLocal[0]*cos(helixCenterLocal[1]);
					double ym_l=helixCenterLocal[0]*sin(helixCenterLocal[1]);

					try{
						phi_intersec= marlintpc::IntersectionCalculator::circle_circle(local[0], xm_l, ym_l,fabs(1. / track->getOmega()));
					}
					catch(std::exception &e) {
						std::cout<<"NO INTERSECTION: " <<local[0]<<" "<<xm_l<<" "<< ym_l<<" : "
							<<track->getOmega()<<" "<< track->getD0()<<" "<<track->getPhi()<<std::endl;
						continue;
					}
				}
				////////////////////////////MODULEDISTORTIONCALCULATOR

				//Figure out which intersection is closest to the hit (i.e., more likely to be the correct intersection)
				//trackX = fabs(intersectionsPhi.first - local[1]) < fabs(intersectionsPhi.second - local[1]) ? 
				//				 intersectionsPhi.first : intersectionsPhi.second;
				trackX = fabs(phi_intersec.first - local[1]) < fabs(phi_intersec.second - local[1]) ? 
					phi_intersec.first : phi_intersec.second;

				trackX *= local[0];

				//Now get the local phi

				if(_track_type == 0)
				{
					gear::Vector2D rowPosGlobal = theModule.localToGlobal(local[0], trackX/local[0]);
					gear::Vector2D modPosGlobal = theModule.getOffset(); 

					double row_x = rowPosGlobal[0] - modPosGlobal[0];
					double row_y = rowPosGlobal[1] - modPosGlobal[1];

					double row_phi = std::atan2(row_y, row_x);

					localPhi = phi - row_phi;
					
				}
				else
				{
					//calculate angle between hit and track
					int padIndex = theModule.getNearestPad(aHit->getPosition()[0],aHit->getPosition()[1]);
					gear::Vector2D padCenterGlobal = theModule.getPadCenter(padIndex);
					gear::Vector2D e_pad (padCenterGlobal[0]-theModule.getOffset()[0],padCenterGlobal[1]-theModule.getOffset()[1]);
					double phi_s=atan2(sin(track->getPhi())-track->getOmega()*(aHit->getPosition()[0]-(-track->getD0()*sin(track->getPhi())+track->getReferencePoint()[0])),
							cos(track->getPhi())+track->getOmega()*(aHit->getPosition()[1]-(track->getD0()*cos(track->getPhi())+track->getReferencePoint()[1])));

					localPhi = phi_s-atan(e_pad[1]/e_pad[0]);
				}

				//Good. Now all we need are the residuals.
				EVENT::DoubleVec residualsIn = trackFitter->calculateResiduals(aHit, track);

				//Make a simple copy of this track to get the exclusive residuals
				//Can't use the LCObjectCopier because it's too complicated to remove a hit
				IMPL::TrackImpl * simpleCopy = new IMPL::TrackImpl();
				simpleCopy->setTypeBit(LCIO::TRBIT_HITS);
				for (EVENT::TrackerHitVec::const_iterator hitItCopy = trackHits.begin(); hitItCopy != trackHits.end(); ++hitItCopy) 
				{
					//Using a const cast because I know these hits are not going to be modified. It would be nice if there were another
					//efficient alternative...
					if ( hitIt != hitItCopy ) simpleCopy->addHit(const_cast<EVENT::TrackerHit*>(*hitItCopy));
				}

				//Copy over aaaall the info...
				simpleCopy->setD0(d0);
				simpleCopy->setPhi(phi);
				simpleCopy->setOmega(omega);
				simpleCopy->setZ0(z0);
				simpleCopy->setTanLambda(tanl);
				simpleCopy->setChi2(chi2);
				simpleCopy->setNdf(ndf);
				float ref[3] = {track->getReferencePoint()[0], track->getReferencePoint()[1], track->getReferencePoint()[2]};
				simpleCopy->setReferencePoint(ref);
				simpleCopy->setCovMatrix(track->getCovMatrix());

				//Fit the exclusive track
				IMPL::TrackImpl * exTrack = trackFitter->fitTrack(simpleCopy);

				//Now get the exclusive residual
				EVENT::DoubleVec residualsEx = trackFitter->calculateResiduals(aHit, exTrack);

				if(_track_type == 1)
				{

					//Verifying residuals/////////////////////
					double dxIn = hitX - trackX;
					//double dPhiIn = trackX/local[0] - hitX/local[0];
					double dPhiIn = dxIn/local[0];

					//center of helix in global coordinates
					double xm, ym;
					xm = ref[0] + (1. / exTrack->getOmega() - exTrack->getD0()) * sin(exTrack->getPhi());
					ym = ref[1] - (1. / exTrack->getOmega()  - exTrack->getD0()) * cos(exTrack->getPhi());
					//transform helix center into local coordinate system
					gear::Vector2D helixCenterLocal = theModule.globalToLocal(xm,ym);
					//local comes on in polar, go back to cartesian
					double xm_l=helixCenterLocal[0]*cos(helixCenterLocal[1]);
					double ym_l=helixCenterLocal[0]*sin(helixCenterLocal[1]);

					try{
						phi_intersec= marlintpc::IntersectionCalculator::circle_circle(local[0], xm_l, ym_l,fabs(1. / exTrack->getOmega()));
					}
					catch(std::exception &e) {
						std::cout<<"NO INTERSECTION: " <<local[0]<<" "<<xm_l<<" "<< ym_l<<" : "
							<<track->getOmega()<<" "<< track->getD0()<<" "<<exTrack->getPhi()<<std::endl;
						continue;
					}
					double exTrackX = fabs(phi_intersec.first - local[1]) < fabs(phi_intersec.second - local[1]) ? 
						phi_intersec.first : phi_intersec.second;

					exTrackX *= local[0];

					double dxEx = hitX - exTrackX;
					//double dPhiEx = exTrackX/local[0] - hitX/local[0];
					double dPhiEx = dxEx/local[0];
					/////////////////////////////////////////

					/////CALCULATING Z TEMPORARILY///////////
					double pca_x = ref[0] + ( -1.0 * track->getD0() ) * std::sin(track->getPhi());
					double pca_y = ref[1] + ( track->getD0() ) * std::cos(track->getPhi());
					double r = 1. / track->getOmega();
					//center of circle
					xm = track->getReferencePoint()[0] + (r - track->getD0()) * sin(track->getPhi());
					ym = track->getReferencePoint()[1] - (r - track->getD0()) * cos(track->getPhi());
					double s = r * ( atan2(pca_y - ym , pca_x - xm)
							-
							atan2( aHit->getPosition()[1]- ym ,  aHit->getPosition()[0] - xm)
							);

					double s_1, s_2;
					s_1 = s;
					if(s < 0)
					{
						s_2 = s + fabs(2.*M_PI/track->getOmega());
					}
					else
					{
						s_2 = s - fabs(2.*M_PI/track->getOmega());
					}

					if(fabs(s_1) <= fabs(s_2)) s = s_1;
					else s = s_2;
					double trackZ = ref[2] + z0 + s*tanl;
					int count = 0;

					_output_tuple->fill(count++, d0); //D0
					_output_tuple->fill(count++, phi); //Phi
					_output_tuple->fill(count++, z0); //Z0
					_output_tuple->fill(count++, omega); //Omega
					_output_tuple->fill(count++, tanl); //TanLambda
					_output_tuple->fill(count++, chi2); //Chi2 (currently sneakily just the x-y chi2)
					_output_tuple->fill(count++, ndf); //Ndf
					_output_tuple->fill(count++, residualsIn[0]);
					_output_tuple->fill(count++, residualsEx[0]);
					_output_tuple->fill(count++, dPhiIn); //Inclusive Phi residuals
					_output_tuple->fill(count++, dPhiEx); //Exclusive Phi residuals
					_output_tuple->fill(count++, residualsIn[1]); //Inclusive z residuals
					_output_tuple->fill(count++, residualsEx[1]); //Exclusive z residuals
					_output_tuple->fill(count++, hitX); //Hit rphi
					_output_tuple->fill(count++, local[1]); //Hit phi
					_output_tuple->fill(count++, trackX); //Track rphi
					_output_tuple->fill(count++, localPhi); //Local track angle
					_output_tuple->fill(count++, aHit->getPosition()[2]); //Z of Hit
					_output_tuple->fill(count++, trackZ); //Z of Track
					_output_tuple->fill(count++, moduleID); //Hit Module ID
					_output_tuple->fill(count++, rowID); //Hit Row ID
					_output_tuple->fill(count++, colID); //Hit Column ID
					_output_tuple->fill(count++, inputCol->getNumberOfElements()); //NTracks
					_output_tuple->fill(count++, (int)trackHits.size()); //NHits
					_output_tuple->fill(count++, (int)aHit->getRawHits().size()); //NPulses
					_output_tuple->fill(count++, evt->getEventNumber()); //Event Number
					_output_tuple->addRow();

				}
				else
				{
					double dPhiIn = residualsIn[0]/local[0];
					double dPhiEx = residualsEx[0]/local[0];
					int count = 0;

					_output_tuple->fill(count++, d0); //D0
					_output_tuple->fill(count++, phi); //Phi
					_output_tuple->fill(count++, z0); //Z0
					_output_tuple->fill(count++, tanl); //TanLambda
					_output_tuple->fill(count++, chi2); //Chi2 (currently sneakily just the x-y chi2)
					_output_tuple->fill(count++, ndf); //Ndf
					_output_tuple->fill(count++, residualsIn[0]);
					_output_tuple->fill(count++, residualsEx[0]);
					_output_tuple->fill(count++, dPhiIn); //Inclusive Phi residuals
					_output_tuple->fill(count++, dPhiEx); //Exclusive Phi residuals
					_output_tuple->fill(count++, residualsIn[1]); //Inclusive z residuals
					_output_tuple->fill(count++, residualsEx[1]); //Exclusive z residuals
					_output_tuple->fill(count++, hitX); //Hit rphi
					_output_tuple->fill(count++, local[1]); //Hit phi
					_output_tuple->fill(count++, trackX); //Track rphi
					_output_tuple->fill(count++, localPhi); //Local track angle
					_output_tuple->fill(count++, aHit->getPosition()[2]); //Z of Hit
					_output_tuple->fill(count++, moduleID); //Hit Module ID
					_output_tuple->fill(count++, rowID); //Hit Row ID
					_output_tuple->fill(count++, colID); //Hit Column ID
					_output_tuple->fill(count++, inputCol->getNumberOfElements()); //NTracks
					_output_tuple->fill(count++, (int)trackHits.size()); //NHits
					_output_tuple->fill(count++, (int)aHit->getRawHits().size()); //NPulses
					_output_tuple->addRow();

				}

			}//End loop over hits

		}//End loop over tracks

	}//End loop over track collection names

}//End processEvent()


void GeometricMeanResidualsTupleProcessor::check( LCEvent *  /*evt*/ ) { 
}


void GeometricMeanResidualsTupleProcessor::end(){ 
}

}
