#include "HitCorrelationProcessor.h"

#include "HistogramWidthExtractionTools.h"
#include "TrackFitterFactory.h"
#include "TrackFitterBase.h"
#include "LCObjectCopier.h"
#include <UTIL/BitField64.h>
#include <UTIL/ILDConf.h>

#include <iostream>
#include <string>

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <IMPL/TrackImpl.h>
#include "TrackFitterBase.h"
#include <marlin/Exceptions.h>
#include <Exceptions.h>

//GEAR
#include "gear/TPCModule.h"
#include <gear/BField.h>
#include <gearimpl/Vector3D.h>

//ROOT
#include "TGraphErrors.h"

// Aida
#ifdef MARLIN_USE_AIDA
//AIDA
#include <marlin/AIDAProcessor.h>
#endif

using namespace lcio;
using namespace marlin;
using namespace std;

namespace marlintpc
{

HitCorrelationProcessor aHitCorrelationProcessor;


HitCorrelationProcessor::HitCorrelationProcessor() : Processor("HitCorrelationProcessor")
{
	// the processor description
	_description = "Determines the Hit correlations of a selectable row with all other rows.";



	registerInputCollection(    LCIO::TRACK,
			"InputCollectionName" ,
			"Name of the input Track collection, default: TPCTracks"  ,
			_inputColName ,
			string("TPCTracks")) ;

	registerProcessorParameter( "xyResolutionBins" ,
			"Number of bins in the xy-residual histogram, default: 256",
			_xyResolutionBins,
			int(256));

	registerProcessorParameter( "xyResolutionRange",
			"Range of xy-residual histogram, default: 3.mm",
			_xyResolutionRange,
			float(3));

	registerProcessorParameter( "zResolutionBins" ,
			"Number of bins in the z-residual histogram, default: 256",
			_zResolutionBins,
			int(256));

	registerProcessorParameter( "zResolutionRange",
			"Range of z-residual histogram, default: 5.mm",
			_zResolutionRange,
			float(5));

	registerProcessorParameter("BFieldScaleFactor",
			"Scales magnetic field (map), use 1.0 (default) for field ON or 0.0 for field OFF",
			_bfieldScaleFactor,
			double(1.0));

	registerProcessorParameter( "observedHitRow" ,
			"Row with which the correlation is calculated: default = 35",
			_observedHitRow,
			int(35));

	registerProcessorParameter( "maxRows" ,
			"maximal number of rows on the endplate : default = 84",
			_maxRow,
			int(84));
}
int _maxRow; ///< maximum number of rows on the endplate


void HitCorrelationProcessor::init()
{
	streamlog_out(DEBUG) << " init() called  " << endl;

	// you might want to list the used parameters by invoking
	printParameters();
	AIDAProcessor::tree(this);
	const gear::TPCParameters* tpcParameters = &marlin::Global::GEAR->getTPCParameters();
	_gearParameters= &marlin::Global::GEAR->getTPCParameters();
	std::vector<gear::TPCModule*> TPCModules = tpcParameters->getModules();

	std::vector<gear::TPCModule*>::const_iterator moduleIter = TPCModules.begin();
	int moduletype0 = (*moduleIter)->getPadLayoutType();

	for( ; moduleIter != TPCModules.end() ; ++moduleIter)
	{
		if((*moduleIter)->getPadLayoutType() != moduletype0)
			streamlog_out(MESSAGE) << "WARNING: Modules with different pad layouts are defined. Will use first one!"<< std::endl;
	}

	_myModuleDistortionCalculator=new ModuleDistortionCalculator( &marlin::Global::GEAR->getTPCParameters());

	//	_zResidualCorrelatedRows = new TH2F("zCorrelatedHits","zCorrelatedHits",_zResolutionBins,-_zResolutionRange,_zResolutionRange,_zResolutionBins,-_zResolutionRange,_zResolutionRange);
	_zResidualUncorrelatedRows = new TH2F("zUncorrelatedHits","zUncorrelatedHits",_zResolutionBins,-_zResolutionRange,_zResolutionRange,_zResolutionBins,-_zResolutionRange,_zResolutionRange);
	//	_xyResidualCorrelatedRows= new TH2F("xyCorrelatedHits","xyCorrelatedHits",_xyResolutionBins,-_xyResolutionRange,_xyResolutionRange,_xyResolutionBins,-_xyResolutionRange,_xyResolutionRange);
	_xyResidualUncorrelatedRows = new TH2F("xyUncorrelatedHits","xyUncorrelatedHits",_xyResolutionBins,-_xyResolutionRange,_xyResolutionRange,_xyResolutionBins,-_xyResolutionRange,_xyResolutionRange);

	gear::Vector3D theTPCCenter(_gearParameters->getDoubleVal("TPC_cylinder_centre_c0"),
			_gearParameters->getDoubleVal("TPC_cylinder_centre_c1"),
			0.);

	gear::Vector3D field = marlin::Global::GEAR->getBField().at(theTPCCenter);

	// const bool bOff ( gear_field[2] == 0.);
	const bool bOff ( field.z()*_bfieldScaleFactor == 0. );
	if (bOff)
	{	streamlog_out(DEBUG) << "BField(GEAR) is OFF" << std::endl;}
	else
	{	streamlog_out(DEBUG) << "BField(GEAR) is ON" << std::endl;}
	// constant magnetic field (in z direction)
	// double Bz = gear_field[2];
	double Bz = field.z()*_bfieldScaleFactor;// rho component is accessible with field.rho()

	unsigned char fitterType= TrackFitterBase::FitterTypes::SIMPLEHELIXFIT;
	if(Bz==0)
	{
		fitterType=TrackFitterBase::FitterTypes::LINEARREGRESSION;
	}
	_trackFitter  = TrackFitterFactory::getTrackFitter(fitterType,0);
}


void HitCorrelationProcessor::processRunHeader(LCRunHeader*  /*run*/)
{
	// do something for each run
}


void HitCorrelationProcessor::processEvent(LCEvent* evt)
{

	AIDAProcessor::tree(this);
	// use an LCCollectionVec instead of LCCollection so we can use iterators later
	LCCollectionVec* inputColVec = 0;

	// do a try / catch in case there is an empty event
	try
	{
		inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColName ) );
	}
	catch (DataNotAvailableException &e)
	{
		// no input data, just return and process the next event
		return;
	}
	UTIL::BitField64 encoder(lcio::ILDCellID0::encoder_string);
	// loop the track collection and do something
	for( LCCollectionVec::iterator inputObjectIter = inputColVec->begin() ;
			inputObjectIter < inputColVec->end() ;
			inputObjectIter++ )
	{
		// dynamic cast the input track object
		Track * inputTrack = dynamic_cast<Track *>( *inputObjectIter );

		if ( inputTrack == 0 )
		{
			throw DataNotAvailableException("No input tracks found.");
		}
		for(int j=0 ; j<_maxRow ;j++)
		{
			if(j==_observedHitRow) continue;
			TrackerHitVec hitVec = inputTrack->getTrackerHits();
			TrackerHitVec hitVecCorrelatedHits;
			TrackerHitVec hitVecUncorrelatedHits;
			TrackImpl * trackWOCorrelatedHits = LCObjectCopier::copyTrackWithoutHits(inputTrack);
			TrackImpl * trackWOUncorrelatedHits = LCObjectCopier::copyTrackWithoutHits(inputTrack);
			TrackerHit *temp;
			for (EVENT::TrackerHitVec::const_iterator hitIter = hitVec.begin();
					hitIter < hitVec.end(); hitIter++)
			{
				temp=(*hitIter);
				encoder.setValue((*hitIter)->getCellID0());
				int row = encoder[ILDCellID0::layer];
				if(row ==_observedHitRow || row ==j )
				{
					hitVecCorrelatedHits.push_back(temp);
					trackWOUncorrelatedHits->addHit(*hitIter);
				}
				else if (row ==8 || row ==_maxRow-8)
				{
					hitVecUncorrelatedHits.push_back(temp);
					trackWOCorrelatedHits->addHit(*hitIter);
				}
				else if(row ==_observedHitRow-1 ||  row ==_observedHitRow+1 || row ==j+1 || row ==j-1)
				{
					trackWOUncorrelatedHits->addHit(*hitIter);
				}
				else
				{
					trackWOCorrelatedHits->addHit(*hitIter);
					trackWOUncorrelatedHits->addHit(*hitIter);
				}

			}

			TrackImpl * refittedWOCorrelatedHitsTrack = _trackFitter->fitTrack(trackWOCorrelatedHits);

			std::vector<double> distValue;
			std::vector<double> distValueZ;
			for (EVENT::TrackerHitVec::const_iterator hitIter = hitVecCorrelatedHits.begin();
					hitIter < hitVecCorrelatedHits.end(); hitIter++)
			{
				try{
					distValue.push_back(_myModuleDistortionCalculator->calculateDistanceRow((*hitIter),refittedWOCorrelatedHitsTrack));
					distValueZ.push_back(_myModuleDistortionCalculator->calculateDistanceZ((*hitIter),refittedWOCorrelatedHitsTrack));
				}
				catch(std::exception &e)
				{
					std::cout<<"calculateDistanceRow for distances failed. Event skipped!"<<std::endl;
					throw marlin::SkipEventException(this);
				}
			}


			//fill the histograms with the distance and residual values
			std::map< int, TH2*>::iterator histoiter;
			histoiter = _xyResidualCorrelatedRow.find(j);
			if(histoiter == _xyResidualCorrelatedRow.end())
			{
				TH2D* tempHist = new TH2D(Form("xyCorrelation_%i_%i",_observedHitRow, j),Form("xyCorrelation_%i_%i",_observedHitRow, j),_xyResolutionBins,-_xyResolutionRange,_xyResolutionRange,_xyResolutionBins,-_xyResolutionRange,_xyResolutionRange);
				TH2D* tempHist2 = new TH2D(Form("zCorrelation_%i_%i",_observedHitRow, j),Form("zCorrelation_%i_%i",_observedHitRow, j),_zResolutionBins,-_zResolutionRange,_zResolutionRange,_zResolutionBins,-_zResolutionRange,_zResolutionRange);
				_xyResidualCorrelatedRow.insert( std::make_pair(j,tempHist));
				_zResidualCorrelatedRow.insert( std::make_pair(j,tempHist2));
			}

			if(distValue.size()==2 && distValueZ.size()==2)
			{

				histoiter = _xyResidualCorrelatedRow.find(j);
				(*histoiter).second->Fill(distValue[0],distValue[1]);
				histoiter = _zResidualCorrelatedRow.find(j);
				(*histoiter).second->Fill(distValueZ[0],distValueZ[1]);

			}

			if(j==0)
			{
				TrackImpl * refittedWOUncorrelatedHitsTrack = _trackFitter->fitTrack(trackWOUncorrelatedHits);
				std::vector<double> distValue2;
				std::vector<double> distValueZ2;

				for (EVENT::TrackerHitVec::const_iterator hitIter = hitVecUncorrelatedHits.begin();
						hitIter < hitVecUncorrelatedHits.end(); hitIter++)
				{
					try{
						distValue2.push_back(_myModuleDistortionCalculator->calculateDistanceRow((*hitIter),refittedWOUncorrelatedHitsTrack));
						distValueZ2.push_back(_myModuleDistortionCalculator->calculateDistanceZ((*hitIter),refittedWOUncorrelatedHitsTrack));

					}
					catch(std::exception &e)
					{
						std::cout<<"calculateDistanceRow for distances failed. Event skipped!"<<std::endl;
						throw SkipEventException(this);
					}
				}
				if(distValue2.size()==2 && distValueZ2.size()==2)
				{
					_xyResidualUncorrelatedRows->Fill(distValue2[0],distValue2[1]);
					_zResidualUncorrelatedRows->Fill(distValueZ2[0],distValueZ2[1]);
				}
				delete refittedWOUncorrelatedHitsTrack;

			}


			delete trackWOCorrelatedHits;
			delete trackWOUncorrelatedHits;
			delete refittedWOCorrelatedHitsTrack;
		}
	}// track loop
}


void HitCorrelationProcessor::check(LCEvent*  /*evt*/)
{
	// check what is needed here
}


void HitCorrelationProcessor::end()
{
	AIDAProcessor::tree(this);
	streamlog_out(MESSAGE4)<<"xy uncorrelated -> "<<_xyResidualUncorrelatedRows->GetCorrelationFactor()<<endl;
	//	streamlog_out(MESSAGE4)<<"xy correlated -> "<<_xyResidualCorrelatedRows->GetCorrelationFactor()<<endl;
	streamlog_out(MESSAGE4)<<"z uncorrelated -> "<<_zResidualUncorrelatedRows->GetCorrelationFactor()<<endl;
	//	streamlog_out(MESSAGE4)<<"z correlated -> "<<_zResidualCorrelatedRows->GetCorrelationFactor()<<endl;
	std::map< int, TH2*>::iterator histoiter = _xyResidualCorrelatedRow.begin();
	std::vector<double> rowDistance;
	std::vector<double> correlationFactor;
	std::vector<double> correlationFactorZ;
	for(; histoiter!=_xyResidualCorrelatedRow.end(); histoiter++)
	{
		int key=(*histoiter).first;
		//		streamlog_out(MESSAGE4)<<"xy correlated -> "<<_xyResidualCorrelatedRows->GetCorrelationFactor()<<endl;
		rowDistance.push_back(key);
		correlationFactor.push_back(_xyResidualCorrelatedRow[key]->GetCorrelationFactor());
		correlationFactorZ.push_back(_zResidualCorrelatedRow[key]->GetCorrelationFactor());
	}
	TGraph* xyGraph=new TGraph(rowDistance.size(),&(rowDistance[0]),&(correlationFactor[0]));
	xyGraph->SetNameTitle("xyCorrelation","xyCorrelation");
	xyGraph->GetXaxis()->SetTitle("row number");
	xyGraph->GetYaxis()->SetTitle("Correlation in xy");
	xyGraph->Write();
	TGraph* zGraph=new TGraph(rowDistance.size(),&(rowDistance[0]),&(correlationFactorZ[0]));
	zGraph->SetNameTitle("zCorrelation","zCorrelation");
	zGraph->GetXaxis()->SetTitle("row number");
	zGraph->GetYaxis()->SetTitle("Correlation in z");
	zGraph->Write();

	delete xyGraph;
	delete zGraph;
}
}
