#include "LinearThreePointResolutionProcessor.h"
//#include "ZBinTemplateProcessor.cc"

// C++
#include <iostream>
#include <sstream>
#include <cmath>
#include <vector>
#include <algorithm>
//#include <pair>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include "UTIL/LCTOOLS.h"

// Aida
#ifdef MARLIN_USE_AIDA
//AIDA
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
#endif


using namespace lcio ;
using namespace marlin ;

namespace marlintpc{

// define the indices for the template parameter
// they are for convenience and better code readability
    const unsigned int LinearThreePointResolutionProcessor::_histo3PIndex   = 0;
    


    LinearThreePointResolutionProcessor aLinearThreePointResolutionProcessor ;
    

    LinearThreePointResolutionProcessor::LinearThreePointResolutionProcessor() : 
    	ZBinTemplateProcessor<N_1D_HISTOS_THREE_POINT_RESOLUTION_PROCESSOR ,
			      N_2D_HISTOS_THREE_POINT_RESOLUTION_PROCESSOR>("LinearThreePointResolutionProcessor"),
	histo3P(0), 
	histo3P2D(0),
	histoNoTrack(0) 

    {


//  Processor("LinearThreePointResolutionProcessor") {

	/* the constructor call first the super constructor which the name 
	 * of the processor
	 * this must by the same as the class name.
	 */
	
	/* description of the processor which will be displayed in the
	 * steering file autometicly generated by 
	 * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
	 */
	_description = "LinearThreePointResolutionProcessor determines the 3 point resolution of a linear track" ;
	
	
	/* register steering parameters: name, description, class-variable, default value
	 * the type will by definded by the type of the default value
	 * string, double, float and int are possible
	 * plaese note to get a float you must cast in C++: (float) 1.3 
	 */
	registerInputCollection( LCIO::TRACK,
				 "InputCollectionName" , 
				 "Name of the input collection"  ,
				 _inputColName ,
				 std::string("TimePixTracks") ) ;
	
	registerProcessorParameter( "HistoRange3P" , 
				    "d3P will have range (-_histoRange3P, _histoRange3P) in each histogram, default _histoRange3P = 1 mm"  ,
				    _histoRange3P ,
				    float(1.) ) ;

	registerProcessorParameter( "HistoBin3P" , 
				    "d3P values will be binned with _histoBin3P in each histgram, default _histoBin3P = 100"  ,
				    _histoBin3P ,
				    int(100) ) ;

	registerProcessorParameter( "HistoRangeTime" , 
				    "Time values will have range (0, _histoRangeTime) in each histgram, default _histoRangeTime = 16000 counts"  ,
				    _histoRangeTime ,
				    float(16000.) ) ;

	registerProcessorParameter( "HistoBinTime" , 
				    "Time values will be binned with _histoBinTime in each histgram, default _histoBinTime = 16000"  ,
				    _histoBinTime ,
				    int(16000) ) ;

	registerProcessorParameter( "HistoBinNoTrack" , 
				    "histoNoTrack will be binned with _histoBinNoTrack and its range is (0,_histoBinNoTrack), default _histoBinNoTrack = 20"  ,
				    _histoBinNoTrack ,
				    int(20) ) ;
	

	// set correct names for the z binned histograms
	_histo1DBaseNames[_histo3PIndex]	   = "ThreePointResolution";
	_histo1DTitles[_histo3PIndex] = "Three Point Resoltuion / mm";
	
    }
    
    
    void LinearThreePointResolutionProcessor::init() { 
	
	// usually a good idea to
	printParameters() ;

	histo3P = AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "histo3P", "Three point resolution", 
			     _histoBin3P , -_histoRange3P, _histoRange3P ); 

	histo3P2D = AIDAProcessor::histogramFactory(this)->
	  createHistogram2D( "histo3P2D", "Three point resolution vs time", 
			     _histoBin3P , -_histoRange3P, _histoRange3P, _histoBinTime, 0, _histoRangeTime); 


	histoNoTrack = AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "histoNoTrack", "number of tracks per event" , 
			     _histoBinNoTrack , 0, _histoBinNoTrack ); 


	// set histo range and one bin per int for templated histograms
	// according to Processor parameters
	_histoRanges1D[_histo3PIndex].min = -_histoRange3P;
	_histoRanges1D[_histo3PIndex].max = _histoRange3P;
	_nBins1D[_histo3PIndex] = _histoBin3P;

	// now call ZBinTemplateProcessor's init which initialises the histograms for us
	ZBinTemplateProcessor<N_1D_HISTOS_THREE_POINT_RESOLUTION_PROCESSOR,
	                      N_2D_HISTOS_THREE_POINT_RESOLUTION_PROCESSOR>::init();

	
    }
    
    void LinearThreePointResolutionProcessor::processRunHeader(lcio::LCRunHeader* run) { 
	
      
      run->parameters().setValue( _processorName + "_revision", "$Rev: 5472 $");

      for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
	if ( ! i->second->isOptional() || i->second->valueSet() ) {
	  run->parameters().setValue( _processorName + "_" + i->second->name(), 
				      i->second->value() );
	}
      }

	
    } 
    
    void LinearThreePointResolutionProcessor::processEvent(lcio::LCEvent * evt ) { 

      // get the input collection from the event
      LCCollection* inputCol = 0;

      try{

	inputCol = evt->getCollection( _inputColName ) ;

      } //end of try {

      catch(DataNotAvailableException){
	// if there is not input collection give a warning to the user
	streamlog_out(WARNING4)<<" Input collection "
		 <<_inputColName<<" cannot be found in event "
		 <<evt->getEventNumber()<<std::endl;

	//just return, no data to process
	return;
      }
      

      streamlog_out(DEBUG0)<<"inputCol = evt->getCollection( "<<_inputColName<<" ) ;"<<std::endl;
      // get the number of elements in the input collection, 
      //and loop over it

      histoNoTrack->fill( inputCol->getNumberOfElements() );

      //only consider single tracks
      if( inputCol->getNumberOfElements() == 1 ) {


      for(int iTrack=0; iTrack< inputCol->getNumberOfElements() ; iTrack++){
	
	/* getElement returns only a LCObject
	 * you have to cast it to the get the right type
	 * please use dynamic_cast which will return 0 if the object cannot be casted
	 */
	Track* track = 
	  dynamic_cast<Track*>( inputCol->getElementAt( iTrack ) ) ;
	
	if ( track != 0 ) {
	  //only proceed if the cast was successful

	  streamlog_out(DEBUG2)<<"new track "<<std::endl;

	  if ( track->getOmega() == 0 ) {
	    //only proceed if track is linear

	    TrackerHitVec HitVec = track->getTrackerHits();
	    
	    double phi =  track->getPhi();
	    double d0 =  track->getD0();

	    typedef std::vector< std::pair<TrackerHit*, double> > HitDouble;
	    //fill pair of Hit and s vector and sort them by s
	    HitDouble HitPairVec;

	    for ( TrackerHitVec::iterator i = HitVec.begin(); i < HitVec.end(); i++ ) {

	      double s = cos( phi )*(*i)->getPosition()[0] + sin( phi )*(*i)->getPosition()[1]
		- 2*d0*cos( phi )*sin( phi );

	      
	      TrackerHit* Hit = *i;
	      std::pair<TrackerHit*, double> HitPair( Hit , s);
	      HitPairVec.push_back( HitPair );

	    }

	    
	    streamlog_out(DEBUG0)<<"SORT"<<std::endl;

	    sort( HitPairVec.begin(), HitPairVec.end(), compareHit );

	    streamlog_out(DEBUG0)<<"sorted!"<<std::endl;

	    int count = 1;
	    for ( HitDouble::iterator i = HitPairVec.begin(); i < HitPairVec.end(); i++ ) {

	      streamlog_out(DEBUG0)<<"Hit "<<count++<<std::endl;

	      if ( i < HitPairVec.end() - 2 ) {

		//P1
		double X1 = (*i->first).getPosition()[0];
		double Y1 = (*i->first).getPosition()[1];
		double Z1 = (*i->first).getPosition()[2];

		streamlog_out(DEBUG0)<<"P1 : "<<X1<<" , "<<Y1<<" , "<<Z1<<std::endl;

		//P2
		double X2 = (*(i+1)->first).getPosition()[0];
		double Y2 = (*(i+1)->first).getPosition()[1];
		double Z2 = (*(i+1)->first).getPosition()[2];

		streamlog_out(DEBUG0)<<"P2 : "<<X2<<" , "<<Y2<<" , "<<Z2<<std::endl;

		//P3
		double X3 = (*(i+2)->first).getPosition()[0];
		double Y3 = (*(i+2)->first).getPosition()[1];
		double Z3 = (*(i+2)->first).getPosition()[2];

		streamlog_out(DEBUG0)<<"P3 : "<<X3<<" , "<<Y3<<" , "<<Z3<<std::endl;
		
		//a
		double a1 = X3 - X1;
		double a2 = Y3 - Y1;
		double a3 = Z3 - Z1;

		streamlog_out(DEBUG0)<<"a : "<<a1<<" , "<<a2<<" , "<<a3<<std::endl;
		
		//a*a
		//double a_square = a1*a1 + a2*a2 + a3*a3;
		double a_square = a1*a1 + a2*a2 ;

		streamlog_out(DEBUG0)<<"a_square : "<<a_square<<std::endl;

		//P2 - P1
		double Pb1 = X2 - X1;
		double Pb2 = Y2 - Y1;
		double Pb3 = Z2 - Z1;

		streamlog_out(DEBUG0)<<"Pb: "<<Pb1<<" , "<<Pb2<<" , "<<Pb3<<std::endl;
		
		//double x_F = (Pb1*a1 + Pb2*a2 + Pb3*a3)/a_square;
		double x_F = (Pb1*a1 + Pb2*a2)/a_square;

		streamlog_out(DEBUG0)<<"x_F : "<<x_F<<std::endl;

		double d1 = X2-X1 - x_F*a1;
		double d2 = Y2-Y1 - x_F*a2;
		double d3 = Z2-Z1 - x_F*a3;
		
		streamlog_out(DEBUG0)<<"d : "<<d1<<" , "<<d2<<" , "<<d3<<std::endl;
		
		//F_i = a_i*x_F + b_i point on line next to P2
		//

		//|d| = z(dxa)
		double d = ( a1*d2 - a2*d1 )/ sqrt(a1*a1+a2*a2);
		
		streamlog_out(DEBUG0)<<"d = "<<d<<std::endl;

		double d3P = d * sqrt(2./3.);

		streamlog_out(DEBUG3)<<"d3P = "<<d3P<<std::endl;

//              //Rescaling 

//		//distance of the middle point to middle between outer points
//		double ab = 0.5 - sqrt( (a1*x_F)*(a1*x_F) + (a2*x_F)*(a2*x_F) )/
//		    sqrt( (X3-X1)*(X3-X1) + (Y3-Y1)*(Y3-Y1) );    
//		
//		//Skalierungsfaktor S
//		double S = 0.8182-0.4505*pow(ab,2);
//		
//		d3P = S*d3P;
		
		double time = (*(i+1)->first).getPosition()[2];
		
		histo3P->fill( d3P );
		histo3P2D->fill( d3P, time);

		// fill the number if hits per track into the histogram, use z0 of the track as z position
		fillHistogram( _histo3PIndex, 
			       d3P , 
			       track->getZ0() );		 
		
	      }

	    }

	  

	  } //end of if ( track->getOmega() == 0 )
	  else{

	    streamlog_out(WARNING)<<"your track is not linear,"
				  <<" can't determine 3 point resolution"<<std::endl;
	    
	    continue;

	  }

	} //end of if ( track != 0 ) {
	
      } //end of for(int iTrack=0; iTrack< nTrack ; iTrack++){
      
      }	    

    }
    
    
    void LinearThreePointResolutionProcessor::check(lcio::LCEvent *  /*evt*/ ) { 

      //nothing to check here, could be used to fill checkplots

    }

    
    
    void LinearThreePointResolutionProcessor::end(){ 
      /*	
      streamlog_out(MESSAGE4) << "LinearThreePointResolutionProcessor::end()  " << name() 
		<< " processed " << evt->getEventNumber() 
		<< " events in " << _nRun << " runs "
		<< std::endl ;
      */	

    }
    
}
