#include "PRFPerformanceProcessor.h"


#include <iostream>
#include <string>

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include <EVENT/TrackerPulse.h>
#include <Exceptions.h>

using namespace lcio;
using namespace marlin;
using namespace std;

//ROOT
#include "TGraphErrors.h"

// GEAR
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>
#include <gearimpl/GlobalPadIndex.h>

// Use AIDA histograms
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>

//MARLIN
#include <marlin/Global.h>
#include "HistogramWidthExtractionTools.h"

namespace marlintpc
{

PRFPerformanceProcessor aPRFPerformanceProcessor;


PRFPerformanceProcessor::PRFPerformanceProcessor() : Processor("PRFPerformanceProcessor"),_prfAll(0),
		_gearParameters(0),_singleRowsTuple(0),_bestRowTuple(0),_allRowsTuple(0)
{
	// the processor description
	_description = "PRFPerformanceProcessor creates the PRF plots for each row and for all row integrated.";


	registerInputCollection(    LCIO::TRACK,
			"InputCollectionName" ,
			"Name of the input Track collection, default: TPCTracks"  ,
			_inputColName ,
			string("TPCTracks")) ;

	/*
        registerOutputCollection(   LCIO::TRACKERDATA,
          "OutputCollectionName" ,
          string("Name of the output TrackerData collection")+
          string("only the data related to a pulse will be stored")+
          string(\, default: TPCDataInPulses),
          _outputCollectionName ,
          string("TPCDataInPulses")) ;

        // It's nive to have the default value in the comment
        registerProcessorParameter( "MinPulsesPerHit" ,
	       "Minimum number of pulses for a valid hit, default: 3",
	       _minPulsesPerHit,
	       int(3));

        // Optional parameters should only be used when set
        // in the steering file.
        // The default value is there only for technical reasons,
        // it is meaningless.
        //
        // If properly implemented the cut is not applied when the parameter is not set:
        //   if (parameterSet("ChageCut")) { //apply the cut }
        //
        registerOptionalParameter( "ChargeCut" ,
            "If the charge is below this value the data will be ignored",
            _chargeCut,
           float(17.8));
	 */
	registerProcessorParameter( "ModuleWithBestRow",
			"Module that has the best row (default:0)",
			_bestModule,
			int(0));

	registerProcessorParameter( "RowWithBestPerformance",
			"Row with the best PRF performance (default:0)",
			_bestRow,
			int(0));

}


void PRFPerformanceProcessor::init()
{ 
	streamlog_out(DEBUG) << " init() called  " << endl;
	printParameters();
	AIDAProcessor::tree(this);
	_gearParameters= &marlin::Global::GEAR->getTPCParameters();

	_allRowsTuple = new TNtuple("allRowsPRF","allRowsPRF","sigma:sigmaError:mean:meanError");
	_singleRowsTuple = new TNtuple("singleRowsPRF","singleRowsPRF","module:row:sigma:sigmaError:mean:meanError");
	_bestRowTuple = new TNtuple("bestRowPRF","bestRowPRF","module:row:sigma:sigmaError:mean:meanError");
	_singleRowsTupleCorrected = new TNtuple("singleRowsPadCorPRF","singleRowsPadCorPRF","module:row:sigma:sigmaError:mean:meanError");
	_bestRowTupleCorrected = new TNtuple("bestRowPadCorPRF","bestRowPadCorPRF","module:row:sigma:sigmaError:mean:meanError");

	_prfAll = new TH2D("prf","prf",200,0,0,200,0,1.1);


}


void PRFPerformanceProcessor::processRunHeader(LCRunHeader*  /*run*/)
{
	// do something for each run
} 


void PRFPerformanceProcessor::processEvent(LCEvent* evt)
{ 
	// do something for each event, usually access the input collection:
	AIDAProcessor::tree(this);
	// use an LCCollectionVec instead of LCCollection so we can use iterators later
	LCCollectionVec* inputColVec = 0;

	// do a try / catch in case there is an empty event
	try
	{
		inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColName ) );
	}
	catch (DataNotAvailableException &e)
	{
		// no input data, just return and process the next event
		return;
	}

	// loop the data collection and do something
	for( LCCollectionVec::iterator inputObjectIter = inputColVec->begin() ;
			inputObjectIter < inputColVec->end() ;
			inputObjectIter++ )
	{
		// dynamic cast the input object (LCObject) to what it really is e.g.
		lcio::Track * myTrack= dynamic_cast<lcio::Track *>( *inputObjectIter );

		lcio::TrackerHitVec trackHits =  myTrack->getTrackerHits();
		lcio::TrackerHitVec::const_iterator itHit = trackHits.begin();

		//loop the hits
		for(; itHit != trackHits.end(); ++itHit)
		{
			TrackerHit* hit = *itHit;
//			if( hit->getQuality() != 0 )
//				continue;
//			if (hit->getRawHits().size()<3 )
//				continue;

			const double *position = hit->getPosition();
			streamlog_out(DEBUG4) << "position[0] "<< position[0] << " ; position[1] "<< position[1] << endl;
			gear::GlobalPadIndex nearestPadIndex = _gearParameters->getNearestPad( position[0],
					position[1]);
			const gear::TPCModule& module = _gearParameters->getModule( nearestPadIndex.getModuleID() );
			gear::Vector2D localHit = module.globalToLocal(position[0],position[1]);
			double radius=localHit[0];
			const gear::PadRowLayout2D& localPadPlane = module.getLocalPadLayout();
			int moduleNumber = module.getModuleID();
			int row = module.getRowNumber(nearestPadIndex.getPadIndex());
			std::pair<int, int> key(moduleNumber,row);
			std::map< std::pair<int,int>, TH2*>::iterator histoiter;
			histoiter = _histogramsRow.find(key);
			if(histoiter == _histogramsRow.end())
			{
				TH2* tempHist = new TH2D(Form("HistPRf_%i_%i",moduleNumber,row),"HistPRf_",200,0,0,200,0,1.1);
				_histogramsRow.insert( std::make_pair(key,tempHist));
			}
//
//			TrackerPulse const* maxPulse=0;
//			double maxCharge=0;
//			for ( LCObjectVec::const_iterator pulseIter = hit->getRawHits().begin() ;
//					pulseIter < hit->getRawHits().end() ; pulseIter++ )
//			{
//				TrackerPulse const * pulse = dynamic_cast<TrackerPulse const *>(*pulseIter);
//				if(pulse->getCharge()>maxCharge)
//				{
//					maxCharge=pulse->getCharge();
//					maxPulse=pulse;
//				}
//			}
//			if (maxPulse->getCellID0()!= nearestPadIndex.getPadIndex() || maxPulse->getCellID1()!= nearestPadIndex.getModuleID())
//				continue;

			for ( LCObjectVec::const_iterator pulseIter = hit->getRawHits().begin() ;
					pulseIter < hit->getRawHits().end() ; pulseIter++ )
			{

				TrackerPulse const * pulse = dynamic_cast<TrackerPulse const *>(*pulseIter);

				gear::Vector2D globalPadCenter = module.getPadCenter( pulse->getCellID0() );
				gear::Vector2D localPadCenter = module.globalToLocal(globalPadCenter[0],globalPadCenter[1]);
				
				double positionInPad=0;
				switch ( localPadPlane.getCoordinateType() )
				  {
				  case gear::PadRowLayout2D::POLAR:
				    //local coordinates are r and phi
				    positionInPad = radius*( localHit[1] - localPadCenter[1] );
				    break;
				  case gear::PadRowLayout2D::CARTESIAN:
				    positionInPad = ( localHit[0] - localPadCenter[0] );
				    break;
				  default:
				    throw gear::Exception("Unknown coordinate type");
				  }

				double normPulseCharge=pulse->getCharge()/hit->getEDep();

				_prfAll->Fill(positionInPad,normPulseCharge);
				(*histoiter).second->Fill(positionInPad,normPulseCharge);
			}
		}

	}
}


void PRFPerformanceProcessor::check(LCEvent*  /*evt*/)
{ 
	// check what is needed here
}


void PRFPerformanceProcessor::end()
{ 
	AIDAProcessor::tree(this);
	HistogramWidthExtractionTools *myTool= new HistogramWidthExtractionTools();
	std::map< std::pair<int,int>, TH2*>::iterator histoiter = _histogramsRow.begin();
	myTool->Calculate(_prfAll,2);
	_allRowsTuple->Fill(myTool->GetWidth(),myTool->GetWidthError(),myTool->GetMean(),myTool->GetMeanError());

	std::vector<double> radius,radiusError,sigmaXY,sigmaXYError;
	std::vector<double> sigmaXYPadCor,sigmaXYPadCorError;
	for(; histoiter!=_histogramsRow.end(); histoiter++)
	{
		std::pair<int,int> key = (*histoiter).first;
		myTool->Calculate((*histoiter).second,2);
//		(*histoiter).second->Scale(1/myTool->GetHeight());
//		myTool->Calculate((*histoiter).second,2);
		sigmaXY.push_back(myTool->GetWidth());
		sigmaXYError.push_back(myTool->GetWidthError());
		_singleRowsTuple->Fill(key.first, key.second, myTool->GetWidth(), myTool->GetWidthError(),myTool->GetMean(),myTool->GetMeanError());
				if (key.first==_bestModule && key.second==_bestRow)
				{
							_bestRowTuple->Fill(key.first, key.second, myTool->GetWidth(), myTool->GetWidthError(),myTool->GetMean(),myTool->GetMeanError());
				}
		//get from to corresponding radius at phi=0
		const gear::TPCModule& theModule = _gearParameters->getModule(key.first);
		gear::Vector2D padcenter=theModule.getPadCenter(theModule.getPadIndex(key.second,0));
		double padWidth;
		if (theModule.getLocalPadLayout().getPadLayoutType()==2)//circular pad plane
		{
			padWidth=theModule.globalToLocal(padcenter[0],padcenter[1])[0]*theModule.getPadWidth(theModule.getPadIndex(key.second,0));
		}
		else if (theModule.getLocalPadLayout().getPadLayoutType()==1)//rectangular pad plane
		{

			padWidth=theModule.getPadWidth(theModule.getPadIndex(key.second,0));
		}
		else
		{
			streamlog_out(ERROR)<<"Error: Invalid pad layout type!"<<std::endl;
			throw gear::Exception("Unknown pad layout type");
		}
		double tempvalue = pow(myTool->GetWidth(),2)-pow(padWidth,2)/12;
		if (tempvalue<0) sigmaXYPadCor.push_back(0);
		else sigmaXYPadCor.push_back(sqrt(tempvalue));
		sigmaXYPadCorError.push_back(myTool->GetWidthError());
		_singleRowsTupleCorrected->Fill(key.first, key.second, sqrt(tempvalue), myTool->GetWidthError(),myTool->GetMean(),myTool->GetMeanError());
				if (key.first==_bestModule && key.second==_bestRow)
				{
							_bestRowTupleCorrected->Fill(key.first, key.second, sqrt(tempvalue), myTool->GetWidthError(),myTool->GetMean(),myTool->GetMeanError());
				}

		radius.push_back(sqrt(padcenter[0]*padcenter[0]+padcenter[1]*padcenter[1]));
		radiusError.push_back((theModule.getRowHeight(key.second))/sqrt(12.));
	}
	TGraphErrors *graphRow=new TGraphErrors(radius.size(),&radius[0],&sigmaXY[0],&radiusError[0],&sigmaXYError[0]);
	graphRow->SetNameTitle("PRFRow","PRFRow");
	graphRow->GetXaxis()->SetTitle("r_{row} [mm]");
	graphRow->GetYaxis()->SetTitle("#sigma_{prf} [mm]");
	graphRow->Write();
	TGraphErrors *graphRowPadCor=new TGraphErrors(radius.size(),&radius[0],&sigmaXYPadCor[0],&radiusError[0],&sigmaXYPadCorError[0]);
	graphRowPadCor->SetNameTitle("PRFRowPadCor","PRFRowPadCor");
	graphRowPadCor->GetXaxis()->SetTitle("r_{row} [mm]");
	graphRowPadCor->GetYaxis()->SetTitle("#sigma_{prf} [mm]");
	graphRowPadCor->Write();

	delete myTool;
	delete graphRow;
	delete graphRowPadCor;
}
} // close namespace brackets
