#include "PadResponseProcessor.h"

//C++
#include <iostream>
#include <sstream>
using namespace std;

//LCIO
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include "IntersectionCalculator.h"
#include <EVENT/TrackerPulse.h>
#include <EVENT/LCRelation.h>
#include "UTIL/LCRelationNavigator.h"

// Aida
//#ifdef MARLIN_USE_AIDA // the whole processor does not make sense without aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
// endif

// Marlin 
#include <marlin/Global.h>

// GEAR
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>
#include <gear/PadRowLayout2D.h>

#include "TrackFitterFactory.h"
#include "TrackFitterBase.h"

using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

PadResponseProcessor aPadResponseProcessor;
    
PadResponseProcessor::PadResponseProcessor(): Processor("PadResponseProcessor")
{
    // modify processor description
    _description = "PadResponseProcessor: Fill biased residuals of hits in tracks into AIDA histgram" ;

    // register steering parameters: name, description, class-variable, default value
    registerInputCollection(    LCIO::TRACK,
				"InputTracks" , 
				"Name of the input tracks collection"  ,
				_inputColName ,
				std::string("TPCSeedTracks") ) ;

    registerOptionalParameter( "MCTruthRelationColName",
			       "Name of the collection with relations to the MC truth",
			       _mcTruthRelationsColName,
			       std::string("TPCTruthTracksRelation") );

  registerProcessorParameter( "ResidualMinValue" , 
			      "Minimum residual value histogram (default: -0.3 mm)"  ,
			      _residualMinValue,
			      float(-0.3) ) ;

  registerProcessorParameter( "ResidualMaxValue" , 
			      "Maximal residual value histogram (default: 0.3 mm)"  ,
			      _residualMaxValue,
			      float(0.3) ) ;

  registerProcessorParameter( "SkewnessMinValue" , 
			      "Minimum skewness value histogram (default: -1 )"  ,
			      _skewnessMinValue,
			      float(-1.) ) ;

  registerProcessorParameter( "SkewnessMaxValue" , 
			      "Maximum skewness value histogram (default: 1 )"  ,
			      _skewnessMaxValue,
			      float(1.) ) ;

  registerProcessorParameter( "VarianceMaxValue" , 
			      "Maxnimum variance value histogram (default: 0.5 )"  ,
			      _varianceMaxValue,
			      float(0.5) ) ;

  registerProcessorParameter( "NPadRowHistos" , 
			      "Number of pad rows for which an individual histo should be created (default: 0 )"  ,
			      _nPadRowHistos,
			      0 ) ;

}

PadResponseProcessor::~PadResponseProcessor()
{
}

void PadResponseProcessor::init()
{ 
	// usually a good idea to
	printParameters() ;

 	_residuals_vs_position_on_pad =   AIDAProcessor::histogramFactory(this)->
	    createHistogram2D( "residuals_vs_pos_on_pad", 
			       "residuals_vs_pos_on_pad", 
		               200, -1, 1,
			       200, _residualMinValue, _residualMaxValue);

 	_residuals_vs_variance =   AIDAProcessor::histogramFactory(this)->
	    createHistogram2D( "residuals_vs_variance", 
			       "residuals_vs_variance", 
		               200, 0, _varianceMaxValue ,
			       200, _residualMinValue, _residualMaxValue);

 	_residuals_vs_skewness =   AIDAProcessor::histogramFactory(this)->
	    createHistogram2D( "residuals_vs_skewness", 
			       "residuals_vs_skewness", 
		               200, _skewnessMinValue, _skewnessMaxValue,
			       200, _residualMinValue, _residualMaxValue);

	_padRowHistos.resize( _nPadRowHistos );
	for ( int i = 0 ; i <_nPadRowHistos ; i++ )
	{
	  std::stringstream histoName;
	  histoName << "residuals_vs_pos__row_" << i ;
	  _padRowHistos[i] = AIDAProcessor::histogramFactory(this)->
	    createHistogram2D( histoName.str() , 
			       histoName.str(), 
		               200, -1, 1,
			       200, _residualMinValue, _residualMaxValue);
	}
   }

void PadResponseProcessor::processRunHeader(lcio::LCRunHeader*  /*run*/ )
{
}
    
void PadResponseProcessor::processEvent(lcio::LCEvent * evt ) 
{ 

    // we need a LCCollectionVec to use iterators, as LCCollectio is not a std::vector
    LCCollectionVec* inputCollection=0;
    try{
	inputCollection = dynamic_cast<LCCollectionVec*>(evt->getCollection( _inputColName )) ;
    }
    catch(DataNotAvailableException &e)
    {
	// no input data
	// just continue ...
	return;
    }

    if( inputCollection != 0 )
    {

        // Try to find a collection with LCRelations to the MC truth.
        // This is an optional parameter. Only use it if it is set
        LCCollection* mcTruthRelationsCol = 0;
	LCRelationNavigator* relationNavigator = 0;
        if (parameterSet( "MCTruthRelationColName" ))
        {
	  try {
	    mcTruthRelationsCol = evt->getCollection( _mcTruthRelationsColName ) ;
	  }
	  catch(DataNotAvailableException){
	    streamlog_out(WARNING) << "MCTruth collection requested, "
				   << "but no collection with name \"" << _mcTruthRelationsColName
				   << "\" found in event! Skipping this event" << std::endl;
	    return;
	  }

	  relationNavigator = new LCRelationNavigator( mcTruthRelationsCol );
        }

	int trackNr=0;

        LCParameters const & trackCollectionParameters = inputCollection->getParameters();

	gear::TPCParameters const & tpcParameters = Global::GEAR->getTPCParameters();

	for(LCCollectionVec::iterator it = inputCollection->begin() ; 
	    it < inputCollection->end(); it ++ )
	{
	    
	    streamlog_out(DEBUG4) <<"\t TrackNumber "<<trackNr<<std::endl;

	    Track * track = dynamic_cast<Track *>(*it) ;
	    Track * referenceTrack = 0; // possible reference track for MCTruth
	    
	    if ( relationNavigator ) // there is MCTruth info
	    {
	      LCObjectVec truthTracks = relationNavigator->getRelatedToObjects(track);
	      if ( truthTracks.size() == 0 )
	      {
		streamlog_out(MESSAGE) << "No mc truth for track " << trackNr++
				       << " in event " << evt->getEventNumber() 
				       << ". Skipping this track!" << std::endl;
		continue; // with next track
	      }
	      
	      // always use the first entry... this should be the one with 
	      /// most overlapping hits
	      referenceTrack = dynamic_cast<Track *>( *truthTracks.begin() );
	    }
	    else // no MCTruth
	    {
	      // use the track itself to calculate the residuals (they will be/ might be biased)
	      referenceTrack = track;
	    }

	    // first try to get a fitter, according to the fitter type defined in the track
	    unsigned char fitterType = TrackFitterBase::getFitterTypeFromTrack( track->getType() );

	    if (fitterType == 0) // no fitter has been set
	    {
	        fitterType = TrackFitterBase::FitterTypes::SIMPLEMINIMIZER;
	    }
	    
	    streamlog_out(DEBUG4) << "Using fitter type "<<static_cast<unsigned short>(fitterType)<< " to calculate residuals"<<std::endl;

	    TrackFitterBase *trackFitter = TrackFitterFactory::getTrackFitter( fitterType ,
									       & trackCollectionParameters);
	    
	    TrackerHitVec hitVector = track->getTrackerHits();
	    
	    for (TrackerHitVec::iterator hit = hitVector.begin();
		 hit <  hitVector.end(); hit ++)
	    {
		const double *position = (*hit)->getPosition();

		try
		{
		    EVENT::DoubleVec residuals = trackFitter->calculateResiduals(*hit, referenceTrack);

		    // calculate distance to pad centre. in polar and cartesian LOCAL geometries...

		    // the global gear coordinates might be polar, while the hit coordinates in lcio are 
		    // always cartesian
		    gear::Vector2D globalGearCoordinates;
		    switch ( tpcParameters.getCoordinateType() )
		    {
 		      case gear::PadRowLayout2D::POLAR:
			globalGearCoordinates[0] = sqrt( position[0]*position[0] + position[1]*position[1] );
			globalGearCoordinates[1] = atan2( position[1], position[0] );
			break;
 		      case gear::PadRowLayout2D::CARTESIAN:
			globalGearCoordinates[0] = position[0];
			globalGearCoordinates[1] = position[1];
			break;
		      default:
			throw gear::Exception("Unknown coordinate type");
		    }

		    gear::GlobalPadIndex nearestPadIndex = tpcParameters.getNearestPad( globalGearCoordinates[0],
											globalGearCoordinates[1]);
		    gear::TPCModule const & module = tpcParameters.getModule( nearestPadIndex.getModuleID() );

		    gear::Vector2D localHitCoordinates = module.globalToLocal( globalGearCoordinates[0],
									       globalGearCoordinates[1] );
		    // wrap to 0..2pi
		    if ( localHitCoordinates[1]<0 )  localHitCoordinates[1]+=2*M_PI; 

		    gear::PadRowLayout2D const & localPadLayout = module.getLocalPadLayout();
		    gear::Vector2D localPadCoordinates = localPadLayout.getPadCenter( nearestPadIndex.getPadIndex() );

		    double padPitch = localPadLayout.getPadPitch( nearestPadIndex.getPadIndex() );

		    double positionInPad;
		    bool localIsPolar;
		    switch ( localPadLayout.getCoordinateType() )
		    {
 		      case gear::PadRowLayout2D::POLAR:
			positionInPad = ( localHitCoordinates[1] - localPadCoordinates[1] ) / padPitch;
			streamlog_out(DEBUG4) << "hit "<< localHitCoordinates[1]  / padPitch
					      << " ; pad "<< localPadCoordinates[1] / padPitch  << std::endl;
			localIsPolar = true;
			break;
 		      case gear::PadRowLayout2D::CARTESIAN:
			positionInPad = ( localHitCoordinates[0] - localPadCoordinates[0] ) / padPitch;
			localIsPolar = false;
		      default:
			throw gear::Exception("Unknown coordinate type");
		    }

		    _residuals_vs_position_on_pad->fill( positionInPad , residuals[0] );

		    if ( module.getRowNumber( nearestPadIndex.getPadIndex() ) < _nPadRowHistos)
		    {
		      _padRowHistos[module.getRowNumber( nearestPadIndex.getPadIndex() )]
			->fill(  positionInPad , residuals[0]);
		    }

		    // now calculate variance and skewness
		    // this only makes sense for hits with more than one pad
		    if ( (*hit)->getRawHits().size() < 2 )
		    {
		      streamlog_out(MESSAGE2) << "hit with less than two pads in event "
					      << evt->getEventNumber()
					      << std::endl;
		      continue;
		    }

		    // for this we need the sums over q, q*x , q*x^2 and q*x^3
		    // x is the coordinate along the pad row, might be phi....
		    double sum_q = 0; 
		    double sum_qx = 0;
		    double sum_qxx = 0;
		    double sum_qxxx = 0;
		    
		    for ( LCObjectVec::const_iterator pulseIter = (*hit)->getRawHits().begin() ;
			  pulseIter < (*hit)->getRawHits().end() ; pulseIter++ )
		    {
		      TrackerPulse const * pulse = dynamic_cast<TrackerPulse const *>(*pulseIter);
		      
		      if (pulse==0)
			break;

		      gear::Vector2D padCentre = localPadLayout.getPadCenter( pulse->getCellID0() );
		      // always in mm
		      double x = (localIsPolar?padCentre[1]*padCentre[0]:padCentre[0]);
		      streamlog_out(DEBUG3) << "x is " << x << std::endl;
		      double q =  pulse->getCharge();
		      streamlog_out(DEBUG3) << "q is " << q << std::endl;

		      sum_q    += q;
		      sum_qx   += q*x;
		      sum_qxx  += q*x*x;
		      sum_qxxx += q*x*x*x;
		    }

		    if (sum_q ==0)
		    {
		      streamlog_out(WARNING2) << "charge sum in hit is 0. check previous processors"
					      << std::endl;
		      continue;
		    }

		    double mu1prime = sum_qx / sum_q  ; // the first algebraic moment (the centre of gravity)
		    double mu2prime = sum_qxx / sum_q ; // the second algebraic moment
		    double mu3prime = sum_qxxx / sum_q; // the third algebraic moment
		    
		    double variance = mu2prime - mu1prime*mu1prime; // the second central moment (aka variance)
		    double mu3 = mu3prime - 3*mu2prime*mu1prime + 2*pow(mu1prime,3);// the third central moment

		    _residuals_vs_variance->fill( variance  , residuals[0]);

		    if ( variance > 0. )
		    {
		      double skewness = mu3 / pow( variance, 1.5 ); // the skewness

		      _residuals_vs_skewness->fill( skewness, residuals[0]);

		      // only use in the calculation if within the histogram range (reduce outliers)
		      // we want it in the histo for out of range bins, but not here
		      if ( ( residuals[0] < _residualMaxValue ) && 
			   ( residuals[0] > _residualMinValue ) &&
			   ( skewness < _skewnessMaxValue ) &&
			   ( skewness > _skewnessMinValue ) )
		      {
			_skewnessResidualsList.push_back( std::pair< double, double  >( skewness, residuals[0]) );
		      }
		    }
		    else
		    {
		      streamlog_out(WARNING0) << "variance in hit is "<< variance <<", hit size is "
					      << (*hit)->getRawHits().size()
					      << std::endl;
		    }

		}
		catch ( IntersectionCalculator::NoIntersectionException &e )
		{
		    cout << e.what() << endl;
		}
	
	    }
	    
	    trackNr++;
        }// input tracks collection for loop

    }
}

void PadResponseProcessor::end()
{
  // do a linear regression on all the entries
  // The straight line is y = a*x +b

  double a = 0;
  double b = 0;
  double sigma = 0; // sigma^2 is sum { (y_i - y( x_i) )^2 } / n
                    // = sum{ ( y_i - a*x_i -b )^2 } / n

  // do we have to do an iteration to reject outliers? -> YES!
  unsigned int numberOfIterations = 18;

  for (unsigned int i=0; i< numberOfIterations; i++)
  {
    // the parameters we need to calculate
    double sumX=0; double sumY=0; double sumXY=0;
    double sumXX=0;  // = Sum (X^2)
    double sumYY=0;  // = Sum (Y^2), needed for sigma calculation

    unsigned int n=_skewnessResidualsList.size();

    if (n>=2)
    {
      for ( std::list< std::pair< double, double  > >::iterator skewnessResidualsIter 
	      = _skewnessResidualsList.begin(); skewnessResidualsIter != _skewnessResidualsList.end();
	    skewnessResidualsIter++ )
      {
	streamlog_out(DEBUG2) << "skewness " << skewnessResidualsIter->first
			      << " ; residual " << skewnessResidualsIter->second << std::endl;
	sumX       += skewnessResidualsIter->first;
	sumY       += skewnessResidualsIter->second;
	sumXY      += skewnessResidualsIter->first  * skewnessResidualsIter->second;
	sumXX      += skewnessResidualsIter->first  * skewnessResidualsIter->first;
	sumYY      += skewnessResidualsIter->second  * skewnessResidualsIter->second;
      }
//      streamlog_out(DEBUG1) << "sumX: " << sumX << " ; " 
//			   << "sumY: " << sumY << " ; " 
//			   << "sumXY: " << sumXY << " ; "  
//			   << "sumXX: "<< sumXX << std::endl;


      // Now calculate a and b
      a  = (sumX*sumY-n*sumXY)/(sumX*sumX-n*sumXX);
      b  = (sumX*sumXY-sumXX*sumY)/(sumX*sumX-n*sumXX);

      // sum out the formula given above for sigma and you get this:
      sigma = sqrt ( ( sumYY - 2*a*sumXY - 2*b*sumY + a*a*sumXX + 2*a*b*sumX +b*b ) / n );
      streamlog_out(DEBUG2) << "sigma is " << sigma << std::endl;
  

      // Print the result. What else do we do with it? Write a contitions data object?
      streamlog_out(DEBUG1) << "Iteration "<<i<< ": residual skewness relation: residual = "
			    << a << "*skewness + "<<b<< std::endl;
      
      if ( i < numberOfIterations-1 ) // no need to run the removal in the last step
      {
	unsigned int nOutliers = 0;
	// remove outliers from vector
	for ( std::list< std::pair< double, double  > >::iterator skewnessResidualsIter 
		= _skewnessResidualsList.begin(); skewnessResidualsIter != _skewnessResidualsList.end();
	      /* do not increase iterator here */ )
	{
	  if ( fabs(skewnessResidualsIter->second - a * skewnessResidualsIter->first - b) > 5 * sigma )
	  {
	    // this is an outlier, remove it
	    // note: incrementation must be done in the erase call to be safe
	    _skewnessResidualsList.erase( skewnessResidualsIter++ );
	    nOutliers++;
	  }
	  else
	  {  
	    skewnessResidualsIter++;
	  }
	} // for skewnessResidualsIter

	streamlog_out(DEBUG1) << "removed "<< nOutliers << " outliers"<< endl;
      }
    }// if n > 2
    else
    {
      streamlog_out(WARNING) << "Could not calculate residual skewness relation, number of hits is only "
	   << n << std::endl;
      break;
    }
  }// for i
  
  cout  << "Final residual skewness relation: residual = "
			<< a << "*skewness + "<<b<< std::endl;


}// processor::end



}// namespace 
