#include "PixelSpectrumProcessor.h"
#include "TimePixMapHandlerProcessor.h"


using namespace lcio ;
using namespace marlin ;
using namespace tpcconddata;

namespace marlintpc{


PixelSpectrumProcessor aPixelSpectrumProcessor ;
    
PixelSpectrumProcessor::PixelSpectrumProcessor():Processor("PixelSpectrumProcessor")
{
  /* the constructor call first the super constructor which the name 
   * of the processor
   * this must by the same as the class name.
   */
	
  /* description of the processor which will be displayed in the
   * steering file autometicly generated by 
   * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
   */
  
  _description = "PixelSpectrumProcessor creates a histgram for TOT and TIME over all events" ;	
	
  /* register steering parameters: name, description, class-variable, default value
   * the type will by definded by the type of the default value
   * string, double, float and int are possible
   * please note to get a float you must cast in C++: (float) 1.3 
   */
	
  registerInputCollection( LCIO::TRACKERDATA,
    "InputCollectionName" , 
    "Name of the input collection"  ,
    _inputColName ,
    std::string("TimePixZeroSuppressedData") ) ;

  registerProcessorParameter( "Mode" , 
    "mode that should be plotted (TOT, TIMEPIX, MEDIPIX, ONEHIT or ANY), default: ANY"  ,
    _mode ,
    std::string("ANY") ) ;
	
  registerProcessorParameter( "Minclock" ,
    "MinZ Minimum z value of lowest bin in case of z binning (default: 2500)",
    _min_clock,
    float(0) );
	
  registerProcessorParameter( "Maxclock" ,
    "MaxZ Maximum z value of highest bin in case of z binning (default: 4000)",
    _max_clock,
    float(4000) ); 

  registerProcessorParameter( "nBins" ,
    "Number ob bins in z, 0 means one Histo without z cut (default: 100)",
    _nBins,
    int(100) );
    
  registerProcessorParameter( "LogFile" , 
    "Position of logfile"  ,
    _logFile ,
    std::string("") ) ;				    
    
  registerProcessorParameter( "HighTH" ,
    "High threshold from wich pixel are collected in _myBadPixelLocation2dHisto",
    _HighTH,
    int(1800) );
			
  registerProcessorParameter( "LowTH" ,
    "Low threshold from wich pixel are collected in _myBadPixelLocation2dHisto",
    _LowTH,
    int(285) );		
				    
  registerProcessorParameter( "LowerBadCountsLimit" ,
    "Minimal Number of Pixels with Bad Counts to create Event Display",
    _lowerBadCountsLimit,
    int(0) );			    
				    
  _timePixMap = 0;
}
    
    
void PixelSpectrumProcessor::init()
{ 
  // usually a good idea to
  printParameters() ;
  _timePixMap = TimePixMapHandlerProcessor::getTimePixMap();

  // CR: not used anywhere, commented out to suppress warnings during compilation
  //~ float _zBinWidth;
  //~ if(_nBins>0) _zBinWidth = (_max_clock - _min_clock)/_nBins;

  std::string histoTitle = "Clock";
  std::string histoName  = "Clock";

  _mySpectrumHisto = AIDAProcessor::histogramFactory(this)->createHistogram1D(
	   histoName, histoTitle, (int)_max_clock, _min_clock-0.5, _max_clock-0.5);
	
  int myXBins = 256;
  double myMinX = -0.5;
  double myMaxX = 255.5; 
  int myYBins = 256;
  double myMinY = -0.5;
  double myMaxY = 255.5;

  _myPixelAboveTH_Location2dHisto = AIDAProcessor::histogramFactory(this)->createHistogram2D(
    "BadPixelLocationAboveTH","Positons of pixels with counts above TH",
     myXBins, myMinX, myMaxX,
     myYBins, myMinY, myMaxY);
  _myPixelBelowTH_Location2dHisto = AIDAProcessor::histogramFactory(this)->createHistogram2D(
    "BadPixelLocationBelowTH","Positons of pixels with counts below TH",
    myXBins, myMinX, myMaxX,
    myYBins, myMinY, myMaxY);					    
}

    
void PixelSpectrumProcessor::processRunHeader(lcio::LCRunHeader* run) 
{ 
  run->parameters().setValue( _processorName + "_revision", "$Rev$");

  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) 
  {
    if ( ! i->second->isOptional() || i->second->valueSet() )
    {
      run->parameters().setValue( _processorName + "_" + i->second->name(),i->second->value() );
    }
  }
}
    

void PixelSpectrumProcessor::processEvent(lcio::LCEvent * evt )
{ 
  streamlog_out(DEBUG1)<<"processEvent( "<<evt->getEventNumber()<<" )"<<std::endl;

  LCCollection* TrackerDataCol;
  try
  {
    TrackerDataCol = evt->getCollection( _inputColName );
  }
  catch(lcio::DataNotAvailableException)
  {
    streamlog_out(DEBUG3) << "Empty event: no zero suppressed TimePix data in event " 
      <<evt->getEventNumber()<<std::endl;
    return;
  }

	
  // Use the TimePix Modemap to define plotmode
  int plotMode;

  if ( _mode == "TOT" ) plotMode = TimePixPixelMode::TOT;
  else if ( _mode == "TIMEPIX" ) plotMode = TimePixPixelMode::TIMEPIX;
  else if ( _mode == "MEDIPIX" ) plotMode = TimePixPixelMode::MEDIPIX;
  else if ( _mode == "ONEHIT") plotMode = TimePixPixelMode::ONEHIT;
  else plotMode = -1; // plot all modes
	
  streamlog_out(DEBUG4)<<"plotmode =  "<<plotMode<<std::endl;

  streamlog_out(DEBUG4)<<"loop over "<<TrackerDataCol->getNumberOfElements()
    <<" Elements in inputCol"<<std::endl;

  // create log file containing eventNumbers of high events		     
  std::ofstream logFile;
  if ( _logFile != "") 
  {
    logFile.open( _logFile.c_str(), std::ios::out | std::ios::app );
  }
	
  int HighCountsFlag = 0;
  int LowCountsFlag = 0;
	
  //Now loop over the modules
  for( int i = 0; i < TrackerDataCol->getNumberOfElements(); i++ )
  {
    streamlog_out(DEBUG4)<<i<<" . "<<std::flush;
    TrackerData* Data = 
      dynamic_cast<TrackerData*>( TrackerDataCol->getElementAt( i ) ) ;
	    
    if ( Data != 0 ) 
    {
      int pixelID = Data->getCellID0();
      int chipID = Data->getCellID1();

	  	
//    const gear::TPCModule &module = marlin::Global::GEAR->getTPCParameters().getModule(chipID);

      FloatVec pixelContentVector  = Data->getChargeValues();
		
      for ( FloatVec::iterator pixelContentIter = pixelContentVector.begin(); 
        pixelContentIter != pixelContentVector.end(); pixelContentIter++ ) 
      {

        //The following part sets the correct mode for each pixel from the modemap
		    
        TimePixMap::const_iterator modeIter = _timePixMap->find( TimePixPixelMode::PixelKey( chipID, pixelID ) );
		    
        if ( modeIter == _timePixMap->end() )
        {
          // The pixel was not found in the mode map
	  // Give an error message and throw an exception
	  streamlog_out(ERROR) << "Pixel key (chipID "<< chipID 
	    << ", pixelID "<< pixelID 
	    << " could not be found in Timepix mode map!"
	    << std::endl;
			
	  throw lcio::DataNotAvailableException("Pixel key not found in mode map");
        }
		    
        int mode = (modeIter->second)->getPixelMode();
		   
        if( ( ( mode == plotMode ) || ( plotMode == -1 ) ) && (*pixelContentIter != 0.) )
        {	
          _mySpectrumHisto->fill(*pixelContentIter); 
	  //std::cout << "myHistoFill:\t" << *pixelContentIter << std::endl;
			
	  // begining to fill _myBadPixelLocation2dHisto
	  if   (*pixelContentIter > _HighTH) 
	  {
            _myPixelAboveTH_Location2dHisto->fill((pixelID % 256), (pixelID / 256)); 
	  
            if ( _logFile != "") 
            {
              logFile << "High Pixel found in " <<  evt->getEventNumber() << std::endl;   
            }
	    HighCountsFlag++; 
	  }
	  if  (*pixelContentIter < _LowTH)  
	  {
            _myPixelBelowTH_Location2dHisto->fill((pixelID % 256), (pixelID / 256));			
	  
            if ( _logFile != "") 
            { 
              logFile << "Low Pixel found in " <<  evt->getEventNumber() << std::endl;   
            }
	    LowCountsFlag++;
	  }
        }

      pixelID++;
      }
	       
		
    } //if ( Data != 0 )
  }//for( int i = 0; i < TrackerDataCol ... - now loop over the modules
 
  if ( _logFile != "") logFile.close();
  streamlog_out(DEBUG4)<<std::endl;

	
  //  create Eventdisplay for  events with bad counts
  if  ((HighCountsFlag > _lowerBadCountsLimit) || (LowCountsFlag > _lowerBadCountsLimit) )
  {
    AIDAProcessor::tree(this);	//changes into directory of this processor in aida-output-file 
    TH2F *EventDisplayBadCountsHisto;
	  
    // build HistoName
    std::stringstream EventNumberString;
    EventNumberString << evt->getEventNumber();
    std::stringstream HighCountsFlagString;
    std::stringstream LowCountsFlagString;
    HighCountsFlagString << HighCountsFlag;
    LowCountsFlagString << LowCountsFlag;
	  
    std::string HistoName;
    if  (HighCountsFlag != 0) 
      HistoName = "HighEvent" + EventNumberString.str() + "HighCounts" + HighCountsFlagString.str();
    if (LowCountsFlag != 0)
      HistoName = "LowEvent" + EventNumberString.str() + "LowCounts" + LowCountsFlagString.str();
    if ( (HighCountsFlag != 0) && (LowCountsFlag != 0) )
      HistoName = "HighAndLowEvent" + EventNumberString.str()  + "HighCounts" 
      + HighCountsFlagString.str() + "LowCounts" + LowCountsFlagString.str();
	  
    EventDisplayBadCountsHisto = new TH2F(HistoName.c_str(), HistoName.c_str(), 
      256, -0.5, 255.5, 
      256, -0.5, 255.5);
	  
    // begining to fill EventDisplayBadCountsHisto Histo  
    for( int i = 0; i < TrackerDataCol->getNumberOfElements(); i++ )
    { 
      TrackerData* Data = dynamic_cast<TrackerData*>( TrackerDataCol->getElementAt( i ) ) ;
      if ( Data != 0 ) 
      {
        int pixelID = Data->getCellID0();
        int chipID = Data->getCellID1();
	    
        FloatVec pixelContentVector  = Data->getChargeValues();	  
	      
        for ( FloatVec::iterator pixelContentIter = pixelContentVector.begin(); 
          pixelContentIter != pixelContentVector.end(); pixelContentIter++ ) 
	{ 
          TimePixMap::const_iterator modeIter = _timePixMap->find( 
            TimePixPixelMode::PixelKey( chipID, pixelID ) );   
	  if ( modeIter == _timePixMap->end() )
	  { 
            // The pixel was not found in the mode map
	    // Give an error message and throw an exception
	    streamlog_out(ERROR) << "Pixel key (chipID "<< chipID 
	      << ", pixelID "<< pixelID 
	      << " could not be found in Timepix mode map!"
	      << std::endl;
	 
            throw lcio::DataNotAvailableException("Pixel key not found in mode map");
	  }
		    
	  int mode = (modeIter->second)->getPixelMode();
		  	    
	  if ( ( ( mode == plotMode ) || ( plotMode == -1 ) ) && ( *pixelContentIter != 0.) )
	  { 
            EventDisplayBadCountsHisto->SetBinContent((pixelID % 256), (pixelID / 256), *pixelContentIter);
          }
	  pixelID++;
	}
      }	    
    }
    EventDisplayBadCountsHisto->Write();
    delete EventDisplayBadCountsHisto;
    HighCountsFlag = 0;  
    LowCountsFlag = 0;
  }
	
  streamlog_out(DEBUG4)<<"end of processEvent( "<<evt->getEventNumber()<<" )"<<std::endl;
}
  
 
void PixelSpectrumProcessor::check(lcio::LCEvent *  /*evt*/ ) 
{ 
}
    
void PixelSpectrumProcessor::end()
{ 
 /*	
  streamlog_out(MESSAGE4) << "PixelSpectrumProcessor::end()  " << name() 
    << " processed " << evt->getEventNumber() 
    << " events in " << _nRun << " runs "
    << std::endl ;
 */
}
      
    
    
}
