#include "PulsePerHitProcessor.h"

#include "marlin/Global.h"

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include <Exceptions.h>

// MarlinTPC
#include "FlagwordDefinitions.h"

// GEAR
#include <gear/TPCModule.h>

// ROOT
#include "TH1.h"
#include "TH1D.h"
#include "TNtuple.h"
#include "TF1.h"
#include "TGraphErrors.h"

using namespace lcio;
using namespace marlin;
using namespace std;


namespace marlintpc
{

PulsePerHitProcessor aPulsePerHitProcessor;


PulsePerHitProcessor::PulsePerHitProcessor() : Processor("PulsePerHitProcessor")
{
	// the processor description
	_description = "PulsePerHitProcessor creates histograms for the number of pulses per hit and average pulse shapes for the maximum pulse in a hit and for its neighbors. It also creates a graph for the number of pulses per hit vs the row of the hit.";
	
	/* register steering parameters: name, description, class-variable, default value
	*/

	registerInputCollection( LCIO::TRACK,
			"InputCollectionName",
			"Name of the input Track collection, default: TPCTracks",
			_inputColName,
			string("TPCTracks") );

	registerProcessorParameter( "PulsesPerHitHistoMin",
			"Minimum of histogram of the pulses per hit, default: 0.",
			_pulsesPerHitHistoMin,
			float(0.) );

	registerProcessorParameter( "PulsesPerHitHistoMax",
			"Maximum of histogram of the pulses per hit, default: 20.",
			_pulsesPerHitHistoMax,
			float(20.) );

}


void PulsePerHitProcessor::init()
{ 
	streamlog_out(DEBUG) << " init() called	" << endl;
	
	// you might want to list the used parameters by invoking
	printParameters();
	AIDAProcessor::tree(this);
	_gearParameters = &marlin::Global::GEAR->getTPCParameters();

	_pulsesPerHitHisto = new TH1D( "PulsesPerHitHisto"
		,"Pulses per Hit;# pulses per hit;# events"
		,_pulsesPerHitHistoMax-_pulsesPerHitHistoMin, _pulsesPerHitHistoMin, _pulsesPerHitHistoMax );


}


void PulsePerHitProcessor::processRunHeader(LCRunHeader*  /*run*/)
{
	// do something for each run
} 


void PulsePerHitProcessor::processEvent(LCEvent* evt)
{ 
	AIDAProcessor::tree(this);

	// use an LCCollectionVec instead of LCCollection so we can use iterators later
	LCCollectionVec* inputColVec = 0;

	// do a try / catch in case there is an empty event
	try
	{
		inputColVec = dynamic_cast<LCCollectionVec *>( evt->getCollection(_inputColName) );
	}
	catch (DataNotAvailableException &e)
	{
		// no input data, just return and process the next event
		streamlog_out(DEBUG) << "No data available in event for " << _inputColName << endl;
		return;
	}

	// loop over all tracks in the collection
	for( LCCollectionVec::iterator inputObjectIter = inputColVec->begin() ; inputObjectIter < inputColVec->end() ; ++inputObjectIter )
	{
		// dynamic cast the input object (LCObject) to what it really is e.g.
		lcio::Track* inputTrack = dynamic_cast<lcio::Track*>( *inputObjectIter );
		lcio::TrackerHitVec hitVec = inputTrack->getTrackerHits();
		streamlog_out(DEBUG4) << "number of hits " << hitVec.size() << endl;

		//loop over hits
		for ( lcio::TrackerHitVec::iterator hitIter = hitVec.begin() ; hitIter < hitVec.end() ; ++hitIter )
		{
			lcio::TrackerHit* myHit = *hitIter;
			//find module and row of the hit
			const double* hitPosition = myHit->getPosition();
			const gear::GlobalPadIndex nearestPadIndex = _gearParameters->getNearestPad( hitPosition[0], hitPosition[1] );
			const gear::TPCModule& myModule = _gearParameters->getModule( nearestPadIndex.getModuleID() );
			const int module = myModule.getModuleID();
			const int row = myModule.getRowNumber( nearestPadIndex.getPadIndex() );
			std::pair<int, int> key(module,row);
			streamlog_out(DEBUG3) << "hitPosition[0] " << hitPosition[0] << "\thitPosition[1] " << hitPosition[1] << std::endl;
			streamlog_out(DEBUG3) << "module: " << module << "\trow: " << row << std::endl;

			std::map<std::pair<int,int>, TH1*>::iterator histoIter = _pulsesPerHitPerRowHistos.find(key);

			//get the number of pulses
			LCObjectVec pulseVec = myHit->getRawHits();
			const int numberOfPulses = pulseVec.size();
			streamlog_out(DEBUG3) << "number of pulses: " << numberOfPulses << std::endl;
			_pulsesPerHitHisto->Fill(numberOfPulses);

			//fill histograms for each module and row
			if(histoIter != _pulsesPerHitPerRowHistos.end())
			{
				((*histoIter).second)->Fill(numberOfPulses);
			}
			else
			{
				TH1* tempHist = new TH1D( Form("PulsesPerHitHisto_%i_%i",module,row),
					"PulsesPerHit;# pulses per hit;# events",
					_pulsesPerHitHistoMax-_pulsesPerHitHistoMin,
					_pulsesPerHitHistoMin, _pulsesPerHitHistoMax );
				tempHist->Fill(numberOfPulses);
				_pulsesPerHitPerRowHistos.insert( std::make_pair(key,tempHist) );
			}

		}//end loop over hits
	}//end loop over tracks
}


void PulsePerHitProcessor::check(LCEvent*  /*evt*/)
{ 
	// check what is needed here 
}


void PulsePerHitProcessor::end()
{ 
	AIDAProcessor::tree(this);

	TF1* fit = new TF1("fit", "gaus", 2, _pulsesPerHitHistoMax);
	_pulsesPerHitHisto->Fit("fit","QRN");

	TNtuple* pulsesPerHitTuple = new TNtuple("PulsesPerHitStats","PulsesPerHitStats","mean:meanError:rms:rmsError:meanRangeMin:meanRangeMax:nEntries");
	pulsesPerHitTuple->Fill( fit->GetParameter(1), fit->GetParError(1), fit->GetParameter(2), fit->GetParError(2), 2, _pulsesPerHitHistoMax, _pulsesPerHitHisto->Integral(2, _pulsesPerHitHistoMax) );

	TNtuple* pulsesPerHitVSRowTuple = new TNtuple("PulsesPerHitRowStats","PulsesPerHitRowStats","module:row:mean:meanError:rms:rmsError:rangeMin:rangeMax:nEntries");
	lcio::FloatVec vecMean, vecMeanErr, vecRMS, vecRMSErr, vecRadius, vecRadiusErr;
	for( std::map< std::pair<int,int>, TH1*>::iterator histoIter = _pulsesPerHitPerRowHistos.begin() ; histoIter != _pulsesPerHitPerRowHistos.end() ; ++histoIter )
	{
		TH1D* tempHisto = dynamic_cast<TH1D*>( (*histoIter).second );
		int module = ((*histoIter).first).first;
		int row = ((*histoIter).first).second;
		const gear::TPCModule& theModule = _gearParameters->getModule(module);
		gear::Vector2D padcenter = theModule.getPadCenter( theModule.getPadIndex(row,0) );
		vecRadius.push_back( sqrt(padcenter[0]*padcenter[0]+padcenter[1]*padcenter[1]) );
		vecRadiusErr.push_back( (theModule.getRowHeight(row))/sqrt(12.) );
		streamlog_out(DEBUG3) << "module: " << module << "\trow: " << row << std::endl;
		tempHisto->Fit("fit","QR");
		pulsesPerHitVSRowTuple->Fill( module, row, fit->GetParameter(1), fit->GetParError(1), fit->GetParameter(2), fit->GetParError(2), 2, _pulsesPerHitHistoMax, tempHisto->Integral(2, _pulsesPerHitHistoMax) );
		vecMean.push_back(fit->GetParameter(1));
		vecMeanErr.push_back(fit->GetParError(1));
		vecRMS.push_back(fit->GetParameter(2));
		vecRMSErr.push_back(fit->GetParError(2));

		delete tempHisto;
	}

	TGraphErrors* pulsesPerHitVSRowGraph = new TGraphErrors(vecRadius.size(),&vecRadius[0],&vecMean[0],&vecRadiusErr[0],&vecMeanErr[0]);
	pulsesPerHitVSRowGraph->SetNameTitle("PulsesPerHitVSRowGraph","PulsesPerHitVSRowGraph;r_{row} [mm];# pulses per hit");
	pulsesPerHitVSRowGraph->Write();

	delete fit;

}
} // close namespace brackets
