#include "PulseShapeProcessor.h"

#include "marlin/Global.h"

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include <EVENT/TrackerPulse.h>
#include <EVENT/TrackerData.h>
#include <Exceptions.h>

// MarlinTPC
#include "FlagwordDefinitions.h"

// ROOT
#include "TH2D.h"
#include "TNtuple.h"
#include "TProfile.h"
#include "TF1.h"

using namespace lcio;
using namespace marlin;
using namespace std;


namespace marlintpc
{

PulseShapeProcessor aPulseShapeProcessor;


PulseShapeProcessor::PulseShapeProcessor() : Processor( "PulseShapeProcessor" )
{
	// the processor description
	_description = "PulseShapeProcessor creates average pulse shapes for the maximum pulse in a hit and for its neighbors.";

	/* register steering parameters: name, description, class-variable, default value
	*/

	registerInputCollection( LCIO::TRACK,
			"InputCollectionName",
			"Name of the input Track collection, default: TPCTracks",
			_inputColName,
			string( "TPCTracks" ) );

	registerProcessorParameter( "NormPulseHistoMax",
			"Maximum relative pulse height value of the normalized pulse histograms",
			_normPulseHistoMax,
			float( 1. ) );

	registerProcessorParameter( "NeighborsUsed",
			"Number of Neighbors used to create norm pulse ( set to 0 to use all ), default, 1",
			_nNeighborsUsed,
			int( 1 ) );

	registerProcessorParameter( "UsePolynomialInFit",
			"Determines if the polynomial is used in the gamma4 fit to approximate the underground, default: true",
			_usePolynomialInFit,
			bool( true ) );

	registerProcessorParameter( "UseOnlyHighQualityPulses",
			"If true, only high quality pulses ( getQuality()=0 ) are used for the average pulse shape, set to false to use all pulses, default: true",
			_useOnlyHighQualityPulses,
			bool( true ) );
}


void PulseShapeProcessor::init()
{
	streamlog_out( DEBUG ) << "init() called" << endl;

	// you might want to list the used parameters by invoking
	printParameters();
	AIDAProcessor::tree( this );

	_pulseHighestHisto2D = new TH2D( "PulseHighestHisto2D", "Pulse of Pads with highest Charge", 30, 0, 30, 1024, 0, 1024 );
	_pulseHighestHisto2D->SetXTitle( "time bin" );
	_pulseHighestHisto2D->SetYTitle( "pulse height [ADC counts]" );

	_pulseNeighborsHisto2D = new TH2D( "PulseNeighborsHisto2D", "Pulse of neighboring Pads", 30, 0, 30, 1024, 0, 1024 );
	_pulseNeighborsHisto2D->SetXTitle( "time bin" );
	_pulseNeighborsHisto2D->SetYTitle( "pulse height [ADC counts]" );

	_normPulseHighestHisto2D = new TH2D( "NormPulseHighestHisto2D", "Norm Pulse of Pads with highest Charge", 30, 0, 30, 1024, 0, _normPulseHistoMax );
	_normPulseHighestHisto2D->SetXTitle( "time bin" );
	_normPulseHighestHisto2D->SetYTitle( "relative pulse height [a.u.]" );

	_normPulseNeighborsHisto2D = new TH2D( "NormPulseNeighborsHisto2D", "Norm Pulse of neighboring Pads", 30, 0, 30, 1024, 0, _normPulseHistoMax );
	_normPulseNeighborsHisto2D->SetXTitle( "time bin" );
	_normPulseNeighborsHisto2D->SetYTitle( "relative pulse height [a.u.]" );

}


void PulseShapeProcessor::processRunHeader( LCRunHeader*  /*run*/ )
{
	// do something for each run
}


void PulseShapeProcessor::processEvent( LCEvent* evt )
{
	AIDAProcessor::tree( this );

	// use an LCCollectionVec instead of LCCollection so we can use iterators later
	LCCollectionVec* inputColVec = 0;

	// do a try / catch in case there is an empty event
	try
	{
		inputColVec = dynamic_cast<LCCollectionVec *>( evt->getCollection( _inputColName ) );
	}
	catch ( DataNotAvailableException &e )
	{
		// no input data, just return and process the next event
		streamlog_out( DEBUG ) << "No data available in event for " << _inputColName << endl;
		return;
	}

	// loop over all tracks in the collection
	for( LCCollectionVec::iterator inputObjectIter = inputColVec->begin() ; inputObjectIter < inputColVec->end() ; ++inputObjectIter )
	{
		// dynamic cast the input object ( LCObject ) to what it really is e.g.
		lcio::Track* inputTrack = dynamic_cast<lcio::Track*>( *inputObjectIter );
		lcio::TrackerHitVec hitVec = inputTrack->getTrackerHits();
		streamlog_out( DEBUG4 ) << "number of hits " << hitVec.size() << endl;

		//loop over hits
		for ( lcio::TrackerHitVec::iterator hitIter = hitVec.begin() ; hitIter < hitVec.end() ; ++hitIter )
		{
			lcio::TrackerHit* myHit = *hitIter;
			//get the pulses
			LCObjectVec pulseVec = myHit->getRawHits();
			const int numberOfPulses = pulseVec.size();
			streamlog_out( DEBUG3 ) << "number of pulses: " << numberOfPulses << std::endl;

			//loop over the pulses in the hit find the highest pulse and get the charge of each pulse
			float highestPulseCharge = 0;
			const lcio::TrackerPulse* highestPulse = 0;
			for( LCObjectVec::const_iterator pulseIter = pulseVec.begin(); pulseIter < pulseVec.end(); ++pulseIter )
			{
				//dynamic cast the pulses to their real data type
				const lcio::TrackerPulse* myPulse = dynamic_cast<const lcio::TrackerPulse*>( *pulseIter );
				if( myPulse->getCharge() > highestPulseCharge )
				{
					highestPulseCharge = myPulse->getCharge();
					highestPulse = myPulse;
				}
			}

			//fill profiles for average pulses
			for ( LCObjectVec::const_iterator pulseIter = pulseVec.begin(); pulseIter < pulseVec.end(); ++pulseIter )
			{
				const lcio::TrackerPulse* myPulse = dynamic_cast<const lcio::TrackerPulse*>( *pulseIter );
				//only use high quality pulses except specified otherwise
				if( _useOnlyHighQualityPulses && myPulse->getQuality()!=0 )
				{
					streamlog_out( DEBUG3 ) << "pulse quality = " << myPulse->getQuality() << " ; pulse is not used for average pulseshape" << std::endl;
					continue;
				}

				//check if the pulse is the highest one or in the range of used neighbors
				const int distanceToHighest = abs( myPulse->getCellID0()-highestPulse->getCellID0() );
				TH2* tempHisto2D = NULL;
				TH2* tempNormHisto2D = NULL;
				if( myPulse==highestPulse )
				{
					tempHisto2D = _pulseHighestHisto2D;
					tempNormHisto2D = _normPulseHighestHisto2D;
				}
				else if( distanceToHighest<=_nNeighborsUsed || _nNeighborsUsed==0 )
				{
					tempHisto2D = _pulseNeighborsHisto2D;
					tempNormHisto2D = _normPulseNeighborsHisto2D;
				}

				if( tempHisto2D==NULL || tempNormHisto2D==NULL ) continue;

				//fill pulse shape in scatterplot
				lcio::FloatVec pulseChargeValues = myPulse->getTrackerData()->getChargeValues();
				for( unsigned short timeBin=0 ; timeBin<pulseChargeValues.size() ; ++timeBin )
				{
					tempHisto2D->Fill( timeBin, pulseChargeValues.at( timeBin ) );
					tempNormHisto2D->Fill( timeBin, pulseChargeValues.at( timeBin ) / myPulse->getCharge() );
				}
			}

			//determine charge interval of the highest pulse and fill the respective histogram
			bool pulseOK ( !pulseflag::isMultiplePulseCandidate( highestPulse->getQuality() )
					&& !pulseflag::isAnomalousShape( highestPulse->getQuality() )
					&& !pulseflag::isUnderflow( highestPulse->getQuality() )
					&& !pulseflag::isSplit( highestPulse->getQuality() )
					&& !pulseflag::isPlateauCutoff( highestPulse->getQuality() ) );
			if( !_useOnlyHighQualityPulses || pulseOK )
			{
				int chargeBin;
				double lgQ = log10( highestPulseCharge );
				if( lgQ>=4 ) chargeBin = 40;
				else if( lgQ<=2 ) chargeBin = 1;
				else chargeBin = static_cast<int>( floor( ( lgQ - 2 ) / 2 * 40 + 1 ) );
				std::map<int, TH2*>::iterator chargeIter = _pulseHighestHistoVSCharge.find( chargeBin );
				if( chargeIter==_pulseHighestHistoVSCharge.end() )
				{
					TH2* tempHisto2D = new TH2D( Form( "PulseHighestHisto2D_Charge%d", chargeBin ), "Pulse of Pads with highest Charge", 30, 0, 30, 1024, 0, 1024 );
					tempHisto2D->SetXTitle( "time bin" );
					tempHisto2D->SetYTitle( "pulse height [ADC counts]" );
					lcio::FloatVec pulseChargeValues = highestPulse->getTrackerData()->getChargeValues();
					for( unsigned int timeBin=0 ; timeBin<pulseChargeValues.size() ; ++timeBin )
					{
						tempHisto2D->Fill( timeBin, pulseChargeValues.at( timeBin ) );
					}
					_pulseHighestHistoVSCharge.insert( make_pair( chargeBin, tempHisto2D ) );
				}
				else
				{
					TH2* tempHisto2D = ( *chargeIter ).second;
					lcio::FloatVec pulseChargeValues = highestPulse->getTrackerData()->getChargeValues();
					for( unsigned int timeBin=0 ; timeBin<pulseChargeValues.size() ; ++timeBin )
					{
						tempHisto2D->Fill( timeBin, pulseChargeValues.at( timeBin ) );
					}
				}
			}
		}//end loop over hits
	}//end loop over tracks
}


void PulseShapeProcessor::check( LCEvent*  /*evt*/ )
{
	// check what is needed here
}


void PulseShapeProcessor::end()
{
	AIDAProcessor::tree( this );

	//make profiles from the scatterplots
	TProfile* pulseHighestProfile = _pulseHighestHisto2D->ProfileX( "pulseHighestProfile" );
	pulseHighestProfile->SetTitle( "pulseHighestProfile" );
	pulseHighestProfile->SetXTitle( "time bin" );
	pulseHighestProfile->SetYTitle( "pulse height [ADC counts]" );

	TProfile* pulseNeighborsProfile = _pulseNeighborsHisto2D->ProfileX( "pulseNeighborsProfile" );
	pulseNeighborsProfile->SetTitle( "pulseNeighborsProfile" );
	pulseNeighborsProfile->SetXTitle( "time bin" );
	pulseNeighborsProfile->SetYTitle( "pulse height [ADC counts]" );

	TProfile* normPulseHighestProfile = _normPulseHighestHisto2D->ProfileX( "normPulseHighestProfile" );
	normPulseHighestProfile->SetTitle( "normPulseHighestProfile" );
	normPulseHighestProfile->SetYTitle( "relative pulse height [a.u.]" );

	TProfile* normPulseNeighborsProfile = _normPulseNeighborsHisto2D->ProfileX( "normPulseNeighborsProfile" );
	normPulseNeighborsProfile->SetTitle( "normPulseNeighborsProfile" );
	normPulseNeighborsProfile->SetXTitle( "time bin" );
	normPulseNeighborsProfile->SetYTitle( "relative pulse height [a.u.]" );

	TNtuple* normPulseHighestFitTuple = new TNtuple( "NormPulseHighestFitParameters", "NormPulseHighestFitParameters", "par0:err0:par1:err1:par2:err2:par3:err3:par4:err4:par5:err5" );
	TNtuple* normPulseNeighborsFitTuple = new TNtuple( "NormPulseNeighborsFitParameters", "NormPulseNeighborsFitParameters", "par0:err0:par1:err1:par2:err2:par3:err3:par4:err4:par5:err5" );
	TNtuple* pulseHighestFitTuple = new TNtuple( "PulseHighestFitParameters", "PulseHighestFitParameters", "par0:err0:par1:err1:par2:err2:par3:err3:par4:err4:par5:err5" );
	TNtuple* pulseNeighborsFitTuple = new TNtuple( "PulseNeighborsFitParameters", "PulseNeighborsFitParameters", "par0:err0:par1:err1:par2:err2:par3:err3:par4:err4:par5:err5" );
	TNtuple* pulseHighestVSChargeFitTuple = new TNtuple( "PulseHighestVSChargeFitParameters", "PulseHighestVSChargeFitParameters", "par0:err0:par1:err1:par2:err2:par3:err3:par4:err4:par5:err5" );
	TNtuple* pulseHighestVSChargeHistFitTuple = new TNtuple( "PulseHighestVSChargeHistFitParameters", "PulseHighestVSChargeHistFitParameters", "par0:err0:par1:err1:par2:err2:par3:err3:par4:err4:par5:err5" );

	TF1* gamma4 = new TF1( "gamma4", "[0]*pol1(4) + ( [0]*exp([1]) * ( (x-[2])/[3] )**[1] * exp( -[1]*((x-[2])/[3]) ) ) * (x>[2])", 0, 25 );
	gamma4->FixParameter( 1, 4 );
	if( !_usePolynomialInFit )
	{
		gamma4->FixParameter( 4, 0 );
		gamma4->FixParameter( 5, 0 );
	}

	gamma4->SetParameters( .2, 4, 3, 2, 0, 0 );
	normPulseHighestProfile->Fit( "gamma4", "RQ" );
	normPulseHighestFitTuple->Fill( gamma4->GetParameter( 0 ), gamma4->GetParError( 0 ),
					gamma4->GetParameter( 1 ), gamma4->GetParError( 1 ),
					gamma4->GetParameter( 2 ), gamma4->GetParError( 2 ),
					gamma4->GetParameter( 3 ), gamma4->GetParError( 3 ),
					gamma4->GetParameter( 4 ), gamma4->GetParError( 4 ),
					gamma4->GetParameter( 5 ), gamma4->GetParError( 5 ) );
	gamma4->SetParameters( .2, 4, 3, 2, 0, 0 );
	normPulseNeighborsProfile->Fit( "gamma4", "RQ" );
	normPulseNeighborsFitTuple->Fill( gamma4->GetParameter( 0 ), gamma4->GetParError( 0 ),
					gamma4->GetParameter( 1 ), gamma4->GetParError( 1 ),
					gamma4->GetParameter( 2 ), gamma4->GetParError( 2 ),
					gamma4->GetParameter( 3 ), gamma4->GetParError( 3 ),
					gamma4->GetParameter( 4 ), gamma4->GetParError( 4 ),
					gamma4->GetParameter( 5 ), gamma4->GetParError( 5 ) );
	gamma4->SetParameters( 150, 4, 3, 2, 0, 0 );
	pulseHighestProfile->Fit( "gamma4", "RQ" );
	pulseHighestFitTuple->Fill( gamma4->GetParameter( 0 ), gamma4->GetParError( 0 ),
					gamma4->GetParameter( 1 ), gamma4->GetParError( 1 ),
					gamma4->GetParameter( 2 ), gamma4->GetParError( 2 ),
					gamma4->GetParameter( 3 ), gamma4->GetParError( 3 ),
					gamma4->GetParameter( 4 ), gamma4->GetParError( 4 ),
					gamma4->GetParameter( 5 ), gamma4->GetParError( 5 ) );
	gamma4->SetParameters( 100, 4, 3, 2, 0, 0 );
	pulseNeighborsProfile->Fit( "gamma4", "RQ" );
	pulseNeighborsFitTuple->Fill( gamma4->GetParameter( 0 ), gamma4->GetParError( 0 ),
					gamma4->GetParameter( 1 ), gamma4->GetParError( 1 ),
					gamma4->GetParameter( 2 ), gamma4->GetParError( 2 ),
					gamma4->GetParameter( 3 ), gamma4->GetParError( 3 ),
					gamma4->GetParameter( 4 ), gamma4->GetParError( 4 ),
					gamma4->GetParameter( 5 ), gamma4->GetParError( 5 ) );

	gamma4->SetParameters( .2, 4, 3, 2, 0, 0 );
	_normPulseHighestHisto2D->Fit( "gamma4", "RQ" );
	normPulseHighestFitTuple->Fill( gamma4->GetParameter( 0 ), gamma4->GetParError( 0 ),
					gamma4->GetParameter( 1 ), gamma4->GetParError( 1 ),
					gamma4->GetParameter( 2 ), gamma4->GetParError( 2 ),
					gamma4->GetParameter( 3 ), gamma4->GetParError( 3 ),
					gamma4->GetParameter( 4 ), gamma4->GetParError( 4 ),
					gamma4->GetParameter( 5 ), gamma4->GetParError( 5 ) );
	gamma4->SetParameters( .2, 4, 3, 2, 0, 0 );
	_normPulseNeighborsHisto2D->Fit( "gamma4", "RQ" );
	normPulseNeighborsFitTuple->Fill( gamma4->GetParameter( 0 ), gamma4->GetParError( 0 ),
					gamma4->GetParameter( 1 ), gamma4->GetParError( 1 ),
					gamma4->GetParameter( 2 ), gamma4->GetParError( 2 ),
					gamma4->GetParameter( 3 ), gamma4->GetParError( 3 ),
					gamma4->GetParameter( 4 ), gamma4->GetParError( 4 ),
					gamma4->GetParameter( 5 ), gamma4->GetParError( 5 ) );
	gamma4->SetParameters( 150, 4, 3, 2, 0, 0 );
	_pulseHighestHisto2D->Fit( "gamma4", "RQ" );
	pulseHighestFitTuple->Fill( gamma4->GetParameter( 0 ), gamma4->GetParError( 0 ),
					gamma4->GetParameter( 1 ), gamma4->GetParError( 1 ),
					gamma4->GetParameter( 2 ), gamma4->GetParError( 2 ),
					gamma4->GetParameter( 3 ), gamma4->GetParError( 3 ),
					gamma4->GetParameter( 4 ), gamma4->GetParError( 4 ),
					gamma4->GetParameter( 5 ), gamma4->GetParError( 5 ) );
	gamma4->SetParameters( 100, 4, 3, 2, 0, 0 );
	_pulseNeighborsHisto2D->Fit( "gamma4", "RQ" );
	pulseNeighborsFitTuple->Fill( gamma4->GetParameter( 0 ), gamma4->GetParError( 0 ),
					gamma4->GetParameter( 1 ), gamma4->GetParError( 1 ),
					gamma4->GetParameter( 2 ), gamma4->GetParError( 2 ),
					gamma4->GetParameter( 3 ), gamma4->GetParError( 3 ),
					gamma4->GetParameter( 4 ), gamma4->GetParError( 4 ),
					gamma4->GetParameter( 5 ), gamma4->GetParError( 5 ) );

	for( std::map<int, TH2*>::iterator iter=_pulseHighestHistoVSCharge.begin() ; iter!=_pulseHighestHistoVSCharge.end() ; ++iter )
	{
		TProfile* tempProfile = ( ( *iter ).second )->ProfileX( Form( "pulseHighestProfile_Charge%d", ( *iter ).first ) );
		tempProfile->SetTitle( Form( "pulseHighestProfile_Charge%d", ( *iter ).first ) );
		tempProfile->SetXTitle( "time bin" );
		tempProfile->SetYTitle( "pulse height [ADC counts]" );

		gamma4->SetParameters( 100, 4, 3, 2, 0, 0 );
		tempProfile->Fit( "gamma4", "RQ" );
		pulseHighestVSChargeFitTuple ->Fill( gamma4->GetParameter( 0 ), gamma4->GetParError( 0 ),
							gamma4->GetParameter( 1 ), gamma4->GetParError( 1 ),
							gamma4->GetParameter( 2 ), gamma4->GetParError( 2 ),
							gamma4->GetParameter( 3 ), gamma4->GetParError( 3 ),
							gamma4->GetParameter( 4 ), gamma4->GetParError( 4 ),
							gamma4->GetParameter( 5 ), gamma4->GetParError( 5 ) );
		gamma4->SetParameters( 100, 4, 3, 2, 0, 0 );
		( ( *iter ).second )->Fit( "gamma4", "RQ" );
		pulseHighestVSChargeHistFitTuple ->Fill( gamma4->GetParameter( 0 ), gamma4->GetParError( 0 ),
							gamma4->GetParameter( 1 ), gamma4->GetParError( 1 ),
							gamma4->GetParameter( 2 ), gamma4->GetParError( 2 ),
							gamma4->GetParameter( 3 ), gamma4->GetParError( 3 ),
							gamma4->GetParameter( 4 ), gamma4->GetParError( 4 ),
							gamma4->GetParameter( 5 ), gamma4->GetParError( 5 ) );
	}

	delete gamma4;

}
} // close namespace brackets
