#include "ResolutionPerformanceProcessor.h"

#include "HistogramWidthExtractionTools.h"
#include "TrackFitterFactory.h"
#include "TrackFitterBase.h"
#include "LCObjectCopier.h"

// the common flag word definitions in MarlinTPC
#include "FlagwordDefinitions.h"

#include <iostream>
#include <string>
#include <algorithm>

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <IMPL/TrackImpl.h>
#include "TrackFitterBase.h"
#include <EVENT/TrackerData.h>

#include <Exceptions.h>

//GEAR
#include "gear/TPCModule.h"
#include <gear/BField.h>
#include <gearimpl/Vector3D.h>

//ROOT
#include "TGraphErrors.h"
#include "TF1.h"
#include "TH1F.h"
#include "TH2F.h"
#include "TProfile.h"

// Aida
#ifdef MARLIN_USE_AIDA
//AIDA
#include <marlin/AIDAProcessor.h>
#endif

using namespace lcio;
using namespace marlin;
using namespace std;

namespace marlintpc
{

ResolutionPerformanceProcessor aResolutionPerformanceProcessor;


ResolutionPerformanceProcessor::ResolutionPerformanceProcessor() : Processor("ResolutionPerformanceProcessor")
{
	// the processor description
	_description = "Determines the resolution via the geometric mean method.";



	registerInputCollection(    LCIO::TRACK,
			"InputCollectionName" ,
			"Name of the input Track collection, default: TPCTracks"  ,
			_inputColName ,
			string("TPCTracks")) ;

	registerProcessorParameter( "xyResolutionBins" ,
			"Number of bins in the xy-residual histogram, default: 254",
			_xyResolutionBins,
			int(254));

	registerProcessorParameter( "xyResolutionRange",
			"Range of xy-residual histogram, default: 3.mm",
			_xyResolutionRange,
			float(3));

	registerProcessorParameter( "zResolutionBins" ,
			"Number of bins in the z-residual histogram, default: 254",
			_zResolutionBins,
			int(254));

	registerProcessorParameter( "zResolutionRange",
			"Range of z-residual histogram, default: 5.mm",
			_zResolutionRange,
			float(5));

	registerProcessorParameter( "ChargeBins",
			"Number of bins on the charge axis in the residuals vs charge histograms, default: 1024",
			_chargeBins,
			int(1024));

	registerProcessorParameter( "ChargeUpperBound",
			"Upper bound of the charge axis in the residuals vs charge histograms, default: 1024",
			_chargeUpperBound,
			int(1024));

	registerProcessorParameter( "meanCalculationMethodZ",
			"Defines the method of calculating the mean of the z-histograms. (default 1:RMS90, 2:Gaussian fit within range of 3 sigma, 3:rms after 5 sigma iterative cut of tails, 4:half gaussian fit in 3 sigma on rising slope)",
			_meanMethodZ,
			int(1));

	registerProcessorParameter( "meanCalculationMethodXY",
			"Defines the method of calculating the mean of the xy-histograms. (default 2:Gaussian fit within range of 3 sigma, 1:RMS90, 3:rms after 5 sigma iterative cut of tails, 4:half gaussian fit in 3 sigma on rising slope)",
			_meanMethodXY,
			int(2));

	registerProcessorParameter( "ModuleWithBestRow",
			"Module that has the best row (default:0)",
			_bestModule,
			int(0));

	registerProcessorParameter( "RowWithBestPerformance",
			"Row with the best resolution performance (default:0)",
			_bestRow,
			int(0));

	registerProcessorParameter("BFieldScaleFactor",
			"Scales magnetic field (map), use 1.0 (default) for field ON or 0.0 for field OFF",
			_bfieldScaleFactor,
			double(1.0));

	registerProcessorParameter("UseHitsWithOverflowPulse",
			"Determines if hits with overflow pulses are evaluated for resolution (default: false)",
			_useHitsWithOverflow,
			false);

	registerOptionalParameter("MaximumPulseCharge",
                              "Optional: Maximum charge of max pulse in hit to determine wether a hit is evaluated for resolution; pulses with the exact cut value are evaluated.",
                              _maxChargeCut,
                              float(1024.));

	registerOptionalParameter("MinimumPulseCharge",
                              "Optional: Minimum charge of max pulse in hit to determine wether a hit is evaluated for resolution; pulses with the exact cut value are evaluated.",
                              _minChargeCut,
                              float(0.));

	registerOptionalParameter("MinBinsRiseTime",
                              "Optional: Minimum number of bins in on the rising slope of the maximum pulse to determine wether hit is evaluated for resolution; pulses with the exact cut value are evaluated.", 
                              _minBinsRiseTimeCut,
                              int(3));

}


void ResolutionPerformanceProcessor::init()
{
	streamlog_out(DEBUG) << " init() called  " << endl;

	// you might want to list the used parameters by invoking
	printParameters();
	if(parameterSet("MaximumPulseCharge") && parameterSet("MinimumPulseCharge") && _maxChargeCut<=_minChargeCut)
	{
		streamlog_out(ERROR) << "MaximumPulseCharge cut is set lower than or equal to MinimumPulseCharge cut! No hit will be evaluated!" << endl;
	}
	AIDAProcessor::tree(this);
	_gearParameters= &marlin::Global::GEAR->getTPCParameters();
	std::vector<gear::TPCModule*> TPCModules = _gearParameters->getModules();

	std::vector<gear::TPCModule*>::const_iterator moduleIter = TPCModules.begin();
	int moduletype0 = (*moduleIter)->getPadLayoutType();

	for( ; moduleIter != TPCModules.end() ; ++moduleIter)
	{
		if((*moduleIter)->getPadLayoutType() != moduletype0)
			streamlog_out(WARNING) << "Modules with different pad layouts are defined. Will use first one!"<< std::endl;
	}

	_myModuleDistortionCalculator=new ModuleDistortionCalculator( &marlin::Global::GEAR->getTPCParameters());

	_zDistanceAll = new TH1F("zDistance","zDistance",_zResolutionBins,-_zResolutionRange,_zResolutionRange);
	_zResidualAll = new TH1F("zResidual","zResiudal",_zResolutionBins,-_zResolutionRange,_zResolutionRange);
	_xyDistanceAll = new TH1F("xyDistance","xyDistance",_xyResolutionBins,-_xyResolutionRange,_xyResolutionRange);
	_xyResidualAll = new TH1F("xyResidual","xyResidual",_xyResolutionBins,-_xyResolutionRange,_xyResolutionRange);

	_zResidualVSChargeAllRows = new TH2F("zResidualVSChargeAllRows","zResidualVSChargeAllRows",_chargeBins,0,_chargeUpperBound,_zResolutionBins,-_zResolutionRange,_zResolutionRange);
	_xyResidualVSChargeAllRows = new TH2F("xyResidualVSChargeAllRows","xyResidualVSChargeAllRows",_chargeBins,0,_chargeUpperBound,_xyResolutionBins,-_xyResolutionRange,_xyResolutionRange);

	_allRowsTuple = new TNtuple("allRowsResolution","allRowsResolution","geoMean_xy:geoMean_xyError:geoMean_z:geoMean_zError");
	_singleRowsTuple = new TNtuple("singleRowsResolution","singleRowsResolution","module:row:geoMean_xy:geoMean_xyError:geoMean_z:geoMean_zError:phi:phiError:phiWidth:phiWidthError");
	_bestRowTuple = new TNtuple("bestRowResolution","bestRowResolution","module:row:geoMean_xy:geoMean_xyError:geoMean_z:geoMean_zError");

	_zResidualVSChargeAllRowsTuple = new TNtuple("zResidualVSChargeAllRowsTuple","zResidualVSChargeAllRowsTuple","offset:offsetError:slope:slopeError:scale:scaleError:exponent:exponentError");
	_zResidualVSChargeSingleRowsTuple = new TNtuple("zResidualVSChargeSingleRowsTuple","zResidualVSChargeSingleRowsTuple","offset:offsetError:slope:slopeError:scale:scaleError:exponent:exponentError");

	gear::Vector3D theTPCCenter(_gearParameters->getDoubleVal("TPC_cylinder_centre_c0"),
			_gearParameters->getDoubleVal("TPC_cylinder_centre_c1"),
			0.);

	gear::Vector3D field = marlin::Global::GEAR->getBField().at(theTPCCenter);

	// const bool bOff ( gear_field[2] == 0.);
	const bool bOff ( field.z()*_bfieldScaleFactor == 0. );
	if (bOff)
	{	streamlog_out(DEBUG) << "BField(GEAR) is OFF" << std::endl;}
	else
	{	streamlog_out(DEBUG) << "BField(GEAR) is ON" << std::endl;}
	// constant magnetic field (in z direction)
	// double Bz = gear_field[2];
	double Bz = field.z()*_bfieldScaleFactor;// rho component is accessible with field.rho()

	unsigned char fitterType= TrackFitterBase::FitterTypes::SIMPLEHELIXFIT;
	if(Bz==0)
	{
		fitterType=TrackFitterBase::FitterTypes::LINEARREGRESSION;
	}
	_trackFitter  = TrackFitterFactory::getTrackFitter(fitterType,0);
}


void ResolutionPerformanceProcessor::processRunHeader(LCRunHeader*  /*run*/)
{
	// do something for each run
}


void ResolutionPerformanceProcessor::processEvent(LCEvent* evt)
{

	AIDAProcessor::tree(this);
	// use an LCCollectionVec instead of LCCollection so we can use iterators later
	LCCollectionVec* inputColVec = 0;

	// do a try / catch in case there is an empty event
	try
	{
		inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColName ) );
	}
	catch (DataNotAvailableException &e)
	{
		// no input data, just return and process the next event
		return;
	}

	// loop the track collection and do something
	for( LCCollectionVec::iterator inputObjectIter = inputColVec->begin() ;
			inputObjectIter < inputColVec->end() ;
			inputObjectIter++ )
	{
		// dynamic cast the input track object
		Track * inputTrack = dynamic_cast<Track *>( *inputObjectIter );

		if ( inputTrack == 0 )
		{
			throw DataNotAvailableException("No input tracks found.");
		}

		TrackImpl * refittedTrack = _trackFitter->fitTrack(inputTrack);
		//TrackImpl * refittedTrack = LCObjectCopier::copy(inputTrack);

		TrackerHitVec hitVec = inputTrack->getTrackerHits();
		int numberHits = hitVec.size();
		int numberCutHits = 0;

		for ( TrackerHitVec::iterator itHit = hitVec.begin(); itHit < hitVec.end(); itHit++ )
		{

			TrackerHit* hit = *itHit;
			double maxCharge = 0;
			int nBinsOnSlope = 0;
			TrackerPulse const * maxPulse = 0;
			//find pulse with maximum charge and calculate the length of it's rising slope
			for ( LCObjectVec::const_iterator pulseIter = hit->getRawHits().begin() ;
						pulseIter < hit->getRawHits().end() ; pulseIter++ )
			{
				TrackerPulse const * pulse = dynamic_cast<TrackerPulse const *>(*pulseIter);
				EVENT::TrackerData* trackerData = pulse->getTrackerData();
				vector<float> theADCSpectrum = trackerData->getChargeValues();
				vector<float>::iterator maxElement = max_element(theADCSpectrum.begin(), theADCSpectrum.end());
				if(*maxElement>maxCharge)
				{
					maxCharge = *maxElement;
					nBinsOnSlope = maxElement-theADCSpectrum.begin()-2;//three bins before threshold are taken; begin()=1
					maxPulse = pulse;
				}

			}
			bool skipHit ( (parameterSet("MinimumPulseCharge") && maxCharge < _minChargeCut)
					|| (parameterSet("MaximumPulseCharge") && maxCharge > _maxChargeCut)
					|| (!_useHitsWithOverflow && hitflag::hasOverflowPulse(hit->getQuality()))
					|| (parameterSet("MinBinsRiseTime") && nBinsOnSlope < _minBinsRiseTimeCut) );
			if(skipHit)
			{
				streamlog_out(DEBUG3) << "Hit failed the following cuts:";
				if(parameterSet("MinimumPulseCharge") && maxCharge < _minChargeCut)
					streamlog_out(DEBUG3) << " MinimumPulseCharge;";
				if(parameterSet("MaximumPulseCharge") && maxCharge > _maxChargeCut)
					streamlog_out(DEBUG3) << " MaximumPulseCharge;";
				if(!_useHitsWithOverflow && hitflag::hasOverflowPulse(hit->getQuality()))
					streamlog_out(DEBUG3) << " HitHasOverflowPulse;";
				if(parameterSet("MinBinsRiseTime") && nBinsOnSlope < _minBinsRiseTimeCut)
					streamlog_out(DEBUG3) << " MinBinsRiseTime;";
				streamlog_out(DEBUG3) << endl;
				streamlog_out(DEBUG2) << "The maximum pulse ADC spectrum:";
				for(vector<float>::const_iterator spectrumIter = maxPulse->getTrackerData()->getChargeValues().begin() ; spectrumIter < maxPulse->getTrackerData()->getChargeValues().end() ; spectrumIter++ )
				{
					streamlog_out(DEBUG2) << " " << *spectrumIter;
				}
				streamlog_out(DEBUG2) << endl;
				++numberCutHits;
				continue;
			}

			TrackImpl * trackWOHit = LCObjectCopier::copyTrackWithoutHits(refittedTrack);
			for (EVENT::TrackerHitVec::const_iterator hitIter = hitVec.begin();
					hitIter < hitVec.end(); hitIter++)
			{
				if(hit!=*hitIter)
					trackWOHit->addHit(*hitIter);
			}

			TrackImpl * refittedHitTrack = _trackFitter->fitTrack(trackWOHit);

			//find out module and row of the hit.
			const gear::TPCModule& theModule = _gearParameters->getNearestModule(hit->getPosition()[0],hit->getPosition()[1]);
			int row=theModule.getRowNumber(theModule.getNearestPad(hit->getPosition()[0],hit->getPosition()[1]));
			int module=theModule.getModuleID();
			std::pair<int, int> key(module,row);
			std::map< std::pair<int,int>, TH1*>::iterator histoiter;
			histoiter = _xyDistanceRow.find(key);
			if(histoiter == _xyDistanceRow.end())
			{
				TH1* tempHist = new TH1F(Form("xyDistance_%i_%i",module,row),"xyDistance",_xyResolutionBins,-_xyResolutionRange,_xyResolutionRange);
				_xyDistanceRow.insert( std::make_pair(key,tempHist));

				TH1* tempHist2 = new TH1F(Form("zDistance_%i_%i",module,row),"zDistance",_zResolutionBins,-_zResolutionRange,_zResolutionRange);
				_zDistanceRow.insert( std::make_pair(key,tempHist2));

				TH1* tempHist3 = new TH1F(Form("xyResidual_%i_%i",module,row),"xyResidual",_xyResolutionBins,-_xyResolutionRange,_xyResolutionRange);
				_xyResidualRow.insert( std::make_pair(key,tempHist3));

				TH1* tempHist4 = new TH1F(Form("zResidual_%i_%i",module,row),"zResidual",_zResolutionBins,-_zResolutionRange,_zResolutionRange);
				_zResidualRow.insert( std::make_pair(key,tempHist4));

				TH1* tempHist5 = new TH1F(Form("phi_%i_%i",module,row),"phi",256,0,0);
				_phiRow.insert( std::make_pair(key,tempHist5));

				TH2* tempHist2d1 = new TH2F(Form("zResidualVSChargeRow_%i_%i",module,row),"zResidualVSCharge",_chargeBins,0,_chargeUpperBound,_zResolutionBins,-_zResolutionRange,_zResolutionRange);
				_zResidualVSChargeRow.insert( std::make_pair(key,tempHist2d1));
			}

			double distValue=0;
			//fill the histograms with the distance and residual values
			histoiter = _xyDistanceRow.find(key);
			try{
				distValue = _myModuleDistortionCalculator->calculateDistanceRow(hit,refittedTrack);
				((*histoiter).second)->Fill( distValue );
				_xyDistanceAll->Fill(distValue);
			}
			catch(std::exception &e)
			{
				streamlog_out(WARNING)<<"calculateDistanceRow for distances failed. Value ignored!"<<std::endl;
			}

			histoiter = _zDistanceRow.find(key);
			distValue = _myModuleDistortionCalculator->calculateDistanceZ(hit,refittedTrack);
			((*histoiter).second)->Fill( distValue );
			_zDistanceAll->Fill(distValue);

			histoiter = _phiRow.find(key);
			try
			{
				TrackerHit * previousHit;
				if( (itHit==hitVec.begin() || row==0) && itHit+1!=hitVec.end() ) previousHit=*(itHit+1);
				else previousHit=*(itHit-1);

				if(module==_gearParameters->getNearestModule(previousHit->getPosition()[0],previousHit->getPosition()[1]).getModuleID())
				{
					double phiValue = calcHitPhi(hit,previousHit);
					((*histoiter).second)->Fill( phiValue );
				}
			}
			catch(std::exception &e)
			{
				streamlog_out(WARNING)<<"calcPhiHit failed. Value ignored!"<<std::endl;
			}

			//TODO: new track fit and change input track to fittrack
			try{
				histoiter = _xyResidualRow.find(key);
				distValue = _myModuleDistortionCalculator->calculateDistanceRow(hit,refittedHitTrack);
				((*histoiter).second)->Fill( distValue );
				_xyResidualAll->Fill(distValue);

				_xyResidualVSChargeAllRows->Fill(maxCharge, distValue); 
			}
			catch(std::exception &e)
			{
				streamlog_out(WARNING)<<"calculateDistanceRow for residuals failed. Value ignored!"<<std::endl;
			}

			histoiter = _zResidualRow.find(key);
			distValue = _myModuleDistortionCalculator->calculateDistanceZ(hit,refittedHitTrack);
			((*histoiter).second)->Fill( distValue );
			_zResidualAll->Fill(distValue);

			std::map< std::pair<int,int>, TH2*>::iterator histoiter2d = _zResidualVSChargeRow.find(key);
			((*histoiter2d).second)->Fill(maxCharge, distValue);
			_zResidualVSChargeAllRows->Fill(maxCharge, distValue); 

			delete trackWOHit;
			delete refittedHitTrack;
		}//hit loop

		streamlog_out(DEBUG4) << "Number of hits (cut / track total): " << numberCutHits << "/" << numberHits << endl;

		delete refittedTrack;
	}// track loop
}


void ResolutionPerformanceProcessor::check(LCEvent*  /*evt*/)
{
	// check what is needed here
}


void ResolutionPerformanceProcessor::end()
{
	//necessary if histograms are created here instead of in init(),
	//to get histograms in the correct folder
	AIDAProcessor::tree(this);

	//make profiles from the scatterplots and fit the z one and store the results
	TF1 * fit = new TF1("fit","pol1(0)+[2]*x**(-[3])",0,900);
	fit->SetParameters(-1,0,10,.3);
	TProfile * _zResidualVSChargeAllRowsProfile = _zResidualVSChargeAllRows->ProfileX("zResidualVSChargeAllRowsProfile");
	_zResidualVSChargeAllRowsProfile->SetTitle("zResidualVSChargeAllRowsProfile");
	_zResidualVSChargeAllRowsProfile->GetYaxis()->SetRangeUser(-_zResolutionRange, _zResolutionRange);
	_zResidualVSChargeAllRowsProfile->SetErrorOption("i");
	_zResidualVSChargeAllRowsProfile->Fit("fit","Q");
	double offset = fit->GetParameter(0);
	double offset_err = fit->GetParError(0);
	double slope = fit->GetParameter(1);
	double slope_err = fit->GetParError(1);
	double scale = fit->GetParameter(2);
	double scale_err = fit->GetParError(2);
	double exponent = fit->GetParameter(3);
	double exponent_err = fit->GetParError(3);
	_zResidualVSChargeAllRowsTuple->Fill(offset,offset_err,slope,slope_err,scale,scale_err,exponent,exponent_err); 

	TProfile * _xyResidualVSChargeAllRowsProfile = _xyResidualVSChargeAllRows->ProfileX("xyResidualVSChargeAllRowsProfile");
	_xyResidualVSChargeAllRowsProfile->SetTitle("xyResidualVSChargeAllRowsProfile");
	_xyResidualVSChargeAllRowsProfile->SetErrorOption("i");

	std::map< std::pair<int,int>, TH2*>::iterator histoiter2d = _zResidualVSChargeRow.begin();
	for(; histoiter2d!=_zResidualVSChargeRow.end(); histoiter2d++)
	{
		std::pair<int,int> key = (*histoiter2d).first;
		TProfile * tempProfile = ((*histoiter2d).second)->ProfileX(Form("zResidualVSChargeProfileRow_%i_%i",key.first,key.second));
		tempProfile->SetTitle("zResidualVSChargeProfile");
		tempProfile->GetYaxis()->SetRangeUser(-_zResolutionRange, _zResolutionRange);
		tempProfile->SetErrorOption("i");
		fit->SetParameters(-1,0,10,.3);
		tempProfile->Fit("fit","Q");
		offset = fit->GetParameter(0);
		offset_err = fit->GetParError(0);
		slope = fit->GetParameter(1);
		slope_err = fit->GetParError(1);
		scale = fit->GetParameter(2);
		scale_err = fit->GetParError(2);
		exponent = fit->GetParameter(3);
		exponent_err = fit->GetParError(3);
		_zResidualVSChargeSingleRowsTuple->Fill(offset,offset_err,slope,slope_err,scale,scale_err,exponent,exponent_err); 
	}

	//all histograms are filled
	//now we need to fit them
	HistogramWidthExtractionTools *myTool= new HistogramWidthExtractionTools();

	myTool->Calculate(_zDistanceAll,_meanMethodZ);
	double z_widthDist=myTool->GetWidth();
	double z_widthDist_err=myTool->GetWidthError();
	myTool->Calculate(_zResidualAll,_meanMethodZ);
	double z_widthRes=myTool->GetWidth();
	double z_widthRes_err=myTool->GetWidthError();

	double resZ = sqrt(z_widthDist*z_widthRes);
	double resZError = 0.5/resZ*sqrt(z_widthRes*z_widthRes*(z_widthDist_err*z_widthDist_err)+z_widthDist*z_widthDist*(z_widthRes_err*z_widthRes_err));

	myTool->Calculate(_xyDistanceAll,_meanMethodXY);
	double xy_widthDist=myTool->GetWidth();
	double xy_widthDist_err=myTool->GetWidthError();
	myTool->Calculate(_xyResidualAll,_meanMethodXY);
	double xy_widthRes=myTool->GetWidth();
	double xy_widthRes_err=myTool->GetWidthError();

	double resXY = sqrt(xy_widthDist*xy_widthRes);
	double resXYError = 0.5/resXY*sqrt(xy_widthRes*xy_widthRes*(xy_widthDist_err*xy_widthDist_err)+xy_widthDist*xy_widthDist*( xy_widthRes_err* xy_widthRes_err));
	_allRowsTuple->Fill(resXY,resXYError,resZ,resZError);

	std::vector<double> radius,radiusError,sigmaXY,sigmaXYError,sigmaZ,sigmaZError;
	std::vector<double> phi,phiError;

	std::map< std::pair<int,int>, TH1*>::iterator histoiter = _zDistanceRow.begin();
	for(; histoiter!=_zDistanceRow.end(); histoiter++)
	{
		std::pair<int,int> key = (*histoiter).first;

		myTool->Calculate(((*histoiter).second),_meanMethodZ);
		double width1=myTool->GetWidth();
		double width1_err=myTool->GetWidthError();
		myTool->Calculate(_zResidualRow[key],_meanMethodZ);
		double width2=myTool->GetWidth();
		double width2_err=myTool->GetWidthError();

		resZ = sqrt(width1*width2);
		resZError = 0.5/resZ*sqrt(width2_err*width2_err*(width1_err*width1_err)+width1_err*width1_err*(width2_err*width2_err));

		myTool->Calculate(_xyDistanceRow[key],_meanMethodXY);
		width1=myTool->GetWidth();
		width1_err=myTool->GetWidthError();
		myTool->Calculate(_xyResidualRow[key],_meanMethodXY);
		width2=myTool->GetWidth();
		width2_err=myTool->GetWidthError();

		resXY = sqrt(width1*width2);
		resXYError = 0.5/resXY*sqrt(width2_err*width2_err*(width1_err*width1_err)+width1_err*width1_err*(width2_err*width2_err));
		myTool->Calculate(_phiRow[key],1);
		double phiValue=myTool->GetMean();
		double phiWidth=myTool->GetWidth();
		double phiValueError=myTool->GetMeanError();
		double phiWidthError=myTool->GetWidthError();


		_singleRowsTuple->Fill(key.first, key.second, resXY,resXYError,resZ,resZError,phiValue,phiValueError,phiWidth,phiWidthError);
		if (key.first==_bestModule && key.second==_bestRow)
			_bestRowTuple->Fill(key.first, key.second, resXY,resXYError,resZ,resZError);

		sigmaXY.push_back(resXY);
		sigmaXYError.push_back(resXYError);
		sigmaZ.push_back(resZ);
		sigmaZError.push_back(resZError);
		phi.push_back(phiValue);
		phiError.push_back(phiValueError);


		//get from to corresponding radius at phi=0
		const gear::TPCModule& theModule = _gearParameters->getModule(key.first);
		gear::Vector2D padcenter=theModule.getPadCenter(theModule.getPadIndex(key.second,0));
		radius.push_back(sqrt(padcenter[0]*padcenter[0]+padcenter[1]*padcenter[1]));
		radiusError.push_back((theModule.getRowHeight(key.second))/sqrt(12.));
	}

	TGraphErrors *graphRow=new TGraphErrors(radius.size(),&radius[0],&sigmaXY[0],&radiusError[0],&sigmaXYError[0]);
	graphRow->SetNameTitle("XYResolutionRow","XYResolutionRow");
	graphRow->GetXaxis()->SetTitle("r_{row} [mm]");
	graphRow->GetYaxis()->SetTitle("#sigma_{row} [mm]");
	graphRow->Write();

	TGraphErrors *graphZ=new TGraphErrors(radius.size(),&radius[0],&sigmaZ[0],&radiusError[0],&sigmaZError[0]);
	graphZ->SetNameTitle("ZResolutionRow","ZResolutionRow");
	graphZ->GetXaxis()->SetTitle("r_{row} [mm]");
	graphZ->GetYaxis()->SetTitle("#sigma_{z} [mm]");
	graphZ->Write();

	TGraphErrors *graphPhi=new TGraphErrors(radius.size(),&radius[0],&phi[0],&radiusError[0],&phiError[0]);
	graphPhi->SetNameTitle("PhiRow","PhiRow");
	graphPhi->GetXaxis()->SetTitle("r_{row} [mm]");
	graphPhi->GetYaxis()->SetTitle("local hit #varphi [rad]");
	graphPhi->Write();

    delete myTool;
    delete graphZ;
    delete graphRow;
    delete graphPhi;

}

double ResolutionPerformanceProcessor::calcHitPhi(TrackerHit* firstHit, TrackerHit* secondHit)
{
	//calculate angle between hit and track
	TrackFitterBase * trackFitter  = TrackFitterFactory::getTrackFitter(TrackFitterBase::FitterTypes::LINEARREGRESSION,0);
	TrackImpl * tempTrack = new TrackImpl();
	tempTrack->addHit(firstHit);
	tempTrack->addHit(secondHit);
	TrackImpl * fittedTrack = trackFitter->fitTrack(tempTrack);
	const double * position = firstHit->getPosition();

	const gear::TPCModule& theModule = _gearParameters->getNearestModule(position[0],position[1]);
	int padIndex = theModule.getNearestPad(position[0],position[1]);
	gear::Vector2D padCenterGlobal = theModule.getPadCenter(padIndex);
	gear::Vector2D e_pad (padCenterGlobal[0]-theModule.getOffset()[0],padCenterGlobal[1]-theModule.getOffset()[1]);
	double phi_s=atan2(sin(fittedTrack->getPhi())-fittedTrack->getOmega()*(position[0]-(-fittedTrack->getD0()*sin(fittedTrack->getPhi())+fittedTrack->getReferencePoint()[0])),
							cos(fittedTrack->getPhi())+fittedTrack->getOmega()*(position[1]-(fittedTrack->getD0()*cos(fittedTrack->getPhi())+fittedTrack->getReferencePoint()[1])));

	delete tempTrack;
	delete fittedTrack;

	return (phi_s-atan(e_pad[1]/e_pad[0]));
}
} // close namespace brackets
