#include "TimePixOccupancyProcessor.h"


using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

TimePixOccupancyProcessor aTimePixOccupancyProcessor;
    
  TimePixOccupancyProcessor::TimePixOccupancyProcessor(): Processor("TimePixOccupancyProcessor"),
                                                          _nColumnModules(1),
                                                          _nRowModules(1)
{
  // register steering parameters: name, description, class-variable, default value
  registerInputCollection(    LCIO::TRACKERDATA,
  				"InputColName" , 
  				"Name of the input collection  with lcio::TrackerData "
  				"(default: TimePixInterpolatedData)"  ,
  				_inputColName ,
  				std::string("TimePixInterpolData") ) ;
  // modify processor description
  _description = "TimePixOccupancyProcessor: Count how many times a pixel has been hit" ;

  registerProcessorParameter( "NColumns" , 
			      "Number of colmns on the chip (default: 256)"  ,
			      _nColumns ,
			      int(256) ) ;

  registerProcessorParameter( "NRows" , 
			      "Number of rows on the chip (default: 256)"  ,
			      _nRows ,
			      int(256) ) ;

  registerOptionalParameter( "nColumnModules",
                             "Number of modules/chips in a column",
			     _nColumnModules,
                             int(1));

  registerOptionalParameter( "nRowModules",
                             "Number of modules/chips in a row",
			     _nRowModules,
                             int(1));

  registerOptionalParameter( "chipOrientation",
                             "Vector containing the position of the zero on the chip. 0 = lower left corner, 1 = lower right corner, 2 = upper right corner, 3 = upper left corner. Make sure the entrys match the the entries in the chip ID vector",
                             _chipDirections,
                             std::vector<int>(1,0) );

  registerOptionalParameter( "ChipIDs" , 
			     "ID number(s) of the chip(s) (default: 0) - make sure they match the entries in the chipOrientation vector, if set"  ,
			     _chipIDs,
			     std::vector<int>(1,0) ) ;


}



TimePixOccupancyProcessor::~TimePixOccupancyProcessor()
{
}



void TimePixOccupancyProcessor::init()
{ 
  // usually a good idea to
  printParameters() ;


  //set vectors with chipIDs and orinatations to theyre according values
  unsigned int nChips = _nColumnModules * _nRowModules;

  if( !parameterSet( "ChipIDs" ) )
  {
     _chipIDs.resize( nChips );

     for( unsigned int iChip = 0; iChip < nChips; ++iChip )
     {
       _chipIDs[iChip] = iChip;
     }
  } // end of if( !parameterSet( "ChipIDs" ) )

  if( !parameterSet( "chipOrientation" ) )
  {
     _chipDirections.resize( nChips );

     for( unsigned int iChip = 0; iChip < nChips; ++iChip )
     {
       _chipIDs[iChip] = 0;
     }
  } // end of if( !parameterSet( "ChipIDs" ) )


  //initialising 
  _allChipsOccupancy = new TH2I("TimePixOccupancyAll"," TimePix Occupancy All Chips",_nColumnModules*_nColumns+1,-0.5,_nColumnModules*_nColumns,_nRows*_nRowModules+1,-0.5,_nRows*_nRowModules+0.5);
}



void TimePixOccupancyProcessor::processRunHeader(lcio::LCRunHeader*  /*run*/ )
{
}


    
void TimePixOccupancyProcessor::processEvent(lcio::LCEvent * evt ) 
{ 
  //read input data
  streamlog_out(MESSAGE3) << "Event: " << evt->getEventNumber() << std::endl;

  // we need a LCCollectionVec to use iterators, as LCCollection is not a std::vector
  LCCollectionVec* InputCollection=0;
  try{
	InputCollection = dynamic_cast<LCCollectionVec*>(evt->getCollection( _inputColName )) ;
  }
  catch(DataNotAvailableException &e)
  {
	// no input data
	// just continue ...
	return;
  }

  // cast did not succeed
  if( InputCollection == 0 )
  {
	streamlog_out(ERROR2) << "InputCollection vector is 0" << std::endl;
	return;
  }


  for(LCCollectionVec::iterator it = InputCollection->begin() ; 
	it < InputCollection->end(); it ++ )
  {
    TrackerData *timePixData =  dynamic_cast<TrackerData *>(*it);
    std::vector<float> adcValues = timePixData->getChargeValues();

    if (timePixData==0)
    {
      throw DataNotAvailableException("Data in collection "+_inputColName +
                                     "is not of type TrackerData");
    }

    int cellID1 = timePixData->getCellID1(); //chipID
    int cellID0 = timePixData->getCellID0();

    streamlog_out(DEBUG2) << "got tracker data - cellID1: " << cellID1 << std::endl;
 
    //get coordinate values
    for(unsigned int iADCvalues = 0; iADCvalues < adcValues.size(); iADCvalues++)
    {
      unsigned int oldX = ((cellID0+1) % _nColumns);
      unsigned int oldY = (cellID0+1-oldX) / _nRows;
    
      //calculate corrdinates for the entry into the allChipsHisto - TODO: work with gear!!
      int newX =0;
      int newY =0;

      //find position in vecs:
      //TODO: Add more parameter to specify the general orientation of a chip in a row
      int pos = 0;
      int row = 0;
      int column = 0;
    
      //&& (_chipIDs.at(iFind) != cellID1)
      /** TODO:       for(int iFind = 0; (iFind < _chipIDs.size()) ; iFind++)
       * causes warning: comparison between signed and unsigned integer
       *  expressions
       */
      for(unsigned int iFind = 0; (iFind < _chipIDs.size()) ; iFind++)
      { 
        if(_chipIDs.at(iFind) == cellID1) 
        {
          pos = iFind;
          for(int iRow = _nRowModules; (iRow > 0) && ((pos+1) <= (_nColumnModules*iRow)); iRow--) row = iRow;
              break;
        }
      }
      
      streamlog_out(DEBUG1) << "found position: " <<  pos+1 << " and row " << row << std::endl;
          
      column = (pos+1) % _nColumnModules;
      if(column == 0)
      { 
        column = _nColumnModules;
      }
      newX = (column-1)*_nColumns;
      newY = (row-1)*_nRows;
    
      if(_chipDirections.at(pos) == 0) 
      { 
        newX += oldX;
        newY += oldY;
      }
      if(_chipDirections.at(pos) == 1)
      {
        newX += (256 - oldX);
        newY += oldX;
      } 
      if(_chipDirections.at(pos) == 2) 
      {  
        newX += (256 - oldX);
        newY += (256 - oldY);
      }
      if(_chipDirections.at(pos) == 3)
      {
        newX += oldY;
        newY += (256-oldX);
      }

      streamlog_out(DEBUG2) << "old/new " << oldX << "/" << oldY << " " << newX << "/" << newY << std::endl; 
      
      _allChipsOccupancy->Fill(newX,newY);

      streamlog_out(DEBUG2) << "wrote data" << std::endl;

    }

    // search for histogram with chip ID CellID1 in the map
    std::map<int, AIDA::IHistogram2D *>::iterator histoIter =
        _timePixOccupancyHistos.find(cellID1);
    
    // if histo was not found create it
    if (histoIter == _timePixOccupancyHistos.end() )
    {
        // create a name for the histogram
        std::stringstream histotitle;
        histotitle << "TimePix Occupancy Chip"<< timePixData->getCellID1();

        std::stringstream histoname;
        histoname << "TimePix_occupancy_"<< _inputColName << "_chip" <<timePixData->getCellID1();

        _timePixOccupancyHistos[timePixData->getCellID1()] = AIDAProcessor::histogramFactory(this)->
            createHistogram2D( histoname.str(),  histotitle.str(), _nColumns, 0, _nColumns , _nRows, 0, _nRows);

        histoIter = _timePixOccupancyHistos.find(timePixData->getCellID1());
    }

    //streamlog_out(MESSAGE1) << "nr of hits " << hitVector.size()<< std::endl;
    //int hitNr=0;
    //streamlog_out(MESSAGE1) << "\t\t hit\t x \t y \t z \t residual_x/r*phi" << std::endl;
    
    unsigned int firstPixelID =  timePixData->getCellID0();
    for ( unsigned int index =  0 ; index < timePixData->getChargeValues().size() ; 
          index++ )

    {
      if (timePixData->getChargeValues()[index] > 0)
      {
        int pixelID = firstPixelID + index;
        // even for DEBUG0 this is too much output
        //      streamlog_out(DEBUG0) << "Filling pixel row " << pixelID/_nColumns
        //                    << " column "<< pixelID%_nColumns  << std::endl;

        // histo iter is an iterator of a pair, the histo pointer is pair.second
        (*histoIter).second->fill(pixelID%_nColumns, pixelID/_nColumns);
      }
    }
  }
}


void TimePixOccupancyProcessor::end()
{
  //Draw global histos
  marlin::AIDAProcessor::tree(this);
  _allChipsOccupancy->Write();
  delete _allChipsOccupancy;
  
}


}// namespace 
