#include "TrackParametersDistributionProcessor.h"

//C++
#include <iostream>
#include <cmath>
using namespace std;

//LCIO
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>

// Aida
//#ifdef MARLIN_USE_AIDA // the whole processor does not make sense without aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
// endif

#include <marlin/Global.h>

using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

TrackParametersDistributionProcessor aTrackParametersDistributionProcessor;
    
TrackParametersDistributionProcessor::TrackParametersDistributionProcessor()
    : Processor("TrackParametersDistributionProcessor"),
      _distributionPhiHisto(0) , _distributionLambdaHisto(0),
      _distributionTanLambdaHisto(0), _distributionD0Histo(0),
      _distributionZ0Histo(0), _distributionOmegaHisto(0),
      _distributionChi2Histo(0), _distributionNdfHisto(0),
      _distributionDEdxHisto(0), _distributionNHitsHisto(0), _distributionDEdxVsOmegaHisto(0)

{
    // register steering parameters: name, description, class-variable, default value
    registerInputCollection(    LCIO::TRACK,
				"InputTracks" , 
				"Name of the input tracks collection"  ,
				_inputColName ,
				std::string("TPCSeedTracks") ) ;
    // modify processor description
    _description = "TrackParametersDistributionProcessor: Fill phi, lambda, and z distribution of hits in tracks into AIDA histgram" ;

  registerProcessorParameter( "HistogramMinValuePhi" , 
			      "Minimum value in phi histogram (default: -pi)"  ,
			      _histogramMinValuePhi ,
			      float(-M_PI) ) ;

  registerProcessorParameter( "HistogramMaxValuePhi" , 
			      "Maximal value in phi histogram (default: pi)"  ,
			      _histogramMaxValuePhi ,
			      float(M_PI) ) ;


  registerProcessorParameter( "HistogramMinValueLambda" , 
			      "Minimum value in lambda histogram (default: pi)"  ,
			      _histogramMinValueLambda ,
			      float(-M_PI) ) ;

  registerProcessorParameter( "HistogramMaxValueLambda" , 
			      "Maximal value in lambda histogram (default:  pi)"  ,
			      _histogramMaxValueLambda ,
			      float(M_PI) ) ;


  registerProcessorParameter( "HistogramMinValueTanLambda" , 
			      "Minimum value in tan(lambda) histogram (default: -20.)"  ,
			      _histogramMinValueTanLambda ,
			      float(-20.) ) ;

  registerProcessorParameter( "HistogramMaxValueTanLambda" , 
			      "Maximal value in tan(lambda) histogram (default:  20. )"  ,
			      _histogramMaxValueTanLambda ,
			      float(20.) ) ;


  registerProcessorParameter( "HistogramMinValueD0" , 
			      "Minimum value in d0 histogram (default: -50. mm)"  ,
			      _histogramMinValueD0 ,
			      float(-100.) ) ;

  registerProcessorParameter( "HistogramMaxValueD0" , 
			      "Maximal value in d0 histogram (default:  50. mm)"  ,
			      _histogramMaxValueD0 ,
			      float(100.) ) ;

 registerProcessorParameter( "HistogramMinValueZ0" , 
			      "Minimum value in z0 histogram (default: -10. mm)"  ,
			      _histogramMinValueZ0 ,
			      float(0.) ) ;

  registerProcessorParameter( "HistogramMaxValueZ0" , 
			      "Maximal value in z0 histogram (default:  300. mm)"  ,
			      _histogramMaxValueZ0 ,
			      float(600.) ) ;

 registerProcessorParameter( "HistogramMinValueOmega" , 
			      "Minimum value in Omega histogram (default: -0.01 mm^-1)"  ,
			      _histogramMinValueOmega ,
			      float(-0.01) ) ;

  registerProcessorParameter( "HistogramMaxValueOmega" , 
			      "Maximal value in Omega histogram (default:  0.01 mm^-1)"  ,
			      _histogramMaxValueOmega ,
			      float(0.01) ) ;
  registerProcessorParameter( "HistogramMinValueMomentum" ,
 			      "Minimum value in Momentum histogram (default: 0 GeV)"  ,
 			      _histogramMinValueMomentum ,
 			      float(0.) ) ;

   registerProcessorParameter( "HistogramMaxValueMomentum" ,
 			      "Maximal value in Momentum histogram (default:  6.5 GeV)"  ,
 			      _histogramMaxValueMomentum ,
 			      float(6.5) ) ;
   registerProcessorParameter( "bFieldStrength" ,
 			      "strength of the b field in T (default: 1 T)"  ,
				  _bFieldStrength ,
 			      float(1.) ) ;

  registerProcessorParameter( "HistogramMinValueChi2" ,
		  "Minimum value in Chi^2 histogram (default: 0)"  ,
		  _histogramMinValueChi2 ,
		  float(0) ) ;

  registerProcessorParameter( "HistogramMaxValueChi2" ,
		  "Maximal value in Chi^2 histogram (default:  1000)"  ,
		  _histogramMaxValueChi2 ,
		  float(1000) ) ;

  registerProcessorParameter( "HistogramMinValueNdf" ,
		  "Minimum value in number of degrees of freedom histogram (default: 0)"  ,
		  _histogramMinValueNdf ,
		  float(0) ) ;

  registerProcessorParameter( "HistogramMaxValueNdf" ,
		  "Maximal value in number of degrees of freedom histogram (default:  100)"  ,
		  _histogramMaxValueNdf ,
		  float(200) ) ;

  registerProcessorParameter( "HistogramMinValueDEdx" ,
		  "Minimum value in dEdx histogram (default: 0)"  ,
		  _histogramMinValueDEdx ,
		  float(0) ) ;

  registerProcessorParameter( "HistogramMaxValueDEdx" ,
		  "Maximal value in dEdx histogram (default:  1000)"  ,
		  _histogramMaxValueDEdx ,
		  float(1000) ) ;

  registerProcessorParameter( "HistogramMinValueNHits" ,
  		  "Minimum value in number of hits histogram (default: 0)"  ,
  		  _histogramMinValueNHits ,
  		  float(0) ) ;

    registerProcessorParameter( "HistogramMaxValueNHits" ,
  		  "Maximal value in number of hits histogram (default:  100)"  ,
  		  _histogramMaxValueNHits ,
  		  float(100) ) ;

}

TrackParametersDistributionProcessor::~TrackParametersDistributionProcessor()
{
}

void TrackParametersDistributionProcessor::init()
{ 
	// usually a good idea to
	printParameters() ;
    AIDAProcessor::tree(this);
	// initialize histograms
	_distributionPhiHisto =   new TH1D( "phiDistributionHisto",
			     "phiDistributionHisto;#varphi [rad];# tracks",
			     1024, _histogramMinValuePhi, _histogramMaxValuePhi);

	_distributionLambdaHisto =   new TH1D( "lambdaDistributionHisto",
			     "lambdaDistributionHisto;#lambda [rad];# tracks",
			     1024, _histogramMinValueLambda, _histogramMaxValueLambda);

	_distributionTanLambdaHisto =   new TH1D( "tanLambdaDistributionHisto",
			     "tanLambdaDistributionHisto;tan#lambda;# tracks",
			     1024, _histogramMinValueTanLambda, _histogramMaxValueTanLambda);

	_distributionD0Histo =   new TH1D( "d0DistributionHisto",
			     "d0DistributionHisto;D_{0} [mm];# tracks",
			     1024, _histogramMinValueD0, _histogramMaxValueD0);

	_distributionZ0Histo =   new TH1D( "z0DistributionHisto",
			     "z0DistributionHisto;Z_{0} [mm];# tracks",
			     1024, _histogramMinValueZ0, _histogramMaxValueZ0);

	_distributionOmegaHisto =   new TH1D( "OmegaDistributionHisto",
			     "OmegaDistributionHisto;#Omega [1/mm];# tracks",
			     1024, _histogramMinValueOmega, _histogramMaxValueOmega);

	_distributionMomentumHisto =   new TH1D( "MomentumDistributionHisto",
			     "MomentumDistributionHisto;#Momentum [GeV];# tracks",
			     1024, _histogramMinValueMomentum, _histogramMaxValueMomentum);

	_distributionChi2Histo =   new TH1D( "Chi2DistributionHisto",
					"Ch2DistributionHisto;Chi^{2};# tracks",
					1024, _histogramMinValueChi2, _histogramMaxValueChi2);

	_distributionNdfHisto =   new TH1D( "NdfDistributionHisto",
					"NdfDistributionHisto;degrees of freedom;# tracks",
					1024, _histogramMinValueNdf, _histogramMaxValueNdf);

	_distributionDEdxHisto =   new TH1D( "dEdxDistributionHisto",
					"dEdxDistributionHisto;#frac{dE}{dx} [a.u.];# tracks",
					1024, _histogramMinValueDEdx, _histogramMaxValueDEdx);

	_distributionDEdxVsOmegaHisto =   new TH2D( "dEdxVsOmegaDistributionHisto",
					"dEdxVsOmegaDistributionHisto;#Omega [1/mm];#frac{dE}{dx} [a.u.];",
					 1024, _histogramMinValueOmega, _histogramMaxValueOmega, 1024, _histogramMinValueDEdx, _histogramMaxValueDEdx);

	_distributionNHitsHisto =   new TH1D( "NHitsDistributionHisto",
					"NHitsDistributionHisto;number of hits on a track;# tracks",
					1024, _histogramMinValueNHits, _histogramMaxValueNHits);

    }

void TrackParametersDistributionProcessor::processRunHeader(lcio::LCRunHeader*  /*run*/ )
{
}
    
void TrackParametersDistributionProcessor::processEvent(lcio::LCEvent * evt ) 
{ 
    //read input data
    streamlog_out(DEBUG3) << "Event: " << evt->getEventNumber() << std::endl;

    // we need a LCCollectionVec to use iterators, as LCCollectio is not a std::vector
    LCCollectionVec* InputCollection=0;
    try{
	InputCollection = dynamic_cast<LCCollectionVec*>(evt->getCollection( _inputColName )) ;
    }
    catch(DataNotAvailableException &e)
    {
	// no input data
	// just continue ...
	streamlog_out(MESSAGE)<<"inputCollection: "<< _inputColName <<"  could not be found in event "
			       << evt->getEventNumber() <<std::endl;
	return;
    }


    // cast did not succeed
    if( InputCollection == 0 )
    {
	streamlog_out(ERROR2) << "InputCollection vector is 0" << std::endl;
	return;
    }

    for(LCCollectionVec::iterator it = InputCollection->begin() ; 
	it < InputCollection->end(); it ++ )
    {
	
	Track * track = dynamic_cast<Track *>(*it) ;
	
	_distributionPhiHisto->Fill(track->getPhi());
	_distributionLambdaHisto->Fill( std::atan( track->getTanLambda() ));
	_distributionTanLambdaHisto->Fill( track->getTanLambda() );
	_distributionD0Histo->Fill( track->getD0() ) ;
	_distributionZ0Histo->Fill( track->getZ0() ) ;
	_distributionOmegaHisto->Fill( track->getOmega() ) ;
	_distributionChi2Histo->Fill( track->getChi2() ) ;
	_distributionNdfHisto->Fill( track->getNdf() ) ;
	_distributionDEdxHisto->Fill( track->getdEdx() ) ;
	_distributionNHitsHisto->Fill( track->getTrackerHits().size() ) ;
	_distributionDEdxVsOmegaHisto->Fill(track->getOmega() ,track->getdEdx() );
	if(track->getOmega()!=0)
		_distributionMomentumHisto->Fill( abs(_bFieldStrength*0.3/1000/track->getOmega()) ) ;

	streamlog_out(DEBUG0)<<"phi = "<<track->getPhi()<<"\t lambda = "<<std::atan( track->getTanLambda() )
			     <<"\nd0 = "<<track->getD0()<<"\tz0 = "<<track->getZ0()
			     <<"\nChi^2 = "<<track->getChi2()<<"\tndf = "<<track->getNdf()
			     <<"\ndEdx = "<<track->getdEdx()<<"\tNHits = "<<track->getTrackerHits().size()<<std::endl;

    }
    
}


}// namespace 
