#include "WriteXrayObjectsToROOTTreeProcessor.h"


namespace marlintpc{

WriteXrayObjectsToROOTTreeProcessor aWriteXrayObjectsToROOTTreeProcessor;

WriteXrayObjectsToROOTTreeProcessor::WriteXrayObjectsToROOTTreeProcessor() : Processor("WriteXrayObjectsToROOTTreeProcessor")
{

	_runType = new int;

	_description = "";

	registerInputCollection(LCIO::LCRELATION,"InputXrayObjectCollection","Collection containing the XrayObjects to be exported as root tree",_inputXrayObjectCollectionName,std::string("TimepixXrayObjects"));

	registerProcessorParameter("OutputROOTFileName","File name for the root file to be produced",_rootFileName,std::string("XrayObjects.root"));

	registerProcessorParameter("OutputROOTFileDescription","",_rootFileDescription,std::string(""));

	registerProcessorParameter("OutputROOTTreeDescription","",_rootTreeDescription,std::string(""));

	registerProcessorParameter("RunType","Type of run (0: background, 1: calibration, 2: x-ray finger, add a leading 1 for data of category testing)",*_runType,int(0));

}

WriteXrayObjectsToROOTTreeProcessor::~WriteXrayObjectsToROOTTreeProcessor()
{

}

void WriteXrayObjectsToROOTTreeProcessor::init()
{

	_eventCounter = 0;
	_runCounter = 0;

	_rootFile = new TFile(_rootFileName.c_str(),"RECREATE",_rootFileDescription.c_str());

	_rootTree = new TTree(_inputXrayObjectCollectionName.c_str(),_rootTreeDescription.c_str());

	//register branches
	_runNumber = new int;
	_rootTree->Branch("RunNumber",_runNumber);

	_eventNumber = new int;
	_rootTree->Branch("EventNumber",_eventNumber);

	_timestamp = new ULong64_t;
	_rootTree->Branch("Timestamp",_timestamp);

	//run type is a processor parameter
	_rootTree->Branch("RunType",_runType);

	_pixelMap = new std::map<int,double>;
	//_rootTree->Branch("EventMap",_pixelMap);

	_xVector = new std::vector<int>;
	_rootTree->Branch("XCoordinatesVector",_xVector);

	_yVector = new std::vector<int>;
	_rootTree->Branch("YCoordinatesVector",_yVector);

	_valueVector = new std::vector<double>;
	_rootTree->Branch("ChargeValuesVector",_valueVector);

	_numberOfPixels = new int;
	_rootTree->Branch("NumberOfPixels",_numberOfPixels);

	_totalCharge = new int;
	_rootTree->Branch("TotalCharge",_totalCharge);

	_energyFromCharge =  new double;
	_rootTree->Branch("EnergyFromCharge",_energyFromCharge);

	_energyFromPixels = new double;
	_rootTree->Branch("EnergyFromPixels",_energyFromPixels);

	_positionX = new double;
	_rootTree->Branch("PositionX",_positionX);

	_positionY = new double;
	_rootTree->Branch("PositionY",_positionY);

	_rotationAngle = new double;
	_rootTree->Branch("RotationAngle",_rotationAngle);

	_rmsLongitudinal = new double;
	_rootTree->Branch("RmsLongitudinal",_rmsLongitudinal);

	_rmsTransverse = new double;
	_rootTree->Branch("RmsTransverse",_rmsTransverse);

	_excentricity = new double;
	_rootTree->Branch("Excentricity",_excentricity);

	_radius = new double;
	_rootTree->Branch("Radius",_radius);

	_width = new double;
	_rootTree->Branch("Width",_width);

	_length = new double;
	_rootTree->Branch("Length",_length);

	_skewnessLongitudinal = new double;
	_rootTree->Branch("SkewnessLongitudinal",_skewnessLongitudinal);

	_skewnessTransverse = new double;
	_rootTree->Branch("SkewnessTransverse",_skewnessTransverse);

	_kurtosisLongitudinal = new double;
	_rootTree->Branch("KurtosisLongitudinal",_kurtosisLongitudinal);

	_kurtosisTransverse = new double;
	_rootTree->Branch("KurtosisTransverse",_kurtosisTransverse);

	_likelihoodMarlin = new double;
	_rootTree->Branch("LikelihoodMarlin",_likelihoodMarlin);

	_fractionWithinRmsTransverse = new double;
	_rootTree->Branch("FractionWithinRmsTransverse",_fractionWithinRmsTransverse);

	_geometryService = new XrayGeometryService();

}

void WriteXrayObjectsToROOTTreeProcessor::processRunHeader(lcio::LCRunHeader * /*run*/)
{

	printParameters();

	_runCounter++;

}

void WriteXrayObjectsToROOTTreeProcessor::processEvent(lcio::LCEvent *evt)
{


	lcio::LCCollectionVec* inputXrayObjectCollection(NULL);

	try{

		inputXrayObjectCollection = dynamic_cast<lcio::LCCollectionVec*>(evt->getCollection(_inputXrayObjectCollectionName));

	}
	catch(DataNotAvailableException &e){

		streamlog_out(MESSAGE3) << "Input collection " << _inputXrayObjectCollectionName << " is not available in event " << evt->getEventNumber() << " of run " << evt->getRunNumber() << std::endl;
		return;

	}

	if(inputXrayObjectCollection == NULL){

		streamlog_out(ERROR2) << "Input collection " << _inputXrayObjectCollectionName << ": Vector is NULL!" << std::endl;
		return;

	}

	for(lcio::LCCollectionVec::iterator iterXrayObject = inputXrayObjectCollection->begin(); iterXrayObject != inputXrayObjectCollection->end(); iterXrayObject++)
	{
		_eventCounter++;

		lcio::LCRelation* xrayObject = dynamic_cast<lcio::LCRelation*>(*iterXrayObject);

		GridPixXrayObject* xray = new GridPixXrayObject(dynamic_cast<lcio::LCGenericObject*>(xrayObject->getFrom()));

		lcio::TrackerHit* hit = dynamic_cast<lcio::TrackerHit*>(xrayObject->getTo());

		*_runNumber = evt->getRunNumber();
		*_eventNumber = evt->getEventNumber();
		*_timestamp = (ULong64_t)evt->getTimeStamp();

		*_numberOfPixels = xray->getPixels();
		*_totalCharge = xray->getCharge();

		*_positionX = xray->getCenterX();
		*_positionY = xray->getCenterY();

		*_energyFromCharge = xray->getEnergyFromCharge();
		*_energyFromPixels = xray->getEnergyFromPixels();

		*_likelihoodMarlin = xray->getLikelihood();

		*_rotationAngle = xray->getRotAngle();

		_geometryService->calculateGeometry(hit,xray->getRotAngle());

		*_rmsLongitudinal = _geometryService->getRmsX();
		*_rmsTransverse = _geometryService->getRmsY();

		*_excentricity = _geometryService->getExcentricity();

		*_radius = _geometryService->getRadius();

		*_length = _geometryService->getLength();
		*_width = _geometryService->getWidth();

		*_skewnessLongitudinal = _geometryService->getSkewnessL();
		*_skewnessTransverse = _geometryService->getSkewnessT();

		*_kurtosisLongitudinal = _geometryService->getKurtosisL();
		*_kurtosisTransverse = _geometryService->getKurtosisT();

		*_fractionWithinRmsTransverse = (double)(_geometryService->numberPixelsWithinRadius(*_rmsTransverse)) / (double)(*_numberOfPixels);

		writeHitToMap(_pixelMap,hit);
		writeHitToVectors(_xVector,_yVector,_valueVector,hit);

		_rootTree->Fill();

		delete xray;
	}



}

void WriteXrayObjectsToROOTTreeProcessor::writeHitToMap(std::map<int,double> *map, lcio::TrackerHit *hit)
{

	//clear map before putting anything into it
	map->clear();

	lcio::LCObjectVec const & dataVec = hit->getRawHits();

	for( lcio::LCObjectVec::const_iterator iterData = dataVec.begin(); iterData != dataVec.end(); iterData++)
	{

		lcio::TrackerData* data = dynamic_cast<lcio::TrackerData *>(*iterData);

		int cellID0 = data->getCellID0();

		lcio::FloatVec pixelVec = data->getChargeValues();

		for(lcio::FloatVec::iterator iterPixel = pixelVec.begin(); iterPixel != pixelVec.end(); iterPixel++)
		{

			cellID0++;

			(*map)[cellID0] = (double)*iterPixel;

		}

	}

}

void WriteXrayObjectsToROOTTreeProcessor::writeHitToVectors(std::vector<int> *xVec, std::vector<int> *yVec, std::vector<double> *valVec, lcio::TrackerHit *hit)
{
	//clear vectors before putting anything into it
	xVec->clear();
	yVec->clear();
	valVec->clear();

	lcio::LCObjectVec const & dataVec = hit->getRawHits();

	int chipID = dynamic_cast<lcio::TrackerData *>(dataVec[0])->getCellID1();

	gear::TPCModule const &module = marlin::Global::GEAR->getTPCParameters().getModule( chipID );

	// get some geometry information from gear
	// This processor implicitly assumes a rectangular matrix of pads/pixels
	int nRows = module.getNRows();
	int nColumns = module.getNPads() / nRows;

	for( lcio::LCObjectVec::const_iterator iterData = dataVec.begin(); iterData != dataVec.end(); iterData++)
	{

		lcio::TrackerData* data = dynamic_cast<lcio::TrackerData *>(*iterData);

		int cellID0 = data->getCellID0();

		lcio::FloatVec pixelVec = data->getChargeValues();

		for(lcio::FloatVec::iterator iterPixel = pixelVec.begin(); iterPixel != pixelVec.end(); iterPixel++)
		{

			cellID0++;

			int x = cellID0 % nColumns;
			int y = (cellID0 - x) / nColumns;

			xVec->push_back(x);
			yVec->push_back(y);
			valVec->push_back((double)*iterPixel);

		}

	}
}

void WriteXrayObjectsToROOTTreeProcessor::end()
{

	_rootFile->Write();
	_rootFile->Close();

	//clean up
	delete _rootFile;
	//_rootFile has ownership of _rootTree and therefore takes care of deleting it when _rootFile is deleted!

	delete _runNumber;
	delete _eventNumber;

	delete _pixelMap;
	delete _xVector;
	delete _yVector;
	delete _valueVector;

	delete _numberOfPixels;
	delete _totalCharge;

	delete _energyFromCharge;
	delete _energyFromPixels;

	delete _positionX;
	delete _positionY;

	delete _rotationAngle;

	delete _rmsTransverse;
	delete _rmsLongitudinal;

	delete _excentricity;

	delete _radius;

	delete _width;
	delete _length;

	delete _skewnessTransverse;
	delete _skewnessLongitudinal;

	delete _kurtosisTransverse;
	delete _kurtosisLongitudinal;

	delete _likelihoodMarlin;

	delete _fractionWithinRmsTransverse;

	delete _geometryService;

	delete _runType;

	streamlog_out(MESSAGE4) << "Wrote " << _eventCounter << " XrayObjects from " << _runCounter << " run(s) to " << _rootFileName << std::endl;

}

}//namespace marlintpc
