#include "XrayAnalysisProcessor.h"

//lcio
#include <lcio.h>
#include <EVENT/LCCollection.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/LCRelation.h>
#include <EVENT/TrackerHit.h>

// Aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>

//marlin
#include <marlin/Global.h>

//c++
#include <sstream>

namespace marlintpc{

XrayAnalysisProcessor aXrayAnalysisProcessor;

XrayAnalysisProcessor::XrayAnalysisProcessor() : Processor("XrayAnalysisProcessor"), _pixelsHisto(NULL), _chargeHisto(NULL), _energyHisto(NULL), _likelihoodHisto(NULL), _centerMap(NULL), _blindRangeTimestamps(2,1414573172), _split(0.), _histoXraysPerEvent(NULL){

	_description = "Creates histograms for XrayObjects";

	registerInputCollection(LCIO::LCRELATION,"InputCollection","InputCollection containing XrayObjects",_inputCol,std::string("TimepixXrayObjects"));

	registerProcessorParameter("EnergySplit","If true likelihood and center position will be histogramed for 10 energy bins",_energySplit,(bool) false);

	registerProcessorParameter("BinsPixels","Number of bins for Pixels histogram",_binsPixels,(int) 200);
	registerProcessorParameter("MinPixels","Minimum number for Pixels histogram",_minPixels,(double) 0.);
	registerProcessorParameter("MaxPixels","Maximum number for pixels histogram",_maxPixels,(double) 500.);

	registerProcessorParameter("BinsCharge","Number of bins for Charge histogram",_binsCharge,(int) 200);
	registerProcessorParameter("MinCharge","Minimum number for Charge histogram",_minCharge,(double) 0.);
	registerProcessorParameter("MaxCharge","Maximum number for pixels histogram",_maxCharge,(double) 4.e6);

	registerProcessorParameter("BinsLikelihood","Number of bins for Likelihood histogram",_binsLikelihood,(int) 200);
	registerProcessorParameter("MinLikelihood","Minimum number for Likelihood histogram",_minLikelihood,(double) -40.);
	registerProcessorParameter("MaxLikelihood","Maximum number for pixels histogram",_maxLikelihood,(double) +10.);

	registerProcessorParameter("BinsEnergy","Number of bins for Energy histogram",_binsEnergy,(int) 100);
	registerProcessorParameter("MinEnergy","Minimum number for Energy histogram",_minEnergy,(double) 0.);
	registerProcessorParameter("MaxEnergy","Maximum number for pixels histogram",_maxEnergy,(double) 10.);

	registerOptionalParameter("BlindRangeTimeStamps","Events with timestamps in the defined range will be ignored if set",_blindRangeTimestamps,_blindRangeTimestamps);

}

XrayAnalysisProcessor::~XrayAnalysisProcessor(){}

void XrayAnalysisProcessor::init(){

	printParameters();

	_pixelsHisto = new TH1D("pixels","Number of pixels",_binsPixels,_minPixels,_maxPixels);
	_chargeHisto = new TH1D("charge","Charge",_binsCharge,_minCharge,_maxCharge);
	_energyHisto = new TH1D("energy","Energy",_binsEnergy,_minEnergy,_maxEnergy);
	_likelihoodHisto = new TH1D("likelihood","Likelihood",_binsLikelihood,_minLikelihood,_maxLikelihood);

	_centerMap = new TH2D("center","Center Positions",100,0.,14.,100,0.,14.);

	_histoXraysPerEvent = new TH1D("xrayperevent","Xrays per event",6,-0.5,5.5);

	if(_energySplit){

		_split = (_maxEnergy - _minEnergy) / 10.;

		std::stringstream histoNameCenter;
		std::stringstream histoTitleCenter;

		std::stringstream histoNameLikelihood;
		std::stringstream histoTitleLikelihood;

		for(int i = 0; i < 10; i++)
		{

			histoNameCenter << "center" << (_minEnergy + i*_split) << "keV";
			histoTitleCenter << "Center positions for " << (_minEnergy + i*_split) << " keV to " << (_minEnergy + (i+1)*_split) << " keV";

			_centerMapVec.push_back(new TH2D(histoNameCenter.str().c_str(),histoTitleCenter.str().c_str(),100,0.,14.,100,0.,14.));

			histoNameCenter.str("");
			histoNameCenter.clear();

			histoTitleCenter.str("");
			histoTitleCenter.clear();

			histoNameLikelihood << "likelihood" << (_minEnergy + i*_split) << "keV";
			histoTitleLikelihood << "Likelihood for " << (_minEnergy + i*_split) << " keV to " << (_minEnergy + (i+1)*_split) << " keV";

			_likelihoodHistoVec.push_back(new TH1D(histoNameLikelihood.str().c_str(),histoTitleLikelihood.str().c_str(),_binsLikelihood,_minLikelihood,_maxLikelihood));

			histoNameLikelihood.str("");
			histoNameLikelihood.clear();

			histoTitleLikelihood.str("");
			histoTitleLikelihood.clear();

		}

	}

	if(parameterSet("BlindRangeTimeStamps") && (_blindRangeTimestamps.size() > 2) )
	{
		streamlog_out(WARNING4) << "More than two values given for range to blind, first and second value are treated as range, rest is ignored!" << std::endl;
	}

	_nBlinded = 0;

}

void XrayAnalysisProcessor::processEvent( LCEvent* evt ){

	lcio::LCCollectionVec* inputCol(NULL);

	try{

		inputCol = dynamic_cast<lcio::LCCollectionVec*>(evt->getCollection(_inputCol));

	}
	catch(DataNotAvailableException &e){

		streamlog_out(MESSAGE3) << "InputCollection not available" << std::endl;
		return;

	}

	if(inputCol == NULL){

		streamlog_out(ERROR2) << "InputCollection vector is 0" << std::endl;
		return;

	}

	if(parameterSet("BlindRangeTimeStamps") && (evt->getTimeStamp() > (long int)_blindRangeTimestamps[0]) && (evt->getTimeStamp() < (long int)_blindRangeTimestamps[1]) )
	{
		streamlog_out(MESSAGE4) << "Blinding event " << evt->getEventNumber() << " (" << evt->getTimeStamp() << ")" << std::endl;
		_nBlinded++;
		return;
	}

	_histoXraysPerEvent->Fill(inputCol->size());

	for(lcio::LCCollectionVec::iterator iterCol = inputCol->begin(); iterCol != inputCol->end(); iterCol++)
	{

		lcio::LCRelation* xrayObject = dynamic_cast<lcio::LCRelation*>(*iterCol);

		GridPixXrayObject* xray = new GridPixXrayObject(dynamic_cast<lcio::LCGenericObject*>(xrayObject->getFrom()));

		_pixelsHisto->Fill(xray->getPixels());
		_chargeHisto->Fill(xray->getCharge());
		_energyHisto->Fill(xray->getEnergyFromCharge());
		_likelihoodHisto->Fill(xray->getLikelihood());

		_centerMap->Fill(xray->getCenterX(),xray->getCenterY());

		if(_energySplit){

			int histNumber = static_cast<int>((xray->getEnergyFromCharge() - _minEnergy) / _split);

			if( histNumber < 10 ){

				_centerMapVec[histNumber]->Fill(xray->getCenterX(),xray->getCenterY());
				_likelihoodHistoVec[histNumber]->Fill(xray->getLikelihood());

			}
		}

		delete xray;

	}

}

void XrayAnalysisProcessor::end(){

	marlin::AIDAProcessor::tree(this);

	_pixelsHisto->Write();
	delete _pixelsHisto;

	_chargeHisto->Write();
	delete _chargeHisto;

	_energyHisto->Write();
	delete _energyHisto;

	_likelihoodHisto->Write();
	delete _likelihoodHisto;

	_centerMap->Write();
	delete _centerMap;

	_histoXraysPerEvent->Write();
	delete _histoXraysPerEvent;

	if(_energySplit){

		for(int i = 0; i < 10; i++)
		{

			_likelihoodHistoVec[i]->Write();
			delete _likelihoodHistoVec[i];

			_centerMapVec[i]->Write();
			delete _centerMapVec[i];

		}

	}

	if(parameterSet("BlindRangeTimeStamps"))
	{
		streamlog_out(MESSAGE4) << _nBlinded << " events have been blinded." << std::endl;
	}

}

}//namespace marlintpc
