#include "XrayCutApplicationProcessor.h"

//c++
#include <cmath>
#include <iostream>

//lcio
#include <EVENT/LCCollection.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCFlagImpl.h>

//marlin
#include <marlin/Global.h>

namespace marlintpc{

XrayCutApplicationProcessor aXrayCutApplicationProcessor;

XrayCutApplicationProcessor::XrayCutApplicationProcessor() : Processor("XrayCutApplicationProcessor"){

	_description = "applies cuts on GridPixXrayObjects";

	registerInputCollection(LCIO::LCRELATION,"InputCollection","Input collection containing XrayObjects",_inputCol,std::string("TimepixXrayObjects"));

	//registerInputCollection(LCIO::TRACKERHIT,"InputHits","Input collection containing Hits associated with GridPixXrayObjects",_inputHits,std::string("XrayHits"));

	registerOutputCollection(LCIO::LCRELATION,"OutputCollection","Output collection containing XrayObjects",_outputCol,std::string("TimepixXrayObjectsCut"));

	//registerOutputCollection(LCIO::TRACKERHIT,"OutputHits","Output collection containing Hits associated with GridPixXrayObjects",_outputHits,std::string("XrayHitsCut"));

	registerProcessorParameter( "SetOutputTransient","if not 0 the output collection is set transient",_outputIsTransient,0);

	registerProcessorParameter("CenterX","x-position of center",_centerX,(double) 6.9);

	registerProcessorParameter("CenterY","y-position of center",_centerY,(double) 8.4);

	registerOptionalParameter("CutRadius","radius around center, dots within this area will be kept",_cutRadius,(double) 2.);

	registerOptionalParameter("LowCutX","low cut on x-position",_lowCutX,(double) 0.);

	registerOptionalParameter("HighCutX","high cut on x-position",_highCutX,(double) 14.);

	registerOptionalParameter("LowCutY","low cut on y-position",_lowCutY,(double) 0.);

	registerOptionalParameter("HighCutY","high cut on y-position",_highCutY,(double) 14.);

	registerOptionalParameter("LowCutExcentricity","low cut on excentricity",_lowCutExcentricity,(double) 1.);

	registerOptionalParameter("HighCutExcentricity","high cut on excentricity",_highCutExcentricity,(double) 1.3);

	registerOptionalParameter("LowCutTransverseSkewness","low cut on transverse skewness",_lowCutSkewnessT,(double) -2.);

	registerOptionalParameter("HighCutTransverseSkewness","high cut on transverse skewness",_highCutSkewnessT,(double) 2.);

	registerOptionalParameter("LowCutTransverseKurtosis","low cut on transverse kurtosis",_lowCutKurtosisT,(double) -2.);

	registerOptionalParameter("HighCutTransverseKurtosis","high cut on transverse kurtosis",_highCutKurtosisT,(double) 2.);

	registerOptionalParameter("LowCutLongitudinalKurtosis","low cut on longitudinal kurtosis",_lowCutKurtosisL,(double) -1.);

	registerOptionalParameter("HighCutLongitudinalKurtosis","high cut on longitudinal kurtosis",_highCutKurtosisL,(double) 1.);

	registerOptionalParameter("LowCutTransverseRms","low cut on transverse rms",_lowCutRmsY,(double) 0.);

	registerOptionalParameter("HighCutTransverseRms","high cut on transverse rms",_highCutRmsY,(double) 1.);

	registerOptionalParameter("LowCutLength","low cut on length",_lowCutLength,(double) 0.);

	registerOptionalParameter("HighCutLength","high cut on length",_highCutLength,(double) 10.);

	registerOptionalParameter("LowCutNpx","low cut on number of pixel",_lowCutPixels,(int) 0);

	registerOptionalParameter("HighCutNpx","high cut on number of pixel",_highCutPixels,(int) 300);

	registerOptionalParameter("LowCutCharge","low cut on charge",_lowCutCharge,(int) 0);

	registerOptionalParameter("HighCutCharge","high cut on charge",_highCutCharge,(int) 50000);

	registerOptionalParameter("LowCutEnergy","low cut on energy",_lowCutEnergy,(double) 0.);

	registerOptionalParameter("HighCutEnergy","high cut on energy",_highCutEnergy,(double) 10.);

	registerOptionalParameter("LowCutLikelihood","low cut on likelihood",_lowCutLikelihood,(double) 0.);

	registerOptionalParameter("HighCutLikelihood","high cut on likelihood",_highCutLikelihood,(double) 10.);

	registerOptionalParameter("HighCutNumberXrays","high cut on number of xrays in event",_highCutNumberXrays,(int) 1);

}

void XrayCutApplicationProcessor::init(){

	printParameters();

	_geoService = new XrayGeometryService();

	_numberBefore = 0;

	_numberAfter = 0;

}

void XrayCutApplicationProcessor::processRunHeader( LCRunHeader* run){


	run->parameters().setValue( _processorName + "_revision", "$Rev: 007 $");

	for ( marlin::ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
		if ( ! i->second->isOptional() || i->second->valueSet() ) {
			run->parameters().setValue( _processorName + "_" + i->second->name(),
					i->second->value() );
		}
	}

}

void XrayCutApplicationProcessor::processEvent( LCEvent* evt ){

	lcio::LCCollectionVec* inputCol(NULL);

	//lcio::LCCollectionVec* inputXrays(NULL);
	//lcio::LCCollectionVec* inputHits(NULL);

	try{

		inputCol = dynamic_cast<lcio::LCCollectionVec*>(evt->getCollection(_inputCol));

		//inputXrays = dynamic_cast<lcio::LCCollectionVec*>(evt->getCollection(_inputXrays));
		//inputHits = dynamic_cast<LCCollectionVec*>(evt->getCollection(_inputHits));
	}
	catch(DataNotAvailableException &e){
		streamlog_out(MESSAGE3) << "Input collection is unavailable" << std::endl;
		return;
	}

	if(inputCol == NULL){

		streamlog_out(ERROR2) << "Input collection vector is 0" << std::endl;
		return;

	}

	if(parameterSet("HighCutNumberXrays") && (int)inputCol->size() > _highCutNumberXrays){
		return;
	}

	LCCollectionVec* outputCol = new LCCollectionVec(LCIO::LCRELATION);

	//LCCollectionVec* outputXrays = new LCCollectionVec(LCIO::LCGENERICOBJECT);
	//LCCollectionVec* outputHits = new LCCollectionVec(LCIO::TRACKERHIT);

	//outputXrays->setTransient( (_outputIsTransient != 0) );

	outputCol->setTransient( (_outputIsTransient != 0) );
	// the output collection is a subset of the input collection
	// only the pointers are stored, the hit is not copied
	outputCol->setSubset();

	for(lcio::LCCollectionVec::iterator inputIter = inputCol->begin(); inputIter != inputCol->end(); inputIter++)
	{

		_numberBefore++;

		lcio::LCRelation* xrayObject = dynamic_cast<lcio::LCRelation*>(*inputIter);

		GridPixXrayObject* xray = new GridPixXrayObject(dynamic_cast<LCGenericObject*>(xrayObject->getFrom()));
		lcio::TrackerHit* hit = dynamic_cast<lcio::TrackerHit*>(xrayObject->getTo());

		double distance = std::sqrt((xray->getCenterX() - _centerX)*(xray->getCenterX() - _centerX) + (xray->getCenterY() - _centerY)*(xray->getCenterY() - _centerY));

		if( parameterSet("CutRadius") && distance > _cutRadius){
			delete xray;
			continue;
		}

		if( parameterSet("LowCutX") && xray->getCenterX() < _lowCutX){
			delete xray;
			continue;
		}

		if( parameterSet("HighCutX") && xray->getCenterX() > _highCutX){
			delete xray;
			continue;
		}

		if( parameterSet("LowCutY") && xray->getCenterY() < _lowCutY){
			delete xray;
			continue;
		}

		if( parameterSet("HighCutY") && xray->getCenterY() > _highCutY){
			delete xray;
			continue;
		}

		if( parameterSet("LowCutNpx") && xray->getPixels() < _lowCutPixels){
			delete xray;
			continue;
		}

		if( parameterSet("HighCutNpx") && xray->getPixels() > _highCutPixels){
			delete xray;
			continue;
		}

		if( parameterSet("LowCutCharge") && xray->getCharge() < _lowCutCharge){
			delete xray;
			continue;
		}

		if( parameterSet("HighCutCharge") && xray->getCharge() > _highCutCharge){
			delete xray;
			continue;
		}

		if( parameterSet("LowCutEnergy") && xray->getEnergyFromCharge() < _lowCutEnergy){
			delete xray;
			continue;
		}

		if( parameterSet("HighCutEnergy") && xray->getEnergyFromCharge() > _highCutEnergy){
			delete xray;
			continue;
		}

		if( parameterSet("LowCutLikelihood") && xray->getLikelihood() < _lowCutLikelihood){
			delete xray;
			continue;
		}

		if( parameterSet("HighCutLikelihood") && xray->getLikelihood() > _highCutLikelihood){
			delete xray;
			continue;
		}

		//TODO: calculate Geometry only if needed!
		_geoService->calculateGeometry(hit,xray->getRotAngle());

		if( parameterSet("LowCutExcentricity") && _geoService->getExcentricity() < _lowCutExcentricity){
			delete xray;
			continue;
		}

		if( parameterSet("HighCutExcentricity") && _geoService->getExcentricity() > _highCutExcentricity){
			delete xray;
			continue;
		}

		if( parameterSet("LowCutTransverseSkewness") && _geoService->getSkewnessT() < _lowCutSkewnessT){
			delete xray;
			continue;
		}

		if( parameterSet("HighCutTransverseSkewness") && _geoService->getSkewnessT() > _highCutSkewnessT){
			delete xray;
			continue;
		}

		if( parameterSet("LowCutTransverseKurtosis") && _geoService->getKurtosisT() < _lowCutKurtosisT){
			delete xray;
			continue;
		}

		if( parameterSet("HighCutTransverseKurtosis") && _geoService->getKurtosisT() > _highCutKurtosisT){
			delete xray;
			continue;
		}

		if( parameterSet("LowCutLongitudinalKurtosis") && _geoService->getKurtosisL() < _lowCutKurtosisL){
			delete xray;
			continue;
		}

		if( parameterSet("HighCutLongitudinalKurtosis") && _geoService->getKurtosisL() > _highCutKurtosisL){
			delete xray;
			continue;
		}

		if( parameterSet("LowCutTransverseRms") && _geoService->getRmsY() < _lowCutRmsY){
			delete xray;
			continue;
		}

		if( parameterSet("HighCutTransverseRms") && _geoService->getRmsY() > _highCutRmsY){
			delete xray;
			continue;
		}

		if( parameterSet("LowCutLength") && _geoService->getLength() < _lowCutLength){
			delete xray;
			continue;
		}

		if( parameterSet("HighCutLength") && _geoService->getLength() > _highCutLength){
			delete xray;
			continue;
		}

		_numberAfter++;

		delete xray;

		outputCol->addElement(xrayObject);

		//outputXrays->addElement(xray);
		//outputHits->addElement(hit);

	}

	if(!outputCol->empty()){

		evt->addCollection(outputCol,_outputCol);

		//evt->addCollection(outputXrays,_outputXrays);
		//evt->addCollection(outputHits,_outputHits);

	}
	else{

		delete outputCol;

		//delete outputXrays;
		//delete outputHits;

	}

}

void XrayCutApplicationProcessor::end(){

	delete _geoService;

	streamlog_out(MESSAGE4) << _numberAfter << " of " << _numberBefore << " passed the cuts applied!" << std::endl;

}

}//namespace marlintpc
