#include "XrayEnergyCalibrationProcessor.h"

namespace marlintpc{
  
  XrayEnergyCalibrationProcessor aXrayEnergyCalibrationProcessor;
  
  XrayEnergyCalibrationProcessor::XrayEnergyCalibrationProcessor() : Processor("XrayEnergyCalibrationProcessor"), _energyFromChargeParameterVector(2,0), _energyFromPixelsParameterVector(2,0), _energyFromChargeHistogram(NULL), _energyFromPixelsHistogram(NULL){
    
    _description = "Derives energy from total charge and number of pixels of an XrayObject and stores value from both methods separately in new XrayObjects";
    
    registerInputCollection(LCIO::LCRELATION,"InputCollection","InputCollection containing XrayObjects",_inputCollection,std::string("TimepixXrayObjects"));
    
    registerOutputCollection(LCIO::LCRELATION,"OutputCollection","OutputCollection containing XrayObjects",_outputCollection,std::string("TimepixXrayObjectsCalibrated"));
    
    registerOutputCollection(LCIO::LCGENERICOBJECT,"OutputXrays","OutputCollection containing calibrated GridPixXrayObjects",_outputXrays,std::string("GridPixXrayObjectsCalibrated"));

    registerProcessorParameter("EnergyFromChargeParameters","Parameters for the conversion of total charge to energy, number of parameters defines the degree of the polynom used",_energyFromChargeParameterVector,_energyFromChargeParameterVector);

    registerProcessorParameter("EnergyFromPixelsParameters","Parameters for the conversion of number of pixels to energy, number of parameters defines the degree of the polynom used",_energyFromPixelsParameterVector,_energyFromPixelsParameterVector);
    
    registerProcessorParameter( "SetOutputTransient","if not 0 the output collection is set transient",_outputIsTransient,0);
    
  }
  
  XrayEnergyCalibrationProcessor::~XrayEnergyCalibrationProcessor(){}
  
  double XrayEnergyCalibrationProcessor::energy(int value, std::vector<float>* parameters){

    //energy calibration: charge [electrons] -> energy [keV] or number of pixels -> energy [keV]
    double e(0.);
    
    for(int i = 0; i < (int)parameters->size(); i++)
    {
    	e += parameters->at(i) * std::pow((double)value, i);
    }
    
    return e;
    
  }
  
  void XrayEnergyCalibrationProcessor::init(){
    
    printParameters();
    
    _energyFromChargeHistogram = new TH1D("EnergyFromCharge","EnergyFromCharge",200,0.,20.);
    _energyFromPixelsHistogram = new TH1D("EnergyFromPixels","EnergyFromPixels",200,0.,20.);
    
  }
  
  void XrayEnergyCalibrationProcessor::processRunHeader( LCRunHeader* run) { 
    
    
    run->parameters().setValue( _processorName + "_revision", "$Rev: 007 $");
    
    for ( marlin::ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
      if ( ! i->second->isOptional() || i->second->valueSet() ) {
	run->parameters().setValue( _processorName + "_" + i->second->name(), 
				    i->second->value() );
      }
    }
    
  }
  
  void XrayEnergyCalibrationProcessor::processEvent( LCEvent * evt ){
    
    lcio::LCCollectionVec* inputCol(NULL);
    
    try{
      
      inputCol = dynamic_cast<lcio::LCCollectionVec*>(evt->getCollection(_inputCollection));
      
    }
    catch(DataNotAvailableException &e){
      
      streamlog_out(MESSAGE3) << "InputCollection not available" << std::endl;
      return;
      
    }
    
    if(inputCol == NULL){
      
      streamlog_out(ERROR2) << "InputCollection vector is 0" << std::endl;
      return;
      
    }
    
    lcio::LCCollectionVec* outputXrays = new lcio::LCCollectionVec(LCIO::LCGENERICOBJECT);
    lcio::LCCollectionVec* outputCol = new lcio::LCCollectionVec(LCIO::LCRELATION);
    
    outputXrays->setTransient( (_outputIsTransient != 0) );
    outputCol->setTransient( (_outputIsTransient != 0) );
    
    for(lcio::LCCollectionVec::iterator iterCol = inputCol->begin(); iterCol != inputCol->end(); iterCol++)
    {
      
      lcio::LCRelation* xrayObject = dynamic_cast<lcio::LCRelation*>(*iterCol);

      GridPixXrayObject* xray = new GridPixXrayObject(dynamic_cast<lcio::LCGenericObject*>(xrayObject->getFrom()));
      
      double xrayEnergyFromCharge(0.);
      double xrayEnergyFromPixels(0.);
      
      xrayEnergyFromCharge = energy(xray->getCharge(),&_energyFromChargeParameterVector);
      xrayEnergyFromPixels = energy(xray->getPixels(),&_energyFromPixelsParameterVector);

      GridPixXrayObject* newXray = new GridPixXrayObject;

      newXray->setCenterX(xray->getCenterX());
      newXray->setCenterY(xray->getCenterY());
      newXray->setCharge(xray->getCharge());
      newXray->setEnergyFromCharge(xrayEnergyFromCharge);
      newXray->setEnergyFromPixels(xrayEnergyFromPixels);
      newXray->setLikelihood(xray->getLikelihood());
      newXray->setPixels(xray->getPixels());
      newXray->setRotAngle(xray->getRotAngle());

      delete xray;
      
      _energyFromChargeHistogram->Fill(xrayEnergyFromCharge);
      _energyFromPixelsHistogram->Fill(xrayEnergyFromPixels);
     
      outputXrays->addElement(newXray);

      lcio::LCRelationImpl* xrayObjectCalibrated = new LCRelationImpl(newXray,xrayObject->getTo());
     
      outputCol->addElement(xrayObjectCalibrated);
      

    }

    if(!outputCol->empty())
    {
    	evt->addCollection(outputXrays,_outputXrays);
    	evt->addCollection(outputCol,_outputCollection);
    }
    else
    {
    	delete outputXrays;
    	delete outputCol;
    }
  }
  
  void XrayEnergyCalibrationProcessor::end(){
    
    marlin::AIDAProcessor::tree(this);
    _energyFromChargeHistogram->Write();
    _energyFromPixelsHistogram->Write();
    delete _energyFromChargeHistogram;
    delete _energyFromPixelsHistogram;
    
  }  
  
}//namespace marlintpc
