#include "XrayEventDisplayProcessor.h"

namespace marlintpc{

XrayEventDisplayProcessor aXrayEventDisplayProcessor;

XrayEventDisplayProcessor::XrayEventDisplayProcessor() : marlin::Processor("XrayEventDisplayProcessor"),
		_occupancy(NULL),
		_eventDisplay(NULL),
		_numberDisplays(0)
{
	registerInputCollection(LCIO::LCRELATION,"InputCollection","Input collection containing TimepixXrayObjects",_inputCollection,std::string("TimepixXrayObjects"));
}

XrayEventDisplayProcessor::~XrayEventDisplayProcessor()
{

}

void XrayEventDisplayProcessor::addHitToEventDisplay(lcio::TrackerHit* hit)
{

	const lcio::LCObjectVec* rawData = & hit->getRawHits();

	if(rawData == NULL)
	{
		streamlog_out(ERROR4) << "RawData vector is NULL" << std::endl;
		return;
	}

	for(lcio::LCObjectVec::const_iterator iter = rawData->begin(); iter != rawData->end(); iter++)
	{
		lcio::TrackerData* data = dynamic_cast<lcio::TrackerData*>(*iter);

		int cellID0 = data->getCellID0();

		EVENT::FloatVec adcValues = data->getChargeValues();

		for( int iADCValues=0; iADCValues < static_cast<int>(adcValues.size()); iADCValues++)
		{

			int nColumn = (cellID0 + iADCValues) % 256;
			int nRow = (cellID0 + iADCValues) / 256;

			_eventDisplay->Fill(nColumn,nRow,adcValues[iADCValues]);

		}
	}

}

void XrayEventDisplayProcessor::addHitToOccupancy(lcio::TrackerHit* hit)
{

	const lcio::LCObjectVec* rawData = & hit->getRawHits();

	if(rawData == NULL)
	{
		streamlog_out(ERROR4) << "RawData vector is NULL" << std::endl;
		return;
	}

	for(lcio::LCObjectVec::const_iterator iter = rawData->begin(); iter != rawData->end(); iter++)
	{
		lcio::TrackerData* data = dynamic_cast<lcio::TrackerData*>(*iter);

		int cellID0 = data->getCellID0();

		EVENT::FloatVec adcValues = data->getChargeValues();

		for( int iADCValues=0; iADCValues < static_cast<int>(adcValues.size()); iADCValues++)
		{

			int nColumn = (cellID0 + iADCValues) % 256;
			int nRow = (cellID0 + iADCValues) / 256;

			_occupancy->Fill(nColumn,nRow);

		}
	}

}

void XrayEventDisplayProcessor::init()
{

	printParameters();

	streamlog_out(WARNING4) << "This processor is not multi-chip capable (yet)" << std::endl;

	_numberDisplays = 0;

	_occupancy = new TH2D("occupancy","Occupancy",256,-0.5,255.5,256,-0.5,255.5);

}

void XrayEventDisplayProcessor::processRunHeader( LCRunHeader* run )
{
	run->parameters().setValue( _processorName + "_revision", "$Rev: 007 $");

	for ( marlin::ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
		if ( ! i->second->isOptional() || i->second->valueSet() ) {
			run->parameters().setValue( _processorName + "_" + i->second->name(),
					i->second->value() );
		}
	}
}

void XrayEventDisplayProcessor::processEvent( LCEvent * evt )
{

	lcio::LCCollectionVec* inputCollection(NULL);

	try{
		inputCollection = dynamic_cast<lcio::LCCollectionVec*>(evt->getCollection(_inputCollection));
	}
	catch(DataNotAvailableException &e)
	{
		streamlog_out(MESSAGE3) << "InputCollection not available in event " << evt->getEventNumber() << std::endl;
		return;
	}

	if(inputCollection == NULL)
	{
		streamlog_out(ERROR4) << "InputCollection vector is NULL in event " << evt->getEventNumber() << std::endl;
		return;
	}

	if(inputCollection->size() == 0)
	{
		streamlog_out(MESSAGE3) << "InputCollection is empty in event " << evt->getEventNumber() << std::endl;
		return;
	}

	std::stringstream histogramName;
	histogramName << "run" << evt->getRunNumber() << "evt" << evt->getEventNumber();

	std::stringstream title;
	title << "Run " << evt->getRunNumber() << " Event " << evt->getEventNumber();

	_eventDisplay = new TH2D(histogramName.str().c_str(),"",256,-.5,255.5,256,-.5,255.5);

	for(lcio::LCCollectionVec::iterator iter = inputCollection->begin(); iter != inputCollection->end(); iter++)
	{

		lcio::LCRelation* xrayObject = dynamic_cast<lcio::LCRelation*>(*iter);

		if(xrayObject == NULL)
		{
			streamlog_out(ERROR4) << "Dynamic cast to lcio::LCRelation failed and returned NULL in event " << evt->getEventNumber() << std::endl;
			continue;
		}

		lcio::TrackerHit* hit = dynamic_cast<lcio::TrackerHit*>(xrayObject->getTo());

		if(hit == NULL)
		{
			streamlog_out(ERROR4) << "Dynamic cast to lcio::TrackerHit failed and returned NULL in event " << evt->getEventNumber() << std::endl;
			continue;
		}

		lcio::LCGenericObject* object = dynamic_cast<lcio::LCGenericObject*>(xrayObject->getFrom());

		if(object == NULL)
		{
			streamlog_out(ERROR4) << "Dynamic cast to lcio::LCGenericObject failed and returned NULL in event " << evt->getEventNumber() << std::endl;
			continue;
		}

		GridPixXrayObject* xray = new GridPixXrayObject(object);

		title << " - " << xray->getEnergyFromCharge() << " keV";

		addHitToOccupancy(hit);

		addHitToEventDisplay(hit);

		delete xray;

	}

	_eventDisplay->SetTitle(title.str().c_str());

	marlin::AIDAProcessor::tree(this);
	_eventDisplay->Write();

	_numberDisplays++;

	delete _eventDisplay;

}

void XrayEventDisplayProcessor::end()
{

	marlin::AIDAProcessor::tree(this);
	_occupancy->Write();

	delete _occupancy;

	streamlog_out(MESSAGE4) << "Produced " << _numberDisplays << " event displays" << std::endl;

}

}//namespace marlintpc
