#include "XrayGeometryService.h"

#define GEOSERVICE_MEMSIZE 9

//GEAR
#include <gear/GEAR.h>
#include <gear/TPCModule.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>

namespace marlintpc{

  XrayGeometryService::XrayGeometryService()
  {
    _memory = new double[GEOSERVICE_MEMSIZE];
    
    //initisalise _memory with 0.
    for(int i = 0; i < GEOSERVICE_MEMSIZE; i++)
    {
      _memory[i] = 0.;
    }
  }
  
  XrayGeometryService::~XrayGeometryService()
  {
    delete _memory;
  }
  
  void XrayGeometryService::calculateGeometry(lcio::TrackerHit* hit, double rotAngle)
  {
    
    if( hit == NULL ){return;}
    
    double xHit = hit->getPosition()[0];
    double yHit = hit->getPosition()[1];
    
    lcio::LCObjectVec const & dataVec = hit->getRawHits();
    
    int chipID = dynamic_cast<lcio::TrackerData *>(dataVec[0])->getCellID1();
    
    gear::TPCModule const &module = marlin::Global::GEAR->getTPCParameters().getModule( chipID );
    
    // get some geometry information from gear
    // This processor implicitly assumes a rectangular matrix of pads/pixels
    int nRows = module.getNRows();
    int nColumns = module.getNPads() / nRows;
    double pitchX = module.getPadPitch(0); // there always should be a pad 0
    double pitchY = module.getRowHeight(0); // there always should be a row 0
    
    double xMax(-20.);
    double xMin(20.);
    double yMax(-20.);
    double yMin(20.);
    
    double sumX2(0.);
    double sumX3(0.);
    double sumX4(0.);
    
    double sumY2(0.);
    double sumY3(0.);
    double sumY4(0.);
    
    double dMax(0.);
    
    int numberPixels(0);
    
    _pixelPositions.clear();

    for( lcio::LCObjectVec::const_iterator iterData = dataVec.begin(); iterData != dataVec.end(); iterData++)
    {
    
      lcio::TrackerData* data = dynamic_cast<lcio::TrackerData *>(*iterData);
      
      int cellID0 = data->getCellID0();
      
      lcio::FloatVec pixelVec = data->getChargeValues();
      
      for(lcio::FloatVec::iterator iterPixel = pixelVec.begin(); iterPixel != pixelVec.end(); iterPixel++)
      {
      
	double x = cellID0 % nColumns;
	double y = static_cast<int>((cellID0 - x) / nColumns);
	
	x = (nColumns - x -0.5)*pitchX;
	y = (y + 0.5)*pitchY;
	
	numberPixels++;
	cellID0++;
	
	double xRot = std::cos(-rotAngle)*(x-xHit) - std::sin(-rotAngle)*(y-yHit);
	double yRot = std::sin(-rotAngle)*(x-xHit) + std::cos(-rotAngle)*(y-yHit);
	
	if(xRot > xMax){xMax = xRot;}
	if(xRot < xMin){xMin = xRot;}
	if(yRot > yMax){yMax = yRot;}
	if(yRot < yMin){yMin = yRot;}
	
	_pixelPositions.push_back(std::pair<double,double>(xRot,yRot));

	double distance = std::sqrt(xRot*xRot + yRot*yRot);
	
	if(distance > dMax){dMax = distance;}
	
	sumX2 += xRot*xRot;
	sumX3 += std::pow(xRot,double(3.));
	sumX4 += std::pow(xRot,double(4.));
	
	sumY2 += yRot*yRot;
	sumY3 += std::pow(yRot,double(3.));
	sumY4 += std::pow(yRot,double(4.));
	
      }
    
    }
    
    _memory[0] = std::sqrt( sumX2 / double(numberPixels) ); //rmsX
    _memory[1] = std::sqrt( sumY2 / double(numberPixels) ); //rmsY
    
    _memory[2] = xMax - xMin; //length
    _memory[3] = yMax - yMin; //width
    
    _memory[4] = sumX3 / double(numberPixels) / std::pow(_memory[0],3.); //skewnessL
    _memory[5] = sumY3 / double(numberPixels) / std::pow(_memory[1],3.); //skewnessT
    
    _memory[6] = (sumX4 / double(numberPixels) / std::pow(_memory[0],4.)) - 3.; //kurtosisL
    _memory[7] = (sumY4 / double(numberPixels) / std::pow(_memory[1],4.)) - 3.; //kurtosisT
    
    _memory[8] = dMax; //radius
    
    return;
    
  }
  
  int XrayGeometryService::numberPixelsWithinRadius(double radius) const
  {
	  if(_pixelPositions.size() == 0)
		  {
		  	  return -1;
		  }
	  else
	  {
		  int pixels = 0;

		  for(std::list<std::pair<double,double> >::const_iterator iter = _pixelPositions.begin(); iter != _pixelPositions.end(); iter++)
		  {
			  double distance = std::sqrt(iter->first*iter->first + iter->second*iter->second);

			  if(distance <= radius)
			  {
				  pixels++;
			  }
		  }

		  return pixels;
	  }
  }

  double XrayGeometryService::getRmsX() const
  {
    return _memory[0];
  }
  
  double XrayGeometryService::getRmsY() const
  {
    return _memory[1];
  }
  
  double XrayGeometryService::getExcentricity() const
  {
    return (_memory[0] / _memory[1]);
  }
  
  double XrayGeometryService::getLength() const
  {
    return _memory[2];
  }
  
  double XrayGeometryService::getWidth() const
  {
    return _memory[3];
  }
  
  double XrayGeometryService::getSkewnessL() const
  {
    return _memory[4];
  }
  
  double XrayGeometryService::getSkewnessT() const
  {
    return _memory[5];
  }
  
  double XrayGeometryService::getKurtosisL() const
  {
    return _memory[6];
  }
  
  double XrayGeometryService::getKurtosisT() const
  {
    return _memory[7];
  }
  
  double XrayGeometryService::getRadius() const
  {
    return _memory[8];
  }

}//namespace marlintpc
