#include "XrayLikelihoodProcessor.h"



namespace marlintpc{
  
  XrayLikelihoodProcessor aXrayLikelihoodProcessor;
  
  XrayLikelihoodProcessor::XrayLikelihoodProcessor() : Processor("XrayLikelihoodProcessor"),
		  _energyBinsLowerEdges(1,0.),
		  _referenceROOTFilesXray(1,"XrayReference.root"),
		  _scaleFactors(3,1.0),
		  _numberEnergyBins(0)

  {
    
    _description = "Calculates log likelihood for XrayObjects based on reference data sets and stores the value in the Likelihood field of the output Xrays";
    
    registerInputCollection(LCIO::LCRELATION,"InputCollectionXray","InputCollection containing XrayObjects",_inputCollectionXray,std::string("TimepixXrayObjects"));
    
    registerOutputCollection(LCIO::LCRELATION,"OutputCollection","OutputCollection containing XrayObjects",_outputCollection,std::string("TimepixXrayObjectsLikelihood"));
    
    registerOutputCollection(LCIO::LCGENERICOBJECT,"OutputXrays","OutputCollection containing GridPixXrayObjects with likelihood",_outputXrays,std::string("GridPixXrayObjectsLikelihood"));
    
    registerProcessorParameter("EnergyBinsLowerEdges","Lower edges of energy bins for X-ray reference data sets, first value has to be 0.",_energyBinsLowerEdges,_energyBinsLowerEdges);

    registerProcessorParameter("ReferenceROOTFilesXray","Xray reference root files, one for each energy bin",_referenceROOTFilesXray,_referenceROOTFilesXray);
    
    registerProcessorParameter( "SetOutputTransient","if not 0 the output collection is set transient",_outputIsTransient,0);
    
    registerOptionalParameter("ScaleFactors","If set the measured values are multiplied by these factors to compensate differences between data and reference sets (eccentricity, rmsY, length)",_scaleFactors,_scaleFactors);

  }
  
  XrayLikelihoodProcessor::~XrayLikelihoodProcessor(){}
  
  int XrayLikelihoodProcessor::energyBin(double energy){

	  int bin = 0;

	  while(energy >= _energyBinsLowerEdges[bin] && bin < _numberEnergyBins)
	  {
		  bin++;
	  }

	  return bin-1;
    
  }
  
  double XrayLikelihoodProcessor::logLikelihood(TH1D* histogram, double value)
  {

	  int binContent = histogram->GetBinContent(histogram->FindBin(value));

	  double likelihood = (double)binContent / (double)histogram->GetEntries();

	  if(binContent == 0)
	  {
		  return -std::numeric_limits<double>::infinity();
	  }
	  else
	  {
		  return std::log(likelihood);
	  }
  }

  void XrayLikelihoodProcessor::init(){
    
    printParameters();
    
    if(_energyBinsLowerEdges.size() != _referenceROOTFilesXray.size())
    {
    	streamlog_out(ERROR1) << "ERROR: Number of energy bins does not match number of given reference root files!" << std::endl;
    	if(_energyBinsLowerEdges.size() > _referenceROOTFilesXray.size()){_numberEnergyBins = _referenceROOTFilesXray.size();}
    	else{_numberEnergyBins = _energyBinsLowerEdges.size();}
    }
    else
    {
    	_numberEnergyBins = _energyBinsLowerEdges.size();
    }

    for(int i = 0; i < _numberEnergyBins; i++)
    {
    	std::stringstream name;
    	name << "likelihood_" << _energyBinsLowerEdges[i];
    	std::stringstream title;
    	title << "logL " << _energyBinsLowerEdges[i];

    	if(i < _numberEnergyBins-1)
    	{
    		name << "to" << _energyBinsLowerEdges[i+1];
    		title << " to " << _energyBinsLowerEdges[i+1];
    	}

    	_histogramsLikelihood.push_back(new TH1D(name.str().c_str(),title.str().c_str(),100,0.,50.));
    	_referenceTFilesXray.push_back(new TFile(_referenceROOTFilesXray[i].c_str()));

    }

    _xrayExcentricity.resize(_numberEnergyBins);
    _xrayKurtosisL.resize(_numberEnergyBins);
    _xraySkewnessL.resize(_numberEnergyBins);
    _xrayRmsT.resize(_numberEnergyBins);
    _xrayLength.resize(_numberEnergyBins);
    _xrayLengthDivByRmsT.resize(_numberEnergyBins);
    _xrayFractionWithinRmsY.resize(_numberEnergyBins);

    std::string pathExcentricity = "MyXrayGeometryAnalysisProcessor/excentricity";
    std::string pathKurtosisL = "MyXrayGeometryAnalysisProcessor/kurtosisl";
    std::string pathSkewnessL = "MyXrayGeometryAnalysisProcessor/skewnessl";
    std::string pathRmsT = "MyXrayGeometryAnalysisProcessor/rmsy";
    std::string pathLength = "MyXrayGeometryAnalysisProcessor/length";
    std::string pathLengthDivByRmsT = "MyXrayGeometryAnalysisProcessor/lengthdivbyrmsy";
    std::string pathFractionWithinRmsY = "MyXrayGeometryAnalysisProcessor/fractionwithinrmsy";

    for(int i = 0; i < _numberEnergyBins; i++)
    {
    	_xrayExcentricity[i] = (TH1D*) _referenceTFilesXray[i]->Get(pathExcentricity.c_str());
    	_xrayKurtosisL[i] = (TH1D*) _referenceTFilesXray[i]->Get(pathKurtosisL.c_str());
    	_xraySkewnessL[i] = (TH1D*) _referenceTFilesXray[i]->Get(pathSkewnessL.c_str());
    	_xrayRmsT[i] = (TH1D*) _referenceTFilesXray[i]->Get(pathRmsT.c_str());
    	_xrayLength[i] = (TH1D*) _referenceTFilesXray[i]->Get(pathLength.c_str());
    	_xrayLengthDivByRmsT[i] = (TH1D*) _referenceTFilesXray[i]->Get(pathLengthDivByRmsT.c_str());
    	_xrayFractionWithinRmsY[i] = (TH1D*) _referenceTFilesXray[i]->Get(pathFractionWithinRmsY.c_str());

    }

    _xrayGeoService = new XrayGeometryService();

    if(parameterSet("ScaleFactors") && _scaleFactors.size() != 3)
    {
    	streamlog_out(ERROR1) << "ScaleFactors are enabled but wrong number (3 values are needed): Disabling ScaleFactors!" << std::endl;
    	_scaleFactors = std::vector<float>(3,1.0);
    }

  }
  
  void XrayLikelihoodProcessor::processRunHeader( LCRunHeader* run) {
    
    
    run->parameters().setValue( _processorName + "_revision", "$Rev: 007 $");
    
    for ( marlin::ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
      if ( ! i->second->isOptional() || i->second->valueSet() ) {
	run->parameters().setValue( _processorName + "_" + i->second->name(), 
				    i->second->value() );
      }
    }
    
  }
  
  void XrayLikelihoodProcessor::processEvent( LCEvent * evt ){
    
    lcio::LCCollectionVec* inputColXray(NULL);

    try{
      
      inputColXray = dynamic_cast<lcio::LCCollectionVec*>(evt->getCollection(_inputCollectionXray));
      
    }
    catch(DataNotAvailableException &e){
      
      streamlog_out(MESSAGE3) << "InputCollection not available" << std::endl;
      return;
      
    }

    if(inputColXray == NULL){
      
      streamlog_out(ERROR2) << "InputCollection vector is 0" << std::endl;
      return;
      
    }

    lcio::LCCollectionVec* outputXrays = new lcio::LCCollectionVec(LCIO::LCGENERICOBJECT);
    lcio::LCCollectionVec* outputCol = new lcio::LCCollectionVec(LCIO::LCRELATION);
    
    outputXrays->setTransient( (_outputIsTransient != 0) );
    outputCol->setTransient( (_outputIsTransient != 0) );
    
    for(lcio::LCCollectionVec::iterator iterCol = inputColXray->begin(); iterCol != inputColXray->end(); iterCol++)
    {
      
      lcio::LCRelation* xrayObject = dynamic_cast<lcio::LCRelation*>(*iterCol);

      GridPixXrayObject* xray = new GridPixXrayObject(dynamic_cast<lcio::LCGenericObject*>(xrayObject->getFrom()));

      int bin = energyBin(xray->getEnergyFromCharge());

      _xrayGeoService->calculateGeometry(dynamic_cast<lcio::TrackerHit*>(xrayObject->getTo()),xray->getRotAngle());

      double lengthDivByRmsT = _xrayGeoService->getLength() / _xrayGeoService->getRmsY();

      double logL(0.);

      logL += logLikelihood(_xrayExcentricity[bin],_scaleFactors[0]*_xrayGeoService->getExcentricity());

      //logL += logLikelihood(_xrayKurtosisL[bin],_xrayGeoService->getKurtosisL());

      //logL += logLikelihood(_xraySkewnessL[bin],_xrayGeoService->getSkewnessL());

      logL += logLikelihood(_xrayLengthDivByRmsT[bin],_scaleFactors[1]*lengthDivByRmsT);

      //logL += logLikelihood(_xrayRmsT[bin],_scaleFactors[1]*_xrayGeoService->getRmsY());

      //logL += logLikelihood(_xrayLength[bin],_scaleFactors[2]*_xrayGeoService->getLength());

      logL += logLikelihood(_xrayFractionWithinRmsY[bin],_scaleFactors[2]*_xrayGeoService->numberPixelsWithinRadius(_xrayGeoService->getRmsY())/(double)xray->getPixels());

      if(logL != logL)
      {
    	  streamlog_out(MESSAGE4) << "logL is NaN!" << std::endl;
      }

      GridPixXrayObject* newXray = new GridPixXrayObject;

      newXray->setCenterX(xray->getCenterX());
      newXray->setCenterY(xray->getCenterY());
      newXray->setCharge(xray->getCharge());
      newXray->setEnergyFromCharge(xray->getEnergyFromCharge());
      newXray->setEnergyFromPixels(xray->getEnergyFromPixels());
      newXray->setLikelihood(-logL);
      newXray->setPixels(xray->getPixels());
      newXray->setRotAngle(xray->getRotAngle());

      delete xray;

      _histogramsLikelihood[bin]->Fill(newXray->getLikelihood());
     
      outputXrays->addElement(newXray);
      
      lcio::LCRelationImpl* xrayObjectCalibrated = new LCRelationImpl(newXray,xrayObject->getTo());
     
      outputCol->addElement(xrayObjectCalibrated);
      
    }
    
    evt->addCollection(outputXrays,_outputXrays);
    evt->addCollection(outputCol,_outputCollection);
    
  }
  
  void XrayLikelihoodProcessor::end(){
    
    marlin::AIDAProcessor::tree(this);
    
    for(int i = 0; i < _numberEnergyBins; i++)
    {

    	_histogramsLikelihood[i]->Write();
    	delete _histogramsLikelihood[i];

    }

    for(int i = 0; i < _numberEnergyBins; i++)
    {
    	delete _referenceTFilesXray[i];
    }

    delete _xrayGeoService;

  }  
  
}//namespace marlintpc
