#include "dEdxProcessor.h"

// C++
#include <iostream>
#include <vector>
#include <string>
#include <sstream>
#include <cmath>
#include <algorithm>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerHit.h>
#include <EVENT/Track.h>
#include <IMPL/LCCollectionVec.h>
#include <UTIL/LCTOOLS.h>
#include <IMPL/LCFlagImpl.h>

//marlin
#include <marlin/ConditionsProcessor.h>
#include <marlin/Global.h>
#include "marlin/Exceptions.h"

//GEAR
#include "gear/GearMgr.h"
#include "gear/gearxml/GearXML.h"
#include "gear/TPCModule.h"
#include "gear/TPCParameters.h"

#include "TrackCircleIntersectionCalculator.hh"

// the common flag word definitions in MarlinTPC
#include "FlagwordDefinitions.h"

#include "TF1.h"

// Aida
#ifdef MARLIN_USE_AIDA
//AIDA
#include <marlin/AIDAProcessor.h>
#endif

//using namespace tpcconddata;
using namespace lcio ;
using namespace marlin ;
using namespace gear;
using namespace std;

namespace marlintpc
{
 const std::pair<int, int> dEdxProcessor::ALLROWS(-1,-1);

  dEdxProcessor adEdxProcessor ;


dEdxProcessor::dEdxProcessor() :
        Processor("dEdxProcessor") 
{
  _description = "dEdxProcessor calculates a factor to convert hit charge in adc counts into primary electrons";

  
    registerInputCollection(LCIO::TRACK,
                             "InputCollectionName" ,
                             "Name of the input collection (default: ClupatraTracks)"  ,
                             _inputColName ,
                             std::string("ClupatraTracks")) ;
    gear::IntVec exampleRow;
    exampleRow.push_back(ALLROWS.first);
    exampleRow.push_back(ALLROWS.second);
    registerOptionalParameter("SkipRow",
            "Define the row to skip. The row must be defined as a pair of integers separated by a space. A negative integer indicates any module/row. Default: -1 -1 (Every row is cut away)",
            _rows2skip,
            exampleRow,
            exampleRow.size());
}


void dEdxProcessor::init()
{

    // usually a good idea to
    printParameters() ;
    AIDAProcessor::tree(this);
    if(parameterSet("SkipRow") )
      {
	if((_rows2skip.size() % 2) != 0)
	  {
	    streamlog_out(ERROR) << "You must specify a list of pair of modules and rows. There seem to be an odd number of inputs." << std::endl;
	    throw marlin::ParseException("The parameter list seems to be incorrect");
	  }
	_skipList.clear();
	
	for(gear::IntVec::iterator it = _rows2skip.begin(); it != _rows2skip.end(); it= it+2)
	  {
	    int mod = *it <= -1 ? ALLROWS.first : *it;
	    int row = *(it+1)<= -1 ? ALLROWS.second : *(it+1);
	    std::pair<int, int> rowid(mod, row);
	    _skipList.insert(rowid);
	    streamlog_out(DEBUG) << "Cutting Row " << *(it +1)  << " in module " << *it <<std::endl;
	  }
      }

    _tuple=new TNtuple("tuple","tuple","EvID:trackID:dEdxMean90:dEdxMean90Err:dEdxRMS90:dEdxRMS90Err:omega");

    _histodEdx = new TH1F("dEdx","dEdx",1000,0,10000);
    _myTool= new HistogramWidthExtractionTools();

}

void dEdxProcessor::processRunHeader(lcio::LCRunHeader* run)
{


    run->parameters().setValue(_processorName + "_revision", "$Rev: 2743 $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(),
                                               i->second->value());
                }
        }



}

void dEdxProcessor::processEvent(lcio::LCEvent * evt)
{


    //necessary if histograms are created here instead of in init(), 
    //to get histograms in the correct folder
    AIDAProcessor::tree(this);
    // get the input collection from the event
    LCCollection* inputCol = 0;
    try
        {
            inputCol = evt->getCollection(_inputColName) ;
        }

    catch (DataNotAvailableException)
        {
            // if there is not input collection give a warning to the user
            streamlog_out(WARNING2)
            << "processEvent: input collection "
            << _inputColName << " cannot be found in event " << evt->getEventNumber() << std::endl;
            return;
        }
    // the main parameter for the geometry description is the gear parameter:
    const gear::TPCParameters& tpcParameters = Global::GEAR->getTPCParameters();
    // loop the input data 
    for (int t=0;t< inputCol->getNumberOfElements();t++)
    {
      _histodEdx->Clear();
      lcio::Track* currentTrack = dynamic_cast<lcio::Track*>(inputCol->getElementAt(t));
      const TrackerHitVec hitvec= currentTrack->getTrackerHits();
      for(unsigned int h=0;h<hitvec.size();h++)
	{
	  lcio::TrackerHit * hit= dynamic_cast<lcio::TrackerHit*>(hitvec[h]);
	  //check hit quality
	  if(hitflag::hasDeadChannel(hit->getQuality()) || hitflag::hasOverflowPulse(hit->getQuality()))
	    continue;
	  //get module where current hit was measured on
	  const gear::TPCModule& theModule =  tpcParameters.getNearestModule(hit->getPosition()[0],hit->getPosition()[1]);
	  int padIndex=theModule.getNearestPad (hit->getPosition()[0],hit->getPosition()[1]); 
	  int rownumber=theModule.getRowNumber(padIndex) ;
	  double rowHeight=theModule.getRowHeight(rownumber);
	  //check if hit is on row that should be skipped
	  int mod = theModule.getModuleID();
	  bool _hitToKeep = true;
	  if(parameterSet("SkipRow") )
	    {
	      if(_skipList.find(std::pair<int,int> (ALLROWS.first, rownumber)) == _skipList.end())
		{
		  if (_skipList.find(std::pair<int,int> (mod, ALLROWS.second)) == _skipList.end())
		    {
		      if (_skipList.find(std::pair<int,int> (mod, rownumber)) == _skipList.end())
			{
			  _hitToKeep = false;
			}
		    }
		}
	    }
	  if(!_hitToKeep)
	      continue;

	  if((theModule.getLocalPadLayout()).getPadLayoutType() == 2)//if circular pad layout
	    {
	      //get radius and center of measurement layer in global coordinates
	      gear::Vector2D cMeasurementLayer = theModule.localToGlobal(0.,0.);
	      std::pair<double, double> centerLayer = std::make_pair(cMeasurementLayer[0], cMeasurementLayer[1]);
	      gear::Vector2D padCenterGlobal = theModule.getPadCenter(theModule.getNearestPad(hit->getPosition()[0],hit->getPosition()[1]));
	      gear::Vector2D padCenterLocal = theModule.globalToLocal(padCenterGlobal[0],padCenterGlobal[1]);
	      
	      //inner and outer radius of padrow
	      double RadiusIn = padCenterLocal[0]-rowHeight/2;
	      double RadiusOut = padCenterLocal[0]+rowHeight/2;
	      //calculate point of itersection between track and measurement layer in global coordinates
	      TrackCircleIntersectionCalculator IntersectionCalculatorInner(centerLayer, RadiusIn, currentTrack->getD0(),
									    currentTrack->getPhi(),currentTrack->getOmega());
	      std::pair<double, double> IntersectionInner = IntersectionCalculatorInner.GetIntersection();
	     
	      gear::Vector2D intersectionIn;
	      intersectionIn[0] = IntersectionInner.first;
	      intersectionIn[1] = IntersectionInner.second;
	      
	      TrackCircleIntersectionCalculator IntersectionCalculatorOuter(centerLayer, RadiusOut, currentTrack->getD0(),
									    currentTrack->getPhi(),currentTrack->getOmega());
	      std::pair<double, double> IntersectionOuter = IntersectionCalculatorOuter.GetIntersection();
	     
	      gear::Vector2D intersectionOut;
	      intersectionOut[0] = IntersectionOuter.first;
	      intersectionOut[1] = IntersectionOuter.second;
	      
	  //     std::cout<<"center :"<<centerLayer.first<<" "<<centerLayer.second<<std::endl;
// 	      std::cout<<"inner: "<<intersectionIn[0]<<" "<<intersectionIn[1]<<std::endl;
// 	      std::cout<<"outer: "<<intersectionOut[0]<<" "<<intersectionOut[1]<<std::endl;

	      //track length on pad row is distance between the two intersection points in mm converted to cm
	      double length=1./10.*sqrt(pow(intersectionOut[0]-intersectionIn[0],2)+pow(intersectionOut[1]-intersectionIn[1],2));
	   
	      _histodEdx->Fill(hit->getEDep()/length);
	    }
	  else if((theModule.getLocalPadLayout()).getPadLayoutType() == 1)
	    {
	      streamlog_out(MESSAGE)<<"ATTENTION! This was not yet implemented. Good luck doing it..."<<std::endl;
	    }
	  else
	    {
	      streamlog_out(ERROR)<<"Error: Invalid pad layout type!"<<std::endl;
	    }
	  
	}//hits
      _myTool->Calculate(_histodEdx,1);
      _tuple->Fill(evt->getEventNumber(),t,_myTool->GetMean(),_myTool->GetMeanError(),_myTool->GetWidth(),
		   _myTool->GetWidthError(),currentTrack->getOmega());
  
    }//tracks
}


void dEdxProcessor::check(lcio::LCEvent*  /*evt*/)
{ 
  // check what is needed here 
}


void dEdxProcessor::end()
{ 
 

}



}
