#ifndef MMADVANCEDCALIBRATIONPROCESSOR_H 
#define MMADVANCEDCALIBRATIONPROCESSOR_H 

#include <marlin/Processor.h>
#include <lcio.h>
#include <string>
#include <map>
#include <vector>

#include "MMPadResponseFunction.h"

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "IMPL/LCCollectionVec.h"
#include "EVENT/TrackerPulse.h"

//Aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/ITupleFactory.h>
#include <AIDA/ITuple.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>

//LCIO
#include <EVENT/LCFloatVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/TrackImpl.h>

// GEAR
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>
#include <gearimpl/GlobalPadIndex.h>

//MARLIN
#include <marlin/Global.h>

//ROOT
#include <Minuit2/MnUserParameters.h>
#include <Minuit2/MnMigrad.h>
#include <Minuit2/FunctionMinimum.h>
#include <Minuit2/Minuit2Minimizer.h>

#include "Math/Minimizer.h"
#include "Math/Functor.h"
#include <TMath.h>
#include <TTree.h>
#include <TH2D.h>
#include <TH2F.h>
#include <TH1F.h>
#include <TF1.h>
#include <TCanvas.h>
#include <TApplication.h>

namespace marlintpc
{
  /**  
   * Basic overview:
   *
   * This processor takes in a collection of tracks. Filters are applied to screen unwanted tracks (trackangle, number of tracks, etc.)
   * The consitutent pulses are then given to an MMRefitTool object, which is then minimized by minuit on an event-by-event basis. 
   * The MMRefitTool will return a chi2 value for a given set of PRF parameters and hit x0 values.
   *
   * Parameters:
   *
   *    InputTrackCollections  =  A list of track collections to be used  [StringVec]
   *    ParamStart             =  Starting values for each PRF parameter  [FloatVec]
   *    ParamSteps             =  Minuit step values for each PRF parameter  [FloatVec]
   *    ParamMin               =  Minimum value for each PRF parameter  [FloatVec]
   *    ParamMax               =  Maximum value for each PRF parameter  [FloatVec]
   *    FitterSource           =  Source for the Root minimizer object (Minuit2 or GSL)  [string]
   *    FitterType             =  Type of minimizer object to use (see Root documentation)  [string]
   *    TrackType              =  Track type (straight or helix)  [integer]
   *    PRFType                =  Name of the PRF to use  [string]
   *    ParOutputPath          =  Output path for the calibrated parameters  [string]
   *    Tolerance              =  Minimizer tolerance  [double]
   *    MaxIterations          =  Maximum iterations of the minimizer  [integer]
   *    MaxFcn                 =  Maximum function calls of the minimizer  [integer]
   *    FitOutput              =  Whether or not to fit the calibrated parameter distributions with gaussians  [boolean]
   *    RowDep                 =  Set whether or not the PRF has a row dependency (as defined in MMHitFinderProcessor)  [boolean]
   *    FitTrack               =  Whether or not the track should be fit in the calibration  [boolean]
   *    PadHits                =  Whether or not the hits should be padded with empty pulses  [boolean]
   *    NormType               =  The type of PRF normalization to be used (0 = sum Amp / sum PRF; 1 = sum Amp; 2 = gaussian amp)  [integer]  
   *    ModuleList             =  A list of modules to be used in the calibration  [IntVec]
   *
   * Author: P. Hayman, Carleton University
   */

  class MMAdvancedCalibrationProcessor : public marlin::Processor
  {
    public:

      virtual Processor*  newProcessor() { return new MMAdvancedCalibrationProcessor; }

      MMAdvancedCalibrationProcessor();

      virtual void init();

      virtual void processRunHeader(EVENT::LCRunHeader* run);

      virtual void processEvent(EVENT::LCEvent* evt);

      virtual void check(EVENT::LCEvent* evt);

      virtual void end();

    protected:

      //Required Parameters
      lcio::StringVec 		_inputColNames; // Name of the input collection
      std::string					_output_path;
      std::string 				_source;
      std::string 				_type;
      std::string 				_prfType;
      int 								_trackType;

      std::vector<float> _stepVec;
      std::vector<float> _minVec;
      std::vector<float> _maxVec;
      std::vector<float> _startVec;


      //Optional Parameters
      double 							_tolerance;
      int 								_iter;
      int 								_fcn;
      int 								_normType;
      bool 								_fitOutput;
      bool 								_rowDep;
      bool 								_fitTrack;
      bool 								_padHits;
      std::vector<int>    _moduleList;

      //Class Variables
      int 								_curEvt; //Keep track of the current event.
      int 								_nPar;
      bool								_moduleDep;

      AIDA::ITuple*      	_fitData;
      AIDA::ITuple*      	_hitData;

      TTree*							_tree;
      std::vector< TH1F * >	_histogramPar;
      std::vector< std::vector<TH2F *> >	_histogramCorrPar;

      ROOT::Math::Minimizer* min;

  };
} // namespace marlintpc
#endif // MMADVANCEDCALIBRATIONPROCESSOR_H 
