#include "ADC2PrimaryElectronFactorCalculator.h"

// C++
#include <iostream>
#include <vector>
#include <string>
#include <sstream>
#include <cmath>
#include <algorithm>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerHit.h>
#include <EVENT/Track.h>
#include <IMPL/LCCollectionVec.h>
#include <UTIL/LCTOOLS.h>
#include <IMPL/LCFlagImpl.h>

//marlin
#include <marlin/ConditionsProcessor.h>
#include <marlin/Global.h>

//GEAR
#include "gear/GearMgr.h"
#include "gear/gearxml/GearXML.h"
#include "gear/TPCModule.h"
#include "gear/TPCParameters.h"

#include "TrackCircleIntersectionCalculator.hh"

#include "TF1.h"

//using namespace tpcconddata;
using namespace lcio ;
using namespace marlin ;
using namespace gear;
using namespace std;

namespace marlintpc
{

  ADC2PrimaryElectronFactorCalculatorProcessor aADC2PrimaryElectronFactorCalculatorProcessor ;


ADC2PrimaryElectronFactorCalculatorProcessor::ADC2PrimaryElectronFactorCalculatorProcessor() :
        Processor("ADC2PrimaryElectronFactorCalculatorProcessor"),
		_inputColName(""),
		_nRun(0),
		_primEl(0),
		_tuple(nullptr),
		_histoL(nullptr),
		_histoQ(nullptr),
		_histodEdx(nullptr),
		_outfile_root(nullptr),
		_outfile(""),
		_outfileName(""),
		_rootoutfileName("")

{
  _description = "ADC2PrimaryElectronFactorCalculatorProcessor calculates a factor to convert hit charge in adc counts into primary electrons";

  
    registerInputCollection(LCIO::TRACK,
                             "InputCollectionName" ,
                             "Name of the input collection (default: ClupatraTracks)"  ,
                             _inputColName ,
                             std::string("ClupatraTracks")) ;
    registerProcessorParameter("outfile", "name of output file in text mode", _outfileName, std::string("ADC2PrimaryElectronFactorCalculator_result.dat"));
    registerProcessorParameter("PrimaryElectronsPercm", "number of primary electrons per cm (default for argon based gases 86.6)", _primEl, double(86.6));
    registerOptionalParameter("rootoutfile", "name of output file in root", _rootoutfileName, std::string("ADC2PrimaryElectronFactorCalculator.root"));
}


void ADC2PrimaryElectronFactorCalculatorProcessor::init()
{

    // usually a good idea to
    printParameters() ;
    
    _tuple=new TNtuple("tuple","tuple","EvID:charge:length");

    _histoL = new TH1F("length","length",1000,0,10);
    _histoQ = new TH1F("hitcharge","hitcharge",1000,0,100000);
    _histodEdx = new TH1F("dEdx","dEdx",1000,0,10000);
    if(parameterSet("rootoutfile"))
      _outfile_root = new TFile(_rootoutfileName.c_str(),"RECREATE");
    _outfile.open(_outfileName.c_str());

}

void ADC2PrimaryElectronFactorCalculatorProcessor::processRunHeader(lcio::LCRunHeader* run)
{


    run->parameters().setValue(_processorName + "_revision", "$Rev: 2743 $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(),
                                               i->second->value());
                }
        }



}

void ADC2PrimaryElectronFactorCalculatorProcessor::processEvent(lcio::LCEvent * evt)
{

    streamlog_out(MESSAGE1) << "Event " << evt->getEventNumber() << std::endl;

    streamlog_out(DEBUG0) << "processEvent( " << evt->getEventNumber() << ")" << std::endl;


    // get the input collection from the event
    LCCollection* inputCol = 0;
    try
        {
            inputCol = evt->getCollection(_inputColName) ;
        }

    catch (DataNotAvailableException &)
        {
            // if there is not input collection give a warning to the user
            streamlog_out(WARNING2)
            << "processEvent: input collection "
            << _inputColName << " cannot be found in event " << evt->getEventNumber() << std::endl;
            return;
        }
    // the main parameter for the geometry description is the gear parameter:
    const gear::TPCParameters& tpcParameters = Global::GEAR->getTPCParameters();
  // loop the input data 
  for (int t=0;t< inputCol->getNumberOfElements();t++)
    {
      lcio::Track* currentTrack = dynamic_cast<lcio::Track*>(inputCol->getElementAt(t));
      const TrackerHitVec hitvec= currentTrack->getTrackerHits();
      for(unsigned int h=0;h<hitvec.size();h++)
	{
	  lcio::TrackerHit * hit= dynamic_cast<lcio::TrackerHit*>(hitvec[h]);
	  //get module where current hit was measured on
	  const gear::TPCModule& theModule =  tpcParameters.getNearestModule(hit->getPosition()[0],hit->getPosition()[1]);
	  int padIndex=theModule.getNearestPad (hit->getPosition()[0],hit->getPosition()[1]); 
	  int rownumber=theModule.getRowNumber(padIndex) ;
	  double rowHeight=theModule.getRowHeight(rownumber);
	  if((theModule.getLocalPadLayout()).getPadLayoutType() == 2)//if circular pad layout
	    {
	      //get radius and center of measurement layer in global coordinates
	      gear::Vector2D cMeasurementLayer = theModule.localToGlobal(0.,0.);
	      std::pair<double, double> centerLayer = std::make_pair(cMeasurementLayer[0], cMeasurementLayer[1]);
	      gear::Vector2D padCenterGlobal = theModule.getPadCenter(theModule.getNearestPad(hit->getPosition()[0],hit->getPosition()[1]));
	      gear::Vector2D padCenterLocal = theModule.globalToLocal(padCenterGlobal[0],padCenterGlobal[1]);
	      
	      //inner and outer radius of padrow
	      double RadiusIn = padCenterLocal[0]-rowHeight/2;
	      double RadiusOut = padCenterLocal[0]+rowHeight/2;
	      //calculate point of itersection between track and measurement layer in global coordinates
	      TrackCircleIntersectionCalculator IntersectionCalculatorInner(centerLayer, RadiusIn, currentTrack->getD0(),
									    currentTrack->getPhi(),currentTrack->getOmega());
	      std::pair<double, double> IntersectionInner = IntersectionCalculatorInner.GetIntersection();
	     
	      gear::Vector2D intersectionIn;
	      intersectionIn[0] = IntersectionInner.first;
	      intersectionIn[1] = IntersectionInner.second;
	      
	      TrackCircleIntersectionCalculator IntersectionCalculatorOuter(centerLayer, RadiusOut, currentTrack->getD0(),
									    currentTrack->getPhi(),currentTrack->getOmega());
	      std::pair<double, double> IntersectionOuter = IntersectionCalculatorOuter.GetIntersection();
	     
	      gear::Vector2D intersectionOut;
	      intersectionOut[0] = IntersectionOuter.first;
	      intersectionOut[1] = IntersectionOuter.second;
	      
	  //     std::cout<<"center :"<<centerLayer.first<<" "<<centerLayer.second<<std::endl;
// 	      std::cout<<"inner: "<<intersectionIn[0]<<" "<<intersectionIn[1]<<std::endl;
// 	      std::cout<<"outer: "<<intersectionOut[0]<<" "<<intersectionOut[1]<<std::endl;

	      //track length on pad row is distance between the two intersection points in mm converted to cm
	      double length=1./10.*sqrt(pow(intersectionOut[0]-intersectionIn[0],2)+pow(intersectionOut[1]-intersectionIn[1],2));
	   
	      _histoQ->Fill(hit->getEDep());
	      _histoL->Fill(length);
	      _histodEdx->Fill(hit->getEDep()/length);
	      _tuple->Fill(evt->getEventNumber(),hit->getEDep(),length);
	    }
	  else if((theModule.getLocalPadLayout()).getPadLayoutType() == 1)
	    {
	      streamlog_out(MESSAGE)<<"ATTENTION! This was not yet implemented. Good luck doing it..."<<std::endl;
	    }
	  else
	    {
	      streamlog_out(ERROR)<<"Error: Invalid pad layout type!"<<std::endl;
	    }
	  
	}//hits
    }//tracks
  
}


/*FIXME: not needed: void ADC2PrimaryElectronFactorCalculatorProcessor::check(lcio::LCEvent* evt)
{ 
  // check what is needed here 
}*/


void ADC2PrimaryElectronFactorCalculatorProcessor::end()
{ 
 

//   TF1 *fit=new TF1("fit","landau");
//   _histodEdx->Fit(fit);
//   std::cout<<"MPV: "<<fit->GetParameter(1)<<" mean: "<<_histodEdx->GetMean()<<std::endl;
  
  double integral = _histodEdx->Integral();
  double mean=0;
  double sumQ=0;
  int b=0;
  //truncate tail at 90% of integral
  while(sumQ<integral*0.9)
    {
      mean+=_histodEdx->GetBinContent(b)*_histodEdx->GetBinCenter(b);
      sumQ+=_histodEdx->GetBinContent(b);
      b++;
    }
  //for argon based gas mixtures we have 86.6 primary electrons per cm
  std::cout<<"---------------------------------------------------"<<std::endl;
  std::cout<<"---- ADCCounts per cm: "<< mean/sumQ <<" --------------------------"<<std::endl;
  std::cout<<"---- Primary electrons per cm: "<< _primEl <<" --------------------"<<std::endl;
  std::cout<<"---- Conversion factor : "<<_primEl/(mean/sumQ)<<" ---------------------------"<<std::endl;
  std::cout<<"---------------------------------------------------"<<std::endl;
  if(parameterSet("rootoutfile"))
    {
      _histoL->Write();
      _histoQ->Write();
      _histodEdx->Write();
      _tuple->Write();
      _outfile_root->Write();
      _outfile_root->Close();
    }
  _outfile<<"# input file \t\t Conversion factor : "<<std::endl;
  _outfile<<Global::parameters->getStringVal("LCIOInputFiles")<<" \t\t"<<_primEl/(mean/sumQ)<<std::endl;
  _outfile.close();

}



}
