#include "MMAdvancedCalibrationProcessor.h"

#include "MMRefitTool.h"

#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
#include <ctime>

// ROOT
#include <TROOT.h>
#include <TH1F.h>
#include <TSystem.h>
#include <TF1.h>
#include <TFitResultPtr.h>
#include <TFitResult.h>
#include <TVector2.h>

// Root Minimizers
#include <Math/Minimizer.h>
#include <Math/Factory.h>
#include <Math/Functor.h>

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/TrackerData.h>
#include <EVENT/TrackerPulse.h>
#include <Exceptions.h>
#include <marlin/Exceptions.h>

//Other
#include "IntersectionCalculator.h"

using namespace lcio;
using namespace marlin;
using namespace std;

namespace marlintpc
{

	MMAdvancedCalibrationProcessor aMMAdvancedCalibrationProcessor;


	MMAdvancedCalibrationProcessor::MMAdvancedCalibrationProcessor() : Processor("MMAdvancedCalibrationProcessor")
	{
		// the processor description
		_description = "MMAdvancedCalibrationProcessor calibrates the PRF by minimizing it while fitting the track.";

		_inputColNames.push_back("TPCTracks");

		registerInputCollections(    LCIO::TRACK,
				"InputTrackCollections" ,
				"Names of the input Track collections, default: TPCTracks"  ,
				_inputColNames ,
				_inputColNames) ;

		registerProcessorParameter("ParamStart",
				"Start value for each parameter.",
				_startVec,
				_startVec);

		registerProcessorParameter("ParamSteps",
				"Minimization Step for each parameter.",
				_stepVec,
				_stepVec);

		registerProcessorParameter("ParamMin",
				"Minimum Value for each parameter",
				_minVec,
				_minVec);

		registerProcessorParameter("ParamMax",
				"Maximum Value for each parameter",
				_maxVec,
				_maxVec);

		registerProcessorParameter("FitterSource",
				"Source for fitter (Minuit2 or GSL)",
				_source,
				std::string("Minuit2"));

		registerProcessorParameter("FitterType",
				"Type of Fitter",
				_type,
				std::string("Simplex"));

		registerProcessorParameter("TrackType",
				"Track type: 0 = Straight, 1 = Helix",
				_trackType,
				int(0));

		registerProcessorParameter("PRFType",
				"Type of PRF. See MMPadResponseFunction.h for options. DEFAULT: \'product\'",
				_prfType,
				std::string("product"));

		registerProcessorParameter("ParOutputPath",
				"Path to save the extracted parameter values",
				_output_path,
				std::string("./"));

		registerOptionalParameter("Tolerance",
				"Tolerance",
				_tolerance,
				double(0.001));

		registerOptionalParameter("MaxIterations",
				"Maximum number of iterations",
				_iter,
				int(100));

		registerOptionalParameter("MaxFcn",
				"Maximum number of function calls",
				_fcn,
				int(100));

		registerOptionalParameter("FitOutput",
				"Fit Output histograms with gaussians?",
				_fitOutput,
				bool(false));

		registerOptionalParameter("RowDep",
				"Add a row dependence to the parameters? (I.e. Param_0 = Param_0_a*row + Param_0_b)",
				_rowDep,
				bool(false));

		registerOptionalParameter("FitTrack",
				"Choose whether or not to include the track chi2 in the minimization process (DEFAULT: true)",
				_fitTrack,
				bool(true));

		registerOptionalParameter("PadHits",
				"Choose whether or not to force a minimum size of hits by addition 0-amplitude pulses",
				_padHits,
				bool(false));

		registerOptionalParameter("NormType",
				"Choose which normalization to use in minimizing the parameters. 0 = sum Amp / sum PRF; 1 = sum Amp; 2 = gaussian amp. (DEFAULT: 0)",
				_normType,
				int(0));

		registerOptionalParameter("ModuleList" ,
				"Optional: List of module numbers to be used in the analysis. This will return a different set of parameters for each module"  ,
				_moduleList,
				_moduleList) ;

	}


	void MMAdvancedCalibrationProcessor::init()
	{ 
		streamlog_out(DEBUG3) << " init() called  " << endl;

		// you might want to list the used parameters by invoking
		printParameters();

		_nPar = MMPadResponseFunction::getNPar(_prfType);

		if(_nPar < 1)
		{
			streamlog_out(ERROR) << "Error : PRF Type \'" << _prfType << "\' not recognized." ;
			throw StopProcessingException(this);
		}

    bool def = true;
    //Warn the user if we are going to have to use default parameters.
		if(_nPar != (int)_startVec.size())
		{
			streamlog_out(WARNING) << "Parameter Length Mismatch : NumPar (" << _nPar << ") does not equal size of parameter start value vector! Using default parameters...";
      def = false;
//			throw StopProcessingException(this);
		}

    else if(_nPar != (int)_stepVec.size())
		{
			streamlog_out(ERROR) << "Parameter Length Mismatch : NumPar (" << _nPar << ") does not equal size of parameter step vector! Using default parameters...";
      def = false;
//			throw StopProcessingException(this);
		}

    else if(_nPar != (int)_minVec.size())
		{
			streamlog_out(ERROR) << "Parameter Length Mismatch : NumPar (" << _nPar << ") does not equal size of parameter min value vector! Using default parameters...";
      def = false;
//			throw StopProcessingException(this);
		}

    else if(_nPar != (int)_maxVec.size())
		{
			streamlog_out(ERROR) << "Parameter Length Mismatch : NumPar (" << _nPar << ") does not equal size of parameter max value vector! Using default parameters...";
      def = false;
//			throw StopProcessingException(this);
		}

		if(parameterSet("ModuleList"))
		{
			_moduleDep = true;
		}
		else
			_moduleDep = false;

		//Now set up the tuples and histograms, accounting for a variable number of parameters
		std::stringstream hitDataString;
		std::stringstream fitDataString;

		hitDataString << "float d0, phi, omega, z0, tanl, chi2, int ndf, double fitX0, diffX0, hitChi2, int hitRow, NPulses, double ";
		fitDataString << "double d0, phi, omega, z0, tanl, fitTrkChi2, fitHitsChi2, int nFitHits, nHits, NCalls, FitResult, float seedD0, seedPhi, seedOmega, seedZ0, seedTanl, seedChi2, int seedNdf, double ";

		std::string sep = "";

		for (int i = 0; i < _nPar; i++)
		{
			if(_moduleDep)
			{
				for(unsigned int j = 0; j < _moduleList.size(); j++)
				{
					hitDataString << sep << "Param_" << i;
					fitDataString << sep << "Param_" << i;

					hitDataString << "_mod" << _moduleList[j];
					fitDataString << "_mod" << _moduleList[j];

					if(_rowDep)
					{
						hitDataString << "_a, Param_" << i << "_mod" << _moduleList[j] << "_b";
						fitDataString << "_a, Param_" << i << "_mod" << _moduleList[j] << "_b";
						hitDataString << ", Param_" << i << "_mod" << _moduleList[j] << "_c";
						fitDataString << ", Param_" << i << "_mod" << _moduleList[j] << "_c";
					}
					sep = ", ";
				}
			}
			else if(_rowDep)
			{
				hitDataString << "_a, Param_" << i << "_b";
				fitDataString << "_a, Param_" << i << "_b";
				hitDataString << ", Param_" << i << "_c";
				fitDataString << ", Param_" << i << "_c";
				sep = ", ";
			}
		}

		_hitData = AIDAProcessor::tupleFactory(this)->create("EventHitData", "EventHitData", hitDataString.str().c_str() );
		_fitData = AIDAProcessor::tupleFactory(this)->create("EventFitData", "EventFitData", fitDataString.str().c_str() );

		//Set up the parameter histograms
		for(int i = 0; i < _nPar; i++)
		{
			if(_moduleDep)
			{
				for(unsigned int j = 0; j < _moduleList.size(); j++)
				{
					if(_rowDep)
					{
						std::stringstream tmp;
						tmp << "Param_" << i << "_mod" << _moduleList[j] << "_a";
						_histogramPar.push_back(new TH1F (tmp.str().c_str(), tmp.str().c_str(), 200, 0., 0.));
						tmp.str(std::string());
						tmp << "Param_" << i << "_mod" << _moduleList[j] << "_b";
						_histogramPar.push_back(new TH1F (tmp.str().c_str(), tmp.str().c_str(), 200, 0., 0.));
						tmp.str(std::string());
						tmp << "Param_" << i << "_mod" << _moduleList[j] << "_c";
						_histogramPar.push_back(new TH1F (tmp.str().c_str(), tmp.str().c_str(), 200, 0., 0.));
						tmp.str(std::string());
					}
					else
					{
						std::stringstream tmp;
						tmp << "Param_" << i << "_mod" << _moduleList[j];
						_histogramPar.push_back(new TH1F (tmp.str().c_str(), tmp.str().c_str(), 200, 0., 0.));
						tmp.str(std::string());
					}
				}
			}
			else if(_rowDep)
			{
				std::stringstream tmp;
				tmp << "Param_" << i << "_a";
				_histogramPar.push_back(new TH1F (tmp.str().c_str(), tmp.str().c_str(), 200, 0., 0.));
				tmp.str(std::string());
				tmp << "Param_" << i << "_b";
				_histogramPar.push_back(new TH1F (tmp.str().c_str(), tmp.str().c_str(), 200, 0., 0.));
				tmp.str(std::string());
				tmp << "Param_" << i << "_c";
				_histogramPar.push_back(new TH1F (tmp.str().c_str(), tmp.str().c_str(), 200, 0., 0.));
			}
			else
			{
				std::stringstream tmp;
				tmp << "Param_" << i;
				_histogramPar.push_back(new TH1F (tmp.str().c_str(), tmp.str().c_str(), 200, 0., 0.));
			}
		}

		//Set up the parameter correlation histograms
		//Note: Let's test the row dependence before going through the effort of working out these correlation histograms
		if(!_rowDep && !_moduleDep)
		{
			for(int i = 0; i < _nPar - 1; i++)
			{
				_histogramCorrPar.push_back(std::vector<TH2F *>());
				for(int j = i+1; j < _nPar; j++)
				{	
					std::stringstream tmp;
					tmp << "Param_" << i << " vs. Param_" << j;
					_histogramCorrPar[i].push_back(new TH2F (tmp.str().c_str(), tmp.str().c_str(), 200, 0., 0., 200, 0., 0.));
				}
			}
		}

    
    
		//Create the minuit minimizer
		min = ROOT::Math::Factory::CreateMinimizer(_source, _type);
		min->Clear();
		min->SetPrintLevel(-1);

		if(parameterSet("Tolerance"))
			min->SetTolerance(_tolerance);
		if(parameterSet("MaxIterations"))
			min->SetMaxIterations(_iter);
		if(parameterSet("MaxFcn"))
			min->SetMaxFunctionCalls(_fcn);
      

    //If no starting values were given, we will use these as defaults.
    if(!def)
    {
      _startVec.insert(_startVec.begin(), _nPar, 1.);
      _stepVec.insert(_stepVec.begin(), _nPar, 1.);
      _minVec.insert(_minVec.begin(), _nPar, -100.); //These limits are just to exclude completely unreasonable parameters.
      _maxVec.insert(_maxVec.begin(), _nPar, 100.);  //There really shouldn't be any parameters on this order.
    }

		//Simplifying parameter procedures
		if(_rowDep || _moduleDep)
		{
			std::vector<float> tmpStart, tmpStep, tmpMin, tmpMax;
			for(int i = 0; i < _nPar; i++)
			{
				if(_moduleDep)
				{
					if(_rowDep)
					{
						for(unsigned int j = 0; j < _moduleList.size(); j++)
						{
							tmpStart.push_back(0);
							tmpStart.push_back(0);
							tmpStart.push_back(_startVec[i]);
							tmpStep.push_back(_stepVec[i]);
							tmpStep.push_back(_stepVec[i]);
							tmpStep.push_back(_stepVec[i]);
							tmpMin.push_back(-10);
							tmpMin.push_back(-40);
							tmpMin.push_back(_minVec[i]);
							tmpMax.push_back(10);
							tmpMax.push_back(40);
							tmpMax.push_back(_maxVec[i]);
						}
					}
					else
					{
						for(unsigned int j = 0; j < _moduleList.size(); j++)
						{
							tmpStart.push_back(_startVec[i]);
							tmpStep.push_back(_stepVec[i]);
							tmpMin.push_back(_minVec[i]);
							tmpMax.push_back(_maxVec[i]);
						}
					}
				}
				else
				{
					tmpStart.push_back(0);
					tmpStart.push_back(0);
					tmpStart.push_back(_startVec[i]);
					tmpStep.push_back(_stepVec[i]);
					tmpStep.push_back(_stepVec[i]);
					tmpStep.push_back(_stepVec[i]);
					tmpMin.push_back(-10);
					tmpMin.push_back(-40);
					tmpMin.push_back(_minVec[i]);
					tmpMax.push_back(10);
					tmpMax.push_back(40);
					tmpMax.push_back(_maxVec[i]);
				}
			}
			_startVec = tmpStart;
			_stepVec = tmpStep;
			_minVec = tmpMin;
			_maxVec = tmpMax;
		}
		if(_moduleDep) _nPar *= _moduleList.size();
		if(_rowDep) _nPar *= 3;
	}


	void MMAdvancedCalibrationProcessor::processRunHeader(LCRunHeader*  /*run*/)
	{
	} 


	void MMAdvancedCalibrationProcessor::processEvent(LCEvent* evt)
	{ 
		/*
		 */

		//TIMER
		clock_t t = clock();

		_curEvt = evt->getEventNumber();

		streamlog_out(DEBUG5) << "Current Event: " << _curEvt << endl;

    

		// use an LCCollectionVec instead of LCCollection so we can use iterators later
		LCCollectionVec* inputColVec = 0;

		for(lcio::StringVec::iterator nameIt = _inputColNames.begin(); nameIt != _inputColNames.end(); ++nameIt)
		{

			// do a try / catch in case there is an empty event
			try
			{
				inputColVec = static_cast<LCCollectionVec *>(evt->getCollection( *nameIt ) );
			}
			catch (DataNotAvailableException &e)
			{
				// no input data, just return and process the next event
				continue;
			}


			//For each track in the collection...
			for( LCCollectionVec::iterator trackIt = inputColVec->begin(); trackIt != inputColVec->end(); ++trackIt )
			{

				IMPL::TrackImpl * aTrack = static_cast<IMPL::TrackImpl *>(*trackIt);	
				const EVENT::TrackerHitVec & trackHits = aTrack->getTrackerHits();

				//Fit for the PRF Parameters
				MMRefitTool refit(_trackType, MMPadResponseFunction::getNPar(_prfType), _prfType, _rowDep, aTrack, _fitTrack, _normType, _padHits);

				if(_moduleDep)
					refit.setModules(_moduleList);

				ROOT::Math::Functor f; 
				f	= ROOT::Math::Functor( refit, _nPar + trackHits.size() );

				//Reset the minimizer
				min->Clear();

				//Set the user parameters
				for(int i = 0; i < _nPar; i++)
				{
					stringstream tmp;
					tmp << "Par_" << i;
          if(_minVec[i] != _maxVec[i])
          {
            min->SetLimitedVariable(i, tmp.str().c_str(), (double)_startVec[i], (double)_stepVec[i], (double)_minVec[i], (double)_maxVec[i]);
          }
          else
          {
            min->SetVariable(i, tmp.str().c_str(), (double)_startVec[i], (double)_stepVec[i]);
          }
				}

				//Set the hit parameters (and remember the start values)
				std::vector<double> hitStartValues((int)trackHits.size());
				std::vector<int> hitRowValues((int)trackHits.size()); //Trying to store row-by-row data

				const gear::TPCParameters& tpcParameters = marlin::Global::GEAR->getTPCParameters();
				gear::Vector2D hitCenterLocal;

				for(unsigned int i = 0; i < trackHits.size(); i++)
				{
					stringstream name;
					name << "HitPos_" << i;

					const EVENT::TrackerHit * aHit = static_cast<const EVENT::TrackerHit *>(trackHits[i]);

					const gear::TPCModule& module = tpcParameters.getModule(static_cast<const EVENT::TrackerPulse *>(aHit->getRawHits()[0])->getCellID1());
					hitCenterLocal = module.globalToLocal( aHit->getPosition()[0], aHit->getPosition()[1]);

					hitStartValues[i] = hitCenterLocal[0]*hitCenterLocal[1];
					hitRowValues[i] = module.getRowNumber(static_cast<const EVENT::TrackerPulse *>(aHit->getRawHits()[0])->getCellID0());

					//Simplified procedure
					min->SetVariable(_nPar + i, name.str().c_str(), hitStartValues[i], 10.);

          //The following commented code is here in case you want to put limits on the hit x0 variables.
					/*
						 int padID =  module.getNearestPad(aHit->getPosition()[0], aHit->getPosition()[1]);
						 gear::Vector2D maxPadCenterLocal = module.getPadCenter( padID );
						 float padWidth = module.getPadWidth( padID ); //Careful. localPadPlane returns width in *local* coordinates.
						 double middle = 0;
						 middle = maxPadCenterLocal[0]*maxPadCenterLocal[1]; // convert to cartesian.
						 padWidth *= maxPadCenterLocal[0]; 

						 double lower = middle - padWidth*1.;
						 double upper = middle + padWidth*1.;

						 min->SetLimitedVariable(_nPar + i, name.str().c_str(), hitCenterLocal[0]*hitCenterLocal[1], 1.e-7, lower, upper );
					 */
    
				}

				// Now set the function to be minimized
				min->SetFunction(f);

				// Minimize the function
				if(!min->Minimize())
				{
          continue;
				}

				const double *results = min->X();

				std::vector<double> fitTrackPar = refit.retrievePar(results);
				std::vector<double> fitHitChi2 = refit.retrieveChi2(results);

				//Fill the tuple
				int index = 0;
				_fitData->fill(index++, fitTrackPar[0]);  
				_fitData->fill(index++, fitTrackPar[1]);
				_fitData->fill(index++, fitTrackPar[2]); 
				_fitData->fill(index++, fitTrackPar[3]); 
				_fitData->fill(index++, fitTrackPar[4]); 
				_fitData->fill(index++, fitTrackPar[5]);
				_fitData->fill(index++, fitTrackPar[6]);
				_fitData->fill(index++, (int)fitTrackPar[7]); 
				_fitData->fill(index++, (int)trackHits.size());
				_fitData->fill(index++, (int)min->NCalls());
				_fitData->fill(index++, (int)min->Status());
				_fitData->fill(index++, aTrack->getD0());
				_fitData->fill(index++, aTrack->getPhi());
				_fitData->fill(index++, aTrack->getOmega());
				_fitData->fill(index++, aTrack->getZ0());
				_fitData->fill(index++, aTrack->getTanLambda());
				_fitData->fill(index++, aTrack->getChi2());
				_fitData->fill(index++, aTrack->getNdf());
				for(int i = 0; i < _nPar; i++)
				{
					_fitData->fill(index++, (double)(results[i]));
					_histogramPar[i]->Fill(results[i], 1./(fitTrackPar[5] + fitTrackPar[6]));
				}
				_fitData->addRow();

				//Fill up the data on the hits 
				for(unsigned int i = 0; i < trackHits.size(); i++)
				{
					index = 0;
					_hitData->fill(index++, aTrack->getD0());
					_hitData->fill(index++, aTrack->getPhi());
					_hitData->fill(index++, aTrack->getOmega());
					_hitData->fill(index++, aTrack->getZ0());
					_hitData->fill(index++, aTrack->getTanLambda());
					_hitData->fill(index++, aTrack->getChi2());
					_hitData->fill(index++, (int)aTrack->getNdf());
					_hitData->fill(index++, results[_nPar + i]);
					_hitData->fill(index++, results[_nPar + i] - hitStartValues[i]);
					_hitData->fill(index++, fitHitChi2[i]);  //Adding Row Data
					_hitData->fill(index++, hitRowValues[i]);//Adding Row Data
					const EVENT::TrackerHit * aHit = static_cast<const EVENT::TrackerHit *>(trackHits[i]);
					_hitData->fill(index++, (int)(aHit->getRawHits().size()));

					for(int j = 0; j < _nPar; j++)
          {
						_hitData->fill(index++, (double)(results[j]));
          }
					_hitData->addRow();
				}

				if(!_rowDep && !_moduleDep)
				{
					for(int i = 0; i < _nPar - 1; i++)
					{
						for(int j = 0; j < _nPar - 1 - i; j++)
						{	
							if(!_rowDep)
              {
								_histogramCorrPar[i][j]->Fill(results[i], results[j + i + 1], 1./(fitTrackPar[5] + fitTrackPar[6]));
              }
							else
              {
								_histogramCorrPar[i][j]->Fill(results[2*i + 1], results[2*j + 2*i + 3], 1./(fitTrackPar[5] + fitTrackPar[6]));
              }
						}
					}
				}

			}//End loop over tracks

		}//End loop over input track collections

		t = clock() - t;
		streamlog_out(MESSAGE4) << "Execution took : " << ((float)t)/CLOCKS_PER_SEC << "s." << std::endl;

	}//End processEvent()


	void MMAdvancedCalibrationProcessor::check(LCEvent*  /*evt*/)
	{ 
	}

	void MMAdvancedCalibrationProcessor::end()
	{ 
		for(int i = 0; i < _nPar; i++)
		{
			Double_t mu_b = _histogramPar[i]->GetMean();
			Double_t sigma_b = _histogramPar[i]->GetRMS();

			//If desired, fit the histograms with gaussians, and use the gaussian mean and sigma as mu and sigma
			if(_fitOutput)	
			{
				TFitResultPtr r_b = 0;

				r_b = _histogramPar[i]->Fit("gaus","wS","", mu_b - 2.*sigma_b, mu_b + 2.*sigma_b);

				if((int)r_b != 0)
				{
					streamlog_out(MESSAGE) << "Error! Fit of the Param_" << i << " Histogram(s) failed! Please investigate." << std::endl;

					ofstream out_error(std::string(_output_path + "/fit_error.log").c_str());
					out_error << "Error! Fit of the Param_" << i << " Histogram(s) failed! See below:" << std::endl << std::endl;
					static_cast<ROOT::Fit::FitResult&>(*r_b).Print(out_error);
					out_error.close();

					throw StopProcessingException(this);
				}
				mu_b = _histogramPar[i]->GetFunction("gaus")->GetParameter(1);
				sigma_b = _histogramPar[i]->GetFunction("gaus")->GetParameter(2);
				streamlog_out(MESSAGE) << "Fit Parameter " << i << "_b = " << mu_b << " +/- " << sigma_b << std::endl;
			}

			//Useful variables
			ofstream out;
			stringstream path;

			//Then save the results to file
			path << _output_path << "/par_" << i;
			out.open(path.str().c_str());
			if(out.is_open())
				out << mu_b << " +/- " << sigma_b;
			else
			{
				streamlog_out(DEBUG3) << "Error! Provided path: " << _output_path << " is invalid!" << std::endl;
				throw StopProcessingException(this);
			}
			out.close();
		}
	}

} //End marlin namespace
