#include "MMScatterProcessor.h"

#include "MMPadResponseFunction.h"
#include "MMHitFinderProcessor.h"

#include <iostream>
#include <fstream>
#include <string>
#include <cmath>

#include <TROOT.h>
#include <TH2D.h>
#include <TF1.h>
#include <TFitResultPtr.h>
#include <TGraph.h>
#include <TGraphErrors.h>
#include <TProfile.h>
#include <TFitResult.h>
#include <TVector2.h>
#include <Math/ParamFunctor.h>

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/TrackerData.h>
#include <EVENT/TrackerPulse.h>
#include <Exceptions.h>
#include <marlin/Exceptions.h>

//Other
#include "IntersectionCalculator.h"

using namespace lcio;
using namespace marlin;

namespace marlintpc
{

  std::vector<double> MMScatterProcessor::CALIB_PARAMS; //Have to "initialize" static vector here.

  MMScatterProcessor aMMScatterProcessor;


  MMScatterProcessor::MMScatterProcessor() : Processor("MMScatterProcessor")
  {
    // the processor description
    _description = "MMScatterProcessor creates necessary scatter-plots and fits appropriate data to calibrate the PRF.";

    _inputColNames.push_back("TPCTracks");

    registerInputCollections(    LCIO::TRACK,
        "InputTrackCollections" ,
        "Names of the input Track collections, default: TPCTracks"  ,
        _inputColNames ,
        _inputColNames) ;

    registerOptionalParameter("SingleTrack",
        "True if only single-track events should be used.",
        _singleTrack,
        false);

    registerOptionalParameter("SmallAngle",
        "True if only hits with local track angles < 5 degrees should be used.",
        _smallAngle,
        false);

    registerProcessorParameter("PRFType",
        "Type of PRF. See MMPadResponseFunction.h for current options.",
        _prfType,
        std::string("product"));

    registerProcessorParameter("NormType",
        "Type of PRF Normalization. Sum of Pulses / Sum of PRF = 0, Sum of Pulses = 1, Gaussian Norm = 2, PRF at Chi2 minimum = 3 (DEFAULT: 1)",
        _normType,
        1);

    registerProcessorParameter("PRF_Parameters",
        "Values of the PRF parameters.",
        _parVec,
        _parVec);

    //Optional Parameters
    registerOptionalParameter("FitPRF",
        std::string("True if the PRF should be fit to the scatter plot(s). "),
        _fitPRF,
        false);

    registerOptionalParameter("CalibRun",
        std::string("Produce a scatter plot after every \"CalibIterations\" events, fit it with the PRF, and store the results in a static variable. "),
        _calibRun,
        false);

    registerOptionalParameter("CalibIterations",
        std::string("If using the calibration mode, this variable sets the number of events-per-iteration to use. "),
        _calibIter,
        200);

    registerOptionalParameter("CalibPad",
        std::string("If using the calibration mode, this variable sets whether or not to force 5-pulse hits by adding data with 0 amplitude. "),
        _calibPad,
        false);

    registerOptionalParameter("RowDep",
        "Optional: Add a row dependence to the parameters? (I.e. Param_0 = Param_0_a*row + Param_0_b)",
        _rowDep,
        bool(false));

    registerOptionalParameter("Profiles",
        "Optional: Make TProfiles out of the scatter plots",
        _profile,
        bool(false));

    registerOptionalParameter("ModuleList" ,
        "Optional: List of module numbers to be used in the analysis. This will return a different set of parameters for each module"  ,
        _moduleList,
        _moduleList) ;

    registerOptionalParameter("ParOutputPath",
        "Path to save the extracted parameter values",
        _output_path,
        std::string("./"));
  }


  void MMScatterProcessor::init()
  { 
    streamlog_out(DEBUG) << " init() called  " << std::endl;

    printParameters();

    // Make sure there's no row or module dependence if using CalibRun. This can be properly implemented later
    if(_calibRun && (_rowDep || _moduleDep))
    {
      streamlog_out(ERROR) << "Error : Currently, no module or row dependence can be used in calibration mode." << std::endl;
      throw StopProcessingException(this);
    }

    //Verify number and type of parameters
    _nPar = MMPadResponseFunction::getNPar(_prfType);

    if(_nPar < 1)
    {
      streamlog_out(ERROR) << "Error : PRF Type not recognized." << std::endl;
      throw StopProcessingException(this);
    }

    if(parameterSet("ModuleList")) _moduleDep = true;

    int reqSize = _nPar;
    if(_rowDep) reqSize *= 3;
    if(_moduleDep) reqSize *= _moduleList.size();

    /** TODO: Set additional braces, according to indentation, to prevent
     *  warnings: suggest explicit braces to avoid ambiguous 'else'
     *  Please check carefully, if braces are still set correctly!
     */
    if(!_calibRun && (_normType == 0 || _normType == 3) && reqSize != (int)_parVec.size())
    {
      if(_rowDep)
      {
        if(_moduleDep)
        {
          streamlog_out(ERROR) << "Error : PRF Type \'" << _prfType << "\' requires " << _nPar << " parameters, but " << (double)_parVec.size()/(2.*_moduleList.size()) << " were provided." ;
        }
        else
          streamlog_out(ERROR) << "Error : PRF Type \'" << _prfType << "\' requires " << _nPar << " parameters, but " << (double)_parVec.size()/(2.) << " were provided." ;
      }
      else if (_moduleDep)
      {
        streamlog_out(ERROR) << "Error : PRF Type \'" << _prfType << "\' requires " << _nPar << " parameters, but " << (double)_parVec.size()/(_moduleList.size()) << " were provided." ;
      }
      else
        streamlog_out(ERROR) << "Error : PRF Type \'" << _prfType << "\' requires " << _nPar << " parameters, but " << (int)_parVec.size() << " were provided." ;

      throw StopProcessingException(this);
    }

    _parArray = new double[reqSize];
    for(int i = 0; i < reqSize; i++)
    {
      if(_calibRun)
      {
        _parArray[i] = 1.;
      }
      else
      {
        _parArray[i] = _parVec[i];
      }
    }

    if(_moduleDep)
    {
      for(unsigned int i = 0; i < _moduleList.size(); i++)
      {
        _moduleMap[_moduleList[i]] = i;
      }
    }

    if(!_calibRun && !_moduleDep && !_rowDep)
    {
      std::stringstream name;
      name << "double ";
      std::string delim = "";
      for(int i = 0; i < _nPar; i++)
      {
        name << delim << "Par_" << i << ", ParErr_" << i;
        delim = ", ";
      }
      fitData = AIDAProcessor::tupleFactory(this)->create("EventFitData", "EventFitData", name.str().c_str() );
    }

    //Store the fit results for each iteration
    if(_calibRun)
    {
      _iterationPar.reserve(reqSize);
    }

  }


  void MMScatterProcessor::processRunHeader(LCRunHeader*  /*run*/)
  {
  } 


  void MMScatterProcessor::processEvent(LCEvent* evt)
  { 

    _curEvt = evt->getEventNumber();
    _nEvt++; //For calibration

    // use an LCCollectionVec instead of LCCollection so we can use iterators later
    LCCollectionVec* inputColVec = 0;

    for(lcio::StringVec::iterator nameIt = _inputColNames.begin(); nameIt != _inputColNames.end(); ++nameIt)
    {

      // do a try / catch in case there is an empty event
      try
      {
        inputColVec = static_cast<LCCollectionVec *>(evt->getCollection( *nameIt ) );
      }
      catch (DataNotAvailableException &e)
      {
        // no input data, just return and process the next event
        continue;
      }

      //Needed for sorting pulses
      static const gear::TPCParameters& tpcParameters = Global::GEAR->getTPCParameters();

      if(_singleTrack && (int)inputColVec->size() != 1) continue; //Cut on multiple-track events

      //For each track in the collection...
      for( LCCollectionVec::iterator trackIt = inputColVec->begin(); trackIt != inputColVec->end(); ++trackIt )
      {
        IMPL::TrackImpl * aTrack = static_cast<IMPL::TrackImpl *>(*trackIt);	

        const EVENT::TrackerHitVec& theHits = aTrack->getTrackerHits();

        //Sort the pulses
        SortedPulseMap sortedPulses;
        _sortPulses(theHits, tpcParameters, sortedPulses);

        //Fill the scatter-plot data
        PointsMap points;

        _fillData(sortedPulses, tpcParameters, aTrack, points);

        if(points.size() == 0) continue;

        MMPadResponseFunction thePRF(_prfType);

        switch(_normType)
        {
          case 1: thePRF.SetFit(MMPadResponseFunction::SUM_FIT); break;
          case 2: thePRF.SetFit(MMPadResponseFunction::GAUS_FIT); break;
          case 3: thePRF.SetFit(MMPadResponseFunction::CHI2_MIN); break;
          default: thePRF.SetFit(MMPadResponseFunction::FIT_BOTH); break;
        }

        //Fill data for the scatter plots 
        for(PointsMap::iterator rowIt = points.begin(); rowIt != points.end(); ++rowIt)
        {

          //Calculate the parameters, if the desired normalization uses them.
          if(_calibRun && CALIB_PARAMS.size() == 0) thePRF.SetFit(MMPadResponseFunction::GAUS_FIT);
          else if(_normType == 0 || _normType == 3)
          {
            int modIndex = -1;
            int rowIndex = -1;
            if(_moduleDep)
            {
              std::map<int, int>::const_iterator modIt = _moduleMap.find(rowIt->first.first);
              if(modIt == _moduleMap.end())
              {
                _moduleMap[rowIt->first.first] = 0;
                modIndex = 0;
              }
              else
                modIndex = modIt->second;
            }
            if(_rowDep) rowIndex = rowIt->first.second;
            //const double * tmpPar = MMHitFinderProcessor::calcParArray(_parArray, _nPar, (int)_moduleList.size(), modIndex, rowIndex);
            const std::vector<double> tmpPar = MMHitFinderProcessor::calcParArray(_parArray, _nPar, (int)_moduleList.size(), modIndex, rowIndex);
            //thePRF.SetParameters(tmpPar, _nPar);
            thePRF.SetParameters(&tmpPar[0], _nPar);
          }

          thePRF.normalizeData(rowIt->second);

          //Now add the data to the total scatter plot(s)
          std::pair<int, int> key;
          if(_moduleDep)
          {
            if(_rowDep)
            {
              key = rowIt->first;
            }
            else
            {
              key = std::make_pair(rowIt->first.first, 0);
            }
          }
          else if(_rowDep)
          {
            key = std::make_pair(0, rowIt->first.second);
          }
          else
          {
            key = std::make_pair(0, 0);
          }

          PointsMap::iterator totalIt = _scatterMap.find(key);
          if(totalIt == _scatterMap.end()) 
          {
            _scatterMap[key] = rowIt->second;
          }
          else
            totalIt->second.insert(totalIt->second.end(), rowIt->second.begin(), rowIt->second.end());

        }//End Scatter-plot row for-loop

      }//End loop over tracks

    }//End loop over input track collections

    if(_calibRun) calibrate();

  }//End processEvent()


  void MMScatterProcessor::check(LCEvent*  /*evt*/)
  { 
  }

  void MMScatterProcessor::end()
  { 
    //If calibRun, just produce the fit PRF parameter plots and go.
    if(_calibRun)
    {
      marlin::AIDAProcessor::tree(this);

      int parNum = 0;
      for(std::vector<std::vector<double> >::iterator parIt = _iterationPar.begin(); 
          parIt != _iterationPar.end(); ++parIt)
      {
        std::stringstream name;
        name << "Fit Par " << parNum;
        TGraph * gr = new TGraph(_iterationNumbers.size(), &(_iterationNumbers[0]), &((*parIt)[0]));
        gr->SetNameTitle(name.str().c_str(), name.str().c_str());
        gr->Write();

        //Output the fit parameters
        std::ofstream out;
        std::stringstream path;
        path.str(std::string());	
        path << _output_path << "/par_" << parNum++;
        out.open(path.str().c_str());
        if(out.is_open())
          out << parIt->back() << " +/- " << 0.000;  //A proper error should be determined later
        else
        {
          streamlog_out(DEBUG3) << "Error! Provided path: " << _output_path << " is invalid!" << std::endl;
          throw StopProcessingException(this);
        }
        out.close();
      }

      return;
    }

    //Create the TH2D's
    for(PointsMap::iterator rowIt = _scatterMap.begin(); rowIt != _scatterMap.end(); ++rowIt)
    {
      //Set up the TH2D
      std::stringstream name;
      name << "Scatter Plot";
      if(_moduleDep)
        name << " for Module " << rowIt->first.first;
      if(_rowDep)
        name << " and Row " << rowIt->first.second;
      TH2D* tmp = new TH2D(name.str().c_str(), name.str().c_str(), 2000, -10, 10, 2000, 0, 1.2);

      //Add the points
      for(std::vector<std::pair<double, float> >::iterator pointIt = rowIt->second.begin(); pointIt != rowIt->second.end(); ++pointIt)
      {
        tmp->Fill(pointIt->first, pointIt->second);
      }

      if(_profile)
      {
        std::string tail = name.str().substr(11);
        name.str(std::string());
        name << "Profile" << tail;
        tmp->RebinX(8);
        //TProfile * tmpP = tmp->ProfileX(name.str().c_str(), 1, -1, "s"); //Sets the errors to be the RMS
        TProfile * tmpP = tmp->ProfileX(name.str().c_str()); //Sets the errors to be the RMS/sqrt(n)
        _scatterPlots[rowIt->first] = tmpP;
      }
      else
        _scatterPlots[rowIt->first] = tmp;
    }

    //If fitting the PRF, extract the fit parameters here.
    if(_fitPRF)
    {

      //For the output...
      std::ofstream out;
      std::stringstream path;

      //Perform the fits:
      std::vector<double> xVals;
      std::vector<double> * yVals = new std::vector<double>[_nPar];
      std::vector<double> * yValErrs = new std::vector<double>[_nPar];
      PRFFcn fcn(_prfType, _nPar);
      TF1 * fit = new TF1("fit", fcn, -5., 5., _nPar + 2);
      for(int i = 0; i < _nPar; i++) fit->SetParameter(i, _parArray[i]);
      fit->SetParameter(_nPar, 1.); //Fit for amplitude
      fit->SetParameter(_nPar+1, 0.); //Fit for x0
      for(std::map<std::pair<int, int>, TH1*>::iterator rowIt = _scatterPlots.begin(); rowIt != _scatterPlots.end(); ++rowIt)
      {
        TFitResultPtr r = 0;
        r = rowIt->second->Fit(fit,"WSQR","");

        if(r == 0)
        {
          //First, retrieve the parameters
          for(int i = 0; i < _nPar; i++)
          {
            yVals[i].push_back(rowIt->second->GetFunction("fit")->GetParameter(i));
            yValErrs[i].push_back(rowIt->second->GetFunction("fit")->GetParError(i));
          }

          //Now, calculate the approriate x-coordinate;
          if(_rowDep)
          {
            static const gear::TPCParameters& tpcParameters = marlin::Global::GEAR->getTPCParameters();

            const gear::TPCModule& theModule = tpcParameters.getModule(rowIt->first.first);
            gear::Vector2D padcenter = theModule.getPadCenter(theModule.getPadIndex(rowIt->first.second,0));

            if(_moduleDep)
            {
              double radius = sqrt(padcenter[0]*padcenter[0]+padcenter[1]*padcenter[1]);
              xVals.push_back(radius);
            }

            else //If not module dependent, then use local coordinates to ensure all modules behave the same
            {
              gear::Vector2D padCenterLocal = theModule.globalToLocal(padcenter[0], padcenter[1]);

              if((theModule.getLocalPadLayout()).getPadLayoutType() == 2)//if circular pad layout
                xVals.push_back(padCenterLocal[0]);
              if((theModule.getLocalPadLayout()).getPadLayoutType() == 1)//if rectangular pad layout
              {
                double radius = sqrt(padCenterLocal[0]*padCenterLocal[0]+padCenterLocal[1]*padCenterLocal[1]);
                xVals.push_back(radius);
              }
            }
          }
          else if(_moduleDep)
          {
            xVals.push_back(rowIt->first.first);
          }
        }
        else std::cout << "ERROR! Fit Failed!" << std::endl;
      }

      if(_moduleDep || _rowDep)
      {
        std::vector<double> xValErrs(xVals.size(), 0.);
        TGraphErrors ** graphs = new TGraphErrors*[_nPar];

        for(int i = 0; i < _nPar; i++)
        {
          graphs[i] = new TGraphErrors(xVals.size(), &(xVals[0]), &(yVals[i][0]), &(xValErrs[0]), &(yValErrs[i][0]));
          std::stringstream name;
          name << "Fit Par " << i;
          graphs[i]->SetNameTitle(name.str().c_str(), name.str().c_str());
          graphs[i]->GetYaxis()->SetTitle(name.str().substr(4).c_str());
          if(_rowDep)
            graphs[i]->GetXaxis()->SetTitle("Row Radius (mm)");
          else
            graphs[i]->GetXaxis()->SetTitle("Module Number");

          graphs[i]->Write();
        }
      }
      else
      {
        for(int i = 0; i < _nPar; i++)
        {
          fitData->fill(2*i, yVals[i][0]);
          fitData->fill(2*i+1, yValErrs[i][0]);
        }
        fitData->addRow();
      }

    }//End fitPRF conditional

  }

  void MMScatterProcessor::calibrate()
  {
    /* At every [interval] events, fit the PRF to the current scatter plot, and save the results in the CALIB_PARAMS vector for use by the HitFinder. */
    /* At the moment, we will assume (and enforce) no module or row dependence */

    static int denom = _calibIter;

    ++_nEvt;
    if((_nEvt % denom) != 0) return; //Enforce the requested number of iterations

    //Create the TH2D
    PointsMap::iterator rowIt = _scatterMap.begin();

    if(rowIt == _scatterMap.end()) 
    {
      streamlog_out(DEBUG) << "No points in iteration " << _nIterations << "!" << std::endl;
      return;
    }

    std::stringstream name;
    name << "Scatter Plot, Iteration " << _nIterations;
    TH2D* tmp = new TH2D(name.str().c_str(), name.str().c_str(), 2000, -5, 5, 2000, 0, 1.2);

    //Add the points
    for(std::vector<std::pair<double, float> >::iterator pointIt = rowIt->second.begin(); pointIt != rowIt->second.end(); ++pointIt)
    {
      tmp->Fill(pointIt->first, pointIt->second);
    }

    PRFFcn fcn(_prfType, _nPar);
    TF1 * fit = new TF1("fit", fcn, -5., 5., _nPar + 2);
    for(int i = 0; i < _nPar; i++) fit->SetParameter(i, _parArray[i]);
    fit->SetParameter(_nPar, 1.); //Fit for amplitude
    fit->SetParameter(_nPar+1, 0.); //Fit for x0

    TFitResultPtr r = 0;
    r = tmp->Fit(fit,"WSQR","");

    if(r == 0)
    {

      MMScatterProcessor::CALIB_PARAMS.clear(); //Wipe the old parameters assuming the fit was successful. If not, try the next iteration.
      streamlog_out(DEBUG) << "Iteration " << _nIterations;
      //First, retrieve the parameters
      for(int i = 0; i < _nPar; i++)
      {
        //If first iteration, ensure vectors exist?
        if(_nIterations == 0)
          _iterationPar.push_back(std::vector<double>());
        _iterationPar[i].push_back(tmp->GetFunction("fit")->GetParameter(i));
        _parArray[i] = tmp->GetFunction("fit")->GetParameter(i);
        MMScatterProcessor::CALIB_PARAMS.push_back(tmp->GetFunction("fit")->GetParameter(i));
        streamlog_out(DEBUG) << "\tPar " << i << " = " << tmp->GetFunction("fit")->GetParameter(i);
      }
      streamlog_out(DEBUG) << std::endl;

      _iterationNumbers.push_back(++_nIterations);

    }
    else
    {
      streamlog_out(DEBUG) << "Failed to fit iteration " << _nIterations << "! Continuing..." << std::endl;
      //delete tmp;
      _nIterations++;
    }

    _scatterMap.clear();

  } //End calibrate() method.

  void MMScatterProcessor::_sortPulses( const	TrackerHitVec& inputHits, const gear::TPCParameters& tpcParameters, 
      SortedPulseMap& sortedPulses	)
  {
    /**  
      Taken almost directly from MMHitFinderProcessor.

      Simple method. Procedure is as follows:

      -Check hits one-by-one:
      -For each pulse in the hit:
      -Check if its module is already in the map:
      -if it is not there, add it;
      -Check if its row is in the map:
      -if it is not there, add it;
      -Add the pulse to the vector of the appropriate row in the appropriate module;
     **/

    for(TrackerHitVec::const_iterator hitIt = inputHits.begin(); hitIt != inputHits.end(); ++hitIt)
    {
      //Get the pulses in the hit
      LCObjectVec hitPulses = (*hitIt)->getRawHits();

      //For each pulse...
      for(LCObjectVec::iterator pulseIt = hitPulses.begin(); pulseIt != hitPulses.end(); ++pulseIt)
      {

        TrackerPulse* aPulse = static_cast<TrackerPulse*>(*pulseIt);
        if (aPulse)
        {
          const int moduleID         = aPulse->getCellID1();
          const int padID            = aPulse->getCellID0();

          const gear::TPCModule& theModule = tpcParameters.getModule(moduleID);
          const int rowID            = theModule.getRowNumber(padID);

          SortedPulseMap::iterator modIt = sortedPulses.find(moduleID); //Look for it in the SortedPulseMap 
          if(modIt != sortedPulses.end())  //If it's there...
          {
            std::map<int, std::vector<TrackerPulse *> >::iterator rowIt = (*modIt).second.find(rowID); //Look for the row
            if(rowIt != (*modIt).second.end()) //If it's also there...
            {
              (*rowIt).second.push_back(aPulse);
            }
            else //Otherwise, we need a new vector there.
            {
              ((*modIt).second)[rowID] = std::vector<TrackerPulse *>();
              ((*modIt).second)[rowID].push_back(aPulse);
            }
          }
          else //Otherwise, we need a new module entry
          {
            sortedPulses[moduleID] = std::map<int, std::vector<TrackerPulse *> >();
            sortedPulses[moduleID][rowID] = std::vector<TrackerPulse *>();	
            sortedPulses[moduleID][rowID].push_back(aPulse);
          }

        }//End valid pulse condition

      }//End pulses for-loop

    }//End hits for-loop

  }//End _sortPulses()

  void MMScatterProcessor::_fillData(  	const SortedPulseMap& sortedPulses, const gear::TPCParameters& tpcParameters, const IMPL::TrackImpl* theTrack,
      PointsMap& points)
  {
    /*
     *	Procedure:
     *
     *	-For each module:
     *	-For each row:
     *	-Calculate the track's position on that row
     *	-For each pulse on the row:
     *	-Calculate its distance from the track
     *	-Add it to the vector of points
     */

    //For each module...
    for(SortedPulseMap::const_iterator modIt = sortedPulses.begin(); modIt != sortedPulses.end(); ++modIt)
    {

      //Get the module and its layout
      const gear::TPCModule &module = tpcParameters.getModule(modIt->first);
      const gear::PadRowLayout2D &localPadPlane = module.getLocalPadLayout();

      //For each row...
      for(std::map<int, std::vector<TrackerPulse*> >::const_iterator rowIt = modIt->second.begin(); rowIt != modIt->second.end(); ++rowIt)
      {

        std::pair<double, double> trackPos = calculateTrackPosition(tpcParameters, theTrack, modIt->first, rowIt->first);

        //Get the track's position in local coordinates
        gear::Vector2D trackPositionLocal = module.globalToLocal(trackPos.first, trackPos.second);

        if(trackPos.first == -9E9)
          streamlog_out(DEBUG) << "ERROR! TrackPos is not calculated!" << std::endl;

        //Cut on local angle
        if(_smallAngle)
        {
          //calculate angle between hit and track
          int padIndex = module.getNearestPad(trackPos.first, trackPos.second);
          gear::Vector2D padCenterGlobal = module.getPadCenter(padIndex);
          gear::Vector2D e_pad (padCenterGlobal[0]-module.getOffset()[0],padCenterGlobal[1]-module.getOffset()[1]);
          double phi_s=atan2(sin(theTrack->getPhi())-theTrack->getOmega()*(trackPos.first-(-theTrack->getD0()*sin(theTrack->getPhi())+theTrack->getReferencePoint()[0])),
              cos(theTrack->getPhi())+theTrack->getOmega()*(trackPos.second-(theTrack->getD0()*cos(theTrack->getPhi())+theTrack->getReferencePoint()[1])));

          double  phiHit=phi_s-atan(e_pad[1]/e_pad[0]);

          if(phiHit > 5.*M_PI/180. || phiHit < -5.*M_PI/180.) continue;
        }

        //Only add a row if it's going to be included, obviously.
        std::vector<std::pair<double, float> > rowVec;

        //If desired, add pulses with 0 amplitude to fill out small hits. This should help bring down tails and widen the PRF.
        if(_calibPad)
        {
          TrackerPulse * maxPulse = NULL;
          double max = 0.;

          for(std::vector<TrackerPulse*>::const_iterator pulseIt = rowIt->second.begin(); pulseIt != rowIt->second.end(); ++pulseIt)
          {
            if((*pulseIt)->getCharge() > max) { maxPulse = *pulseIt; max = (*pulseIt)->getCharge(); }
          }

          if(!maxPulse)	continue;

          int maxCol = module.getPadNumber(maxPulse->getCellID0());

          for(int iC = maxCol - 2; iC < maxCol + 3; iC++)
          {
            std::vector<TrackerPulse*>::const_iterator pulseIt;
            for(pulseIt = rowIt->second.begin(); pulseIt != rowIt->second.end(); ++pulseIt)
            {
              if(module.getPadNumber((*pulseIt)->getCellID0()) == iC) { break; }
            }

            if(pulseIt == rowIt->second.end() && iC >= 0 && iC < (int)module.getPadsInRow(rowIt->first).size())
            {

              int curPad = module.getPadIndex(rowIt->first, iC);
              gear::Vector2D	curPadLocal = module.globalToLocal(module.getPadCenter(curPad)[0], module.getPadCenter(curPad)[1]); 
              double dist = trackPositionLocal[1] - curPadLocal[1];
              rowVec.push_back(std::pair<double, float>(trackPositionLocal[0]*dist, 0.));

              continue;
            }
          }
        }
        ///////////////////////////////////////////////////////////////////////

        for(std::vector<TrackerPulse*>::const_iterator pulseIt = rowIt->second.begin(); pulseIt != rowIt->second.end(); ++pulseIt)
        {

          //Get the pad index
          int padID = (*pulseIt)->getCellID0();

          //Get the location of the pad center.
          gear::Vector2D padPositionGlobal = module.getPadCenter(padID);

          //Change the coordinates to local coordinates
          gear::Vector2D pulsePositionLocal = module.globalToLocal(padPositionGlobal[0], padPositionGlobal[1]);

          double dist = -9E9;

          if (localPadPlane.getCoordinateType() == gear::PadRowLayout2D::CARTESIAN){
            dist = trackPositionLocal[0] - pulsePositionLocal[0]; // cartesian x
          }
          else if (localPadPlane.getCoordinateType() == gear::PadRowLayout2D::POLAR){
            double dPhi = trackPositionLocal[1] - pulsePositionLocal[1];
            dist = dPhi*trackPositionLocal[0]; 
          }

          if(dist == -9E9)
            streamlog_out(DEBUG) << "ERROR! Dist is not calculated!" << std::endl;

          // Get the amplitude of the pad.
          float amp = (*pulseIt)->getCharge();
          //Add the point to the scatter-plot
          rowVec.push_back(std::pair<double, float>(dist, amp));

        }//End pulse for-loop

        points[std::pair<int,int>(modIt->first, rowIt->first)] = rowVec;

      }//End row for-loop

    }//End module for-loop

  }//End _fillData()

  std::pair<double, double> MMScatterProcessor::calculateTrackPosition (	const gear::TPCParameters& tpcParameters, const IMPL::TrackImpl* theTrack, int mod, int row )
  {
    //Get the track parameters
    float d0 = theTrack->getD0();
    float phi0 = theTrack->getPhi();
    float omega = theTrack->getOmega();
    const float * ref = theTrack->getReferencePoint();

    if(omega == 0)
    { 
      return calculateStraightTrackPosition(tpcParameters, d0, phi0, ref, mod, row);
    }

    else
    {
      return calculateTrackPosition(tpcParameters, d0, phi0, omega, ref, mod, row);
    }
  }

  std::pair<double, double> MMScatterProcessor::calculateStraightTrackPosition (	const gear::TPCParameters& tpcParameters, double d0, double phi0, const float * ref, int mod, int row )
  {
    /*
       Writing this after the advent of the IntersectionCalculator, so that will be used here.
     */

    const gear::TPCModule &module = tpcParameters.getModule(mod);
    gear::Vector2D aPadInRow = module.getPadCenter(module.getPadsInRow(row)[0]); 
    gear::Vector2D local  = module.globalToLocal(aPadInRow[0], aPadInRow[1]); 

    std::pair<double,double> phi_intersec(0,0); 
    try
    {
      //calculate two points on the track and transform them into local padplane coordinates to obtain a track parametrisation
      // in local coordionates
      //point1=point of closest approach, should inlcude referencepoint in case it is not 0

      double pca_x = ( -1.0 * d0 ) * std::sin(phi0) + ref[0];
      double pca_y = ( d0 ) * std::cos(phi0) + ref[1];
      //point 2
      double p2x=pca_x+std::cos(phi0);
      double p2y=pca_y+std::sin(phi0);
      //transform both points to local coordinate system
      //this gives back r and phi, not x and y for polar padlayout!
      gear::Vector2D pca_local = module.globalToLocal(pca_x,pca_y);
      gear::Vector2D p2_local = module.globalToLocal(p2x,p2y);
      double pca_local_x=pca_local[0]*cos(pca_local[1]);
      double pca_local_y=pca_local[0]*sin(pca_local[1]);
      double p2_local_x=p2_local[0]*cos(p2_local[1]);
      double p2_local_y=p2_local[0]*sin(p2_local[1]);
      double phi=atan2(p2_local_y-pca_local_y,p2_local_x-pca_local_x);
      phi_intersec= marlintpc::IntersectionCalculator::circle_straightLine(local[0], pca_local_x,pca_local_y,phi);
    }
    catch(std::exception &e) 
    {
      std::cout<<"NO INTERSECTION: " <<local[0]<<" "
        <<" "<< d0 <<" "<< phi0 << std::endl;
      return std::pair<double, double>(-9E9, -9E9);
    }
    double trackPhi = fabs(phi_intersec.first - local[1]) < fabs(phi_intersec.second - local[1]) ? 
      phi_intersec.first : phi_intersec.second;

    gear::Vector2D trackLocal = module.localToGlobal(local[0], trackPhi);

    return std::pair<double, double>(trackLocal[0], trackLocal[1]);

  }//End find track for straight tracks

  std::pair<double, double> MMScatterProcessor::calculateTrackPosition (	const gear::TPCParameters& tpcParameters, double d0, double phi0, double omega, const float * ref, int mod, int row )
  {
    /*
     * Taken from the Kalman Libraries TCircle class. 
     *
     */

    //Calculate the centre of curvature (Note: taken from the LCIO Track Parameters note)
    double xC = (double)(ref[0]) + (1./omega - d0)*sin(phi0);
    double yC = double(ref[1]) - (1./omega - d0)*cos(phi0);

    //Get the radius of the row, as well as the centre of curvature of the module
    const gear::TPCModule &module = tpcParameters.getModule(mod);
    gear::Vector2D aPadInRow = module.getPadCenter(module.getPadsInRow(row)[0]); 
    gear::Vector2D offset = module.getOffset(); 

    //Account for this offset
    aPadInRow[0] -= offset[0]; 
    aPadInRow[1] -= offset[1]; 

    double rR2 = aPadInRow[0]*aPadInRow[0] + aPadInRow[1]*aPadInRow[1]; //Coordinates are global, so they are Cartesian, by convention.
    double rR = sqrt(rR2); //Coordinates are global, so they are Cartesian, by convention.
    double rH = fabs(1./(omega));

    TVector2 xCv(offset[0], offset[1]);
    TVector2 x12(xC - offset[0], yC - offset[1]);

    //Now calculate the radii we need.
    double a     =  x12.Mod2();
    double sqa   = TMath::Sqrt(a);
    double radd  = rR + rH;
    double rsub  = rR - rH;
    double arsub = TMath::Abs(rsub);
    if (sqa > radd || sqa <= 0. || sqa < arsub) {
      // no intersection
      return std::pair<double, double>(-9E9, -9E9);
    } else if (sqa == radd) {
      // single intersection
      TVector2 val = xCv + (rR/sqa)*x12;
      return std::pair<double, double>(val.X(), val.Y());
      //return std::pair<double, double>(-9E9, -9E9);
    } else {
      // two intersection
      double d = radd*rsub + a;
      a   = 1./sqa;
      d  *= 0.5*a;
      double dp  = (rR + d)*(rR - d);
      if (dp <= 0.) dp = 0.;
      else          dp = TMath::Sqrt(dp)*a;
      d  *= a;
      TVector2 d1 = xCv + d*x12;
      TVector2 d2(dp*x12.Y(), -dp*x12.X());
      if(module.isInsideModule(d1.X() + d2.X(), d1.Y() + d2.Y())) //Success!
      {
        return std::pair<double, double> (d1.X() + d2.X(), d1.Y() + d2.Y());
      }
      else if(module.isInsideModule(d1.X() - d2.X(), d1.Y() - d2.Y())) //Success!
      {
        return std::pair<double, double> (d1.X() - d2.X(), d1.Y() - d2.Y());
      }
      else
        return std::pair<double, double>(-9E9, -9E9);
    }
  }//End _calculateTrackPosition()

} //End marlin namespace
