#include "SimplePedestalCalculatorProcessor.h"

//C++
//#include <iostream>
//#include <map>
#include <cmath>
//#include <utility>

//LCIO
//#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <IMPL/TrackerRawDataImpl.h>
#include <IMPL/LCCollectionVec.h>
//#include "lcio.h"

#include "IO/LCWriter.h"
//#include "EVENT/LCIO.h"

#include "IMPL/LCEventImpl.h"
//#include "IMPL/LCRunHeaderImpl.h"
//#include "IMPL/LCTOOLS.h"

using namespace lcio;
using namespace marlin;
using namespace tpcconddata;

namespace marlintpc
{

    SimplePedestalCalculatorProcessor aSimplePedestalCalculatorProcessor;

    SimplePedestalCalculatorProcessor::SimplePedestalCalculatorProcessor() :
	Processor("SimplePedestalCalculatorProcessor")
    {
        // modify processor description
        _description = "SimplePedestalCalculatorProcessor: calculate the pedestal (mean and width) from raw data without iteration." ;
        
 	registerInputCollection(  LCIO::TRACKERRAWDATA,
				  "InputCollectionName", 
				  "Name of the input collection (default: TPCRawData)"  ,
				  _inputColName ,
				  std::string("TPCRawData") ) ;
    
	// this intentionally is not defined as registerOutputCollection
	// because it is not written to the Marlin data flow but to a different lcio file
	registerProcessorParameter( "OutputCollectionName" , 
				    "Name of the output collection (default: TPCPedestals)",
				    _outputColName,
				    std::string("TPCPedestals") );

	registerProcessorParameter( "OutputFileName", 
				    std::string("Name of the lcio output file (existing file will")+
				    std::string(" be overwritten). Default: TPCPedestals.slcio"),
				    _outputFileName ,
				    std::string("TPCPedestals.slcio") );
    }

    
    void SimplePedestalCalculatorProcessor::init()
    {
        printParameters();
    }

    
    void SimplePedestalCalculatorProcessor::processRunHeader(lcio::LCRunHeader*  /*run*/)
    {
//        streamlog_out(DEBUG)<<"SimplePedestalCalculatorProcessor::processRunHeader"<<std::endl;
//        run->parameters().setValue( _processorName + "_revision", "$Rev: 1593 $");
//        for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ )
//        {
//            if ( ! i->second->isOptional() || i->second->valueSet() )
//            {
//                run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
//            }
//        }
//
//        _NRun++ ;
    }

    
    void SimplePedestalCalculatorProcessor::processEvent(lcio::LCEvent * evt )
    {
        //read input data

	LCCollectionVec* inputCollection = 0;
	try
	{
	    inputCollection = dynamic_cast<LCCollectionVec*>
		( evt->getCollection( _inputColName ) );
	    // return in case the dynamic cast failed
	    if( inputCollection == 0 ) return;
	}
	catch(DataNotAvailableException)
	{
	    streamlog_out(MESSAGE3) << "No TPCRawData collection "
				    << _inputColName << " in event "
				    << evt->getEventNumber() << std::endl;
	    return;
	}

	// loop all entries and add the ADC values to the pedestalCreator objects
	for (LCCollectionVec::const_iterator rawDataIter = inputCollection->begin();
	     rawDataIter < inputCollection->end(); rawDataIter++)
	{
	    TrackerRawData const * rawData = dynamic_cast<TrackerRawData const *>
		                                         (*rawDataIter);

	    std::pair<int, int> hardwareID( rawData->getCellID0(),
					    rawData->getCellID1() );
	    
	    ShortVec const & adcValuesVector = rawData->getADCValues();

	    for (ShortVec::const_iterator adcValuesIter = adcValuesVector.begin();
		 adcValuesIter < adcValuesVector.end(); adcValuesIter++)
	    {
		// n.b. if the hardwareID key has not been used before
		// the entry is created using the default constructor of the pedestalCreator
		_pedestalCreatorMap[hardwareID].addSample(*adcValuesIter);
	    }
	}
    }

    
    void SimplePedestalCalculatorProcessor::check(lcio::LCEvent *  /*evt*/ )
    {
        // nothing to check here - could be used to fill checkplots in reconstruction processor
    }

    
    void SimplePedestalCalculatorProcessor::end()
    {
	// create sio writer and open the lcio file
	LCWriter* lcWrt = LCFactory::getInstance()->createLCWriter() ;
	lcWrt->open( _outputFileName, LCIO::WRITE_NEW ) ;

	// create a new event and the output collection
	LCEventImpl *pedestalEvent = new LCEventImpl;
	LCCollectionVec *pedestalCollection = new LCCollectionVec (LCIO::LCGENERICOBJECT);

	// fill the pedestals into the collection
	for (PedestalCreatorMap_t::const_iterator pedestalCreatorIter = _pedestalCreatorMap.begin();
	     pedestalCreatorIter != _pedestalCreatorMap.end() ; pedestalCreatorIter++)
	{
	    pedestalCollection->addElement( new Pedestal( pedestalCreatorIter->first, 
							  pedestalCreatorIter->second.getPedestalValue(),
							  pedestalCreatorIter->second.getPedestalWidth()
                                                        ) 
		                          );
	}

	// add the collection to the event
	pedestalEvent->addCollection( pedestalCollection , _outputColName );

	// write the event
	lcWrt->writeEvent( pedestalEvent ) ;

	// delete the event
	delete pedestalEvent;

	// close the file
	lcWrt->close() ;	

	// delete the writer
	delete lcWrt;
    }

    SimplePedestalCalculatorProcessor::pedestalCreator::pedestalCreator() 
	: _nSamples(0), _sumX(0), _sumXSquare(0)
    {}

    size_t SimplePedestalCalculatorProcessor::pedestalCreator::getNSamples() const
    {
	return _nSamples;
    }

    double SimplePedestalCalculatorProcessor::pedestalCreator::getPedestalValue() const
    {
	if (_nSamples == -0)
	{
	    // avoid divide by zero error, return 0
	    return 0;
	}
	else
	{
	    // cast to double before division to perform floating point division
	    return static_cast<double>(_sumX)/_nSamples;
	}
    }
    
    double SimplePedestalCalculatorProcessor::pedestalCreator::getPedestalWidth() const
    {
	
	if (_nSamples == -0)
	{
	    // avoid divide by zero error, return 0
	    return 0;
	}
	else
	{
	    // cast to double before division to perform floating point division
	    double sigma_square = static_cast<double>(_sumXSquare) /_nSamples
		- static_cast<double>(_sumX)/_nSamples * static_cast<double>(_sumX)/_nSamples;

	    if (sigma_square < 0)
	    {
		// this value cannot be negative, except for rounding errors
		// give a warning and return 0
		streamlog_out(WARNING4) << "Negative width of pedestal due to rounding errors"
					<<", returning zero." << std::endl;
		return 0;
	    }
	    else
	    {
		return std::sqrt(sigma_square);
	    }
	}// if _nSamples
    }
    
    void SimplePedestalCalculatorProcessor::pedestalCreator:: addSample(int adcValue)
    {
	_nSamples++;
	_sumX += adcValue;
	_sumXSquare += adcValue*adcValue;
    }
}// namespace marlinTPC
