#ifndef LCTPCViewer_h
#define LCTPCViewer_h 1

#include "marlin/Processor.h"
#include "lcio.h"
#include <string>
#include <vector>
#include <map>
#include "gear/TPCParameters.h"

//MarliNTPC
#include "ChannelMappingListener.h"

//LCCD
#include "ADCChannelMapping.h"
#include "lccd/ConditionsMap.hh"
#include "TPCCondDataExeptions.h"


// -- Gear
#include "gear/GEAR.h"

//---- LCCD ----
typedef lccd::ConditionsMap< tpcconddata::ADCChannelMapping::key_type , tpcconddata::ADCChannelMapping > ChannelMap ;



namespace marlintpc
{
class ChannelPosMap ;

  /** Simple event display that visualizes data, simulation and gear geometry for the LCTPC large prototype with CED.
   * TrackerRawData, TrackerHit, Tracks, MCParticles and SimTrackerHits are supported.
   * All inputs are optional and can be switched off/on
   *
   *  <h4>Input </h4>
   *
   * @param RawDataCollectionName:     Name of the TrackerRawData collection (default: AltroRawData )
   * @param HitCollectionName:         Name of the TrackerHit collection   (optional, default: TPCTrackerHits)
   * @param TrackCollectionName:       Name of the Track collection        (optional, default: TPCTracks)
   * @param MCParticleCollectionName:  Name of the MCParticle collection        (optional, default: MCParticle)
   * @param SimHitCollectionName:      Name of the SimTrackerHit collection        (optional, default: SimTPCTrackerHits)
   * @param ChannelMappingCollection:  Name of the LCCD collection with channel mapping (default: ADCChannelMapping)
   * @param DriftVelocity:             Drift velocity in micron/ns (default: 75.) needed for TrackerRawData
   * @param ColorScaleMaxADC:          ADC value used for the maximum of the color scales (default: 500)
   * @param ColorScheme:               Color scheme - 1: hot , 2 : cold    (default: 1)
   * @param WaitForKeyBoard:           true: display one event and wait for 'return' - false: continuously display events
   * @param DrawGear:                  default true: draw gear geometry
   * @param DrawFullAdcMap:            default false: draws a projection of TrackerRawData with an entry for every mapped channel. This only works if ChannelMappingCollection is given.
   * @param DrawHitsOnTrack:           default true: the hits belonging to a tracks are drawn for the tracks
   *
   *
   *
   *  @author F. Mueller, DESY
   *
   */

  class LCTPCViewer : public marlin::Processor
  {
	  /** Helper struct for running mean of ADC counts on pads */
	  struct PadMeanADC
	  {

		  PadMeanADC() : Total(0.), N(0) {}

		  float Total ;
		  int N ;

		  void operator+=(float adc)
	        		  {
			  Total += adc ;
			  //      ++N ;
	        		  }
	  };

	  typedef std::map< std::pair<int, int>,  PadMeanADC > ADCMap ;

    public:

      virtual Processor*  newProcessor()
      {
        return new LCTPCViewer ;
      }


      LCTPCViewer() ;

      /** Called at the begin of the job before anything is read.
       * Use to initialize the processor, e.g. book histograms.
       */
      virtual void init() ;

      /** Called for every run.
       */
      virtual void processRunHeader(lcio::LCRunHeader* run) ;

      /** Called for every event - the working horse.
       */
      virtual void processEvent(lcio::LCEvent * evt) ;

      bool getPadCenter(double * padCenter, int cellID0, int cellID1);

      virtual void check(lcio::LCEvent * evt) ;


      /** Called after data processing for clean up.
       */
      virtual void end() ;


    protected:

      static const int ncol = 20 ;
      static const int nscheme = 10 ;

      static const int Red         =   0 ;
      static const int Orange      =   1 ;
      static const int Plum        =   2 ;
      static const int Violet      =   3 ;
      static const int Blue        =   4 ;
      static const int LightBlue   =   5 ;
      static const int Aquamarine  =   6 ;
      static const int Green       =   7 ;
      static const int Olive       =   8 ;
      static const int Yellow      =   9 ;

      static const int Dark        =   10 ;
      static const int Light       =   11 ;
      static const int Classic     =   12 ;
      /** Input collection name.
       */
      std::string _mcParticleColName ;
      std::string _rawColName;
      std::string _hitColName;
      std::string _trkColName;
      std::string _simHitColName;
      std::string _chMapCollection;

      const gear::TPCParameters* _tpcParams  ;
      ChannelMap* _chMap ;
      int _colorScaleMaxADC ;
      int _colorScheme ;
      bool _waitForKeyboard ;
      bool _drawGearGeo;
      std::vector<int> _colors ;
      bool _drawHitsOnTrack;
      bool _drawFullAdcMap;
      float _driftVelocity;
      ChannelMappingListener* _channelMappingListener;
      int _nRun ;
      int _nEvt ;

      ADCMap _adcMap ;
      ADCMap _adcMapInt ;
  } ;
}
#endif



