#ifndef TPCTrackViewer_h
#define TPCTrackViewer_h 1

#include "marlin/Processor.h"
#include "lcio.h"
#include <string>
#include <vector>
#include <map>


// -- Gear
#include "gear/GEAR.h"
#include "gear/TPCParameters.h"



namespace marlintpc
{

  class ChannelPosMap ;

  /** Simple event display that visualizes TrackerRawData for the LCTPC large prototype with CED.
   *  Draws all ADC channels with a color code on layer 1. Optionally, hits and tracks are displayed
   *  on layers 2 and 3. On Layer 4 a color coded map of the mean number of integrated ADC counts
   *  per pad is drawn with a color code.
   *
   *  <h4>Input </h4>
   *  A collection of TrackerRawData (and optionally collections of TrackerHits and Tracks).
   *
   *
   * @param CollectionName:            Name of the TrackerRawData collection (default: AltroRawData )
   * @param ChannelMappingCollection:  Name of the LCCD collection with channel mapping (default: ADCChannelMapping)
   * @param ChannelPositionTextFile:   Optionally use a text file for the hardware channel to position mapping - overwrites mapping from LCCD and GEAR
   * @param DriftVelocity:             Drift velocity in micron/ns (default: 80.)
   * @param ColorScaleMaxADC:          ADC value used for the maximum of the color scales (default: 32)
   * @param WaitForKeyBoard:           true: display one event and wait for 'return' - false: continiously display events
   * @param ColorScheme:               Color scheme - 1: hot , 2 : cold    (default: 1)
   * @param HitCollectionName:         Name of the TrackerHit collection   (optional, default: TPCTrackerHits)
   * @param TrackCollectionName:       Name of the Track collection        (optional, default: TPCTracks)
   *
   *
   *
   *  @author F. Gaede, DESY; modifications by Ch. Rosemann, DESY
   *  @version $Id: TPCTrackViewer.h 3394 2012-05-03 15:29:53Z gaede $
   */

  class TPCTrackViewer : public marlin::Processor
  {

    public:

      virtual Processor*  newProcessor()
      {
        return new TPCTrackViewer ;
      }


      TPCTrackViewer() ;

      /** Called at the begin of the job before anything is read.
       * Use to initialize the processor, e.g. book histograms.
       */
      virtual void init() ;

      /** Called for every run.
       */
      virtual void processRunHeader(lcio::LCRunHeader* run) ;

      /** Called for every event - the working horse.
       */
      virtual void processEvent(lcio::LCEvent * evt) ;


      virtual void check(lcio::LCEvent * evt) ;


      /** Called after data processing for clean up.
       */
      virtual void end() ;


    protected:

      static const int ncol = 20 ;
      static const int nscheme = 10 ;

      static const int Red         =   0 ;
      static const int Orange      =   1 ;
      static const int Plum        =   2 ;
      static const int Violet      =   3 ;
      static const int Blue        =   4 ;
      static const int LightBlue   =   5 ;
      static const int Aquamarine  =   6 ;
      static const int Green       =   7 ;
      static const int Olive       =   8 ;
      static const int Yellow      =   9 ;

      static const int Dark        =   10 ;
      static const int Light       =   11 ;
      static const int Classic     =   12 ;
      /** Input collection name.
       */
      std::string _trkColName ;
      std::string _hitColName ;

      const gear::TPCParameters* _tpcParams  ;

      float _driftVelocity ;
      int _colorScaleMaxADC ;
      int _colorScheme ;
      bool _waitForKeyboard ;

      std::vector<int> _colors ;

      int _nRun ;
      int _nEvt ;
  } ;
}
#endif



