
#ifndef ALTROADCConvolutedPulse_H_
#define ALTROADCConvolutedPulse_H_
#include "ADCPulse.h"
#include "TF1.h"

namespace marlintpc
{

/** Implementation of the ALTRO electronics shaping and ADC.
  *  the ADC and its shaping amplifier.
  *
  *  The shaping is implemented with a Gamma4 function. The height is given by the number of
  *  electrons.
  *
  *  Charge arriving at different times can be added to one pulse. However, there are two
  *  limitations to get proper results:
  *  \li The (start) time of the new charge must not be later than the end of the charge already
  *      stored in the pulse. In this case the pulses would in reality be separated. The
  *      calling code has to check this, otherwise you will get an exception.
  *  \li The (start) time of the new charge must not be earlier than the start time of the
  *      existing pulse. Although the pulses might overlap in reality there is the problem that
  *      the data vector would have to be reorganised. The calling code can simply avoid this
  *      by sorting the charges by time before adding them. Like this it is also ensured
  *      that the overlap detection is working (correct end time of all charges that have arrived before).
  *
  *  For convenience this class also holds the channel number.
  */

class ALTROADCConvolutedPulse: public ADCPulse
{

    public:
        /// The constructor.
        /// @param frequency ADC readout frequency in MHz
        /// @param peakingTime Peaking time ALTRO electronics in ns (time from start of pulse to maximum)
        /// @param nElectronsDynamicRange The number of electrons which correspond to the dynamic range
        ///                               of the ADC (ADCMaxValue)
        /// @param ADCMaxValue The dynamic range of the ADC in ADC counts
        /// @param threshold Minimum number of ADC counts for a sample to be accepted.<br>
        ///        <B>Attention:</B> A threshold of 0 turns off part of the zero suppression. 1 is the smallest
        ///        reasonable value.<br> The parameter intentionally is an int. It will be converted to float
        ///        internally.

        ALTROADCConvolutedPulse(double frequencyMHz, double peakingTime, double nElectronsDynamicRange, int ADCMaxValue,
                      int threshold);

        void add(double startTime, int nElectrons); ///< Add signal from charge (nElectrons) arriving
        ///< at \c startTime to the pulse
        void convoluteSignal();

    protected:
        double _gamma4(double x); ///<gamma4 shaping function
        double _g4Par1;       ///<shaping function order
        double _g4expPar1;    ///<helper variable to avoid recalculation: exp(shaping function order)
        double _peakingTime; ///<Peaking time ALTRO electronics in ns(time from start of pulse to maximum)
};

}


#endif /* ALTROADCConvolutedPulse_H_ */
