
#ifndef GAUSSIANADCPULSE_H_
#define GAUSSIANADCPULSE_H_
#include "ADCPulse.h"
namespace marlintpc
{

/** A simple Gaussian shaping.
  *
  *  The shaping is implemented as a Gaussian smearing, the left edge
  *  (-2\f$\sigma\f$ from the centre of the Gaussian) is the time when the
  *  charge arrives. Afterwards the number of electrons arriving in the time
  *  bins is calculated by integrating the Gaussian within this time bin.
  *  An asymmetry can be assigned to the Gaussian to account for slower falling tails
  *
  *  Charge arriving at different times can be added to one pulse. However, there are two
  *  limitations to get proper results:
  *  \li The (start) time of the new charge must not be later than the end of the charge already
  *      stored in the pulse. In this case the pulses would in reality be separated. The
  *      calling code has to check this, otherwise you will get an exception.
  *  \li The (start) time of the new charge must not be earlier than the start time of the
  *      existing pulse. Although the pulses might overlap in reality there is the problem that
  *      the data vector would have to be reorganised. The calling code can simply avoid this
  *      by sorting the charges by time before adding them. Like this it is also ensured
  *      that the overlap detection is working (correct end time of all charges that have arrived before).
  *
  *  For convenience this class also holds the channel number.
  */

class GaussianADCPulse: public ADCPulse
{

    public:
        /// The constructor.
        /// @param frequency ADC readout frequency in MHz
        /// @param riseTime The rise time of the pulse in ns
        /// @param asymmetry between width left and right of maximum: w_left=asym*w_right
        /// @param nElectronsDynamicRange The number of electrons which correspond to the dynamic range
        ///                               of the ADC (ADCMaxValue)
        /// @param ADCMaxValue The dynamic range of the ADC in ADC counts
        /// @param threshold Minimum number auf ADC counts for a sample to be accepted.<br>
        ///        <B>Attention:</B> A threshold of 0 turns off part of the zero suppression. 1 is the smallest
        ///        reasonable value.<br> The parameter intentionally is an int. It will be converted to float
        ///        internally.

        GaussianADCPulse(double frequencyMHz, double riseTime, double asymmetry, double nElectronsDynamicRange, int ADCMaxValue,
                         int threshold);

        void add(double startTime, int nElectrons); ///< Add signal from charge (nElectrons) arriving
        ///< at \c startTime to the pulse


    protected:

        double _riseTime; ///< Rise time  of the pulse
        double _asym;//asymmetry of pulse shape: width left = _asym*width right
        double _sqrt2_sigma; ///< \f$\sqrt{2}\sigma\f$ is often needed during calculation, where
        ///< \f$\sigma\f$ = riseTime / 3 is the width of the Gaussian
};

}


#endif /* GaussianADCPULSE_H_ */
