#ifndef IonBackDriftProcessor_h
#define IonBackDriftProcessor_h 1

#include <cstdlib>

//MARLIN
#include "marlin/Processor.h"

//LCIO
#include "lcio.h"

//MARLINTPC
#include "VoxelTPC.h"

namespace marlintpc{

/* The following comment is tranfered to doxygen documentation
 * which begins with a second asterix, or in case of on line comment an extra third slash 
 */

/** This processor takes the charges on the pads an calculates 
 *  the amount of backdrifting ions which appear at the x,y-
 *  position of the pad in the TPC.
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection with TPCVoxels
 *
 *  <h4>Output</h4> 
 *  Collection with TPCVoxels
 *
 * @param InputCollectionName Name of input collection, TPCVoxel
 * @param OutputCollectionName Name of output collection, TPCVoxel
 * @param Transient If not 0 the output collection is set transient
 * @param OutputType If not 0 all backdrifting ions in the TPC are returned
 * @param ZBinLength Length of the bins in z direction
 * @param DriftField Strength of the drift field in V/cm
 * @param ETrans1 Strength of the first transfer field in V/cm
 * @param ETrans2 Strength of the second transfer field in V/cm
 * @param EInd Strenght of the induction field in V/cm
 * @param UGEM1 Voltage on the first GEM in V
 * @param UGEM2 Voltage on the second GEM in V
 * @param UGEM3 Voltage on the thirst GEM in V
 * @param ParamFile Name of file with GEM parameters
 *
 * @author Thorsten Krautscheid
 */

class IonBackDriftProcessor : public marlin::Processor {
  
 public:
  
  virtual Processor*  newProcessor() { return new IonBackDriftProcessor;}

  IonBackDriftProcessor();
  virtual  ~IonBackDriftProcessor(){}

  /** Called at the begin of the job before anything is read.
   * Use to initialize the processor, e.g. book histograms.
   */
  virtual void init() ;
  
  /** Called for every run.
   */
  virtual void processRunHeader(lcio::LCRunHeader* run ) ;
  
  /** Called for every event - the working horse.
   */
  virtual void processEvent(lcio::LCEvent * evt ) ; 
  

  void LoadGEMSettings();



protected: 

  // in-/output collection names
  std::string _inputCollectionName;
  std::string _outputCollectionName;

  //steering parameters
  std::string _paramFile;///< Name of the file with parametrisation of the GEM-stack
  int _transient;
  int _outputType;///< Output type: event based or total

  //VoxelTPC for built up of ion disc
  marlintpc::VoxelTPC* _ionDisc;

  //gear pad layout
  gear::PadRowLayout2D const* _padLayout;

  double _zBinLength;///< Length of the bins in z direction

  int _nRun ;
  int _nEvt ;

  //Ionbackdrift parameter.
  double _ionBackDrift;

  double _E,_ETrans1,_ETrans2,_EInd;///<Drift field, 1st, 2nd and 3th transfer field

  double _UGEM1,_UGEM2,_UGEM3;///<GEM voltages 
};

} //end of namespace marlintpc

#endif 
