#ifndef TPCElectronicsProcessor_h
#define TPCElectronicsProcessor_h 1

//MARLIN
#include <marlin/Processor.h>

//LCIO
#include <lcio.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/TrackerRawData.h>

//C++
#include <string>
#include <list>

//GEAR
#include <gear/PadRowLayout2D.h>

//MARLINTPC
#include "VoxelTPC.h"
#include "ADCPulse.h"

namespace marlintpc
{

/** This processor applies the electronics shaping to the incoming electron signals.
 *  The signal shape is a gaussian which is cut off at +- 3 \f$\sigma\f$.
 *  The width of the gaussian is defined as \f$\sigma\f$ = rise time / 3.
 *  The electronics amplification is calculated from the number of of electrons and
 *  the number of ADC counts which correspond to the dynamic range of the ADC.
 *
 *  In order to use the ImproveMemoryFootprint option this processor has to be
 *  run together with (after) the ChargeDistributionProcessor in the same Marlin run.
 *  The ChargeDistributionProcessor also needs the ImproveMemoryFootprint option
 *  to be turned on. MokkaToVoxelProcessor does not support this option yet.
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection of TPCVoxel with charges on pads
 *
 *  <h4>Output</h4>
 *  Collection with TPCRawData
 *
 * @param InputCollectionName Name of input collection, (default: TPCVoxel)
 * @param OutputCollectionName Name of output collection, (default: TrackerRawData)
 * @param WriteOutputToStorage Output collection will be written if set to true (default: true)
 * @param ReadoutFrequency The frequency of readout electronics in MHz (default: 40.).
 * @param RiseTime Signal rise time in ns (default: 60.).
 * @param Asymmetry Signal asymmetry around center of gaussian shape (default: 1.).
 * @param ADCMaxValue Dynamic range of the ADC in ADC counts (default: 1023)
 * @param NElectronsDynamicRange Dynamic range of the ADC in electrons (default: 1e6)
 * @param VoxelZLength Length of a voxel in ns (default: 1.);
 * @param Threshold Minimum number auf ADC counts for a sample to be accepted
 *        (default: 1, just zero suppression)<br>
 *        <B>Attention:</B> A threshold of 0 turns off part of the zero suppression. 1 is the smallest
 *        resonable value.
 * @param ElectronicsSwitch Switch between different electronics: 0=Gaussian shaping, 1=ALTRO electronics, 2=ALTRO convolution (does not work with ClockJitter) (default: 0)
 * @param NoiseSwitch Switch between different noise model for pedestal width distribution (0=no noise, 1=gaussian, 2=landau)
 * @param ClockJitter Clock jitter for the digitisation in nanoseconds  (cannot be used in case of ElectronicsSwitch == 2) (default: 0)
 * @param NoiseLevel Central value of the pedestal width used to add noise to the ADC spectrum (mean for the gaussian noise width distribution, most probable value for the landau (which is, by the way, (almost but) not exactly the parameter root calls "MPV" if you fit your distribution with a Landau, see e.g. the TMath documentation...)
 * @param NoiseWidth Width of the pedestal width distribution used to add noise to the ADC spectrum
 * @param NoiseCutoff Cutoff for random pedestal width - e.g. to discard tails from the Landau distribution, since really noisy channels probably wouldn't be used in the data either. (default: 4)

 * @param MinimumPulseHeight Minimal value of the highest sample in a pulse (default: 1)
 * @param ImproveMemoryFootprint Flag whether to use ChargeDistributionProcessor::fetchVoxelCollection
 *        instead of the input collection (default:false)
 *
 * @author Astrid Muennich, RWTH Aachen
 * @author Thorsten Krautscheid, University of Bonn
 * @author Martin Killenberg, CERN
 */


class TPCElectronicsProcessor : public marlin::Processor
{

    public:

        virtual Processor*  newProcessor()
        {
            return new TPCElectronicsProcessor ;
        }

        TPCElectronicsProcessor() ;
        virtual ~TPCElectronicsProcessor();

        /* Instantiate the ADCPulse with the parameters from the steering file.
         */
        virtual void init() ;

        /** Called for every run.
         */
        virtual void processRunHeader(lcio::LCRunHeader* run) ;

        /** Called for every event - the working horse.
         */
        virtual void processEvent(lcio::LCEvent* evt) ;




    protected:

        /** function to sort the list of std::pair< VoxelTPC::VoxelIndex, int > by pad index, then time
         */
        static bool voxelPadComp(const std::pair< VoxelTPC::VoxelIndex, int >& left,
                                 const std::pair< VoxelTPC::VoxelIndex , int >& right)
        {
            return(left.first.moduleID == right.first.moduleID ?
                   (left.first.padIndex == right.first.padIndex  ?
                    left.first.zBin < right.first.zBin :
                    left.first.padIndex  < right.first.padIndex) :
                   left.first.moduleID < right.first.moduleID);
        }


        /** function to sort the list of std::pair< VoxelTPC::VoxelIndex, int > by time, then pad index
         */
        static bool voxelTimeComp(const std::pair< VoxelTPC::VoxelIndex, int >& left,
                                  const std::pair< VoxelTPC::VoxelIndex , int >& right)
        {
            return(left.first.zBin == right.first.zBin ?
                   (left.first.moduleID == right.first.moduleID  ?
                    left.first.padIndex < right.first.padIndex :
                    left.first.moduleID  < right.first.moduleID) :
                   left.first.zBin < right.first.zBin);
        }

        bool _outputIsPersistent;

        //Input-/output collection names
        std::string _inputColName;///<Name of input collection
        std::string _outputColName;///<Name of output collection

        double _frequency; ///< Readout frequency of the ADC
        double _riseTime; ///< Rise time of the signal
        double _asym;//asymmetry of pulse shape: width left = _asym*width right
        int  _ADCMaxValue; ///< The maximum of the dynamic range in ADC counts
        int _nElectronsDynamicRange;  ///< The number of electrons which correspond
        ///< to the dynamic range of the ADC
        double _voxelZLength; ///<  Length of a voxel in ns
        int _ADCThreshold; ///< The threshold of the ADC (in ADC counts)
        int _minimumPulseHeight; ///< The minimal value for the highest sample value in a pulse

        int   _electronicsSwitch; ///<Switch between different electronics implementations:
        ///<0=Gaussian shaping, 1=ALTRO electronics
        double _peakingTime; ///<Peaking time ALTRO electronics (time from start of pulse to maximum)
        // to be implemented later
        //  int _ADCThreshold; // /<Threshold of the ADC
        //  double _cutOff; // /<Cut off value for the integration

        int _noiseSwitch; ///<Switch between different noise model for pedestal width distribution:
        ///< 0=no noise, 1=gaussian, 2=landau

        double _clockJitter; ///< Time jitter for the digitisation in nanoseconds
        std::map<int, double> _timeJitter; // map of time (zBin) and the clock jitter for that bin

        double _noiseLevel; ///< Central value of the pedestal width used to add noise to the ADC spectrum

        double _noiseWidth; ///< Width of the pedestal width distribution used to add noise to the ADC spectrum

        double _noiseCutoff; ///< Cutoff for random pedestal width - e.g. to discard tails from the Landau distribution, since really noisy channels probably wouldn't be used in the data either. (default: 4)

        /// The instance of the ADCPulse, as pointer since it can only be initialised in the init()
        ADCPulse* pulse;

        bool _improveMemoryFootprint; ///< Flag whether to use the memory saving shortcut.

        double _integrationTime;///<IntegrationTime of the electronics. Voxel within this time window will be shaped in one go.
};



} //end of namespace marlintpc

#endif
