#include <algorithm>
//#include <vector>
//#include <string>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <Exceptions.h>
#include <streamlog/streamlog.h>
#include <lcio.h>

#include <IMPL/TrackerRawDataImpl.h>
#include "ADCPulse.h"


namespace marlintpc
{

ADCPulse::ADCPulse(double frequencyMHz, int ADCMaxValue, int threshold)
    :  _frequencyGHz(frequencyMHz / 1000.), // internally use frequency in GHz (avoid many multiplications with 1000)
       _ADCMaxValue(ADCMaxValue), _startBin(-1),
       _channelNumber(-1), _moduleID(-1), _ADCThreshold(threshold)
{}

double ADCPulse::getEndTime() const
{
    return (_startBin + _rawData.size() + 1) / _frequencyGHz ;
}

int ADCPulse::getModuleID() const
{
    return _moduleID;
}

int ADCPulse::getChannelNumber() const
{
    return _channelNumber;
}

void ADCPulse::clear(int newChannelNumber, int modID, int noiseSwitch, double noiseLevel, double noiseWidth, double noiseCutoff)
{
    _rawData.clear();
    _channelNumber = newChannelNumber;
    _moduleID = modID;
    _startBin = -1;

    const double mpvshift = -0.22278298;
    streamlog_out(DEBUG1) << "noise switch is set to " << noiseSwitch << std::endl;
    switch(noiseSwitch)
    {
        case 0:
            _addNoise = false;
            break;
        case 1:
            _addNoise = true;
            do
            {
                _padNoise = CLHEP::RandGauss::shoot(noiseLevel, noiseWidth);
            }
            while(_padNoise > noiseCutoff || _padNoise < 0); // keep shooting randoms until the noise is in the accepted range
            break;
        case 2:
            _addNoise = true;
            do
            {
                _padNoise = CLHEP::RandLandau::shoot() * noiseWidth + noiseLevel - mpvshift * noiseWidth ;
            }
            while(_padNoise > noiseCutoff || _padNoise < 0);
            /* random landau is not in terms of MPV and width! --> need to
            - scale with sigma
            - shift the mean from zero to npoiseLevel
            - apply MPV correction factor (mean<->MPV) for the approximation
            (cernlib used approximantion based on Kolbig & Schorr Comp Phys Comm 31 p. 97 (1984))
            */
            break;
        default:
            streamlog_out(ERROR) << "Unknown noise type selected!" << std::endl;
            throw lcio::Exception("ADCPulse: Internal error: Unknown noise type selected!");
    }
    if(true == _addNoise)
    {
        streamlog_out(DEBUG1) << "ADCPulse: channel numner " <<  _channelNumber << " , noise " << _padNoise << std::endl;
    }
}



void ADCPulse::setJitterValues(std::map<int, double> timeJitter)
{
    _timeJitter = timeJitter;
    _useJitter = true;
    streamlog_out(DEBUG3) << "size of jitter map " << _timeJitter.size() << std::endl;
}



const std::vector<float>& ADCPulse::getRawDataVec() const
{
    return _rawData;
}



std::list<EVENT::TrackerRawData*> ADCPulse::getLCIORawData() const
{
    std::list<EVENT::TrackerRawData*> returnList;

    if(_rawData.size() == 0)
    {
        streamlog_out(DEBUG3) << "Empty undigitised raw data on channel "
                              << _channelNumber << std::endl;
        return returnList;
    }

    std::vector<float>::const_iterator StartOfPulse = _rawData.begin();

    // flag for debugging
    bool pulseWasClipped = false;

    // the outer loop runs until no more pulses are found
    while(StartOfPulse < _rawData.end())
    {
        // search the beginning of the pulse, i. e. the first adc count above the threshold
        while((StartOfPulse < _rawData.end()) && (*StartOfPulse < (_ADCThreshold - 0.5)))
        {
            streamlog_out(DEBUG1) << "startsignal " << std::scientific << *StartOfPulse << std::endl;

            StartOfPulse++;
            //       lcioStartBin++;

            // streamlog_out(DEBUG1) << " ; startBin now is "<< lcioStartBin  << std::endl;
            pulseWasClipped = true;
        }

        // in case all samples are below theshold the iterator is at the end
        // There is no further pulse ans StartOfPulse would be invalid.
        // exit the loop here
        if(StartOfPulse == _rawData.end())
        {
            break;
        }

        // search the end of the pulse (not included)
        std::vector<float>::const_iterator EndOfPulse = StartOfPulse;
        EndOfPulse++;

        // scan for the first value below the threshold
        while((EndOfPulse < _rawData.end()) && (*EndOfPulse >= (_ADCThreshold - 0.5)))
        {
            streamlog_out(DEBUG0) << "in pulse " << std::scientific << *EndOfPulse << std::endl;

            EndOfPulse++;
        }

        // fill the lcio object
        std::vector<short> shortRawDataVec(std::distance(StartOfPulse, EndOfPulse));
        for(std::vector<float>::const_iterator binIter = StartOfPulse;
                binIter < EndOfPulse; binIter++)
        {
            if(*binIter < _ADCMaxValue)
            {
                if(true == _addNoise)
                {
                    // add random noise from gaussian distribution centred on zero, with given width for this channel
                    shortRawDataVec[ std::distance(StartOfPulse, binIter) ] = static_cast<short>(round(*binIter + CLHEP::RandGauss::shoot(0.0, _padNoise)));
                    streamlog_out(DEBUG2) << "ADCPulse would have been " << static_cast<short>(round(*binIter))
                                          << ", with added noise from distribution with width " << _padNoise
                                          << " it is now " << shortRawDataVec[ std::distance(StartOfPulse, binIter) ]  << std::endl;
                }
                else
                {
                    shortRawDataVec[ std::distance(StartOfPulse, binIter) ] = static_cast<short>(round(*binIter));
                }

                if(shortRawDataVec[ std::distance(StartOfPulse, binIter) ] == 0)
                {
                    streamlog_out(DEBUG2) << "zero from original data " << *binIter << std::endl;
                }
            }
            else
            {
                streamlog_out(DEBUG3) << "pulse on channel " << _channelNumber
                                      << " out of range (" << *binIter << "/" << _ADCMaxValue
                                      << ") in bin " << std::distance(StartOfPulse, binIter)
                                      << std::endl;
                shortRawDataVec[ std::distance(StartOfPulse, binIter) ] =  _ADCMaxValue;
            }
        }

        if(EndOfPulse < _rawData.end())
        {
            pulseWasClipped = true;
        }

        if(pulseWasClipped)
        {
            streamlog_out(DEBUG2) <<  "Pulse on channel " << _channelNumber
                                  << " was clipped. Raw data now is ";
            for(std::vector<short>::iterator rawDataIter = shortRawDataVec.begin();
                    rawDataIter < shortRawDataVec.end(); rawDataIter++)
            {
                streamlog_out(DEBUG2) <<  *rawDataIter  << " ";
            }
            streamlog_out(DEBUG2) <<  std::endl;
        }

        IMPL::TrackerRawDataImpl* lcioRawData = new IMPL::TrackerRawDataImpl;
        lcioRawData->setADCValues(shortRawDataVec);
        lcioRawData->setCellID0(_channelNumber);
        lcioRawData->setCellID1(_moduleID);
        lcioRawData->setTime(_startBin + std::distance(_rawData.begin(), StartOfPulse));

        returnList.push_back(lcioRawData);

        StartOfPulse = EndOfPulse;

    }//while ( StartOfPulse < _rawData.end() )

    return returnList;
}

}// namespace marlintpc
