#include <algorithm>
//#include <vector>
//#include <string>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <Exceptions.h>
#include <streamlog/streamlog.h>
#include <lcio.h>

#include <IMPL/TrackerRawDataImpl.h>
#include "ALTROADCPulse.h"


namespace marlintpc
{

ALTROADCPulse::ALTROADCPulse(double frequencyMHz, double peakingTime, double nElectronsDynamicRange, int ADCMaxValue, int threshold)
    :  ADCPulse(frequencyMHz, ADCMaxValue, threshold), _peakingTime(peakingTime)
{

    _ADCChargeFactor = _ADCMaxValue / nElectronsDynamicRange ;
    _g4Par1 = 4.0;
    _g4expPar1 = exp(_g4Par1);

    _useJitter = false;
}



// implementation as a simple function is much faster than TF1::Eval
double ALTROADCPulse::_gamma4(double x, double ADCChargeValue, double startTime)
{
    // parameters of the gamma4 function:
    // charge, order, start time, peaking time
    return ((ADCChargeValue * _g4expPar1 * pow(((x - startTime) / _peakingTime), _g4Par1) * exp(-_g4Par1 * ((x - startTime) / _peakingTime))) * (x > startTime));
}



void ALTROADCPulse::add(double startTime, int nElectrons)
{
    // Check that the pulses really overlap
    if((startTime > getEndTime()) && (_startBin >= 0))
    {
        //    throw Exception("ADCPulse: ADCPulses do not overlap. Check the calling code!") ;
        throw lcio::Exception("ADCPulse: ADCPulses do not overlap. Check the calling code!") ;
    }

    // Check that the start time of the new charge is not before the start time of the existing pulse.
    // This cannot be handles by this algorithm, the calling code has to take care of this.
    if(startTime < _startBin / _frequencyGHz)
    {
        throw lcio::Exception("ALTROADCPulse: Error: Start time of new charge is before the start of the existing pulse. Check the calling code!") ;
    }

    double ADCChargeValue = nElectrons * _ADCChargeFactor;

    //shaping function starts at t0=startTime.

    int bin = static_cast<int>(startTime * _frequencyGHz); // bin number for the first bin with charge
    if(_startBin < 0) // if there is no charge in the pulse yet the pulse starts at "bin"
    {
        _startBin = bin;
    }
    streamlog_out(DEBUG2) << "GAMMA4: st " << startTime << ", sb " << _startBin << ", pt " << _peakingTime << ", ne " << nElectrons << ", adc " << ADCChargeValue << ", f" << _frequencyGHz << std::endl;
    double binTime = (_startBin + 0.5) / _frequencyGHz;

    std::map<int, double>::iterator it;
    double   ADCValue ;
    do
    {
        // calculate the ADC value for this bin by evaluating gamma4 function in bin center.
        if(_useJitter)
        {
            it = _timeJitter.find(bin);
            if(it == _timeJitter.end())
                throw lcio::Exception("ALTROADCPulse: Error: Jitter map does not contain a value for this time bin!") ;
            else
	      ADCValue = _gamma4(binTime + it->second, ADCChargeValue, startTime);
            streamlog_out(DEBUG1) << "bin " << bin << ", binTime " << binTime << ", ADCValue " << ADCValue << " using jitter of " << it->second << " (jitter map size " << _timeJitter.size() << ")" << std::endl;
        }
        else
        {
	    ADCValue = _gamma4(binTime, ADCChargeValue, startTime);
        }
        // debug: check for 0
        if(ADCValue == 0 || ADCValue < 0)
        {
            streamlog_out(DEBUG2) << "evaluation of gamma4 function at " << binTime
                                  << " resulted in zero" << std::endl;
        }

        // check if the bin already exists in the vector, else append it
        if((bin - _startBin) < static_cast<int>(_rawData.size()))
        {
            // time bin is already there, add the charge
            _rawData[bin - _startBin] += ADCValue;
            streamlog_out(DEBUG1) << "ADCValue " << ADCValue << ", _rawData for bin " << bin - _startBin << " is now " << _rawData[bin - _startBin] << std::endl;

        }
        else // the time bin did not exist yet
        {
            _rawData.push_back(ADCValue);
        }


        // move to the next time bin
        bin++;
        binTime = (bin + 0.5) / _frequencyGHz;


    }
    while(binTime < startTime + 4 * _peakingTime);

}



}// namespace marlintpc
