#include <HitDistributor.h>
#include <PadRowLayout2D.h>

//C++
#include <cmath>
#include <complex>

//MARLIN
#include <streamlog/streamlog.h>

using namespace lcio;

namespace marlintpc{

  //Constructor
  HitDistributor::HitDistributor(const gear::TPCParameters* tpcParameters, 
                                 double zBinLength, double inductionFactor) {
    
    _tpcParameters = tpcParameters;
    _zBinLength    = zBinLength;
    _tpcCoordinateType = _tpcParameters->getCoordinateType();
    //change from percent to ratio
    _inductionFactor = inductionFactor/100.;
  }

  HitDistributor::~HitDistributor() {}
  
  // Steering function. The only function accessible from the outside.
  // Takes a hit for input and calls all other functions for the smearing
  // depending on the pad layout of the readout plane.
  std::list<TPCVoxel> HitDistributor::distributeHit(SimTrackerHit* hit, 
						    double sigmaTrans, 
						    double sigmaLong)
  {
    //Hit(coordinate of electron) values in cartesian coordinates
    double xHit = hit->getPosition()[0];
    double yHit = hit->getPosition()[1];
    
    //Check what coordinate system the TPC should use
    //if polar we need to transform the incoming lcio hit from cartesian to polar
    if( _tpcCoordinateType== gear::PadRowLayout2D::POLAR)
      {
	double rHit   = sqrt(xHit*xHit+yHit*yHit);
	double phiHit = atan2(yHit,xHit);
	_c0Hit=rHit;
	_c1Hit=phiHit;
      }
    else
      {
	_c0Hit=xHit;
	_c1Hit=yHit;
      }

    //Distribute charge of the hit according to the pad layout type (polar or rectangular geometry)
    _padLayoutType= _tpcParameters->getNearestModule(_c0Hit,_c1Hit).getLocalPadLayout().getCoordinateType();
  
    std::list<GlobalPadIndex> padIndexList = getPads( sigmaTrans );
    streamlog_out(DEBUG4) << "padIndexList size: " << padIndexList.size() << std::endl;
  
    if( padIndexList.empty() )
      { 
	streamlog_out(DEBUG4) << "list empty!!!" << std::endl;
	return std::list<TPCVoxel>();
      }
    
    switch( _padLayoutType )
      {
      case gear::PadRowLayout2D::POLAR: 
	{
	  streamlog_out(DEBUG4) << "distributeHit ("<<xHit<<","<<yHit<<"): POLAR" << std::endl;
	  //rows are counted up along r, the first coordinate
	  return distributeHitOnPadsPolar( padIndexList, hit, sigmaTrans, sigmaLong );
	  break;
	}
      case gear::PadRowLayout2D::CARTESIAN:
	{
	  streamlog_out(DEBUG4) << "distributeHit ("<<xHit<<","<<yHit<<"): CARTESIAN" << std::endl;
	  //rows are counted up along y, the second coordinate
	  return distributeHitOnPadsRect( padIndexList, hit, sigmaTrans, sigmaLong );
	  break;
	}
      default:
	throw gear::Exception("Unknown pad layout type");
      };
  
    
    return std::list<TPCVoxel>();    
  }


  // Gets all pads within the smearing radius and stores their indices in list
  std::list<GlobalPadIndex> HitDistributor::getPads( double sigmaTrans )
  {
    std::list<GlobalPadIndex> padIndexList;
    GlobalPadIndex hitIndex(-1,-1);
    const gear::TPCModule & tpcModule = _tpcParameters->getNearestModule(  _c0Hit , _c1Hit );
    try{
      gear::GlobalPadIndex ghitIndex = _tpcParameters->getNearestPad( _c0Hit , _c1Hit );
      hitIndex.setPadIndex(ghitIndex.getPadIndex());
      hitIndex.setModuleID(ghitIndex.getModuleID());
    }
    catch(std::out_of_range& e){
      streamlog_out(DEBUG4) << "getNearestPad for hit "<<_c0Hit <<" "<< _c1Hit<<" failed"<<std::endl;
      return padIndexList;
    }
    
    if( checkDistance( hitIndex, sigmaTrans ) )
      {
	padIndexList.push_back( hitIndex );
	streamlog_out(DEBUG4) <<"Central pad: "<< 
		  tpcModule.getRowNumber(hitIndex.getPadIndex()) <<" "<<
		  tpcModule.getPadNumber(hitIndex.getPadIndex())<<std::endl;
	getPadsOnTheLeft(  padIndexList, hitIndex, sigmaTrans );
	getPadsOnTheRight( padIndexList, hitIndex, sigmaTrans );
      }
    else
      return padIndexList;

    std::list<int> directions;
    directions.push_back(+1);
    directions.push_back(-1);

    for (std::list<int>::iterator deltaRow = directions.begin();
	 deltaRow != directions.end(); ++deltaRow)
      {
	GlobalPadIndex cursor = hitIndex;
	//go back to original row
	int row = tpcModule.getRowNumber(hitIndex.getPadIndex());
	
	// step up a row
	streamlog_out(DEBUG4) <<"Stepping the rows: "<<*deltaRow<< std::endl;
	while(1)
	  {
	    row += *deltaRow;
	    if( row >= 0 && row < tpcModule.getNRows() )
	      {
		try{
		  //direction of rows has to be done in local coordinates
		  const gear::PadRowLayout2D &localPadPlane=tpcModule.getLocalPadLayout();
		  gear::Vector2D local_padCenter=localPadPlane.getPadCenter( cursor.getPadIndex() );
		  if(_padLayoutType==gear::PadRowLayout2D::CARTESIAN)//CARTESIAN: row along y
		    {
		      //x stays constant
		      double x_local=local_padCenter[0];
		      //y moves to next row = half the height of current row plus half the height of next row
		      double y_local= local_padCenter[1]+(*deltaRow)*(tpcModule.getRowHeight(tpcModule.getRowNumber(hitIndex.getPadIndex()))/2.
								   +tpcModule.getRowHeight(row)/2.);
		      cursor.setPadIndex(localPadPlane.getNearestPad(x_local,y_local));
		    }
		  else//POLAR: row along r
		    {
		      //r moves to next row
		      double r_local=local_padCenter[0]+(*deltaRow)*(tpcModule.getRowHeight(tpcModule.getRowNumber(hitIndex.getPadIndex()))/2.
								  +tpcModule.getRowHeight(row)/2.);
		      //phi stays the same
		      double phi_local=local_padCenter[1];
		      cursor.setPadIndex(localPadPlane.getNearestPad(r_local,phi_local));
		    }
		}
		catch(std::out_of_range& e){
		  break;
		}
		if( checkDistance( cursor, sigmaTrans) ) 
		  {
		    padIndexList.push_back( cursor );
		    streamlog_out(DEBUG4) <<"Adding pad (m,r,p): "<< 
		      cursor.getModuleID()<<" "<<
		      tpcModule.getRowNumber(cursor.getPadIndex()) <<" "<<
		      tpcModule.getPadNumber(cursor.getPadIndex())<<std::endl;
		    getPadsOnTheLeft(  padIndexList, cursor, sigmaTrans );
		    getPadsOnTheRight( padIndexList, cursor, sigmaTrans );
		  }
		else
		  break;
	      }
	    else
	      break;
	  }
      }
  
    return padIndexList;
  }

  // Checks whether the distance between the given pad and the hit is lower than the smearing 
  // range. Returns true when any part of the pad is within the smearing radius, even when it 
  // is only a little corner of the pad.
  bool HitDistributor::checkDistance(GlobalPadIndex globalPadIndex, double sigmaTrans)
  {
    //integration of gaussian up to 4 sigma
    static const double SMEAR = 4;

    const gear::TPCModule & tpcModule = _tpcParameters->getModule(globalPadIndex.getModuleID());

    // add 1e-6 mm = 1 nm for rounding errors in case of sigmaTrans = 0
    if ( tpcModule.getDistanceToPad(_c0Hit, _c1Hit, globalPadIndex.getPadIndex()) <= (SMEAR*sigmaTrans + 1e-6) )
    {
      return true;
    }

    return false;
  }

  // Gets the indices of all pads in the row, which are left from the given pad 
  // and wthin the smearing radius, and adds them to the list which contains
  // all pad indices of hit. 
  void HitDistributor::getPadsOnTheLeft(std::list<GlobalPadIndex>& padList, 
					GlobalPadIndex  centerIndex, 
					double sigmaTrans)
  {
    GlobalPadIndex np = centerIndex;
    const gear::TPCModule & tpcModule =  _tpcParameters->getModule(np.getModuleID());
    
    while(1){
      
      int index;
      try
      {
        index = tpcModule.getLeftNeighbour(np.getPadIndex());
      }
      catch(std::exception& e)
	{
	  streamlog_out(DEBUG4) << "left pad is not available..." << std::endl;
	  return;
	}
      
      np.setPadIndex( index );
      if( checkDistance(np, sigmaTrans) && !(np == centerIndex) )
	{
	  padList.push_back(np);  
	  streamlog_out(DEBUG4) <<"Adding pad left (m,r,p): "<< 
	    np.getModuleID()<<" "<<
	    tpcModule.getRowNumber(np.getPadIndex()) <<" "<<
	    tpcModule.getPadNumber(np.getPadIndex())<<std::endl;
	}  
      else 
      {
        streamlog_out(DEBUG4) << "left is not inside range ..." << std::endl;
        return;
      }
    }
  }

  //  Gets the indices of all pads in the row, which are right from the given pad 
  //  and wthin the smearing radius, and adds them to the list which contains
  //  all pad indices of hit. 
  void HitDistributor::getPadsOnTheRight(std::list<GlobalPadIndex>& padList, GlobalPadIndex centerIndex, double sigmaTrans){
    GlobalPadIndex np = centerIndex;

    const gear::TPCModule & tpcModule = _tpcParameters->getModule(np.getModuleID());

    while(1){
      int index = 0;

      try{
        index = tpcModule.getRightNeighbour(np.getPadIndex());  
      }catch(std::exception& e){
	return;
      }

      np.setPadIndex( index );

      if( checkDistance( np, sigmaTrans) )
	{
	  streamlog_out(DEBUG4) <<"Adding pad right (m,r,p): "<< 
	    np.getModuleID()<<" "<<
	    tpcModule.getRowNumber(np.getPadIndex()) <<" "<<
	    tpcModule.getPadNumber(np.getPadIndex())<<std::endl;
	padList.push_back( np );
	}
      else 
	{
	  streamlog_out(DEBUG4) << "right is not inside range ..." << std::endl;
	  return;
	}
    }
  }

  // Smears the distributes the charge of the given hit in voxels according to the given pads, 
  //transversal and longitudinal diffusion for rectangular pad layout
  std::list<TPCVoxel> HitDistributor::distributeHitOnPadsRect( std::list<GlobalPadIndex> padIndexList, 
							   SimTrackerHit* hit, 
							   double sigmaTrans, 
							   double sigmaLong ) {
    std::list<TPCVoxel> voxelList;
    
    //integration should completely happen in local coordinates of the padplane
    //as modules can be translated and/or rotated
    //first get the center of the gaussian into local coordinates:
    const gear::TPCModule & tpcModule = _tpcParameters->getNearestModule(_c0Hit,_c1Hit);
    gear::Vector2D local_hit = tpcModule.globalToLocal(_c0Hit,_c1Hit);

    if( hit->getEDep() > 0 ){
      
      //integration for all the stored pad indices
      while(padIndexList.begin()!=padIndexList.end()) {
        GlobalPadIndex gpadIndex = padIndexList.front();
        const gear::TPCModule & tpcModule = _tpcParameters->getModule(gpadIndex.getModuleID());

	//calculate the width of the pad (with pitch!)
	int row = tpcModule.getRowNumber( gpadIndex.getPadIndex() );
	double halfPadPitch_x = tpcModule.getPadPitch( gpadIndex.getPadIndex() )/2;
	//calculate the height of the pad 
	double halfRowHeight_y = tpcModule.getRowHeight(row)/2.;

	gear::Vector2D padCenter=tpcModule.getPadCenter( gpadIndex.getPadIndex() );
	gear::Vector2D local_padCenter=tpcModule.globalToLocal(padCenter[0],padCenter[1]);

	//start and end values for the integration
        double xStart, xEnd, yStart, yEnd; 
	//check if the pad is left or right of the gaussian center and change 
	//start position for integration has to be the one closest to the center of the gaussian=local_hit
        if( local_padCenter[0] < local_hit[0] ) {
          xStart = local_padCenter[0]+ halfPadPitch_x;
    	  xEnd   = local_padCenter[0]- halfPadPitch_x;
        }else {
          xStart = local_padCenter[0]- halfPadPitch_x;
	  xEnd   = local_padCenter[0]+ halfPadPitch_x;
        }  
        if(  local_padCenter[1] < local_hit[1]) {
          yStart = local_padCenter[1]+halfRowHeight_y;
	  yEnd   = local_padCenter[1]-halfRowHeight_y;
        }else {
	  yStart = local_padCenter[1]-halfRowHeight_y;
	  yEnd   = local_padCenter[1]+halfRowHeight_y;
        }
	
	// integration over a 2d gaussian
	double charge =hit->getEDep()*CalculateGausIntegral(sigmaTrans,xStart,xEnd,local_hit[0])*
	  CalculateGausIntegral(sigmaTrans,yStart,yEnd,local_hit[1]);

	streamlog_out(DEBUG4) <<"integrating mod "<<gpadIndex.getModuleID()<< " pad "<<tpcModule.getPadNumber( gpadIndex.getPadIndex() )
			      <<" in row "<<row<<" to "<<fabs(charge)<<" electrons."<<std::endl;
	charge = fabs(charge);
	if ( charge > 0.5 ) // only perform integration if there can be charge after integration and rounding
	{
	  FillZBins(hit->getTime(),sigmaLong,charge,gpadIndex, voxelList );
	  //induction on left and right neighbour
	  if( charge*_inductionFactor > 0.5 && _inductionFactor>0.0)
	    ApplyInduction(hit->getTime(),sigmaLong,charge*_inductionFactor,gpadIndex, voxelList);
	} // if ( charge > 0.5 ) 

	padIndexList.erase(padIndexList.begin());
      }// while  padIndexList
    }
    return voxelList;
  }

  // Smears the distributes the charge of the given hit in voxels according to the given pads, 
  //transversal and longitudinal diffusion for polar pad layout
  std::list<TPCVoxel> HitDistributor::distributeHitOnPadsPolar( std::list<GlobalPadIndex> padIndexList, 
							   SimTrackerHit* hit, 
							   double sigmaTrans, 
							   double sigmaLong ) {
    std::list<TPCVoxel> voxelList;
    
    //integration should completely happen in local coordinates of the padplane
    //as modules can be translated and/or rotated
    //first get the center of the gaussian into local coordinates:
    const gear::TPCModule & tpcModule = _tpcParameters->getNearestModule(_c0Hit,_c1Hit);
    gear::Vector2D local_hit = tpcModule.globalToLocal(_c0Hit,_c1Hit);
    //now transform r and phi into cartesian x and y coordinates for integration
    double x_local_hit=local_hit[0]*cos(local_hit[1]);
    double y_local_hit=local_hit[0]*sin(local_hit[1]);

    if( hit->getEDep() > 0 ){
      
      //integration for all the stored pad indices
      while(padIndexList.begin()!=padIndexList.end()) {
        GlobalPadIndex gpadIndex = padIndexList.front();
        const gear::TPCModule & tpcModule = _tpcParameters->getModule(gpadIndex.getModuleID());

	//calculate the width of the pad (with pitch!)
	int row = tpcModule.getRowNumber( gpadIndex.getPadIndex() );
	double halfPadPitch_phi = tpcModule.getPadPitch( gpadIndex.getPadIndex() )/2;
	//calculate the height of the pad 
	double halfRowHeight  = tpcModule.getRowHeight(row)/2.;


	gear::Vector2D padCenter=tpcModule.getPadCenter( gpadIndex.getPadIndex() );
	gear::Vector2D local_padCenter=tpcModule.globalToLocal(padCenter[0],padCenter[1]);
	//now transform r and phi into cartesian x and y coordinates for integration
	double sinPhi = sin(local_padCenter[1]);
        double cosPhi = cos(local_padCenter[1]);
      	double x_local_padCenter=local_padCenter[0]*cosPhi;
	double y_local_padCenter=local_padCenter[0]*sinPhi;

	double halfPadPitch_x=halfPadPitch_phi*local_padCenter[0];

	//we now have to rotate the coordinate system so that the y axis is along the r direction of the pad
	//rotation angle is 90deg-phi
	double xHitRot = x_local_hit*sinPhi  - y_local_hit*cosPhi;
        double yHitRot = x_local_hit*cosPhi + y_local_hit*sinPhi;
        double xPadRot = x_local_padCenter*sinPhi - y_local_padCenter*cosPhi;
        double yPadRot = x_local_padCenter*cosPhi + y_local_padCenter*sinPhi;

	//start and end values for the integration
        double xStart, xEnd, yStart, yEnd; 
	//check if the pad is left or right of the gaussian center and change 
	//sign of integration direction to get a positive charge
        if( xPadRot < xHitRot ) {
          xStart = xPadRot+ halfPadPitch_x;
    	  xEnd   = xPadRot- halfPadPitch_x;
        }else {
          xStart = xPadRot- halfPadPitch_x;
	  xEnd   = xPadRot+ halfPadPitch_x;
        }  
        if(   yPadRot < yHitRot) {
          yStart = yPadRot+halfRowHeight;
	  yEnd   = yPadRot-halfRowHeight;
        }else {	   
	  yStart = yPadRot-halfRowHeight;
	  yEnd   = yPadRot+halfRowHeight;
        }

	// integration over a 2d gaussian
	double charge =hit->getEDep()*CalculateGausIntegral(sigmaTrans,xStart,xEnd,xHitRot)*
	  CalculateGausIntegral(sigmaTrans,yStart,yEnd,yHitRot);
	
	streamlog_out(DEBUG4) <<"integrating mod "<<gpadIndex.getModuleID()<< " pad "<<tpcModule.getPadNumber( gpadIndex.getPadIndex() )
			      <<" in row "<<row<<" to "<<fabs(charge)<<" electrons."<<std::endl;

	charge = fabs(charge);
	if ( charge > 0.5 ) // only perform integration if there can be charge after integration and rounding
	{
	  FillZBins(hit->getTime(),sigmaLong,charge,gpadIndex, voxelList );
	  if( charge*_inductionFactor > 0.5 && _inductionFactor>0)
	    ApplyInduction(hit->getTime(),sigmaLong,charge*_inductionFactor,gpadIndex, voxelList);	
	} // if ( charge > 0.5 ) 

	padIndexList.erase(padIndexList.begin());
      }// while  padIndexList
    }
    return voxelList;
  }
  
  
  double HitDistributor::CalculateGausIntegral(double sigma, double start, double end, double center)
  {
    return 0.5 *((erf((end  - center)/(sqrt(2.)*sigma))
		  -erf((start- center)/(sqrt(2.)*sigma))));
  }

  void HitDistributor::ApplyInduction(double hitTime, double sigmaLong,double charge,GlobalPadIndex gpadIndex, std::list<TPCVoxel> &voxelList)
  {
    const gear::TPCModule & tpcModule =  _tpcParameters->getModule(gpadIndex.getModuleID());
    GlobalPadIndex np = gpadIndex;
    try{
      int left=tpcModule.getLeftNeighbour(gpadIndex.getPadIndex());
      np.setPadIndex( left );
      FillZBins(hitTime,sigmaLong,charge,np, voxelList );
    }
    catch(std::exception& e){}
    try{
      int right=tpcModule.getRightNeighbour(gpadIndex.getPadIndex());
      np.setPadIndex( right );
      FillZBins(hitTime,sigmaLong,charge,np, voxelList );
    }
    catch(std::exception& e){}
  }

  void HitDistributor::FillZBins(double hitTime, double sigmaLong,double charge,GlobalPadIndex gpadIndex, std::list<TPCVoxel> &voxelList )
  {
     int zBin = static_cast<int>( hitTime  /_zBinLength ); 
 
	  if( sigmaLong )
	  {
	    const double tDiff = 4;

	    int minBin = static_cast<int>( (hitTime - (tDiff*sigmaLong)) / _zBinLength  );
	    int maxBin = static_cast<int>( ceil( (hitTime + (tDiff*sigmaLong)) / _zBinLength  ));

	    if( minBin < 0 )
	      minBin = 0;

	    for(int j = minBin; j <= maxBin; j++)
	    { 
	      
	      int tDiffCharge = static_cast<int> ( round (charge * 0.5 * ( erf(((j+1)*_zBinLength-hitTime)/(sqrt(2.)*sigmaLong))-
								     erf(( j   *_zBinLength-hitTime)/(sqrt(2.)*sigmaLong))  )));


	      voxelList.push_back( TPCVoxel(gpadIndex.getPadIndex(), j, tDiffCharge, gpadIndex.getModuleID() ) );
	    }
	  }
	  else // if sigmalong
	  { 
	    voxelList.push_back(TPCVoxel(gpadIndex.getPadIndex(), zBin, static_cast<int>(round(charge)),
					 gpadIndex.getModuleID() )); 
	  }
  }
      
} //namespace marlintpc
