/** Processor to sort primary ions into voxel
 */
#include "IonsInVoxelsProcessor.h"

//C++
#include <iostream>
#include <cmath>
#include <complex>

//LCIO
#include <EVENT/SimTrackerHit.h>

//GEAR
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>

//MARLIN
#include <marlin/Global.h>


#include <TMath.h>

using namespace lcio ;
using namespace marlin ;
using namespace std;

namespace marlintpc{

  
IonsInVoxelsProcessor aIonsInVoxelsProcessor ;


  IonsInVoxelsProcessor::IonsInVoxelsProcessor() : Processor("IonsInVoxelsProcessor") ,
						   _chargeInVoxelsTree(0), _rootFile(0)
  {
  
  _description = "IonsInVoxelsProcessor puts primary ions in voxels..." ;
  
  
  // register steering parameters: name, description, class-variable, default value


  registerInputCollection(    LCIO::SIMTRACKERHIT,
                              "InputCollectionName" , 
			      "e- from PrimaryIonisation (default: PrimaryIonisation)"  ,
			      _inputIonsColName ,
			      string("PrimaryIonisation") ) ;

  registerOutputCollection(   LCIO::LCGENERICOBJECT ,
                              "OutputCollectionName" ,
                              "Name of the Output GEM collection (default: IonOccupancy)" ,
                              _outputVoxelColName ,
			      string("IonOccupancy") );

  registerProcessorParameter( "BinLength" , 
			      "Length of a z bin in mm (default 2.5)"  ,
			      _binLength,
			      double(2.5)) ;

  registerProcessorParameter("WriteOutputToStorage",
			     "if true the output collection is written",
			     _outputIsPersistent,
			     true);

  registerProcessorParameter( "OutputType" , 
			      "Determines if output describes only the ions of the event or all ions in the chamber (default: 0)"  ,
			      _outputType,
			      int(0)) ;

  registerProcessorParameter( "IonisationEnergy" ,
			      "The energy in GeV needed to ionize a gas particle (default: 26.E-9 = Argon) ",
			      _ionisationEnergy,
			      26.E-9); //default value is for argon

  registerOptionalParameter(  "WriteAfterNEvents" ,
			      "Force readout of the full TPC after overlaying n events (optional)" ,
			      _writeAfterNEvents,
			      312 );

  registerOptionalParameter( "WriteRootFile" , 
			     "File name for a root tree with the charge per voxel for debugging (optional)",
			      _rootFileName,
			     std::string("")) ;
  
  registerProcessorParameter( "UseNegativeHalfTPC" ,
			      "Use the negative instead of the positive half TPC, default: fase",
			      _useNegativeHalf,
			      false);


  _voxelTPC = 0;  
  _padLayout = 0; 
  _padLayoutType = 0; 

}

void IonsInVoxelsProcessor::init() { 

  streamlog_out(MESSAGE2) << _description << endl;

  printParameters() ;

  if (parameterSet("WriteAfterNEvents") && _outputType==0)
  {
    streamlog_out(WARNING) << "WriteAfterNEvents is activated, forcing OutputType to 1 (was 0)"
			   << std::endl;
  }

  //Get the pad layout from gear for the later binning of the ions
  //in the x,y-plane.
  try{
    _padLayout = &(Global::GEAR->getTPCParameters().getPadLayout());
    _padLayoutType = _padLayout->getPadLayoutType();
  }
  catch(std::exception& e){
    streamlog_out(ERROR)<< e.what() << std::endl;
    return;
  }

  _maxDriftLength = Global::GEAR->getTPCParameters().getMaxDriftLength();

  //Instanciate a VoxelTPC to store the ions in voxels.
  if( _voxelTPC )
    delete _voxelTPC;
  _voxelTPC = new marlintpc::VoxelTPC( & Global::GEAR->getTPCParameters(), _binLength );


  // initialise the root tree if requested
  
  // the default name is empty. It's safer to check if the name is empty than querying if
  // the parameter is set because someone could set an empty name...
  if (!_rootFileName.empty())
  {
    _rootFile = new TFile( _rootFileName.c_str(), "RECREATE");

    _chargeInVoxelsTree = new TTree("chargeInVoxels", "chargeInVoxels");

    _chargeInVoxelsTree->Branch("zBinLength", &_binLength);

    _chargeInVoxelsTree->Branch("chargePerVolume",&_leaf_chargePerVolume);
    _chargeInVoxelsTree->Branch("charge",&_leaf_charge);
    _chargeInVoxelsTree->Branch("padIndex",&_leaf_padIndex);
    _chargeInVoxelsTree->Branch("zIndex",&_leaf_zIndex);
    _chargeInVoxelsTree->Branch("x",&_leaf_x);
    _chargeInVoxelsTree->Branch("y",&_leaf_y);
    _chargeInVoxelsTree->Branch("z",&_leaf_z);

  }
}



void IonsInVoxelsProcessor::processRunHeader(lcio::LCRunHeader* run) {

  run->parameters().setValue( _processorName + "_revision", "$Rev: 3907 $" );

  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
    if ( ! i->second->isOptional() || i->second->valueSet() ) {
      run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
    }
  }

} 

 
void IonsInVoxelsProcessor::processEvent(lcio::LCEvent * evt ) {


  //Gets the input collection of primary ions..
  LCCollection* inputIonCol = 0;
  try{
    inputIonCol = evt->getCollection( _inputIonsColName );
  }
  catch(DataNotAvailableException& e){
    streamlog_out(WARNING4) << "Could not get input collection " 
			    << _inputIonsColName << endl;
    return;
  }

  //Loops over all ions in the input collection. If the x,y coordinates
  //of the ions are within the pad plane there charge is put in a voxel.
  for(int i = 0; i < inputIonCol->getNumberOfElements() ; i++)
  {

    SimTrackerHit* myIon = dynamic_cast<SimTrackerHit*>( inputIonCol->getElementAt(i) );

    //Set padIndex to the index of the pad nearest to the hit of myIon. 
    //If the hit is outside of the pad plane it is set to minus one.
    int padIndex = getPad( _padLayout , myIon );

    if( padIndex >= 0 && (_useNegativeHalf ? myIon->getPosition()[2] < 0
			                   : myIon->getPosition()[2] >= 0) )
    {
      int charge = 0;

      // if the deposited energy is more than 1 GeV it is the old format with number of electrons
      // instead of energy
      if( myIon->getEDep() >= 1 )
	charge = static_cast<int>( myIon->getEDep() );
      else
	charge =  static_cast<int>(round( myIon->getEDep() / _ionisationEnergy ));

      double zBinning = _maxDriftLength - fabs(myIon->getPosition()[2]);

      _voxelTPC->putChargeInVoxel( padIndex , zBinning , charge );  
    }
            
  }//for i

  bool forcedWriteInThisEvent=false;
  
  // check if writing out after a fixed number of events is activated
  if ( parameterSet("WriteAfterNEvents") )
  {
    if ( evt->getEventNumber() % _writeAfterNEvents == (_writeAfterNEvents - 1) )
    {
      // if the reqested number of events is reached read out the full TPC 
      forcedWriteInThisEvent=true;
    }
    else
    {
      // Writing only after n events is requested (parameter is active),
      // but n is not large enough. Don't write anything, just return.
      return;
    }
  }
      
  LCCollectionVec* outputVoxelCol = 0;
      
  if( (_outputType==0) || forcedWriteInThisEvent)
  {
    // write only data from one event or read out in this one:
    // empty the voxel map when reading it
    outputVoxelCol = _voxelTPC->fetchVoxelCollection();
  }
  else
  {
    // only get the collection, but leave the ions in the map to sum up
    outputVoxelCol = _voxelTPC->getVoxelCollection();
  }

  if ( !outputVoxelCol ){
    streamlog_out(DEBUG2) << "Empty voxel collection while input was not empty." << std::endl;
    return;
  }

  //Set write mode.
  if(  !_outputIsPersistent ) 
  {
    outputVoxelCol->setTransient(true);
  }

  //Without giving the binnig in z the voxels cannot be used properly.
  float binLength=static_cast<float>(_binLength);
  outputVoxelCol->parameters().setValue("zBinLength",binLength);

  //Adds a collection with all voxelized ions in the TPC to the event.
  evt->addCollection( outputVoxelCol , _outputVoxelColName );

  if ( (_outputType==0) || forcedWriteInThisEvent)
  {
    fillTree(outputVoxelCol);
  }

}

int IonsInVoxelsProcessor::getPad(gear::PadRowLayout2D const* padLayout , SimTrackerHit* hit){

  //First and second coordinate of the hit. They will be set to x,y or 
  //or r,phi depending on the pad layout.
  double hitFirst;
  double hitSecond;

  //The type of ppad layout is used to determine whether
  //the coordinates of the hit have to be in x,z or in 
  //r,phi.
  int padLayoutType = padLayout->getPadLayoutType();

  //Sets hitFirst and hitSecond depending on the pad layout.
  switch( padLayoutType ) {
  case gear::PadRowLayout2D::CARTESIAN:{
    hitFirst  = hit->getPosition()[0];
    hitSecond = hit->getPosition()[1];
    break;
  }
  case gear::PadRowLayout2D::POLAR: {
    complex<double> compHit( hit->getPosition()[0], hit->getPosition()[1] );
    hitFirst = std::abs(compHit);
    hitSecond = std::arg(compHit);
    if( hitSecond < 0 )
      hitSecond = (2*M_PI)+hitSecond;
    break;
  }
  default:
    throw gear::Exception("Unknown pad layout");
  };

  //Return value of the function. Negative values indicate a hit not 
  //on the padplane.
  
  int padIndex = -1;

  //Checks whether hit is on pad plane. PadRowLayout2D.getPlaneExtend() 
  //return x,y or r,phi coordinates depending on the pad layout.
   
  if( hitFirst  > padLayout->getPlaneExtent()[0] &&
      hitFirst  < padLayout->getPlaneExtent()[1] &&
      hitSecond > padLayout->getPlaneExtent()[2] &&
      hitSecond < padLayout->getPlaneExtent()[3]    ) {
    try{
      //Get the pad nearest to the hit
      padIndex = padLayout->getNearestPad( hitFirst , hitSecond );
    }
    catch( std::out_of_range& e ){
      //If try does not work set padIndex to minus one to prevent
      //crashing of the processor.    
      padIndex = -1;
    }
  }
  else
    padIndex = -1;

  //Returns pad index. This should have been set to a negative value if  
  //the hit is not on pad plane or try did not work.
  return padIndex;
}

void  IonsInVoxelsProcessor::fillTree(LCCollectionVec*outputVoxelCollection)
{
  if (!_chargeInVoxelsTree) 
  {
    // if the tree is not turned on just don't do anything
    return;
  }


  if (outputVoxelCollection->size() == 0)
  {
    // do not write empty events to the tree
    return;
  }

  // clear the leaves
  _leaf_chargePerVolume.clear();
  _leaf_charge.clear();
  _leaf_padIndex.clear();
  _leaf_zIndex.clear();
  _leaf_x.clear();
  _leaf_y.clear();
  _leaf_z.clear();

  for (LCCollectionVec::iterator voxelIter= outputVoxelCollection->begin();
       voxelIter< outputVoxelCollection->end(); voxelIter++)
  {                            ;
    TPCVoxel *voxel = dynamic_cast<TPCVoxel *>(*voxelIter);

    _leaf_charge.push_back( voxel->getCharge() );
    _leaf_padIndex.push_back( voxel->getPadIndex() );
    _leaf_zIndex.push_back( voxel->getTimeIndex() );
    
    gear::Vector2D padCentre = _padLayout->getPadCenter( voxel->getPadIndex() );

    double x,y, width;
    if (_padLayout->getCoordinateType() == gear::PadRowLayout2D::CARTESIAN)
    {
      x = padCentre[0];
      y = padCentre[1];
      width = _padLayout->getPadPitch( voxel->getPadIndex() );
    }
    else // without further checking I assume polar...
    {
      x = padCentre[0]*cos( padCentre[1]);
      y = padCentre[0]*sin( padCentre[1]);
      // the pitch is in radian
      width = _padLayout->getPadPitch( voxel->getPadIndex() )*padCentre[0];      
    }
    
    _leaf_x.push_back(x);
    _leaf_y.push_back(y);
    _leaf_z.push_back( _maxDriftLength - voxel->getTimeIndex()*_binLength );
    
    _leaf_chargePerVolume.push_back( voxel->getCharge() * TMath::Qe() /
				     _padLayout->getRowHeight( voxel->getPadIndex())  /
				     width / _binLength );
  }

  _chargeInVoxelsTree->Fill();
}

void  IonsInVoxelsProcessor::end()
{
  if (_rootFile)
  {
    fillTree( _voxelTPC->fetchVoxelCollection() );

    _rootFile->Write();
    _rootFile->Close();
    delete _rootFile;
    _rootFile=0;
    _chargeInVoxelsTree=0;
  }
}

  IonsInVoxelsProcessor::~IonsInVoxelsProcessor()
{
  delete _voxelTPC;
}

} // end of namespace marlintpc
