//*****************************************************
// Part of program to simulate a TPC with GEM readout *
//                                                    *
// Author: Astrid Muennich                            *
//                                                    *
//   Astrid Muennich                                  *
//   RWTH Aachen, III. Phys. Inst. B                  *
//   Office: 28 A 112, Phone: +49/(0)241/8            *
//   Email: muennich@physik.rwth-aachen.de            *
//                                                    *
//*****************************************************

/** Program to apply electronic effects like shaping and binning of charge into ADCCounts.
 */

//MarlinTPC
#include "TPCElectronicsProcessor.h"
#include "ChargeDistributionProcessor.h"
#include "TPCVoxel.h"
#include "VoxelTPC.h"
#include "GaussianADCPulse.h"
#include "ALTROADCPulse.h"
#include "ALTROADCConvolutedPulse.h"

//tpcconddata
#include "Pedestal.h"

//LCIO
//#include <IO/LCWriter.h>
#include <EVENT/LCCollection.h>
#include <IMPL/TrackerRawDataImpl.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCRunHeaderImpl.h>
#include <IMPL/LCFlagImpl.h>

//Marlin
#include <marlin/Global.h>
#include <marlin/Exceptions.h>

//lccd
#include <lccd.h>
#include <lccd/LCConditionsMgr.hh>
#include <lccd/DataFileHandler.hh>

//C++
#include <iostream>
#include <cmath>
#include <set>
#include <cstdlib>

//CLHEP
#include <CLHEP/Random/RandGauss.h>

using namespace lcio ;
using namespace marlin ;
using namespace std;
using namespace tpcconddata;

namespace marlintpc
{


TPCElectronicsProcessor aTPCElectronicsProcessor;

TPCElectronicsProcessor::TPCElectronicsProcessor() : Processor("TPCElectronicsProcessor")
{

    _description = "TPCElectronicsProcessor takes care of binning the times and charges of electrons into an ADC.";

    registerInputCollection(LCIO::LCGENERICOBJECT,
                            "InputCollectionName" ,
                            "Name of the Input collection with TPCVoxels (default: DistributedElectrons)"  ,
                            _inputColName,
                            std::string("DistributedElectrons")) ;

    registerOutputCollection(LCIO::TRACKERRAWDATA,
                             "OutputCollectionName",
                             "Name of the output collection with adc values (default: TPCRawData)",
                             _outputColName,
                             std::string("TPCRawData")),

    registerProcessorParameter("ReadoutFrequency" ,
                             "Readout Frequency in Mhz (default: 40.)"  ,
                             _frequency ,
                             40.) ;

    registerProcessorParameter("RiseTime" ,
                               "Rise Time of Shaper in ns (default: 60)"  ,
                               _riseTime ,
                               60.) ;

    registerProcessorParameter("Asymmetry" ,
                               "Asymmetry of the Gaussian shape of signal (default: 1.0)"  ,
                               _asym ,
                               1.0) ;

    registerProcessorParameter("ADCMaxValue" ,
                               " Dynamic range of the ADC in ADC counts (default 1023)"  ,
                               _ADCMaxValue ,
                               1023) ;

    registerProcessorParameter("NElectronsDynamicRange",
                               "Dynamic range of the ADC in electrons (default 1e6)",
                               _nElectronsDynamicRange,
                               int(1e6));

    registerProcessorParameter("ADCThreshold",
                               "Threshold for ADC in ADC bins (default: 1, just zero suppression)",
                               _ADCThreshold,
                               1);

    registerProcessorParameter("MinimumPulseHeight",
                               "Minimal value of the highest sample in a pulse (default: 1)",
                               _minimumPulseHeight,
                               1);

    registerProcessorParameter("ElectronicsSwitch",
                               "Switch between different electronics: 0=Gaussian shaping, 1=ALTRO electronics, 2=ALTRO convolution (does not work with ClockJitter) (default: 0)",
                               _electronicsSwitch,
                               (int)0);

    registerProcessorParameter("PeakingTime",
                               "Peaking time ALTRO electronics in ns (time from start of pulse to maximum) (default 190 ns)",
                               _peakingTime,
                               190.0);

    registerProcessorParameter("VoxelZLength",
                               " Length of a voxel in ns (default 1.)",
                               _voxelZLength,
                               1.);

    registerProcessorParameter("NoiseSwitch",
                               "Switch between different noise model for pedestal width distribution: 0=no noise, 1=gaussian, 2=landau   (default: 0). If Gaus or Landau are chosen, you also have to set the parameters NoiseLevel and NoiseWidth for the choice to have any effect.",
                               _noiseSwitch,
                               (int)0);

    registerOptionalParameter("ClockJitter",
                               "Clock jitter for the digitisation in nanoseconds (cannot be used in case of ElectronicsSwitch == 2)  (default: 0).",
                               _clockJitter,
                               0.0);

    registerProcessorParameter("WriteOutputToStorage",
                               "Output collection will be written if set to true (default: true)",
                               _outputIsPersistent,
                               true);

    registerProcessorParameter("ImproveMemoryFootprint",
                               "Flag whether to use ChargeDistributionProcessor::fetchVoxelCollection instead of the input collection (default:false)",
                               _improveMemoryFootprint,
                               false);

    registerOptionalParameter("IntegrationTime",
                              "IntegrationTime of the electronics. Voxel within this time window will be shaped in one go. (default:120 ns)",
                              _integrationTime,
                              120.0);

    registerOptionalParameter("NoiseLevel",
                              "Central value of the pedestal width used to add noise to the ADC spectrum. (default: 0)",
                              _noiseLevel,
                              0.0);

    registerOptionalParameter("NoiseWidth",
                              "Width of the pedestal width distribution used to add noise to the ADC spectrum. (default: 0)",
                              _noiseWidth,
                              0.0);

    registerOptionalParameter("NoiseCutoff",
                              "Cutoff for random pedestal width - e.g. to discard tails from the Landau distribution, since really noisy channels probably wouldn't be used in the data either. (default: 4)",
                              _noiseCutoff,
                              4.0);
}

TPCElectronicsProcessor::~TPCElectronicsProcessor()
{
    // delete the pulse which was allocated in init()
    delete pulse;
}

void TPCElectronicsProcessor::init()
{
    streamlog_out(MESSAGE2) << _description << endl;
    printParameters() ;
    switch(_electronicsSwitch)
    {
        case 0:
            pulse = new GaussianADCPulse(_frequency, _riseTime, _asym, _nElectronsDynamicRange, _ADCMaxValue, _ADCThreshold);
            break;
        case 1:
            pulse = new ALTROADCPulse(_frequency, _peakingTime, _nElectronsDynamicRange, _ADCMaxValue, _ADCThreshold);
            break;
        case 2:
            pulse = new ALTROADCConvolutedPulse(_frequency, _peakingTime, _nElectronsDynamicRange, _ADCMaxValue, _ADCThreshold);
            break;
        default:
            streamlog_out(ERROR) << "Unknown electronics type selected!" << std::endl;
            throw lcio::Exception("TPCElectronicsProcessor: Internal error: Unknown electronics type selected!");
    }
}


void TPCElectronicsProcessor::processRunHeader(lcio::LCRunHeader* run)
{
    run->parameters().setValue(_processorName + "_revision", "$Rev: 5141 $");

    for(ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
    {
        if(! i->second->isOptional() || i->second->valueSet())
        {
            run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
        }
    }
}


void TPCElectronicsProcessor::processEvent(lcio::LCEvent* evt)
{

    streamlog_out(DEBUG2) << "Hi, it's me again " << evt->getEventNumber() << std::endl;

    //Sort TPCVoxels by padIndex first, then by time index
    std::list< std::pair<VoxelTPC::VoxelIndex, int> >* padSortedVoxels;

    if(_improveMemoryFootprint)
    {
        streamlog_out(DEBUG4) << "fetching voxel list" << std::endl;
        padSortedVoxels = ChargeDistributionProcessor::fetchVoxelList();

        if(padSortedVoxels == 0)
        {
            // there was no voxel list in this event
            streamlog_out(DEBUG4) << "No voxel list in ChargeDistributionProcessor in event " <<
                                  evt->getEventNumber() << std::endl;
            return;
        }

        streamlog_out(DEBUG4) << "size of voxel list is " << padSortedVoxels->size()
                              << std::endl;
    }
    else
    {
        LCCollectionVec* inputColVec = 0;

        try
        {
            inputColVec = dynamic_cast<LCCollectionVec*>(evt->getCollection(_inputColName));
        }
        catch(DataNotAvailableException& e)
        {

            // the collection of electron clouds with that name wasn't in the LCIO event file, skip simulating this event
            //    throw marlin::SkipEventException( this );

            return; // just return, otherwise the data from the previous processors is not written since the
            //  OutputProcessor is also skipped
        }

        streamlog_out(DEBUG4) << "Size of input collection is " << inputColVec->size()  << std::endl;
        if(inputColVec->empty())
        {
            streamlog_out(WARNING) << "Input collection exists, but is empty!" << std::endl;
            return;
        }

        padSortedVoxels = new  std::list< std::pair<VoxelTPC::VoxelIndex, int> >;

        for(LCCollectionVec::iterator voxelIter = inputColVec->begin() ;
                voxelIter < inputColVec->end() ; voxelIter++)
        {
            TPCVoxel voxel(*voxelIter);

            padSortedVoxels->push_back(std::pair< VoxelTPC::VoxelIndex, int> (
                                           VoxelTPC::VoxelIndex(voxel.getPadIndex(), voxel.getTimeIndex(), voxel.getModuleID()),
                                           voxel.getCharge()));
        }
    }


    //Initialise and set transient for output collection
    LCCollectionVec* outputColVec = new LCCollectionVec(LCIO::TRACKERRAWDATA)  ;
    if(!_outputIsPersistent)
    {
        outputColVec->setTransient(true);
    }

    LCFlagImpl flag(0);
    flag.setBit(LCIO:: TRAWBIT_ID1);
    outputColVec->setFlag(flag.getFlag());

    // if clock jitter should be simulated, generate it now for each time bin
    if(parameterSet("ClockJitter"))
    {
        // to get length of the event: first and last time bin
        padSortedVoxels->sort(voxelTimeComp);
        std::list< std::pair<VoxelTPC::VoxelIndex, int> >::iterator voxelTimeIter = padSortedVoxels->begin();
        voxelTimeIter = padSortedVoxels->begin();
        int firstJitterTime = voxelTimeIter->first.zBin;
        voxelTimeIter = --padSortedVoxels->end();
        int lastJitterTime = voxelTimeIter->first.zBin;
        switch(_electronicsSwitch)
        {
            case 0:
                lastJitterTime += (_riseTime + _asym * _riseTime); // (endTime in GaussianPulse::add)
                break;
            case 1:
                lastJitterTime += (4 * _peakingTime); // (see end of while-loop in ALTROADCPulse::add)
                break;
            case 2:
                streamlog_out(ERROR) << "ElectronicsSwitch==2 does not work with ClockJitter!" << std::endl;
                throw lcio::Exception("TPCElectronicsProcessor: Internal error: ElectronicsSwitch==2 does not work with ClockJitter!");
            default:
                streamlog_out(ERROR) << "Unknown electronics type selected!" << std::endl;
                throw lcio::Exception("TPCElectronicsProcessor: Internal error: Unknown electronics type selected!");
        }
        // convert zbins to readout frequency bins (time is in ns, frequency in MHz)
        int startJitterBin = static_cast<int>(firstJitterTime * _frequency / 1000); // bin number for the first bin with charge in this event
        int endJitterBin   = static_cast<int>(lastJitterTime * _frequency / 1000) + 1; // bin number for the last pulse bin with charge

        // chose some clock jitter for each time bin recorded in this event
        _timeJitter = std::map<int, double>(); // clear the jitter map
        for(int izbin = startJitterBin; izbin <= endJitterBin; izbin++)
        {
            _timeJitter.insert(std::pair<int, double>(izbin, CLHEP::RandGauss::shoot(0.0, _clockJitter)));
        }
        pulse->setJitterValues(_timeJitter);
        streamlog_out(DEBUG3) << "first zBin " << startJitterBin << " with jitter " << _timeJitter[startJitterBin] << endl;
        streamlog_out(DEBUG3) << "last zBin " << endJitterBin << " with jitter " << _timeJitter[endJitterBin] << endl;
    }


    // As the list in a non-sorted container we have to sort manually.
    // resort the output of the VoxelTPC by module first, then pad, then time
    padSortedVoxels->sort(voxelPadComp);

    //integrate charge of voxel within integration time window before entering shaping
    double _deadTime = 0;
    if(parameterSet("IntegrationTime"))
    {
        streamlog_out(DEBUG3) << "IntegrationTime: " << _integrationTime << std::endl;
        std::list< std::pair<VoxelTPC::VoxelIndex, int> >::iterator FirstvoxelIter = padSortedVoxels->begin();
        streamlog_out(DEBUG3) << "Start: " << FirstvoxelIter->first.padIndex << " " << FirstvoxelIter->first.moduleID << " " << FirstvoxelIter->first.zBin
                              << " " << FirstvoxelIter->first.zBin* _voxelZLength << " " << FirstvoxelIter->second << " " << std::endl;
        double center = FirstvoxelIter->first.zBin * _voxelZLength * FirstvoxelIter->second;

        for(std::list< std::pair<VoxelTPC::VoxelIndex, int> >::iterator voxelIter
                = ++padSortedVoxels->begin(); voxelIter != padSortedVoxels->end() ; voxelIter++)
        {
            streamlog_out(DEBUG1) << "Input: " << voxelIter->first.padIndex << " " << voxelIter->first.moduleID << " " << voxelIter->first.zBin
                                  << " " << voxelIter->first.zBin* _voxelZLength << " " << voxelIter->second << " " << std::endl;
            //voxel is on different channel
            if((voxelIter->first.padIndex != FirstvoxelIter->first.padIndex)
                    || (voxelIter->first.moduleID != FirstvoxelIter->first.moduleID))
            {
                FirstvoxelIter->first.zBin = static_cast<int>(center / FirstvoxelIter->second / _voxelZLength);
                streamlog_out(DEBUG1) << "Final: " << FirstvoxelIter->first.zBin << " " << FirstvoxelIter->second << std::endl;
                FirstvoxelIter = voxelIter;
                center = FirstvoxelIter->first.zBin * _voxelZLength * FirstvoxelIter->second;
            }
            else
            {
                //voxel is within integration time
                if(voxelIter->first.zBin * _voxelZLength < FirstvoxelIter->first.zBin * _voxelZLength + _integrationTime)
                {
                    FirstvoxelIter->second += voxelIter->second;
                    center += voxelIter->first.zBin * _voxelZLength * voxelIter->second;
                    streamlog_out(DEBUG1) << "Int " << FirstvoxelIter->second << " " << center << "("
                                          << voxelIter->first.zBin* _voxelZLength << " " << voxelIter->second << ")" << std::endl;
                    padSortedVoxels->erase(voxelIter);
                    voxelIter--;
                }
                else if(voxelIter->first.zBin * _voxelZLength < FirstvoxelIter->first.zBin * _voxelZLength + _integrationTime + _deadTime)
                {
                    padSortedVoxels->erase(voxelIter);
                    voxelIter--;
                    streamlog_out(DEBUG3) << "Removed voxel due to dead time" << std::endl;
                }
                else
                {
                    //one pulse finished, set time of charge to center of gravity of charge cloud
                    FirstvoxelIter->first.zBin = static_cast<int>(center / FirstvoxelIter->second / _voxelZLength);
                    streamlog_out(DEBUG1) << "Final 2: " << FirstvoxelIter->first.zBin << " " << FirstvoxelIter->second << std::endl;
                    FirstvoxelIter = voxelIter;
                    center = FirstvoxelIter->first.zBin * _voxelZLength * FirstvoxelIter->second;
                }
            }
        }//list loop
        FirstvoxelIter->first.zBin = static_cast<int>(center / FirstvoxelIter->second / _voxelZLength);
        streamlog_out(DEBUG1) << "Final 3: " << FirstvoxelIter->first.zBin << " " << FirstvoxelIter->second << std::endl;
    }//(parameterSet("IntegrationTime"))

    // initialise the pulse with the first pad index
    if(!padSortedVoxels->empty()) pulse->clear(padSortedVoxels->begin()->first.padIndex,
                                                   padSortedVoxels->begin()->first.moduleID);

    int nPulsesWhichHaveBeenSplit = 0;
    int nPulsesAfterSplitting = 0;

    for(std::list< std::pair<VoxelTPC::VoxelIndex, int> >::iterator voxelIter
            = padSortedVoxels->begin(); voxelIter != padSortedVoxels->end() ; voxelIter++)
    {
        streamlog_out(DEBUG1) << "Voxel:  pad=" << voxelIter->first.padIndex << " mod=" << voxelIter->first.moduleID
                              << "  time=" << voxelIter->first.zBin* _voxelZLength << " Q=" << voxelIter->second << std::endl;
        if((voxelIter->first.padIndex != pulse->getChannelNumber())
                || (voxelIter->first.moduleID != pulse->getModuleID())
                // The charge in this voxel belongs to a new pulse, since it is on another channel

                || ((pulse->getStartBin() >= 0) &&
                    (voxelIter->first.zBin * _voxelZLength > pulse->getEndTime())))
            // The pulse is not empty and the voxel time is after the end time of the pulse,
            // so they don't overlap
        {
            // Write the old pulse, then initialise a new one
        	if(_electronicsSwitch==2) pulse->convoluteSignal();
            std::list<TrackerRawData*> lcioPulses = pulse->getLCIORawData();
            for(std::list<TrackerRawData*>::iterator pulseIter = lcioPulses.begin();
                    pulseIter != lcioPulses.end(); pulseIter++)  // loop the list
            {
                //check for minimum pulse height
                short pulseMaxium = 0;
                for(ShortVec::const_iterator sampleIter = (*pulseIter)->getADCValues().begin();
                        sampleIter < (*pulseIter)->getADCValues().end(); sampleIter++)
                {
                    if(*sampleIter > pulseMaxium)
                    {
                        pulseMaxium =  *sampleIter;
                    }
                }

                if(pulseMaxium >=  _minimumPulseHeight)
                {
                    outputColVec->addElement(*pulseIter);
                    streamlog_out(DEBUG2) << "Pulse length is " << (*pulseIter)->getADCValues().size()
                                          << " on channel " << (*pulseIter)->getCellID0() << std::endl;
                }
                else
                {
                    streamlog_out(DEBUG3) << "Rejected pulse on channel " << (*pulseIter)->getCellID0()
                                          << " in module " << (*pulseIter)->getCellID1()
                                          << ", pulse height only " << pulseMaxium
                                          << std::endl;
                }
            }

            if(lcioPulses.empty())
            {
                streamlog_out(DEBUG2) << "empty pulse arrived in processor" << std::endl;
            }

            if(lcioPulses.size() > 1)
            {
                nPulsesAfterSplitting += lcioPulses.size();
                nPulsesWhichHaveBeenSplit++;
            }
            // when preparing the next pulse, pass the chosen noise simulation parameters along
            if(parameterSet("NoiseLevel") && parameterSet("NoiseWidth"))
            {
                pulse->clear(voxelIter->first.padIndex, voxelIter->first.moduleID , _noiseSwitch, _noiseLevel, _noiseWidth, _noiseCutoff);
            }
            else
            {
                pulse->clear(voxelIter->first.padIndex, voxelIter->first.moduleID , 0, -1.0, -1.0, -1.0);
            }
        }

        pulse->add(voxelIter->first.zBin * _voxelZLength ,  voxelIter->second);

    }// for voxelIter

    delete padSortedVoxels;

    //Add ouput collection to event
    if(outputColVec->empty())
        delete outputColVec;
    else
        evt->addCollection(outputColVec , _outputColName) ;

    streamlog_out(DEBUG4) << nPulsesWhichHaveBeenSplit << " pulses have been split in event "
                          << evt->getEventNumber() << ", resulting in "
                          << nPulsesAfterSplitting << " pulses." << std::endl;
}


}//end of namespace marlintpc
