#ifndef ChannelCorrectionListener_h
#define ChannelCorrectionListener_h 1

//C++
#include <iostream>
#include <utility>
#include <stdexcept>

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "IMPL/LCCollectionVec.h"
#include "UTIL/LCTime.h"

//LCCD
#include "lccd.h"
#include "lccd/DBInterface.hh"
#include "lccd/IConditionsChangeListener.hh"
#include "lccd/IConditionsHandler.hh"

//Marlin
#include "marlin/ConditionsProcessor.h"

//TPCCondData
#include "ChannelCorrectionNew.h"

// the package schould be capsulated in a name space
namespace marlintpc
{

/** Helper Class which provides access to the ChannelCorrections for every event.
 *  It is derived from IConditionsChangeListener, so the ChannelCorrections are
 *  updated every time the conditions data change.
 *  If the requested value for a given pair of channel/readoutGroup or pad/module
 *  is not found, a tpcconddata::CondDataNotAvailableException is thrown.
 * @author R. Diener (DESY)
 */

class ChannelCorrectionListener : public lccd::IConditionsChangeListener
{
    public:

        // Several Typedefs used in following declarations
        /// Integer pair of Hardware IDs: first entry: channelID, second: readoutGroup OR first entry: padID, second: moduleID;
        typedef std::pair<int, int> Int_Pair_t;
        /// Map of \link tpcconddata::ChannelCorrectionNew ChannelCorrectionNew objects \endlink indexed by integer pairs.
        typedef std::map< Int_Pair_t, tpcconddata::ChannelCorrectionNew* > ChannelCorrectionNewMap_t;

        /** Constructor
         * @param ColName name of the collection in which the ChannelCorrectionNews are stored
         */
        ChannelCorrectionListener(std::string ColName = tpcconddata::ChannelCorrectionNew::getDefaultColName());

        /// Destructor
        virtual ~ChannelCorrectionListener();

        /// Returns the default name of the conditions data collection.
        static std::string getDefaultColName()
        {
            return  tpcconddata::ChannelCorrectionNew::getDefaultColName();
        };

        /// Updates stored ChannelCorrectionNew values if conditions have change.
        /** Called by the Marlin ConditionsProcessor, if the condition change
         *  inplemetentation of the IConditionsChangeListener interface
         */
        void conditionsChanged(EVENT::LCCollection* col);

        /// Implementation of the listener class, takes care of default collection settings and registration of the handler.
        void registeredWithHandler(lccd::IConditionsHandler* condHandler);

        /// Implementation of the listener class, cleans up after deregistration and gives warning message.
        void deRegisteredWithHandler(lccd::IConditionsHandler* condHandler);

        /// Get the padID, moduleID for given values of channelID and readoutGroup.
        /** Returns the stored pair of padID, moduleID (GEAR) for the given channelID, readoutGroup (readout electronics)
         *  @param hardwareIDpair Integer pair of channel ID (cellID0) and readout group (cellID1) of the readout
         *  @return hardwareID Integer pair of pad ID (cellID0) and module group (cellID1) of GEAR
         */
        std::pair<int, int> getGearMapping(Int_Pair_t hardwareIDpair);

        /// Get the channelID and readoutGroup for given values of padID, moduleID.
        /** Returns the stored pair of channelID, readoutGroup (readout electronics) for the given padID, moduleID (GEAR).
         *  @param hardwareID Integer pair of pad ID (cellID0) and module group (cellID1) of GEAR
         *  @return hardwareID Integer pair of channel ID (cellID0) and readout group (cellID1) of the readout
         */
        std::pair<int, int> getHardwareMapping(Int_Pair_t gearIDpair);

        /// Get ChannelCorrectionNew Object for given values of channelID and readoutGroup..
        /** Returns the stored \link tpcconddata::ChannelCorrectionNew ChannelCorrectionNew object \endlink for the given channel.
         *  @param channelID,readoutGroup channel ID (cellID0) and readout group (cellID1) of the readout.
         *  @return \link tpcconddata::ChannelCorrectionNew ChannelCorrectionNew object \endlink
         */
        tpcconddata::ChannelCorrectionNew* getChannelCorrectionFromHardwareID(Int_Pair_t hardwareIDpair);

        /// Get ChannelCorrectionNew Object for given values of channelID and readoutGroup..
        /** Returns the stored \link tpcconddata::ChannelCorrectionNew ChannelCorrectionNew object \endlink for the given channel.
         *  @param channelID,readoutGroup channel ID (cellID0) and readout group (cellID1) of the readout.
         *  @return \link tpcconddata::ChannelCorrectionNew ChannelCorrectionNew object \endlink
         */
        tpcconddata::ChannelCorrectionNew* getChannelCorrectionFromGearID(Int_Pair_t gearIDpair);

        float getCalibrationFactorFromGearID(Int_Pair_t gearIDpair);
        float getCalibrationFactorFromHardwareID(Int_Pair_t hardwareIDpair);
        float getTimeOffsetFromGearID(Int_Pair_t gearIDpair);
        float getTimeOffsetFromHardwareID(Int_Pair_t hardwareIDpair);
        float getCalibrationOffsetFromGearID(Int_Pair_t gearIDpair);
        float getCalibrationOffsetFromHardwareID(Int_Pair_t hardwareIDpair);
        bool  isDeadFromGearID(Int_Pair_t gearIDpair);
        bool  isDeadFromHardwareID(Int_Pair_t hardwareIDpair);
        bool  isNoisyFromGearID(Int_Pair_t gearIDpair);
        bool  isNoisyFromHardwareID(Int_Pair_t hardwareIDpair);
        /** One  of two not predefined quality bits */
        bool  isAddonBitASetFromGearID(Int_Pair_t gearIDpair);
        /** One  of two not predefined quality bits */
        bool  isAddonBitASetFromHardwareID(Int_Pair_t hardwareIDpair);
        /** One  of two not predefined quality bits */
        bool  isAddonBitBSetFromGearID(Int_Pair_t gearIDpair);
        /** One  of two not predefined quality bits */
        bool  isAddonBitBSetFromHardwareID(Int_Pair_t hardwareIDpair);


        /// Get the validity start time of the current conditions collection or default collection.
        lcio::long64 collectionValidSince();
        /// Get the validity end time of the current conditions collection or default collection.
        lcio::long64 collectionValidTill();

        /// Set if an exception should the thrown in case of no data available in the database.
        void useLastValidIfNoDataAvailable(bool useLastValid);
        /// Return if currently data is available in the database/file or not.
        bool noConditionsDataAvailable();

    private:

        /// fill the internal conditions map from the given collection
        void fillConditionsMaps(lcio::LCCollection* col);

        /// Holds the name of the conditions data collection
        std::string _collectionName;

        /// Default collection that is registered and will be returned from the handler if no data is available.
        lcio::LCCollectionVec* _thisEmptyCollection;

        /// Map of registered default collections for differnet handlers. Will hold here only one entry, since this listener registers itself with only one handler.
        std::map<lcio::LCCollection* , lccd::IConditionsHandler* > _handlerDefaultCollectionMap;

        /// Maps to store the ChannelCorrectionNews
        ChannelCorrectionNewMap_t  _readoutChannelCorrectionNewMap;
        ChannelCorrectionNewMap_t  _gearChannelCorrectionNewMap;

        /// Internal member to hold the "since" of the validity range of the current conditions collection.
        lcio::long64 _timeStampSince;
        /// Internal member to hold the "till" of the validity range of the current conditions collection.
        lcio::long64 _timeStampTill;

        /// Used to set if the last valid collection should be used if no current data are in the database.
        bool _useLastValidIfNoDataAvailable;
        /// Holds the indormation, if there are no current data available in the database.
        bool _noDataAvailable;

        /// Pointer to hold the address of the handler this listener is registered with.
        lccd::IConditionsHandler* _condHandler;
};

}
#endif
