#ifndef FieldSettingListener_h
#define FieldSettingListener_h 1

//C++
#include <iostream>
#include <utility>
#include <stdexcept>

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "IMPL/LCCollectionVec.h"
#include "UTIL/LCTime.h"

//LCCD
#include "lccd.h"
#include "lccd/DBInterface.hh"
#include "lccd/IConditionsChangeListener.hh"
#include "lccd/IConditionsHandler.hh"

//Marlin
#include "marlin/ConditionsProcessor.h"

//TPCCondData
#include "FieldSetting.h"

// the package schould be capsulated in a name space
namespace marlintpc
{

/** Helper Class which provides access to the field settings for every event.
 *  It is derived from IConditionsChangeListener,
 *  so the Field Settings are updated every time the conditions data change.
 * @author F. Mueller (DESY)
 */

class FieldSettingListener : public lccd::IConditionsChangeListener
{
    public:

        /** Constructor
         * @param ColName name of the collection in which the field settings are stored
         */
        FieldSettingListener(std::string ColName = tpcconddata::FieldSetting::getDefaultColName());

        /// Destructor
        virtual ~FieldSettingListener();

        /// Returns the default name of the conditions data collection.
        static std::string getDefaultColName()
        {
            return  tpcconddata::FieldSetting::getDefaultColName();
        };

        /** Called by the Marlin ConditionsProcessor, if the condition change
         *  implementation of the IConditionsChangeListener interface
         */
        void conditionsChanged(EVENT::LCCollection* col);

        /// Implementation of the listener class, takes care of default collection settings and registration of the handler.
        void registeredWithHandler(lccd::IConditionsHandler* condHandler);

        /// Implementation of the listener class, cleans up after deregistration and gives warning message.
        void deRegisteredWithHandler(lccd::IConditionsHandler* condHandler);

        /// Get FieldSetting Object
        /** Returns the stored \link tpcconddata::FieldSetting FieldSetting object \endlink
         *  @return \link tpcconddata::FieldSetting FieldSetting object \endlink
         */
        tpcconddata::FieldSetting* getFieldSetting();


        /// return the Number of GEMs
        int   getNGEMs() const;

        /** return the status of the GEM System
         *  0: off
         *  1: running normally
         *  2: ramping up
         *  3: ramping down
         *  4: tripped
         *  5: other error
         */
        int   getGEMStatus() const;

        /** return the status of the DriftField System
         *  0: off
         *  1: running normally
         *  2: ramping up
         *  3: ramping down
         */
        int getDriftFieldStatus() const;

        /** return the status of the Magnet
         *  0: off
         *  1: running normally
         *  2: ramping up
         *  3: ramping down
         */
        int getMagneticFieldStatus() const;

        /** return the voltage of the given number
         *  GEM 0 is the nearest the drift region
         */
        float getGEMVoltage( int GEMID ) const;

        /** return the field (in V/mm) between the GEM N and N+1
         */
        float getTransferField( int GEMID ) const;

        /** return the field (in V/mm) between the last GEM and the Pad-Readout-Plane
         */
        float getInductionField() const;

        /** return voltage (in V) of the guard ring on the gem modules
         */
        float getGuardRingVoltage() const;

        /** return voltage (in V) of the cathode of the field cage
         */
        float getCathodeVoltage() const;

        /** return voltage (in V) of the anode of the field cage
         *  (e.g. the shield at the position of the first GEM)
         */
        float getAnodeVoltage() const;

        /** return the drift field (in V/mm) if the length of the drift region (in mm) is given
         */
        float getDriftField( float maxDriftLegth ) const;

        /// return the magnetic field (in T)
        float getMagneticField() const;

        /// Get the validity start time of the current conditions collection or default collection.
        lcio::long64 collectionValidSince();
        /// Get the validity end time of the current conditions collection or default collection.
        lcio::long64 collectionValidTill();

        /// Set if an exception should the thrown in case of no data available in the database.
        void useLastValidIfNoDataAvailable(bool useLastValid);
        /// Return if currently data is available in the database/file or not.
        bool noConditionsDataAvailable();

    private:

        /// Holds the name of the conditions data collection
        std::string _collectionName;

        /// Default collection that is registered and will be returned from the handler if no data is available.
        lcio::LCCollectionVec* _thisEmptyCollection;

        /// Map of registered default collections for differnet handlers. Will hold here only one entry, since this listener registers itself with only one handler.
        std::map<lcio::LCCollection* , lccd::IConditionsHandler* > _handlerDefaultCollectionMap;

        /// Internal member to hold the "since" of the validity range of the current conditions collection.
        lcio::long64 _timeStampSince;
        /// Internal member to hold the "till" of the validity range of the current conditions collection.
        lcio::long64 _timeStampTill;

        /// Used to set if the last valid collection should be used if no current data are in the database.
        bool _useLastValidIfNoDataAvailable;
        /// Holds the indormation, if there are no current data available in the database.
        bool _noDataAvailable;

        /// Pointer to hold the address of the handler this listener is registered with.
        lccd::IConditionsHandler* _condHandler;

        /// Pointer to the TPC conditions object
        tpcconddata::FieldSetting* _fieldSettings;
};

}
#endif
