#ifndef PedestalListener_h
#define PedestalListener_h 1

//C++
#include <iostream>
#include <utility>
#include <stdexcept>

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "IMPL/LCCollectionVec.h"
#include "UTIL/LCTime.h"

//LCCD
#include "lccd.h"
#include "lccd/DBInterface.hh"
#include "lccd/IConditionsChangeListener.hh"
#include "lccd/IConditionsHandler.hh"

//Marlin
#include "marlin/ConditionsProcessor.h"

//TPCCondData
#include "Pedestal.h"

// the package schould be capsulated in a name space
namespace marlintpc
{

/** Helper Class which provides access to the pedestals for every event.
 *  It is derived from IConditionsChangeListener,
 *  so the pedestals are updated every time the conditions data change.
 * @author M. E. Janssen (DESY), R. Diener (DESY)
 */

class PedestalListener : public lccd::IConditionsChangeListener
{
    public:

        // Several Typedefs used in following declarations
        /// Integer pair of Hardware IDs, first entry: channelID, second: readoutGroup;
        typedef std::pair<int, int> ChannelReadoutGroupPair_t;
        /// Map of pedestals.
        typedef std::map< ChannelReadoutGroupPair_t, tpcconddata::Pedestal* > PedestalMap_t;

        /** Constructor
         * @param ColName name of the collection in which the pedestals are stored
         */
        PedestalListener(std::string ColName = tpcconddata::Pedestal::getDefaultColName());

        /// Destructor
        virtual ~PedestalListener();

        /// Returns the default name of the conditions data collection.
        static std::string getDefaultColName()
        {
            return  tpcconddata::Pedestal::getDefaultColName();
        };

        /// Updates stored pedestal values if conditions have change.
        /** Called by the Marlin ConditionsProcessor, if the condition change
         *  inplemetentation of the IConditionsChangeListener interface
         */
        void conditionsChanged(EVENT::LCCollection* col);

        /// Implementation of the listener class, takes care of default collection settings and registration of the handler.
        void registeredWithHandler(lccd::IConditionsHandler* condHandler);

        /// Implementation of the listener class, cleans up after deregistration and gives warning message.
        void deRegisteredWithHandler(lccd::IConditionsHandler* condHandler);

        /// Get Pedestal Object for a channel.
        /** Returns the stored \link tpcconddata::Pedestal Pedestal object \endlink for the given channel
         *  @param hardwareID Integer pair of channel ID (cellID0) and readout group (cellID1) of the pad
         *  @return \link tpcconddata::Pedestal Pedestal object \endlink
         */
        tpcconddata::Pedestal* getPedestal(ChannelReadoutGroupPair_t hardwareID) const;

        /// Get Pedestal Object for a channel.
        /** Returns the stored \link tpcconddata::Pedestal Pedestal object \endlink for the given channel.
         *  @param channelID,readoutGroup channel ID (cellID0) and readout group (cellID1) of the pad
         *  @return \link tpcconddata::Pedestal Pedestal object \endlink
         */
        tpcconddata::Pedestal* getPedestal(int channelID, int readoutGroup) const;

        /// Get pedestal value for a channel.
        /** Returns the pedestal value for the given channel.
         *  @param hardwareID Integer pair of channel ID (cellID0) and readout group (cellID1) of the pad
         *  @return Pedestal mean value (0.0 if no value is stored for the ID).
         */
        float getPedestalValue(ChannelReadoutGroupPair_t hardwareID) const;

        /// Get pedestal value for a channel.
        /** Returns the pedestal value for the given channel.
         *  @param channelID,readoutGroup Channel ID (cellID0) and readout group (cellID1) of the pad
         *  @return Pedestal mean value (0.0 if no value is stored for the ID).
         */
        float getPedestalValue(int channelID, int readoutGroup) const;

        /// Get pedestal width for a channel.
        /** Returns the pedestal width for the given channel.
         *  @param hardwareID Integer pair of channel ID (cellID0) and readout group (cellID1) of the pad
         *  @return Pedestal width (-1.0 if no value is stored).
         */
        float getPedestalWidth(ChannelReadoutGroupPair_t hardwareID) const;

        /// Get pedestal width for a channel.
        /** Returns the pedestal width for the given channel.
         *  @param channelID,readoutGroup Channel ID (cellID0) and readout group (cellID1) of the pad
         *  @return Pedestal width (-1.0 if no value is stored).
         */
        float getPedestalWidth(int channelID, int readoutGroup) const;

        /// Get the validity start time of the current conditions collection or default collection.
        lcio::long64 collectionValidSince();
        /// Get the validity end time of the current conditions collection or default collection.
        lcio::long64 collectionValidTill();

        /// Set if an exception should the thrown in case of no data available in the database.
        void useLastValidIfNoDataAvailable(bool useLastValid);
        /// Return if currently data is available in the database/file or not.
        bool noConditionsDataAvailable();

    private:

        /// fill the internal conditions map from the given collection
        void fillConditionsMap(lcio::LCCollection* col);

        /// Holds the name of the conditions data collection
        std::string _collectionName;

        /// Default collection that is registered and will be returned from the handler if no data is available.
        lcio::LCCollectionVec* _thisEmptyCollection;

        /// Map of registered default collections for differnet handlers. Will hold here only one entry, since this listener registers itself with only one handler.
        std::map<lcio::LCCollection* , lccd::IConditionsHandler* > _handlerDefaultCollectionMap;

        /// Map to store the pedestals
        PedestalMap_t  _pedestalMap;

        /// Internal member to hold the "since" of the validity range of the current conditions collection.
        lcio::long64 _timeStampSince;
        /// Internal member to hold the "till" of the validity range of the current conditions collection.
        lcio::long64 _timeStampTill;

        /// Used to set if the last valid collection should be used if no current data are in the database.
        bool _useLastValidIfNoDataAvailable;
        /// Holds the indormation, if there are no current data available in the database.
        bool _noDataAvailable;

        /// Pointer to hold the address of the handler this listener is registered with.
        lccd::IConditionsHandler* _condHandler;
};

}
#endif
