/*
 * RawDataCutterProcessor.h
 *
 *  Created on: Oct 13, 2011
 *      Author: caiazza
 */

#ifndef RAWDATACUTTERPROCESSOR_H_
#define RAWDATACUTTERPROCESSOR_H_
//MARLINTPC includes
#include "TPCConditionsListener.h"
#include "ADCElectronicsListener.h"

//MARLIN includes
#include "marlin/Processor.h"

//LCIO includes
#include "lcio.h"

//LCCD Includes
#include "lccd.h"

//standard includes
#include <string>


namespace marlintpc
{

/**  This processor is used to select in the raw data only a part of the time spectrum,
 * in fact shortening the time window of the data acquired.
 *
 *  This processor cuts all the data arriving later than the maximum drift time
 *  to clean up the raw data before the main processing phase.
 *
 *  The user can set the length of this time window manually specifying the maximum
 *  number of time bins allowed, or this value can be calculated from the condition data
 *
 *  <h4>Input - Prerequisites</h4>
 *  Needs the collection of TrackerRawData
 *
 *  <h4>Output</h4>
 *  Collection of TrackerRawData
 *
 * @param InputCollectionName:string Name of the input tracker raw data collection
 * @param OutputCollectionName:string Name of the collection where to store the corrected data
 * @param WriteOutputToStorage:bool If true the output collection is written to file
 * @param MaxTimeBinOverride:int Set this parameter if you want to define manually the maximum time bin length. No calculation will be executed
 * @param StartOffset:int Defines the time bin of the anode position
 * @param DriftVelocityOverride:float Defines the drift velocity in mm/us. If set the data in the condition database will not be considered
 * @param DriftLengthOverride:float Defines the maximum drift length in mm of the TPC. If set the data in the GEAR file will not be considered
 * @param ReadoutFrequencyOverride:float Defines the electronics readout frequency in MHz. If set the data in the condition database will not be considered
 * @param EndOffset:int Defines the time bin to add to the calculated maximum as a safety factor
 * @param TPCConditionsCollName:string Defines the name of the condition collection containing the drift velocity
 * @param ADCElectronicsCollName:string Defines the name of the condition collection containing the readout frequency
 *
 * @author S. Caiazza, DESY
 * @version $Id: RawDataCutterProcessor.h,v 1.0 2011-10-13 12:57:39 caiazza Exp $
 */
class RawDataCutterProcessor : public marlin::Processor
{

public:

    virtual Processor*
    newProcessor()
    {
        return new RawDataCutterProcessor;
    }

    RawDataCutterProcessor();

    /** Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms.
     */
    virtual void
    init();

    /** Called for every run.
     */
    virtual void
    processRunHeader(lcio::LCRunHeader* run);

    /** Called for every event - the working horse.
     */
    virtual void
    processEvent(lcio::LCEvent * evt);

    virtual void
    check(lcio::LCEvent * evt);

    /** Called after data processing for clean up.
     */
    virtual void
    end();

protected:

    /**
     * Input collection name.
     */
    std::string _inputcolName;

    /**
     * Output collection name
     */
    std::string _outcolName;

    /**
     *  Set the persistency flag of the output collection.
     */
    bool _outputIsPersistent;

    /**
     * Maximum time value allowed in the tracker raw data.
     * The processor will cut all data after this value
     */
    int _maxTimeBin;

    /**
     * Number of time bins at the start of the collection window which would correspond to data
     * collected at negative drift lengths.
     * I.e. The DAQ could store some time bin before the trigger comes or the delays of the DAQ
     * could be set in such a way that the electrons produced at the anode of the TPC are not at
     * time bin 0.
     */
    int _startOffset;

    /**
     * Electron drift velocity (mm/us) used to compute the maximum time necessary to drift through the TPC
     */
    float _driftVelocity;

    /**
     * Maximum TPC drift length (mm). The length of the TPC used to calculate the maximum drift time
     */
    float _driftLength;

    /**
     * Readout frequency of the electronics (MHz)
     */
    float _readoutFreq;

    /**
     * Safety Offset to be added at the Maximum drift time calculated
     */
    int _endOffset;

    /**
     * Name of the collection containing the TPC Conditions
     * i.e. drift velocity
     */
    std::string _TPCConditionsCollName;

    /**
     * Name of the collection containing the GenericADCElectronicsParameters
     * i.e. readout frequency
     */
    std::string _ADCElectronicsCollName;

    /**
     * Boolean selector that specifies whether to shorten the pulses started before the filter time but continuing afterwards
     */
    bool _shortenPulse;

private:
    int _nRun;
    int _nEvt;

    TPCConditionsListener* _TPCConditionsListener;
    ADCElectronicsListener* _ADCElectronicsListener;

}; //End class definition
} //End Namespace definition
#endif /* RAWDATACUTTERPROCESSOR_H_ */
