/*
 * TPCDistortionsMapListener.h
 *
 *  Created on: Apr 3, 2014
 *      Author: zenker
 */

#ifndef TPCDistortionsMapListener_h
#define TPCDistortionsMapListener_h 1

//C++
#include <string>
#include <vector>
#include <utility>

// CLHEP
#include <CLHEP/Vector/TwoVector.h>

//LCCD
#include "lccd.h"
#include "lccd/DBInterface.hh"
#include "lccd/IConditionsChangeListener.hh"
#include "lccd/IConditionsHandler.hh"

//Marlin
#include "marlin/ConditionsProcessor.h"

//MarlinTPC
#include "TPCDistortionMeasurement.h"
#include "PhotoelectricDistortionsCalculator.h"

// the package should be capsulated in a name space
namespace marlintpc
{
/**
 * Helper Class which provides access to the distortions information
 * produced with the photodot system. In fact the data stored and accessed
 * here is only the true photodot position and the reconstructed photodot
 * position.
 * This class is derived from IConditionsChangeListener,
 * so the distortions are updated every time the conditions data change.
 * In principle this should only happen per run, since the distortions should
 * be measured for 0T and 1T and than be attached to runs with 0T and 1T respectively.
 * @author K. Zenker (DESY)
 */
class TPCDistortionsMapListener : public lccd::IConditionsChangeListener{

public:
  /**
   * Constuctor
   * \param ColName name of the collection in which the TPCDistortionMeasurement is stored
   */
  TPCDistortionsMapListener(std::string ColName = tpcconddata::TPCDistortionMeasurement::getDefaultColName());

  ///Destructor
  virtual ~TPCDistortionsMapListener();

  /// Returns the default name of the conditions data collection.
  static std::string getDefaultColName()
  {
      return  tpcconddata::TPCDistortionMeasurement::getDefaultColName();
  };

  /** Called by the Marlin ConditionsProcessor, if the condition change
   *  inplemetentation of the IConditionsChangeListener interface
   */
  void conditionsChanged(EVENT::LCCollection* col);

  /// Implementation of the listener class, takes care of default collection settings and registration of the handler.
  void registeredWithHandler(lccd::IConditionsHandler* condHandler);

  /// Implementation of the listener class, cleans up after deregistration and gives warning message.
  void deRegisteredWithHandler(lccd::IConditionsHandler* condHandler);

  /// Get TPCConditions Object
  /** Returns the stored \link tpcconddata::TPCDistortionMeasurement TPCDistortionMeasurement object \endlink
   *  @return \link tpcconddata::TPCDistortionMeasurement TPCDistortionMeasurement object \endlink
   */
  std::vector<tpcconddata::TPCDistortionMeasurement*> getTPCDistortionsMap();

  /** Return the true position of the photodot with the given id.
   * If the position of this photodot could not be reconstructed it will not be in the
   * TPCDistortionMeasurement object. In this case an lccd::LCCDException is thrown.
   */
  CLHEP::Hep3Vector getTruePosition(int photodot_id) const;

  /** Return the reconstructed position of the photodot with the given id.
   * If this photodot could not be reconstructed it will not be in the
   * TPCDistortionMeasurement object. In this case an lccd::LCCDException is thrown.
   */
  CLHEP::Hep3Vector getReconstructedPosition(int photodot_id) const;

  /**Returns true photodot positions for photodots that could be reconstructed.
   */
  std::vector<std::pair<int,CLHEP::Hep3Vector> > getTruePositions() const;

  /**Returns reconstructed photodot positions.
   */
  std::vector<std::pair<int,CLHEP::Hep3Vector> > getReconstructedPositions() const;

  /**Returns the estimated shift of a given point by the field distortions stored in the data base.
   * If this is not possible an InvalidDataException will be thrown.
   * \param measuredPosition The measured position in the global coordinate system.
   * \param moduleInterpolation If false only dots from the module with the measured point are considered.
   * \return The expected shift due to field distortion in the local coordinate system!
   * \remark Using the moduleInterpolation is not recommended since the shift at the module boundaries can be
   *         wrong in case of field distortions at the module boundary.
   */
  gear::Vector2D getEstimatedShift2D(const CLHEP::Hep3Vector measuredPosition, bool moduleInterpolation = false);

  /// Get the validity start time of the current conditions collection or default collection.
  lcio::long64 collectionValidSince();
  /// Get the validity end time of the current conditions collection or default collection.
  lcio::long64 collectionValidTill();

  /// Set if an exception should the thrown in case of no data available in the database.
  void useLastValidIfNoDataAvailable(bool useLastValid);
  /// Return if currently data is available in the database/file or not.
  bool noConditionsDataAvailable();

private:
  /// Holds the name of the conditions data collection
  std::string _collectionName;

  /// Default collection that is registered and will be returned from the handler if no data is available.
  lcio::LCCollectionVec* _thisEmptyCollection;

  /// Map of registered default collections for differnet handlers. Will hold here only one entry, since this listener registers itself with only one handler.
  std::map<lcio::LCCollection* , lccd::IConditionsHandler* > _handlerDefaultCollectionMap;

  /// Internal member to hold the "since" of the validity range of the current conditions collection.
  lcio::long64 _timeStampSince;
  /// Internal member to hold the "till" of the validity range of the current conditions collection.
  lcio::long64 _timeStampTill;

  /// Used to set if the last valid collection should be used if no current data are in the database.
  bool _useLastValidIfNoDataAvailable;
  /// Holds the indormation, if there are no current data available in the database.
  bool _noDataAvailable;

  /// Pointer to hold the address of the handler this listener is registered with.
  lccd::IConditionsHandler* _condHandler;

  /// Pointer to the TPC conditions object
  std::vector<tpcconddata::TPCDistortionMeasurement*> _tpcDistortionsMap;

  /// Pointer to the object used to calculate distortions
  PhotoelectricDistortionsCalculator* _calculationTool;
};

}
#endif
