#include "CountsToPrimaryElectronsProcessor.h"

// C++
#include <iostream>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerPulse.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/LCFlagImpl.h>

// Marlin(TPC)
#include <marlin/Exceptions.h>
#include <LCObjectCopier.h>

// tpcconddata
#include "TPCCondDataExeptions.h"

using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

CountsToPrimaryElectronsProcessor::CountsToPrimaryElectronsProcessor() : Processor("CountsToPrimaryElectronsProcessor")
{
    /* description of the processor which will be displayed in the
     * steering file autometicly generated by
     * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
     */
    _description = "CountsToPrimaryElectronsProcessor converts FADC counts into primary electrons, the conversion factors are provided by LCCD. Gain corrections should be done in the GainCorrectorProcessor";

    /* register steering parameters: name, description, class-variable, default value
     * the type will by definded by the type of the default value
     * string, double, float and int are possible
     * plaese note to get a float you must cast in C++: (float) 1.3
     */
    registerInputCollection(LCIO::TRACKERPULSE ,
                            "InputCollectionName",
                            "Name of the input collection as stored in the LCIO file",
                            _inputColName,
                            std::string("TPCPulses"));

    registerOptionalParameter(LCIO::TRACKERPULSE ,
                              //                               " OutputCollectionName",
                              "Name of the collection to be stored in the LCIO file",
                              _outputColName,
                              std::string("TPCPulses"));

    registerProcessorParameter("WriteOutputToStorage",
                               "if true the output collection is written",
                               _outputIsPersistent,
                               true);

    registerOptionalParameter("TPCConditionsCollection" ,
                              "Name of the collection containing the TPC conditions",
                              _tpcConditionsCollectionName,
                              tpcconddata::TPCConditions::getDefaultColName());

    registerOptionalParameter("MissingConditionsDataOverrideTimespan",
                              "The timespan in seconds for which the last valid collection will be used if no conditions data are available via LCCD.",
                              _missingConditionsDataOverrideTimespan,
                              (int)0);

}

void CountsToPrimaryElectronsProcessor::init()
{

    // usually a good idea to
    printParameters();

    _number_of_runs = 0;
    _number_of_events = 0;

    // create the TPC conditions listener
    _tpcConditionsListener = new TPCConditionsListener(_tpcConditionsCollectionName);
    if (parameterSet("MissingConditionsDataOverrideTimespan"))
        {
            _missingConditionsDataOverrideLCIOTime = 1000000000 * (lcio::long64)(_missingConditionsDataOverrideTimespan);
            streamlog_out(WARNING) << _processorName << ": Processing will go on using the last valid conditions collection even if there are no data available in the database for " << _missingConditionsDataOverrideTimespan << " seconds!" << std::endl;
            //tell the listeners that we want to override
            _tpcConditionsListener->useLastValidIfNoDataAvailable(true);
        }
}

void CountsToPrimaryElectronsProcessor::processRunHeader(lcio::LCRunHeader* run)
{

    _number_of_runs++ ;
    run->parameters().setValue(_processorName + "_revision", "$Rev: 2598 $");
    for (marlin::ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
                }
        }
}

void CountsToPrimaryElectronsProcessor::processEvent(lcio::LCEvent * evt)
{
    // Get input collection and check if there are data in it
    // Cast the input to LCCollectionVec if we want to use the iterator or modify it
    IMPL::LCCollectionVec* inputColl = 0;
    try
        {
            inputColl = dynamic_cast<IMPL::LCCollectionVec*>(evt->getCollection(_inputColName));
            if (inputColl->size() == 0)
                throw DataNotAvailableException(std::string("Empty TrackerPulses collection")
                                                + _inputColName);
        }
    catch (EVENT::DataNotAvailableException& e)
        {
            std::cout << _processorName << ": Empty event " << evt->getEventNumber() << std::endl;
            throw marlin::SkipEventException(this);
        }

    streamlog_out(DEBUG) << "Processing " << inputColl->getNumberOfElements() << " tracker pulses. The conversion factor (ADC->prim.e) is " << _tpcConditionsListener->getADCtoPrimaryElectronsFactor() << ".\n";

    // The TrackerData collection is the "output".
    // If the input collection is writeable and no output collection name
    // is given we use the input collection
    IMPL::LCCollectionVec* outputColl = 0;

    // Test if the collection is writeable
    try
        {
            dynamic_cast<IMPL::TrackerPulseImpl*>(inputColl->getElementAt(0))->setTime(
                dynamic_cast<EVENT::TrackerPulse*>(inputColl->getElementAt(0))->getTime());
        }
    catch (lcio::ReadOnlyException& e)
        {
            // Check if output collection name is given
            if (! parameterSet("OutputCollectionName"))
                {
                    streamlog_out(ERROR) << "Error: Collection is not writeable." << std::endl;
                    streamlog_out(ERROR) << "You have to specify an OutputCollectionName." << std::endl;
                    streamlog_out(ERROR) << "Remember to adapt the input collection name of the next processor!" << std::endl;
                    throw lcio::ReadOnlyException(_inputColName + " not writeable");
                }
        }

    // Clone input if OutputCollectionName is given
    // (note that this is mandatory if input is writeOnly)
    if (parameterSet("OutputCollectionName"))
        {
            outputColl = new LCCollectionVec("TrackerPulse");
            for (IMPL::LCCollectionVec::iterator it = inputColl->begin(); it < inputColl->end() ; it++)
                {
                    outputColl->addElement(lcio::LCObjectCopier::copy(dynamic_cast<IMPL::TrackerPulseImpl*>(*it)));
                }

            // set the transient flag as definded by the processor parameter
            outputColl->setTransient(!_outputIsPersistent);
        }
    else
        {
            outputColl = inputColl;
        }
    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1) ;
    outputColl->setFlag(trkFlag.getFlag());

    // convert ADC counts to primary electrons
    for (IMPL::LCCollectionVec::iterator it = outputColl->begin(); it < outputColl->end(); it++)
        {
            IMPL::TrackerPulseImpl* trackerPulse = dynamic_cast<IMPL::TrackerPulseImpl*>(*it);
            float oldCharge = trackerPulse->getCharge();
            trackerPulse->setCharge(oldCharge * _tpcConditionsListener->getADCtoPrimaryElectronsFactor());
            streamlog_out(DEBUG2) << " - ADC counts (before) =" << oldCharge << ", Primary Electrons (after) =" << oldCharge * _tpcConditionsListener->getADCtoPrimaryElectronsFactor() << std::endl;
        }

    // if the pulses collection has been created by cloning
    // we have to add it to the event
    if (parameterSet("OutputCollectionName"))
        {
            evt->addCollection(outputColl, _outputColName);
        }

    _number_of_events++;
}

void CountsToPrimaryElectronsProcessor::check(lcio::LCEvent *  /*evt*/)
{
    // nothing to check here - could be used to fill checkplots in reconstruction processor
}

void CountsToPrimaryElectronsProcessor::end()
{

}
}

