/*
 * ModuleDistortionCalculatorProcessor.cc
 *
 *  Created on: June 6, 2013
 *      Author: amunnich
 */

#include "ModuleDistortionCalculatorProcessor.h"
        
#include "marlin/Global.h"

#include "gear/GearMgr.h"
#include "gear/gearxml/GearXML.h"
#include "gear/TPCModule.h"
#include "gear/TPCParameters.h"
#include "EVENT/LCCollection.h"
#include "EVENT/Track.h"
#include "EVENT/TrackerPulse.h"
#include "EVENT/TrackerHit.h"

#include <fstream>
#include "TF1.h"
#include "TFile.h"

namespace marlintpc
{
  ModuleDistortionCalculatorProcessor aModuleDistortionCalculatorProcessor;

  ModuleDistortionCalculatorProcessor::ModuleDistortionCalculatorProcessor() : Processor("ModuleDistortionCalculatorProcessor")
  {
    _description = "Processor to analyze distortions for one run";

    registerInputCollection( LCIO::TRACK,
			     "TrackInputCollectionName",
			     "Name of the input track collection of final fit, default: FoundTracks",
			     _inputColNameTrackFit, 
			     std::string("FoundTracks"));

    registerOptionalParameter("HitInputCollectionName",
			     "Name of the input Track which hits to use, if different from fitted track",
			     _inputColNameTrackHits, 
			     std::string(""));

    registerOptionalParameter("DistortionXYOutputFileName",
                              "Output file name for distortions in xy",
                              _distortionXYOutputFileName,
                              std::string(""));

    registerOptionalParameter("DistortionZOutputFileName",
                              "Output file name for distortions in z",
                              _distortionZOutputFileName,
                              std::string(""));

    registerOptionalParameter("DistortionRowOutputFileName",
                              "Output file name for distortions along rows",
                              _distortionRowOutputFileName,
                              std::string(""));
 
    registerOptionalParameter("rootOutputFileName",
                              "Output file name for debug: store all histograms in root file",
                              _rootOutputFileName,
                              std::string(""));
  }

  void ModuleDistortionCalculatorProcessor::init()
  {
    const gear::TPCParameters* tpcParameters = &marlin::Global::GEAR->getTPCParameters();
    _GearParameters= &marlin::Global::GEAR->getTPCParameters();
    std::vector<gear::TPCModule*> TPCModules = tpcParameters->getModules();

    std::vector<gear::TPCModule*>::const_iterator moduleIter = TPCModules.begin();
    int moduletype0 = (*moduleIter)->getPadLayoutType();

    for( ; moduleIter != TPCModules.end() ; ++moduleIter)
      {
	if((*moduleIter)->getPadLayoutType() != moduletype0)
	  streamlog_out(MESSAGE) << "WARNING: Modules with different pad layouts are defined. Will use first one!"<< std::endl;
      }

    _myModuleDistortionCalculator=new ModuleDistortionCalculator( &marlin::Global::GEAR->getTPCParameters());
    _usedEvents=0;
    _fillCount=0;
  }

  void ModuleDistortionCalculatorProcessor::processRunHeader(lcio::LCRunHeader*  /*run*/)
  {
  }

  void ModuleDistortionCalculatorProcessor::processEvent(lcio::LCEvent* evt)
  {
    LCCollection* inputColFit;
    LCCollection* inputColHit;

    try
      {
	inputColFit = evt->getCollection(_inputColNameTrackFit);
      }
    catch (DataNotAvailableException &e)
      {
	streamlog_out(DEBUG) << "No data available in event for " <<_inputColNameTrackFit << std::endl;
	return;
      }
    if(parameterSet("HitInputCollectionName"))
      {
	try
	  {
	    inputColHit = evt->getCollection(_inputColNameTrackHits);
	  }
	catch (DataNotAvailableException &e)
	  {
	    streamlog_out(DEBUG) << "No data available in event for " <<_inputColNameTrackHits << std::endl;
	    return;
	  }
      }
    else //take the hits on the original track
      inputColHit =inputColFit;

    //we loop over the fitted tracks and then test all hits of the original found track
    //if the hits are not the one from the fitted track this might go wrong if there is more than one 
    //track in the event, so we skip these
    //Note that this can lead to reduced statistics if there is a significant amount of data with multiple tracks.
    bool eventOk=false;

    if(parameterSet("HitInputCollectionName"))
      {	
	if(inputColFit->getNumberOfElements()==1 && inputColHit->getNumberOfElements()==1)
	  eventOk=true;
      }
    else
      eventOk=true;

    if(eventOk)
      {
	_usedEvents++;
	for(int t=0;t<inputColFit->getNumberOfElements();t++)
	  {
	    lcio::Track *track=dynamic_cast<lcio::Track*>(inputColFit->getElementAt(t));
	    lcio::Track *trackhits=dynamic_cast<lcio::Track*>(inputColHit->getElementAt(t));
	    lcio::TrackerHitVec HitsinTrack = trackhits->getTrackerHits();
	    lcio::TrackerHitVec::const_iterator itHit = HitsinTrack.begin();

	    for(; itHit != HitsinTrack.end(); ++itHit)
	      {
		TrackerHit* hit = *itHit;
		//find out module and row of the hit. need pulses for that
		lcio::LCObjectVec PulseVect = hit->getRawHits();
		if(PulseVect.size() == 0)
		  {
		    std::cout << "WARNING: In Event " << evt->getEventNumber() << " for Hit " << ( itHit - HitsinTrack.begin() )
			      << ": The vector that should contain the pulse list is empty."
			      << "Impossible to decode row and module number " << std::endl;
		    continue;
		  }
		
		lcio::TrackerPulse* thePulse = dynamic_cast<lcio::TrackerPulse*>(*( PulseVect.begin() ));
		if(thePulse == 0)
		  {
		    std::cout << "WARNING: In Event " <<evt->getEventNumber()<< " for Hit on track "
			      << ( itHit - HitsinTrack.begin() ) << ": The raw hits do not contain valid TrackerPulse objects."
			      << "Impossible to decode row and module number " << std::endl;
		    continue;
		  }
		
		int currentModule = thePulse->getCellID1();
		int currentRow = 0;
		
		if(_GearParameters == 0)
		  {
		    currentRow = ( 0xffff0000 & ( thePulse->getCellID0() ) ) >> 16;
		  }
		else
		  {
		    const gear::TPCModule& theModule = _GearParameters->getModule(thePulse->getCellID1());
		    currentRow = theModule.getRowNumber(thePulse->getCellID0());
		  }
		
		std::pair<int, int> key(currentModule, currentRow);
	
		if(parameterSet("DistortionXYOutputFileName"))
		  {
		    std::map< std::pair<int,int>, TH1*>::iterator histoiter;
		    histoiter = _HistogramsXY.find(key);
		    
		    if(histoiter != _HistogramsXY.end())
		      {
			((*histoiter).second)->Fill( _myModuleDistortionCalculator->calculateDistanceXY(hit,track) );
		      }
		    else
		      {
			TH1* tempHist = new TH1F(Form("tempHistXY_%i_%i",currentModule,currentRow),"tempHistXY",200,-5,5);
			tempHist->Fill( _myModuleDistortionCalculator->calculateDistanceXY(hit,track) );
			_HistogramsXY.insert(std::make_pair(key,tempHist));
		      }
		  }
		if(parameterSet("DistortionZOutputFileName"))
		  {
		    std::map< std::pair<int,int>, TH1*>::iterator histoiter;
		    histoiter = _HistogramsZ.find(key);
		    
		    if(histoiter != _HistogramsZ.end())
		      {
			((*histoiter).second)->Fill( _myModuleDistortionCalculator->calculateDistanceZ(hit,track) );
		      }
		    else
		      {
			TH1* tempHist = new TH1F(Form("tempHistZ_%i_%i",currentModule,currentRow),"tempHistZ",200,-5,5);
			tempHist->Fill(_myModuleDistortionCalculator->calculateDistanceZ(hit,track) );
			_HistogramsZ.insert( std::make_pair(key,tempHist));
		      }
		  }
		if(parameterSet("DistortionRowOutputFileName"))
		  {
		    std::map< std::pair<int,int>, TH1*>::iterator histoiter;
		    histoiter = _HistogramsRow.find(key);
		    
		    try{
		      if(histoiter != _HistogramsRow.end())
			{
			  ((*histoiter).second)->Fill( _myModuleDistortionCalculator->calculateDistanceRow(hit,track) );
			  _fillCount++;
			}
		      else
			{
			  TH1* tempHist = new TH1F(Form("tempHistRow_%i_%i",currentModule,currentRow),"tempHistRow",200,-5,5);
			  tempHist->Fill(_myModuleDistortionCalculator->calculateDistanceRow(hit,track) );
			  _HistogramsRow.insert( std::make_pair(key,tempHist));
			  _fillCount++;
			}
		    }
		    catch(std::exception &e) {
		      std::cout<<"calculateDistanceRow failed. Value ignored!"<<std::endl;
		    }
		  }
	      }//hit loop
	  }//track loop
      }//event ok
  }
  
  void ModuleDistortionCalculatorProcessor::check(lcio::LCEvent*  /*evt*/)
  {
  }

  void ModuleDistortionCalculatorProcessor::end()
  {
 
    //all histograms are filled
    //now we need to fit them
    std::cout<<"Used Events: "<<_usedEvents<<", fill counter="<<_fillCount<<std::endl;
   if(parameterSet("DistortionXYOutputFileName"))
      {
	std::map< std::pair<int,int>, TH1*>::iterator histoiter = _HistogramsXY.begin();
	for(; histoiter!=_HistogramsXY.end(); histoiter++)    
	  {
	    std::pair<int,int> key = (*histoiter).first;
	    ((*histoiter).second)->Fit("gaus","Q");
	    std::pair<double,double> distortions;
	    if(((*histoiter).second)->GetFunction("gaus")!=0)
	      distortions = std::make_pair(((*histoiter).second)->GetFunction("gaus")->GetParameter(1),
					   ((*histoiter).second)->GetFunction("gaus")->GetParError(1));
	    else
	      distortions =  std::make_pair(0.0,0.0);

	    _ResultsXY.insert(std::make_pair(key,distortions));
	  }
	//results are available, now write them into given file
	WriteDistortionFiles(_distortionXYOutputFileName, _ResultsXY);
	
      }
    if(parameterSet("DistortionZOutputFileName"))
      {
	std::map< std::pair<int,int>, TH1*>::iterator histoiter = _HistogramsZ.begin();
	for(; histoiter!=_HistogramsZ.end(); histoiter++)    
	  {
	    std::pair<int,int> key = (*histoiter).first;
	    ((*histoiter).second)->Fit("gaus","Q");
	    std::pair<double,double> distortions;
	    if(((*histoiter).second)->GetFunction("gaus")!=0)
	      distortions = std::make_pair(((*histoiter).second)->GetFunction("gaus")->GetParameter(1),
					   ((*histoiter).second)->GetFunction("gaus")->GetParError(1));
	    else
	      distortions =  std::make_pair(0.0,0.0);

	    _ResultsZ.insert(std::make_pair(key,distortions));
	  }
	//results are available, now write them into given file

	WriteDistortionFiles(_distortionZOutputFileName,_ResultsZ);

      }
    if(parameterSet("DistortionRowOutputFileName"))
      {
	std::map< std::pair<int,int>, TH1*>::iterator histoiter = _HistogramsRow.begin();
	for(; histoiter!=_HistogramsRow.end(); histoiter++)    
	  {
	    std::pair<int,int> key = (*histoiter).first;
	    ((*histoiter).second)->Fit("gaus","Q");
	    std::pair<double,double> distortions;
	    if(((*histoiter).second)->GetFunction("gaus")!=0)
	      distortions = std::make_pair(((*histoiter).second)->GetFunction("gaus")->GetParameter(1),
					   ((*histoiter).second)->GetFunction("gaus")->GetParError(1));
	    else
	      distortions =  std::make_pair(0.0,0.0);

	    _ResultsRow.insert(std::make_pair(key,distortions));
	  }
	//results are available, now write them into given file

	WriteDistortionFiles(_distortionRowOutputFileName, _ResultsRow);
	
      }
    if(parameterSet("rootOutputFileName"))
      WriteRootFile();
  }

  void ModuleDistortionCalculatorProcessor::WriteDistortionFiles(std::string filename,std::map< std::pair<int,int>, std::pair<double,double> > results)
  {
    std::map< std::pair<int,int>, std::pair<double,double> >::iterator distiter =  results.begin();
    
    std::ofstream outfile;
    outfile.open(filename.c_str());
    
    for(; distiter != results.end(); ++distiter)
      {
	int module = ((*distiter).first).first;
	int row = ((*distiter).first).second;
	double distortion = ((*distiter).second).first;
	double distortionerr = ((*distiter).second).second;

	outfile << Form("%i %i %.7f %.7f\n",module,row,distortion,distortionerr);
      }
    outfile.close();
  }
  
  void ModuleDistortionCalculatorProcessor::WriteRootFile()
  {
    TFile *f=new TFile(_rootOutputFileName.c_str(),"RECREATE");
    std::map< std::pair<int,int>, TH1*>::iterator histoiter = _HistogramsRow.begin();
    for(; histoiter!=_HistogramsRow.end(); histoiter++)    
      ((*histoiter).second)->Write();
    histoiter = _HistogramsXY.begin();
    for(; histoiter!=_HistogramsXY.end(); histoiter++)    
      ((*histoiter).second)->Write();  
    histoiter = _HistogramsZ.begin();
    for(; histoiter!=_HistogramsZ.end(); histoiter++)    
      ((*histoiter).second)->Write();  
    f->Write();
    f->Close();
}

}//end namespace
