#include "PedestalListener.h"

//C++
#include <utility>
#include <map>
#include <iostream>
#include <string.h>
#include <exception>
#include <sstream>

//lccd
#include <lccd.h>
#include <lccd/LCConditionsMgr.hh>
#include <lccd_exceptions.h>

//lcio
#include "lcio.h"
#include "EVENT/LCIO.h"
#include "UTIL/LCFixedObject.h"
#include "UTIL/LCTime.h"

//MarlinTPC
#include "TPCCondDataExeptions.h"

namespace marlintpc
{

PedestalListener::PedestalListener(std::string ColName)
{
    //initialize first the members, else it goes wrong
    _collectionName = ColName;
    _pedestalMap.clear();
    _thisEmptyCollection = new LCCollectionVec(LCIO::LCGENERICOBJECT);
    _thisEmptyCollection->parameters().setValue("CollectionName", "EmptyCollectionPedestalListener");
    _timeStampSince = 0;
    _timeStampTill = 0;
    _useLastValidIfNoDataAvailable = false;
    _noDataAvailable = true;
    _condHandler = NULL;
    // try to register with the Marlin conditions processor
    if (!(marlin::ConditionsProcessor::registerChangeListener(this, ColName)))
        {
            std::stringstream message;
            message << " PedestalListener couldn't register collection " << ColName << " with the ConditionsProcessor.";
            throw lccd::LCCDException(message.str());
        }
}



PedestalListener::~PedestalListener()
{
    delete _thisEmptyCollection;
}



void PedestalListener::conditionsChanged(EVENT::LCCollection* col)
{
    streamlog_out(DEBUG) << "PedestalListener::conditionsChanged()" << std::endl;

    // get collection time parameters and set internal members
    std::stringstream sstrSince(col->getParameters().getStringVal(lccd::DBSINCE));
    sstrSince >> _timeStampSince;
    lcio::LCTime timeSince(_timeStampSince) ;
    std::stringstream sstrTill(col->getParameters().getStringVal(lccd::DBTILL));
    sstrTill >> _timeStampTill;
    lcio::LCTime timeTill(_timeStampTill) ;
    streamlog_out(DEBUG) << "PedestalListener::conditionsChanged(): _timeStampSince=" << _timeStampSince << " _timeStampTill=" << _timeStampTill << std::endl;

    // look into the map to see if we have accepted this collection as a default
    std::map<lcio::LCCollection* , lccd::IConditionsHandler* >::iterator itDefaultCollectionMap = _handlerDefaultCollectionMap.find(col);

    // check if the collection is our default collection
    if (itDefaultCollectionMap != _handlerDefaultCollectionMap.end())
        {
            _noDataAvailable = true;

            streamlog_out(MESSAGE) << "PedestalListener::conditionsChanged(): default collection " << col->getParameters().getStringVal("CollectionName") << " sent " << std::endl;
            streamlog_out(DEBUG) << "PedestalListener::conditionsChanged(): address of default collection: " << col << std::endl;

            //give some information on the missing conditions data
            streamlog_out(DEBUG) << "PedestalListener::conditionsChanged(): CollectionName: " << _collectionName << " not available at this timestamp." << std::endl;
            streamlog_out(DEBUG) << "PedestalListener::conditionsChanged(): No data in database from: " << timeSince.timeStamp() << " (" << timeSince.year() << " " << timeSince.month() << " " << timeSince.day() << " " << timeSince.hour() << " " << timeSince.min() << " " << timeSince.sec() << " " << timeSince.ns() << ")" << std::endl;
            streamlog_out(DEBUG) << "PedestalListener::conditionsChanged(): No data in database till: " << timeTill.timeStamp() << " (" << timeTill.year() << " " << timeTill.month() << " " << timeTill.day() << " " << timeTill.hour() << " " << timeTill.min() << " " << timeTill.sec() << " " << timeTill.ns() << ")" << std::endl;

            //check what to do if there are not data available
            if (_useLastValidIfNoDataAvailable)
                {
                    // check if there has been a last valid collection
                    if (_condHandler->lastValidCollection() != NULL)
                        {
                            fillConditionsMap(_condHandler->lastValidCollection());
                        }
                    else
                        {
                            // no last valid collection available, so we bail out by throwing an exception
                            std::stringstream out;
                            out << "PedestalListener::conditionsChanged(): Tried to use last valid collection but found none. Exiting processing";
                            throw lccd::DataNotAvailableException(out.str());
                        }
                }
            else
                {
                    // no conditions data available, so we bail out by throwing an exception
                    std::stringstream out;
                    out << "PedestalListener::conditionsChanged(): No conditions data stored for current time stamp. Exiting processing";
                    throw lccd::DataNotAvailableException(out.str());
                }
        }
    else                     // it is not a default so we can do the updating
        {
            _noDataAvailable = false;

            streamlog_out(MESSAGE) << "PedestalListener::conditionsChanged(): new values available. Updating parameter map." << std::endl;
            streamlog_out(DEBUG) << "PedestalListener::conditionsChanged(): CollectionName: " << col->getParameters().getStringVal("CollectionName") << std::endl;
            streamlog_out(DEBUG) << "PedestalListener::conditionsChanged(): Valid Since: " << col->getParameters().getStringVal("DBSince") << std::endl;
            streamlog_out(DEBUG) << "PedestalListener::conditionsChanged(): Valid Till: " << col->getParameters().getStringVal("DBTill") << std::endl;

            fillConditionsMap(col);
        }
}



void PedestalListener::fillConditionsMap(lcio::LCCollection* col)
{
    // before clearing the map, delete stored objects to free memory
    PedestalMap_t::iterator pediter;
    for (pediter = _pedestalMap.begin(); pediter != _pedestalMap.end(); ++pediter)
    {
        delete (*pediter).second;
    }
    // clear map
    _pedestalMap.clear();


    int nEntries = col->getNumberOfElements();

    ChannelReadoutGroupPair_t channelID;
    tpcconddata::Pedestal*    Entry = 0;
    std::pair<PedestalMap_t::iterator, bool> iInsertReturn;
    streamlog_out(DEBUG) << "PedestalListener: update pedestals  (data for " << nEntries << " channels in collection)" << std::endl;

    // fill LCCD collection into the map(map(pedestal)) structure
    for (int iEntry = 0; iEntry < nEntries; ++iEntry)
        {
            Entry = new tpcconddata::Pedestal(dynamic_cast<LCGenericObject*>(col->getElementAt(iEntry)));
            channelID.first  = (Entry->getHardwareID()).first;
            channelID.second = (Entry->getHardwareID()).second;
            iInsertReturn = _pedestalMap.insert(std::make_pair((std::make_pair(channelID.first, channelID.second)), Entry));
            if (!iInsertReturn.second)
                {
                    streamlog_out(ERROR) << "PedestalListener: could not store Pedestal for hardwareID: (" << channelID.first << "," << channelID.second << ") in the pedestal map because there was already a value stored for this channel" << std::endl;
                }
        }                        //loop over all entries in the LCCollectionVec
}

tpcconddata::Pedestal* PedestalListener::getPedestal(int channelID, int readoutGroup) const
{
    return this->getPedestal(std::make_pair(channelID, readoutGroup));
}

tpcconddata::Pedestal* PedestalListener::getPedestal(ChannelReadoutGroupPair_t hardwareID) const
{
    PedestalMap_t::const_iterator result = _pedestalMap.find(hardwareID);
    if (result != _pedestalMap.end())
        {
            return dynamic_cast<tpcconddata::Pedestal*>((*result).second);
        }
    else
        {
            std::stringstream message;
            message << "No entry could be found for channel/readgroup: " << hardwareID.first << "/" << hardwareID.second << ".";
            throw tpcconddata::CondDataNotAvailableException(message.str());
        }
}



float PedestalListener::getPedestalValue(ChannelReadoutGroupPair_t hardwareID) const
{
    tpcconddata::Pedestal* Entry = getPedestal(hardwareID);
    return Entry->getPedestalValue();
}



float PedestalListener::getPedestalValue(int channelID, int readoutGroup) const
{
    ChannelReadoutGroupPair_t hardwareID(channelID, readoutGroup);
    return this->getPedestalValue(hardwareID);
}



float PedestalListener::getPedestalWidth(ChannelReadoutGroupPair_t hardwareID) const
{
    tpcconddata::Pedestal* Entry = getPedestal(hardwareID);
    return Entry->getPedestalWidth();
}



float PedestalListener::getPedestalWidth(int channelID, int readoutGroup) const
{
    ChannelReadoutGroupPair_t hardwareID(channelID, readoutGroup);
    return this->getPedestalWidth(hardwareID);
}



void PedestalListener::registeredWithHandler(lccd::IConditionsHandler* condHandler)
{
    _condHandler = condHandler;
    streamlog_out(MESSAGE) << "PedestalListener::registeredWithHandler(): registered: " << condHandler->name() << std::endl;
    streamlog_out(DEBUG) << "PedestalListener::registeredWithHandler(): register default collection:" << std::endl;
    // try to get the default collection
    try
        {
            LCCollection* col = condHandler->defaultCollection();

            if (! col)           // it will be null if none has so far been registered. So let's register ours
                {
                    condHandler->registerDefaultCollection(_thisEmptyCollection);
                    streamlog_out(DEBUG) << "PedestalListener::registeredWithHandler(): default collection registered; address: " << _thisEmptyCollection << std::endl;
                    // and put in the map for this handler
                    _handlerDefaultCollectionMap[_thisEmptyCollection] = condHandler;
                }
            //then the default handler was already registered, that's odd ... ;)
            else if (col == _thisEmptyCollection)
                {
                    streamlog_out(DEBUG) << "PedestalListener::registeredWithHandler(): default collection is already set to thisEmptyCollection:" << std::endl;
                }
            else                 // somebody has got there before us, let's see if we like the default
                {
                    // if it has the same CollectionName, we can use it, no problem ...
                    if (col->getParameters().getStringVal("CollectionName") == "EmptyCollectionPedestalListener")
                        {
                            streamlog_out(DEBUG) << "PedestalListener::registeredWithHandler(): I like your default ;)" << std::endl;
                            _handlerDefaultCollectionMap[col] = condHandler;
                        }
                    // if not, an exception is thrown because things should be handled in a uniform standardized way
                    else
                        {
                            streamlog_out(ERROR) << "PedestalListener::registeredWithHandler(): a non-standard empty default collection has already been registered with the conditions handler" << std::endl;
                            throw lccd::InconsistencyException("PedestalListener::registeredWithHandler() found a default collection that differs from the standard one has already been registered with the handler");
                        }
                }
        }
    // catch MemberNotImplemented exception, which happens, if a simple file is used as input
    catch (lccd::MemberNotImplementedException)
        {
            streamlog_out(MESSAGE) << "PedestalListener::registeredWithHandler: Could not register default collection. Seems simple file input is used for pedestal conditions." << std::endl;
        }
}



void PedestalListener::deRegisteredWithHandler(lccd::IConditionsHandler* condHandler)
{
    streamlog_out(WARNING) << "PedestalListener::deRegisteredWithHandler() deregistered with:" << condHandler->name() << std::endl;
    _condHandler = NULL;
}



lcio::long64 PedestalListener::collectionValidSince()
{
    return _timeStampSince;
}



lcio::long64 PedestalListener::collectionValidTill()
{
    return _timeStampTill;
}



void PedestalListener::useLastValidIfNoDataAvailable(bool useLastValid)
{
    _useLastValidIfNoDataAvailable = useLastValid;
}



bool PedestalListener::noConditionsDataAvailable()
{
    return _noDataAvailable;
}


}
