//lcio
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "IMPL/LCCollectionVec.h"
#include "UTIL/LCTime.h"

//MarlinTPC
#include "TPCCondDataExeptions.h"

#include "TPCDistortionsMapListener.h"

namespace marlintpc
{
TPCDistortionsMapListener::TPCDistortionsMapListener(std::string ColName){
  //initialize first the members, else it goes wrong
  _collectionName = ColName;
  _thisEmptyCollection = new LCCollectionVec(LCIO::LCGENERICOBJECT);
  _thisEmptyCollection->parameters().setValue("CollectionName", "EmptyCollectionTPCConditionsListener");
  _timeStampSince = 0;
  _timeStampTill = 0;
  _useLastValidIfNoDataAvailable = false;
  _noDataAvailable = true;
  _condHandler = NULL;
  _calculationTool = NULL;
  // try to register with the Marlin conditions processor
  if (!(marlin::ConditionsProcessor::registerChangeListener(this, ColName))){
    std::stringstream message;
    message << " TPCConditionsListener couldn't register collection " << ColName << " with the ConditionsProcessor.";
    throw lccd::LCCDException(message.str());
  }
}

TPCDistortionsMapListener::~TPCDistortionsMapListener(){
  if(_calculationTool)
    delete _calculationTool;
  delete _thisEmptyCollection;
}

void TPCDistortionsMapListener::conditionsChanged(EVENT::LCCollection* col){
  streamlog_out(DEBUG) << "TPCDistortionsMapListener::conditionsChanged()" << std::endl;
  // delete the latest calculation tool since distortions are updated
  if(_calculationTool)
    delete _calculationTool;
  // get collection time parameters and set internal members
  std::stringstream sstrSince(col->getParameters().getStringVal(lccd::DBSINCE));
  sstrSince >> _timeStampSince;
  lcio::LCTime timeSince(_timeStampSince) ;
  std::stringstream sstrTill(col->getParameters().getStringVal(lccd::DBTILL));
  sstrTill >> _timeStampTill;
  lcio::LCTime timeTill(_timeStampTill) ;
  streamlog_out(DEBUG) << "TPCDistortionsMapListener::conditionsChanged(): _timeStampSince=" << _timeStampSince << " _timeStampTill=" << _timeStampTill << std::endl;

  // look into the map to see if we have accepted this collection as a default
  std::map<lcio::LCCollection* , lccd::IConditionsHandler* >::iterator itDefaultCollectionMap = _handlerDefaultCollectionMap.find(col);
  // check if the collection is our default collection
  if (itDefaultCollectionMap != _handlerDefaultCollectionMap.end()){
    _noDataAvailable = true;

    streamlog_out(MESSAGE) << "TPCDistortionsMapListener::conditionsChanged(): default collection " << col->getParameters().getStringVal("CollectionName") << " sent " << std::endl;
    streamlog_out(DEBUG) << "TPCDistortionsMapListener::conditionsChanged(): address of default collection: " << col << std::endl;

    //give some information on the missing conditions data
    streamlog_out(DEBUG) << "TPCDistortionsMapListener::conditionsChanged(): CollectionName: " << _collectionName << " not available at this timestamp." << std::endl;
    streamlog_out(DEBUG) << "TPCDistortionsMapListener::conditionsChanged(): No data in database from: " << timeSince.timeStamp() << " (" << timeSince.getDateString() << ")" << std::endl;
    streamlog_out(DEBUG) << "TPCDistortionsMapListener::conditionsChanged(): No data in database till: " << timeTill.timeStamp() << " (" << timeTill.getDateString() << ")" << std::endl;

    //check what to do if there are not data available
    if (_useLastValidIfNoDataAvailable){
      // check if there has been a last valid collection
      if (_condHandler->lastValidCollection() != NULL){
        // there should be only one object in the collection:
        if (col->getNumberOfElements() > 1){
          std::stringstream out;
          out << "TPCDistortionsMapListener::conditionsChanged(): Found more than one object (#=" << col->getNumberOfElements() << ") in the conditions collection, although it should be only one.";
          throw lccd::InconsistencyException(out.str());
        } else {
          for(int distortionMapEntry = 0;
              distortionMapEntry < (_condHandler->lastValidCollection())->getNumberOfElements();
              distortionMapEntry++){
            _tpcDistortionsMap.push_back((dynamic_cast<tpcconddata::TPCDistortionMeasurement*>((_condHandler->lastValidCollection())->getElementAt(distortionMapEntry))));
          }
        }
      } else {
        // no last valid collection available, so we bail out by throwing an exception
        std::stringstream out;
        out << "TPCDistortionsMapListener::conditionsChanged(): Tried to use last valid collection but found none. Exiting processing";
        throw lccd::DataNotAvailableException(out.str());
      }
    } else {
      // no conditions data available, so we bail out by throwing an exception
      std::stringstream out;
      out << "TPCDistortionsMapListener::conditionsChanged(): No conditions data stored for current time stamp. Exiting processing";
      throw lccd::DataNotAvailableException(out.str());
    }
  } else {
    // it is not a default so we can do the updating
    _noDataAvailable = false;

    streamlog_out(MESSAGE) << "TPCDistortionsMapListener::conditionsChanged(): new values available. Updating parameter map." << std::endl;
    streamlog_out(DEBUG) << "TPCDistortionsMapListener::conditionsChanged(): CollectionName: " << col->getParameters().getStringVal("CollectionName") << std::endl;
    streamlog_out(DEBUG) << "TPCDistortionsMapListener::conditionsChanged(): Valid Since: " << col->getParameters().getStringVal("DBSince") << std::endl;
    streamlog_out(DEBUG) << "TPCDistortionsMapListener::conditionsChanged(): Valid Till: " << col->getParameters().getStringVal("DBTill") << std::endl;

    streamlog_out(DEBUG) << "TPCDistortionsMapListener::conditionsChanged(): Found: " << col->getNumberOfElements() << " in the DistortionsMap."<< std::endl;

    for(int distortionMapEntry = 0; distortionMapEntry < col->getNumberOfElements();distortionMapEntry++){
        _tpcDistortionsMap.push_back(new tpcconddata::TPCDistortionMeasurement(dynamic_cast<LCGenericObject*>(col->getElementAt(distortionMapEntry))));
    }
  }
}

void TPCDistortionsMapListener::registeredWithHandler(lccd::IConditionsHandler* condHandler)
{
    _condHandler = condHandler;
    streamlog_out(MESSAGE) << "TPCDistortionsMapListener::registeredWithHandler(): registered: " << condHandler->name() << std::endl;
    streamlog_out(DEBUG) << "TPCDistortionsMapListener::registeredWithHandler(): register default collection:" << std::endl;
    // try to get the default collection
    try {
      LCCollection* col = condHandler->defaultCollection();
      if (! col) {
        // it will be null if none has so far been registered. So let's register ours
        condHandler->registerDefaultCollection(_thisEmptyCollection);
        streamlog_out(DEBUG) << "TPCDistortionsMapListener::registeredWithHandler(): default collection registered; address: " << _thisEmptyCollection << std::endl;
        // and put in the map for this handler
        _handlerDefaultCollectionMap[_thisEmptyCollection] = condHandler;
      } else if (col == _thisEmptyCollection) {
        //then the default handler was already registered, that's odd ... ;)
        streamlog_out(DEBUG) << "TPCDistortionsMapListener::registeredWithHandler(): default collection is already set to thisEmptyCollection:" << std::endl;
      } else {
        // somebody has got there before us, let's see if we like the default
         // if it has the same CollectionName, we can use it, no problem ...
        if (col->getParameters().getStringVal("CollectionName") == "EmptyCollectionTPCDistortionsMapListener") {
          streamlog_out(DEBUG) << "TPCDistortionsMapListener::registeredWithHandler(): I like your default ;)" << std::endl;
          _handlerDefaultCollectionMap[col] = condHandler;
        } else {
          // if not, an exception is thrown because things should be handled in a uniform standardized way
          streamlog_out(ERROR) << "TPCDistortionsMapListener::registeredWithHandler(): a non-standard empty default collection has already been registered with the conditions handler" << std::endl;
          throw lccd::InconsistencyException("TPCDistortionsMapListener::registeredWithHandler() found a default collection that differs from the standard one has already been registered with the handler");
        }
      }
    } catch (lccd::MemberNotImplementedException &e) {
      // catch MemberNotImplemented exception, which happens, if a simple file is used as input
      streamlog_out(MESSAGE) << "TPCDistortionsMapListener::registeredWithHandler: Could not register default collection." << std::endl;
    }
}



void TPCDistortionsMapListener::deRegisteredWithHandler(lccd::IConditionsHandler* condHandler)
{
    streamlog_out(WARNING) << "TPCDistortionsMapListener::deRegisteredWithHandler() deregistered with:" << condHandler->name() << std::endl;
    _condHandler = NULL;
}



lcio::long64 TPCDistortionsMapListener::collectionValidSince()
{
    return _timeStampSince;
}



lcio::long64 TPCDistortionsMapListener::collectionValidTill()
{
    return _timeStampTill;
}



void TPCDistortionsMapListener::useLastValidIfNoDataAvailable(bool useLastValid)
{
    _useLastValidIfNoDataAvailable = useLastValid;
}



bool TPCDistortionsMapListener::noConditionsDataAvailable()
{
    return _noDataAvailable;
}

std::vector<tpcconddata::TPCDistortionMeasurement*> TPCDistortionsMapListener::getTPCDistortionsMap(){
  return _tpcDistortionsMap;
}

CLHEP::Hep3Vector TPCDistortionsMapListener::getTruePosition(int photodot_id) const{
  for(std::vector<tpcconddata::TPCDistortionMeasurement*>::const_iterator it = _tpcDistortionsMap.begin();
      it != _tpcDistortionsMap.end(); it++){
    if((*it)->getDotID() == photodot_id){
      return (*it)->getTruePosition();
    }
  }
  std::stringstream out;
  out << "TPCDistortionsMapListener::getTruePosition(): Could not find photodot id: " << photodot_id
      << "\n Could be that the id is wrong or this photodot was not reconstructed.";
  throw lccd::LCCDException(out.str());
}

CLHEP::Hep3Vector TPCDistortionsMapListener::getReconstructedPosition(int photodot_id) const{
  for(std::vector<tpcconddata::TPCDistortionMeasurement*>::const_iterator it = _tpcDistortionsMap.begin();
      it != _tpcDistortionsMap.end(); it++){
    if((*it)->getDotID() == photodot_id){
      return (*it)->getReconstructedPosition();
    }
  }
  std::stringstream out;
  out << "TPCDistortionsMapListener::getTruePosition(): Could not find photodot id: " << photodot_id
      << "\n Could be that the id is wrong or this photodot was not reconstructed.";
  throw lccd::LCCDException(out.str());
}

std::vector<std::pair<int,CLHEP::Hep3Vector> > TPCDistortionsMapListener::getTruePositions() const{
  std::vector<std::pair<int,CLHEP::Hep3Vector> > v;
  for(std::vector<tpcconddata::TPCDistortionMeasurement*>::const_iterator it = _tpcDistortionsMap.begin();
      it != _tpcDistortionsMap.end(); it++){
    v.push_back(std::make_pair((*it)->getDotID(),(*it)->getTruePosition()));
  }
  return v;
}

std::vector<std::pair<int,CLHEP::Hep3Vector> > TPCDistortionsMapListener::getReconstructedPositions() const{
  std::vector<std::pair<int,CLHEP::Hep3Vector> > v;
  for(std::vector<tpcconddata::TPCDistortionMeasurement*>::const_iterator it = _tpcDistortionsMap.begin();
      it != _tpcDistortionsMap.end(); it++){
    v.push_back(std::make_pair((*it)->getDotID(),(*it)->getReconstructedPosition()));
  }
  return v;
}

gear::Vector2D TPCDistortionsMapListener::getEstimatedShift2D(const CLHEP::Hep3Vector measuredPosition, bool moduleInterpolation){
  if(!_calculationTool){
    _calculationTool = new PhotoelectricDistortionsCalculator(_tpcDistortionsMap, moduleInterpolation);
  }
  return _calculationTool->getEstimatedShift2D(measuredPosition);
}
}
