#ifndef BIASCORRECTORPROCESSOR_H
#define BIASCORRECTORPROCESSOR_H

#include <marlin/Processor.h>
#include <lcio.h>

//C++
#include <fstream>

//GSL
#include <gsl/gsl_spline.h>

namespace marlintpc
{
  /* Processor to correct bias (i.e., residuals as a function of hit-position on pad)
   * 
   * Parameters:
   * 
   *    HitInputCollectionName    =  Name of the input hits collection to be corrected  [string]
   *    HitOutputCollectionName   =  Name of the output corrected hits collection  [string]
   *    BiasInputFileName_RPhi    =  R-Phi input file name  [string]
   *    BiasInputFileName_Z       =  Z input file name  [string] 
   *    SplineType                =  Type of spline to use. Options are 0 = Cubic spline, 1 = Polynomial, 2 = Akima (Default: 0)  [integer] 
   *    RelativeBias              =  Whether or not to use x_relative (i.e. (x_hit - x_pad) / pad_width )  [boolean] 
   * 
   * Author: P. Hayman, Carleton University
   */

  class BiasCorrectorProcessor : public marlin::Processor
  {
    public:

      virtual Processor*  newProcessor() { return new BiasCorrectorProcessor; }

      BiasCorrectorProcessor();

      virtual void init();

//FIXME: not used:     virtual void processRunHeader(lcio::LCRunHeader* run);

      virtual void processEvent(lcio::LCEvent* evt);

//FIXME: not used:      virtual void check(lcio::LCEvent* evt);

      virtual void end();

    protected:
      std::string _inputHits; ///< Name of the input track collection 
      std::string _outputHits; ///< Name of the output track collection 

    private:

      std::string _biasInputFileName_rphi;
      std::string _biasInputFileName_z;
      int  _spline; //What type of spline to use
      bool _relative;

      std::map< std::pair<int,int>, gsl_spline*> _interp_rphi; //Store the GSL interpolators
      std::map< std::pair<int,int>, gsl_interp_accel*> _interp_acc_rphi; //Store the associated "accelerator" objects

      std::map< std::pair<int,int>, gsl_spline*> _interp_z; //Store the GSL interpolators
      std::map< std::pair<int,int>, gsl_interp_accel*> _interp_acc_z; //Store the associated "accelerator" objects

      void fillData(std::ifstream& file, int biasType);
  };
}

#endif //BIASCORRECTORPROSECCOR_H
