#ifndef PRFGEM_H
#define PRFGEM_H 1

//LCIO
#include "lcio.h"
#include <EVENT/TrackerHit.h>
#include <EVENT/Track.h>

// GEAR
#include <gear/TPCParameters.h>

//ROOT
#include "TGraph.h"
#include "TF1.h"


namespace marlintpc
{
  /** Functionality to evaluate the pad response function for a hit.
   * This class is for GEM data where the PRF can be described as a gaussian.
   * Goals are to improve the position of the hit, enable the correction of
   * dead channels and channels in overrange.
   *
   ** @author A.Muennich, DESY
   */

class PadResponseFunctionGEM
{
 public:
  /// The constructor.
  PadResponseFunctionGEM(double defocussing, double diffusion, const gear::TPCParameters& gearparameter);
  ///returns true if a PRF correction was applied, returns false if KeepCentreOfGravityInfo was used
  bool  EvaluateHit(lcio::TrackerHit *hit, lcio::Track *track=NULL);
  ///PRF improved position in global coordinates. PRF acts in rphi or xy not in z, so this is a 2D position in global coordinates.
  double* GetPRFPosition();
  ///Hit charge given by the integral of the PRF
  double GetPRFCharge(){return _charge;}
  ///Error on the hit charge
  double GetPRFChargeError(){return _chargeErr;}
  ///Option to check if the fit failed. A way to distinguish between the reseon the PRF was not used, either because enough pads or fit failed
  bool GetPRFSucceeded(){return _fitOK;}
  ///Option to check if the fit failed. A way to distinguish between the reseon the PRF was not used, either because enough pads or fit failed
  void SetPRCOnly(int value){_PRConly=value;}


 protected:
  /** Calculate the width to be fixed for the PRF fit from information from the hit and/or track
   * based on a parametrisation.
   */
  void CalculateFixedWidth(lcio::TrackerHit *hit, lcio::Track *track=NULL);
  ///prepare hit data, transfer info to fit object
  void PrepareHitData(lcio::TrackerHit *hit);
   ///Carry out the fit
  void FitPRF(lcio::TrackerHit *hit, bool fixCharge=false);
  ///calculate only a position correction for the center of gravity based on parametrisation
  void CalculatePRC(lcio::TrackerHit *hit);
  ///In case a PRF is not needed we keep the information of the original hit via the centre of gravity
  void KeepCentreOfGravityInfo(lcio::TrackerHit *hit);


  ///First paramter for the parametrisation of expected width: diffusion within gem stack sigma_0^2 [mm^2]
  double _defocussing;
  ///Second paramter for the parametrisation of expected width: diffusion in drift region D^2 [mm]
  double _diffusion;
  /// width for the gaussian fit
  double _fixedWidth;
  ///position of the hit along the row in global coordinates
  double _position;
  ///position of the row of the hit in global coordinates, unaffected by PRF fit
  double _positionRow;
  ///position to give back
  double _prfpos[2];
  ///position of the row of the hit in local coordinates, unaffected by PRF fit
  double _positionRowLocal;
  ///charge of the hit as given by the integral
  double _charge;
  double _chargeErr;
  ///status of prf evaluation, true if fits have worked
  bool _status;
  ///status of prf fit
  bool _fitOK;
  ///object for fitting or calculation in local coordinates
  TGraph *_graph;
  ///max charge filled in graph
  double _maxCharge;
  ///max position in graph
  double _maxPos;
  ///range of graph
  double _min,_max;
  ///access to gear information
  const gear::TPCParameters& _tpcParameters;
  ///prf parametrisation for flat region
  TF1 *_PRFShift;
  ///prf parametrisation
  TF1 *_PRF;
  ///Switch to disable the fitting, just do the calculation of a correction (like in MultiFit)
  int _PRConly;

};

}// namespace marlintpc
#endif
