/*
 * PedSubRawDataFilterProcessor.h
 *
 *  Created on: Oct 11, 2011
 *      Author: caiazza
 */

#ifndef PEDSUBRAWDATAFILTERPROCESSOR_H
#define PEDSUBRAWDATAFILTERPROCESSOR_H

// C++
#include <string>

// LCIO
#include "lcio.h"
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>

// Marlin
#include "marlin/Processor.h"

namespace marlintpc
{

/**  Filter Processor for the Pedestal Subtracted Raw Data
 *
 *   Select events based on the number of pulses in it and the average amplitude
 *   (max adc value recorded) of those pulses
 *   Can be use to perform a preselection of the interesting events prior to
 *   reconstruction.
 *
 *  <h4>Input - Prerequisites</h4>
 *  a collection of TrackerRawData
 *
 *  <h4>Output</h4>
 *  Collection containing TrackerRawData.
 *
 * @param InputCollectionName Name of input raw data collection
 * @param Minimum number of pulses cut
 * @param Minimum average pulse amplitude cut
 *
 * @author S. Caiazza, DESY
 * @version $Id: PedSubRawDataFilter,v 1.0 2011/10/11 22:52:04 caiazza Exp $
 */

class PedSubRawDataFilterProcessor : public marlin::Processor
{
public:
    virtual Processor*
    newProcessor()
    {
        return new PedSubRawDataFilterProcessor;
    }

    /* the default constructor
     * here the processor parameters are registered to the marlin package
     * other initialisation should be placed in the init method
     */
    PedSubRawDataFilterProcessor();

    /* Called at the begin of the job before anything is read.
     * Use to initialize the processor, e.g. book histograms
     */
    virtual void
    init();

    /* Called for every run.
     * in this processor it is not used
     */
    virtual void
    processRunHeader(lcio::LCRunHeader* run);

    /* Called for every event - the working horse.
     * Here the real conversion take place
     */
    virtual void
    processEvent(lcio::LCEvent* evt);

    /* This method is only called if the check flag is set (default) in the main processor
     * (where the input files are given)
     * It should be used to create check plots which are not needed in a complete analysis
     * (where normaly the check flag is deactivated)
     */
    virtual void
    check(lcio::LCEvent* evt);

    /* Called after data processing for clean up.
     * e.g. saving contol histograms, delete created objects
     * (which are not stored in a collection/event), etc.
     */
    virtual void
    end();

protected:
    /**
     * The name the input collection is stored under
     */
    std::string _inputColName;

    /**
     * The name the output collection will be stored under
     */
    LCCollection* inputCol;

    /**
     * The minimum number of pulses to be present in an event to be kept
     */
    int _minNumPulses;

    /**
     * The average amplitude of the raw pulses in an event for it to be kept
     */
    float _minPeakAverage;

    int _nRun;    //<- counts the runs processed
    int _nEvt;    //<- counts the events processed

    bool _numPulsesOK; //<- true if the number of pulses in the event is higher than the minimum required
    bool _peakAverageOK; //<- true if the average amplitude of the pulses in the event is higher than the minimum required

};  // End class declaration
}// End namespace declaration
#endif
