#ifndef PULSEFINDER_H
#define PULSEFINDER_H

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "EVENT/TrackerData.h"
#include <IMPL/TrackerDataImpl.h>
#include "EVENT/TrackerPulse.h"

#include "PedestalListener.h"
#include "ADCElectronicsListener.h"

//C++
#include <string>

namespace marlintpc
{

  /** The PulseFinder class is working horse of the ModularPulseFinderProcessor. 
   *  It contains the finding and splitting algorithms which are used in the processor.
   *  This allows for unit testing of the methods which is not possible if they are members of 
   *  the processor, as processors cannot be instantiated outside of Marlin.

   *  First the pulses are searched using a threshold based method, 
   *  where the start and end thresholds are calculated from the pedestal widths and the 
   *  PedestalWidthFactor parameters.
   *  Then these candidates are further processed by splitting multiple pulses.
   *  In a last step quality flags are assigned to account for
   *    - multiple pulses (changed shape/charge information)
   *    - overflows, where the ADC values are saturated
   *    - anomalous shape, currently only pulses with a long plateau which has been cut off
   *    - whether a pulse has been split
   *
   *  There are different modes to calculate the pulse time or the pulse charge:
   *
   *  Pulse charge:
   *  \li SUMQ The sum of all ADC values on the pulse (the default value).
   *  \li FIT_PARABOLA Parabolic fit around the maximum of the pulse.
   *  \li MAXQ The maximal ADC value in the pulse.
   *
   *  Pulse time:
   *  \li INFELXION_POINT The infexion point on the rising slope (the default value).
   *       (In fact it is the COG of the derivatives. Is this really the infelxion point? MK)
   *  \li FIT_PARABOLA Parabolic fit around the maximum of the pulse.
   *  \li CENTRE_OF_GRAVITY The centre of grafity of all charge in the pulse.
   *  \li FIRST_ABOVE_THRESHOLD The first value abve the threshold. This method has a jitter and 
   *         a walk due to threshold effects
   *  \li AT_MAXIMUM Time of the sample with the maximumal value in the pulse.
   *  \li CONSTANT_FRACTION Signal crossing at a constant fraction of the maximal pulse height.
   *  \li GAMMA_FIT Fits a Gamma_4 function to the pulses
   *  \li COG_ATMAX Center of gravity around maximum.
   *
   *  parameter for time methods:  (default 0.9 for constant fraction)
   *  - fraction for constant fraction, 
   *  - no. of neighboring bins for center of gravity at maxbin
   *
   *  By default also the initial TrackerData (=ADC) spectrum is stored. By setting the 
   *  \c forceSectrumSave flag to false one can choose to store the spectrum only when one of the
   *  quality flags is set. This feature is always on.
   *
  * 
   * @author Ch. Rosemann, DESY (original author of the methods)
   * @author Martin Killenberg (implementation as a separate class)
   */
  class PulseFinder
  {

    public:

    /**
     * The constructor. 
     * @param minimumPulseHeight The minimum pulse height in adc count
     * @param minimumPulseLength The minimum pulse length in time samples
     *
     * @param pulseStartPedestalWidthFactor The start threshold is calculated as 
     *                                      pulseStartPedestalWidthFactor * pedestalWidth.
     *                                      @FIXME pedestal should be added if not already subtracted
     *
     * @param pulseEndPedestalWidthFactor   The end threshold is calculated as 
     *                                      pulseEndPedestalWidthFactor * pedestalWidth.
     *                                      @FIXME pedestal should be added if not already subtracted
     *
     * @param NoiseValuePedestalWidthFactor The noise value is calculated as 
     *                                      noiseValuePedestalWidthFactor * pedestalWidth.
     *
     * @param pedestalListener Pointer to the pedestal listener. If you do not have a pedestal
     *                         listener you can set it to NULL. In this case you have to provide
     *                         override parameters for pedestalValue and pedestalWidth.
     *
     * @param adcElectronicsListener Pointer to the electronics listener. If you do not have an
     *                               electronics listener you can set it to NULL.
     *                               In this case you have to provide override values for 
     *                               maximumADCValue, readoutFrequency, isPedestalSubtracted and
     *                               polarity.
     *
     * @param saveNBinsBeforeStart Store n bins before the first bin which is above
     *                             the start threshold (default 0).
     *
     * @param saveNBinsAfterEnd    Store n bins after the last bin which is above
     *                             the end threshold (default 0).
     *
     * @param timeCalculationMode  The mode used to calculate the pulse time 
     *                             (default INFELXION_POINT)
     *
     * @param timeCalculationParameter  parameter for time methods: fraction for constant fraction, no. of 		
     *					neighboring bins for center of gravity at maxbin
     *                             	(default 0.9 for constant fraction)
     *
     * @param chargeCalculationMode The mode used to calculate the pulse charge
     *                              (default SUMQ).
     *
     */

    PulseFinder( float minimumPulseHeight,
		 int minimumPulseLength,
		 float pulseStartPedestalWidthFactor,
		 float pulseEndPedestalWidthFactor,
		 float noiseValuePedestalWidthFactor,
		 PedestalListener const * pedestalListener,
		 ADCElectronicsListener const * adcElectronicsListener,
		 int saveNBinsBeforeStart = 0,
		 int saveNBinsAfterEnd = 0,
		 int timeCalculationMode = INFLEXION_POINT,
		 double timeCalculationParameter = 0.9,
		 int chargeCalculationMode = SUMQ
		 );

    //The modes for charge and time calculation
    static const int INFLEXION_POINT = 0; ///< Time is the inflexion point of the rising slope.
    static const int FIT_PARABOLA = 1;    ///< Time and/or charge is determined by parabolic fit.
    static const int CENTRE_OF_GRAVITY = 2; ///< Time is the center of gravity of the charge.
    static const int FIRST_ABOVE_THRESHOLD = 3; ///< Time is the first sample above the start threshold
    static const int AT_MAXIMUM = 4; ///< Time is the first sample at the maximum
    static const int CONSTANT_FRACTION = 5; ///< Time is at the constant fraction of the maximum pulse
    static const int GAMMA_FIT = 6; ///< Time from a Gamma_4 function fit to the pulse
    static const int COG_ATMAX = 7; ///< COG for 2 bin range around maximum
    static const int BOX = 8; ///< Box Method
    static const int GAUSSIAN_INFLEXION = 9; ///< Inflexion point of a gaussian fit


    static const int SUMQ = 0; ///< Charge is the sum of all ADC counts.
    // FIT_PARABOLA = 1 already defined in the time section.
    static const int MAXQ = 2; ///< Charge is the maximal ADC value.
    static std::string getRevisionNumber(); ///< Log the revision number

  protected:

    float _pulseStartPedestalWidthFactor; ///< factor to calculate the start threshold
    float _pulseEndPedestalWidthFactor; ///< factor to calculate the end threshold
    float _noiseValuePedestalWidthFactor; ///< factor to calculate noise value

    
    int _timeCalculationMode;  ///< The mode used for time calculation
    int _COGAtMaxParameter;///<parameters for time methods
    double _ConstFracParameter;///<parameters for time methods
    int _chargeCalculationMode; ///< The mode used for charge calculation

    float _minimumPulseHeight; ///< minimal value for the pulse maximum in adc counts
    int _minimumPulseLength;   ///< minimal length for a pulse in number of time bins
    int _saveNBinsBeforeStart; ///< # of bins stored before the start threshold
    int _saveNBinsAfterEnd;    ///< # of bins stored after the end threshold
    
    /// the recommended way of providing the access to the pedestal values:
    PedestalListener const * _pedestalListener;
    /// the recommended way to access the electronics parameters
    ADCElectronicsListener const * _adcElectronicsListener;
    
     // override variables

    // A set of variables for which values the override value should be used.
    bool _usePedestalValueOverride;
    bool _usePedestalWidthOverride;
    bool _use_isPedestalSubtractedOverride;
    bool _usePolarityOverride;
    bool _useReadoutFrequencyOverride;
    bool _useMaxADCValueOverride;


      /** maximum adc value, needed to determine if a pulse is maxed out
       */
    float _maximumADCValueOverride;

      /** The readout frequency of every readout group (in MHz). This overrides any information provided by LCCD
       */
    float _readoutFrequencyOverride;
    
    /** The default pedestal value which is used if the pedestalListener is not set
     */
    float _pedestalValueOverride;

    /** The default pedestal width which is used if the pedestalListener is not set
     */
    float _pedestalWidthOverride;

      /** Sets, if the data is already pedestal subtracted in the read-out electronics. This overrides any information provided by LCCD
       */
    bool _isPedestalSubtractedOverride;

      /** the polarity of all readout groups, overrides LCCD
         */
    int _polarityOverride;

  
    bool _forceSpectrumSave; ///< flag whether to always save the spectrum
    bool _switchOffSplitting; ///< flag whether to try to split pulses

    int _maxPlateauLength; ///<The maximum length of a plateau within a pulse in time bins. 


    /// Helper class to store the time and it's error
      class PulseTime
      {
        public:
          PulseTime() : _time(0.), _timeError(0.)
          {
            ;
          };
          void setTime(float t)
          {
            _time = t;
          };
          void setTimeError(float e)
          {
            _timeError = e;
          };
          float getTime()
          {
            return _time;
          };
          float getTimeError()
          {
            return _timeError;
          };
        private:
          float _time, _timeError;
      };

    /// Helper class to store the charge and it's error
      class PulseCharge
      {
        public:
          PulseCharge() : _charge(0.), _chargeError(0.)
          {
            ;
          };
          void setCharge(float t)
          {
            _charge = t;
          };
          void setChargeError(float e)
          {
            _chargeError = e;
          };
          float getCharge()
          {
            return _charge;
          };
          float getChargeError()
          {
            return _chargeError;
          };
        private:
          float _charge, _chargeError;
      };

      /// Helper class which combined PulseCharge and PulseTime
      class Pulse : public PulseCharge, public PulseTime
      {
        public:
          Pulse() : PulseCharge(), PulseTime()
          {
            ;
          };
          void set(PulseCharge c)
          {
            setCharge(c.getCharge());
            setChargeError(c.getChargeError());
          };
          void set(PulseTime t)
          {
            setTime(t.getTime());
            setTimeError(t.getTimeError());
          };
      };

      /** Helper class to make more sense to the splitting phase:
       *  actually keep track of the "split" flag
        */
      struct InternalSplitPulse
      {
        InternalSplitPulse(std::vector<float>::const_iterator s, 
			   std::vector<float>::const_iterator e, bool f)
        {
          startBin = s;
          endBin = e;
          isSplit = f;
        }
        std::vector<float>::const_iterator startBin;
        std::vector<float>::const_iterator endBin;
        bool isSplit;

        InternalSplitPulse()
        {
          ;
        };
      };

  public:
      /** Create TrackerPulses from the TrackerData input. It uses findPulseCandidates and composes
       *  TrackerPulse objects out of the TrackerData.
       *
       *  All TrackerPulses and the contained TrackerData objects have been created on the heap.
       *  Do not forget to store them in a collection or delete them to avoid memory leaks!
       *
       *  @attention When storing the pulses do not forget to store the tracker data objects
       *  they contain to a collection! Check for NULL before adding to the collection!
       */
      std::vector<TrackerPulse *> findPulses(TrackerData* inputData);
      
      /** This wrapper function handles the calculation of the pulse time and the pulse charge.
       *  They are combined into one function because it is faster to calculate both values at the same time if the fit method is used.
       *  It can also be modified to add new methods of calculating these values without modifying the main event loop.
       *  New addition: errors on all the values!
       *  Convention: first time and time error, then charge and charge error
      @param data the pulse data
      */
      Pulse calculatePulseTimeAndCharge(const EVENT::TrackerData *tracker_data);


      // the internal functions
  protected:
      /** Find pulses by threshold in the ADC information and closer inspection.
      *  This is the somewhat "master" function that determines all aspects of the pulses,
      *  it calls some further functions for easier reading.
      *  @param thisData 'raw' data as TrackerData
      *  @param tempPulse a helper data construct, a pair of the TrackerData of a pulse and the quality bit
      *  @param noiseValue the width of the pedestal
       */
      void findPulseCandidates(lcio::TrackerData* thisData, 
			       std::vector<std::pair<lcio::TrackerDataImpl*, int> >* tempPulse);



      /** The first function that is called during the pulse finding,
      it simply looks for values above the start threshold and below the end thresholds,
      the start and end positions are stored by their iterators in the original TrackerData ADC value vector.
      Here already pulses which are too short are suppressed.
      @param ADCSpectrum the ADC spectrum to be searched
      @param theCandidatePulses a vector of the pairs of start and end positions of the pulse candidates (as iterators)
      */

      void brutePulseSearch(TrackerData const * inputTrackerData,
                            std::vector<std::pair<std::vector<float>::const_iterator,
			    std::vector<float>::const_iterator> >* theCandidatePulses);



      /** The second part in the pulse finding process.
      Check the pulse candidates for possible multiple pulses by looking more closely at the ADC spectrum.
      Then replace the /old/ pairs of start/end iterators with the appropriate new number of pairs.
      TODO: adjust the charges in the bins at the joined boundary of the split pulses
      @param theCandidatePulses a vector of the pairs of start and end positions of the pulse candidates (as iterators)
      @param noiseThreshold Threshold to avoid splitting at minima caused by noise on flat signals.
                            The threshold has to be given here because it cannot be determined from 
			    the cellID as only iterators are given to this function.
      */

      std::vector<InternalSplitPulse> splitMultiplePulses(
		std::vector< std::pair< std::vector<float>::const_iterator,
		                        std::vector<float>::const_iterator> >* theCandidatePulses,
		float noiseThreshold);


   

      /** This function calculates the pulse time and charge by fitting a parabola to the log of the data surrounding the extreme point.
      @param data the pulse data
      */
      Pulse calculatePulseWithFit(const EVENT::TrackerData *tracker_data);

      /** This function calculates the pulse time and charge by fitting a Gamma_n to the pulse
      @param data the pulse data
            */
      PulseTime calcPulseTimeWithGammaFit(const lcio::TrackerData *thisPulse);

      /** Calculate the time information as the inflextion point of the rising slope.
       *  This is the  default way.
       */
      PulseTime calcPulseTimeInflexionPoint(const lcio::TrackerData* thisPulse);

     /** Calculate the time information as the time of the bin with pulse maximum.
      */
      PulseTime calcPulseTimeAtMaximum(const lcio::TrackerData* thisPulse);

      /** Calculate the time information as the time of crossing the threshold set by a fraction of the max. charge in a pulse.
       */
      PulseTime calcPulseTimeAtConstantFraction(const lcio::TrackerData* thisPulse);

     /** Calculate the time information as centre of gravity.
      */
      PulseTime calcPulseTimeCentreOfGravity(const lcio::TrackerData* thisPulse);

      /** Calculate the time information as the cog for 2 bin range around the maximum.
      */
      PulseTime calcPulseTimeCOGAtMaximum(const lcio::TrackerData* thisPulse);

      /** Calculate the time information as the first bin above threshold.
      */
      PulseTime calcPulseTimeFirstBinAboveThreshold(const lcio::TrackerData* thisPulse);

      /** Calculate the time information from the box method.
            */
      PulseTime calcPulseTimeBoxMethod(const lcio::TrackerData* thisPulse);

      /** Calculate the time information from the inflexion point of a gausssian fit.
            */
      PulseTime  calcuPulseTimeGaussianInflexion(const lcio::TrackerData* thisPulse);
      /** Calculate the integrated charge of the given pulse by summing up all bins of the ADC spectrum.
       *  The polarity of the pulse is taken into acount
       *  @param thisPulse ADC spectrum for charge calculation
       */
      PulseCharge calcPulseChargeSumQ(const lcio::TrackerData* thisPulse);

      /** Use the maximum value as pulse charge.
       *  FIXME: Does not use polarity. I think negative polarity does not work anyway from reading
       *  the rest of the code. (MK)
       */
      PulseCharge calcPulseChargeMaxQ(const lcio::TrackerData* thisPulse);


      /** Returns the pedestal value for a given TrackerData / TrackerPulse object taking into consideration whether it should be overwritten by the override parameter.
       */
      float getPedestalValue(const int hardware_channel, const int readout_group);



      /** Returns the pedestal width for a given TrackerData / TrackerPulse object taking into consideration whether it should be overwritten by the override parameter.
       */
      float getPedestalWidth(const int hardware_channel, const int readout_group);



      /** Returns if the data has been pedestal subtracted in the read-out electronics or not
       */
      bool isPedestalSubtracted(const int readout_group);



      /** Returns the polarity of a TrackerData / TrackerPulse object taking into consideration whether is should be overwritten by the override parameter.
       */
      int getPolarity(const int readout_group);



      /** Returns the readout frequency of a readout group (in Mhz)
       *  taking into consideration whether it should be overwritten by the override parameter.
       */
      float getReadoutFrequency(const int readout_group);



      /** Returns the maximum ADC value of a readout group taking into consideration
       *  whether it should be overwritten by the override parameter.
       */
      float getMaxADCValue(const int readout_group);
  public:

      /** Set an overide value for the maximal ADC value.
       */
      void setMaxADCValueOverride(float maxADCValue);

      /** Set an overide value for the pedestal value (in ADC counts).
       */      
      void setPedestalValueOverride(float pedestalValue);

      /** Set an overide value for the pedestal width (in ADC counts).
       */
      void setPedestalWidthOverride(float pedestalWidth);
      
      /** Set an override value for the readout frequency in MHz.
       */
      void setReadoutFrequencyOverride(float readoutFrequency);

      /** Set an override value whether the pedestal has been subtracted.
       */
      void set_isPedestalSubtractedOverride(bool isPedestalSubtracted);

      /** Set an override value for the polarity. It should be +1 or -1.
       */
      void setPolarityOverride(int polarity);

      /** Flag whether the spectrum is always stored. By default this only happend if the
       *  quality word is not 0.
       */
      void setForceSpectrumSave(bool doSave = true);
 
      /** Flag whether to try to split pulses.
       */
      void setSwitchOffSplitting(bool doNotSplit = false);

      /** Set he maximum length of a plateau allowed within a pulse in time bins. 
       *  This used to cut off strange pulses that go into a plateau.
       */
      void setPlateauCutOff(int maxPlateauLength);

 // declare the tester as friend so we can unit-test the protected member functions
      friend class PulseFinderTester;
  };
}
#endif // PULSEFINDER_H
