#ifndef ROWCUTAPPLICATIONPROCESSOR_H
#define ROWCUTAPPLICATIONPROCESSOR_H

// C++
#include <string>
#include <set>

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "marlin/EventModifier.h"

namespace marlintpc
{

/**
 * \section Introduction
 * This processor is used to select in a hit collection those hits that are not formed by pulses
 * contained in a custom set of rows or modules
 * This processor is meant to be used in the reconstruction chain of data from row based readout
 *
 * @author Martin Ljunggren (Lund University), Stefano Caiazza (DESY)
 * @version $Id: RowCutApplicationProcessor.h 2993 2010-02-13 20:50:00Z caiazza $
 *
 * \section Input
 * @param InputCollectionName: A collection of TrackerHit containing the uncut data sample. Default: TPCHits
 * @param OutputCollectionName: A collection of TrackerHit where the output is going to be stored. Default: TPCHitsCut
 * @param SetOutputTransient (Optional): if true the output will not be stored to disk. Default: false
 * @param SkipRow: A pair of integers, separated by spaces identifying the module and row to select.
 * A negative number indicates any row and/or any module. For example -1 -1 signals to remove all rows from all modules
 * The user can use this parameter an arbitrary number of times in each processor. Default: -1 -1
 * @param isSubset: Is a boolean to specify whether the output collection is to be a subset of the input
 * (only pointers to the original data are stored). If this value is false the algorithm assumes only the
 * flagging mechanism and no output collection will created
 * @param ActivateFlagging: Is a boolean specifying whether to activate the flagging mechanism. If this parameter
 * is set to be true the original input collection will be modified adding a flag of type RemoveFromTrackFinding
 * which should be interpreted by the following processor accordingly. WARNING: Setting this parameter to be true
 * WILL MODIFY THE INPUT COLLECTION
 *
 */

class RowCutApplicationProcessor : public marlin::Processor
{

public:

    /**
     * Creates and register a new processor of type RowCutApplicationProcessor
     */
    virtual Processor*
    newProcessor()
    {
        return new RowCutApplicationProcessor;
    }

    /**
     * Default constructor
     * Used to create objects and register the processor parameters
     */
    RowCutApplicationProcessor();

    /**
     * Variable initialization
     */
    virtual void
    init();

    /**
     * Method to modify the RunHeader of the LCIO file
     */
    virtual void
    processRunHeader(lcio::LCRunHeader* run);

    /**
     * All the processing is done here
     */
    virtual void
    processEvent(lcio::LCEvent * evt);

 
    /**
     * @return Returns a string with the name of the processor
     */
    virtual const std::string &
    name() const {return Processor::name();}

protected:

    std::string _inputColName;  ///< the name of the input collection
    std::string _outputColName;  ///< the name of the output collection

    bool _outputIsTransient;///< if this flag is true the output will not be written to file

    IntVec _rows2skip;///< Vector of rows to skip

    bool _isSubset; ///< If true the output collection will be stored as a subset of the input containing only the "good" hits. Otherwise no output collection will be created
    bool _flaggingON;      ///< If true, hits in rows marked for removal will be flagged in the input collection modifying the input collection

    std::set<std::pair<int,int> > _skipList; ///< The list of the rows to skip
    static const std::pair<int, int> ALLROWS; ///< The pair that identifies the default value of the SkipRow parameter and the special combination meaning "all rows and modules"

};

}

#endif
