#ifndef TOPO_FINDER_H
#define TOPO_FINDER_H 1

#include <map>
#include <list>

#include <lcio.h>
#include <EVENT/TrackerHit.h>
#include <EVENT/TrackerPulse.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>
#include <gearimpl/GlobalPadIndex.h>

namespace marlintpc
{

/** TopoFinder contains finding algorithms for clusters of TrackerPulses.
 *  It operates on a std::multimap<int, TrackerPulse*> which contains
 *  the pulses to search in a multimap sorted by the corresponding pad index.
 *  \attention !! Be careful with the iterators! They are deleted from the map
 *     and are invalid afterwards!!
 *
 *  @author Martin Killenberg, University of Bonn
 *
 *  This class is not a processor, but called from them.
 */

class TopoFinder
{
	public:
		virtual ~TopoFinder() {}

    protected:

        /// The constructor.
        TopoFinder();

//    void setPulseMap(std::multimap<int, EVENT::TrackerPulse*> *map);

        /** Find the hit which contains the given pulse.
         *  It looks for pulses in neighbouring pads.
         */
        EVENT::TrackerHit* findHit(
            std::multimap<gear::GlobalPadIndex, EVENT::TrackerPulse*>::iterator currentPulseIter);

        /** Iteratively searches for pulses that are on neighbouring pads.
         *  Erase the iterator of this pulse from the global map before calling this function to avoid
         *  finding it twice.
         *  The reference to the module is passed so it does not have to be retrieved over and over again...
         */
        std::list<EVENT::TrackerPulse *> findNeighbouringPulses(
            gear::GlobalPadIndex currentPadIndex,
            EVENT::TrackerPulse *  currentPulse,
            const gear::TPCModule &module);

        /** Find Hits in the neighbouring rows. If empty rows are allowed this does not necessarrily
         *  mean next neighbouring rows. (Empty row means there is no hit in this row which matches the
         *  given hit. It does not have to be completely empty.)
         */
        EVENT::TrackerHitVec *findHitsInNeighbouringRows(EVENT::TrackerHit* hit, int moduleID);

        /** Find all the hits in the given pad range. There might be more than one hit in
         *  case of a looper or an interaction.
         */
        EVENT::TrackerHitVec *findHitsInRange(gear::GlobalPadIndex index1, gear::GlobalPadIndex index2,
                                              float minADCIndex, float maxADCIndex);

        int _maxEmptyPads;   ///< Maximum number of consecuitve empty pads in cluster
        int _minHitSize; ///< Minimum number of pads per hit
        float _maxTimeSpread;///< Maximum time between pulses in a cluster in ??
        int _minTrackHits; ///< Minimum number of hits on track
        int _maxSkipRows;    ///< Maximum number of empty rows

        float _vDrift; ///< Optional: Set drift velocity in case there is no conditions data

        //  /// The pad layout as defined in GEAR
        //    int _padLayoutType; ///< Pad layout type (PadRowLayout2D::CARTESIAN or PadRowLayout2D::POLAR)
        //    gear::PadRowLayout2D const *_padLayout;

        /// The position of the readout anode (positive value, TPC currently assumed to be symmetrical)
        float _zAnode;

        /** Flag to determine if the second half TPC is reconstructed, which looks in the
         *  negative z direction.
         *  In this case the x and z coordinate are multiplied with -1 (endplate rotated by
         *  180 degress around y axis wrt. first endplate)
         */
        int _negativeZDirection;

        /** A multimap containing the pad index and the corresponding pulse.
         *  Sorted by pad index.
         */
        std::multimap<gear::GlobalPadIndex, EVENT::TrackerPulse*> *_pulsesByIndex;

//    /**
//     * A nested helper class for sorting gear::GlobalPadIndex.
//     * They are sorted by moduleID, and within moduleID in padIndex
//     */
//    class  sortGlobalIndex {
//    public:
//      bool operator()(gear::GlobalPadIndex index1, gear::GlobalPadIndex index2) {
//  if (index1.getModuleID() == index2.getModuleID() )
//    { // the modules IDs are equal, distiguish using pad index
//      return ( index1.getPadIndex() < index2.getPadIndex() );
//    }
//  else
//    { // the modules IDs are equal, distiguish using pad index
//      return ( index1.getModuleID() < index2.getModuleID() );
//    }
//      }
//    };
        unsigned int _recursionCounter;

    public:
        /** A helper function to add the contents of one TrackerHitVec to another.
         *  Be aware that you need handlers of TrackerHitVec because the pointers are modified!
         *  \attention vectorToBeAdded is 0 after calling this function!
         */
        static void addTrackerHitVecs(EVENT::TrackerHitVec **vectorToAddTo, EVENT::TrackerHitVec **vectorToBeAdded);

        /** Get the revision of the TopoFinder.
         *  Use this information in the processor which uses or is derived from this class:
         *  \verbatim     run->parameters().setValue( _processorName + "_revision",
         *                          std::string("$Rev:$  using TopoFinder " + myTopoFinder->getRevision()));
         *  \endverbatim
         */
        std::string getRevision() ;

};

}// namespace marlintpc

#endif //TOPO_FINDER_H
