//C++
#include <iostream>
#include <stdexcept>
#include <cmath>
#include <map>
#include <list>

//LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/LCFloatVec.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/TrackImpl.h>
#include <IMPL/LCFlagImpl.h>

#include <marlin/Global.h>


#include "HitTrackFinderTopoProcessor.h"
#include "ADCChannelMapping.h"
#include <gearimpl/RectangularPadRowLayout.h>
#include <gearimpl/FixedPadSizeDiskLayout.h>
#include <gear/PadRowLayout2D.h>
#include <gear/TPCParameters.h>
#include <gearimpl/GlobalPadIndex.h>

#include "TPCConditions.h"

using namespace lcio ;
using namespace marlin ;
using namespace tpcconddata ;
using namespace std ;
using namespace gear ;

namespace marlintpc
{

HitTrackFinderTopoProcessor aHitTrackFinderTopoProcessor;


HitTrackFinderTopoProcessor::HitTrackFinderTopoProcessor() : Processor("HitTrackFinderTopoProcessor")
{


    /**
     * description of the processor which will be displayed in the
     * steering file autometicly generated by
     * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
     */
    _description = "HitTrackFinderTopoProcessor calculates TrackerHits from TrackerPulses" ;


    // register steering parameters: name, description, class-variable, default value
    registerInputCollection(LCIO::TRACKERPULSE,
                            "InputTrackerPulses" ,
                            "Name of the input TrackerPulses collection"  ,
                            _inputTrackerPulsesCollectionName ,
                            std::string("TPCPulses")) ;

    registerOutputCollection(LCIO::TRACKERHIT,
                             "OutputTrackerHits" ,
                             "Name of the output TrackerHits collection"  ,
                             _outputTrackerHitsCollectionName ,
                             std::string("TPCHits")) ;

    registerOutputCollection(LCIO::TRACK,
                             "OutputTrackCandidates" ,
                             "Name of the output Tracks collection"  ,
                             _outputTracksCollectionName ,
                             std::string("TPCTrackCandidates")) ;

    registerProcessorParameter("InputTPCConditions",
                               "Name of the input collection containing the TPC conditions data",
                               _inputTPCConditionsCollectionName,
                               tpcconddata::TPCConditions::getDefaultColName()) ;

    registerProcessorParameter("SetOutputHitsTransient" ,
                               "if not 0 the output hits collection is set transient (default: 0)"  ,
                               _outputHitsTransient,
                               int(0));

    registerProcessorParameter("SetOutputTrackCandidatesTransient" ,
                               "if not 0 the output trackscollection is set transient (default: 0)"  ,
                               _outputTrackCandidatesTransient,
                               int(0));

    registerProcessorParameter("MaxEmptyPads" ,
                               "Maximum number of empty consecuitve pads in hit (default: 1)"  ,
                               _maxEmptyPads ,
                               int(1)) ;

    registerProcessorParameter("MinHitSize" ,
                               "Minimum size of hit / Minimum number of Pads (default: 1)"  ,
                               _minHitSize ,
                               int(1)) ;

    registerProcessorParameter("MaxTimeSpread" ,
                               "Maximum time between pulses in a hit in ns (default: 200.)"  ,
                               _maxTimeSpread ,
                               float(200.)) ;

    registerProcessorParameter("MinTrackHits" ,
                               "Minimum number of hits on track (default: 5)",
                               _minTrackHits ,
                               int(5)) ;

    registerProcessorParameter("MaxSkipRows" ,
                               "Maximum number of subsequnetly missing hits (default: 1)",
                               _maxSkipRows ,
                               int(1)) ;

    registerOptionalParameter("VDriftOverride" ,
                              "Optional: Override drift velocity from conditons data or in case there is no conditions data [in mm/us]"  ,
                              _vDrift ,
                              float(0.)) ;

    registerProcessorParameter("NegativeZDirection" ,
                               "Set to 1 for the second half TPC, which look into the negative z direction (default: 0)"  ,
                               _negativeZDirection,
                               int(0));

    //  _padLayoutType = -1;
    //  _padLayout = NULL;
}


void HitTrackFinderTopoProcessor::init()
{

    // usually a good idea to
    // FIXME!!! This if statement will become obsolete / should be removed
    //          once the new logging mechanism is implemented in printParameters()
    if (streamlog::out.write< streamlog::MESSAGE3 >())
        printParameters() ;

    // this has to be per module
    //    _padLayout = &Global::GEAR->getTPCParameters().getPadLayout();
    //    _padLayoutType = _padLayout->getPadLayoutType();

    //get max drift distance from gear
    //    _maxDriftDistance = Global::GEAR->getTPCParameters().getMaxDriftLength ();

    // set the correct position of the anode from the gear file
    try
        {
            _zAnode =  marlin::Global::GEAR->getTPCParameters().getDoubleVal("zAnode");
        }
    catch (gear::UnknownParameterException &)
        {
            // no Anode position set in the GEAR file, use max drift length instead
            streamlog_out(WARNING) << "No zAnode tag found in gear file. Using maxDriftLength instead."
            << std::endl;
            _zAnode = Global::GEAR->getTPCParameters().getMaxDriftLength();
        }

//  _nRun = 0 ;
//  _nEvt = 0 ;

}

void HitTrackFinderTopoProcessor::processRunHeader(lcio::LCRunHeader* run)
{
    run->parameters().setValue(_processorName + "_revision", std::string("$Rev: 2711 $ :: TopoFinder "
                                                                         + TopoFinder::getRevision()));

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
                }
        }

//    _nRun++ ;
}


void HitTrackFinderTopoProcessor::processEvent(lcio::LCEvent * evt)
{
    // some status output
    if (evt->getEventNumber() % 1000 == 0)
        streamlog_out(MESSAGE4) << "\r" << _processorName << " processing event " << evt->getEventNumber()
        << std::endl;

    // get the pulses collections from the event
    LCCollection *inputPulses = 0;
    try
        {
            inputPulses = evt->getCollection(_inputTrackerPulsesCollectionName);
        }
    catch (DataNotAvailableException &e)
        {
            // there are not TPC TrackerPuses in this event
            // just continue ...
            return;
        }

    // the output collections:
    LCCollectionVec *tpcHitsCollection = 0;
    LCCollectionVec *tpcTrackCandidatesCollection = 0;
    // try to find the output collections in the event
    try
        {
            tpcHitsCollection = dynamic_cast<LCCollectionVec*>(evt->getCollection(_outputTrackerHitsCollectionName));

            // test whether collection is writeable
            try
                {
                    dynamic_cast<IMPL::TrackerHitImpl*>(tpcHitsCollection->getElementAt(0))->setType(
                        dynamic_cast<EVENT::TrackerHit*>(tpcHitsCollection->getElementAt(0))->getType());
                }
            catch (lcio::ReadOnlyException& e)
                {
                    throw lcio::ReadOnlyException(_outputTrackerHitsCollectionName +
                                                  " already exists in event and is not writeable");
                }
        }
    catch (DataNotAvailableException &)
        {
            // allocate the output collections (hits and tracks)
            tpcHitsCollection = new LCCollectionVec(LCIO::TRACKERHIT);
        }

    try
        {
            tpcTrackCandidatesCollection =  dynamic_cast<LCCollectionVec*>(evt->getCollection(_outputTracksCollectionName));

            // test whether collection is writeable
            try
                {
                    dynamic_cast<IMPL::TrackImpl*>(tpcTrackCandidatesCollection->getElementAt(0))->setD0(
                        dynamic_cast<EVENT::Track*>(tpcTrackCandidatesCollection->getElementAt(0))->getD0());
                }
            catch (lcio::ReadOnlyException& e)
                {
                    throw lcio::ReadOnlyException(_outputTracksCollectionName +
                                                  " already exists in event and is not writeable");
                }
        }
    catch (DataNotAvailableException &)
        {
            tpcTrackCandidatesCollection = new LCCollectionVec(LCIO::TRACK);
        }

    // set the transient flag as definded by the processor parameter
    tpcHitsCollection->setTransient((_outputHitsTransient != 0));
    tpcTrackCandidatesCollection->setTransient((_outputTrackCandidatesTransient != 0));

    LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(LCIO::TRBIT_HITS) ;
    tpcTrackCandidatesCollection->setFlag(trkFlag.getFlag());


    // get drift velicity, either measured value from calibration
    // or nominal value given as process parameter

    // check whether the drift velocity is stored as conditions data
    try
        {
            LCCollection *TPCConditionsCollection = evt->getCollection(_inputTPCConditionsCollectionName);
            TPCConditions myTPCConditions(TPCConditionsCollection->getElementAt(0));

            // conditions data are available, but can be
            // overwritten by using processor parameter
            // give a warning in this case
            if (parameterSet("VDriftOverride"))
                {
                    streamlog_out(WARNING4) << _processorName << ": WARNING: drift velicity from conditions data is"
                    << " overwritten by processor parameter!" << std::endl
                    << "Remove parameter \"VDriftOverride\" from steering file to use vDrift from TPCCondData."
                    << std::endl;
                }
            else
                _vDrift = myTPCConditions.getDriftVelocity();
        }
    catch (DataNotAvailableException &e)
        {
            // no conditions data available, try to use processor parameter
            if (!parameterSet("VDriftOverride"))
                {
                    throw(DataNotAvailableException(_processorName +
                                                    std::string(" no drift velocity available, provide either as TPCCondData or processor parameter")));
                }
        }

    // map the input TrackerPulses by their cell number
    _pulsesByIndex = new std::multimap<gear::GlobalPadIndex, TrackerPulse*>;

    for (int iPulses = 0; iPulses < inputPulses->getNumberOfElements(); iPulses++)
        {
            TrackerPulse* inputPulse = dynamic_cast<TrackerPulse*>(inputPulses->getElementAt(iPulses)) ;

            if (inputPulse)
                {
                    // create pairs of the pulses pad index and the puls itself and insert it into the map
                    _pulsesByIndex->insert(std::pair<gear::GlobalPadIndex, TrackerPulse*>(
                                               gear::GlobalPadIndex(inputPulse->getCellID0(), inputPulse->getCellID1()),
                                               inputPulse));
                }
        }

    // an iterator as a handle to the currently invesigated pulse
    std::multimap<gear::GlobalPadIndex, TrackerPulse*>::iterator currentPulse = _pulsesByIndex->begin();

    // loop over all pulses to make clusters
    while (currentPulse != _pulsesByIndex->end())
        {
            // start a cluster with currentPulse
            //
            streamlog_out(DEBUG3) << "starting new track with pad index " << currentPulse->first.getPadIndex() << std::endl;

            int moduleID = currentPulse->first.getModuleID();

            TrackerHit* seedHit = findHit(currentPulse);

            // if currentPulse is not part of a valid hit
            if (seedHit == 0)
                {
                    // set currentPulse to the beginning to the remaining map
                    currentPulse = _pulsesByIndex->begin();
                    // and continue the search
                    continue;
                }

            // the hit is valid
            // add it to the collection of hits
            tpcHitsCollection->addElement(seedHit);

            // create a track candidate which contains all hit in a contiguous area
            TrackImpl* trackCandidate = new TrackImpl();
            // add the found hit
            trackCandidate->addHit(seedHit);

            // search for contiguous hits, just call it once, it's recursive
            TrackerHitVec *contiguousHits = findHitsInNeighbouringRows(seedHit, moduleID);

            if (contiguousHits)  // 0 pointer if no hits have been found
                {
                    // add the found contiguous hits to the track candidate and the hits collection
                    for (TrackerHitVec::iterator hit = contiguousHits->begin();
                            hit < contiguousHits->end(); hit++)
                        {
                            trackCandidate->addHit(*hit); // dereference the iterator
                            tpcHitsCollection->addElement(*hit); // dereference the iterator
                        }
                }

            // check whether the track candidate is valid and add it to the tracks collection
            if (trackCandidate->getTrackerHits().size() >= static_cast<unsigned int>(_minTrackHits))
                {
                    tpcTrackCandidatesCollection->addElement(trackCandidate);
                }
            else // delete the invalid track candidate
                delete trackCandidate;

            // set currentPulse to the beginning to the remaining map
            currentPulse = _pulsesByIndex->begin();

        }// while(currentPulse != pulsesByIndex.end())

    // All loops are finished, add the collections to the event
    if (tpcHitsCollection->getNumberOfElements() > 0)
        {
            // test if the collection already is in the event, add it if not
            try
                {
                    evt->getCollection(_outputTrackerHitsCollectionName);
                }
            catch (DataNotAvailableException &)
                {
                    evt->addCollection(tpcHitsCollection, _outputTrackerHitsCollectionName);
                }

            streamlog_out(MESSAGE4) << "Found " << tpcHitsCollection->getNumberOfElements()
            << " hits." << std::endl;

            // dump the found hits in MESSAGE1 level
            if (streamlog::out.write< streamlog::MESSAGE1 > ())
                {
                    for (IMPL::LCCollectionVec::iterator hitIter = tpcHitsCollection->begin() ;
                            hitIter != tpcHitsCollection->end() ; hitIter++)
                        {
                            TrackerHit *hit = 0;
                            try
                                {
                                    hit = dynamic_cast<EVENT::TrackerHit*>(*hitIter);
                                }
                            catch (std::bad_cast &)
                                {
                                    streamlog::out() << "ERROR: bad cast when dumping hit coordinates."
                                    << std::endl;
                                    break;
                                }

                            const double * hitPos = hit->getPosition();
                            streamlog_out(MESSAGE1) << "Found hit at " << hitPos[0] << "\t"
                            << hitPos[1] << "\t" << hitPos[2] << std::endl;

                        }

                }// if MESSAGE1
        } // tpcHitsCollection not empty
    else
        delete tpcHitsCollection;

    if (tpcTrackCandidatesCollection->getNumberOfElements() > 0)
        {
            // test if the collection already is in the event, add it if not
            try
                {
                    evt->getCollection(_outputTracksCollectionName);
                }
            catch (DataNotAvailableException &)
                {
                    evt->addCollection(tpcTrackCandidatesCollection, _outputTracksCollectionName);
                }

            streamlog_out(MESSAGE4) << "Found " << tpcTrackCandidatesCollection->getNumberOfElements()
            << " tracks." << std::endl;

            if (streamlog::out.write< streamlog::MESSAGE3 > ())
                {
                    for (IMPL:: LCCollectionVec:: iterator trackIter = tpcTrackCandidatesCollection->begin() ;
                            trackIter < tpcTrackCandidatesCollection->end() ; trackIter++)
                        {
                            Track *track = 0;
                            try
                                {
                                    track = dynamic_cast<EVENT::Track*>(*trackIter);
                                }
                            catch (std::bad_cast &)
                                {
                                    streamlog_out(MESSAGE3) << "ERROR: bad cast when dumping track info."
                                    << std::endl;
                                    break;
                                }

                            streamlog_out(MESSAGE3) << "Track " << distance(tpcTrackCandidatesCollection->begin() , trackIter)
                            << " has " << track->getTrackerHits().size() << " hits" << std::endl;
                        } // for trackIter
                } // if MESSAGE3


            // dump the found hits in tracks in MESSAGE2 level
            if (streamlog::out.write< streamlog::MESSAGE2 > ())
                {
                    for (IMPL:: LCCollectionVec:: iterator trackIter = tpcTrackCandidatesCollection->begin() ;
                            trackIter < tpcTrackCandidatesCollection->end() ; trackIter++)
                        {
                            Track *track = 0;
                            try
                                {
                                    track = dynamic_cast<EVENT::Track*>(*trackIter);
                                }
                            catch (std::bad_cast &)
                                {
                                    streamlog::out() << "ERROR: bad cast when hits in tracks."
                                    << std::endl;
                                    break;
                                }

                            EVENT::TrackerHitVec trackerHits = track->getTrackerHits();
                            for (EVENT::TrackerHitVec::iterator hitIter = trackerHits.begin() ;
                                    hitIter < trackerHits.end() ; hitIter ++)
                                {
                                    const double * hitPos = (*hitIter)->getPosition();
                                    streamlog_out(MESSAGE2) << "Found hit at " << hitPos[0] << "\t"
                                    << hitPos[1] << "\t" << hitPos[2] << "\t"
                                    << "on track " << distance(tpcTrackCandidatesCollection->begin() , trackIter)
                                    << std::endl;
                                }

                        } // for trackIter
                } // if MESSAGE2


        }
    else
        delete tpcTrackCandidatesCollection;

//   _nEvt ++ ;
}



void HitTrackFinderTopoProcessor::check(lcio::LCEvent *  /*evt*/)
{
    // nothing to check here - could be used to fill checkplots in reconstruction processor
}


void HitTrackFinderTopoProcessor::end()
{

//   std::cout << "HitTrackFinderTopoProcessor::end()  " << name()
//      << " processed " << _nEvt << " events in " << _nRun << " runs "
//      << std::endl ;

}

}// namespace marlintpc
