#include "MMHitFinderProcessor.h"

//C++
#include <iostream>
#include <sstream>
#include <algorithm>

//LCIO
#include <EVENT/LCFloatVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/LCFlagImpl.h>

// GEAR
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>
#include <gearimpl/GlobalPadIndex.h>

//MARLIN
#include <marlin/Global.h>
#include <Exceptions.h>
#include <marlin/Exceptions.h>

//MARLINTPC
#include <MMPadResponseFunction.h>

//lccd
#include <lccd_exceptions.h>

// the exception by TPCCondData in MarlinTPC
#include "TPCCondDataExeptions.h"
#include "TPCConditions.h"

// the common flag word definitions in MarlinTPC
#include "FlagwordDefinitions.h"

// OTHER
#include "LCObjectCopier.h"
#include "MMScatterProcessor.h"

//ROOT
#include <Math/Minimizer.h>
#include <Math/Factory.h>

using namespace lcio;
using namespace gear;
using namespace marlin;
using namespace tpcconddata;

namespace marlintpc
{

	MMHitFinderProcessor aMMHitFinderProcessor;

	MMHitFinderProcessor::MMHitFinderProcessor() : Processor("MMHitFinderProcessor")
	{
		_description = "MMHitFinderProcessor creates TrackerHits from TrackerPulses" ;

		// register steering parameters: name, description, class-variable, default value
		registerInputCollection(LCIO::TRACKERPULSE,
				"InputTrackerPulses" ,
				"Name of the input TrackerPulses collection"  ,
				_inputTrackerPulsesCollectionName ,
				std::string("TPCPulses")) ;

		registerOutputCollection(LCIO::TRACKERHIT,
				"OutputTrackerHits" ,
				"Name of the output TrackerHits collection"  ,
				_outputTrackerHitsCollectionName ,
				std::string("TPCHits")) ;

		registerOutputCollection(LCIO::TRACKERPULSE,
				"OutputTrackerHitPulses" ,
				"Name of the output TrackerHit Pulses collection"  ,
				_outputTrackerHitPulsesCollectionName ,
				std::string("TPCHitPulses")) ;

		registerProcessorParameter("WriteOutputToStorage" ,
				"if true, the output hits collection is written to storage (default: true)"  ,
				_outputHitsPersistent,
				true);

		registerProcessorParameter("MaxEmptyPads" ,
				"Maximum number of empty consecutive pads in hit (default: 1)"  ,
				_maxEmptyChannelsOverride ,
				1) ;

		registerProcessorParameter("MinHitSizeOverride" ,
				"Minimum size of hit / Minimum number of Pads (default: 1)"  ,
				_minHitSizeOverride ,
				1) ;

		registerProcessorParameter("VDriftOverride" ,
				"Set drift velocity"  ,
				_driftVelocity,
				float(0.)) ;

		registerProcessorParameter("ModuleList" ,
				"List of module numbers to be used in the analysis."  ,
				_moduleListParam,
				_moduleListParam) ;

		registerProcessorParameter("MaxTimeSpreadOverride" ,
				"Maximum time between pulses in a hit in ns (default: 200.)"  ,
				_maxTimeSpreadOverride ,
				float(10000.)) ;

		registerProcessorParameter("PRFType",
				"Type of PRF. See MMPadResponseFunction for options.",
				_prfType,
				std::string("product"));

		registerProcessorParameter("PRF_Parameters",
				"Values of the PRF parameters.",
				_parVec,
				_parVec);

		//Optional Parameters

		registerOptionalParameter("RowDep",
				"Add a row dependence to the parameters? (I.e. Param_0 = Param_0_a*row + Param_0_b)",
				_rowDep,
				bool(false));

		registerOptionalParameter("DeadRowList" ,
				"List of row numbers *not* to be used in the analysis. Format: MODID ROWID (e.g., 0 2 = Module 0, Row 2)"  ,
				_deadRowListParam,
				_deadRowListParam) ;

		registerOptionalParameter("CentreColumns" ,
				"List of centre columns per module. Format: MODID COLID (e.g., 0 2 = Module 0, Col 2)"  ,
				_centreColListParam,
				_centreColListParam) ;

		registerOptionalParameter("Threshold" ,
				"Minimum Signal for a Hit (default: 30)"  ,
				_threshold,
				int(30)) ;

		registerOptionalParameter("CalibRun" ,
				"Is this a calibration run? (I.e., just use the weighted mean for finding hits)"  ,
				_calibrationRun,
				bool(false)) ;

		registerOptionalParameter("ExcludeRows" ,
				"Boolean: Whether or not the indicated rows should be excluded."  ,
				_excludeRows,
				false) ;

		registerOptionalParameter("SelectCentralCols" ,
				"Boolean: Whether or not only the central 10 columns should be used."  ,
				_selectCentralCols,
				false) ;

		registerOptionalParameter("ModuleDependence" ,
				"Boolean: Whether or not the provided PRF parameters are module dependent."  ,
				_moduleDep,
				false) ;

		registerOptionalParameter("PadHits",
				"Choose whether or not to force a minimum size of hits by addition 0-amplitude pulses",
				_padHits,
				bool(false));

	}//End Constructor

	void MMHitFinderProcessor::init()
	{
		printParameters();

		//Verify number and type of parameters
		_nPar = MMPadResponseFunction::getNPar(_prfType);

		if(_nPar < 1)
		{
			streamlog_out(ERROR) << "Error : PRF Type not recognized." ;
			throw StopProcessingException(this);
		}

    int reqSize = _nPar;
    if(_rowDep) reqSize *= 3;  //This depends on the form of the row dependency and should be retrieved from elsewhere
    if(_moduleDep) reqSize *= _moduleListParam.size();

		else if(!_calibrationRun)
		{
			if(reqSize != (int)_parVec.size())
			{
				if(_rowDep)
				{
					if(_moduleDep)
					{
						streamlog_out(ERROR) << "Error : PRF Type \'" << _prfType << "\' requires " << _nPar << " parameters, but " << (double)_parVec.size()/(2.*_moduleListParam.size()) << " were provided." ;
					}
					else
					{
						streamlog_out(ERROR) << "Error : PRF Type \'" << _prfType << "\' requires " << _nPar << " parameters, but " << (double)_parVec.size()/(2.) << " were provided." ;
					}
				}
				else if (_moduleDep)
				{
						streamlog_out(ERROR) << "Error : PRF Type \'" << _prfType << "\' requires " << _nPar << " parameters, but " << (double)_parVec.size()/(_moduleListParam.size()) << " were provided." ;
				}
				else
				{
					streamlog_out(ERROR) << "Error : PRF Type \'" << _prfType << "\' requires " << _nPar << " parameters, but " << (int)_parVec.size() << " were provided." ;
				}

				throw StopProcessingException(this);
			}
		}

		_parArray = new double[reqSize];
		for(int i = 0; i < reqSize; i++)
		{
      if(_calibrationRun)
      {
        _parArray[i] = 0;
      }
      else
      {
        _parArray[i] = _parVec[i];
      }
		}

		if(_moduleDep)
		{
			for(unsigned int i = 0; i < _moduleListParam.size(); i++)
			{
				_moduleMap[_moduleListParam[i]] = i;
			}
		}
		//////////

		//Set up the minimizer
		min = new ROOT::Minuit2::Minuit2Minimizer(ROOT::Minuit2::kCombined);
		min->SetTolerance(0.001);
		min->SetMaxIterations(100000);
		min->SetPrintLevel(-1);

		_moduleList = std::set<int>(_moduleListParam.begin(), _moduleListParam.end());

		if(_deadRowListParam.size() % 2) streamlog_out(MESSAGE) << "Error : Invalid Dead Row parameters" << std::endl;

		//Fill the list of appropriate rows.
		for(unsigned int id = 0; id != _deadRowListParam.size(); id += 2 )
		{

			if(_deadRowList.find(_deadRowListParam[id]) == _deadRowList.end())  //If the module is being used...
			{

				_deadRowList[_deadRowListParam[id]] = std::set<int>();

			}//End module conditional

			_deadRowList[_deadRowListParam[id]].insert(_deadRowListParam[id+1]);

		}

		if(_centreColListParam.size() % 2) streamlog_out(MESSAGE) << "Error : Invalid Centre Column List" << std::endl;

		//Fill the list of appropriate centre columns.
		for(unsigned int id = 0; id != _centreColListParam.size(); id += 2 )
		{

			_centreColList[_centreColListParam[id]] = _centreColListParam[id+1];

		}

    //These histograms should be improved
		for(std::set<int>::iterator modIt = _moduleList.begin(); modIt != _moduleList.end(); ++modIt)
		{

			int iMod = *modIt;

			std::stringstream title1, title2, title3, title4, title5, title6, title7, title8;
			title1 << "Number of Pulses per Hit (Module #" << iMod <<")";
			title2 << "Average Number of Pulse Per Hit Per Row (Module #" << iMod <<")";
			title3 << "Number of Hits per Event (Module #" << iMod <<")";
			title4 << "Chi-Square Distribution of Hits(Module #" << iMod <<")";
			title5 << "Results of Hit Fits(Module #" << iMod <<")";
			title6 << "Distribution of Hits(Module #" << iMod <<")";
			title7 << "Distribution of Hit Errors(Module #" << iMod <<")";
			title8 << "Distribution of Hit Shapes(Module #" << iMod <<")";

			// Create AIDA Histograms.
			numberOfPulsesPerHit[iMod] = AIDAProcessor::histogramFactory(this)->createHistogram1D(title1.str().c_str(), 10, -0.5, 9.5);
			averagePulsesPerHitPerRow[iMod] = AIDAProcessor::histogramFactory(this)->createHistogram1D(title2.str().c_str(), 24, -0.5, 23.5);
			numberOfHitsPerEvent[iMod] = AIDAProcessor::histogramFactory(this)->createHistogram1D(title3.str().c_str(), 200, 0.5, 200.5);

			chiSquareDistrubtionOfHits[iMod] = AIDAProcessor::histogramFactory(this)->createHistogram1D(title4.str().c_str(), 100, -0.5, 99.5); 
			hitFitResults[iMod] = AIDAProcessor::histogramFactory(this)->createHistogram1D(title5.str().c_str(), 10, -0.5, 9.5);  

			hitDistribution[iMod] = AIDAProcessor::histogramFactory(this)->createHistogram1D(title6.str().c_str(), 2000, -10., 300.);;
			hitErrorDistribution[iMod] = AIDAProcessor::histogramFactory(this)->createHistogram1D(title7.str().c_str(), 400, -0.5, 1.5);;
			hitShapes[iMod] = AIDAProcessor::histogramFactory(this)->createHistogram1D(title8.str().c_str(), 100, -5., 5.);;

			//This should be changed. 
			numberOfHitsInRow[iMod] = std::vector<int>(24, 0);
			numberOfPulsesInRow[iMod] = std::vector<int>(24, 0);

		}

	}//End init()


	void MMHitFinderProcessor::processRunHeader(LCRunHeader* run)
	{
		run->parameters().setValue(_processorName + "_revision", "$Rev: 1842 $");

		for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
		{
			if (!i->second->isOptional() || i->second->valueSet())
				run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
		}

	}//End processRunHeader()

	void MMHitFinderProcessor::processEvent(LCEvent* evt)
	{
		/** the processor is supposed to find "hits" in "rows"
			The input is a collection of pulses: (charge,time,moduleID,cellID + qualityFlag)
			The output is a collection of hits: ( [x,y,z] + covariance matrix, charge + qualityFlag )

			steering input: - minimal hit size
			- minimum signal for a hit
			- max number of dead/empty pads within a hit
			- max distance in z/time between pulses
			- drift velocity

			structurally this happens in a decomposition of:
			(A) sort the pulses by module & row
			(B) LOOP: look for pulses that belong together in a row (for all rows in a module, for all modules)
			(C) create the "Hit" by calculating all of its parameters 
		 **/

		_evt = evt->getEventNumber();

		// Get the input pulse collection
		LCCollection* inputPulses;
		try
		{
			inputPulses = evt->getCollection(_inputTrackerPulsesCollectionName);
		}
		catch (DataNotAvailableException &e)
		{
			// if no pulses are available, skip to next event
			streamlog_out(DEBUG) << "No pulses available in event " << evt->getEventNumber() << std::endl;
			return;
		}

		// create output hit collection and set the persistency flag
		LCCollectionVec* tpcHitCollection = new LCCollectionVec(LCIO::TRACKERHIT);
		tpcHitCollection->setTransient(!_outputHitsPersistent);

		// set the flags that cellID1 should be stored
		lcio::LCFlagImpl trkFlag(0);
		trkFlag.setBit(lcio::LCIO::RTHBIT_ID1);
		tpcHitCollection->setFlag(trkFlag.getFlag());

		// the main parameter for the geometry description is the gear parameter:
		const gear::TPCParameters& tpcParameters = Global::GEAR->getTPCParameters();

		/** ********************************************************************************
			the basic data structure for the hit search is a single row;
			the row is a simple vector of the Pulse pointers:
			vector<TrackerPulse*>

			Pulses are grouped into rows, and the rows into modules. This is represented by a
			map of maps of vectors of TrackerPulses, typedefed to SortedPulseMap.
			--> SortedPulseMap ==> map<int, map<int, vector<TrackerPulse*> >;	

			a helper class is instantiated to hold the hit candidates: "HitCandidate"
			the temporary result of the hit candidate search is stored in a vector
		 ********************************************************************************
		 **/

		SortedPulseMap					sortedPulses;
		SortedHitMap					hitCandidates;

		// (A): sort pulses
		_sortPulses(inputPulses, tpcParameters, sortedPulses);

		// (B)
		_findHitCandidates(sortedPulses, tpcParameters, hitCandidates);

		// Clean up the pulse map
		sortedPulses.clear();

		// (C)
		_createHitCollection(hitCandidates, tpcParameters, tpcHitCollection);

		// Save the collection
		evt->addCollection(tpcHitCollection, _outputTrackerHitsCollectionName);

		if(_outputHitsPersistent)
		{
			LCCollectionVec* tpcHitPulseCollection = new LCCollectionVec(LCIO::TRACKERPULSE);
			tpcHitCollection->setTransient(!_outputHitsPersistent);

			// Set the flags that cellID1 should be stored
			lcio::LCFlagImpl trkFlag(0);
			trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1);
			trkFlag.setBit(lcio::LCIO::TRAWBIT_CM);
			tpcHitPulseCollection->setFlag(trkFlag.getFlag());

			for(LCCollectionVec::iterator hitIt = tpcHitCollection->begin(); hitIt != tpcHitCollection->end(); ++hitIt)
			{

				const LCObjectVec& pulses = (static_cast<const TrackerHitImpl*>(*hitIt))->getRawHits();

				for(LCObjectVec::const_iterator pulseIt = pulses.begin(); pulseIt != pulses.end(); ++pulseIt)
				{

					tpcHitPulseCollection->addElement(static_cast<TrackerPulse *>(*pulseIt));
				}

			}

			// Save the collection
			evt->addCollection(tpcHitPulseCollection, _outputTrackerHitPulsesCollectionName );

		}

		streamlog_out(DEBUG3) << "Found " << tpcHitCollection->getNumberOfElements() << " hits in event " << _evt << "." << std::endl;

		streamlog_out(DEBUG3) << "There are " << inputPulses->getNumberOfElements()
			<< " pulses to begin with, and a total of "
			<< tpcHitCollection->getNumberOfElements() << " hits found." << std::endl;

	} //End processEvent()


	void MMHitFinderProcessor::check(LCEvent*  /*evt*/) 
	{
		//Nothing to do here
	} //End check()

	void MMHitFinderProcessor::end() 
	{
		for(std::set<int>::iterator modIt = _moduleList.begin(); modIt != _moduleList.end(); ++modIt)
		{
			int iMod = *modIt;
			//Should be generic.
			for(int iRow = 0; iRow < 24; iRow++){
				float average = float(numberOfPulsesInRow[iMod][iRow]) / float(numberOfHitsInRow[iMod][iRow]);
				if(numberOfHitsInRow[iMod][iRow] != 0) averagePulsesPerHitPerRow[iMod]->fill(iRow, average);
				else averagePulsesPerHitPerRow[iMod]->fill(iRow, 0.);
			}
		}
	} //End end()

	void MMHitFinderProcessor::_sortPulses(	LCCollection* inputPulses, const gear::TPCParameters& tpcParameters, 
			SortedPulseMap& sortedPulses	)
	{
		/**  
			Simple method. Procedure is as follows:

			-Check pulses one-by-one:
			-Check if the pulse's module is in the moduleList:
			-if it is, check if the module is already in the map:
			-if it is not there, add it;
			-Check if the pulse's row is in the map:
			-if it is not there, add it;
			-Add the pulse to the vector of the appropriate row in the appropriate module;
		 **/

		for(int i = 0; i < inputPulses->getNumberOfElements(); ++i)
		{
			TrackerPulse* inputPulse = static_cast<TrackerPulse*>(inputPulses->getElementAt(i));
			if (inputPulse){
				const int moduleID         = inputPulse->getCellID1();
				const int padID            = inputPulse->getCellID0();

				const TPCModule& theModule = tpcParameters.getModule(moduleID);
				const int rowID            = theModule.getRowNumber(padID);

				if(_moduleList.find(moduleID) != _moduleList.end())  //If the module is being used...
				{
					SortedPulseMap::iterator modIt = sortedPulses.find(moduleID); //Look for it in the SortedPulseMap 
					if(modIt != sortedPulses.end())  //If it's there...
					{
						std::map<int, std::vector<TrackerPulse *> >::iterator rowIt = (*modIt).second.find(rowID); //Look for the row
						if(rowIt != (*modIt).second.end()) //If it's also there...
						{
							(*rowIt).second.push_back(inputPulse);
						}
						else //Otherwise, we need a new vector there.
						{
							((*modIt).second)[rowID] = std::vector<TrackerPulse *>();
							((*modIt).second)[rowID].push_back(inputPulse);
						}
					}
					else //Otherwise, we need a new module entry
					{
						sortedPulses[moduleID] = std::map<int, std::vector<TrackerPulse *> >();
						sortedPulses[moduleID][rowID] = std::vector<TrackerPulse *>();	
						sortedPulses[moduleID][rowID].push_back(inputPulse);
					}

				}//End valid module condition

			}//End valid pulse condition

		}//End for-loop

	}//End _sortPulses()

	void MMHitFinderProcessor::_findHitCandidates( SortedPulseMap& pulseMap, const gear::TPCParameters& tpcParameters,
			SortedHitMap& theHitCandidateMap){
		/**
			Similar iterative procedure to _sortPulses. For each module, each row is checked as follows:

			-While there are pulses in the row,
			-The largest pulse is found.
			-If its charge is below the threshold, the rest of the row is discarded
			-The neighbours of the largest pulse are found and collected into a Hit Candidate
			-If the HitCandidate is too small, it is discarded.	

		 **/

		for(SortedPulseMap::iterator modIt = pulseMap.begin(); modIt != pulseMap.end(); ++modIt)
		{
			for(std::map<int, std::vector<TrackerPulse *> >::iterator rowIt = modIt->second.begin(); 
					rowIt != modIt->second.end(); ++rowIt)
			{

				while(rowIt->second.size() > 0)
				{
					TrackerPulse * max = _findHighestPulseInRow(rowIt->second);
					if(max != NULL && max->getCharge() >= _threshold) //If there is a valid max pulse
					{
						//			streamlog_out(MESSAGE) << "Found a max pulse in the row. It is on row " << rowIt->first << " and has charge " << max->getCharge() << std::endl;

						HitCandidate tmpHit(max);
						_addNeighbourPulses(tmpHit, tpcParameters, rowIt->second);

						/*			streamlog_out(MESSAGE) << "Found the following hit candidate: " << std::endl;
										for(std::deque<TrackerPulse *>::const_iterator it = tmpHit.getPulses().begin(); it != tmpHit.getPulses().end(); ++it)
										streamlog_out(MESSAGE) << (*it)->getCharge() << "  ";
										streamlog_out(MESSAGE) << std::endl;
						 */
						if(tmpHit.getNumberOfPulses() >= _minHitSizeOverride) //Can add a quality check here later too
						{
							SortedHitMap::iterator checkIt = theHitCandidateMap.find(modIt->first);
							if(checkIt != theHitCandidateMap.end()) //If the module is not empty
							{
								checkIt->second.push_back(tmpHit); //Add the hit
							}
							else //Otherwise...
							{
								theHitCandidateMap[modIt->first] = std::vector<HitCandidate>(); //Add the module
								theHitCandidateMap[modIt->first].push_back(tmpHit); //Then add the hit
							}
						}//End hit quality condition

					}
					else{	break; }//End threshold condition

				}//End row while-loop

			}//End row for-loop

		}//End module for-loop

	}//End _findHitCandidates()


	TrackerPulse* MMHitFinderProcessor::_findHighestPulseInRow(std::vector<TrackerPulse*>& pulsesInARow)
	{
		/**

			Simple search algorithm. Returns the pulse with the highest charge. In cases of multiple pulses with the same
			highest charge, the first is returned.

			This method is (almost) unmodified from the RowBasedHitFinderProcessor()

		 **/

		if(pulsesInARow.size() == 0)
			return NULL;

		TrackerPulse* highestPulse = pulsesInARow.at(0);

		std::vector<TrackerPulse*>::iterator eraser = pulsesInARow.begin();
		for (std::vector<TrackerPulse*>::iterator iter = pulsesInARow.begin(); iter < pulsesInARow.end(); ++iter)
		{
			if (fabs((*iter)->getCharge()) > fabs(highestPulse->getCharge()))
			{
				highestPulse = *iter;
				eraser = iter;
			}
		}
		pulsesInARow.erase(eraser);
		return highestPulse;

	}//End _findHighestPulseInRow()

	void MMHitFinderProcessor::_addNeighbourPulses( HitCandidate& theCandidate,  
			const gear::TPCParameters& tpcParameters, std::vector<TrackerPulse*>& pulsesInRow)
	{
		/**
			Again, this is a fairly simple procedure. It is as follows:

			-Check the pulse vector for pulses to the left of the max pulse (i.e., the one in the hit candidate), 
			from closest to furthest
			-Add the pulse to the front if: a) the pulse times are close enough, and b) few enough pads have been skipped.
			-Repeat the process (inverted, of course) for the pulses to the right of the max pulse.

		 **/

		// the initial values for searching
		const TrackerPulse * max 		= theCandidate.getMaximumPulse();
		const int padID				    = max->getCellID0();
		const int modID					= max->getCellID1();
		const TPCModule& theModule      = tpcParameters.getModule(modID);
		const int maxCol				= theModule.getPadNumber(padID);
		const int maxRow				= theModule.getRowNumber(padID);
		//const float maxTime		        = max->getTime();
		const float maxTime		        = max->getTrackerData()->getTime();
		const float maxCharge	        = max->getCharge();

		int count = 1; //Keeps track of how many pads to the left or right we are looking for
		int padsSkipped = 0; //Keeps track of how many pads have been skipped
		bool left = true, right = false;
		float curTime = maxTime;
		float curCharge = maxCharge;

		streamlog_out(DEBUG) << "Searching Module " << modID << ", Row " << maxRow << std::endl;
		streamlog_out(DEBUG) << "Max pulse on column " << maxCol << std::endl;

		while (left || right) //While we still have a direction to search
		{
			streamlog_out(DEBUG) << "Searching left = " << left << ", right = " << right << ", count = " << count << ", desired column = "
				<< (left ? maxCol - count : maxCol + count) << std::endl;

			//Check that we're not at a boundary
			if(left && maxCol - count < 0)
			{
				count = 1;
				curTime = maxTime;
				curCharge = maxCharge;
				left = false;
				right = true;
				continue;
			}
			else if (right && maxCol + count >= (int)theModule.getPadsInRow(maxRow).size())
			{
				right = false;
				continue;
			}

			//Otherwise, get the desired pad ID
			const int searchID = left ? theModule.getPadIndex(maxRow, maxCol - count) : theModule.getPadIndex(maxRow, maxCol + count); 

			//A vector of possible pulses
			std::vector<TrackerPulse*> possibleNeighbours;

			for(std::vector<TrackerPulse*>::iterator pulseIt = pulsesInRow.begin(); pulseIt != pulsesInRow.end(); ++pulseIt)
			{

				//		const int checkCol = theModule.getPadNumber((*pulseIt)->getCellID0());
				//		if((left && (maxCol - checkCol > 0 && maxCol - checkCol == count)) ||
				//			(right && (checkCol - maxCol > 0 && checkCol - maxCol == count))) //Check if it's the pulse for which we are looking
				if((*pulseIt)->getCellID0() == searchID)
				{

					possibleNeighbours.push_back(*pulseIt);			

				}//End distance conditional

			}//End left search pulses for-loop

			streamlog_out(DEBUG) << "Found " << possibleNeighbours.size() << " suitable candidates. Pads Skipped = " << padsSkipped << ", maxEmptyChannels = " 
				<< _maxEmptyChannelsOverride << std::endl;

			if(possibleNeighbours.size() == 0) //If no appropriate match was found...
			{
				if(padsSkipped < _maxEmptyChannelsOverride) //Check if we can skip a pad. If so, do it.
				{
					padsSkipped++;
				}
				else //Otherwise, give up and either search to the right, or finish up
				{

					//	  padsSkipped = 0; //Hm. Shouldn't be resetting this one.
					count = 1;
					curTime = maxTime;
					curCharge = maxCharge;
					if(left)
					{
						left = false;
						right = true;
					}
					else
					{
						left = false;
						right = false;
					}

				}//End skipped pulses conditional
			}
			else //Otherwise, at least one pulse was found in the appropriate column. Search these for the closest in time,
				//and make sure that they are decreasing in amplitude.
				{
					//		float bestTime = 99999;
					TrackerPulse* nearest = NULL;
					streamlog_out(DEBUG) << "curCharge = " << curCharge << ", curTime = " << curTime << std::endl;
					for(std::vector<TrackerPulse*>::iterator possPulseIt = possibleNeighbours.begin();
							possPulseIt != possibleNeighbours.end(); ++possPulseIt)
					{
						float tmpTime = ((*possPulseIt)->getTime() - curTime);
						float tmpCharge = (curCharge - (*possPulseIt)->getCharge());

						streamlog_out(DEBUG) << "Checking charge = " << (*possPulseIt)->getCharge() << ", time = " << (*possPulseIt)->getTime() << std::endl;

						//Guard against anomalous small pulses showing up before the max
						//I don't actually remember why I was doing this...
						/*			if(count > 1 && tmpTime < 360)
										{
										streamlog_out(DEBUG) << "REJECTED" << std::endl;
										continue;
										}
						 */			
						if( tmpCharge >= 0 && tmpTime > -160 && tmpTime < _maxTimeSpreadOverride )  //160 is ~4 time bins.
							//	    if( tmpCharge >= 0 && tmpTime < _maxTimeSpreadOverride ) 
							//		  if( tmpTime < bestTime && tmpTime < _maxTimeSpreadOverride ) 
							//if( tmpTime > -180 && tmpTime < _maxTimeSpreadOverride ) 
							//if( tmpTime < _maxTimeSpreadOverride ) 
						{
							if(nearest == NULL || tmpCharge < (curCharge - nearest->getCharge())) //Checks charge
								//				if(nearest == NULL || tmpTime < (nearest->getTime() - curTime)) //Checks time
							{
								streamlog_out(DEBUG) << "ACCEPTED Charge" << std::endl;
								//					bestTime = tmpTime;
								nearest = *possPulseIt;
							}
							else if(tmpCharge == (curCharge - nearest->getCharge()) && tmpTime < (nearest->getTime() - curTime))
							{
								streamlog_out(DEBUG) << "ACCEPTED Time" << std::endl;
								//					bestTime = tmpTime;
								nearest = *possPulseIt;
							}
							else
								streamlog_out(DEBUG) << "REJECTED" << std::endl;

						}//End time difference conditional
						else
							streamlog_out(DEBUG) << "REJECTED" << std::endl;

					}//End possible neighbours for-loop

					if(nearest != NULL)
					{
						theCandidate.addPulse(nearest, left);
						//		  curTime = nearest->getTime();
						curCharge = nearest->getCharge();

						//Here we will directly remove this pulse from the vector of pulses in the row. In the future, consider first transferring the
						//contents of the vector to a list and using that to find/remove elements. Could be faster.
						for(std::vector<TrackerPulse*>::iterator pulseIt = pulsesInRow.begin(); pulseIt != pulsesInRow.end(); ++pulseIt)
						{

							if(*pulseIt == nearest) //Check if it's the pulse for which we are looking
							{

								pulsesInRow.erase(pulseIt);
								break;			

							}//End equality conditional

						}//End for-loop to remove the found pulse
						streamlog_out(DEBUG) << "Successfully added a pulse from column " << (left ? maxCol - count : maxCol + count) << std::endl;
						count++;
					}
					else
					{
						if(padsSkipped < _maxEmptyChannelsOverride) //Check if we can skip a pad. If so, do it.
						{
							padsSkipped++;
						}
						else //Otherwise, give up and either search to the right, or finish up
						{

							//	  padsSkipped = 0; //Hm. Shouldn't be resetting this one.
							count = 1;
							curTime = maxTime;
							curCharge = maxCharge;
							if(left)
							{
								left = false;
								right = true;
							}
							else
							{
								left = false;
								right = false;
							}

						}//End skipped pulses conditional
					}


				}//End possible neighbours conditional

		}//End search pulses while-loop

	}//End _addNeighbourPulses()

	void MMHitFinderProcessor::_createHitCollection (	SortedHitMap& hitCandidateMap, const gear::TPCParameters& tpcParameters, 
			LCCollectionVec* tpcHitCollection	)
	{
		/** Now, loop over all hits on all modules and:
			-check their validity (nothing to do here yet, but possibly in the future)
			-calculate the fixed and free coordinates (given by the row [y or r], and the x or phi)
			-calculate the z coordinate
			-store the hit collection!
		 **/
		//		streamlog_out(MESSAGE) << "In _createHitCollection() " << endl;
		//		streamlog_out(MESSAGE) << "There are " << hitCandidateMap[3].size() << " hits in the candidate hit map." << endl;

		int hitCounter = 0;  //Keep track of number of hits in the event;

		// Loop over modules
		for (	SortedHitMap::iterator modIt = hitCandidateMap.begin(); modIt != hitCandidateMap.end(); ++modIt	)
		{

			const TPCModule& theModule = tpcParameters.getModule(modIt->first);

			// Loop over HitCandidates:
			for (	std::vector<HitCandidate>::iterator hitIt = modIt->second.begin(); hitIt != modIt->second.end(); ++hitIt	)
			{

				// check validity
				if (hitIt->getQuality())
					continue;

				// change output coordinate system
				double position[3];

				/************************************************************************************************/
				//                       CALCULATE THE HIT COORDINATES                                          //
				/************************************************************************************************/

				std::pair<double, double> freeCoord = std::pair<double, double>(0., 0.);
        freeCoord 		= _calculateFreeCoordinate	(	*hitIt, tpcParameters	);
				if(freeCoord.first == -9E9) //In case the fit failed
					continue;

				std::pair<double, double> fixedCoord 	= _calculateFixedCoordinate	(	*hitIt, tpcParameters	);
				std::pair<double, double> zCoord 		= _calculateZCoordinate		(	*hitIt					);

				gear::Vector2D positionGlobal2D = theModule.localToGlobal(fixedCoord.first, freeCoord.first);
				position[0] = positionGlobal2D[0];
				position[1] = positionGlobal2D[1];
				//position[2] = zCoord.first; 
				position[2] = 600; //For testing x-y reconstruction; Always run MMHitTimeCorrectionProcessor after this processor.

				//Again we'll assume there is no correlation between these parameters. This could, of course, be altered in the future.
				float covMatrix[6] = 	
				  {	float(freeCoord.second*freeCoord.second), 
					float(0.),							float(fixedCoord.second*fixedCoord.second),
					float(0.),							float(0.),							float(zCoord.second*zCoord.second )
				};

				/************************************************************************************************/
				/*                       CREATE THE HIT OBJECT AND SET ITS VARIABLES                            */
				/************************************************************************************************/
				TrackerHitImpl* theHit = new TrackerHitImpl();

				theHit->setPosition		(	position				);
				theHit->setCovMatrix	(	covMatrix				);
				//		theHit->setEDep			(	hitIt->getMaxCharge()	);
				//		theHit->setEDepError	(	sqrt(theHit->getEDep())	);

				float hitCharge = 0.;

				std::deque<const TrackerPulse*> hitPulses = hitIt->getPulses();
				for (	std::deque<const TrackerPulse*>::const_iterator aPulse = hitPulses.begin(); aPulse != hitPulses.end(); ++aPulse)
				{

					//TrackerPulseImpl* newPulse = new TrackerPulseImpl();
					TrackerPulseImpl* newPulse = LCObjectCopier::copy(*aPulse);

					hitCharge += newPulse->getCharge();

					theHit->rawHits().push_back(newPulse);


					//We can deal with the quality flags later.
					/*
					// determine the quality bits from the pulses to let hit inherit them
					int pulseQuality = (*aPulse)->getQuality();
					if (pulseflag::isSplit(pulseQuality)) // split pulse?
					hitflag::setSplitPulse(candidate->qualityFlag);
					if (pulseflag::isOverflow(pulseQuality)) // overflow
					hitflag::setOverflowPulse(candidate->qualityFlag);
					if (pulseflag::isAnomalousShape(pulseQuality)) // anomalous shape
					hitflag::setAnomalousPulseShape(candidate->qualityFlag);
					 */ 

				}

				theHit->setEDep			(	hitCharge	);
				theHit->setEDepError	(	sqrt(hitCharge)	);
				theHit->setQuality(hitIt->getQuality());

				// the time is inherited from the pulse with the largest charge
				theHit->setTime(hitIt->getMaximumPulse()->getTime());

				int padIndex = -999999;

				padIndex = hitIt->getMaximumPulse()->getCellID0();

				theHit->setCellID0(hitIt->getMaximumPulse()->getCellID0());
				theHit->setCellID1(hitIt->getMaximumPulse()->getCellID1());

				/////
				int rowNum = theModule.getRowNumber(padIndex);
				int colNum = theModule.getPadNumber(padIndex);
				/////

//				std::cout << "_excludeRows = " << _excludeRows << std::endl;
//				std::cout << "Module = " << modIt->first << "\t rowNum = " << rowNum << "\t colNum = " << colNum << std::endl;
//				getchar();

				if(_excludeRows && (_deadRowList.find(modIt->first) != _deadRowList.end()) && (_deadRowList[modIt->first].find(rowNum) != _deadRowList[modIt->first].end()))  //Extra Temp!!
					//		if(_excludeRows && (rowNum == 0 || rowNum == 11 || rowNum == 22 || rowNum == 23 || colNum < 32 || colNum > 41))  
				{
//					std::cout << "In the exit conditional!" << std::endl; 
					continue;
				}

				if(_selectCentralCols && (_centreColList.find(modIt->first) != _centreColList.end()) && (_centreColList[modIt->first] - 5 > colNum  || _centreColList[modIt->first] + 4 < colNum ))  
				{
					continue;
				}

				tpcHitCollection->addElement(theHit);	    

				numberOfHitsInRow[modIt->first][theModule.getRowNumber(padIndex)]++;
				numberOfPulsesInRow[modIt->first][theModule.getRowNumber(padIndex)] += hitIt->getNumberOfPulses();
				numberOfPulsesPerHit[modIt->first]->fill( hitIt->getNumberOfPulses() );
				hitCounter++;

			}//End HitCandidate for-loop

			numberOfHitsPerEvent[modIt->first]->fill( hitCounter );
			hitCounter = 0;

		}//End Module for-loop

	}//End _createHitCollection();

	std::pair<double, double> MMHitFinderProcessor::_calculateFixedCoordinate(	const HitCandidate& cand, const gear::TPCParameters& tpcParameters	)
	{
		//Taken almost directly from RowBasedHitFinderProcessor.

		std::pair<double, double> fixedCoord;

		const TrackerPulse* maxPulse = cand.getMaximumPulse();

		const gear::TPCModule &module = tpcParameters.getModule(maxPulse->getCellID1());
		const gear::PadRowLayout2D &localPadPlane = module.getLocalPadLayout();
		Vector2D padCenterLocal = localPadPlane.getPadCenter(maxPulse->getCellID0());

		if (localPadPlane.getCoordinateType() == PadRowLayout2D::CARTESIAN)
			fixedCoord.first = padCenterLocal[1]; // this is y
		else if (localPadPlane.getCoordinateType() == PadRowLayout2D::POLAR)
			fixedCoord.first = padCenterLocal[0]; // this is r
		else
			throw gear::Exception("Coordinate type unknown!");

		const double SQRT12 = 3.46410161513775458705;

		int rowNumber    = module.getRowNumber(maxPulse->getCellID0());
		double rowHeight = module.getRowHeight(rowNumber);

		fixedCoord.second = rowHeight / SQRT12;

		return fixedCoord;
	}

	std::pair<double, double> MMHitFinderProcessor::_calculateFreeCoordinate(	const HitCandidate& cand, const gear::TPCParameters& tpcParameters)
	{

		if(_calibrationRun)
		{
      //This is for the new calibration method involving the MMScatterProcessor.
			if(MMScatterProcessor::CALIB_PARAMS.size() == 0)
				return _calculateFreeCoordinateWeightedMean(cand, tpcParameters);
			else
			{
				for(unsigned int i = 0; i < MMScatterProcessor::CALIB_PARAMS.size(); i++)
				{
					_parArray[i] = MMScatterProcessor::CALIB_PARAMS[i];
				}
				return _calculateFreeCoordinatePRF(cand, tpcParameters);
			}
		}
		else
		{
			return _calculateFreeCoordinatePRF(cand, tpcParameters);
		}
	}

	std::pair<double, double> MMHitFinderProcessor::_calculateFreeCoordinateWeightedMean(	const HitCandidate& cand, const gear::TPCParameters& tpcParameters)
	{
		/*
			 If we're doing a calibration run, all we need is to use the weighted mean for a good first estimate.
		 */

		std::pair<double, double> freeCoord;

		const TrackerPulse* maxPulse = cand.getMaximumPulse();
		const gear::TPCModule &module = tpcParameters.getModule(maxPulse->getCellID1());
		const gear::PadRowLayout2D &localPadPlane = module.getLocalPadLayout();

		std::deque<const TrackerPulse*> thePulses = cand.getPulses();

		if(thePulses.size() == 1)
		{
			// Depending on the coordinate type convert the pad center into a position.
			if (localPadPlane.getCoordinateType() == PadRowLayout2D::CARTESIAN){
				freeCoord.first = localPadPlane.getPadCenter( maxPulse->getCellID0() )[0];
			}
			else if (localPadPlane.getCoordinateType() == PadRowLayout2D::POLAR){
				freeCoord.first = localPadPlane.getPadCenter( maxPulse->getCellID0() )[1];
			}

			freeCoord.second = sqrt(12);  //FIXME This will probably be dependent on the PRF in the future.

			return freeCoord;
		}

		float  padAmp = 0.;
		float  padAmpSum = 0.;
		double positionSum = 0.;

		Vector2D padCenterLocal;

		for (std::deque<const TrackerPulse*>::const_iterator aPulse = thePulses.begin(); aPulse < thePulses.end(); ++aPulse){

			// Get the amplitude of the pad.
			padAmp = (*aPulse)->getCharge();

			padAmpSum += padAmp;

			// Get the location of the pad center.
			padCenterLocal = localPadPlane.getPadCenter( (*aPulse)->getCellID0() );

			positionSum += padAmp*padCenterLocal[1]; 

		}//End HitCandidate for-loop

		freeCoord.first = positionSum/padAmpSum;
		freeCoord.second = localPadPlane.getPadWidth( maxPulse->getCellID0() )/2.; //Inflate error to ~1/2 a pad's width. Not that it works. 

		return freeCoord;
	}
	std::pair<double, double> MMHitFinderProcessor::_calculateFreeCoordinatePRF(	const HitCandidate& cand, const gear::TPCParameters& tpcParameters)
	{
		/*

			 The process here is the following:

			 -Use the data points to create a PRF object that contains a chi2 function to be minimized
			 -Minimize the chi2 function

       N.B. We are assuming polar local coordinates, since that is what we generally use with the Micromegas
		 */

		std::pair<double, double> freeCoord;

		const TrackerPulse* maxPulse = cand.getMaximumPulse();
		const gear::TPCModule &module = tpcParameters.getModule(maxPulse->getCellID1());
		const gear::PadRowLayout2D &localPadPlane = module.getLocalPadLayout();

		std::deque<const TrackerPulse*> thePulses = cand.getPulses();

		if(thePulses.size() == 1)
		{
			freeCoord.first = localPadPlane.getPadCenter( maxPulse->getCellID0() )[1];
			freeCoord.second = sqrt(12);  //FIXME This will probably be dependent on the PRF in the future.

			return freeCoord;
		}

		float  padAmp = 0.;
		double position = 0.;

		int nPulses = thePulses.size();
		int iter = 0;
		int maxPos = 0;

		Vector2D padCenterLocal;

		std::vector< std::pair <double, float> > fitData;

		for (std::deque<const TrackerPulse*>::const_iterator aPulse = thePulses.begin(); aPulse < thePulses.end(); ++aPulse){

			if(*aPulse == maxPulse)	maxPos = iter;

			iter++;

			// Get the amplitude of the pad.
			padAmp = (*aPulse)->getCharge();

			// Get the location of the pad center.
			padCenterLocal = localPadPlane.getPadCenter( (*aPulse)->getCellID0() );

			position = padCenterLocal[0]*padCenterLocal[1]; // convert to arclength.

			// Add the point to the vector of fit data.
			fitData.push_back(std::pair<double, float>(position, padAmp));

		}//End HitCandidate for-loop

		//If desired, add pulses with 0 amplitude to fill out small hits. This should help bring down tails and widen the PRF.
		//NOTE: If this becomes permanent, this implementation will be replaced by the better algorithm in the MMRefitTool.
		if(_padHits)
		{
			int maxCol = module.getPadNumber(maxPulse->getCellID0());

			for(int iC = maxCol - 2; iC < maxCol + 3; iC++)
			{
				std::deque<const TrackerPulse*>::const_iterator pulseIt;
				for(pulseIt = thePulses.begin(); pulseIt != thePulses.end(); ++pulseIt)
				{
					if(module.getPadNumber((*pulseIt)->getCellID0()) == iC) { break; }
				}

				if(pulseIt == thePulses.end() && iC >= 0 && iC < (int)module.getPadsInRow(module.getRowNumber(maxPulse->getCellID0())).size())
				{

					int curPad = module.getPadIndex(module.getRowNumber(maxPulse->getCellID0()), iC);
					gear::Vector2D	curPadLocal = module.globalToLocal(module.getPadCenter(curPad)[0], module.getPadCenter(curPad)[1]); 
					fitData.push_back(std::pair<double, float>(curPadLocal[0]*curPadLocal[1], 0.));
				}
			}
		}

		float delta = nPulses - 2*maxPos - 1;
		hitShapes[maxPulse->getCellID1()]->fill(delta);

		// Create an instance of the PRF which is minimized to find the hit location.
		MMPadResponseFunction aPRF(_prfType);

		aPRF.SetData(fitData);

		int modIndex = -1;
		int rowIndex = -1;
		if(_moduleDep)
		{
			std::map<int, int>::const_iterator modIt = _moduleMap.find(module.getModuleID());
			if(modIt == _moduleMap.end())
			{
				_moduleMap[module.getModuleID()] = 0;
				modIndex = 0;
			}
			else
				modIndex = modIt->second;
		}
		if(_rowDep) rowIndex = module.getRowNumber(maxPulse->getCellID0());
		//const double * tmpPar = calcParArray(_parArray, _nPar, (int)_moduleListParam.size(), modIndex, rowIndex);
		const std::vector<double> tmpPar = calcParArray(_parArray, _nPar, (int)_moduleListParam.size(), modIndex, rowIndex);
		//aPRF.SetParameters(tmpPar, _nPar);
		aPRF.SetParameters(&tmpPar[0], _nPar);

		// Create an instance of the minimizer   
		// N.B. Can choose minimization type:
		//     kMigrad, kSimplex, kCombined, 
		//     kScan, kFumili

		//Clear the minimizer to prepare for another minimization.
		min->Clear();

		//Try letting Minuit choose these
		//		min->SetTolerance(0.001);
		//		min->SetMaxIterations(100000);

		ROOT::Math::Functor f = ROOT::Math::Functor(aPRF,1);

		//Now we need some limits on our fit, to make it easier. We will look at a range of
		//3 pad-widths, centred on the main pad.
		Vector2D maxPadCenterLocal = localPadPlane.getPadCenter( maxPulse->getCellID0() );
		float padWidth = localPadPlane.getPadWidth( maxPulse->getCellID0() ); //Careful. localPadPlane returns width in *local* coordinates.
		double middle = 0;
		middle = maxPadCenterLocal[0]*maxPadCenterLocal[1]; // convert to cartesian.
		padWidth *= maxPadCenterLocal[0]; 

		//double lower = middle - padWidth*3.;
		//double upper = middle + padWidth*3.;

		min->SetVariable(0, "hitX", middle, 1.e-7);
//		min->SetVariable(0, "hitX", middle, 1.);

		// Now set the function to be minimized
		min->SetFunction(f);

		// Minimize the function
		if(!min->Minimize())
		{
			hitFitResults[maxPulse->getCellID1()]->fill(min->Status());
			freeCoord.first  = -9E9;
			freeCoord.second = -9E9;
			return freeCoord;
		}

		//Get the errors from Minos
		double errUp  = -9e9;
		double errLow = -9e9; 
		double hitError[1] = {min->Errors()[0]};
		
		if(!min->GetMinosError(0, errLow, errUp))
		{
			hitFitResults[maxPulse->getCellID1()]->fill(min->Status());
			freeCoord.first  = -9E9;
			freeCoord.second = -9E9;
			return freeCoord;
		}
		
		hitError[0] = std::max(fabs(errLow), fabs(errUp));

		// Get the results of the fit.
		const double *hitPos = min->X();
		double chi2 = min->MinValue(); 
		int resultFlag = min->Status();

		//////Diagnostic tools. /////////
		chiSquareDistrubtionOfHits[maxPulse->getCellID1()]->fill(chi2);
		hitFitResults[maxPulse->getCellID1()]->fill(resultFlag);

		hitDistribution[maxPulse->getCellID1()]->fill(hitPos[0]);
		hitErrorDistribution[maxPulse->getCellID1()]->fill(hitError[0]);
		///////////////////////////////////////////

	  freeCoord.first  = hitPos[0]/maxPadCenterLocal[0];
	  freeCoord.second = hitError[0]/maxPadCenterLocal[0];

		return freeCoord;

	}//End _calculateFreeCoordinate();

	std::pair<double, double> MMHitFinderProcessor::_calculateZCoordinate(	const HitCandidate& cand	)
	{
		/*
			 Taken almost directly from RowBasedHitFinderProcessor.
		 */

		std::pair<double, double> zCoord;

		zCoord.first = Global::GEAR->getTPCParameters().getMaxDriftLength()
			-  _driftVelocity*(cand.getMaximumPulse())->getTime() / 1000.; // time is in [ns], drift velocity in [mm/mus]

		double pulseTimeError = sqrt((cand.getMaximumPulse()->getCovMatrix())[2]);
		zCoord.second = _driftVelocity * pulseTimeError / 1000.;

		return zCoord;

	}//End _calculateZCoordinate();

}// namespace marlintpc
