#include "PRFHitBuilderProcessor.h"


// the common flag word definitions in MarlinTPC
#include "FlagwordDefinitions.h"

// Marlin
#include <marlin/ConditionsProcessor.h>
#include <marlin/Exceptions.h>

//LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerHit.h>
#include <EVENT/Track.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/LCFlagImpl.h>
#include "LCObjectCopier.h"

//Streamlog
#include <streamlog/streamlog.h>
// GEAR
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>

// Aida
#ifdef MARLIN_USE_AIDA
//AIDA
#include <marlin/AIDAProcessor.h>
#endif

// the exception by TPCCondData in MarlinTPC
#include "TPCCondDataExeptions.h"
#include "TPCConditions.h"

#include <math.h> 

using namespace lcio;
using namespace marlin;
using namespace std;


namespace marlintpc
{
  PRFHitBuilderProcessor aPRFHitBuilderProcessor;

  PRFHitBuilderProcessor::PRFHitBuilderProcessor() 
    : Processor("PRFHitBuilderProcessor"),_tpcConditionsListener(0),_PadResponseFunctionGEM(0)
  {

    _description = "PRFHitBuilderProcessor: finds pulses out of the raw data and calculate the time and charge (in FADC) information";

    registerInputCollection(LCIO::TRACKERHIT,
                            "InputCollectionName",
                            "Name of the input TrackerHit collection (default: TPCHits)",
                            _inputColName,
                            string("TPCHits")); 

    //for a later stage
//     registerInputCollection(LCIO::TRACK,
//                             "InputCollectionName",
//                             "Name of the input Track collection (default: TPCTracks)",
//                             _inputTracksColName,
//                             string("TPCTracks")); 

    registerOutputCollection(LCIO::TRACKERHIT,
                             "OutputCollectionName",
                             "Name of the output TrackerHit collection (default: TPCPRFHits)",
                             _outputColName,
                             string("TPCPRFHits"));

    registerProcessorParameter("defocussing",
                               string("Parameter for the PRF: defocussing = diffusion in amplification region sigma_0^2 [mm^2] (e.g. GEM stack) (default: 0.01)"),
                               _defocussing,
                               double(0.01));

    registerProcessorParameter("diffusion",
                               string("Parameter for the PRF: diffusion = diffusion in drift region D^2 [mm] (default: 0.001)"),
                               _diffusion,
                               double(0.001));
    
    registerProcessorParameter("PRConly",
                               string("Parameter to switch off the fitting of the PRF and just do the calculation of a correction (default 1)"),
                               _PRConly,
                               int(true));
    
    registerProcessorParameter("offsetx", "sigma0 from resolution fit in xy", _offsetx, double(0.02));
    registerProcessorParameter("slopex", "diffusion term from resolution  fit in xy", _slopex, double(0.0004));
    registerProcessorParameter("offsetz", "sigma0 from resolution  fit in z", _offsetz, double(0.4));
    registerProcessorParameter("slopez", "diffusion term from resolution fit in z", _slopez, double(0.));
    
    registerOptionalParameter("InputTPCConditions",
			      "Optional: Name of the input collection containing the TPC conditions data",
			      _inputTPCConditionsCollectionName,
			      tpcconddata::TPCConditions::getDefaultColName()) ;
    
 }



  void PRFHitBuilderProcessor::init()
  {
    // usually a good idea to
    streamlog_out(DEBUG) << "   init called  " << std::endl ;
    if(parameterSet("InputTPCConditions"))
      {
	try
	  {
	    _tpcConditionsListener = new TPCConditionsListener(_inputTPCConditionsCollectionName);
	    _offsetx=_tpcConditionsListener->getTransDefocussing();
	    _slopex=_tpcConditionsListener->getTransDiffusionCoef();
	    _offsetz=_tpcConditionsListener->getLongDefocussing();
	    _slopez=_tpcConditionsListener->getLongDiffusionCoef();
	  }
	
	catch(lccd::LCCDException &)
	  {
	    streamlog_out(ERROR) << "Collection named \"" << _inputTPCConditionsCollectionName << "\" with TPCConditions is not available." << std::endl;
	    streamlog_out(ERROR) << "Use the ConditionsProcessor to provide it, or set the resolution parameter (and comment the InputTPCConditions parameter) if this is not possible." << std::endl;
	    throw;
	  }
      }
    else
      {
	streamlog_out(WARNING) << "The resolution parameters are taken from the XML input" << std::endl;
      }

    printParameters();
    AIDAProcessor::tree(this);
    _testTuple=new  TNtuple("testTuple","testTuple","prf:npulses:sigma:in:out");
    _gearTPC = &Global::GEAR->getTPCParameters() ;


    _PadResponseFunctionGEM=new PadResponseFunctionGEM(_defocussing, _diffusion, Global::GEAR->getTPCParameters());
    _PadResponseFunctionGEM->SetPRCOnly(_PRConly);
    _PRF_used=0;
    _PRF_notused=0;
    _PRF_failed=0;  
  
  }



  void PRFHitBuilderProcessor::processRunHeader(lcio::LCRunHeader*  /*run*/)
  {
  }



  void PRFHitBuilderProcessor::processEvent(lcio::LCEvent* evt)
  { 
    LCCollectionVec * inputCollection;
    
    try
      {
	// this cast should be safe. The CollectionVec is the only implementation of
      // the abstract interface class.
	inputCollection = dynamic_cast<LCCollectionVec *>(evt->getCollection(_inputColName));
      }
    catch(DataNotAvailableException &e)
      {
        // if there is not input collection print a message and just continue. It's just an 
	// empty event, might happen...
        streamlog_out(MESSAGE) << "Input collection " << _inputColName 
		       << " cannot be found in event " << evt->getEventNumber() << endl;
	
	return; // just return, nothing special about an empty event.
      }
    
    //output collection with improved hits
    IMPL::LCCollectionVec* outputCol = new IMPL::LCCollectionVec(LCIO::TRACKERHIT);
    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0);
    trkFlag.setBit(lcio::LCIO::RTHBIT_ID1);
    outputCol->setFlag(trkFlag.getFlag());

    // loop the input data 
    for (LCCollectionVec::const_iterator inputIter = inputCollection->begin();
	 inputIter != inputCollection->end(); ++inputIter)
      {
	TrackerHit* inputTrackerHit = dynamic_cast<TrackerHit*>(*inputIter);
	//------------PRF part-------------------------
	//use PRF to improve position and correct charge if dead channel or overrange pulse present
	bool status=_PadResponseFunctionGEM->EvaluateHit(inputTrackerHit);
	if(status)
	  _PRF_used++;
	else
	  _PRF_notused++;
	if(!_PadResponseFunctionGEM->GetPRFSucceeded())
	  _PRF_failed++;
	double *pos=_PadResponseFunctionGEM->GetPRFPosition();
	double charge=_PadResponseFunctionGEM->GetPRFCharge();
	//make a copy of hit and change properties
	TrackerHitImpl* theHit = LCObjectCopier::copy( inputTrackerHit );
	double newpos[3];
	newpos[0]=pos[0];
	newpos[1]=pos[1];
	newpos[2]=inputTrackerHit->getPosition()[2];
	theHit->setPosition(newpos);
	if(hitflag::hasDeadChannel(inputTrackerHit->getQuality()) || hitflag::hasOverflowPulse(inputTrackerHit->getQuality()))
	  theHit->setEDep(charge);
	/*	
	//--------------ERROR part--------------------------
	if(hitflag::isAtModuleBorder(hit->getQuality()))
	std::cout<<"Hit at border"<<std::endl;
	if(hitflag::isNextToDeadChannel(hit->getQuality()))
	std::cout<<"Hit next to dead channel"<<std::endl;

	//error application for tracking
	// all hits get diffusion term depending on z position of hit
	// Determine the covariance matrix according to the given resolution
	double zdrift = _gearTPC->getMaxDriftLength() - (fabs(theHit->getPosition()[2]));
	// Sigma squared along Z
	double zSigmaSquare = _offsetz  +  _slopez * zdrift;
	// Sigma squared in RPhi plane
	double rphiSigmaSquare = _offsetx  +  _slopex * zdrift;
	
	// The phi angle of the hit.
	const gear::TPCModule &module = _gearTPC->getNearestModule(theHit->getPosition()[0],theHit->getPosition()[1]);
	gear::Vector2D correction = module.getOffset();
	//CHK (azimuthal) measurement direction = phi(hit) + pi/2 !
	double phi =  atan2(newpos[0] - correction[0], correction[1] - newpos[1]);
	float covMatrix[6] = {0., 0., 0., 0., 0., 0.};
	
	covMatrix[0] = pow(cos(phi), 2) * rphiSigmaSquare;
	covMatrix[1] = sin(phi) * cos(phi) * rphiSigmaSquare;
	covMatrix[2] = pow(sin(phi), 2) * rphiSigmaSquare;
	covMatrix[5] = float(zSigmaSquare);
	
	theHit->setCovMatrix(covMatrix);
	*/

	outputCol->addElement( theHit );         
	
      }
	evt->addCollection(outputCol, _outputColName);
  }



  void PRFHitBuilderProcessor::check(lcio::LCEvent*  /*evt*/)
  {
  }



  void PRFHitBuilderProcessor::end()
  {
    std::cout<<"Used PRF "<<_PRF_used<<" not used : "<<_PRF_notused<<" ( used in % : "
	     <<double(_PRF_used)/(_PRF_used+_PRF_notused)*100.<<" )"<<std::endl;
    std::cout<<"PRF fit failed: "<<_PRF_failed<<" ("<<double(_PRF_failed)/(_PRF_used)*100.<<" %)"<<std::endl;
    delete _PadResponseFunctionGEM;

  }



}// namespace marlintpc

