#include "PadResponseFunctionGEM.h"

#include "marlin/Global.h"

// the common flag word definitions in MarlinTPC
#include "FlagwordDefinitions.h"

//LCIO
#include <EVENT/TrackerPulse.h>

// GEAR
#include <gear/TPCModule.h>

//ROOT
#include "TF1.h"

#include <algorithm>
#include <math.h>
#include "streamlog/streamlog.h"


using namespace lcio;
using namespace std;


namespace marlintpc
{

  PadResponseFunctionGEM::PadResponseFunctionGEM(double defocussing, double diffusion,const gear::TPCParameters& gearparameter):
    _defocussing(defocussing),_diffusion(diffusion),
    _fixedWidth(0),_position(0), _positionRow(0), _positionRowLocal(0),
    _charge(0),_chargeErr(0),_status(false),
    _fitOK(true),_graph(0),
    _maxCharge(0),_min(0),_max(0),
    _tpcParameters(gearparameter)
  {
    _PRF      =new TF1("PRF","[0]*0+[1]*x+[2]*x^(1/2)+((1-[1])/2-[2]/sqrt(2))*(x*2)^(1/3)",0,0.5);
    _PRF->SetParameter(0,0);
    _PRFShift =new TF1("PRFShift", "[1]*0+[0]+[2]*x^(1/2)+(0.5-[0]-[2]/sqrt(2))*(x*2)^(1/3)",0,0.5);
    _PRFShift->SetParameter(1,0);
    _prfpos[0]=0;
    _prfpos[1]=0;
  }

  bool  PadResponseFunctionGEM::EvaluateHit(lcio::TrackerHit *hit, lcio::Track *track)
  {
    bool doPRF=false;
    bool fixCharge=false;
    _status=false;
    //check if PRF is needed:
    //If hit contains pulse in overflow or a dead channel use PRF with a fit
    if(hitflag::hasDeadChannel(hit->getQuality()) || hitflag::hasOverflowPulse(hit->getQuality()))
      {
	doPRF=true;
	fixCharge=true;
      }
     
    //If hit has less than 5 pulses use PRF calculation
    int npulses=hit->getRawHits().size();
    if(npulses < 5)
      doPRF=true;
   
    streamlog_out(DEBUG0)<<"-------------------------------"<<std::endl;
    streamlog_out(DEBUG0)<<"# pulses: "<<npulses<<" quality: "<<fixCharge<<std::endl;
    streamlog_out(DEBUG0)<<"Position in: "<<hit->getPosition()[0]<<" "<<hit->getPosition()[1]<<std::endl;

    if (doPRF)
      {
	//use parametrisation to fix width of gaussian fit
	CalculateFixedWidth(hit,track);
	if(fixCharge && npulses>2 && !_PRConly)
	  {
	    //prepare hit data
	    PrepareHitData(hit);
	    FitPRF(hit,fixCharge);
	  }
	else
	  CalculatePRC(hit);
	//check if PRF worked, if not fall back to KeepCentreOfGravityInfo
	//do some sanity checks on PRF result based on corrected position
	double distance=sqrt((hit->getPosition()[0]-_positionRow)*(hit->getPosition()[0]-_positionRow)+(hit->getPosition()[1]-_position)*(hit->getPosition()[1]-_position));
	//very crude sanity check if the corrected position is more than 1 mm away consider the PRF failed
	streamlog_out(DEBUG0)<<"Distance: "<<distance<<std::endl;
	if(distance>1)
	  _status=false;	
	if(! _status)
	  KeepCentreOfGravityInfo(hit);
      }
    else
      KeepCentreOfGravityInfo(hit);
    
    //_status will only tell if the prf has been carried out, but not if it was tried and failed
    //for that we have GetPRFSucceeded() function
    return _status;
  }

  void  PadResponseFunctionGEM::CalculateFixedWidth(lcio::TrackerHit *hit, lcio::Track * /*track*/)
  {
    //something along the lines of in global coordinates!
    
    //conversion of z-position into driftlength along z
    double L = _tpcParameters.getMaxDriftLength();
    _fixedWidth=sqrt(_defocussing+_diffusion*fabs(L-hit->getPosition()[2]));
    streamlog_out(DEBUG0)<<"---- Diffusion  fixed width = "<<_fixedWidth<<" at z="<<L-hit->getPosition()[2]<<std::endl;

    //need fixed width in local coordinates in projection along the row, which in polar coordinates is in phi
    const gear::TPCModule& theModule = _tpcParameters.getNearestModule(hit->getPosition()[0],hit->getPosition()[1]);
    gear::Vector2D localHitPos  = theModule.globalToLocal(hit->getPosition()[0],hit->getPosition()[1]);
    if((theModule.getLocalPadLayout()).getPadLayoutType() == 2)
      _fixedWidth= _fixedWidth/localHitPos[0];

    //plus eventually something based on track angle
  }
  
  void  PadResponseFunctionGEM::PrepareHitData(lcio::TrackerHit *hit)
  {
    if(_graph!=NULL)
      _graph->Clear();
    //container for location and charge 
    std::vector<double> pos,Q;
    _maxCharge=0;
    _maxPos=0;
    //everything will be done in local padplane coordinates until the end
    //loop the pulses
    lcio::LCObjectVec PulseVect = hit->getRawHits();
   
    for(unsigned int p=0;p<PulseVect.size();p++)
      {
	lcio::TrackerPulse* thePulse = dynamic_cast<lcio::TrackerPulse*>(PulseVect[p] );
	int currentModule = thePulse->getCellID1();
	const gear::TPCModule& theModule = _tpcParameters.getModule(currentModule);
	int padindex = thePulse->getCellID0();
	//get local coordinates of pad center
	gear::Vector2D localPadCenter, globalPadCenter;
	globalPadCenter = theModule.getPadCenter(padindex);
	localPadCenter  = theModule.globalToLocal(globalPadCenter[0], globalPadCenter[1]);
	//figure out in which coordinate the row is oriented
	//in case of polar coordinates all pads have same radius = first coordinate
	//in case of cartesian, in the local system rows should be along y
	if((theModule.getLocalPadLayout()).getPadLayoutType() == 2)
	  {
	    _positionRowLocal=localPadCenter[0];
	    //this is the coordinate along the row
	    pos.push_back(localPadCenter[1]);
	  }
	else
	  { 
	    _positionRowLocal=localPadCenter[1];
	    pos.push_back(localPadCenter[0]);
	  }

	Q.push_back(thePulse->getCharge());
	if(thePulse->getCharge()>_maxCharge)
	  {
	    _maxCharge=thePulse->getCharge();
	    if((theModule.getLocalPadLayout()).getPadLayoutType() == 2)
	      _maxPos=localPadCenter[1];
	    else
	      _maxPos=localPadCenter[0];
	  }
	streamlog_out(DEBUG0)<<theModule.getRowNumber(padindex)<<" " <<localPadCenter[0]<<" "<<localPadCenter[1]<< "  "<<thePulse->getCharge()<<std::endl;
      }
    _graph=new TGraph(pos.size(),&pos[0],&Q[0]);
    _min=*(std::min_element(pos.begin(),pos.end()));
    _max=*(std::max_element(pos.begin(),pos.end()));
  }
  
  void  PadResponseFunctionGEM::FitPRF(lcio::TrackerHit *hit, bool fixCharge)
  {
    _fitOK=false;
    streamlog_out(DEBUG0)<<"---- FitPRF start:  fixed width = "<<_fixedWidth<<" "<<_maxCharge<<" "<<_graph->GetMean()<<" "<<_min<<" "<<_max<<std::endl;
    TF1 *fit=new TF1("fit","gaus",_min,_max);
    //    fit->SetParLimits(2, _fixedWidth,_fixedWidth);
    fit->SetParameter(2, _fixedWidth);
    fit->SetParameter(1,_maxPos);
    fit->SetParameter(0,_maxCharge);
    //fix parameters with SetParLimits and B option in fit, but that fails
    int result =  _graph->Fit(fit,"NBQR");
    streamlog_out(DEBUG0)<<"---- FitPRF end ("<<result<<"):  fixed width = "<<fit->GetParameter(2)<<" "<<fit->GetParameter(0)<<" "<<fit->GetParameter(1)<<std::endl;
    if(fit->GetParameter(1)<_min || fit->GetParameter(1)>_max)
      result=1;
    if(result==0)
      {
	_status=true;
	_fitOK=true;
      }
    else
      streamlog_out(DEBUG0)<<"Fit failed: "<<result<<std::endl;
    //transform position back into global coordinate system
    const gear::TPCModule& theModule = _tpcParameters.getNearestModule(hit->getPosition()[0],hit->getPosition()[1]);
    gear::Vector2D globalPosition2D =  theModule.localToGlobal(_positionRowLocal,fit->GetParameter(1));
    if((theModule.getLocalPadLayout()).getPadLayoutType() == 2)
      {
	_positionRow=globalPosition2D[0];
	_position=globalPosition2D[1];
      }
    else
      {
	_positionRow=globalPosition2D[1];
	_position=globalPosition2D[0];
      }
    //integrate charge within 5 sigma
    //only change charge for hits with over-range or dead channel
    if(fixCharge)
      {
	_charge=fit->Integral(fit->GetParameter(1)-5*fit->GetParameter(2),fit->GetParameter(1)+5*fit->GetParameter(2));
	_chargeErr=fit->IntegralError(fit->GetParameter(1)-5*fit->GetParameter(2),fit->GetParameter(1)+5*fit->GetParameter(2));
      }
    else
      {
	_charge=hit->getEDep();
	_chargeErr=hit->getEDepError();
      }
    delete fit;
    streamlog_out(DEBUG0)<<"DONE with FIT"<<std::endl;
  }
  
  void  PadResponseFunctionGEM::CalculatePRC(lcio::TrackerHit *hit)
  {
    streamlog_out(DEBUG0)<<"CalculatePRC:"<<std::endl;
    const gear::TPCModule& theModule = _tpcParameters.getNearestModule(hit->getPosition()[0],hit->getPosition()[1]);
    int padindex = theModule.getNearestPad(hit->getPosition()[0],hit->getPosition()[1]);
    double padwidth=theModule.getPadWidth(padindex);
    gear::Vector2D localPadCenter,globalPadCenter, localHitPos, globalHitPos;
    globalPadCenter = theModule.getPadCenter(padindex);
    localPadCenter  = theModule.globalToLocal(globalPadCenter[0], globalPadCenter[1]);
    localHitPos = theModule.globalToLocal(hit->getPosition()[0],hit->getPosition()[1]);
    //convert _fixedWidth into units of padwidth. _fixedWidth should have already been converted to radian in case of polar 			coordinate system
    double sigma=_fixedWidth/padwidth;
    //calculate the position of the hit within the pad, results in a value between -0.5 and 0.5.
    double cog_pos=0;
    if((theModule.getLocalPadLayout()).getPadLayoutType() == 2)
      cog_pos = (localHitPos[1]-localPadCenter[1])/padwidth;
    else
      cog_pos = (localHitPos[0]-localPadCenter[0])/padwidth;

    streamlog_out(DEBUG0)<<"HitPos = "<<localHitPos[1]<<std::endl;
    streamlog_out(DEBUG0)<<"PadCenter = "<<localPadCenter[1]<<std::endl;
    streamlog_out(DEBUG0)<<"PadWidth = "<<padwidth<<std::endl;
	    
    double cog_pos_cor=0;
    int sign =(cog_pos>0?1:-1);
    cog_pos*=sign;
    //now the prametrisation from MultiFit:
    streamlog_out(DEBUG0)<<"sigma = "<<sigma<<std::endl;
    if (sigma>0.128)
      {
	_PRF->SetParameter(1,(9.52257-141.994*sigma+739.665*sigma*sigma-1739.82*sigma*sigma*sigma
			      +1936.17*sigma*sigma*sigma*sigma-832.538*sigma*sigma*sigma*sigma*sigma));
	_PRF->SetParameter(2,(-21.2726+292.479*sigma-1399.78*sigma*sigma+3126.27*sigma*sigma*sigma
			      -3350.91*sigma*sigma*sigma*sigma+1398.15*sigma*sigma*sigma*sigma*sigma));
	cog_pos_cor=sign*(_PRF->Eval(cog_pos));
      }
    else
      {
	//< flat region
	_PRFShift->SetParameter(0,0.499*(1-sigma/0.128));
	_PRFShift->SetParameter(2,-4.21678*sigma);
	cog_pos_cor=sign*(_PRFShift->Eval(cog_pos));
      }
    streamlog_out(DEBUG0)<<"In/Out: "<<cog_pos<<" "<<cog_pos_cor<<std::endl;
   	
    //transform back from -0.5 to 0.5 range to real position on pad
    double cog_pos_new=0;
    if((theModule.getLocalPadLayout()).getPadLayoutType() == 2)
      cog_pos_new=cog_pos_cor*padwidth+localPadCenter[1];
    else
      cog_pos_new=cog_pos_cor*padwidth+localPadCenter[0];
    //transform back to global
    if((theModule.getLocalPadLayout()).getPadLayoutType() == 2)
      globalHitPos = theModule.localToGlobal(localHitPos[0],cog_pos_new);
    else
      globalHitPos = theModule.localToGlobal(cog_pos_new,localHitPos[1]);
    
    _positionRow=globalHitPos[0];
    _position=globalHitPos[1];
  
    //do not touch charge
    _charge=hit->getEDep();
    _chargeErr=hit->getEDepError();
    _status=true;
  }

  void  PadResponseFunctionGEM::KeepCentreOfGravityInfo(lcio::TrackerHit *hit)
  {
    streamlog_out(DEBUG0)<<"KeepCentreOfGravityInfo"<<std::endl;
    _positionRow=hit->getPosition()[0];
    _position=hit->getPosition()[1];
    _charge=hit->getEDep();
    _chargeErr=hit->getEDepError();

  }

  double* PadResponseFunctionGEM::GetPRFPosition()
  {
    _prfpos[0]=_positionRow;
    _prfpos[1]=_position;
    streamlog_out(DEBUG0)<<"Position out: "<<_positionRow<<" "<<_position<<std::endl;
    return _prfpos;
  }

}//namespace
