/*
 * RowCutApplicationProcessor.cc
 *
 *      Author: ljungren
 */

#include "RowCutApplicationProcessor.h"

#include <sstream>

// LCIO
#include <EVENT/LCCollection.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <EVENT/TrackerHit.h>
#include <IMPL/LCFlagImpl.h>
#include <IMPL/TrackerPulseImpl.h>

//MARLIN
#include "marlin/Global.h"
#include "marlin/Exceptions.h"
#include "FlagwordDefinitions.h"

// GEAR
#include "gear/TPCParameters.h"
#include "gear/TPCModule.h"

namespace marlintpc
{
const std::pair<int, int> RowCutApplicationProcessor::ALLROWS(-1,-1);

RowCutApplicationProcessor aRowCutApplicationProcessor;


RowCutApplicationProcessor::RowCutApplicationProcessor()
: Processor("RowCutApplicationProcessor")
{

    /*
     * Processor for removing hits from user specified rows (in user specified pad modules).
     *
     *
     */
    _description = "Processor for removing hits from user specified rows (in user specified pad modules).";

    registerInputCollection(LCIO::TRACKERHIT,
            "InputCollectionName",
            "Name of the input collection",
            _inputColName,
            std::string("TPCHits"));

    registerOutputCollection(LCIO::TRACKERHIT,
            "OutputCollectionName",
            "Name of the output collection",
            _outputColName,
            std::string("TPCHitsCut"));

    registerProcessorParameter("SetOutputTransient",
            "if not true the output collection is set transient. default: true",
            _outputIsTransient, false);

    IntVec exampleRow;
    exampleRow.push_back(ALLROWS.first);
    exampleRow.push_back(ALLROWS.second);
    registerProcessorParameter("SkipRow",
            "Define the row to skip. The row must be defined as a pair of integers separated by a space. A negative integer indicates any module/row. Default: -1 -1 (Every row is cut away)",
            _rows2skip,
            exampleRow,
            exampleRow.size());

    //Modifying the event is frowned upon and only works on collections already written to file not the ones created by processors
    //running in the same steering file as this one.
    //Therefore the modify option is removed and the output collection is mandatory
    /*
    registerOptionalParameter("isSubset",
            "If true the output collection will be stored as a subset of the input collection, otherwise the hits not selected will be flagged not to be used in the track finding. Default: true",
            _isSubset, true);
    
    registerOptionalParameter("ActivateFlagging",
            "If true the hit of the input collection that are in the rows to cut will be flagged as not to be added in the track finding. WARNING: Turning this feature on will modify the input collection",
            _flaggingON, false);
    */

}

void
RowCutApplicationProcessor::init()
{

    printParameters();

    if((_rows2skip.size() % 2) != 0)
    {
        streamlog_out(ERROR) << "You must specify a list of pair of modules and rows. There seem to be an odd number of inputs." << std::endl;
        throw marlin::ParseException("The parameter list seems to be incorrect");
    }

    _skipList.clear();

    for(IntVec::iterator it = _rows2skip.begin(); it != _rows2skip.end(); it= it+2)
    {
        int mod = *it <= -1 ? ALLROWS.first : *it;
        int row = *(it+1)<= -1 ? ALLROWS.second : *(it+1);
        std::pair<int, int> rowid(mod, row);
        _skipList.insert(rowid);
        streamlog_out(DEBUG) << "Cutting Row " << *(it +1)  << " in module " << *it <<std::endl;
    }

    streamlog_out(DEBUG) << "Initialization completed" << std::endl;
}

void
RowCutApplicationProcessor::processRunHeader(lcio::LCRunHeader* run)
{
    run->parameters().setValue(_processorName + "_revision", "$Rev: 1842 $");

    for (marlin::ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
    {
        if (!i->second->isOptional() || i->second->valueSet())
            run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
    }
    streamlog_out(DEBUG) << "Run Header processed" << std::endl;

}

void
RowCutApplicationProcessor::processEvent(lcio::LCEvent * evt)
{
  //flagging is switched off for now as a new collection has to be written anyways
  _flaggingON=false;
   //Output the event number on multiples of 1000
    int eventNumber = evt->getEventNumber();
    if((eventNumber % 1000) == 0)
    {
        streamlog_out(MESSAGE) << "Event: " << eventNumber << std::endl;
    }
    //Reads the input collection
    LCCollectionVec* inputCol = 0;
    try
    {
        inputCol = dynamic_cast<LCCollectionVec*>(evt->getCollection(_inputColName));
    }  //end of try
    catch (DataNotAvailableException &e)
    {
        // if there is not input collection give a message to the user
        streamlog_out(ERROR) << " processEvent:" << " input collection " << _inputColName << " cannot be found in event " << evt->getEventNumber()
                                                    << std::endl;
        return;
    }

    //Creation of the output collection
    //The output collection is always created
    LCCollectionVec* outputCol = 0;
    streamlog_out(MESSAGE) << "A new subset collection named " << _outputColName << " is being created";
    outputCol = new LCCollectionVec(LCIO::TRACKERHIT);
    outputCol->setTransient(_outputIsTransient);
    // the output collection is a subset of the input collection
    // only the pointers are stored, the hits are not copied
    _isSubset=true;
    outputCol->setSubset(_isSubset);
    
    /*
     * If the user selected the subset mode the algorithm will search for the hits
     * to be accepted and load them in the subset collection
     * If the flagging mode is selected the algorithm searches for the hits to be
     * cut away flagging them
     */
    for (LCCollectionVec::iterator hitIter = inputCol->begin(); hitIter < inputCol->end(); hitIter++)
    {
        TrackerHitImpl* hit = 0;
        TrackerPulseImpl* pulse = 0;
        try
        {
            hit = dynamic_cast<TrackerHitImpl*>(*hitIter);
            pulse = dynamic_cast<TrackerPulseImpl*>(*(hit->getRawHits().begin()));
        }
        catch(EVENT::Exception &e)
        {
            streamlog_out(WARNING) << "Unable to access data in the hit collection" << std::endl;
            return;
        }

        if(_skipList.find(ALLROWS) != _skipList.end())
        {
            if(_flaggingON)
            {
                int quality = hit->getQuality();
                hitflag::setRemoveFromTrackFinding(quality);
                hit->setQuality(quality);
            }
        }
        else
        {
            const gear::TPCParameters& tpcParameters = marlin::Global::GEAR->getTPCParameters();
            const gear::TPCModule& theModule = tpcParameters.getModule(pulse->getCellID1());
            int row = theModule.getRowNumber(pulse->getCellID0());
            int mod = pulse->getCellID1();
	    bool _hitToFlag = true;
            if(_skipList.find(std::pair<int,int> (ALLROWS.first, row)) == _skipList.end())
            {
	      if (_skipList.find(std::pair<int,int> (mod, ALLROWS.second)) == _skipList.end())
                {
		  if (_skipList.find(std::pair<int,int> (mod, row)) == _skipList.end())
                    {
		      if(_isSubset)
			outputCol->addElement(hit);
		      _hitToFlag = false;
                    }
                }
            }

            if(_hitToFlag && _flaggingON)
            {
                int quality = hit->getQuality();
                hitflag::setRemoveFromTrackFinding(quality);
                hit->setQuality(quality);
            }//end iftoskip
        }//end if not all rows are to skip
    }//end for loop

    if(outputCol!= 0 && _isSubset)
    {
      evt->addCollection(outputCol, _outputColName);
    }
}

}  // namespace marlintpc
