#include "SignalSplitter.h"
#include <iostream>

namespace marlintpc
{

std::vector< std::pair< size_t, std::vector<float> > >
SignalSplitter::splitSignal(std::vector<float> const & inputSignal, double noise)
{
    // first search for the maxima in the signal
  _maxima = findMaxima(inputSignal, noise);

  //std::cout << "#found " << _maxima.size() << " maxima" << std::endl;

    // now split according to number of maxima
    switch (_maxima.size())
        {
        case 1: // as long as mulit-splitting is not implemented don't split at all
            _splitSignals.resize(1);
            _splitSignals[0].first = 0;
            _splitSignals[0].second = inputSignal;
            break;
        case 2:
            splitSignalTwoPeaks(inputSignal);
            break;
        default:
	  splitSignalAtMinima(inputSignal, noise);
        }

    return _splitSignals;
}

std::vector< size_t >  SignalSplitter::findMaxima(std::vector<float> const & inputSignal, double noise)
{
    std::vector< size_t > retVal;
    float lastValue = 0;
    float lastMinimum = 0;
    
    bool risingEdge = true;
    for (size_t i = 0; i < inputSignal.size(); ++i)
        {
	  // std::cout<<"T: "<<risingEdge<<" "<<inputSignal[i]<<" "<<lastValue<<" "<<lastMinimum<<std::endl;
	  if (inputSignal[i] >= lastValue)
                {
		  //just starting to rise again after minimum
		  if(risingEdge==false)
		    lastMinimum=lastValue;
		  risingEdge = true;
                    if (i == inputSignal.size() - 1)
                        {
                            // this is the last sample, and it's larger than the previous. This is a maimum
			  if(noise >0 && inputSignal[i]-lastMinimum>noise)
			    retVal.push_back(i);
			  else //do as before without noise
			    retVal.push_back(i); 
                            //  std::cout << "pushing back " << i << std::endl;
                        }
                }
            else
                {
                    if (risingEdge == true)
                        {
			   // the previous value was larger, so it's the maximum
			  //std::cout<<"B: "<<risingEdge<<" "<<inputSignal[i]<<" "<<lastValue<<" "<<lastMinimum<<std::endl;
			  if(noise >0 && inputSignal[i]-lastMinimum>noise)
			    {
			      //its a maximum above noise level
			      retVal.push_back(i - 1);
			      //std::cout << "pushing back " << i-1 << " "<<inputSignal[i]<<" "<<lastMinimum<<std::endl;
			    }
			  else
			    retVal.push_back(i - 1);
                            risingEdge = false;
                        }
                }
            lastValue = inputSignal[i];
            //    std::cout << "setting lastValue to " << lastValue << std::endl;
        }

    return retVal;
}

//  std::vector< std::pair< size_t, std::vector<float> > > SignalSplitter::splitSignalTwoPeaks( std::vector<float> const & inputSignal)
void SignalSplitter::splitSignalTwoPeaks(std::vector<float> const & inputSignal)

{
    /*
     *     1   2
     *     1 M 2
     *   A 1 M 2 B
     * A A 1 M 2 B B
    */

    _splitSignals.resize(2);

    double sumM = 0;
    for (size_t i = _maxima[0] + 1 ; i < _maxima[1] ; i++)
        {
            sumM += inputSignal[i];
        }
    //std::cout << "sumM=" << sumM << std::endl;

    size_t widthOfM = _maxima[1] - _maxima[0] - 1 ;
    size_t widthOfA = _maxima[0];
    size_t widthOfB = inputSignal.size() - _maxima[1] - 1;
    //    std::cout << "widthes" << widthOfA << " " << widthOfM << " "<< widthOfB << std::endl;

    // There are a few cases this algorithm does not perform well
    // 1. The first or the last bin is the maximum, so the the width of A or B is zero
    //    In this case the normalisation does not work.
    // 2. The distance between the maxima is larger than the sum of the widthes of A and B
    //    The pulses should not mix if the widthes are ok, but they do...
    //
    // In these cases we can just split at the minimum

    if ((widthOfA == 0) || (widthOfB == 0) || (widthOfA + widthOfB < widthOfM))
        {
            //std::cout << "Splitting at minimum instead of using overlap method"
            //          << std::endl;

            splitSignalAtMinima(inputSignal);

            return;
        }

    // perform the normal spliting
    double sumAInM = 0;
    for (int i = _maxima[0] - 1 ; (i >= static_cast<int>(_maxima[0] - widthOfM)) && (i >= 0); --i)
        {
            //      std::cout << "summing A in M , i=" << i << std::endl;
            sumAInM += inputSignal[i];
        }
    //std::cout << "sumAInM=" << sumAInM << std::endl;

    double sumBInM = 0;
    for (size_t i = _maxima[1] + 1; (i <= (_maxima[1] + widthOfM))  && (i < inputSignal.size()) ; ++i)
        {
            sumBInM += inputSignal[i];
        }
    //std::cout << "sumBInM=" << sumBInM << std::endl;

    double max1 = inputSignal[_maxima[0]]; // the height of the first maximum
    double max2 = inputSignal[_maxima[1]]; // the height of the second maximum

    // weigh the halfs A and B so that the following is true
    // 1.) the fraction of A' in M is the same as the first maximum to the sum of the maxima
    //     (sumAInM * f1) / (sumBInM * f2) = max1 / max2
    // 2.) A' in M and B' in M sum up to the original sum in M
    //     (sumAInM * f1) + (sumBInM * f2) = sumM

    // resolving these equations to f1 and f2 one gets

    double f1 = sumM / sumAInM * max1 / (max1 + max2);
    double f2 = sumM / sumBInM * max2 / (max1 + max2);
    // std::cout << "f1=" << f1<< " , f2="<<f2 << std::endl;

    double firstMaximumCorrected = 0;
    // check whether to correct the first maximum for a contribution from second signal
    if (widthOfB > widthOfM)
        {
            // the first maximum also contains a contribution from the second signal, correct for it
            firstMaximumCorrected = inputSignal[_maxima[0]] - f2 * inputSignal[_maxima[1] + widthOfM + 1];
            //      std::cout << "fm1 = " << firstMaximumCorrected << std::endl;

            // if the corrected value is below 0 don't correct it
            if (firstMaximumCorrected < 0) firstMaximumCorrected = 0;

            if (_maxima[0] > 0)
                {
                    // check for the bin before and after the maximum
                    if ((firstMaximumCorrected < inputSignal[_maxima[0] - 1]) &&
                            (firstMaximumCorrected < inputSignal[_maxima[0] + 1]))
                        {
                            // the previous maximum now realy would be a minimum, which is suspicious
                            // don't correct
                            firstMaximumCorrected = 0.;
                        }
                }
            else // no bin before the maximum, check the bin after maximum
                {
                    if (firstMaximumCorrected < inputSignal[_maxima[0] + 1])
                        {
                            // the previous maximum now realy would be a minimum, which is suspicious
                            // don't correct
                            firstMaximumCorrected = 0.;
                        }
                }
        }
    //std::cout << "firstMaxCorrected"<< firstMaximumCorrected << std::endl;

    double secondMaximumCorrected = 0;
    // check whether to correct the first maximum for a contribution from second signal
    if (widthOfA > widthOfM)
        {
            // the first maximum also contains a contribution from the second signal, correct for it
            secondMaximumCorrected = inputSignal[_maxima[1]] - f1 * inputSignal[_maxima[0] - widthOfM - 1];

            // if the corrected value is below 0 don't correct it
            if (secondMaximumCorrected < 0) secondMaximumCorrected = 0;

            if (_maxima[1] < (inputSignal.size() - 1))
                {
                    // check for the bin before and after the maximum
                    if ((secondMaximumCorrected < inputSignal[_maxima[1] - 1]) &&
                            (secondMaximumCorrected < inputSignal[_maxima[1] + 1]))
                        {
                            // the previous maximum now realy would be a minimum, which is suspicious
                            // don't correct
                            secondMaximumCorrected = 0.;
                        }
                }
            else // no bin after the maximum, check the bin after maximum
                {
                    if (secondMaximumCorrected < inputSignal[_maxima[0] - 1])
                        {
                            // the previous maximum now realy would be a minimum, which is suspicious
                            // don't correct
                            secondMaximumCorrected = 0.;
                        }
                }
        }
    //std::cout << "secondMaxCorrected"<< secondMaximumCorrected << std::endl;

    // the signal will only be taken up to the next maximum
    size_t widthOfFirstSignal = widthOfA + 1 + (widthOfM < widthOfA ? widthOfM : widthOfA) ;
    // increase the width by one if the second maximum has been corrected
    // for a contribution from the first signal
    if (secondMaximumCorrected != 0) widthOfFirstSignal++;

    size_t widthOfSecondSignal = (widthOfM < widthOfB ? widthOfM : widthOfB) + 1 + widthOfB ;
    // increase the width by one if the first maximum has been corrected
    // for a contribution from the second signal
    if (firstMaximumCorrected != 0) widthOfSecondSignal++;

    std::vector<float> &firstSignal = _splitSignals[0].second;
    firstSignal.resize(widthOfFirstSignal);
    // first fill the A part of the pulse
    for (size_t i = 0; i < _maxima[0]; i++)
        {
            firstSignal[i] = inputSignal[i];
        }

    // fill the (possibly corrected) maximum
    if (firstMaximumCorrected != 0)
        {
            firstSignal[_maxima[0]] = firstMaximumCorrected;
        }
    else
        {
            firstSignal[_maxima[0]] = inputSignal[_maxima[0]];
        }

    // fill the part between the pulses
    // the widthOfSignal already contains the information if the second maximum is included or not
    for (size_t i = 1; i < widthOfFirstSignal - _maxima[0]; ++i)
        {
            firstSignal[_maxima[0] + i] = inputSignal[_maxima[0] - i] * f1;
        }

    // now the second signal
    std::vector<float> & secondSignal = _splitSignals[1].second;
    secondSignal.resize(widthOfSecondSignal);

    size_t positionOfMaxInSecondPulse = _maxima[1] + widthOfSecondSignal - inputSignal.size();
    // = widthOfSecondSignal - ( inputSignal.size() -  _maxima[1] )
    //    std::cout << "positionOfMaxInSecondPulse = " << positionOfMaxInSecondPulse<< std::endl;

    // fill the parts between the pulses
    for (size_t i = 0 ; i < positionOfMaxInSecondPulse ; ++i)
        {
            secondSignal[i] =
                inputSignal[_maxima[1] + (widthOfSecondSignal - 1 - widthOfB) - i ] * f2;
            // widthOfM or widthOfM+1
            // or widthOfB if widthOfB<widthOfM
        }

    // fill the maximum
    if (secondMaximumCorrected != 0)
        {
            secondSignal[positionOfMaxInSecondPulse] = secondMaximumCorrected;
        }
    else
        {
            secondSignal[positionOfMaxInSecondPulse] = inputSignal[_maxima[1]];
        }

    // fill the B part
    for (size_t i = 1 ; i <= widthOfB ; ++i)
        {
            secondSignal[positionOfMaxInSecondPulse+i] = inputSignal[_maxima[1] + i];
        }

    _splitSignals[0].first = 0; // the index of the first pulse, always 0
    _splitSignals[1].first = inputSignal.size() - secondSignal.size(); // the index of the second pulse
}

//  void SignalSplitter::splitSignalAtMinima( std::vector<float> const & inputSignal)
//  {
//    _splitSignals.resize( _maxima.size());
//
//    // store the minima in a vector
//    std::vector<size_t> minima;
//
//    for ( unsigned int minimumIndex = 0; minimumIndex < _maxima.size()-1 ; minimumIndex++)
//    {
//      // find the minimum between the two maxima around the minimum
//      // start at the one of the maxima
//      size_t minPosition = _maxima[minimumIndex];
//      float minValue = inputSignal[minPosition];
//
//      for (size_t i = _maxima[minimumIndex]+1 ; i < _maxima[minimumIndex+1] ; i++)
//      {
//  if (inputSignal[i] <minValue)
//  {
//    minPosition = i;
//    minValue = inputSignal[i];
//  }
//      }
//
//      //now that the minimum is found store it in the vector
//      minima.push_back(minPosition);
//    }
//
//    // prepare the first signal, needs special treatment because it starts at 0
//    _splitSignals[0].first=0;
//    _splitSignals[0].second.resize(minima[0]+1);
//
//    // prepare the other signals
//    for (size_t maxIndex = 1; size_t maxIndex < (_maxima.size()-1) ; maxIndex++ )
//    {
//      _splitSignals[maxIndex].first=minima[maxIndex-1];
//      _splitSignals[maxIndex].second.resize(minima[maxIndex]-minima[maxIndex-1]+1);
//      // +1 because both minima are part of the signal
//    }
//
//    // the last signal again needs special treatment
//    _splitSignals[_maxima.size()-1].first=minima[minima.size()-1];// starts at last minimum
//    _splitSignals[_maxima.size()-1].second.resize(inputSignal.size()-minima[minima.size()-1]);
//
//    // fill the first signal up to the minimum
//    for (size_t i = 0 ; i < minima[0] ; i++)
//    {
//      _splitSignals[0].second[i]= inputSignal[i];
//    }
//
//    _splitSignals[1].first=minPosition;
//    _splitSignals[1].second.resize(inputSignal.size()-minPosition);
//
//    // fill the first pulse up to the minimum
//    for (size_t i = 0 ; i < minPosition ; i++)
//    {
//      _splitSignals[0].second[i]= inputSignal[i];
//    }
//
//    // fill the other signals
//    for (size_t maxIndex = 1; size_t maxIndex < (_maxima.size()-1) ; maxIndex++ )
//    {
//      // start with the minimum itself
//      // the minimum is still missing in the previous signal
//      size_t previousSignalSize =  _splitSignals[maxIndex-1].second.size();
//      _splitSignals[maxIndex-1].second[previousSignalSize-1] = inputSignal[ minima[maxIndex-1] ] /2.;
//      _splitSignals[maxIndex].second[0] = inputSignal[ minima[maxIndex-1] ] /2.;
//
//      // fill the pulse up to the next minimum
//      for (size_t i = 1 ; i < (minima[maxIndex]-minima[maxIndex-1]) ; i++)
//      {
//  _splitSignals[0].second[i]= inputSignal[i];
//      }
//
//      _splitSignals[maxIndex].first=minima[maxIndex-1];
//      _splitSignals[maxIndex].second.resize(minima[maxIndex]-minima[maxIndex-1]+1);
//      // +1 because both minima are part of the signal
//    }
//
//    // split the minimum to both pulses
//    _splitSignals[0].second[minPosition] = minValue/2.;
//    _splitSignals[1].second[0] = minValue/2.;
//
//    // fill the rest of the second pulse
//    for (size_t i = 1 ; i < inputSignal.size()-minPosition ; i++)
//    {
//      _splitSignals[1].second[i]= inputSignal[minPosition+i];
//    }
//
//  }// splitSignalAtMinima
//

  void SignalSplitter::splitSignalAtMinima(std::vector<float> const & inputSignal, double noise)
{
    _splitSignals.clear();

    // this part of the algorithm is much clearer with iterators since it is not necessary to
    // read from several vectors using the same index

    std::vector< std::vector<float>::const_iterator > minima;
    //if there is a noise threshold a real minima has to stick out above the noise

    for (std::vector<size_t>::iterator maxIter = _maxima.begin();
            maxIter < (_maxima.end() - 1); maxIter ++)
        {
            // start search at maximum
            std::vector<float>::const_iterator minIter =  inputSignal.begin() + (*maxIter);
            float minValue = *minIter;
            for (std::vector<float>::const_iterator inputIter = inputSignal.begin() + *maxIter ;
                    inputIter <  inputSignal.begin() + *(maxIter + 1) ; inputIter++)
                {
		  if (*inputIter < minValue + noise)
                        {
                            minValue = *inputIter;
                            minIter = inputIter;
                        }
                }
	    // save the minimum
            minima.push_back(minIter);
        }

    std::pair< size_t, std::vector<float> > aSplitSignal;
    aSplitSignal.first = 0;
    std::vector< std::vector<float>::const_iterator >::iterator nextMinimum = minima.begin();

    // fill the split signal. At the minimum push it to the _splitSignals collection
    for (std::vector<float>::const_iterator signalIter = inputSignal.begin() ;
            signalIter < inputSignal.end() ; signalIter++)
        {
            // we are at the next minimum, split this charge between two signals
            if ((nextMinimum != minima.end()) && (signalIter == *nextMinimum))
                {
                    // only half of the signal goes into this signal
                    aSplitSignal.second.push_back(*signalIter / 2.);
                    // signal complete, store it
                    _splitSignals.push_back(aSplitSignal);

                    // clear the signal and set the new starting point
                    aSplitSignal.first = std::distance(inputSignal.begin(), signalIter);
                    aSplitSignal.second.clear();
                    // add the second half of the charge to this signal
                    aSplitSignal.second.push_back(*signalIter / 2.);
                    // increase minimum iterator
                    nextMinimum++;
                }
            else
                {
                    // just add the charge to the signal
                    aSplitSignal.second.push_back(*signalIter);
                }
        }
    // finally add the last signal
    _splitSignals.push_back(aSplitSignal);
}


}//namespace MarlinTPC
