#ifndef NEIGHBOURING_PAD_HANDLER_H
#define NEIGHBOURING_PAD_HANDLER_H

#include <vector>
#include <map>

namespace marlintpc
{

    /**
     *  A helper class which stores information about the neighbouring pads of a specific pad.
     *  It allows to obtain the preceding and successing pad indices separately which adjoin the given pad.
     *
     *  @author Oliver Sch&aumlfer
     */
    class NeighbouringPadsObject
	{

		public:

			/// Default Constructor, needed for pointers.
			NeighbouringPadsObject();
			
			/// Copy Constructor, needed for copy of vectors? Maybe not.
			NeighbouringPadsObject(const NeighbouringPadsObject& originalObject);
			
			/// Constructor. The maps with information about neighbours are initialized here.
			NeighbouringPadsObject(const int padIndex);
			
			/// Returns a vector with all the neighbouring pads with a smaller pad index.
			std::vector<int> getPrecedingPads() const;
			
			/// Returns a vector with all the neighbouring pads with a larger pad index.
			std::vector<int> getSucceedingPads() const;
			
			/// Adds a Pad Index to either the vector of preceding or successing pads.
			void addNeighbouringPad(const int padIndex);

			/// Set the x- or phi-coordinates of the corners of the pad.
			void setPadSideRowCoordinates(double cLeft, double cRight);

			/// Returns the x- or phi-coordinate of the left side of the pad.
			double getLeftSide() const;

			/// Returns the x- or phi-coordinate of the right side of the pad.
			double getRightSide() const;
			
            /// Returns the pad index of the pad.
            int getPadIndex() const;

		protected:
		
			int _thisPadsPadIndex;

			double _leftRowCoordinate; ///< x or phi-coordinate of the left corner of the pad
			double _rightRowCoordinate; ///< x or phi-coordinate of the right corner of the pad
			
			std::vector<int> _precedingPads;
			std::vector<int> _succeedingPads;
			
	};

	class NeighbouringPadHandler
    {
        public:

			/// Constructor. The maps with information about neighbours are initialized here.
			NeighbouringPadHandler();

			/// The test function. It returns true, if 
			bool areNeighbouringPads(int moduleIndex, int padIndex1, int padIndex2) const;

			/// Returns the smallest pad index of all the neighbouring pads of the given one.
			/// This can be used to break the search early, if objects are sorted by pad index.
			int getSmallestNeighbouringPadIndex(int moduleIndex, int padIndex) const;
			
			/// Returns the largest pad index of all the neighbouring pads of the given one.
			/// This can be used to break the search early, if objects are sorted by pad index.
			int getLargestNeighbouringPadIndex(int moduleIndex, int padIndex) const;

			/// Returns a vector with all the neighbouring pads with a smaller pad index.
			std::vector<int> getPrecedingNeighbouringPads(int moduleIndex, int padIndex) const;
			
			/// Returns a vector with all the neighbouring pads with a larger pad index.
			std::vector<int> getSucceedingNeighbouringPads(int moduleIndex, int padIndex) const;

			/// Returns a pointer to the NeighbouringPadsObject of the specified pad- and module index.
			NeighbouringPadsObject* getNeighbouringPadsObject(int moduleIndex, int padIndex);

		protected:

			std::map<int, std::map<int, NeighbouringPadsObject> > _endPlateMap; ///< Map of all pads of the GEAR geometry. First index is for modules, second for pads.

    };
} // namespace marlintpc

#endif // NEIGHBOURING_PAD_HANDLER_H
