#ifdef USE_LCCD

#ifndef PHOTODOTRECONSTRUCTIONPROCESSOR_H_
#define PHOTODOTRECONSTRUCTIONPROCESSOR_H_ 1

//Marlin
#include <marlin/Processor.h> // Integral for any processor

//gear
#include <gear/PadRowLayout2D.h>
#include <gearimpl/GlobalPadIndex.h>

//lcio
#include <IMPL/TrackerPulseImpl.h>

//marlinTPC
#include "HitCandidate.h"

namespace marlintpc
{


/** Analyses the effect of variable magnetic field on photodot positions
 *
 *  PhotodotReconstructionProcessor uses the electron cloud from a laser pulse
 *  in the TPC to examine the effect of the magnetic field.
 *  Locations of the centres of images of photodots on the readout
 *  electronics are compared with the positions of dots on the cathode.
 *
 * <h4>Prerequisites</h4>
 * A TrackerPulse collection containing charges from a number of photodots
 *
 * @author P. Conley (UVic)<br>
 */
class PhotodotReconstructionProcessor : public marlin::Processor
{

    public: // Functions for interaction with Marlin

        /** Functions used directly by Marlin
         */
        virtual Processor * newProcessor()
        {
            return new PhotodotReconstructionProcessor ;
        }
        PhotodotReconstructionProcessor();
        ~PhotodotReconstructionProcessor();

        /** Called at the beginning of the job to initialise the processor
         */
        virtual void init();

        /** Called for every run
         */
        virtual void processRunHeader(lcio::LCRunHeader * run);

        /** Called for every event - the body of the processor
         */
        virtual void processEvent(lcio::LCEvent * evt);

        /** Called after the event has been processed
         */
        virtual void check(lcio::LCEvent * evt);

        /** Called after data processing for clean-up
         */
        virtual void end();

    private:

        /** Structure holding the information about photodots from the
         *  geometry_file. The diameter won't be used, but exists in the data.
         */
        struct Photodot
        {
            // Cartesian constructor
            Photodot(double x, double y, double diameter) :
                    _x(x),
                    _y(y),
                    _r(sqrt(x*x + y*y)),
                    _phi(atan2(y, x)),
                    _diameter(diameter) {};
            // Polar constructor
            /* The extra integer argument serves simply to differentiate the two
             * constructor types. Its value is irrelevant.
             */
            Photodot(double r, double phi, double diameter, int  /*alt_type*/) :
                    _x(r * cos(phi)),
                    _y(r * sin(phi)),
                    _r(r),
                    _phi(phi),
                    _diameter(diameter) {};
            double _x;
            double _y;
            double _r;
            double _phi;
            double _diameter;
        };

        /** Functions to keep the code clean while finding the hit centre
         */

        /** Find the hit nearest the position of a photodot, by comparing distances
         *  to pulses.
         *  Returns a cellID0 and cellID1 inside the hit.
         *  Operates on a photodot
         *  \parameter thisPhotodot The input photodot
         *  \parameter module Here CellID1 is returned
         *  \parameter pad Here CellID0 is returned
         */
        void findNearestHit(Photodot thisPhotodot, int &module, int &pad);

        /** Find the centre of a hit containing a given pulse, under the assumption
         *  that deposited charge displays a Gaussian profile, and pads at the centre
         *  of a hit receive the greatest charge.
         *  Returns the cellID of the photodot at the bottom of the centre column.
         *  Operates on a pulse somewhere in a hit, and a container of pulses.
         */
        int findCentreColumn(int module, int pad);

        /** Find the precise co-ordinate of the centre of a hit.
         *  Returns the coordinates of the hit centre.
         *  Operates on a pulse at the bottom of the centre column of a hit (on the given module)
         *   and a collection of pulses.
         *  Assumes a gaussian distribution on each row and therefore needs to know the expected diffusion.
         */
        gear::Vector2D findGausHitCentre(int module, int columnCentre);

        /** Find the precise co-ordinate of the centre of a hit.
         *  Returns the coordinates of the hit centre.
         *  Operates on a pulse at the bottom of the centre column of a hit (on the given module)
         *   and a collection of pulses.
         *  Calculates the charge weighted mean taking all pad in the cluster into account.
         *  Therefore it is not done row wise!
         */
        gear::Vector3D findMeanHitCentre(int module, int columnCentre, float* covMatrix);

        /**Calculate the centre of the hit.
         * This is done along the row like in the RowBasedHitFinderProcessor.
         * The same algorithm is used for the position transverse to the rows.
         * The position in z corresponds to the arrival time of the pulse with the
         * highest charge.
         */
        gear::Vector3D calculatePositionsAndErrors(const HitCandidate &cand, float* covMatrix);
        /** Determine which pulses arrived at the readout electronics as part of
         *  the clouds resulting from the laser pulse.
         *  No returned value (operations performed on input collection)
         *  Operates on the the photodots and input pulses
         *  The ElectronClouds are used to filter pulses in the correct time range.
         */
        void findPhotodotPulses(LCCollection* electron_cloud_collection);

        /** Calculate the total charge from multiple pulses incident on one pad.
         *  This isn't an optimal solution, as there should be only one valid
         *  pulse on a pad, but there's no obvious way around the problem.
         *  Returns the sum of all charges on the pad.
         *  Operates on the index of a pad.
         */
        double getPadCharge(int module, int padIndex);

        /** The name of the file containing the positions of the photodots.
         */
        std::string _geometry_filename;

        /** The name of the collection containing the photoelectric tracker pulses.
         */
        std::string _input_pulses_collection_name;

        /** The name of the collection containing the expected photodot positions after the drift.
         */
        std::string _input_hitPosition_collection_name;

        /** The name of the output collection which are TrackerHits with the
         * photodot position
         */
        std::string _output_reconstuctedPhototdots_collection_name;

        /** The name of the output collection which are TrackerHits with the
         * photodot position
         */
        std::string _output_photodot_collection_name;

        /** The positions of the photodots.
         *  This vector is built from the _geometry_filename, so is not initialised
         *  with the processor
         */
        std::vector< Photodot > _photodots;
        std::vector< Photodot > _expected_photodots;

        /** Map of the pulses in the event, sorted by module and pad id.
         */
        std::multimap<gear::GlobalPadIndex , EVENT::TrackerPulse * > * _pulses;

        /** The geometry type of the file in _geometry_filename
         */
        std::string _geometry_type;

        /** Diffusion of the photodots through the TPC.
         */
        double _diffusion;

        /* The following parameters describe the geometry of the TPC
         */
        /** The gas parameterisation of this piece of the TPC
         */
        double _omega_tau;

        /** The paramaterised drift velocity in this section of the TPC (mm/us)
         */
        double _drift_velocity;

        /** The outer radius of the TPC (mm)
         */
        double _tpc_outer_radius;

        /** The inner radius of the TPC (mm)
         */
        double _tpc_inner_radius;

        /** The absolute value of the z coordinate of the outer position of the
         *  TPC (mm)
         */
        double _outer_z;

        /** The absolute value of the z coordinate of the inner position of the
         *  TPC (mm)
         */
        double _inner_z;


        /** temporary correction
         */
        double _sigma;

        /** If true the HitCentre is calculated assuming a gaus distribution on each row -> the sigma must be known
         * Otherwise the charge weighted mean is calculated.
         */
        bool _useGaus;

}; // /PhotodotReconstructionProcessor

} // /namespace marlintpc

#endif //PHOTODOTRECONSTRUCTIONPROCESSOR_H_

#endif //USE_LCCD
