/*
 * PhotoelectricClusterTools.h
 *
 *  Created on: Apr 24, 2014
 *      Author: zenker
 */

#ifndef PHOTOELECTRICCLUSTERTOOLS_H_
#define PHOTOELECTRICCLUSTERTOOLS_H_

#include <set>
#include <vector>

// CLHEP
#include <CLHEP/Vector/TwoVector.h>

// LCIO
#include <lcio.h>
#include <EVENT/TrackerPulse.h>
#include <IMPL/TrackerHitImpl.h>

// GEAR
#include <gear/TPCParameters.h>
#include <gearimpl/GlobalPadIndex.h>

/* NOTE this uses the C++ version of Minuit - not the Fortran version */
#include <Minuit2/FCNBase.h>

namespace marlintpc{

/**
 * The given TrackerPulse set is used to sum up all positions with the charge of the pulse as weight.
 * Finally everything is divided by the total charge:
 * \f[x_{\mathrm{seed}} = \sum^{N_{\mathrm{pulses}}}_{i=1}\frac{x_i q_i}{Q}\f]
 *  with \f[Q = \sum^{N_{\mathrm{pulses}}}_{i=1} q_i\f]
 * Total charge of the cluster is stored in EDep.
 * TrackerPulses are stored in the RawHits.
 */
CLHEP::Hep2Vector calculateSeed(const std::set<lcio::TrackerPulse*> &pulses, const gear::TPCParameters& tpc_parameters, IMPL::TrackerHitImpl *seed_photodot = 0);

/**
 * Perform the Likelihood fit using the PhotodotLikelihood class.
 * EDep is copied from the seed and also the RawHits are copied. The fit covariance is stored in the covariance matrix of the fit_result:
 * - cov(x,x) = fit(0,0)
 * - cov(y,x) = fit(0,1)
 * - cov(y,y) = fit(1,1)
 * - cov(z,x) = 0
 * - cov(z,y) = 0
 * - cov(z,z) = sigma (this is the fitted sigma of the cluster)
 * \param seed A tracker hit with its pulses stored in order to be able to call \c seed->getRawHits()
 * \param fit_result Resulting position. The position is set to (x,y,sigma). This is the same as the seed if only one pulse is inside the seed hit.
 * \param sigma The expected cluster size.
 * \param fitSigma If true sigma is limited in the minimization to [0,5] and also minimized. Else it is fixed to the given value.
 * \return False if no raw data are stored in the given seed. True in case only one pulse is inside the seed. Else it is the result of the
 * minimizer given by \c minimizer.IsValid().
 */
bool doLikelihoodFit(const lcio::TrackerHit &seed, lcio::TrackerHitImpl* fit_result, double sigma, bool fitSigma = false);

/**
 * This helper class is used to calculate the likelihood function while fitting photodots
 */
class PhotodotLikelihood : public ROOT::Minuit2::FCNBase
{

  public:

    PhotodotLikelihood(const lcio::TrackerHit &photodot_data);

    virtual double operator()(const std::vector<double>& x) const;

    virtual double Up() const;

  private:

    const lcio::TrackerHit &_photodot_data;
};
}
#endif /* PHOTOELECTRICCLUSTERTOOLS_H_ */
