#ifndef PHOTOELECTRICDISTORTIONSCALCULATOR_H_
#define PHOTOELECTRICDISTORTIONSCALCULATOR_H_

// C++
#include <vector>

// CLHEP
#include <CLHEP/Vector/TwoVector.h>

// MarlinTPC
#include "TPCDistortionMeasurement.h"

// GEAR
#include "gear/TPCModule.h"

namespace marlintpc {
struct Element2D{
  gear::Vector2D p[4];
  gear::Vector2D shift[4];
};

/**
 * \author K. Zenker, DESY
 * \class PhotoelectricDistortionsCalculator
 * This class can be used to calculate field distortions measured with
 * the system of photodots on the TPC cathode. It considers the information
 * of the true photodot position on the cathode and the reconstructed photodot
 * position on the pad plane.
 * With this information a given measured point can be converted into a "true"
 * point inverting the distortions. Keep in mind the measured distortion are integrated
 * over the whole TPC volume. So if you have a measured point close to the anode
 * the result might be wrong.
 * In detail the grid of points is used to construct elements with four corners.
 * Next the element containing the point is searched for. In this element the distortions
 * are linearly shaped using the information at the corners of the element.
 * Finally the shift needed to move the given point to the estimated position without
 * distortions is given.
 * \remark This shift is given in local coordinates c0, c1 of the module!
 *
 * \note I use points in 3 dimensions because maybe one could also do something
 *  on the z position sometime...
 */
class PhotoelectricDistortionsCalculator{
public:
  /* The constructor takes the distortions information and fills it into an internal map.
   * If fastSearch is true the information is stored per module else it is stored in a single vector.
   * Later only dots from a single module are used in case of no module interpolation is used.
   * Thus the search is faster, but at the boundaries of the module the shift can not be calculated,
   * since for an element at the boundary also a dot from a neighboring module is needed.
   * \remark Use the module interpolation only if there are no field distortions at the module boundary.
   *         Else the resulting distortions at the module boundary are wrong!
   */
  PhotoelectricDistortionsCalculator(const std::vector<tpcconddata::TPCDistortionMeasurement*> &map, bool moduleInterpolation = false);
  /* Calculates the estimated shift to move the given point to the estimated point without distortions.
   * Give the measured point in global coordinates.
   * \remark This shift is given in local coordinates c0, c1 of the module!
   */
  gear::Vector2D getEstimatedShift2D(const CLHEP::Hep3Vector &measuredPosition);

  /* Set the maximum distance from a corner of the constructed element to the given measured point.
   */
  void setMaxDistance(double dist){_maxDist = dist;}

private:
  /* Map containing the module id and vector of photodot distortions.
   * The module id is taken from the reconstructed photodot position.
   */
  std::map<int,std::vector<tpcconddata::TPCDistortionMeasurement*> > _tpcDistortionsMap;

  Element2D _element;

  int _module; ///< Module that contains the current point under investigation.

  double _maxDist; ///< The maximum distance from a corner of the constructed element to the given measured point

  gear::Vector2D _point; ///< The measured point under investigation in local coordinates

  bool _moduleInterpolation; ///< If true the element construction is faster, but no distortions can be calculated at module boundaries

  /* Construct the element to be used to calculate field distortions.
   * Find top left/right and bottom left/right photodot with respect to a given position.
   * \return True if 4 corners were found.
   */
  bool constructElement();
  /* After the element is found now inside the element the shift is calculated.
   */
  gear::Vector2D calculateEstimatedShift();

  /* Here the measured point in local coordinates of the module is converted into
   * the coordinates of the isoparametric element.
   */
  gear::Vector2D getElementCoordinates();
};
}



#endif /* PHOTOELECTRICDISTORTIONSCALCULATOR_H_ */
