#ifndef PHOTOELECTRICPULSEEXTRACTIONPROCESSOR_H
#define PHOTOELECTRICPULSEEXTRACTIONPROCESSOR_H 1

#include "PhotoelectricXMLHelper.h"

// Marlin
#include <marlin/Processor.h>

// Gear
#include <gearimpl/GlobalPadIndex.h>

// LCIO
#include <lcio.h>
#include <EVENT/TrackerPulse.h>

//ROOT
#include "TH2D.h"
#include "TH1D.h"

//AIDA
namespace AIDA
{

class ITuple;
}

// stl
#include <string>
#include <set>

/**
 * Store the cluster information for each photodot:
 * 1. id of the photodot/cluster
 * 2. All Pulses in this cluster.
 */
typedef std::map<int, std::set<EVENT::TrackerPulse*> > photodotCluster;

namespace marlintpc
{

/** PhotoelectricPulseExtractionProcessor: takes the input pulse collection, finds the ones which are most-likely to be photoelectric pulses and puts them into another collection
 *    @author J. Abernathy (UVic)<br>
 */
class PhotoelectricPulseExtractionProcessor : public marlin::Processor
{

    public:

        PhotoelectricPulseExtractionProcessor() ;

        virtual void init() ;

        virtual Processor*  newProcessor()
        {
            return new PhotoelectricPulseExtractionProcessor();
        }

        virtual void processRunHeader(lcio::LCRunHeader* run);

        virtual void processEvent(lcio::LCEvent * evt);

        virtual void check(lcio::LCEvent * evt);

        virtual void end();

    private:

        /** the name of the collection the input pulses will be stored under
        */
        std::string _input_tracker_pulses_collection_name;

        /** the name of the output collection which will contain the photodot pulses
         */
        std::string _output_photoelectric_tracker_pulses_collection_name_dots;

        /** the name of the output collection which will contain the photoline pulses
         */
        std::string _output_photoelectric_tracker_pulses_collection_name_lines;

        /** The width of the acceptance interval (in ns) which is centred on the mode of the arrival times
         */
        float _arrival_time_width;

        /** determines whether the output of this processor will be stored in a file after program termination (false) or not (true - default)
        */
        bool _output_is_transient;

        /** specifies the name of the collection which stores the results of the projection finding
         */
        std::string _projected_photogeometry_file_name;

        /** the beginning (in ns) of the interval of the arrival time for photoelectric data
         */
        float _photopulse_interval_start;

        /** the end (in ns) of the interval of the arrival time for photoelectric data
         */
        float _photopulse_interval_end;

        /** the set of global pad indices which should contain photodot data
         */
        std::set< gear::GlobalPadIndex > _global_pad_indices_dots;

        /** the set of global pad indices which should contain photoline data
         */
        std::set< gear::GlobalPadIndex > _global_pad_indices_lines;

        /** this needs to be stored in order to keep informations got in init() in memory until end()
         *  which is needed if Distortions are extracted.
         */
        PhotoelectricXMLFile* _photodotFile; ///< XML file with the photoelectric informations.

        /** @name Things used to write Distortions slcio file */
        /*@{*/
        /**
         * if distortions are extracted this is the name of the output file
         */
        std::string _distortion_file_name;

        /**
         * This map stores all pulses for a certain photodot cluster if distortions are to be extracted
         */
        photodotCluster _photodot_pulse_sets;

        /**
         * Here for each pad the photodot id is stored
         * in order to fill pulses into the pulse set corresponding to a certain photodot
         */
        std::map<gear::GlobalPadIndex, int > _photodot_pulse_set_by_global_index;

        int _runNumber; ///< Only used if the distortions file is written

        std::string _detectorName; ///< Only used if the distortions file is written

        std::vector<std::vector<marlintpc::Photodot>::const_iterator> _dots; ///<Photodots taken from the xml \c _projected_photogeometry_file_name file

        double _sigma; ///< The expected cluster size needed for the likelihood fit
        
        bool _fitSigma; ///<Decide if sigma should be fitted. Else it is fixed.

        TH2D * _numberOfPadsWithPulsesPerCluster;
        TH2D * _numberOfPulsesPerCluster;
        AIDA::ITuple *_pulse_tuple;
        AIDA::ITuple *_cluster_tuple;
        /*@}*/
};

}

#endif // PHOTOELECTRICPULSEEXTRACTIONPROCESSOR_H
