#include "PhotoelectricDistortionsCalculator.h"
#include "MarlinTPCExceptions.hh"

// C++
#include <sstream>
#include <iostream>
#include <algorithm>

// GEAR
#include "gear/TPCParameters.h"

// Marlin
#include <marlin/Global.h>
#include <marlin/Processor.h>

// Root
#include "TMath.h"
#include "TVector2.h"

using namespace std;
using namespace tpcconddata;
using namespace marlin;
using namespace CLHEP;
using namespace gear;

namespace marlintpc{

PhotoelectricDistortionsCalculator::PhotoelectricDistortionsCalculator(const vector<TPCDistortionMeasurement*> &map, bool moduleInterpolation){
  _moduleInterpolation = moduleInterpolation;
  const TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();
  for(vector<TPCDistortionMeasurement*>::const_iterator it = map.begin(); it != map.end();it++){
    int module = (tpc_parameters.getNearestModule((*it)->getDoubleVal(3),(*it)->getDoubleVal(4))).getModuleID();
    if(_moduleInterpolation)
      _tpcDistortionsMap[0].push_back((*it));
    else
      _tpcDistortionsMap[module].push_back((*it));
  }
  _maxDist = -1;
  _module = 0;
}

Vector2D PhotoelectricDistortionsCalculator::getEstimatedShift2D(const Hep3Vector &measuredPosition){
  const TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();
  _module = (tpc_parameters.getNearestModule(measuredPosition.x(),measuredPosition.y())).getModuleID();
  const TPCModule& module = tpc_parameters.getModule(_module);
  _point = module.globalToLocal(measuredPosition.x(),measuredPosition.y());
  bool foundElement = constructElement();
  Vector2D pos;
  if(foundElement){
    pos = calculateEstimatedShift();
  } else {
    std::stringstream message;
    message << " PhotoelectricDistortionsCalculator couldn't construct an element to calculate distortions. ";
    throw InvalidDataException(message.str());
  }

  return pos;
}

/*
  Element corner numbering:

  c1 direction
  /\
  |  3-------2
  |  |       |
  |  |       |
  |  0-------1
  |
  |-------------> c0 direction
*/

bool PhotoelectricDistortionsCalculator::constructElement(){
  // distance of the corner points to the given point
  double distance[4] = {-1,-1,-1,-1};
  bool foundCorner[4] = {false,false,false,false};
  Hep2Vector tmp;
  double tmp_dist;
  const TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();
  const TPCModule& module = tpc_parameters.getModule(_module);
  vector<TPCDistortionMeasurement*>::const_iterator it_begin, it_end;
  if(_moduleInterpolation){
    // loop over all reconstructed positions
    it_begin = _tpcDistortionsMap[0].begin();
    it_end = _tpcDistortionsMap[0].end();
  } else {
    // loop over all reconstructed positions of the current module
    it_begin = _tpcDistortionsMap[_module].begin();
    it_end = _tpcDistortionsMap[_module].end();
  }

  for(vector<TPCDistortionMeasurement*>::const_iterator it = it_begin;
        it != it_end;it++){
    Vector2D reco_Photodot = module.globalToLocal((*it)->getDoubleVal(3),(*it)->getDoubleVal(4));
    Vector2D true_Photodot = module.globalToLocal((*it)->getDoubleVal(0),(*it)->getDoubleVal(1));
    for(int corner = 0; corner < 4; corner++){
      if(module.getLocalPadLayout().getCoordinateType() == PadRowLayout2D::POLAR && (reco_Photodot[1] > TMath::Pi()/2.)){
        reco_Photodot[1] -= 2. * TMath::Pi();
      }

      if(
         (corner == 0 && (reco_Photodot[0] <= _point[0] && reco_Photodot[1] <= _point[1])) ||
         (corner == 1 && (reco_Photodot[0] >  _point[0] && reco_Photodot[1] <= _point[1])) ||
         (corner == 2 && (reco_Photodot[0] >  _point[0] && reco_Photodot[1] >  _point[1])) ||
         (corner == 3 && (reco_Photodot[0] <= _point[0] && reco_Photodot[1] >  _point[1]))){
        Vector2D shift;
        if(module.getLocalPadLayout().getCoordinateType() == PadRowLayout2D::POLAR && (true_Photodot[1] > TMath::Pi()/2.)){
          true_Photodot[1] -= 2. * TMath::Pi();
        }
        shift = Vector2D((true_Photodot[0] - reco_Photodot[0]), (true_Photodot[1]-reco_Photodot[1]));
        if(module.getLocalPadLayout().getCoordinateType() == PadRowLayout2D::POLAR){
          tmp_dist = TMath::Sqrt(TMath::Power((reco_Photodot[0]*TMath::Sin(reco_Photodot[1]) - _point[0]*TMath::Sin(_point[1])),2)
                               + TMath::Power((reco_Photodot[0]*TMath::Cos(reco_Photodot[1]) - _point[0]*TMath::Cos(_point[1])),2));
        } else {
          tmp_dist = TMath::Sqrt(TMath::Power((reco_Photodot[0] - _point[0]),2) + TMath::Power((reco_Photodot[1] - _point[1]),2));
        }
        if(_maxDist > 0 && tmp_dist > _maxDist)
          break;
        if(tmp_dist < distance[corner] || distance[corner] < 0){
          // save the point as a possible corner of the element - if later a point is closer to the measurement point the corner position is updated
          distance[corner] = tmp_dist;
          _element.p[corner] = reco_Photodot;
          _element.shift[corner] = shift;
          foundCorner[corner] = true;
          break;
        }
      }
    }
  }
  streamlog_out(DEBUG) << "Found element: " << endl;
  streamlog_out(DEBUG) << "\t point in question: " << _point[0] << " , " << _point[1] << endl;
  for(int corner = 0; corner < 4; corner++){
    streamlog_out(DEBUG) <<  "\t " <<   _element.p[corner][0] << " , " << _element.p[corner][1] << " shift: " << _element.shift[corner][0] << " , " << _element.shift[corner][1] << endl;
  }

  return (foundCorner[0] && foundCorner[1] && foundCorner[2] && foundCorner[3]);
}

Vector2D PhotoelectricDistortionsCalculator::getElementCoordinates(){
  /*
   * You have to solve this system of equations for m and n:
   * \f[x = \frac[1][4]\left((1-m)(1-n)x_1+(1+m)(1-n)x_2+(1+m)(1+n)x_3+(1-m)(1+n)x_4\\
   *    y = \frac[1][4]\left((1-m)(1-n)y_1+(1+m)(1-n)y_2+(1+m)(1+n)y_3+(1-m)(1+n)y_4\f]
   */
  double b = _point[0]*(_element.p[0][1]-_element.p[1][1]+_element.p[2][1]-_element.p[3][1]);
  double a = _element.p[0][0]*(_element.p[3][1]-_point[1]) - _element.p[1][0]*(_element.p[2][1] - _point[1]) + _element.p[2][0]*(_element.p[1][1]-_point[1]) - _element.p[3][0]*(_element.p[0][1]-_point[1]) + b;
  double a1 = _element.p[0][0]*(_point[1]-_element.p[1][1]) - _element.p[1][0]*(_point[1]-_element.p[0][1]) + _element.p[2][0]*(_point[1]-_element.p[3][1]) - _element.p[3][0]*(_point[1]-_element.p[2][1]) -b;
  double t = (_element.p[0][0] + _element.p[1][0])*(_element.p[2][1]+_element.p[3][1]-2*_point[1]) - (_element.p[2][0]+ _element.p[3][0])*(_element.p[0][1]+_element.p[1][1]-2*_point[1]) + 2*_point[0]*(_element.p[0][1] + _element.p[1][1] -_element.p[2][1] - _element.p[3][1]);
  double den  = (_element.p[1][0] - _element.p[0][0])*(_element.p[2][1] - _element.p[3][1]) + (_element.p[3][0] - _element.p[2][0])*(_element.p[1][1] - _element.p[0][1]);
  double den1 = (_element.p[3][0] - _element.p[0][0])*(_element.p[1][1] - _element.p[2][1]) + (_element.p[1][0] - _element.p[2][0])*(_element.p[0][1] - _element.p[3][1]);
  double s = TMath::Sqrt(TMath::Power(-1.*a,2) - den*t);

  double m,n;
  // check if the denominator den1 is 0
  if( fabs(_element.p[3][0] - _element.p[0][0]) < 1e-8 && fabs(_element.p[1][0] - _element.p[2][0]) < 1e-8){
    n = (_element.p[0][0]*(_element.p[1][1] + _element.p[2][1] - 2*_point[1]) + _element.p[1][0]*(2*_point[1] - _element.p[0][1] - _element.p[3][1]) +
        _point[0]*(_element.p[0][1] - _element.p[1][1] - _element.p[2][1] + _element.p[3][1]))/(b + _element.p[1][0]*(_element.p[3][1] - _element.p[0][1]) + _element.p[0][0]*(_element.p[1][1] - _element.p[2][1]));
  } else {
    double n1 = (a1+s)/den1;
    if(fabs(n1) <= 1)
      n = n1;
    else
      n = (a1-s)/den1;;
  }
  // check if the denominator den is 0
  if(fabs(_element.p[0][1] - _element.p[1][1]) < 1e-8 && fabs(_element.p[2][1] - _element.p[3][1]) < 1e-8){
    m = (_point[1]*(_element.p[0][0] + _element.p[1][0] - _element.p[2][0] - _element.p[3][0]) + _element.p[0][1]*(_element.p[2][0] + _element.p[3][0] - 2*_point[0]) + _element.p[3][1]*(2*_point[0] - _element.p[0][0] - _element.p[1][0]))/
        ((_element.p[2][0] - _element.p[3][0])*(_point[1] - _element.p[0][1]) + _element.p[0][0]*(_point[1] - _element.p[3][1]) - _element.p[1][0]*(_point[1] - _element.p[3][1]));
  } else {
    double m1 = (a - s)/den;
    if(fabs(m1) <= 1)
        m = m1;
    else
      m = (a + s)/den;
  }

  if(fabs(m) > 1 || fabs(n) > 1){
    streamlog_out(ERROR) << "In element with corners: " << endl;
    for(int corner = 0; corner < 4; corner++){
      streamlog_out(ERROR) << "\t " << _element.p[corner][0] << " , " << _element.p[corner][1] << " with shift "
           << "\t " << _element.shift[corner][0] << " , " << _element.shift[corner][1] << endl;
    }
    streamlog_out(ERROR) << "\t x0 && x3: " << (_element.p[3][0] == _element.p[0][0]) << " x0-x3: " << _element.p[3][0] - _element.p[0][0] <<
        "\n\t x1 && x2: " << (_element.p[1][0] == _element.p[2][0]) << " x1-x2: " << _element.p[1][0] - _element.p[2][0] << endl;
    streamlog_out(ERROR) << "\t The denuminators are: " << den << " , "  << den1 << endl;
    streamlog_out(ERROR) << "\t Measurement point is: " << _point[0] << " , " << _point[1] << endl;
    streamlog_out(ERROR) << "\t Local coordinates: " << m << " , " << n << endl;

    std::stringstream message;
    message << " PhotoelectricDistortionsCalculator couldn't construct an element to calculate distortions. ";
    throw InvalidDataException(message.str());
  }

  return Vector2D(m,n);
}

Vector2D PhotoelectricDistortionsCalculator::calculateEstimatedShift(){
  // measured point in element coordinates
  Vector2D localPoint = getElementCoordinates();

  double c0 = 1./4.*((1-localPoint[0])*(1-localPoint[1])*_element.shift[0][0] +
      (1+localPoint[0])*(1-localPoint[1])*_element.shift[1][0] +
      (1+localPoint[0])*(1+localPoint[1])*_element.shift[2][0] +
      (1-localPoint[0])*(1+localPoint[1])*_element.shift[3][0]);
  double c1 = 1./4.*((1-localPoint[0])*(1-localPoint[1])*_element.shift[0][1] +
        (1+localPoint[0])*(1-localPoint[1])*_element.shift[1][1] +
        (1+localPoint[0])*(1+localPoint[1])*_element.shift[2][1] +
        (1-localPoint[0])*(1+localPoint[1])*_element.shift[3][1]);

  return Vector2D(c0,c1);
}
}




