// MarlinTPC
#include "PhotoelectricLikelihoodSeederProcessor.h"
#include "PhotoelectricXMLHelper.h"
#include "PhotoelectricClusterTools.h"

// Marlin
#include <marlin/Exceptions.h>
#include <marlin/Global.h>

// AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/ITupleFactory.h>
#include <AIDA/ITuple.h>

// CLHEP
#include <CLHEP/Vector/TwoVector.h>

// GEAR
#include <gear/TPCModule.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>

// LCIO
#include <EVENT/LCIO.h>
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerPulse.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>

using namespace CLHEP;
using namespace EVENT;
using namespace IMPL;
using namespace lcio;
using namespace gear;
using namespace marlin;
using namespace std;

namespace marlintpc {

PhotoelectricLikelihoodSeederProcessor aPhotoelectricLikelihoodSeederProcessor;

PhotoelectricLikelihoodSeederProcessor::PhotoelectricLikelihoodSeederProcessor() :
    Processor("PhotoelectricLikelihoodSeederProcessor"),

    _input_photoelectric_tracker_pulses_collection_name(
        "TPCPhotoelectricPulses"), _output_photodot_collection_name(
        "TPCPhotodots"), _output_photoline_collection_name("TPCPhotolines"), _output_is_transient(
        true), _projected_photogeometry_file_name(
        "projected_photogeometry.xml"), _write_seed_tuple(false), _sigma(2.5) {

  /* description of the processor which will be displayed in the
   * steering file autometicly generated by
   * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
   */

  _description =
      "PhotoelectricLikelihoodSeederProcessor collects the TrackerPulses which belong to each projected photogeometry element and produces \"seed\" objects for the likelihood reconstruction algorithm.";

  /* register steering parameters: name, description, class-variable, default value
   * the type will by definded by the type of the default value
   * string, double, float and int are possible
   * plaese note to get a float you must cast in C++: (float) 1.3
   */

  registerInputCollection(LCIO::TRACKERPULSE, "InputPhotoelectricTrackerPulses",
      "The name of the input collection of TPCTrackerPulses from the photoelectric system.",
      _input_photoelectric_tracker_pulses_collection_name,
      std::string("TPCPhotoelectricPulses"));

  registerOutputCollection(LCIO::TRACKERHIT, "OutputSeedPhotodots",
      "The name of the output collection of TrackerHits which represent seed photodots.",
      _output_photodot_collection_name, std::string("TPCSeedPhotodots"));

  registerOutputCollection(LCIO::TRACK, "OutputSeedPhotolines",
      "The name of the output collection of Tracks which represent seed photolines.",
      _output_photoline_collection_name, std::string("TPCSeedPhotolines"));

  registerOptionalParameter("SetOutputTransient",
      "Controls whether the output collection should be saved in the LCIO file (false) or if it should be deleted at the end (true - default).",
      _output_is_transient, true);

  registerOptionalParameter("ProjectedPhotogeometryFilename",
      "This is the name of the collection which stores the true positions of the photodots as SimTrackerHit objects. It is optional but it provides the number of photodots and, if provided, the processor will order the seed photodots to match the order of true photodots (useful for analysis).",
      _projected_photogeometry_file_name, _projected_photogeometry_file_name);

  registerOptionalParameter("WriteSeedsToRootFile",
      "Controls whether the seed positions found by this processor are written to the root file or not.",
      _write_seed_tuple, _write_seed_tuple);

}

void PhotoelectricLikelihoodSeederProcessor::init() {

  // usually a good idea to
  printParameters();

  // prepare tuple to write seeds if necessary
  _seed_position_tuple = 0;
  if (_write_seed_tuple) {
    try {
      _seed_position_tuple = AIDAProcessor::tupleFactory(this)->create(
          "PhotodotSeeds", "Photodot Seeds", "int i_dot, double x, double y",
          "");
    } catch (lcio::Exception &e) {
      streamlog_out(DEBUG)<< e.what() << endl;
      return;
    }
  }

  // load AIDA (if it's available)
  AIDA::ITuple *projected_photodot_tuple = 0;
  try {

    projected_photodot_tuple = AIDAProcessor::tupleFactory(this)->create(
        "PhotodotProjectionInfo", "Photodot Projection Info",
        "int i_dot, int max_pads", "");
  } catch (lcio::Exception &e) {

    streamlog_out(DEBUG)<< e.what() << endl;
    return;
  }

  PhotoelectricXMLFile photodotFile(_projected_photogeometry_file_name);
  std::map<int, std::set<gear::GlobalPadIndex> > pads =
      photodotFile.GetPhotodotPads();
  if (pads.size() == 0) {
    streamlog_out(ERROR)<< "No pads could be retrieved from projected geometry file (either file ." << endl;
  }
  for (std::map<int, std::set<gear::GlobalPadIndex> >::iterator it_dots =
      pads.begin(); it_dots != pads.end(); it_dots++) {
    // create a set to hold the pulses for this photodot
    // even if there is no projection and it won't be used for fitting it must be stored in the lcio collection to keep all of the photodots in the correct order
    std::list<std::set<EVENT::TrackerPulse*> >::iterator new_pulse_set =
        _photodot_pulse_sets.insert(_photodot_pulse_sets.end(),
            set<EVENT::TrackerPulse*>());
    if ((*it_dots).second.size() == 0) {
      streamlog_out(DEBUG)<< "photodot " << (*it_dots).first << " doesn't have a projection." << endl;
      projected_photodot_tuple->fill(0, (*it_dots).first);
      projected_photodot_tuple->fill(1, 0);
      projected_photodot_tuple->addRow();
      continue;
    }
    int pad_count = 0;
    for (std::set<gear::GlobalPadIndex>::iterator it_pads =
        (*it_dots).second.begin(); it_pads != (*it_dots).second.end();
        it_pads++, pad_count++) {
      _photodot_pulse_set_by_global_index[*it_pads] = new_pulse_set;
    }
    projected_photodot_tuple->fill(0, (*it_dots).first);
    projected_photodot_tuple->fill(1, pad_count);
    projected_photodot_tuple->addRow();
  }

}

void PhotoelectricLikelihoodSeederProcessor::processRunHeader(
    lcio::LCRunHeader*  /*run*/) {

}

void PhotoelectricLikelihoodSeederProcessor::processEvent(lcio::LCEvent * evt) {

  // Step 1. Reset the pulse data because the TrackerPulses will be different from the last event
  for (std::list<std::set<EVENT::TrackerPulse*> >::iterator pulse_set =
      _photodot_pulse_sets.begin(); pulse_set != _photodot_pulse_sets.end();
      pulse_set++) {

    pulse_set->clear();
  }

  // Step 2. Associate the TrackerPulses with the correct photodot or photoline
  LCCollection* input_photoelectric_pulses = evt->getCollection(
      _input_photoelectric_tracker_pulses_collection_name);

  streamlog_out(MESSAGE)<< "Using " << input_photoelectric_pulses->getNumberOfElements() << " pulses from the collection " << _input_photoelectric_tracker_pulses_collection_name << " to seed " << _photodot_pulse_sets.size() << " photodots." << endl;

  // now determine if this is a photoelectric pulse
  for (int i_tracker_pulse = 0;
      i_tracker_pulse < input_photoelectric_pulses->getNumberOfElements();
      i_tracker_pulse++) {

    TrackerPulse* input_pulse =
        dynamic_cast<TrackerPulse*>(input_photoelectric_pulses->getElementAt(
            i_tracker_pulse));

    if (!input_pulse) {

      throw(SkipEventException(this));
    }

    const GlobalPadIndex global_index(input_pulse->getCellID0(),
        input_pulse->getCellID1());

    // I can't wait for the new auto keyword...
    const std::map<gear::GlobalPadIndex,
        std::list<std::set<EVENT::TrackerPulse*> >::iterator>::iterator i_pulse_set =
        _photodot_pulse_set_by_global_index.find(global_index);

    if (i_pulse_set != _photodot_pulse_set_by_global_index.end()) {

      i_pulse_set->second->insert(input_pulse);
    }
  }

  // Step 3. Loop through the photodots, creating TrackerHits with seed positions
  const TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();

  int i_dot = 0;

  // create a collection to store the photodot seeds
  LCCollectionVec* output_seed_photodots = new LCCollectionVec(
      LCIO::TRACKERHIT);
  output_seed_photodots->setTransient(_output_is_transient);
  output_seed_photodots->setSubset(false);

  for (list<set<TrackerPulse*> >::const_iterator i_photodot =
      _photodot_pulse_sets.begin(); i_photodot != _photodot_pulse_sets.end();
      i_photodot++, i_dot++) {

    TrackerHitImpl *seed_photodot = new TrackerHitImpl();

    // again, make sure that the photodot is added to the collection to preserve ordering
    output_seed_photodots->addElement(seed_photodot);

    if (i_photodot->size() == 0) {

      streamlog_out(DEBUG)<< "photodot " << i_dot << " did not receive any data this event." << endl;
      continue;
    }
    Hep2Vector seed_position = calculateSeed(*i_photodot, tpc_parameters,seed_photodot);


    // fill tuple
    if (_write_seed_tuple) {
      _seed_position_tuple->fill(0, i_dot);
      _seed_position_tuple->fill(1, seed_position[0]);
      _seed_position_tuple->fill(2, seed_position[1]);
      _seed_position_tuple->addRow();
    }
    streamlog_out(DEBUG)<< "Seeded photodot " << i_dot << " with position " << seed_position << "." << endl;
  }

  // now add the collection to the event
  evt->addCollection(output_seed_photodots, _output_photodot_collection_name);
}

void PhotoelectricLikelihoodSeederProcessor::check(lcio::LCEvent *  /*evt*/) {

  // nothing to check here - could be used to fill checkplots in reconstruction processor
}

void PhotoelectricLikelihoodSeederProcessor::end() {

}

} // namespace marlintpc
