//#ifdef USE_LCCD

// MarlinTPC
#include "PhotoelectricProjectionFinder.h"
#include "PhotoelectricXMLHelper.h"

// tpcconddata
#include "TPCDistortionMeasurement.h"

// Marlin
#include <marlin/Exceptions.h>
#include <marlin/Global.h>
#include <marlin/Processor.h>

// AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/ITupleFactory.h>
#include <AIDA/ITuple.h>

// GEAR
#include <gear/TPCModule.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>
#include <gearxml/tinyxml.h>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerPulse.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <UTIL/LCTime.h>

// CLHEP
#include <CLHEP/Vector/TwoVector.h>

// stl
#include <algorithm>
#include <stack>
#include <fstream>


/*
 #include "TAxis.h"
 #include "TApplication.h"
 #include "TCanvas.h"
 #include "TColor.h"
 #include "TGraph.h"
 #include "TPaletteAxis.h"
 #include "TROOT.h"
 #include "TStyle.h"
 #include "TSystem.h"
 #include "TMath.h"
 */

using namespace CLHEP;
using namespace lcio;
using namespace gear;
using namespace marlin;
using namespace std;

namespace marlintpc {

PhotoelectricProjectionFinder aPhotoelectricProjectionFinder;

PhotoelectricProjectionFinder::PhotoelectricProjectionFinder() :
    Processor("PhotoelectricProjectionFinder"), _input_tracker_pulses_collection_name(
        "TPCPulses"), _projected_photogeometry_file_name(
        "projected_photogeometry.xml"), _event_fraction_threshold(0.75), _maximum_dot_displacement(30),
        _distortion_file_name("DistortionInformation.sclio") {

  /* description of the processor which will be displayed in the
   * steering file autometicly generated by
   * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
   */

  _description =
      "PhotoelectricProjectionFinder finds the projected pattern produced by the photoelectric system on the pad plane. It then creates two collections. One collection is of TrackerHits which store the pads to perform the photodot fitting on. The other collection contains one TrackerHit which stores the entire collection of pads which contain photoelectric data.";

  /* register steering parameters: name, description, class-variable, default value
   * the type will by definded by the type of the default value
   * string, double, float and int are possible
   * plaese note to get a float you must cast in C++: (float) 1.3
   */
  registerInputCollection(LCIO::TRACKERPULSE, "InputTrackerPulses",
      "The name of the input collection of TPCPulses found by a pulse finder.",
      _input_tracker_pulses_collection_name, std::string("TPCPPulses"));

  registerProcessorParameter("MaxSeperation",
      "Specifies the maximum distance (in mm) that two points can be separated before they are considered to be neighbours.",
      _max_separation, (float) 10.);

  registerProcessorParameter("PhotopulseIntervalStart",
      "The beginning (in ns) of the interval of the arrival time for photoelectric data.",
      _photopulse_interval_start, (float) 40.);

  registerProcessorParameter("PhotopulseIntervalEnd",
      "The end (in ns) of the interval of the arrival time for photoelectric data.",
      _photopulse_interval_end, (float) 40.);

  registerOptionalParameter("ProjectedPhotogeometryFilename",
      "This is the name of the collection which stores the true positions of the photodots as SimTrackerHit objects. It is optional but it provides the number of photodots and, if provided, the processor will order the seed photodots to match the order of true photodots (useful for analysis).",
      _projected_photogeometry_file_name, _projected_photogeometry_file_name);

  registerOptionalParameter("PhotogeometryFilename",
      "This file provides the locations of the photodots and photolines on the TPC cathode.",
      _photogeometry_file_name, _photogeometry_file_name);

  registerOptionalParameter("EventFractionThreshold",
      "This is fraction of events which a pad has to be in to be extracted for having photoelectric data.",
      _event_fraction_threshold, _event_fraction_threshold);

  registerOptionalParameter("MaximumDotProjectionDisplacement",
      "This is the maximum distance (in mm) that the projection of a photodot can be displaced and still be accepted.",
      _maximum_dot_displacement, _maximum_dot_displacement);

  registerOptionalParameter("DistortionOutputFile",
      "If this option is used the true and the reconstructed photodot position is stored as TPCDistortionsMap in the slcio file.",
      _distortion_file_name, _distortion_file_name);

}

void PhotoelectricProjectionFinder::init() {

  // usually a good idea to
  printParameters();
}

void PhotoelectricProjectionFinder::processRunHeader(lcio::LCRunHeader* run) {
  _detectorName = run->getDetectorName();
  _runNumber = run->getRunNumber();
}

void PhotoelectricProjectionFinder::processEvent(lcio::LCEvent * evt) {

  LCCollection* input_pulses = evt->getCollection(
      _input_tracker_pulses_collection_name);

  streamlog_out(DEBUG)<< "Analysing " << input_pulses->getNumberOfElements() << " pulses." << endl;

  int pulse_count = 0;
  int out_of_time = 0;

  // add the pad data to the event
  for (int i_tracker_pulse = 0;
      i_tracker_pulse < input_pulses->getNumberOfElements();
      i_tracker_pulse++) {

    TrackerPulse* input_pulse =
        dynamic_cast<TrackerPulse*>(input_pulses->getElementAt(i_tracker_pulse));

    if (!input_pulse) {

      throw(lcio::Exception(
          "Input collection " + _input_tracker_pulses_collection_name
              + " does not contain TrackerPulses."));
    }

    const float time = input_pulse->getTime();
    bool test_start = (!_parameters->isParameterSet("PhotopulseIntervalStart") ||  (_parameters->isParameterSet("PhotopulseIntervalStart") && (time >= _photopulse_interval_start)));
    bool test_end = (!_parameters->isParameterSet("PhotopulseIntervalEnd") ||  (_parameters->isParameterSet("PhotopulseIntervalEnd") && (time <= _photopulse_interval_end)));
    if (test_start && test_end) {
      GlobalPadIndex global_index(input_pulse->getCellID0(),
          input_pulse->getCellID1());

      _counts_by_pad[global_index] += 1;

      pulse_count += 1;
    } else {
      out_of_time++;
    }

  }

  _total_event_count += 1;

  streamlog_out(DEBUG)<< "Found " << pulse_count << " photopulse candidates. (" << out_of_time << " not in time window)" << endl;
}

void PhotoelectricProjectionFinder::check(lcio::LCEvent *  /*evt*/) {
  // nothing to check here - could be used to fill checkplots in reconstruction processor
}

void PhotoelectricProjectionFinder::end() {

  // Step 1. Break the pulse data into sets of points. Give each point a weight
  vector<Hep2Vector*> points;
  map<Hep2Vector*, GlobalPadIndex> pad_by_point;
  create_points(points, pad_by_point);

  if (points.size() == 0) {

    streamlog_out(WARNING)<< "Unable to find any photoelectric pulses." << endl;
    return;
  }

  streamlog_out(DEBUG)<< "Created " << points.size() << " points." << endl;

  /*    // need to create a ROOT application to draw a canvas, it does not need to be run
   TApplication* app = new TApplication("app", new int(0), new char* );
   app->SetReturnFromRun( kFALSE );
   if (gROOT->IsBatch()) {
   gROOT->SetBatch( kFALSE );
   }

   TGraph graph ( points.size() );

   int i_point = 0;
   for ( vector< Hep2Vector* >::const_iterator point = points.begin(); point != points.end(); point++ ) {

   graph.SetPoint ( i_point++, (*point)->x(), (*point)->y() );
   }

   graph.Draw("AP");

   // need some wait loop to keep canvas open
   std::cout << "Press return to continue ..." << std::endl;
   int flag = fcntl(1, F_GETFL, 0);
   int key = 0;
   while(true) {
   gSystem->ProcessEvents();
   fcntl(1, F_SETFL, flag | O_NONBLOCK);
   key = getchar();
   if((key == '\n') || (key == '\r'))
   break;
   usleep(1000);
   }
   fcntl(1, F_SETFL, flag);*/

  // just create some points to test the processor THIS IS FOR DEBUGGING
  /*Hep2Vector v1(0,   0);
   Hep2Vector v2(28, 28);
   Hep2Vector v3(100,100);
   Hep2Vector v4(50,  14);
   Hep2Vector v5(-50,  -50);

   points.push_back(&v1);
   points.push_back(&v2);
   points.push_back(&v3);
   points.push_back(&v4);
   points.push_back(&v5);*/

  // Step 2. Find each points' neighbours
  map<Hep2Vector*, set<Hep2Vector*> > neighbours_by_point;
  find_neighbours(points, _max_separation, neighbours_by_point);
  //find_neighbours( points, 40, neighbours_by_point );

  streamlog_out(DEBUG)<< "Found the neighbours." << endl;

  // Step 3. Use the neighbour information to create point clusters
  set<set<Hep2Vector*> > point_clusters;
  calculate_point_clusters(points, neighbours_by_point, point_clusters);

  streamlog_out(DEBUG)<< "Created " << point_clusters.size() << " point clusters." << endl;

  // Step 4. Classify the point clusters as photodots or photolines
  set<const set<Hep2Vector*>*> photodots;
  set<const set<Hep2Vector*>*> photolines;
  classify_point_clusters(point_clusters, pad_by_point, photodots, photolines);

  streamlog_out(MESSAGE)<< "Found " << photodots.size() << " photodots and " << photolines.size() << " photolines." << endl;

  // Step 5. Store the results
  store_results(photodots, photolines, pad_by_point,_parameters->isParameterSet("DistortionOutputFile"));

  // Step 6. Produce tuples
  produce_tuples(photodots, photolines);

  streamlog_out(DEBUG)<< "Stored the results in " << _projected_photogeometry_file_name << "." << endl;

  // Step 7. Clean up! ;) ....... I can't wait for C++0x to use anonymous functions :)
  for (vector<Hep2Vector*>::iterator point = points.begin();
      point != points.end(); point++) {

    if (*point) {

      delete *point;
    }
  }

}

void PhotoelectricProjectionFinder::create_points(vector<Hep2Vector*> &points,
    map<Hep2Vector*, GlobalPadIndex> &pad_by_point) {

  // get the TPC parameters
  const TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();

  // resize the vector of points
  points.reserve(_counts_by_pad.size());

  // load AIDA (if it's available)
  AIDA::ITuple *pad_event_count_tuple = 0;
  try {

    pad_event_count_tuple = AIDAProcessor::tupleFactory(this)->create(
        "PhotoelectricPadInfo", "Photoelectric Pad Info",
        "int pad_index, int module_id, int event_count", "");
  } catch (lcio::Exception &e) {
    streamlog_out(WARNING) << "No AIDA processor found. Will not produce a root file..." << endl;
    streamlog_out(DEBUG)<< e.what() << endl;
  }

  // the pad must be in at least "_event_fraction_threshold" of the events to count as one which receives photoelectric data
  const int threshold = static_cast<int>(ceil(
      _event_fraction_threshold * _total_event_count));

  streamlog_out(DEBUG)<< "Pad count threshold is " << threshold << "." << endl;

  // loop through the pads
  for (std::map<gear::GlobalPadIndex, int>::const_iterator i_pad =
      _counts_by_pad.begin(); i_pad != _counts_by_pad.end(); i_pad++) {

    const GlobalPadIndex global_index = i_pad->first;
    const int count = i_pad->second;

    if (pad_event_count_tuple) {

      pad_event_count_tuple->fill(0, global_index.getPadIndex());
      pad_event_count_tuple->fill(1, global_index.getModuleID());
      pad_event_count_tuple->fill(2, count);
      pad_event_count_tuple->addRow();
    }

    if (count < threshold) {

      streamlog_out(DEBUG)<< "Pad " << global_index.getPadIndex() << " from module " << global_index.getModuleID() << " rejected. Reason? count: " << count << "." << endl;
      continue;
    }

    streamlog_out(DEBUG)<< "Pad " << global_index.getPadIndex() << " from module " << global_index.getModuleID() << " accepted. Reason? count: " << count << "." << endl;

    // break this cluster into points
    const PadRowLayout2D &local_layout = tpc_parameters.getModule(
        global_index.getModuleID()).getLocalPadLayout();

    // add some points, make sure that the distance between points in the width and height direction is about 1mm
    Vector2D local_pad_centre = local_layout.getPadCenter(
        global_index.getPadIndex());
    float pad_height = local_layout.getPadHeight(global_index.getPadIndex());
    float pad_width = local_layout.getPadWidth(global_index.getPadIndex());

    if (local_layout.getCoordinateType() == PadRowLayout2D::POLAR) {

      // approximate the pad width
      pad_width *= local_pad_centre[0];

      // reverse the coordinates
      float temp = local_pad_centre[0];
      local_pad_centre[0] = local_pad_centre[1];
      local_pad_centre[1] = temp;
    }

    const int NW = static_cast<int>(ceil(pad_width));
    const int NH = static_cast<int>(ceil(pad_height));

    const float width_seperation = local_layout.getPadWidth(
        global_index.getPadIndex()) / NW;
    const float height_seperation = pad_height / NH;

    float x_phi = local_pad_centre[0] - (0.5 * (NW - 1) * width_seperation);

    // produce a grid of points on the pad
    for (int i_w = 1; i_w <= NW; i_w++) {

      float y_r = local_pad_centre[1] - (0.5 * (NH - 1) * height_seperation);

      for (int i_h = 1; i_h <= NH; i_h++) {

        // create a point
        Vector2D global_point;

        if (local_layout.getCoordinateType() == PadRowLayout2D::POLAR) {

          global_point =
              tpc_parameters.getModule(global_index.getModuleID()).localToGlobal(
                  y_r, x_phi);
        } else {

          global_point =
              tpc_parameters.getModule(global_index.getModuleID()).localToGlobal(
                  x_phi, y_r);
        }

        if (tpc_parameters.getCoordinateType() == PadRowLayout2D::POLAR) {

          // convert it to cartesian coordinates
          const float r = global_point[0];
          global_point[0] = r * cos(global_point[1]);
          global_point[1] = r * sin(global_point[1]);
        }

        // convert it to a Hep2Vector because Vector2D can't do any vector operations
        Hep2Vector *new_point = new Hep2Vector(global_point[0],
            global_point[1]);
        points.push_back(new_point);
        pad_by_point.insert(make_pair(new_point, global_index));

        GlobalPadIndex new_index = pad_by_point.find(new_point)->second;
        streamlog_out(DEBUG)<< "Inserted point " << *new_point << " for pad " << new_index.getPadIndex() << endl;

        y_r += height_seperation;
      }

      x_phi += width_seperation;
    }
  }
}

void PhotoelectricProjectionFinder::find_neighbours(
    const vector<Hep2Vector*> &points, const float max_separation,
    map<Hep2Vector*, set<Hep2Vector*> > &neighbours_by_point) {

  // square the distance to save processing time in the loop
  const float max_separation_squared = max_separation * max_separation;

  // this will be implemented using a quad tree map
  map<Hep2Vector*, vector<Hep2Vector*> > quad_tree;

  // let the first point be the head node
  vector<Hep2Vector*>::const_iterator point = points.begin();

  // initialize the head node
  Hep2Vector *head_node = *point;
  vector<Hep2Vector*> &children = quad_tree[head_node];
  children.resize(4);

  // now add the other points to the tree, recording the edges
  for (++point; point != points.end(); point++) {

    // a stack of points whose proximity with the new point should be checked
    stack<Hep2Vector*> check_points;

    Hep2Vector *current_node = head_node;

    while (true) {

      // determine where this node should be placed, relative to the current one by finding the correct quadrant
      const Hep2Vector difference = **point - *current_node;

      // this is a neat way to calculate the index of the next node to search
      const int negative_y = (difference[1] < 0) << 1;
      const int negative_x = (difference[0] < 0);
      const int index = negative_y + negative_x;

      // get the list of children
      vector<Hep2Vector*> &children = quad_tree[current_node];

      // check the children for proximity (but not the next node at index - it will be checked automatically)
      check_points.push(children[index ^ 0x01]);
      check_points.push(children[index ^ 0x02]);

      // check the current node for proximity
      if (difference.mag2() <= max_separation_squared) {

        // check the node in the opposite direction
        check_points.push(children[index ^ 0x03]);

        // add the neighbours
        neighbours_by_point[current_node].insert(*point);
        neighbours_by_point[*point].insert(current_node);
      }

      // check if the next node exists
      if (children[index] == NULL) {

        // if not, then add the new point as a node
        children[index] = *point;

        // and initialize it's children
        vector<Hep2Vector*> &new_children = quad_tree[*point];
        new_children.resize(4);
        break;
      }

      current_node = children[index];
    }

    // now follow the insertion backwards looking for neighbours
    while (!check_points.empty()) {

      current_node = check_points.top();
      check_points.pop();

      // skip this node if it doesn't exist or if it's checking itself
      if (!current_node || (current_node == *point)) {

        continue;
      }

      // determine where this node should be placed, relative to the current one by finding the correct quadrant
      const Hep2Vector difference = **point - *current_node;

      // this is a neat way to calculate the index of the next node to search
      const int negative_y = (difference[1] < 0) << 1;
      const int negative_x = (difference[0] < 0);
      const int index = negative_y + negative_x;

      // and check the children which could be close to this node
      vector<Hep2Vector*> &children = quad_tree[current_node];

      check_points.push(children[index]);
      check_points.push(children[index ^ 0x01]);
      check_points.push(children[index ^ 0x02]);

      // check if this node is close to the newly added node
      if (difference.mag2() <= max_separation_squared) {

        // add the remaining child node
        check_points.push(children[index ^ 0x03]);

        // add a neighbour for each one of the points
        neighbours_by_point[current_node].insert(*point);
        neighbours_by_point[*point].insert(current_node);
      }
    }

  }
}

void PhotoelectricProjectionFinder::calculate_point_clusters(
    const vector<Hep2Vector*> &points,
    const map<Hep2Vector*, set<Hep2Vector*> > &neighbours_by_point,
    set<set<Hep2Vector*> > &point_clusters) {

  // create a set for the unclustered points
  set<Hep2Vector*> unclustered_points;
  unclustered_points.insert(points.begin(), points.end());

  // Step 3. Create clusters by joining neighbouring core points their boundaries
  set<Hep2Vector*>::iterator unclustered_point = unclustered_points.begin();

  while (unclustered_point != unclustered_points.end()) {

    // create a cluster for this point
    set<Hep2Vector*> new_cluster;

    // the neighbours of these neighbours need to be inserted into this cluster
    stack<Hep2Vector*> added_points;

    // add the first point to this stack
    added_points.push(*unclustered_point);

    while (!added_points.empty()) {

      Hep2Vector *added_point = added_points.top();
      added_points.pop();

      // add this point to the new cluster
      new_cluster.insert(added_point);

      // and remove it from the list of unclustered points (so it doesn't get added more than once)
      unclustered_points.erase(added_point);

      // and add it's unclustered neighbours
      map<Hep2Vector*, set<Hep2Vector*> >::const_iterator i_neighbours =
          neighbours_by_point.find(added_point);

      if (i_neighbours == neighbours_by_point.end()) {

        continue;
      }

      const set<Hep2Vector*> &neighbours = i_neighbours->second;

      set<Hep2Vector*> unclustered_neighbours;
      set_intersection(neighbours.begin(), neighbours.end(),
          unclustered_points.begin(), unclustered_points.end(),
          inserter(unclustered_neighbours, unclustered_neighbours.begin()));

      for (set<Hep2Vector*>::const_iterator unclustered_neighbour =
          unclustered_neighbours.begin();
          unclustered_neighbour != unclustered_neighbours.end();
          unclustered_neighbour++) {

        added_points.push(*unclustered_neighbour);
        unclustered_points.erase(*unclustered_neighbour);
      }

    }

    // now add this cluster to the set of clusters which have been found
    point_clusters.insert(new_cluster);

    // and move on to the next core point in the collection
    unclustered_point = unclustered_points.begin();

    streamlog_out(DEBUG)<< "Created a cluster of " << new_cluster.size() << " points." << endl;
  }

}

void PhotoelectricProjectionFinder::classify_point_clusters(
    const set<set<Hep2Vector*> > &point_clusters,
    const map<Hep2Vector*, GlobalPadIndex> pad_by_point,
    set<const set<Hep2Vector*>*> &photodots,
    set<const set<Hep2Vector*>*> &photolines) {

  for (set<set<Hep2Vector*> >::const_iterator cluster = point_clusters.begin();
      cluster != point_clusters.end(); cluster++) {

    // count the number of unique pads
    set<GlobalPadIndex> pad_cluster;

    for (set<Hep2Vector*>::const_iterator point = cluster->begin();
        point != cluster->end(); point++) {

      GlobalPadIndex global_index = pad_by_point.find(*point)->second;

      pad_cluster.insert(global_index);
    }

    if (pad_cluster.size() <= 10) {

      // it's most likely a dot
      photodots.insert(&(*cluster));

    } else {

      // it's most likely a line
      photolines.insert(&(*cluster));
    }
  }
}

void PhotoelectricProjectionFinder::store_results(
    const set<const set<Hep2Vector*>*> &photodots,
    const set<const set<Hep2Vector*>*> &photolines,
    const map<Hep2Vector*, GlobalPadIndex> pad_by_point, bool storeLCIO) {

  // calculate the centroids of the projected photodots
  set<pair<Hep2Vector, set<GlobalPadIndex> > > centroids_and_photodots; // and put them in here

  for (set<const set<Hep2Vector*>*>::const_iterator photodot =
      photodots.begin(); photodot != photodots.end(); photodot++) {

    // calculate the centre of the photodot and add the TrackerPulses to the TrackerHit
    Hep2Vector centroid(0., 0.);

    set<GlobalPadIndex> pad_cluster;

    for (set<Hep2Vector*>::const_iterator point = (*photodot)->begin();
        point != (*photodot)->end(); point++) {

      centroid += **point;

      GlobalPadIndex global_index = pad_by_point.find(*point)->second;

      pad_cluster.insert(global_index);
    }

    centroid *= 1. / (*photodot)->size(); // Hep2Vectors can't do division... weird

    centroids_and_photodots.insert(make_pair(centroid, pad_cluster));
  }
  const double max_dot_displacement_squared = _maximum_dot_displacement
      * _maximum_dot_displacement;

  // load AIDA (if it's available)
  AIDA::ITuple *centroid_tuple = 0;
  AIDA::ITuple *ignored_dots = 0;
  try {

    centroid_tuple = AIDAProcessor::tupleFactory(this)->create(
        "PhotoelectricCentroids",
        "Photodot centroid and estimated cetroid in data",
        "int i_dot, double x, double y, double x_data, double y_data", "");
    ignored_dots =  AIDAProcessor::tupleFactory(this)->create(
        "PhotodotsIgnored",
        "Photodots which have not been considered. Either they are not above modules or they could not be associated.",
        "int i_dot, double x, double y, int onModule","");
  } catch (lcio::Exception &e) {
    streamlog_out(WARNING) << "No AIDA processor found. Will not produce a root file..." << endl;
    streamlog_out(DEBUG)<< e.what() << endl;
  }
  const TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();
  PhotoelectricXMLFile photodotData(_photogeometry_file_name);
  vector<pair<double, vector<Photodot>::const_iterator> > sortedPhotodots = photodotData.GetSortedPhotodots(
      tpc_parameters.getDoubleVal("TPC_cylinder_centre_c0"),
      tpc_parameters.getDoubleVal("TPC_cylinder_centre_c1"));
  photodotData.SetTimeInfo(_photopulse_interval_start,_photopulse_interval_end);
  streamlog_out(DEBUG) << "Working on " << sortedPhotodots.size() << " sorted photodots." << endl;

  //prepare slcio file with reconstructed and true position to be used in the data base and to correct distortions
  lcio::LCWriter* lcWrt = NULL;
  lcio::LCEventImpl*  evt = NULL;
  lcio::LCCollectionVec* mappingCol = NULL;
  if(storeLCIO){
    lcWrt = lcio::LCFactory::getInstance()->createLCWriter();
    lcWrt->open( _distortion_file_name, lcio::LCIO::WRITE_NEW );
    // create new lcio event
    evt = new lcio::LCEventImpl();
    evt->setDetectorName(_detectorName);
    evt->setRunNumber(_runNumber);
    // set time to the time of creation
    LCTime time;
    evt->setTimeStamp(time.timeStamp());
    mappingCol = new lcio::LCCollectionVec( lcio::LCIO::LCGENERICOBJECT );
  }
  int foundPhotodotCentroids = centroids_and_photodots.size();
  for (vector<pair<double, vector<Photodot>::const_iterator> >::iterator it = sortedPhotodots.begin(); it != sortedPhotodots.end();
      it++) {
    Hep2Vector centroid = (*it).second->_position;
    int id_photodot = (*it).second->_id;

    // check if point is on a module (to remove points which have wrong alignment on the cathode, since they are all in the photodot file
    // get the TPC parameters
    if(ignored_dots && (tpc_parameters.getCoordinateType() == PadRowLayout2D::CARTESIAN)){
      if(!tpc_parameters.isInsideModule(centroid[0],centroid[1])){
        streamlog_out(DEBUG) << "The following photodot point from the photodot file is not on any module and will be ignored: ("
            << centroid[0] << ","
            << centroid[1] << ")." << endl;
        ignored_dots->fill(0,id_photodot);
        ignored_dots->fill(1,centroid[0]);
        ignored_dots->fill(2,centroid[1]);
        ignored_dots->fill(3,0);
        ignored_dots->addRow();
        continue;
      }
    } else if(ignored_dots && (tpc_parameters.getCoordinateType() != PadRowLayout2D::CARTESIAN)) {
      streamlog_out(WARNING) << "Global coordinate system is not cartesian and thus it was not tested if the phototdot from the given photodot file is on any module." << endl;
    }

    // check to see if any projections remain
    if (centroids_and_photodots.size() == 0) {

      streamlog_out(DEBUG)<< "\tPhotodot " << id_photodot << " has no projection. No projections remain." << endl;
      continue;
    }

    // find the closest projected photodot
    set<pair<Hep2Vector, set<GlobalPadIndex> > >::iterator closest_projection;
    double closest_distance_squared = 999999999.;
    const Hep2Vector* RecoPos = 0;
    for (set<pair<Hep2Vector, set<GlobalPadIndex> > >::iterator centroid_and_photodot =
        centroids_and_photodots.begin();
        centroid_and_photodot != centroids_and_photodots.end();
        centroid_and_photodot++) {

      const double distance_squared =
          (centroid - centroid_and_photodot->first).mag2();

      if (distance_squared < closest_distance_squared) {

        closest_distance_squared = distance_squared;
        closest_projection = centroid_and_photodot;
        RecoPos = &centroid_and_photodot->first;
        streamlog_out(DEBUG)<< "\tcloser projection at " << centroid_and_photodot->first << " is " << sqrt(distance_squared) << " mm away." << endl;
      }
    }

    // add this closest projection to the photodot element

    // make sure that it's close enough
    if (closest_distance_squared > max_dot_displacement_squared) {

      streamlog_out(DEBUG)<< "\tPhotodot " << id_photodot << " doesn't have a projection. The closest one ("
      << closest_distance_squared << ") is further than the maximum allowed displacement("
      << max_dot_displacement_squared << ")." << endl;
      if(ignored_dots){
        ignored_dots->fill(0,id_photodot);
        ignored_dots->fill(1,centroid[0]);
        ignored_dots->fill(2,centroid[1]);
        ignored_dots->fill(3,1);
        ignored_dots->addRow();
      }
      continue;
    }
    photodotData.AddPhotodotProjection(id_photodot,closest_projection->first, closest_projection->second);
    if (centroid_tuple) {
      centroid_tuple->fill(0, id_photodot);
      centroid_tuple->fill(1, centroid[0]);
      centroid_tuple->fill(2, centroid[1]);
      centroid_tuple->fill(3, closest_projection->first[0]);
      centroid_tuple->fill(4, closest_projection->first[1]);
      centroid_tuple->addRow();
    }
    //store data to slcio file
    if(storeLCIO){
      Hep3Vector truePos;
      truePos.set(centroid[0],centroid[1],0);
      Hep3Vector recoPos;
      recoPos.set(RecoPos->x(),RecoPos->y(),0);
      tpcconddata::TPCDistortionMeasurement * distortionmap = new tpcconddata::TPCDistortionMeasurement(&truePos, &recoPos, id_photodot);
      mappingCol->addElement( distortionmap );
    }
    centroids_and_photodots.erase(closest_projection);
  }
  foundPhotodotCentroids -= centroids_and_photodots.size();
  int foundCentroids = centroids_and_photodots.size();
  streamlog_out(MESSAGE) << "Photodot centroids associated to a photodot: " << foundPhotodotCentroids << endl;
  streamlog_out(MESSAGE) << "Photodot centroids not associated to a photodot: " << foundCentroids << endl;

  for (set<pair<Hep2Vector, set<GlobalPadIndex> > >::const_iterator centroid_and_photodot =
      centroids_and_photodots.begin();
      centroid_and_photodot != centroids_and_photodots.end();
      centroid_and_photodot++) {
    photodotData.AddPhotodotProjection(centroid_and_photodot->first, centroid_and_photodot->second);
  }

  // store the photolines
  for ( set< const set< Hep2Vector* >* >::const_iterator photoline = photolines.begin(); photoline != photolines.end(); photoline++ ) {
    set<GlobalPadIndex> pads;
    for ( set< Hep2Vector* >::const_iterator point = (*photoline)->begin(); point != (*photoline)->end(); point++ ) {
      GlobalPadIndex global_index = pad_by_point.find( *point )->second;
      pads.insert(global_index);
    }
    photodotData.AddPhotolineProjection((*pads.begin()).getModuleID(),pads);
  }
  photodotData.SaveFile(_projected_photogeometry_file_name);

  //store slcio file with reconstructed and true position to be used in the data base and to correct distortions
  if(storeLCIO){
    // add the mapping collection to the event
    evt->addCollection( mappingCol , "TPCDistortionMeasurement" ) ;
    // write the event to the file
    lcWrt->writeEvent( evt ) ;
    // dump the event to the screen
    lcWrt->close() ;
    // ------------ IMPORTANT ------------- !
    // we created the event so we need to delete it ...
    delete evt ;
  }

}

void PhotoelectricProjectionFinder::produce_tuples(
    const set<const set<Hep2Vector*>*> &photodots,
    const set<const set<Hep2Vector*>*> & /*photolines*/) {

  AIDA::ITuple *photodot_tuple = 0;
  try {

    photodot_tuple = AIDAProcessor::tupleFactory(this)->create("PadCounts",
        "Pad Counts per Photodot", "int n_pads", "");

  } catch (lcio::Exception &e) {

    streamlog_out(DEBUG)<< e.what() << endl;
    return;
  }

  // store the photodots
  for (set<const set<Hep2Vector*>*>::const_iterator photodot =
      photodots.begin(); photodot != photodots.end(); photodot++) {

    photodot_tuple->fill(0, static_cast<int>((*photodot)->size()));
    photodot_tuple->addRow();
  }

  // store the photolines
//  for (set<const set<Hep2Vector*>*>::const_iterator photoline =
//      photolines.begin(); photoline != photolines.end(); photoline++) {
//
//  }
}

}   // end namespace marlintpc

//#endif // USE_LCCD

