// MarlinTPC
#include "PhotoelectricPulseExtractionProcessor.h"
#include "PhotoelectricClusterTools.h"
#include "TPCDistortionMeasurement.h"
#include "LCObjectCopier.h"

// Marlin
#include <marlin/Exceptions.h>
#include <marlin/Global.h>

// AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>

//ROOT
#include "TMath.h"

// LCIO
#include <EVENT/LCCollection.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/LCFlagImpl.h>
#include <IMPL/LCEventImpl.h>
#include <UTIL/LCTime.h>

#include <map>
// GEAR
#include <gearxml/tinyxml.h>
#include "gear/TPCModule.h"

using namespace lcio;
using namespace marlin;
using namespace std;
using namespace gear;

namespace marlintpc
{

PhotoelectricPulseExtractionProcessor aPhotoelectricPulseExtractionProcessor;


PhotoelectricPulseExtractionProcessor::PhotoelectricPulseExtractionProcessor() : Processor("PhotoelectricPulseExtractionProcessor"),

        _input_tracker_pulses_collection_name("TPCPulses"),
        _output_photoelectric_tracker_pulses_collection_name_dots("TPCPhotodotsPulses"),
        _output_photoelectric_tracker_pulses_collection_name_lines("TPCPhotolinesPulses"),
        _output_is_transient(true),
        _projected_photogeometry_file_name("projected_photogeometry.xml"),
        _distortion_file_name("DistortionInformation.sclio"),
        _sigma(2.5)
{

    /* description of the processor which will be displayed in the
    * steering file autometicly generated by
    * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
    */

    _description = "PhotoelectricPulseExtractionProcessor searches for pulses from the photoelectric system and puts then into a subset collection.";

    /* register steering parameters: name, description, class-variable, default value
    * the type will by definded by the type of the default value
    * string, double, float and int are possible
    * plaese note to get a float you must cast in C++: (float) 1.3
    */
    registerInputCollection(LCIO::TRACKERPULSE ,
                            "InputTrackerPulses",
                            "Name of the input collection as stored in the LCIO file",
                            _input_tracker_pulses_collection_name,
                            std::string("TPCPulses"));

    registerOutputCollection(LCIO::TRACKERPULSE ,
                             "OutputPhotodotTrackerPulses",
                             "The name of the output collection of tracker pulses from the photodots.",
                             _output_photoelectric_tracker_pulses_collection_name_dots,
                             std::string("TPCPhotodotPulses"));
    registerOutputCollection(LCIO::TRACKERPULSE ,
                             "OutputPhotollineTrackerPulses",
                             "The name of the output collection of tracker pulses from the photolines.",
                             _output_photoelectric_tracker_pulses_collection_name_lines,
                             std::string("TPCPhotolinePulses"));

    registerOptionalParameter("SetOutputTransient",
                              "if not 0 the output collection is set transient",
                              _output_is_transient,
                              true);

    registerOptionalParameter("ProjectedPhotogeometryFilename",
                              "This is the name of the collection which stores the true positions of the photodots as SimTrackerHit objects. It is optional but it provides the number of photodots and, if provided, the processor will order the seed photodots to match the order of true photodots (useful for analysis).",
                              _projected_photogeometry_file_name,
                              _projected_photogeometry_file_name);

    registerOptionalParameter("DistortionOutputFile",
        "If this option is used the true and the reconstructed photodot position is stored as TPCDistortionMeasurement in the slcio file.",
        _distortion_file_name, _distortion_file_name);
    registerOptionalParameter("ExpectedClusterSize",
        "If you have choosen to write out the DistortionOutputFile than give here the expected cluster size in order to perform a likelihood fit.",
        _sigma,
        _sigma);
    registerOptionalParameter("FitSigma",
        "If true also sigma is fitted.",
        _fitSigma,
        false);


}


void PhotoelectricPulseExtractionProcessor::init()
{

    // usually a good idea to
    printParameters();

    _photodotFile = new PhotoelectricXMLFile(_projected_photogeometry_file_name);
    if(!_photodotFile->GetTimeInfo(_photopulse_interval_start, _photopulse_interval_end)){
      throw DataNotAvailableException("Projected geometry file doesn't contain an <arrival_time> element.");
    }

    if(_parameters->isParameterSet("DistortionOutputFile")){
      _dots = _photodotFile->GetPhotodots();
    }

    _numberOfPadsWithPulsesPerCluster = 0;
    _numberOfPulsesPerCluster = 0;
    _pulse_tuple = 0;
    _cluster_tuple = 0;

    try {
      AIDAProcessor::tree(this);
      _numberOfPadsWithPulsesPerCluster = new TH2D("PulsesPerCluster","PulsesPerCluster", 392, 0, 392, 20, -0.5, 19.5);
      _numberOfPulsesPerCluster = new TH2D("PadsPerClusterWithPulses","PadsPerClusterWithPulses", 392, 0, 392, 10, -0.5, 9.5);
      _pulse_tuple = AIDAProcessor::tupleFactory(this)->create(
          "PhotodotSeeds", "Cluster pulses", "int i_dot, double x, double y, float q",
          "");
      _cluster_tuple  = AIDAProcessor::tupleFactory(this)->create(
          "PhotodotCluster", "Cluster Reco", "int i_dot, double seed_x, double seed_y, float seed_variance_pad_x, float seed_variance_pad_y, float seed_variance_charge_x, float seed_variance_charge_y, double seed_err_x, double seed_err_y, double fit_x, double fit_y, float fit_err_x, float fit_err_y, double fit_sigma, floats sigma_err","");
    } catch (lcio::Exception &e) {

      streamlog_out(DEBUG)<< e.what() << endl;
      return;
    }


    streamlog_out(DEBUG0) << "Photopulse interval is from " << _photopulse_interval_start << " to " << _photopulse_interval_end << " ns." << endl;

    // add pads from photodots
    map<int, set<GlobalPadIndex> > photodotPads = _photodotFile->GetPhotodotPads();
    if(photodotPads.size() == 0){
      streamlog_out(ERROR) << "No pads associated to photodots found in the projection file: " << _projected_photogeometry_file_name.c_str() << endl;
    }
    int dots = 0;
    for(map<int, set<GlobalPadIndex> >::iterator it = photodotPads.begin(); it != photodotPads.end();it++){
      if((*it).second.size() != 0){
	if(_parameters->isParameterSet("DistortionOutputFile") || _numberOfPadsWithPulsesPerCluster || _numberOfPulsesPerCluster){
          streamlog_out(DEBUG) << "Add pads for photodot: " << (*it).first << endl;
          _photodot_pulse_sets[(*it).first] = set<EVENT::TrackerPulse*>();
          for(set<GlobalPadIndex>::iterator it_pad = (*it).second.begin(); it_pad != (*it).second.end(); it_pad++){
            _photodot_pulse_set_by_global_index[*it_pad] = (*it).first;
          }
	}
        _global_pad_indices_dots.insert((*it).second.begin(),(*it).second.end());
        dots++;
      }
    }

    // add pads from photolines
    int lines = 0;
    map<int, set<GlobalPadIndex> > photolinePads = _photodotFile->GetPhotolinePads();
    if(photolinePads.size() == 0){
      streamlog_out(ERROR) << "No pads associated to photolines found in the projection file: " << _projected_photogeometry_file_name.c_str() << endl;
    }
    for(map<int, set<GlobalPadIndex> >::iterator it = photolinePads.begin(); it != photolinePads.end();it++){
      _global_pad_indices_lines.insert((*it).second.begin(),(*it).second.end());
      if((*it).second.size() != 0)
        lines++;
    }
    streamlog_out(MESSAGE) << "Found " << dots << " photodots in the projection file with " << _global_pad_indices_dots.size() << " pads associated." << endl;
    streamlog_out(MESSAGE) << "Found " << lines << " photolines in the projection file with " << _global_pad_indices_lines.size() << " pads associated." << endl;

}


void PhotoelectricPulseExtractionProcessor::processRunHeader(lcio::LCRunHeader* run)
{
  _detectorName = run->getDetectorName();
  _runNumber = run->getRunNumber();
}

void PhotoelectricPulseExtractionProcessor::processEvent(lcio::LCEvent * evt)
{


    LCCollection* input_pulses = evt->getCollection(_input_tracker_pulses_collection_name);

    streamlog_out(DEBUG0) << "Searching " << input_pulses->getNumberOfElements() << " pulses from the collection " << _input_tracker_pulses_collection_name << endl;

    // create the output collection with a string defining the TYPE (see lcio.h)
    LCCollectionVec* output_photodot_pulses = new LCCollectionVec(LCIO::TRACKERPULSE);
    // set the transient flag as definded by the processor parameter
    output_photodot_pulses->setTransient(_output_is_transient);
    output_photodot_pulses->setSubset(true);

    LCCollectionVec* output_photoline_pulses = new LCCollectionVec(LCIO::TRACKERPULSE);
    // set the transient flag as definded by the processor parameter
    output_photoline_pulses->setTransient(_output_is_transient);
    output_photoline_pulses->setSubset(true);

    // set the flags that cellID1 should be stored
//    lcio::LCFlagImpl trkFlag(0) ;
//    trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
//    output_photodot_pulses->setFlag( trkFlag.getFlag() );
//    output_photoline_pulses->setFlag( trkFlag.getFlag() );

    std::map<int,int> pulses_perCluster;
    std::map<int,std::set<gear::GlobalPadIndex> > pads_perCluster;

    // now determine if this is a photoelectric pulse
    for (int i_tracker_pulse = 0; i_tracker_pulse < input_pulses->getNumberOfElements(); i_tracker_pulse++)
    {

      TrackerPulse* input_pulse = dynamic_cast<TrackerPulse*>(input_pulses->getElementAt(i_tracker_pulse)) ;

      if (!input_pulse)
      {

        throw "Input collection " + _input_tracker_pulses_collection_name + " does not contain TrackerPulses.";
      }

      const float time = input_pulse->getTime();

      // check if it's in the required time interval
      if ((time >= _photopulse_interval_start) && (time <= _photopulse_interval_end))
      {

        // determine if it is one of the required pads
        const GlobalPadIndex global_index(input_pulse->getCellID0(), input_pulse->getCellID1());

        if (_global_pad_indices_dots.find(global_index) != _global_pad_indices_dots.end())
        {

          output_photodot_pulses->addElement(input_pulse);
          streamlog_out(DEBUG1) << "Extracted pad " << global_index.getPadIndex() << " from module " << global_index.getModuleID() << " with charge " << input_pulse->getCharge() << " associated to a photodot." << endl;
          int id = _photodot_pulse_set_by_global_index[global_index];
          if(_numberOfPadsWithPulsesPerCluster || _numberOfPulsesPerCluster){
            pads_perCluster[id].insert(global_index);
            pulses_perCluster[id]++;
          }
          if(_parameters->isParameterSet("DistortionOutputFile")){
            TrackerPulse* input_pulse_copy = LCObjectCopier::copy(input_pulse);
            _photodot_pulse_sets[id].insert(input_pulse_copy);
            Vector2D pos = (Global::GEAR->getTPCParameters()).getModule((input_pulse_copy)->getCellID1()).getPadCenter((input_pulse_copy)->getCellID0());
            if(_pulse_tuple){
              _pulse_tuple->fill(0,id);
              _pulse_tuple->fill(1,pos[0]);
              _pulse_tuple->fill(2,pos[1]);
              _pulse_tuple->fill(3,input_pulse->getCharge());
              _pulse_tuple->addRow();
            }
          }

        }
        if (_global_pad_indices_lines.find(global_index) != _global_pad_indices_lines.end())
        {
          output_photoline_pulses->addElement(input_pulse);
          streamlog_out(DEBUG1) << "Extracted pad " << global_index.getPadIndex() << " from module " << global_index.getModuleID() << " with charge " << input_pulse->getCharge() << " associated to a photoline." << endl;
        }
      }
      else
      {

        streamlog_out(DEBUG1) << "Pulse rejected. Reason? time: " << time << "." << endl;
      }
    }

    if(_numberOfPadsWithPulsesPerCluster || _numberOfPulsesPerCluster){
      std::map<int,int>::iterator it = pulses_perCluster.begin();
      std::map<int,std::set<gear::GlobalPadIndex> >::iterator it_set = pads_perCluster.begin();
      while(it != pulses_perCluster.end()){
        _numberOfPadsWithPulsesPerCluster->Fill((*it).first,(*it).second);
        _numberOfPulsesPerCluster->Fill((*it_set).first,(*it_set).second.size());
        it++;
        it_set++;
      }
    }

    streamlog_out(DEBUG0) << "Found " << output_photodot_pulses->getNumberOfElements() << " pulses which could be from photodots." << endl;
    streamlog_out(DEBUG0) << "Found " << output_photoline_pulses->getNumberOfElements() << " pulses which could be from photoline." << endl;
    try{
      evt->addCollection(output_photodot_pulses, _output_photoelectric_tracker_pulses_collection_name_dots);
    } catch (EventException &e) {
      streamlog_out(ERROR) << "Failed to add the photodot collection with error: " << e.what() << endl;
    }
    try{
      evt->addCollection(output_photoline_pulses, _output_photoelectric_tracker_pulses_collection_name_lines);
    } catch (EventException &e) {
      streamlog_out(ERROR) << "Failed to add the photoline collection with error: " << e.what() << endl;
    }
}



void PhotoelectricPulseExtractionProcessor::check(lcio::LCEvent *  /*evt*/)
{
    // nothing to check here - could be used to fill checkplots in reconstruction processor
}


void PhotoelectricPulseExtractionProcessor::end()
{
  if(_parameters->isParameterSet("DistortionOutputFile")){
    //prepare slcio file with reconstructed and true position to be used in the data base and to correct distortions
    lcio::LCWriter* lcWrt = NULL;
    lcio::LCEventImpl*  evt = NULL;
    lcio::LCCollectionVec* mappingCol = NULL;
    lcio::LCCollectionVec* seedCol = NULL;
    lcio::LCCollectionVec* fitCol = NULL;
    lcWrt = lcio::LCFactory::getInstance()->createLCWriter();
    lcWrt->open( _distortion_file_name, lcio::LCIO::WRITE_NEW );
    // create new lcio event
    evt = new lcio::LCEventImpl();
    evt->setDetectorName(_detectorName);
    evt->setRunNumber(_runNumber);
    // set time to the time of creation
    LCTime time;
    evt->setTimeStamp(time.timeStamp());
    mappingCol = new lcio::LCCollectionVec( lcio::LCIO::LCGENERICOBJECT );
    seedCol = new lcio::LCCollectionVec( LCIO::TRACKERHIT );
    fitCol = new lcio::LCCollectionVec( LCIO::TRACKERHIT );
    for(vector<vector<Photodot>::const_iterator>::iterator it = _dots.begin();
        it != _dots.end(); it++){
      TrackerHitImpl *seed_photodot = new TrackerHitImpl();
      _photodot_pulse_sets.find((*it)->_id);
      streamlog_out(DEBUG) << "testing dot " << (*(*it))._id << endl;
      if(_photodot_pulse_sets.find((*it)->_id) == _photodot_pulse_sets.end())
        continue;
      streamlog_out(MESSAGE) << "Found projections for dot " << (*it)->_id << "  try to find seeds and fit it..." << endl;
      CLHEP::Hep2Vector seed_position = calculateSeed(_photodot_pulse_sets[(*it)->_id], Global::GEAR->getTPCParameters(),seed_photodot);
      streamlog_out(MESSAGE) << "...found seed " << seed_photodot->getPosition()[0] << " : " << seed_photodot->getPosition()[1] << "(1 error: " <<TMath::Sqrt(seed_photodot->getCovMatrix()[0]) << ", 2 error: " << TMath::Sqrt(seed_photodot->getCovMatrix()[1]) << ")" <<  endl;
      if(_cluster_tuple){
        _cluster_tuple->fill(0,(*it)->_id);
        _cluster_tuple->fill(1,seed_photodot->getPosition()[0]);
        _cluster_tuple->fill(2,seed_photodot->getPosition()[1]);
        _cluster_tuple->fill(3,seed_photodot->getCovMatrix()[0]);
        _cluster_tuple->fill(4,seed_photodot->getCovMatrix()[1]);
        _cluster_tuple->fill(5,seed_photodot->getCovMatrix()[2]);
        _cluster_tuple->fill(6,seed_photodot->getCovMatrix()[3]);                                
        _cluster_tuple->fill(7,TMath::Sqrt(seed_photodot->getCovMatrix()[0]+seed_photodot->getCovMatrix()[2]));
        _cluster_tuple->fill(8,TMath::Sqrt(seed_photodot->getCovMatrix()[1]+seed_photodot->getCovMatrix()[3]));        
      }
      seedCol->addElement( seed_photodot);
      TrackerHitImpl *fit_photodot = new TrackerHitImpl();
      bool isValid = doLikelihoodFit(*seed_photodot, fit_photodot, _sigma, _fitSigma);
      CLHEP::Hep3Vector truePos, recoPos;
      truePos.set((*it)->_position.x(),(*it)->_position.y(),0);
      if(isValid){
        streamlog_out(MESSAGE) << "...fit success " << fit_photodot->getPosition()[0] << " : " << fit_photodot->getPosition()[1] << endl;
        recoPos.set(fit_photodot->getPosition()[0],fit_photodot->getPosition()[1],0);
        fitCol->addElement(fit_photodot);
        if(_cluster_tuple){
          streamlog_out(MESSAGE) << "Filling : " << TMath::Sqrt(fit_photodot->getCovMatrix()[0]) << " and " << TMath::Sqrt(fit_photodot->getCovMatrix()[1]) << endl;
          _cluster_tuple->fill(9,fit_photodot->getPosition()[0]);
          _cluster_tuple->fill(10,fit_photodot->getPosition()[1]);
          _cluster_tuple->fill(11,fit_photodot->getCovMatrix()[0]);
          _cluster_tuple->fill(12,fit_photodot->getCovMatrix()[1]);
          _cluster_tuple->fill(13,fit_photodot->getPosition()[2]);
          _cluster_tuple->fill(14,fit_photodot->getCovMatrix()[2]);                                                  
        }
      } else {
        recoPos.set(seed_photodot->getPosition()[0],seed_photodot->getPosition()[1],0);
      }
      tpcconddata::TPCDistortionMeasurement * distortionmap = new tpcconddata::TPCDistortionMeasurement(&truePos, &recoPos, (*it)->_id);
      mappingCol->addElement( distortionmap );
  	  _cluster_tuple->addRow();
    }
    // add the mapping collection to the event
    evt->addCollection( mappingCol , "TPCDistortionMeasurement" ) ;
    evt->addCollection( seedCol , "PhotodotSeeds" ) ;
    evt->addCollection( fitCol , "PhotodotFittedPosition" ) ;
    // write the event to the file
    lcWrt->writeEvent( evt ) ;
    // dump the event to the screen
    lcWrt->close() ;
    // ------------ IMPORTANT ------------- !
    // we created the event so we need to delete it ...
    delete evt ;
  }
  delete _photodotFile;
}

}
