#ifndef TIMEPIXCALIBRATIONPROCESSOR_H
#define TIMEPIXCALIBRATIONPROCESSOR_H

// C++
#include <string>
#include <fstream>

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "TimePixPixelMode.h"
#include "TimePixPixelInterpolationProcessor.h"
#include "TimePixCalibParam.h"
#include "lccd/ConditionsMap.hh"

//AIDA
#include <AIDA/AIDA.h>

namespace AIDA
{
class IHistogram2D;
}

namespace marlintpc
{

/* the following comment is tranfered to doxygen documentation
 * which begins with a second asterix, or in case of on line comment an extra third slash
 */

/**  Calibration data converting Processor of the MarlinTPC package .
 *
 *   modify and corrects the ToT and Time pixel values with the chip calibration data
 *
 *  <h4>Input - Prerequisites</h4>
 *  a collection of TrackerData
 *  a collection of calibration parameters
 *
 *  <h4>Output</h4>
 *  Collection containing calibrated TrackerData.
 *
 * @param InputCollectionName Name of TimePix data collection
 * @param InputCalibrationData Name of calibration parameters collection
 * @param OutputCollectionName Name of calibrated data collection
 * @param NRow Number of rows on the chip, default 256
 * @param NColumn Number of columns on the chip, default 256
 * @param NParameters Number of calibration parameters per chip, default 8
 * @param MinNoPixelToInterpolate Minimum number of surrounding pixels needed to interpolate, default 2
 * @param VTHL Value of the THL used for the data
 * @param SetOutputTransient If not 0 the output collection is set transient
 *
 *
 * @author R.Ulman, Bonn
 * @author modified and corrected by C.Brezina
 */

class TimePixCalibrationProcessor : public marlin::Processor
{
    public:

        /* This method will be called by the marlin package
         * It returns a processor of the currend type (here TimePixCalibrationProcessor)
         */
        Processor* newProcessor()
        {
            return new TimePixCalibrationProcessor;
        }

        /* the default constructor
         * here the processor parameters are registered to the marlin package
         * other initialisation should be placed in the init method
         */
        TimePixCalibrationProcessor();

        /* Called at the begin of the job before anything is read.
         * Use to initialize the processor, e.g. book histograms
         */
        void init();

        /* Called for every run.
         * in this processor it is not used
         */
        void processRunHeader(lcio::LCRunHeader* run);

        /* Called for every event - the working horse.
         * Here the real conversion take place
         */
        void processEvent(lcio::LCEvent* evt);

        /* Called after data processing for clean up.
         * e.g. saving contol histograms, delete created objects
         * (which are not stored in a collection/event), etc.
         */
        void end();


    protected:

        /* normally the doxygen documentation is placed before the method/variable
         * but for one line comments in case of variables is may be more convenient to you
         * to place it after the variable. In this case you use an arrow to tell doxygen
         * to which variable the comment belongs to.
         */

        /* For input and output maps: 4 histograms are created in the processor for each of the chips,
         * to compare the pixel map before and after the calibration.
         * (for single chips only one of the histograms will be full, while the others are empty)
         */
        AIDA::IHistogram2D * _inputMapHisto0;
        AIDA::IHistogram2D * _outputMapHisto0;
        AIDA::IHistogram2D * _inputMapHisto1;
        AIDA::IHistogram2D * _outputMapHisto1;
        AIDA::IHistogram2D * _inputMapHisto2;
        AIDA::IHistogram2D * _outputMapHisto2;
        AIDA::IHistogram2D * _inputMapHisto3;
        AIDA::IHistogram2D * _outputMapHisto3;

        std::string _inputColName;  ///<- the name of the input collection
        std::string _inputCalibrationData;  ///<-  the name of the calibration data collection
        std::string _outputColName; ///<- the name of the output collection

        std::string _dataFileList;  ///<- name of input calibration parameters file list
        std::ifstream _infile;    ///<- input for calibration parameters

        int _nColumn;  ///<- number of columns on the chip
        int _nRow;  ///<- number of rows on the chip
        int _nParameters;  ///<- number of parameters in calibration data per chip
        int _minNoPixelToInterpolate;  ///<- minimum number of pixels needed to interpolate
        double _vTHL;  ///<- THL value used for data, must be entered with value used in specific run

        /** give the status which is set to the transient flag of the output collection
         */
        bool _outputIsTransient;

        int _nRun;  ///<- counts the runs processed
        int _nEvt;  ///<- counts the events processed

        /** Called to create input map with pixel values for the interpolation (needed for Time-Walk correction)
         */
        void createInputPixelMap(LCCollection* inCol);

        /** Called to preform calibration on pixel in TIME mode (correct for Clock distribution & Time-Walk)
         */
        float calibrateTimePixel(int chipID, int pixelID, float pixelValue);

        /** Called to preform calibration on pixel in TOT mode (convert into electrons and add THL offset)
         */
        float calibrateChargePixel(int chipID, float pixelValue);

        /** Called to interpolate charge for pixels in TIME mode (by using surrounding pixels in TOT mode)
         */
        float interpolatePixelCharge(int chipID, int pixelID);

        tpcconddata::TimePixMap* map;    ///<- input map needed for the interpolation function
        std::map< tpcconddata::TimePixPixelMode::PixelKey, float> inMap;


        /* Called to read the calibration parameters by their index (param)
         *
         */
        float getParameter(int chipID, int param);

        bool _isFirstEvent;
        LCCollection* _vCalib;

};
} // class declaration
#endif



