#ifndef TimePixHitCenterCalculatorProcessor_h
#define TimePixHitCenterCalculatorProcessor_h 1

// C++
#include <string>

// LCIO
#include "lcio.h"

// Aida
//#ifdef MARLIN_USE_AIDA // the whole processor does not make sense without aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
// endif

//root
#include <TH1F.h>

// Marlin
#include "marlin/Processor.h"
#include "TimePixPixelMode.h"
#include "lccd/ConditionsMap.hh"
#include "TimePixMapListener.h"

namespace AIDA
{
  class IHistogram2D;
  class IHistogram3D;

}

namespace marlintpc
{

/* the following comment is tranfered to doxygen documentation
  * which begins with a second asterix, or in case of on line comment an extra third slash
  */
/**
 *  calculates the center of gravity of each cluster, weighted by charge information.
 *
 *  <h4>Input - Prerequisites</h4>
 *  a collection of TrackerHits, position not set
 *
 *  <h4>Output</h4>
 *  Collection of TrackerHits with the position and dEdx set.
 *
 * @param InputCollectionName: Name of input collection, TrackerHits without position
 * @param OutputCollectionName: Name of output collection, TrackerHits with position
 * @param ModeMapCollectionName: Name of mode map collection
 * @param _max_zDiffusion_perCluster: pixels further than _max_zDiffusion_perCluster [clock cycles] away from the mean cluster time are ignored during cluster center calculation
 * @param SetOutputTransient: If not 0 the output collection is set transient
 * @param nColumn: number of columns on the chip
 * @param nRow: number of rows on the chip
 * @param pitch: pitch of the pixel
 * @param _minNrOfPixelsToCalcTime: if there are less than _minNrOfPixelsToCalcTime with 4 or more neighbors with charge all time pixels are taken into account for mean time calculation
 * @param _keepClusterHistoFlag: decide whether you want to keep a histo of z-Distribution for each Cluster - default is FALSE
 * @param rmsCut: pixels further away in z than rmsCut * RMSofDistributon are ignorred for Meancalculation
 * @param SetOutputTransient: If not 0 the output collection is set transient
 *
 * @author Simone Zimmermann
 */


class TimePixHitCenterCalculatorProcessor : public marlin::Processor
{

    public:

        /* This method will be called by the marlin package
         * It returns a processor of the currend type (here TimePixHitCenterCalculatorProcessor)
         */
        virtual Processor*  newProcessor()
        {
            return new TimePixHitCenterCalculatorProcessor ;
        }

        /* the default constructor
         * here the processor parameters are registered to the marlin package
         * other initialisation should be placed in the init method
         */
        TimePixHitCenterCalculatorProcessor() ;

        /* Called at the beginning of the job before anything is read.
         * Use to initialize the processor, e.g. book histograms
         */
        virtual void init() ;

        /* Called for every run.
         * in this processor it is not used
         */
        virtual void processRunHeader(lcio::LCRunHeader* run) ;

        /* Called for every event - the working horse.
         * Here the real conversion take place
         */
        virtual void processEvent(lcio::LCEvent * evt) ;

        /* This method is only called if the check flag is set (default) in the main processor
         * (where the input files are given)
         * It should be used to create check plots which are not needed in a complete analysis
         * (where normaly the check flag is deactivated)
         */
        virtual void check(lcio::LCEvent * evt) ;


        /* Called after data processing for clean up.
         * e.g. saving control histogramms, delete created objects
         * (which are not stored in a collection/event), etc.
         */
        virtual void end() ;

    protected:

        /* normally the doxygen documentation is palced before the method/variable
         * but for one line comments in case of variables it may be more convenied for you
         * to place it after the variable. In this case you use an arrow to tell doxygen
         * to which variable the commet belongs to.
         */

        std::string _inputColName ;  ///<- the name of the input collection
        std::string _outputColName ;  ///<- the name of the output collection
        std::string _modeMapColName ;  ///<- the name of the mode map collection
        std::string _modeMapCorrName ;  ///<- the name of the correction to mode map collection
        
        /*upper threshold of clock cycles, which a pixel is allowed to have to 
         *be part of the cluster center calculation
         */   
        int _max_zDiffusion_perCluster;
   
        double _shutter;  ///<- length of shutter signal in us
        double _readoutFrequency; ///<- readout frequency of TimePix chip in MHz, optional, should come from gear
        double _driftVelocity; ///<- drift velocity of gas in mm/us
        bool _multiShutter; ///<- true if multiple shutters are allowed during acquisition
        bool _keepClusterHistoFlag; ///<- true if you want to keep a histo of the z-distribution of each cluster
	
        /*if there are less than _minNrOfPixelsToCalcTime with 4 or more neighbors with charge 
         *all time pixels are taken into account for mean time calculation
         */
        int _minNrOfPixelsToCalcTime;
        
        /*threshold for the meancalculation, all pixel further away in z than 
         *rmsCut * RMSofDistributon are ignorred
         */
        int _rmsCut;

        /* give the status which is set to the transient flag of the output collection
         * it is an int instead of a boolean, because the processor parameter cannot be a boolean
         */
        int _outputIsTransient;
        
        AIDA::IHistogram1D * _zDistancesPixelsInClusters_beforeCleaning; ///<- _zDistancesPixelsInClusters_beforeCleaning
	AIDA::IHistogram1D * _zDistancesPixelsInClusters_afterCleaning; ///<- _zDistancesPixelsInClusters_afterCleaning
	AIDA::IHistogram1D * _zclusterCenterBeforeMinusAfterCleaning; ///<- z shift of Clustercenter introduced by cutting on special Pixels
	TH1F * _clusterHisto; ///< Root histo containing z Values of Pixels in a Cluster
       
        tpcconddata::TimePixMap * _timePixMap;

        int countNrOfNeighboringPixelsWithHit( int chipID, int pixelID, std::vector<LCObject*> DataVec );


    private:
//  AIDA::IHistogram2D * histoCluster;
//  AIDA::IHistogram2D * histoSP;
//  AIDA::IHistogram3D * histo3D;

} ;

}

#endif



