#ifndef TimePixZeroSuppressionProcessor_h
#define TimePixZeroSuppressionProcessor_h 1

// C++
#include <string>

// LCIO
#include "lcio.h"

// tpcconditions
#include "TPCCondDataExeptions.h"
#include "ADCChannelMapping.h"
#include "ChannelMappingListener.h"
// Marlin
#include "marlin/Processor.h"

namespace AIDA
{
class IHistogram2D;
}

namespace marlintpc
{

/* the following comment is tranfered to doxygen documentation
 * which begins with a second asterix, or in case of on line comment an extra third slash
 */
/**  zero suppression Processor of the MarlinTPC package .
 *
 *   converts TrackerRawData of a TimePix into zero suppressed TrackerRawData of a TimePix
 *
 *  <h4>Input - Prerequisites</h4>
 *  a collection of TrackerRawData
 *
 *  <h4>Output</h4>
 *  Collection containing zero suppressed TrackerRawData.
 *
 * @param InputCollectionName Name of input collection, TrackerRawData
 * @param OutputCollectionName Name of output collection, zero suppressed TrackerRawData
 * @param SetOutputTransient If not 0 the output collection is set transient
 * @param TOTThreshold Threshold on the TOT value of a single pixel
 * @param NColumn number of columns of the chip
 * @param NRow number of rows of the chip
 * @param SkipEmptyEvents Throw SkipEventException if TimePix data is empty.
 *        All subsequent processors are skipped, including the LCIOOutputProcessor.
 *        You might want to switch this off in case there is data from several
 *        subdetectors in the event.
 *        Default: 1
 * @param ChannelMappingCollectionName Optional parameter:  Name of the LCCD collection with channel mapping (default: ADCChannelMapping).
 * 		  If parameter is not given, the cellids of the input will be used.
 * @param SingleRawData Optional parameter: create a TrackerRawData object for every pixel? default: false
 *
 * @author Simone Zimmermann; edit Felix Mueller
 */
class TimePixZeroSuppressionProcessor : public marlin::Processor
{

    public:

        /* This method will be called by the marlin package
         * It returns a processor of the currend type (here TimePixZeroSuppressionProcessor)
         */
        virtual Processor*  newProcessor()
        {
            return new TimePixZeroSuppressionProcessor ;
        }

        /* the default constructor
         * here the processor parameters are registered to the marlin package
         * other initialisation should be placed in the init method
         */
        TimePixZeroSuppressionProcessor() ;

        /* Called at the beginning of the job before anything is read.
         * Use to initialize the processor, e.g. book histograms
         */
        virtual void init() ;

        /* Called for every run.
         * in this processor it is not used
         */
        virtual void processRunHeader(lcio::LCRunHeader* run) ;

        /* Called for every event - the working horse.
         * Here the real conversion take place
         */
        virtual void processEvent(lcio::LCEvent * evt) ;

        /* This method is only called if the check flag is set (default) in the main processor
         * (where the input files are given)
         * It should be used to create check plots which are not needed in a complete analysis
         * (where normaly the check flag is deactivated)
         */
        virtual void check(lcio::LCEvent * evt) ;


        /* Called after data processing for clean up.
         * e.g. saving control histogramms, delete created objects
         * (which are not stored in a collection/event), etc.
         */
        virtual void end() ;


    protected:

        /* normally the doxygen documentation is palced before the method/variable
         * but for one line comments in case of variables it may be more convenied for you
         * to place it after the variable. In this case you use an arrow to tell doxygen
         * to which variable the commet belongs to.
         */

        std::string _inputColName ;  ///<- the name of the input collection
        std::string _outputColName ; ///<- the name of the output collection

        /** optional parameter: the name of the channel map
         * if the parameter is set, the processor maps the cellIDs according to the channelmap
         * otherwise the cellIDs are not mapped
         */
        std::string _chMapCollection ; ///<- the name of the channel map

        double _totThreshold ; ///<-Threshold on the TOT value of a single pixel
        int _nColumn ; ///<-Number of columns of the chip
        int _nRow ; ///<-Number of rows of the chip

        /** give the status which is set to the transient flag of the output collection
         * it is an int instead of a boolean, because the processor parameter cannot be a boolean
         */
        int _outputIsTransient;

        /** is true till the first output collection is wirten to an event
         *  used to to store parameters only in the first collection
         *  (also done, if collection is transient)
         */
        bool isFirstCollection;

        /** if true a TrackerRawData object is created for every single pixel
         *  default is false: neighboring pixels are added into a single TrackerRawData object
         *  by writing their value into the adcvector
         */
        bool _singleRawData;

        int _nRun ; ///<- counts the processed runs
        int _nEvt ; ///<- counts the processed events

        // this  is an int instead of a boolean, because the processor parameter cannot be a boolean
        int _skipEmptyEvents; ///< Flag whether to throw SkipEventException for epmty events

    private:
        AIDA::IHistogram2D* histoCluster;
        AIDA::IHistogram2D* histoClusterRaw;
        ChannelMappingListener* _ChannelMappingListener;
} ;

}

#endif



