#include "TimePixCalibrationProcessor.h"
#include "TimePixPixelInterpolationProcessor.h"
#include "TimePixPixelMode.h"
#include "TimePixCalibParam.h"

// C++
#include <iostream>
#include <math.h>

// LCIO
#include <EVENT/LCCollection.h>
//#include "EVENT/LCObject.h"
#include <EVENT/TrackerRawData.h>
#include <EVENT/TrackerData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/LCFlagImpl.h>

//Marlin
#include <marlin/Exceptions.h>

// Aida
#ifdef MARLIN_USE_AIDA
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/ICloud1D.h>
#endif



using namespace lcio;
using namespace marlin;
using namespace tpcconddata;

namespace marlintpc
{
TimePixCalibrationProcessor aTimePixCalibrationProcessor;

TimePixCalibrationProcessor::TimePixCalibrationProcessor() : Processor("TimePixCalibrationProcessor"),
		_inputMapHisto0(0),_outputMapHisto0(0),_inputMapHisto1(0),_outputMapHisto1(0),_inputMapHisto2(0),_outputMapHisto2(0),_inputMapHisto3(0),_outputMapHisto3(0),
		_nRun(0),_nEvt(0),map(0),_isFirstEvent(false),_vCalib(0)
{
    /* the constructor call first the super constructor with the name
     * of the processor
     * this must be the same as the class name.
     */

    /* description of the processor which will be displayed in the
     * steering file autometicly generated by
     * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
     */
    _description = "TimePixCalibrationProcessor modify and corrects the ToT and Time pixel values with the chip calibration data. I guess we agreed to ignore the possible existence of the DistributionDelay therefore the input map should be empty!";

    /* register steering parameters: name, description, class-variable, default value
     * the type will by definded by the type of the default value
     * string, double, float and int are possible
     * please note to get a float you must cast in C++: (float) 1.3
     */
    registerInputCollection(LCIO::TRACKERDATA,
                            "InputCollectionName" ,
                            "Name of the input collection"  ,
                            _inputColName ,
                            std::string("TimePixZeroSuppressedData"));

    registerInputCollection(LCIO::LCGENERICOBJECT,
                            "InputCalibrationData" ,
                            "Name of the input calibration collection"  ,
                            _inputCalibrationData ,
                            std::string("CalibrationParameters"));

    registerOutputCollection(LCIO::TRACKERDATA,
                             "OutputCollectionName" ,
                             "Name of the output collection"  ,
                             _outputColName ,
                             std::string("TimePixCalibratedData"));

    registerProcessorParameter("nColumn" ,
                               "number of columns of the chip"  ,
                               _nColumn ,
                               256) ;

    registerProcessorParameter("nRow" ,
                               "number of rows of the chip"  ,
                               _nRow ,
                               256) ;

    registerProcessorParameter("nParameters" ,
                               "number of extra parameters in calibration data"  ,
                               _nParameters ,
                               8) ;

    registerProcessorParameter("MinNoPixelToInterpolate" ,
                               "Minimum number of surrounding pixels needed to interpolate"  ,
                               _minNoPixelToInterpolate,
                               2) ;

    registerProcessorParameter("VTHL" ,
                               "value of THL used for data"  ,
                               _vTHL ,
                               400.0) ;

    registerProcessorParameter("SetOutputTransient" ,
                               "if not 0 the output collection is set transient"  ,
                               _outputIsTransient,
                               false);

    //prevent _minNoPixelToInterpolate to be set to values smaller than 1
    if(_minNoPixelToInterpolate <= 0)
    {
    	streamlog_out(WARNING) << "MinNoPixelToInterpolate is <= 0: This will NOT work! Setting MinNoPixelToInterpolate to 1" << std::endl;
    	_minNoPixelToInterpolate = 1;
    }
}



void TimePixCalibrationProcessor::init()
{
    // usually a good idea to
    printParameters();

    _isFirstEvent = true;

    _nRun = 0 ;
    _nEvt = 0 ;

    // histograms of input and output pixel maps for a quad-board
    _inputMapHisto0 =   AIDAProcessor::histogramFactory(this)->
                        createHistogram2D("inputMapHisto0",   "inputMapHisto0",
                                          _nRow,    0, _nRow - 1,
                                          _nColumn, 0, _nColumn - 1);

    _outputMapHisto0 =   AIDAProcessor::histogramFactory(this)->
                         createHistogram2D("outputMapHisto0",   "outputMapHisto0",
                                           _nRow,    0, _nRow - 1,
                                           _nColumn, 0, _nColumn - 1);

    _inputMapHisto1 =   AIDAProcessor::histogramFactory(this)->
                        createHistogram2D("inputMapHisto1",   "inputMapHisto1",
                                          _nRow,    0, _nRow - 1,
                                          _nColumn, 0, _nColumn - 1);

    _outputMapHisto1 =   AIDAProcessor::histogramFactory(this)->
                         createHistogram2D("outputMapHisto1",   "outputMapHisto1",
                                           _nRow,    0, _nRow - 1,
                                           _nColumn, 0, _nColumn - 1);

    _inputMapHisto2 =   AIDAProcessor::histogramFactory(this)->
                        createHistogram2D("inputMapHisto2",   "inputMapHisto2",
                                          _nRow,    0, _nRow - 1,
                                          _nColumn, 0, _nColumn - 1);

    _outputMapHisto2 =   AIDAProcessor::histogramFactory(this)->
                         createHistogram2D("outputMapHisto2",   "outputMapHisto2",
                                           _nRow,    0, _nRow - 1,
                                           _nColumn, 0, _nColumn - 1);

    _inputMapHisto3 =   AIDAProcessor::histogramFactory(this)->
                        createHistogram2D("inputMapHisto3",   "inputMapHisto3",
                                          _nRow,    0, _nRow - 1,
                                          _nColumn, 0, _nColumn - 1);

    _outputMapHisto3 =   AIDAProcessor::histogramFactory(this)->
                         createHistogram2D("outputMapHisto3",   "outputMapHisto3",
                                           _nRow,    0, _nRow - 1,
                                           _nColumn, 0, _nColumn - 1);

    // get input Timepix pixel map
        map = TimePixMapHandlerProcessor::getTimePixMap() ;
        // create map with pixel values for pixel interpolation

}


void TimePixCalibrationProcessor::processRunHeader(lcio::LCRunHeader* run)
{
    run->parameters().setValue(_processorName + "_revision", "$Rev: $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
                }
        }

    _nRun++ ;
}


void TimePixCalibrationProcessor::processEvent(lcio::LCEvent* evt)
{

    // get the input collection from the event
    LCCollection* inputCol = 0;
    try
        {
            inputCol = evt->getCollection(_inputColName);
//       std::cout << inputCol->getNumberOfElements() << std::endl;
        }
    catch (DataNotAvailableException&)
        {
            // if there is no input collection give a warning to the user and skip event
            streamlog_out(WARNING0) << "Warning of TimePixCalibrationProcessor::processEvent: input collection "
            << _inputColName
            << " cannot be found in event "
            << evt->getEventNumber() << std::endl;
            return;
        }
    streamlog_out(MESSAGE0) << "Processing "
    << inputCol->getNumberOfElements()
    << " data elements from collection "
    << _inputColName << "." << std::endl;

    // create the output collection with a string defining the TYPE (see lcio.h)
    LCCollectionVec* outputCol = new LCCollectionVec(LCIO::TRACKERDATA);

    // set the transient flag as defined by the processor parameter
    outputCol->setTransient(_outputIsTransient != 0);

    // Calibration parameters collection

    if (_isFirstEvent)
        {

            _vCalib = 0 ;
            try
                {
                    _vCalib = evt->getCollection(_inputCalibrationData);
                    _isFirstEvent = false;
                }
            catch (DataNotAvailableException&)
                {
                    // if there is no input collection give a warning to the user
                    streamlog_out(WARNING0) << "Warning of TimePixCalibrationProcessor::processEvent: calibration collection "
                    << _inputCalibrationData
                    << " cannot be found in event "
                    << evt->getEventNumber() << std::endl;
                    return;
                }
        }
    streamlog_out(MESSAGE0) << "Processing "
    << _vCalib->getNumberOfElements()
    << " calibration elements from parameters collection "
    << _inputCalibrationData << "." << std::endl;

    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1) ;
    outputCol->setFlag(trkFlag.getFlag());

    streamlog_out(MESSAGE0) << "call to createInputPixelMap() " << std::endl;

    createInputPixelMap(inputCol) ;

    // get the number of elements in the input collection and loop over it
    for (int iData = 0; iData < (inputCol->getNumberOfElements()); iData++)
        {
            /* getElement return only a LCObject
             * you have to cast it to get the right type
             * please use dynamic_cast which will return 0 if the object can not be casted
             */
            TrackerData* nData = dynamic_cast<TrackerData*>(inputCol->getElementAt(iData));

            if (nData != 0)
                {
                    // only proceed if the cast was successful

                    /* create a new object of the TrackerData type
                     * you have to use the IMPL, otherwise you can not set values
                     */
                    TrackerDataImpl* calibratedData = new TrackerDataImpl();

                    // transfer the values

                    // first the simple
                    calibratedData->setCellID0(nData->getCellID0());
                    calibratedData->setCellID1(nData->getCellID1());
                    calibratedData->setTime(nData->getTime());

                    // and now the vector
                    FloatVec CDataADCVal =  nData->getChargeValues() ;

                    // get chip ID
                    int chipID = nData->getCellID1() ;

                    int element_pixelID = nData->getCellID0();

                    for (unsigned int i = 0; i < CDataADCVal.size(); i++)
                        {
                            // get pixel ID
                            int pixelID = element_pixelID + i ;

                            // get pixel mode
                            int pm = ((map->find(TimePixPixelMode::PixelKey(chipID, pixelID)))->second)->getPixelMode() ;

                            // get pixel value
                            float PV = CDataADCVal[i] ;

                            //std::cout << " \t pixelID: " << pixelID << " \t chipID: " << chipID <<  "\t element: " << iData <<  "\t mode: " << pm << std::endl;

                            // if pixel in Timepix mode
                            if (pm == TimePixPixelMode::TIMEPIX)
                                {
                                    // correct Time value for Clock-delay and Time-Walk
                                    (calibratedData->chargeValues()).push_back(calibrateTimePixel(chipID, pixelID, PV)) ;
                                }
                            else
                                // if pixel in ToT mode
                                if (pm == TimePixPixelMode::TOT)
                                    {
                                        // convert ToT value into electrons and add THL offset
                                        (calibratedData->chargeValues()).push_back(calibrateChargePixel(chipID, PV)) ;
                                    }
                                else
                                    // if pixel in Medipix/Hit mode
                                    // if( pm == TimePixPixelMode::MEDIPIX || pm == TimePixPixelMode::ONEHIT )
                                    (calibratedData->chargeValues()).push_back(0);

                            // output histogram pixel index for column (OC) and for row (OR)
                            int OC = pixelID % _nRow ;
                            int OR = static_cast<int>((pixelID - OC) / _nColumn) ;

                            // fill histograms using pixelman notation for chip ID
                            switch (chipID)
                                {
                                case 1:
                                    _outputMapHisto0->fill(OC, OR, calibratedData->chargeValues()[i]) ;     // histogram of output pixel map chip 0
                                    break;
                                case 0:
                                    _outputMapHisto1->fill(OC, OR, calibratedData->chargeValues()[i]) ;     // histogram of output pixel map chip 1
                                    break;
                                case 2:
                                    _outputMapHisto2->fill(OC, OR, calibratedData->chargeValues()[i]) ;     // histogram of output pixel map chip 2
                                    break;
                                case 3:
                                    _outputMapHisto3->fill(OC, OR, calibratedData->chargeValues()[i]) ;     // histogram of output pixel map chip 3
                                    break;
                                default:
                                    break;
                                }

                        }

                    /* fill the TrackerData in the collection
                     * please note: DO NOT DELETE the object ,this will be done by LCIO
                     */
                    outputCol->addElement(calibratedData);

                }           // if( nData != 0 )
        }            // for(int iData=0; iData< inputCol->getNumberOfElements(); iData++)

    /* after converting all raw data objects
     * add the collection to the event and asign a name to it
     * please note again: DO NOT DELETE the collection ,this will be done by LCIO
     */
    evt->addCollection(outputCol, _outputColName);

}           // void TimePixCalibrationProcessor::processEvent(lcio::LCEvent* evt)



void TimePixCalibrationProcessor::createInputPixelMap(LCCollection* inCol)
{


    //create input map of inCol values with PixelKey as key
    inMap.clear();

    for (int j = 0; j < inCol->getNumberOfElements(); j++)
        {

            TrackerData* Data = dynamic_cast<TrackerData*>(inCol->getElementAt(j)) ;

            if (Data != 0)    // if cast didn't fail
                {

                    int cellid1 = Data->getCellID1();
                    int cellid0 = Data->getCellID0();
                    FloatVec ChargeVec = Data->getChargeValues();

                    for (FloatVec::iterator f = ChargeVec.begin(); f != ChargeVec.end(); f++)
                        {
                            // loop over the pixel values and insert into the input map
                            inMap[ TimePixPixelMode::PixelKey(cellid1, cellid0)] = *f;


                            // input histogram pixel index for column (IC) and for row (IR)
                            int IC = cellid0 % _nRow ;
                            int IR = static_cast<int>((cellid0 - IC) / _nColumn);

                            // fill histogram of input pixel map ( pixelman notation: 0<-->1 )
                            switch (cellid1)
                                {
                                case 1:
                                    _inputMapHisto0->fill(IC, IR, *f) ;     //  chip 0
                                    break;
                                case 0:
                                    _inputMapHisto1->fill(IC, IR, *f) ;     //  chip 1
                                    break;
                                case 2:
                                    _inputMapHisto2->fill(IC, IR, *f) ;     //  chip 2
                                    break;
                                case 3:
                                    _inputMapHisto3->fill(IC, IR, *f) ;     //  chip 3
                                    break;
                                default:
                                    //streamlog_out(WARNING0) << "Could not create histogram as there exists no chipID = " << cellid1 << std::endl;
                                    break;
                                }

                            cellid0++;
                        }
                }
        }

}


float TimePixCalibrationProcessor::getParameter(int chipID, int param)
{


    //  get calibration parameters from the collection

    // calibration parameter value
    float cT = 0.;

    // match parameter to the correct chip
    param = _nParameters * chipID + param ;
    // get element from calibration parameters collection
    LCGenericObject* CP = dynamic_cast<LCGenericObject*>(_vCalib->getElementAt(param));
    if (CP != 0)   // if cast didn't fail
        {

            // check if parameter belongs to chipID
            if (chipID != CP->getIntVal(0))   streamlog_out(WARNING4) << "Wrong chipID for parameter " << param << " at chip" << chipID << std::endl;

            // get calibration parameter
            cT  = CP->getFloatVal(0) ;
        }

    return cT ;
}


float TimePixCalibrationProcessor::calibrateTimePixel(int chipID, int pixelID, float pixelValue)
{

    // need to compensate for TimeWalk (Fit and Formula) and DistributionDeleay (map)
    // TimeWalk[clock cycles] = m*PulseHeight [mV]   +   a / (PulseHeight [mV] +b)
    //           = [ m*(ChargeValue[electrons] / 50.0)   +  (50.0*  a) / (ChargeValue[electrons] +b) ]
    // factor 50 found in thesis X.Llopart...

    // Time-Walk calibration parameters
    float timeWalk_a = getParameter(chipID, 4);
    float timeWalk_b = getParameter(chipID, 5);
    // float timeWalk_C = getParameter(  chipID, 6);     // this one is 0 by default
    float timeWalk_m = getParameter(chipID, 7);
    float timeWalkOffset = 0;

    // get calibration value                    // I guess we agreed to ignore the possible existence of the DistributionDeleay therefore the input map should be empty!
    int distDelay = ((map->find(TimePixPixelMode::PixelKey(chipID, pixelID)))->second)->getPixelCalibration() ;

    // Time-Walk correction needs a charge information, therefore interpolate the Pixel Charge from neighbors
    //  interpolation returns value in Clock Cyces - converting to Charge in electrons as requested by formula
    float interpolatedPixelCharge = interpolatePixelCharge(chipID, pixelID);
    if (interpolatedPixelCharge > 0)
        {
            interpolatedPixelCharge = calibrateChargePixel(chipID, interpolatePixelCharge(chipID, pixelID));

            // fit Time-Walk equation: Y=a/(X+b) ,with Y[Time] & X[electrons]
            timeWalkOffset = timeWalk_m * (interpolatedPixelCharge / 50)  + (timeWalk_a) / (timeWalk_b + (interpolatedPixelCharge / 50)) ;
        };
    streamlog_out(MESSAGE2) << "pixel " << pixelID / 256 << "," << pixelID % 256 << " was not TimeWalk corrected, not enough pixels available for charge interpolation." << std::endl;
    // return the corrected pixel value (Clock distribution and Time-Walk)
    return (pixelValue - distDelay + timeWalkOffset);

}


float TimePixCalibrationProcessor::calibrateChargePixel(int chipID, float pixelValue)
{

    // measured and fitted is ToT[clock cycles] in dependency of TestPuseHeight [mV]
    // fit function is:         ToT[clock cycles]       = a*TestPuseHeight [mV]  + b - ( c / (TestPuseHeight [mV] -t) )
    // isolating TestPuseHeight gives:  TestPuseHeight [mV] = 1/(2*a) * (ToT[clock cycles] - (b - a*t) + SQRT( (ToT[clock cycles] - (b - a*t))^2 + 4*(a*b*t + a*c - a*t*ToT[clock cycles]) ) )
    // (wow - that looks nasty)
    // we are not interested in the corresponding testpulse, but in the charge in elctrons - conversion is again: elcectrons = 50 * testpulse in mV
    // so we have:          Charge[electrons]       = 50 / (2*a) * (ToT[clock cycles] - (b - a*t) + SQRT( (ToT[clock cycles] - (b - a*t))^2 + 4*(a*b*t + a*c - a*t*ToT[clock cycles]) ) )
    // (still wow)

    // ToT & THL calibration parameters
    float charge_a = getParameter(chipID, 0);
    float charge_b = getParameter(chipID, 1);
    float charge_c = getParameter(chipID, 2);
    float charge_t = getParameter(chipID, 3);
    //float _cTHL1 = getParameter(  chipID, 4);         // I think this measurement was bullshit, since the offset has already been taken into account in prev. calibration
    //float _cTHL2 = getParameter(  chipID, 5);

    // creating some abbreviations:
    float p = pixelValue - (charge_b - charge_a * charge_t);                        // 1.sum term
    float q = 4 * (charge_a * charge_b * charge_t  +  charge_a * charge_c  -  charge_a * charge_t * pixelValue);  // 2. term of sqrt - neither is exactly the p or q from pq formula

    float charge_inElectrons = (50 / (2 * charge_a))  * (p + sqrt(p * p + q));    // well that doesn't look nasty at all does it?

    return charge_inElectrons ;
}


float TimePixCalibrationProcessor::interpolatePixelCharge(int chipID, int pixelID)
{

    /* For Time-Walk correction need to interpolate pixels in TIME mode with surrounding pixels in TOT mode so that will get Charge value for those pixels.
     * To accomplish this the interpolation function from TimePixPixelInterpolationProcessor is used.
     */

    // check if pixel is in TIME mode
    int mode = ((map->find(TimePixPixelMode::PixelKey(chipID, pixelID)))->second)->getPixelMode();
    if (mode != TimePixPixelMode::TIMEPIX) streamlog_out(WARNING0) << "Warning of TimePixCalibrationProcessor::interpolatePixelCharge: interpolation pixel NOT in Time mode !" << std::endl;

    std::vector<int> NeighborVec;

    //fill with the PixelIDs of the surrounding pixel:
    NeighborVec.push_back(pixelID - _nColumn - 1);    // |-----------|
    NeighborVec.push_back(pixelID - _nColumn);        // | 0 | 1 | 2 |
    NeighborVec.push_back(pixelID - _nColumn + 1);    // |-----------|
    NeighborVec.push_back(pixelID - 1);               // | 3 | x | 4 |
    NeighborVec.push_back(pixelID + 1);               // |-----------|
    NeighborVec.push_back(pixelID + _nColumn - 1);    // | 5 | 6 | 7 |
    NeighborVec.push_back(pixelID + _nColumn);        // |-----------|
    NeighborVec.push_back(pixelID + _nColumn + 1);    //

    //set all pixelIDs to -1 that can not be used to interpolate
    for (unsigned int n = 0; n < NeighborVec.size(); n++)
        {

            if (n == 0 || n == 3 || n == 5)
                {

                    //the pixel lies on the left edge of the chip
                    //the left neighbors therefore are on the other side of the chip
                    //and should not be used
                    if (NeighborVec[n] % _nColumn == _nColumn - 1)
                        NeighborVec[n] = -1;
                }

            if (n == 2 || n == 4 || n == 7)
                {

                    //the pixel lies on the right edge of the chip
                    //the right neighbors therefore are on the other side of the chip
                    //and should not be used
                    if (NeighborVec[n] % _nColumn == 0)
                        NeighborVec[n] = -1;
                }

            //these pixels are not on the chip anymore
            if (NeighborVec[n] > (_nColumn*_nRow - 1) || NeighborVec[n] < 0)
                NeighborVec[n] = -1;

        }  // for ( int n = 0; n < NeighborVec.size(); n++ ){

    // calculate mean of the counting surrounding pixels
    int noPixel = 0;
    float sumADC = 0;
    // int maxVal = 0;

    for (std::vector<int>::iterator iN = NeighborVec.begin(); iN < NeighborVec.end(); iN++)
        {
            if (*iN > 0)
                {

                    // get pixel mode & status
                    int mode2 = ((map->find(TimePixPixelMode::PixelKey(chipID, *iN)))->second)->getPixelMode();
                    int status2 = ((map->find(TimePixPixelMode::PixelKey(chipID, *iN)))->second)->getPixelStatus();


                    if (mode2 == TimePixPixelMode::TOT && status2 == TimePixPixelMode::INTACT)
                        {
                            //pixel can be used to interpolate

                            if (inMap[ TimePixPixelMode::PixelKey(chipID, *iN)]  > 0)
                                {
                                    sumADC += inMap[ TimePixPixelMode::PixelKey(chipID, *iN)];
                                    noPixel++;
                                    // if ( inMap[ TimePixPixelMode::PixelKey( chipID, *iN ) ] > maxVal)
                                    //   maxVal = inMap[ TimePixPixelMode::PixelKey( chipID, *iN ) ];
                                };
                        }

                } // if ( *iN > 0 )

        } //end for ( std::vector<int>::iterator iN = NeighborVec.begin();

    float interpolatedValue;

    //set mean of the counting surrounding pixels as new charge value
    if (noPixel >= _minNoPixelToInterpolate)
        {
            interpolatedValue = sumADC / static_cast<float>(noPixel) ;
            //if (noPixel < 3 ) {
//  interpolatedValue = maxVal;
//      };

        }


    //if we cannot calculate interpolated charge value we must keep the old one!
    else
        {
            streamlog_out(DEBUG4) << "pixel could not be interpolated, only " << noPixel << " available for interpolation." << std::endl;
            interpolatedValue =  0 ;
        }

    return interpolatedValue;

}


void TimePixCalibrationProcessor::end()
{
    //   nothing done here
}


}     // namespace marlintpc


