#include "TimePixMapHandlerProcessor.h"
#include "TimePixPixelMode.h"
#include "TimePixMapListener.h"
#include "TimePixHighTOTAnalyserProcessor.h"

// C++
#include <iostream>
#include <queue>
#include <string>
#include <sstream>
#include <cmath>
#include <vector>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/TrackerDataImpl.h>
#include "UTIL/LCTOOLS.h"

// Aida
#ifdef MARLIN_USE_AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IPlotter.h>
#include <AIDA/IPlotterFactory.h>
#include <RAIDA/RAIDAUtil.h>
#include <RAIDA/IBrushStyleROOT.h>
#include <RAIDA/IPlotterFactoryROOT.h>
#include <AIDA/IMarkerStyle.h>


//marlin
#include <marlin/ConditionsProcessor.h>


//Root
#include "TROOT.h"
#include "TCanvas.h"
#include "TFile.h"
#include "TH2F.h"
#include "THStack.h"
#include "TLine.h"

#endif

using std::string;
using namespace tpcconddata;
using namespace lcio ;
using namespace marlin ;


namespace marlintpc
{

tpcconddata::TimePixMapListener* TimePixMapHandlerProcessor::_timePixMapListener = 0;
TimePixHighTOTAnalyserProcessor* TimePixMapHandlerProcessor::_HighTOTProcessor = 0;
TimePixMapHandlerProcessor* TimePixMapHandlerProcessor::_MapHandlerProcessor = 0;

TimePixMapHandlerProcessor aTimePixMapHandlerProcessor ;


TimePixMapHandlerProcessor::TimePixMapHandlerProcessor() :
        Processor("TimePixMapHandlerProcessor")
{

    /* the constructor calls first the super constructor which is the name
     * of the processor
     * this must by the same as the class name.
     */

    /* description of the processor which will be displayed in the
     * steering file autometicly generated by
     * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
     */
    _description = "TimePixMapHandlerProcessor further seperates the clusters found in the TimePixClusterFinderProcessor by projecting them onto their own axis" ;


    /* register steering parameters: name, description, class-variable, default value
     * the type will by definded by the type of the default value
     * string, double, float and int are possible
     * plaese note to get a float you must cast in C++: (float) 1.3
     */

    registerInputCollection(LCIO::LCGENERICOBJECT,
                            "ModeMapCollectionName" ,
                            "Name of the mode map collection"  ,
                            _modeMapColName ,
                            std::string("TimePixModeMap")) ;

    registerInputCollection(LCIO::LCGENERICOBJECT,
                            "StatusMapCollectionName" ,
                            "Name of the status map collection"  ,
                            _statusMapColName ,
                            std::string("TimePixStatusMap")) ;

    registerProcessorParameter("SetOutputTransient" ,
                               "if not 0 the output collection is set transient"  ,
                               _outputIsTransient,
                               0);

    _MapHandlerProcessor = this;

}


void TimePixMapHandlerProcessor::init()
{

    // usually a good idea to
    printParameters() ;

    if (_timePixMapListener == 0)
        _timePixMapListener = new TimePixMapListener();
    else
        std::cout << "Warning: TimePixMapHandlerProcessor has already been initialised!"
                  << std::endl;

    // register it for automatic update
    ConditionsProcessor::registerChangeListener(_timePixMapListener->getStatusMapListener(),
                                                _statusMapColName);
    ConditionsProcessor::registerChangeListener(_timePixMapListener->getModeMapListener(),
                                                _modeMapColName);


}


void TimePixMapHandlerProcessor::processRunHeader(lcio::LCRunHeader* run)
{

    run->parameters().setValue(_processorName + "_revision", "$Rev: 2624 $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
                }
        }



}

void TimePixMapHandlerProcessor::processEvent(lcio::LCEvent * evt)
{

    streamlog_out(MESSAGE1) << "Event " << evt->getEventNumber() << std::endl;


}


void TimePixMapHandlerProcessor::check(lcio::LCEvent * evt)
{

    streamlog_out(DEBUG1) << "check( evt " << evt->getEventNumber() << ")" << std::endl;


}


void TimePixMapHandlerProcessor::end()
{



    //std::cout<<"end()"<<std::endl;
    /*
    streamlog_out(MESSAGE4) << "TimePixMapHandlerProcessor::end()  " << name()
      << " processed " << evt->getEventNumber() << " events in " << _nRun << " runs "
      << std::endl ;
    */
}

TimePixMap * TimePixMapHandlerProcessor::getTimePixMap()
{
    if (!_timePixMapListener)
        throw Exception("TimePixMapHandlerProcessor  is not properly initialized -"
                        " you need to have it in your execute section before"
                        " the processor which uses it!") ;

    if (_HighTOTProcessor != 0)
        return _HighTOTProcessor->getUpdatedTimePixMap();

    else
        return _timePixMapListener->getMap();

}

TimePixMap * TimePixMapHandlerProcessor::getUnmodifiedTimePixMap()
{
    if (!_timePixMapListener)
        throw Exception("TimePixMapHandlerProcessor  is not properly initialized -"
                        " you need to have it in your execute section before"
                        " the processor which uses it!") ;

    return _timePixMapListener->getMap();

}

TimePixMapHandlerProcessor* TimePixMapHandlerProcessor::registerHighTOTProcessor
(TimePixHighTOTAnalyserProcessor* htp)
{

    if (_HighTOTProcessor    == 0 &&
            _MapHandlerProcessor != 0) _HighTOTProcessor = htp;

    else throw Exception("TimePixMapHandlerProcessor::registerHighTOTProcessor: "
                             "TimePixHighTOTProcessor has already been registerd -"
                             "only register it once!");

    return _MapHandlerProcessor;

}










}
