#include "TimePixRemoveMaskedPixelsProcessor.h"
#include "TimePixMapHandlerProcessor.h"
#include "TimePixPixelMode.h"

// C++
#include <iostream>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerRawDataImpl.h>
#include "UTIL/LCTOOLS.h"
#include <IMPL/LCFlagImpl.h>

//Marlin
#include <marlin/Exceptions.h>

using namespace lcio ;
using namespace marlin ;
using namespace tpcconddata;

namespace marlintpc{


TimePixRemoveMaskedPixelsProcessor aTimePixRemoveMaskedPixelsProcessor ;


TimePixRemoveMaskedPixelsProcessor::TimePixRemoveMaskedPixelsProcessor() :
			Processor("TimePixRemoveMaskedPixelsProcessor"),   isFirstCollection(false)
{
	/* the constructor call first the super constructor which the name 
	 * of the processor
	 * this must by the same as the class name.
	 */

	/* description of the processor which will be displayed in the
	 * steering file autometicly generated by 
	 * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
	 */
	_description = "TimePixRemoveMaskedPixelsProcessor converts TrackerRawData of TimePix into zerosuppressed TrackerRawData by only keeping ADC values above a certain threshold" ;


	/* register steering parameters: name, description, class-variable, default value
	 * the type will by definded by the type of the default value
	 * string, double, float and int are possible
	 * plaese note to get a float you must cast in C++: (float) 1.3 
	 */
	registerInputCollection( LCIO::TRACKERRAWDATA,
			"InputCollectionName" ,
			"Name of the input collection"  ,
			_inputColName ,
			std::string("TimePixZeroSuppressedRawData") ) ;

	registerOutputCollection( LCIO::TRACKERRAWDATA,
			"OutputCollectionName" ,
			"Name of the output collection"  ,
			_outputColName ,
			std::string("TimePixCleanedRawData") ) ;

	registerProcessorParameter( "SetOutputTransient" ,
			"if not 0 the output collection is set transient"  ,
			_outputIsTransient,
			0);

	registerProcessorParameter( "SkipEmptyEvents" ,
			"Throw SkipEventException if TimePix data is empty"  ,
			_skipEmptyEvents,
			(bool) true);

	registerProcessorParameter( "RemoveNoisyPixels",
			"If true noisy pixels are removed from data",
			_removeNoisyPixels,
			(bool) true );

	registerProcessorParameter( "RemoveDeadPixels",
			"If true dead pixels are removed from data",
			_removeDeadPixels,
			(bool) true );

	_timePixMap = 0;

}


void TimePixRemoveMaskedPixelsProcessor::init() {

	// usually a good idea to
	printParameters() ;

	_timePixMap = TimePixMapHandlerProcessor::getTimePixMap();

	_nRun = 0 ;

}

void TimePixRemoveMaskedPixelsProcessor::processRunHeader( LCRunHeader* run) {


	run->parameters().setValue( _processorName + "_revision", "$Rev: 2739 $");

	for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
		if ( ! i->second->isOptional() || i->second->valueSet() ) {
			run->parameters().setValue( _processorName + "_" + i->second->name(),
					i->second->value() );
		}
	}

	_nRun++ ;

}

void TimePixRemoveMaskedPixelsProcessor::processEvent( LCEvent * evt ) {

	streamlog_out(MESSAGE1)<<"Event "<<evt->getEventNumber()<<std::endl;


	// get the input collection from the event
	LCCollection* inputCol = 0;

	try{

		inputCol = evt->getCollection( _inputColName ) ;

	} //end of try {

	catch(DataNotAvailableException &e){
		// if there is not input collection give a warning to the user
		streamlog_out(WARNING0)<<"ProcessEvent:"
				<<" input collection "
				<<_inputColName<<" cannot be found in event "
				<<evt->getEventNumber()<<std::endl;
		return;
	}


	// create the output collection with a string defining the TYPE (see lcio.h)
	LCCollectionVec* outputCol = new LCCollectionVec( LCIO::TRACKERRAWDATA );

	// set the transient flag as definded by the processor parameter
	outputCol->setTransient( (_outputIsTransient != 0) );

	// set the flags that cellID1 should be stored
	lcio::LCFlagImpl trkFlag(0) ;
	trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
	outputCol->setFlag( trkFlag.getFlag() );

	// get the number of elements in the input collection,
	// for TimePixRawData usually 1
	int nRawData = inputCol->getNumberOfElements();


	//and loop over it
	for(int iRawData=0; iRawData< nRawData ; iRawData++){

		/* getElement returns only a LCObject
		 * you have to cast it to the get the right type
		 * please use dynamic_cast which will return 0 if the object cannot be casted
		 */
		TrackerRawData* RawData = dynamic_cast<TrackerRawData*>( inputCol->getElementAt( iRawData ) ) ;

		if ( RawData != 0 ) {
			//only proceed if the cast was successful
			ShortVec RawDataADCVal = RawData->getADCValues();

			EVENT::ShortVec ZSADCValues;
			int cellid0 = RawData->getCellID0();

			for ( ShortVec::iterator iRawDataADCVal = RawDataADCVal.begin(); iRawDataADCVal < RawDataADCVal.end(); iRawDataADCVal++ ) {

				int status = ((_timePixMap->find( TimePixPixelMode::PixelKey( RawData->getCellID1(), cellid0 ) ))->second)->getPixelStatus();

				if ( !(status == TimePixPixelMode::DEAD && _removeDeadPixels) && !(status == TimePixPixelMode::NOISY && _removeNoisyPixels) ) {

					TrackerRawDataImpl* ZSRawData = new TrackerRawDataImpl();
					ZSRawData->setCellID1( RawData->getCellID1() );
					ZSRawData->setCellID0( cellid0 );

					int jCellID0 = cellid0;

					for ( ShortVec::iterator j = iRawDataADCVal; j < RawDataADCVal.end(); j++) {

						int jStatus;

						jStatus = ((_timePixMap->find( TimePixPixelMode::PixelKey( RawData->getCellID1(), jCellID0 ) ))->second)->getPixelStatus();

						if ( !(jStatus == TimePixPixelMode::DEAD && _removeDeadPixels) && !(jStatus == TimePixPixelMode::NOISY && _removeNoisyPixels) )
						{
							ZSRawData->adcValues().push_back( *j );
						}
						else
						{
							break;
						}

						jCellID0++;

					}

					cellid0 += ZSRawData->getADCValues().size()-1;
					iRawDataADCVal += ZSRawData->getADCValues().size()-1;
					if ( ZSRawData->getADCValues().size() > 0 )
					{
						outputCol->addElement( ZSRawData );
					}
					else
					{
						delete ZSRawData;
					}

				}

				cellid0++;

			}// end of for ( iRawDataADCVal = RawDataADCVal.begin(); ...

		} //end of if ( RawData != 0 ) {

	} //end of for(int iRawData=0; iRawData< nRawData ; iRawData++){


	// save the processor parameters in the first collection stored
	if (isFirstCollection){
		outputCol->parameters().setValue("InputCollectionName",_inputColName);
		// TO DO: does this make sence?
		outputCol->parameters().setValue("SetOutputTransient",_outputIsTransient);
		isFirstCollection = false;
	}

	/* after converting all raw data objects
	 * add the collection to the event and asign a name to it
	 * please note again: DO NOT DELETE the collection this will be done by LCIO
	 */
	if (outputCol->size() > 0)
		evt->addCollection( outputCol, _outputColName );
	else
	{	  // have to delete the collection if not added to the event
		// the outputCol deletes all the LCObjects it contains
		delete outputCol;
		if (_skipEmptyEvents && !isFirstEvent() )
			throw SkipEventException (this);
	}

}


void TimePixRemoveMaskedPixelsProcessor::check( LCEvent *  /*evt*/ )
{

}



void TimePixRemoveMaskedPixelsProcessor::end()
{

}

}
