#include "TimePixZeroSuppressionProcessor.h"

// C++
#include <iostream>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerRawDataImpl.h>
#include "UTIL/LCTOOLS.h"
#include <IMPL/LCFlagImpl.h>

//Marlin
#include <marlin/Exceptions.h>

// Aida
#ifdef MARLIN_USE_AIDA
//AIDA
#include <AIDA/IAnalysisFactory.h>
#include <AIDA/ITreeFactory.h>
#include <AIDA/ITree.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/ICloud1D.h>
#include <AIDA/IHistogram2D.h>
#endif


using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{


TimePixZeroSuppressionProcessor aTimePixZeroSuppressionProcessor ;


TimePixZeroSuppressionProcessor::TimePixZeroSuppressionProcessor() :
        Processor("TimePixZeroSuppressionProcessor"),   isFirstCollection(false)
{
    /* the constructor call first the super constructor which the name
     * of the processor
     * this must by the same as the class name.
     */

    /* description of the processor which will be displayed in the
     * steering file autometicly generated by
     * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
     */
    _description = "TimePixZeroSuppressionProcessor converts TrackerRawData of TimePix into zerosuppressed TrackerRawData by only keeping ADC values above a certain threshold" ;


    /* register steering parameters: name, description, class-variable, default value
     * the type will by definded by the type of the default value
     * string, double, float and int are possible
     * plaese note to get a float you must cast in C++: (float) 1.3
     */
    registerInputCollection(LCIO::TRACKERRAWDATA,
                            "InputCollectionName" ,
                            "Name of the input collection"  ,
                            _inputColName ,
                            std::string("TimePixRawData")) ;

    registerOutputCollection(LCIO::TRACKERRAWDATA,
                             "OutputCollectionName" ,
                             "Name of the output collection"  ,
                             _outputColName ,
                             std::string("TimePixZeroSuppressedRawData")) ;

    registerProcessorParameter("TOTThreshold" ,
                               "Threshold on the TOT value of a single pixel "  ,
                               _totThreshold ,
                               (double) 0.) ;

    registerProcessorParameter("NColumn" ,
                               "Number of columns of the chip "  ,
                               _nColumn ,
                               (int) 256) ;

    registerProcessorParameter("NRow" ,
                               "Number of rows of the chip "  ,
                               _nRow ,
                               (int) 256) ;

    registerProcessorParameter("SetOutputTransient" ,
                               "if not 0 the output collection is set transient"  ,
                               _outputIsTransient,
                               0);

    registerProcessorParameter("SkipEmptyEvents" ,
                               "Throw SkipEventException if TimePix data is empty"  ,
                               _skipEmptyEvents,
                               1);

    registerOptionalParameter("ChannelMappingCollectionName" ,
                                   "optional parameter: Name of the collection containing the ChannelMappings. If parameter is not given, the cellids of the input will be used.",
                                   _chMapCollection,
                                   tpcconddata::ADCChannelMapping::getDefaultColName()
                                  );

    registerOptionalParameter("SingleRawData" ,
                                   "create a TrackerRawData object for every pixel? default: false",
                                   _singleRawData,
                                   bool(false)
                                  );


}


void TimePixZeroSuppressionProcessor::init()
{

    // usually a good idea to
    printParameters() ;

    _nRun = 0 ;

}

void TimePixZeroSuppressionProcessor::processRunHeader(lcio::LCRunHeader* run)
{


    run->parameters().setValue(_processorName + "_revision", "$Rev: 2732 $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(),
                                               i->second->value());
                }
        }
    if( parameterSet( "ChannelMappingCollectionName" ) )
         {
        	 _ChannelMappingListener = new ChannelMappingListener(_chMapCollection);
         }
    _nRun++ ;

}

void TimePixZeroSuppressionProcessor::processEvent(lcio::LCEvent * evt)
{

    streamlog_out(MESSAGE1) << "Event " << evt->getEventNumber() << std::endl;


    // get the input collection from the event
    LCCollection* inputCol = 0;

    try
        {

            inputCol = evt->getCollection(_inputColName) ;

        } //end of try {

    catch (DataNotAvailableException)
        {
            // if there is not input collection give a warning to the user
            streamlog_out(WARNING0) << "ProcessEvent:"
            << " input collection "
            << _inputColName << " cannot be found in event "
            << evt->getEventNumber() << std::endl;
            return;
        }


    // create the output collection with a string defining the TYPE (see lcio.h)
    LCCollectionVec* outputCol = new LCCollectionVec(LCIO::TRACKERRAWDATA);

    // set the transient flag as definded by the processor parameter
    outputCol->setTransient((_outputIsTransient != 0));

    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1) ;
    outputCol->setFlag(trkFlag.getFlag());

    // get the number of elements in the input collection,
    // for TimePixRawData usually 1
    int nRawData = inputCol->getNumberOfElements();


    //and loop over it
    for (int iRawData = 0; iRawData < nRawData ; iRawData++)
        {

            /* getElement returns only a LCObject
             * you have to cast it to the get the right type
             * please use dynamic_cast which will return 0 if the object cannot be casted
             */
            TrackerRawData* RawData =
                dynamic_cast<TrackerRawData*>(inputCol->getElementAt(iRawData)) ;

            if (RawData != 0)
                {
                    //only proceed if the cast was successful
                    ShortVec RawDataADCVal = RawData->getADCValues();

                    int cellid0 = 0;
                    if ( (parameterSet( "SingleRawData" ) && !_singleRawData) || !parameterSet( "SingleRawData" ) )
                    {
                    	for (ShortVec::iterator iRawDataADCVal = RawDataADCVal.begin();
                            iRawDataADCVal < RawDataADCVal.end(); iRawDataADCVal++)
                        {

                            if (*iRawDataADCVal > _totThreshold)
                                {

                                    TrackerRawDataImpl* ZSRawData = new TrackerRawDataImpl();
                        			if(parameterSet("ChannelMappingCollectionName"))
                        			{
                        	            std::pair<int, int> gearIDpair = _ChannelMappingListener->getGearMapping(std::make_pair(cellid0,RawData->getCellID1()));
                        				ZSRawData->setCellID1(gearIDpair.second);
                        				ZSRawData->setCellID0(gearIDpair.first);
                        			}
                        			else
                        			{
                               			ZSRawData->setCellID1(RawData->getCellID1());  // normally 0 for TPCdata
                               			ZSRawData->setCellID0(cellid0);
                        			}

                                    for (ShortVec::iterator j = RawDataADCVal.begin() + cellid0;
                                            j < RawDataADCVal.end(); j++)
                                        {



                                            if (*j > _totThreshold) ZSRawData->adcValues().push_back(*j);
                                            else break;


                                        }



                                    //test if line break happend

                                    if (cellid0 % _nColumn  + ZSRawData->getADCValues().size() >
                                            static_cast<unsigned int>(_nColumn))
                                        {
                                            //line break happend
                                            streamlog_out(DEBUG4) << "cellid0%_nColumn  + ZSRawData->getADCValues().size() "
                                            << cellid0 % _nColumn  + ZSRawData->getADCValues().size() << " > "
                                            << static_cast<unsigned int>(_nColumn) << std::endl;


                                            streamlog_out(DEBUG4) << "cellid0 " << cellid0 << " % " << _nColumn << " = "
                                            << cellid0 % _nColumn << std::endl;

                                            ShortVec ZSRawDataVec = ZSRawData->getADCValues();

                                            streamlog_out(DEBUG4) << "ZSRawDataVec.size() = " << ZSRawDataVec.size() << std::endl;

                                            ZSRawDataVec.erase(ZSRawDataVec.begin()  + _nColumn - cellid0 % _nColumn,
                                                               ZSRawDataVec.end());

                                            streamlog_out(DEBUG4) << _nColumn - cellid0 % _nColumn << " =? "
                                            << ZSRawDataVec.size() << std::endl;

                                            ZSRawData->setADCValues(ZSRawDataVec);
                                            if (ZSRawData->getADCValues().size() > 0) outputCol->addElement(ZSRawData);
                                            else delete ZSRawData;


                                            iRawDataADCVal += _nColumn - cellid0 % _nColumn - 1;

                                            cellid0 += _nColumn - cellid0 % _nColumn - 1; //set cellid0 to end of column


                                        }
                                    else
                                        {

                                            cellid0 += ZSRawData->getADCValues().size() - 1;
                                            iRawDataADCVal += ZSRawData->getADCValues().size() - 1;
                                            if (ZSRawData->getADCValues().size() > 0) outputCol->addElement(ZSRawData);
                                            else delete ZSRawData;


                                        }

                                }//end of if ( *iRawDataADCVal > _totThreshold ) {

                            cellid0++;

                        }// end of for ( iRawDataADCVal = RawDataADCVal.begin(); ...
                    }
                    else// end of if ( (parameterSet( "SingleRawData" ) && !_singleRawData) || !parameterSet( "SingleRawData" ) )
                    {
                    	for (ShortVec::iterator iRawDataADCVal = RawDataADCVal.begin();
                    			iRawDataADCVal < RawDataADCVal.end(); iRawDataADCVal++)
                    	{

                    		if (*iRawDataADCVal > _totThreshold)
                    		{

                    			TrackerRawDataImpl* ZSRawData = new TrackerRawDataImpl();
                    			if(parameterSet("ChannelMappingCollectionName"))
                    			{
                    	            std::pair<int, int> gearIDpair = _ChannelMappingListener->getGearMapping(std::make_pair(cellid0,RawData->getCellID1()));
                    				streamlog_out(DEBUG)<<gearIDpair.second<<" "<<gearIDpair.first<<" "<<cellid0<<" "<< RawData->getCellID1()<<" "<<std::endl;
                    	            ZSRawData->setCellID1(gearIDpair.second);
                    				ZSRawData->setCellID0(gearIDpair.first);
                    			}
                    			else
                    			{
                           			ZSRawData->setCellID1(RawData->getCellID1());  // normally 0 for TPCdata
                           			ZSRawData->setCellID0(cellid0);
                    			}
                    			ZSRawData->adcValues().push_back(*iRawDataADCVal);
                    			outputCol->addElement(ZSRawData);

                    		}//end of if ( *iRawDataADCVal > _totThreshold ) {

                    		cellid0++;

                    	}// end of for ( iRawDataADCVal = RawDataADCVal.begin(); ...
                    }
                } //end of if ( RawData != 0 ) {

        } //end of for(int iRawData=0; iRawData< nRawData ; iRawData++){


    // save the processor parameters in the first collection stored
    if (isFirstCollection)
        {
            outputCol->parameters().setValue("InputCollectionName", _inputColName);
            // TO DO: does this make sence?
            outputCol->parameters().setValue("SetOutputTransient", _outputIsTransient);
            isFirstCollection = false;
        }

    /* after converting all raw data objects
     * add the collection to the event and asign a name to it
     * please note again: DO NOT DELETE the collection this will be done by LCIO
     */
    if (outputCol->size() > 0)
        evt->addCollection(outputCol, _outputColName);
    else
        {   // have to delete the collction if not added to the event
            // the outputCol deletes all the LCObjects it contains
            delete outputCol;
            if (_skipEmptyEvents && !isFirstEvent())
                throw SkipEventException(this);
        }

}


void TimePixZeroSuppressionProcessor::check(lcio::LCEvent * evt)
{

    LCCollection* TrackerRawDataCol;
    try
        {
            TrackerRawDataCol = evt->getCollection(_outputColName);
        }
    catch (lcio::DataNotAvailableException)
        {
            streamlog_out(WARNING0) << "Empty event: no zero suppredded TimePix data  in event "
            << evt->getEventNumber() << std::endl;
            return;
        }

    //LCTOOLS::dumpEvent( evt );

    std::stringstream sEvt;
    sEvt << evt->getEventNumber();

    std::string histoname = "clustersevt" + sEvt.str();


    AIDAProcessor::histogramFactory(this)->destroy(histoCluster);

    histoCluster = AIDAProcessor::histogramFactory(this)->
                   createHistogram2D(histoname.c_str(), histoname.c_str(),
                                     256, 0, 256, 256, 0, 256);

    for (int i = 0; i < TrackerRawDataCol->getNumberOfElements(); i++)
        {

            TrackerRawData* RawData =
                dynamic_cast<TrackerRawData*>(TrackerRawDataCol->getElementAt(i)) ;

            if (RawData != 0)
                {

                    int cellid0 = RawData->getCellID0();
                    int x = cellid0 % _nColumn;
                    int y = static_cast<int>((cellid0 - x) / _nColumn) + 1;

                    ShortVec ADCVec  = RawData->getADCValues();

                    for (ShortVec::iterator sv = ADCVec.begin(); sv != ADCVec.end(); sv++)
                        {

                            histoCluster->fill(x, y, *sv);
                            x++;
                            if (x == _nColumn)
                                {
                                    x = 0;
                                    y++;
                                }
                        }


                }


        }


    std::string histonameraw = "clustersevtraw" + sEvt.str();

    LCCollection* TrackerRawDataColRaw = evt->getCollection(_inputColName);

    AIDAProcessor::histogramFactory(this)->destroy(histoClusterRaw);

    histoClusterRaw = AIDAProcessor::histogramFactory(this)->
                      createHistogram2D(histonameraw.c_str(), histonameraw.c_str(),
                                        256, 0, 256, 256, 0, 256);

    for (int i = 0; i < TrackerRawDataColRaw->getNumberOfElements(); i++)
        {

            TrackerRawData* RawData =
                dynamic_cast<TrackerRawData*>(TrackerRawDataColRaw->getElementAt(i)) ;

            if (RawData != 0)
                {

                    int cellid0 = RawData->getCellID0();
                    int x = cellid0 % _nColumn;
                    int y = static_cast<int>((cellid0 - x) / _nColumn) + 1;

                    ShortVec ADCVec  = RawData->getADCValues();

                    for (ShortVec::iterator sv = ADCVec.begin(); sv != ADCVec.end(); sv++)
                        {

                            histoClusterRaw->fill(x, y, *sv);
                            x++;
                            if (x == _nColumn)
                                {
                                    x = 0;
                                    y++;
                                }
                        }


                }


        }

    // AIDAProcessor::tree(this)->commit();

}



void TimePixZeroSuppressionProcessor::end()
{
    /*
    streamlog_out(MESSAGE4) << "TimePixZeroSuppressionProcessor::end()  " << name()
      << " processed " << evt->getEventNumber()
      << " events in " << _nRun << " runs "
      << std::endl ;
    */

}

}
