#ifndef HELIKALTRACK_H
#define HELIKALTRACK_H

#include "KalTrackingTemplate.h"

//KalTest
#include <kaltest/TKalTrack.h>

namespace marlintpc
{

/** This is the helix track class, which inherits from KalTrackingTemplate and TKalTrack.
 *  Kalman filter algorithm comes from TKalTrack which is a class in KalTest package.
 *
 *  @author Bo Li, Tsinghua University
 *  @author Keisuke Fujii, KEK
 */

class KalTrackingHelix : public KalTrackingTemplate<TKalTrack>
{
    public:
        KalTrackingHelix(const gear::GearMgr &gearMgr, 
			 int minTrackHits, 
			 int maxSkipRows, 
			 double maxDeltaChi2,
			 int hitDistanceFactor,
			 int hitStartingMode,
			 double vDrift,
			 std::string modulesNotUsed);

        /** Get track type: helix
         */
        inline int getType()
        {
            return 1;
        }

   protected:

        /** Add a new site to track and update track parameters
         */
        bool doAddAndFilter(TKalTrack * track, ModuleRowHit hit) ;

	/** Create a track candidate by making a KalHit from the ModuleRowHit and
	 *  creating a TrackSite from it. Afterwards it copies the _kalTrack of this class
	 *  and adds and filters with the new site.
	 *  It also hands over the ownership of the new track, so don't forget to delete it.
	 *  Attention: The new track is not the owner of the sites, as they belong to the
	 *  original _kalTrack, except for the newly deleted site. Before deleting the 
	 *  candidate do not forget to delete the last added site on the track!
	 *
	 *  If the add and filter fails you get a 0-pointer;
	 */
	//	virtual TKalTrack * createTrackCandidate( ModuleRowHit hit );

        /** Search track starting from hit
         */
        virtual bool searchTrack(ModuleRowHit hit);

	/// Calculate the initial track state starting from three hits.
	virtual TMatrixD calculateStartHelix(ModuleRowHit const & hit1, 
					     ModuleRowHit const & hit2,
					     ModuleRowHit const & hit3);

	//        TVector3 getTangent();

        /** Print track parameters
         */
        virtual void printStateVector();

        static double toBaseRange(double phi);

        /** Fill the track state (incl. cov matrix) at the beginning (4th hit) of the track into the given LCIO track.
        *  Smoothes back to the 4th hit and returns that state.
        */
	virtual void setLCIOTrackParameters(IMPL::TrackImpl* trk) ;

	/** Get the deltaChi2 from the track candidate
	 */
	virtual double calculateDeltaChi2(TKalTrack * trackCandidate);
	virtual Intersection extrapolateIntoNextLayer(int targetDistance,
						      bool useFarHit);

	virtual const kaldet::GearTPCMeasLayer*  extrapolateIntoNextLayer(int targetDistance);

	virtual gear::Vector2D getNearestPointInNextLayer(ModuleRowHit firstHit);

	virtual double getTrackTangentAndLayerAngle(TKalTrack* track);
	virtual TKalTrackSite* createInitialSite(ModuleRowHit& mrh1, ModuleRowHit& mrh2, ModuleRowHit& mrh3);

 private:
    KalTrackingHelix(const KalTrackingHelix& orig);
	KalTrackingHelix & operator=(KalTrackingHelix const &);

	// a friend class for unit testing the protected member functions
	friend class KalTrackingHelixTester;
};


}
#endif  //HELIKALTRACK_H
