#ifndef ROWTRIPLETBASEDTRACKFINDERPROCESSOR_H
#define ROWTRIPLETBASEDTRACKFINDERPROCESSOR_H

#include <marlin/Processor.h>
#include <marlin/Global.h>
#include <lcio.h>
#include <string>

#include <gear/TPCModule.h>

// LCIO:
#include "EVENT/TrackerHit.h"

#include "TFile.h"
#include "TTree.h"
#include "TMath.h"
#include "TVectorD.h"
#include "TMatrixD.h"
#include "TMatrixDSym.h"

namespace marlintpc {
/// Track finder based on triplets of rows.

/**
 * Inspired by H1 pattern recognition by V. Blobel (Univ. Hamburg).
 *
 * Input are the (position, measurement direction and errors) of the TPC hits
 * and module parameters (offset and extend) from GEAR.
 *
 * In each module (with hits):
 *  - Triplets of hits in adjacent even or odd rows are constructed
 *  - Triplets matching in position and direction are connected to segments
 *  - Segments with compatible track parameters are merged
 *  - Unused hits are added to matching segments
 *
 * Segments from each module are combined with compatible segments from
 * the corresponding neighbouring modules to build track candidates. As
 * reference point for the track canidates the first hit is used (optional: PCA).
 *
 * The hits on the track candidates are ordered by row number, no multiple
 * hits in a row are allowed. All matching criteria use chi2 cuts based on the
 * hit measurement errors.
 * The calculation of the measurement (pad) direction in XY uses
 * "module.getLocalPadLayout().getCoordinateType()" to select polar or cartesian pad row geometry.
 *
 * \param  "InputHits":string           The name of the input collection of TPC hits (default: "TPCHits")
 * \param  "OutputTracks":string        The name of the output collection with the found tracks (default: "TripletTracks")
 * \param  "BFieldScaleFactor":double   \c Optional parameter, scales magnetic field (map), use 1.0 (default) for field ON or 0.0 for field OFF
 * \param  "TripletPreSelectionDistCut":double  \c Optional parameter, coarse cut on XY and Z residuals for triplet preselection (default 10.)
 * \param  "TripletDefinitionChi2Cut":double  \c Optional parameter, Chi2 cut for triplet definition on XY and Z residuals (default 20.)
 * \param  "TripletPositionMatchingChi2Cut":double   \c Optional parameter, Chi2 cut for triplet position matching in XY and Z (default 20.)
 * \param  "TripletDirectionMatchingChi2Cut":double  \c Optional parameter, Chi2 cut for triplet direction matching in XY and Z (default 20.)
 * \param  "SegmentMatchingChi2Cut":double    \c Optional parameter, Chi2/Ndf cut for segment matching (Ndf=4 for B off, 5 for B on) (default 30.)
 * \param  "UnusedHitMatchingChi2Cut":double  \c Optional parameter, Chi2 cut for matching unused hits in XY and Z (default 20.)
 * \param  "UnusedHitMaxGapCut:int  \c Optional parameter, maximal (row) gap for matching unused hits in XY and Z (default 4)
 * \param  "EncodedModuleID":bool  \c Optional parameter, flag for encoding of module ID in CellID0 (default true)
 * \param  "ReferencePointAtPca":bool \c Optional parameter,  use PCA as reference point, else 1. hit (default false)
 *
 * @author  C. Kleinwort (130729)
 * 
 * \par Credits:
 * The processor skeleton was generated by the script \c createProcessor.py
 *
 * \todo
 *   - definition of module neighbourhood from POLAR module extend ?
 *   - better resolve equivalent (indirectly matching) overlapping segments ?
 *
 * A detailed description is available as <a href="http://www-flc.desy.de/lcnotes/">lcnote</a> LC-TOOL-2014-004.
 * Changes with respect to this note:
 *   - Triplet measurement direction is average of outer hits instead of center hit (, already
 *     used for doublets).
 *   - Segment fit is iterated as weights depend on track direction (due to radial hit error).
 */
class RowTripletBasedTrackFinderProcessor: public marlin::Processor {
	public:

		virtual Processor* newProcessor() {
			return new RowTripletBasedTrackFinderProcessor;
		}

		RowTripletBasedTrackFinderProcessor();

		virtual void init();

		virtual void processRunHeader(EVENT::LCRunHeader* run);

		virtual void processEvent(EVENT::LCEvent* evt);

		virtual void check(EVENT::LCEvent* evt);

		virtual void end();

	protected:
		/* the place for protected and private member data and functions */
		std::string _inputColName; ///< Name of the input collection
		std::string _outputColName;  ///< Name of the output collection
		double _bfieldScaleFactor; ///< scale factor for magnetic field (default: 1.0)
		double _distCut; ///< Coarse cut on XY and Z residuals for triplet preselection
		double _trpCut; ///< Chi2 cut for triplet definition on XY and Z residuals
		double _posCut; ///< Chi2 cut for triplet position matching in XY and Z
		double _dirCut; ///< Chi2 cut for triplet direction matching in XY and Z
		double _segCut; ///< Chi2/Ndf cut for segment matching (Ndf=4 for B off, 5 for B on)
		double _unusedCut; ///< Chi2 cut for matching unused hits in XY and Z
		int _maxGap; ///< Cut for (row) distance to segment for matching unused hits in XY and Z
		bool _encodedModuleID; ///< Module ID is encoded in CellID0
		bool _refAtPCA; ///< Use Pca as reference point (else 1. hit)

	private:

		double _Bzc; ///< magnetic field strength (Bz*c)
		std::map<int, std::vector<int> > _modNeighbours; ///< neighbour modules
		double _Xcenter; ///< TPC center, X coordinate
		double _Ycenter; ///< TPC center, Y coordinate

		bool areNeighbourModules(gear::TPCModule*, gear::TPCModule*, bool);
};

/// Row based hit.
/**
 * Has identification and measurements with directions and errors.
 */
class rb_Hit {
	public:
		rb_Hit(const int iHit, const int mod, const int row, const EVENT::TrackerHit& aHit);
		void print() const;
		int getHitNum() const;
		int getMod() const;
		int getRow() const;
		double getX() const;
		double getY() const;
		double getZ() const;
		void getPos(double*) const;
		double getVarXY(const double = 0.) const;
		double getVarZ(const double = 0.) const;
		double getPhiMeas() const;
		bool getUsed() const;
		void setUsed(const double);
		double getCosBeta() const;
		double getDistXY(const double, const double, const double, const double) const;
		double getDistZ(const double) const;

	private:
		const int _hitNum; ///< input hit collection index
		const int _hitModule; ///< module number
		const int _hitRow; ///< row number
		const double _hitX; ///< X position
		const double _hitY; ///< Y position
		const double _hitZ; ///< Z position
		const double _hitPhiMeas; ///< measurement direction in XY (tangential to row)
		const double _hitVarXY; ///< variance for XY measurement
		const double _hitVarR; ///< variance in radial direction
		const double _hitVarZ; ///< variance for Z measurement
		bool _used; ///< used flag
		double _cosb; ///< cos(beta)

		double _calcPhiMeas(const EVENT::TrackerHit& aHit);
		double _getVarXY(FloatVec);
		double _getVarR(FloatVec);
		double _getVarZ(FloatVec);

};

typedef std::vector<rb_Hit*> hitListType;
typedef std::map<int, hitListType> rowHitMapType;
typedef std::map<int, rowHitMapType> modHitMapType;

/// Row based doublet.
/**
 * Is interpolation from outer hits (of triplet candidate).
 * Has a position (with covariance) and (track and measurement) directions 
 * determined from first and last hit in XY and ZS.
 */
class rb_Doublet {
	public:
		rb_Doublet(rb_Hit*, rb_Hit*);
		bool match(rb_Hit*, const double, const double) const;
		rb_Hit* getHit(const int) const;
		void getParameters(double&, double&, double&, double&, double&, double&, double&, double&, double&, double&) const;

	private:
		rb_Hit* _hit[2]; ///< pointers to hits
		double _phiMeas; ///< measurement direction in XY (tangential to row)
		double _cosPhi; ///< cos(_hitPhiMeas)
		double _sinPhi; ///< sin(_hitPhiMeas)
		double _xav; ///< average X coordinate
		double _yav; ///< average Y coordinate
		double _zav; ///< average Z coordinate
		double _phi; ///< XY direction
		double _tanl; ///< slope in ZS
		double _length; ///< (arc) length
		double _cosb; ///< cos(beta)
		double _der2XY; ///< (dxy/dr)^2
		double _der2Z; ///< (z/dr)^2
		double _varXY; ///< variance in XY
		double _varZ; ///< variance in Z
};

/// Row based triplet.
/**
 * Is build from matching hits in even or odd rows and
 * has a position and a direction (determined from seeding doublet)
 * in XY and ZS with corresponding covariances.
 */
class rb_Triplet {
	public:
		rb_Triplet(rb_Doublet&, rb_Hit*);
		int getRow() const;
		double getX() const;
		double getY() const;
		double getZ() const;
		void getPos(double*) const;
		double getVarXYPos(const double = 0.) const;
		double getVarXYDir() const;
		double getVarZPos(const double = 0.) const;
		double getVarZDir() const;
		double getPhiMeas() const;
		double getPhi() const;
		double getTanl() const;
		void getPosCloseTo(const double, const double, double*, double&) const;
		rb_Hit* getHit(const int) const;
		bool match(rb_Triplet*, const double, const double) const;
		double getCosBeta() const;

	private:
		rb_Hit* _hit[3]; ///< pointers to hits
		double _phiMeas; ///< (average) measurement direction in XY
		double _posX; ///< X position
		double _posY; ///< Y position
		double _posZ; ///< Z position
		double _varXY[2]; ///< (diagonal of) XY covariance matrix (position, slope)
		double _phi; ///< XY direction
		double _tanl; ///< slope in ZS
		double _varZ[2]; ///< (diagonal of) ZS covariance matrix (position, slope)
		double _length; ///< length
		double _cosb; ///< cos(beta)
};

typedef std::vector<rb_Triplet*> trpListType;

/// Row based segment.
/**
 * Is build from (the unused hits) of matching triplets
 * (from equivalence classes) and fitted in XY and ZS
 * (or from other matching segments).
 */
class rb_Segment {
	public:
		rb_Segment(double, trpListType);
		rb_Segment(double, hitListType, unsigned int = 1);
		rb_Segment(std::vector<rb_Segment*> segments);
		double getBzc() const;
		int getFirstRow() const;
		int getLastRow() const;
		void getRefPoint(double*) const;
		int getNdf() const;
		double getChi2() const;
		void getPar(double *) const;
		const hitListType& getHitList() const;
		const TVectorD& getPar() const;
		const TMatrixDSym& getCov() const;
		bool match(rb_Segment*, const double, const int) const;
		bool match(rb_Hit*, const double) const;
		void match(hitListType&, hitListType&, const double) const;
		void getLCIOStateAtRefPoint(const double*, TVectorD&, TMatrixDSym&) const;
		void fillRowMap(std::map<int, int>&) const;
		int getId() const;
		void setId(int);
		void addHit(rb_Hit *);

	private:
		int _segId; ///< segment ID
		hitListType _hitList; ///< pointers to hits
		double _bzc; ///< magnetic field strength (Bz*c)
		double _refX; ///< X of reference point
		double _refY; ///< Y of reference point
		double _refZ; ///< Z of reference point
		int _npar; ///< number of parameters (5: helix, 4: line)
		int _ndf; ///< number of degrees of freedom of segment fit
		double _chi2; ///< chi2 from segment fit
		TVectorD _parameters; ///< parameter vector
		TMatrixDSym _covariance; ///< covariance matrix
		void calcRefPoint();
		void fitSegment(double);
};

typedef std::vector<rb_Segment*> segListType;
typedef std::map<int, segListType> modSegMapType;

/// Equivalence classes
/**
 * From (integer) indices of matching (list) objects.
 *
 * Adopted from (H1) Fortran version by V. Blobel
 */
class simpleEquiClasses {
	public:
		simpleEquiClasses();
		void addIndex(int);
		void addMatch(std::pair<int, int>);
		std::map<int, std::vector<int> > getClasses();

	private:
		std::map<int, int> _index; ///< index list
		std::vector<std::pair<int, int> > _matches;  ///< list of matches
};

} // namespace marlintpc
#endif // ROWTRIPLETBASEDTRACKFINDERPROCESSOR_H
