#ifndef SIMPLEHELIXGENERALBROKENLINEINTERFACEPROCESSOR_H
#define SIMPLEHELIXGENERALBROKENLINEINTERFACEPROCESSOR_H

#include "SimpleHelixTrackModel.h"
#include "GeneralBrokenLineInterfaceHelpers.h"

#include <marlin/Processor.h>
#include <marlin/Global.h>

#include <lcio.h>
#include <string>

// GBL:
#include "GblTrajectory.h"
#include "MilleBinary.h"

// LCIO:
#include <EVENT/Track.h>
#include "EVENT/TrackerHit.h"

#include "TFile.h"
#include "TTree.h"
#include "TMath.h"

namespace marlintpc {

typedef std::pair<int, double> indexArcPair;
typedef std::pair<double, double> arcVarPair;

/// Using GBL with a simple helix.
/**  Processor interface to simple helix (or line) fitting with General Broken Lines, written by Claus Kleinwort
 *
 *  @author Ch. Rosemann, DESY (straight line), C. Kleinwort, DESY (extention to helix)
 *
 * This processor interfaces the re-fitting functionality of the general broken line package by Claus Kleinwort.
 * For information on the General Broken Lines, please refer to the publication and/or the wiki pages:
 *   - C. Kleinwort, General Broken Lines as advanced track fitting method, NIM A, 673 (2012), 107-110, doi:10.1016/j.nima.2012.01.024
 *   - http://www.wiki.terascale.de/index.php/GeneralBrokenLines
 *
 * The standard behaviour is that a single, straightforward fit without any outlier rejection is performed.
 * For multiple turns, including some outlier rejection; extend the fit options in the according steering parameter.
 * The calculation of the measurement (pad) direction in XY uses
 * "module.getLocalPadLayout().getCoordinateType()" to select polar or cartesian pad row geometry.
 *
 * <h4> Inputs and requirements </h4>
 * \li Input to the processor are the found tracks with their associated hits.
 * \li The curvilinear system is used as local system at each hit.
 * \li The curvilinear track parameters are used as local parameters at each hit.
 * \li A form of hit cleaning should be performed before calling the fit -- before using the fit results.
 *
 * <h4> Output </h4>
 * \li Output are the fitted tracks with the best estimation of track parameters and their uncertainties.
 * \li The specific energy loss is estimated from the median (and the MAD) of the 1/sqrt(E_scaled) distribution.
 * \li \c Optional output is a Millepede-II binary file, which can be inspected
 *
 * \param  "InputSeedTracks":string     The name of the input collection of track candidates (default: "TPCSeedTracks")
 * \param  "OutputTracks":string        The name of the output collection with the fitted tracks(default: "TPCTracks")
 * \param  "WriteOutputToStorage":bool  To set whether the output should be written to storage
 * \param  "BFieldScaleFactor":double   \c Optional parameter, scales magnetic field (map), use 1.0 (default) for field ON or 0.0 for field OFF
 * \param  "FitOptions":string          \c Optional parameter, defines if multiple passes with "h" huber, "t" tukey or "c" cauchy downweighting should be performed
 * \param  "WriteMillepedeBinary":bool  \c Optional parameter, set to "true" if you want a Millepede binary file (default:"false")
 * \param  "MillepedeFilename":string   \c Optional parameter to set the name of the Millepede binary file (default:"")
 * \param  "MillePedeCalcMethod":int    \c Optional parameter to define set of (global) parameters (default:0 = module alignment)
 * \param  "MillePedeMinHits":int       \c Optional parameter to define the minimum number of hits on track for output to Millepede binary file (default: 42)
 * \param  "EncodedModuleID":bool       \c Optional parameter, flag for encoding of module ID in CellID0 (default true)
 * \param  "X0PerUnitLength":double     \c Optional parameter, X0 per unit length (TPC gas as homogeneous scatterer) (default: 0.)
 * \param  "ThickScatterer":bool        \c Optional parameter, selects thick (otherwise thin) scatterers for multiple scattering (default: false)
 * \param  "DefaultMomentum":double     \c Optional parameter, default momentum for multiple scattering (for Bfield off, GeV) (default: 10.) 
 *
 *
 * WORK IN PROGRESS: currently only works for modules with concentric pad rows (using center from GEAR)
 * \todo
 *        - implement additional output of pull distributions, residuals, track paramters for each measurement point
 *
 * A detailed description is available as <a href="http://www-flc.desy.de/lcnotes/">lcnote</a> LC-TOOL-2014-008.
 * Changes with respect to this note:
 *   - Additional parameter "MillePedeMinHits" for help class "trackerAlcaSelector"
 *   - 160222 CHK: Implement TPC gas as homogeneous scatterer (between first and last hit) modelled by thin or thick scatterers (per hit)
 */
class SimpleHelixGeneralBrokenLineInterfaceProcessor: public marlin::Processor {
	public:

		virtual Processor* newProcessor() {
			return new SimpleHelixGeneralBrokenLineInterfaceProcessor;
		}

		SimpleHelixGeneralBrokenLineInterfaceProcessor();

		virtual void init();

		virtual void processRunHeader(EVENT::LCRunHeader* run);

		virtual void processEvent(EVENT::LCEvent* evt);

		virtual void check(EVENT::LCEvent* evt);

		virtual void end();
	protected:
		std::string _inputCollectionName; ///< Name of the input collection -- track seeds
		std::string _outputTrackCollectionName; ///< Name of the output track collection
		double _bfieldScaleFactor; ///< scale factor for magnetic field (default: 1.0)
		bool _outputIsPersistent;  ///< whether the output is to be stored or not (default: true)
		std::string _fitOptions; ///< list of iterations with 'h' Huber, 't' Tukey or 'c' Cauchy down-weighting
		bool _writeMillepedeOut; ///< selects output to Millepede-II binary file
		int _milleCalcMethod; ///< defines which parameter should be calculated for Millepede-II binary file
		int _milleMinHits; ///< defines minimum number of hits on track for output to Millepede binary file
		std::string _fileNameMillepedeFile; ///< name of Millepede-II binary file
		bool _encodedModuleID; ///< Module ID is encoded in CellID0
		double _x0PerUnitLength; ///< X0 per unit length (TPC gas as homogeneous scatterer) (default: 0.)
		bool _thickScatterer; ///< selects thick (otherwise thin) scatterers for multiple scattering (default: false)
		double _defaultMomentum; ///< default momentum for multiple scattering (for Bfield off, GeV) (default: 10.)

	private:
		gbl::GblTrajectory* _trajectory; ///< GBL trajectory
		gbl::MilleBinary* _milleBinary; ///< Millepede-II binary file
		bool _curvature; ///< flag for curved track (helix, else straight line)
		std::vector<unsigned int> _theLabels; ///< Labels of (global) MP-II parameters
		TVectorD* _correctionVector; ///< correction vector from GBL fit (for track parameters)
		TMatrixDSym* _covarianceMatrix; ///< covariance matrix from GBL
		int _ndf; ///< number of degrees of freedom in GBL fit
		double _chisquare; ///< chi2 from GBL fit
		double _lostweight; ///< weight lost by down-weighting
		double _refPointS; ///< arc-length of reference point
		unsigned int _refPointIndex; ///< label of reference point
		double _Xcenter; ///< TPC center, X coordinate
		double _Ycenter; ///< TPC center, Y coordinate

		void calcLineSeed(const std::vector<TrackerHit*>, const double*, double &, double &, double &, double &, double &) const;
};

} // namespace marlintpc
#endif // SimpleHelixGENERALBROKENLINEINTERFACEPROCESSOR_H
