#ifndef TRACKMAKINGKALMANFILTERPROCESSOR_H
#define TRACKMAKINGKALMANFILTERPROCESSOR_H

#include <marlin/Processor.h>
#include <lcio.h>
#include <gear/TPCParameters.h>

//Aida
#include <AIDA/AIDA.h>
#include <AIDA/ITuple.h>

#include "KalTrackingBase.h"

namespace marlintpc
{
/** This is processor which uses KalTrackingBase class and
 *  its derived classes(i.e. KalTrackingStraight and KalTrackingHelix) for finding and fitting tracks.
 *  In those track classes,  Kalman filter algorithm and track hits are included.
 *
 *  @author Bo Li, Tsinghua University
 *  @author Keisuke Fujii, KEK
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection of TrackerHits
 *
 *  <h4>Output</h4>
 *  Collection of Tracks
 *
 *  @param InputTrackerHits Name of the Input TrackerHits collection
 *  @param OutputTracks Name of the output Tracks collection
 *  @param MinTrackHits Minimum number of hits on track (default: 15)
 *  @param MaxSkipRows Maximum number of subsequnetly missing hits (default: 5)
 *  @param MaxDeltaChi2 Maximum number of chi2 increment (default: 30.)
 *  @param TrackType Track type (default: 1, 0=straight line 1=helix)
 */

class TrackMakingKalmanFilterProcessor : public marlin::Processor
{

    public:
        TrackMakingKalmanFilterProcessor();

        ~TrackMakingKalmanFilterProcessor();

        Processor * newProcessor();

        /** Called at the begin of the job before anything is read.
         */
        void init();

        /** Called for every run.
         */
        void processRunHeader(lcio::LCRunHeader *run);

        /** Called for every event - the working horse.
         */
        virtual void processEvent(lcio::LCEvent *evt);

        virtual void check(lcio::LCEvent *evt);

        /** Called after data processing for clean up.
         */
        virtual void end();

    private:
        /** Name of the input TrackerHitsCollection
         */
        std::string _InputTrackerHitsCollectionName;

        /** Name of the outout TracksCollection
         */
        std::string _outputTracksCollectionName;

        /** The minimum hit number to discriminate a track
         */
        int _minTrackHits;

        /** The maximum step over which a track can skip for one time
         */
        int _maxSkipRows;

        /** Maximum number of chi2 increment 
         */
        double _maxDeltaChi2;

		int _hitDistanceFactor;
		
		int _startingHitMode;

	double _vDrift;

        /** Track type: Straight line(0) or helix(1)
         */
        int _trackType;

        /** The pointer to a TPCHitMap which arranges hits by module and row
         */
	//        TPCHitMap*                  _hitMap;

        /** The geometry pointer for TPC
         */
        const gear::TPCParameters*        _tpcParameters;

        /** Tuple for track parameters
         */
        AIDA::ITuple*               _trackTuple;

        /** The pointer to the tracking class
         */
	KalTrackingBase*  _kalTracking;

	std::string _modulesNotUsed;
};

}// namespace marlintoc

#endif // TRACKMAKINGKALMANFILTERPROCESSOR_H
