// marlin
#include "CutHitsOnTracksProcessor.h"

using namespace marlin;
using namespace lcio;

namespace marlintpc
{
 
CutHitsOnTracksProcessor aCutHitsOnTracksProcessor;
  
CutHitsOnTracksProcessor::CutHitsOnTracksProcessor() : Processor( "CutHitsOnTracksProcessor" )
{
  _description = "Reads Tracks and removes Hits on this Track according to the given Parameters";
    
  registerInputCollection(LCIO::TRACK, "InputTrackCandidates", 
                        "The the name of the input collection of track candidates (default: CompleteTPCTrackCandidates)",
			_input_track_candidates_collection_name,
                        std::string("CompleteTPCTrackCandidates"));
				 
  registerOutputCollection(LCIO::TRACK, "OutputSeedTracks", "The name of the collection the seed tracks be stored under (default: CutTPCSeedTracks)",
                        _output_seed_tracks_collection_name,  
                        std::string("CutTPCSeedTracks"));

  registerProcessorParameter( "lowXCut" , 
                        "minimal Position of Hits on X-Axis in mm",
			_lowXCut, 
                        float(0.5));

  registerProcessorParameter( "lowYCut" , 
                        "minimal Position of Hits in Y-Axis in mm",
			_lowYCut,  
                        float(0.0));
   
  registerProcessorParameter( "_highXCut" , 
                        "highest Position of Hits on X-Axis in mm",
			_highXCut, 
                        float(13.5));

  registerProcessorParameter( "_highYCut" , 
                        "highest Position of Hits in Y-Axis in mm",
			_highYCut, 
                        float(14.0));  
    
  registerProcessorParameter( "_minNrOfHitsOnTrack" , 
                        "If a Track has less than N Hits it is thrown away",
			_minNrOfHitsOnTrack, 
                        int( 5 ) );    
				  
  //registerProcessorParameter( "SetOutputTransient" , 
  //                      "if not 0 the output hits collection is set transient (default: 0)",
  //                      _outputIsTransient,
  //                      int(0));    
}

  
CutHitsOnTracksProcessor::~CutHitsOnTracksProcessor() {}
  
  
void CutHitsOnTracksProcessor::init()
{
   printParameters(); 
}  // end init
  
  
void CutHitsOnTracksProcessor::processRunHeader(lcio::LCRunHeader *run)
{
  run->parameters().setValue( _processorName + "_revision", "$Rev: 5472 $");
    
  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ )
  {
    if ( ! i->second->isOptional() || i->second->valueSet() )
    {
      run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
    }
  }   
}  // end processRunHeader
  
  
void CutHitsOnTracksProcessor::processEvent(lcio::LCEvent *evt)
{
  streamlog_out(DEBUG4) << "CutHitsOnTracksProcessor" <<std::endl;
    
  IMPL::LCCollectionVec* trackCandidates=0;
   
  try
  {
    // the tracker hit collection
    trackCandidates = dynamic_cast<IMPL::LCCollectionVec*>(evt->getCollection(
      _input_track_candidates_collection_name));
  }
  catch (DataNotAvailableException &) 
  {
    streamlog_out(MESSAGE1) << "No seed tracks collection \"" <<_input_track_candidates_collection_name
      <<"\" in event "<< evt->getEventNumber()<<std::endl;
    return;
  }
  catch (std::bad_cast)
  {
    streamlog_out(MESSAGE1) << "Could not convert collection \"" <<_input_track_candidates_collection_name
      <<"\" to LCCollectionVec, skipping event " << evt->getEventNumber()<<std::endl;
    return;
  }
    
  // now we hava a valid input track collection named trackCandidates
       
  IMPL::LCCollectionVec* outputTrackCollection = new IMPL::LCCollectionVec(EVENT::LCIO::TRACK);
   
  LCFlagImpl trkFlag(0) ;
  trkFlag.setBit( LCIO::TRBIT_HITS ) ;
  outputTrackCollection->setFlag( trkFlag.getFlag() );
    
  // now we hava a valid output track Collection named outputTrackCollection
    
   
  //loop all tracks
  for (IMPL::LCCollectionVec::iterator trackIter=trackCandidates->begin(); trackIter < trackCandidates->end() ; 
    trackIter++)
  {
    EVENT::Track *inputTrack = dynamic_cast<EVENT::Track *>(*trackIter);
      
    if (inputTrack == 0)
    {
      streamlog_out(WARNING2) << "bad_cast: Track candidates collection"
        << _input_track_candidates_collection_name << " does not contain EVENT::Tracks!"<< std::endl;
      throw DataNotAvailableException("No track candidates found.");
    }
      
    // now we have hands on a Track and we can start to apply the needed cuts
      
    EVENT::TrackerHitVec inputTrackerHits = inputTrack->getTrackerHits(); // these are the hits on the track
      
    IMPL::TrackImpl* outputTrack = new IMPL::TrackImpl();
    outputTrack->setTypeBit(LCIO::TRBIT_HITS);
      
    // loop over all hits on the track
    for (EVENT::TrackerHitVec::iterator hit = inputTrackerHits.begin(); hit < inputTrackerHits.end(); hit++)
    {
      // if hit is within range keep it
      if ( ( (*hit)->getPosition()[0] >= _lowXCut) && ( (*hit)->getPosition()[0] <= _highXCut) &&  
        ( (*hit)->getPosition()[1] >= _lowYCut) && ( (*hit)->getPosition()[1] <= _highYCut) )
      {
        outputTrack->addHit(*hit);
      }
      // else not ;)
    } // end loop over all hits on the track
      
    // copy all relevant parameters
    outputTrack->setOmega( inputTrack->getOmega() );
    outputTrack->setPhi( inputTrack->getPhi() );
    outputTrack->setD0( inputTrack->getD0() );
    outputTrack->setTanLambda( inputTrack->getTanLambda() );
    outputTrack->setZ0( inputTrack->getZ0() );
    outputTrack->setdEdx( inputTrack->getdEdx() );
    outputTrack->setChi2( inputTrack->getChi2() );
      
    float reference[3];
    reference[0] = inputTrack->getReferencePoint()[0];
    reference[1] = inputTrack->getReferencePoint()[1];
    reference[2] = inputTrack->getReferencePoint()[2];
    outputTrack->setReferencePoint( reference );
        
    // attach new track to output collection      
    if ( outputTrack->getTrackerHits().size() >= static_cast <unsigned int> (_minNrOfHitsOnTrack) )
      outputTrackCollection->push_back(outputTrack);
    else delete outputTrack;      
  } // end loop all tracks
    
  // now add the collection to the event:
  try
  {
    evt->addCollection(outputTrackCollection, _output_seed_tracks_collection_name);
  }
  catch (EVENT::EventException e)
  {
    std::cout << "CutHitsOnTracksProcessor::processEvent(). " << e.what()
      << "Collection " << _output_seed_tracks_collection_name 
      << " is being deleted." << std::endl;
      
    std::vector<EVENT::LCObject*>::iterator iter = outputTrackCollection->begin();
      
    for(; iter != outputTrackCollection->end(); iter++)
    {
      if(*iter)
      {
        delete (*iter);
	(*iter) = NULL;
      }
    }      
    outputTrackCollection->clear();
    delete outputTrackCollection;
  }
} // end processEvent
  
  
  
Processor* CutHitsOnTracksProcessor::newProcessor() 
{ 
  return new CutHitsOnTracksProcessor();
}
  
  
void CutHitsOnTracksProcessor::check(lcio::LCEvent * /*evt*/)
{
} // end check
  
  
void CutHitsOnTracksProcessor::end()
{  
} // end end
   
  
} // end namespace marlintpc
