#ifdef USE_TRACK_FITTER_KALMAN
#include "KalTrackingStraight.h"

#include <marlin/Global.h>
#include <gear/PadRowLayout2D.h>
#include <gearimpl/RectangularPadRowLayout.h>
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>
#include <gear/GEAR.h>
#include "TPCHitMap.h"
#include "IntersectionCalculator.h"

#include <kaldet/GearTPCMeasLayer.h>
#include <kaldet/GearTPCCylinderMeasLayer.h>

#include <streamlog/streamlog.h>

using namespace gear;
using namespace std;
using namespace kaldet;

namespace marlintpc
{

KalTrackingStraight::KalTrackingStraight(const gear::GearMgr &gearMgr, 
			    unsigned int minTrackHits, 
			    unsigned int maxSkipRows, 
			    double maxDeltaChi2,
				int hitDistanceFactor,
				int hitStartingMode,
			    double vDrift,
				std::string modulesNotUsed)
  : KalTrackingTemplate<StraightTrackParameters>
                              ( gearMgr, minTrackHits, maxSkipRows, maxDeltaChi2, 
								hitDistanceFactor, hitStartingMode, vDrift, modulesNotUsed)
{
  if (gearMgr.getTPCParameters().getCoordinateType() != PadRowLayout2D::CARTESIAN)
  {
    streamlog_out(ERROR) << "KalTrackingStraight only works with global cartesian TPC " 
			 << "coordinates." << std::endl;
    throw gear::NotImplementedException("KalTrackingStraight does not run with polar coordinates.");
  }
}

bool KalTrackingStraight::searchTrack(ModuleRowHit firstHit)
{
    streamlog_out(DEBUG3) << "---------------------------" << endl
    << "In SearchTrack()"            << endl
    << "---------------------------" << endl;

    bool foundTrack = false;

    // fixme: coordinate transformation to gear coordinates for polar TPC coordinates in gear
    const double *globalFirstPos = firstHit.getHitPosition();

    streamlog_out(DEBUG3) << "hit1_x=" << globalFirstPos[0] 
			  << " hit1_y=" << globalFirstPos[1] << endl;

    // Extrapolate the hit into the next pad row (not measurement layer, this might be the
    // neighbouring cylinder segment at the same radius).

    std::set< std::pair<int, int> > moduleRowsInDomain =  
      _domain->getModuleRows(firstHit.module, firstHit.row+1 );

	streamlog_out(MESSAGE4) << "domain of hit 2: ---- module: " << firstHit.module
		    << ",  row: " << firstHit.row + 1 << std::endl;	

	ModuleRowHit defaultHit;	
	ModuleRowHit secondHit = _tpcHitMap.getHitInDomain( moduleRowsInDomain, firstHit);
	if(secondHit==defaultHit)return false;
      
    //reset track
    reset();
	_hitsOnTrack.push_back(firstHit);
	const double *secondPos = secondHit.getHitPosition();

    streamlog_out(MESSAGE4) << "hit2_x=" << secondPos[0] 
			  << " hit2_y=" << secondPos[1] << endl;
      _hitsOnTrack.push_back(secondHit);  //Put the first hit into the list
      if ( !doAddAndFilter(this->_containerTrack, secondHit) ) // Calculate the first track state
      {
	streamlog_out(ERROR) << "KalTrackingStraight: Filtering the second hit should never fail."
			     <<std::endl;
	return false;
      }
      searchTrackHit();//it's recursive

      if (_hitsOnTrack.size() >= _minTrackHits)foundTrack = true;
    
    return foundTrack;
}

// the kalTrack is not needed in this implementation
bool KalTrackingStraight::doAddAndFilter(StraightTrackParameters* stp, ModuleRowHit hit) 
{
    ///////FIXME///////////
    const double SIGMA = 0.3;
    /** TODO: warning: unused variables:
    * const double N_SIGMA = 5;
    * const double SIGMA_TIME = 50;
    */
    const double MAX_TIME = 200;   //max time difference
    ///////FIXME///////////

    streamlog_out(DEBUG3) << "--------------------------" << endl
    << "In AddAndFilter()"          << endl
    << "--------------------------" << endl;
    if (_hitsOnTrack.empty())
    {
      streamlog_out(ERROR) << "KalTrackingStraight: cannot addAndFilter with no hit in the list "
			   << " to start with" << std::endl;
      return false;
    }

    if (_hitsOnTrack.size() == 2)//if the track has two hits, calculate the state vector
    {
      const double* hp1 = _hitsOnTrack.begin()->getHitPosition();
      const double* hp2 = _hitsOnTrack.rbegin()->getHitPosition();
      
      streamlog_out(DEBUG3) << "hit1_x=" << hp1[0] << " hit1_y=" << hp1[1] << endl;
      streamlog_out(DEBUG3) << "hit2_x=" << hp2[0] << " hit2_y=" << hp2[1] << endl;
      
      stp->_sv1[0][0] = (hp2[1] - hp1[1]) / (hp2[0] - hp1[0]);//slope
      stp->_sv1[1][0] = hp2[1] - stp->_sv1[0][0] * hp2[0];//intercept
      
      stp->_sv = stp->_sv1;
      
      streamlog_out(DEBUG3) << "Initial state vector: " 
			    << "slope=" << stp->_sv[0][0] 
			    << "  intercept="  << stp->_sv[1][0] << endl;

	  stp->_deltaChi2 = 0.;

      return true;//a hit is added,retrun true
    }
    else
      {
	    const double* hitPos1 = _hitsOnTrack.rbegin()->getHitPosition();
        const double* hitPos2 = hit.getHitPosition();

        double predHitPos = stp->_sv[0][0] * hitPos2[0] 
	      + stp->_sv[1][0];//predicted hit position
            double delta = fabs(hitPos2[1] - predHitPos);

            streamlog_out(DEBUG3) << "hit_x=" << hitPos2[0] << " hit_y=" << hitPos2[1];
            streamlog_out(DEBUG3) << "k=" << stp->_sv[0][0] << ", b="
			                        << stp->_sv[1][0] 
									<< ", prediction_y=" << predHitPos  
				  << " delta=" << delta << endl;
                    double timeDiff = fabs(hitPos2[2] - hitPos1[2]);//time difference
                    if (timeDiff > MAX_TIME)return false;

                    //update state vector
                    double z = hitPos2[1];
                    stp->_H[0][0] = stp->_H_t[0][0] = hitPos2[0];
                    stp->_K = stp->_P1 * stp->_H_t;
                    TMatrixD t = (stp->_H * stp->_P1 * stp->_H_t 
				  + stp->_R).Invert();

                    stp->_K[0][0] = stp->_K[0][0] * t[0][0];
                    stp->_K[1][0] = stp->_K[1][0] * t[0][0];

                    stp->_sv = stp->_sv1 + 
		      stp->_K * (z - stp->_H * stp->_sv1);
                    stp->_P = (stp->_M_unit - 
					  stp->_K * stp->_H) * stp->_P1;
                    stp->_sv1 = stp->_sv;
                    stp->_P1 = stp->_P;

                    stp->_deltaChi2 = pow(delta / SIGMA, 2);

					stp->_chi2 += stp->_deltaChi2;
                    streamlog_out(MESSAGE4) << "k=" << stp->_sv1[0][0] 
					  << "   b=" << stp->_sv1[1][0] << ", dchi2=" 
					  << stp->_deltaChi2 << endl;

                    return true;
        }
}

void KalTrackingStraight::printStateVector()
{
    streamlog_out(DEBUG4) << "slope=" << _containerTrack->_sv[0][0]
    << " intercept=" << _containerTrack->_sv[1][0]
    << endl;
}

double KalTrackingStraight::calculateDeltaChi2(StraightTrackParameters * trackCandidate)
{
  return trackCandidate->_deltaChi2;
}

void KalTrackingStraight::setLCIOTrackParameters(IMPL::TrackImpl* trk)
{
  //straight line parametrization: y = slope * x + intercept
  double m = _containerTrack->_sv[0][0];   //slope
  double y0 = _containerTrack->_sv[1][0];  //intercept

  trk->setD0( y0*cos(atan( m ) )) ; // d0 = y0 * cos(phi)
  trk->setPhi( atan( m ) ) ;  // phi = atan( m )
  trk->setOmega(0.); // it's a straight track
  trk->setZ0(0) ; // not in the algoritm 
  trk->setTanLambda(0); // not in the algoritm
  
  trk->subdetectorHitNumbers().push_back(_hitsOnTrack.size()) ;
  trk->subdetectorHitNumbers().push_back(_hitsOnTrack.size()) ;

  trk->setChi2( _containerTrack->_chi2 ) ;
  
  // the Jacobian matrix to transform the covariance matrix to helix parameters
  TMatrixD J(2,2);

  // phi = atan( m )
  // d0  = y0 * cos( phi ) = y0 * cos( atan( m ) )

  // d phi / d y0 = 0
  J(0,0)=0;
  
  // d phi / d m = 1/(1+m^2)
  J(0,1)=1 / (1+ m*m);

  // d d0 / d y0 = cos( atan( m ) )
  J(1,0) = cos( atan( m ) );

  //d d0 / d m = - y0 * m/ (1+m^2)^(3/2)
  J(1,1) = -y0*m/pow(1+m*m, 1.5);

  //the transposed J matrix
  TMatrixD J_t(J); // create a copy
  J_t.T(); // transpose it

  TMatrixD covHelix = J * _containerTrack->_P * J_t;

  EVENT::FloatVec cov(15)  ;
  cov[ 0] =   covHelix(0 , 0)   ;   //   d0,   d0
  
  cov[ 1] = - covHelix(1 , 0)   ;   //   phi0, d0
  cov[ 2] =   covHelix(1 , 1)   ;   //   phi0, phi

  cov[ 3] =  0;
  cov[ 4] =  0;
  cov[ 5] =  0;

  cov[ 6] =  0; 
  cov[ 7] =  0; 
  cov[ 8] =  0;
  cov[ 9] =  0; 

  cov[10] =  0; 
  cov[11] =  0; 
  cov[12] =  0;
  cov[13] =  0; 
  cov[14] =  0; 
    
  trk->setCovMatrix(cov) ;
  
  float const referencePoint[] = {0., 0., 0.};
  trk->setReferencePoint(referencePoint) ;
  trk->setNdf( _hitsOnTrack.size() -2 ) ;

}

  KalTrackingTemplate<StraightTrackParameters>::Intersection KalTrackingStraight::extrapolateIntoNextLayer(int targetDistance,	bool /*useFarHit*/)
  // A straight track cannot have a far hit, ignore the flag
{
  // the straight track can only propagate outwards, so just count up the Kalman layers
  if (targetDistance <= 0)
  {
    throw gear::Exception("straight track can only go outwards");
  }

  // the last hit on the track
  ModuleRowHit mrh = _hitsOnTrack.back();
  GearTPCMeasLayer const * thisMeasLayer = _kalDetector.GetMeasLayer(mrh.module, mrh.row);
  int thisMeasLayerIndex = thisMeasLayer->GetIndex(); // the index in the cradle

  int nextIndex = thisMeasLayerIndex;
  int distance = 0;

  TVMeasLayer const * nextMeasLayer=0;

 // skip "skip" layers plus all inactive ones
  do
  { 
    ++nextIndex;

    if ( nextIndex >= _detectorCradle.GetEntries() )
     {
         return Intersection( TVector3(), Intersection::OutOfDetector);
     }
     
     nextMeasLayer = dynamic_cast<TVMeasLayer const *> (_detectorCradle.At(nextIndex));
     if (nextMeasLayer==0)
     {
       throw std::bad_cast();
     }
     
     if (!nextMeasLayer->IsActive())
     {
       // An inactive layer. Just ignore it and continue with the next layer.
       // Do not increase the distance.
       continue;
     }

     ++distance;
   }
  while (distance <= targetDistance);

  // ok the active layer should be a GearTPCMeasLayer
  //     GearTPCMeasLayer const * gearTPCMeasLayer 
  //       =  dynamic_cast<GearTPCMeasLayer const *>(nextMeasLayer);

  // try to cast to a gearTPCClinderMeasLayer, 
  GearTPCCylinderMeasLayer const * cylinderMeasLayer 
    = dynamic_cast<GearTPCCylinderMeasLayer const *>(nextMeasLayer);
  if ( cylinderMeasLayer==0 )
  {
    //FIXME: Implementation for planar layer missing
    throw gear::NotImplementedException("Only cylindrical layers are implemented");
  }

  //calculate the intersection with the layer
  double m = _containerTrack->_sv[0][0]; // the slope
  double b    = _containerTrack->_sv[1][0]; // y axis intersept
  
  TVector3 cylinderCentre = cylinderMeasLayer->GetXc();
  
  // now calculate the x-intersection in local coordinates of the layer
  // Although rather straight forward I write down the transformation in case I have a 
  // bug in the formula:
  // coodinates in local cylinder coordinates (with prime):
  // the cylinder centre is at x0', y0'
  // x' = x - x0'
  // y' = y - y0'
  // The track equation is y = m*x + b   => y'+y0' = m(x'+x0') + b
  // soling this equation for y'=0 results in
  // x' = -x0' - (b-y0')/m
  double x_prime = -cylinderCentre[0] - (b - cylinderCentre[1]) / m;

  // Now use the intersection calculator (it runs in local measCylinder coordinates)
  std::pair< std::pair<double,double>, std::pair<double,double> >  intersections;
  try{
    intersections = IntersectionCalculator::circle_straightLine_xy( cylinderMeasLayer->GetR(), 
                                                                    m, 
                                                                    x_prime);
  }
  catch(IntersectionCalculator::NoIntersectionException &) 
  {
    // no intersection, return an invalid Intersection class
    return Intersection(TVector3(), Intersection::NoIntersection);
  }

  // transfer intersections to global coordinates
  TVector3 firstIntersection( intersections.first.first,  intersections.first.second, 0.);
  TVector3 secondIntersection( intersections.second.first,  intersections.second.second, 0.);

  // use TVector to conveniently calculate the distances
  TVector3 lastHit( mrh.getHitPosition()[0], mrh.getHitPosition()[1], 0.);
  
  // pick the nearest and return it (what about the z coordinate?)
  return Intersection ( (lastHit-firstIntersection).Mag2() < 
			(lastHit-secondIntersection).Mag2() ?
			firstIntersection : secondIntersection,
			Intersection::IntersectionOK );
  
}

const GearTPCMeasLayer*  KalTrackingStraight::extrapolateIntoNextLayer(int targetDistance)
{
    streamlog_out(DEBUG3) << "In extrapolateIntoNextLayer" << endl;

   	// the last hit on the track
    ModuleRowHit mrh = _hitsOnTrack.back();

    GearTPCMeasLayer const* thisMeasLayer = _kalDetector.GetMeasLayer(mrh.module, mrh.row);
    GearTPCMeasLayer const* nextMeasLayer = 0;
    
	int thisMeasLayerIndex = thisMeasLayer->GetIndex(); // the index in the cradle
    
    // Determine the layer index which is distance away, whithout counting inactive layers.
    // Do we have to count the index and the distance separately.
    int nextIndex = thisMeasLayerIndex;
    int currentDistance = 0;

    // incease/decrease the distance until we reach the target value

    if (targetDistance<=0)
    { 
		// we are already there and know that the layer is active
		//nextMeasLayer = thisMeasLayer;
		return 0;
    }
    else
    { 
		do //while (currentDistance != targetDistance)
		{  
				nextIndex++;
				
				if ( nextIndex >= _detectorCradle.GetEntries() )
				{   
					//if out of detector return null pointer.
					return 0;
				}   
			
			nextMeasLayer = dynamic_cast<GearTPCMeasLayer const*> (_detectorCradle.At(nextIndex));
   
			if (nextMeasLayer==0)
			{
				throw bad_cast();
			}
       
			// only count active layers for the distance
	
			if (!nextMeasLayer->IsActive())
			{
				// An inactive layer. Just ignore it and continue with the next layer.
				// Do not increase/decrease the current distance.
				continue;
			}

				++currentDistance;

			streamlog_out(DEBUG3) << "targetDistance: " << targetDistance 
								  << ", currentDistance: "<< currentDistance << endl;
		}while (currentDistance != targetDistance);
    }

    streamlog_out(DEBUG3) << "end of extrapolateIntoNextLayer()" << endl;

	return nextMeasLayer;
}


}//namespace
#endif
