#include "LinearTrackRegression.h"

#include <EVENT/LCIO.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>

#include <streamlog/streamlog.h>
#include <stdexcept>
#include <string>

namespace marlintpc
{

LinearTrackRegression::StraightLine LinearTrackRegression::doRegressionX(const EVENT::Track *track)
{
    // Do the actual linear regression

    StraightLine straightline;

    // Initialize the straight line parameters.
    // In case there is only one or no hits (which should not
    // be the case, but who knows...) these parameters are kept.
    // It's a straight line along the y axis
    // through the origin
    straightline.slope1 = 0;
    straightline.intersept1 = 0;
    straightline.slope2 = 0;
    straightline.intersept2 = 0;


    // These are the sums needed to calculate the
    // track parameters (see documentation in .h file)
    float sumX = 0;
    float sumY = 0;
    float sumXY = 0;
    float sumZ = 0;
    float sumYZ = 0;
    float sumYsqare = 0;  // = Sum (Y^2)

    // get the vector of hits from the track candidate
    EVENT::TrackerHitVec hitVec = track->getTrackerHits();

    // save the number of hits in a separate variable
    // makes the code nicer
    unsigned int n = hitVec.size();

    // some debug output
    streamlog_out(DEBUG0) << "hitVec.size() is " << hitVec.size() << std::endl;

    // Regression can only be performed if there are at least
    // two hits on the track. Otherwise we get a divide by zero
    if (n >= 2)
        {
            // debug:
            streamlog_out(DEBUG0) << "n is " << n << std::endl;

            // loop all hits and calculate the sums needed
            for (unsigned int i = 0; i < n; i++)
                {
                    // get the particular hit from the vector
                    EVENT::TrackerHit *hit = hitVec[i];

                    // check whether the hit pointer is valid
                    // If the algorithm is working this should always
                    // be the case. Otherwise throw an exception
                    // to simplify debugging.
                    if (hit == NULL)
                        {
                            streamlog_out(ERROR) << "something went wrong, got empty hit " << i << "from the vector" << std::endl;
                            throw std::runtime_error(std::string("TrackFinderRectangularProcessor::doRegression hit is NULL "));
                        }

                    // add the variables to the sums
                    sumX += hit->getPosition()[0];
                    sumY += hit->getPosition()[1];
                    sumXY += hit->getPosition()[0] * hit->getPosition()[1];
                    sumZ += hit->getPosition()[2];
                    sumYZ += hit->getPosition()[1] * hit->getPosition()[2];
                    sumYsqare += hit->getPosition()[1] * hit->getPosition()[1];
                }

            // Now that the sums are calculated we can compute the
            // track parameters
            straightline.slope1     = (sumX * sumY - n * sumXY) / (sumY * sumY - n * sumYsqare);
            straightline.intersept1 = (sumY * sumXY - sumYsqare * sumX) / (sumY * sumY - n * sumYsqare);
            straightline.slope2     = (sumZ * sumY - n * sumYZ) / (sumY * sumY - n * sumYsqare);
            straightline.intersept2 = (sumY * sumYZ - sumYsqare * sumZ) / (sumY * sumY - n * sumYsqare);

        }// if  n>=2

    return straightline;
}

LinearTrackRegression::StraightLine LinearTrackRegression::doRegressionY(const EVENT::Track *track)
{
    // Do the actual linear regression

    StraightLine straightline;

    // Initialize the straight line parameters.
    // In case there is only one or no hits (which should not
    // be the case, but who knows...) these parameters are kept.
    // It's a straight line along the y axis
    // through the origin
    straightline.slope1 = 0;
    straightline.intersept1 = 0;
    straightline.slope2 = 0;
    straightline.intersept2 = 0;


    // These are the sums needed to calculate the
    // track parameters (see documentation in .h file)
    float sumX = 0;
    float sumY = 0;
    float sumXY = 0;
    float sumZ = 0;
    float sumXZ = 0;
    float sumXsqare = 0;  // = Sum (X^2)

    // get the vector of hits from the track candidate
    EVENT::TrackerHitVec hitVec = track->getTrackerHits();

    // save the number of hits in a separate variable
    // makes the code nicer
    unsigned int n = hitVec.size();

    // some debug output
    streamlog_out(DEBUG0) << "hitVec.size() is " << hitVec.size() << std::endl;

    // Regression can only be performed if there are at least
    // two hits on the track. Otherwise we get a divide by zero
    if (n >= 2)
        {
            // debug:
            streamlog_out(DEBUG0) << "n is " << n << std::endl;

            // loop all hits and calculate the sums needed
            for (unsigned int i = 0; i < n; i++)
                {
                    // get the particular hit from the vector
                    EVENT::TrackerHit *hit = hitVec[i];

                    // check whether the hit pointer is valid
                    // If the algorithm is working this should always
                    // be the case. Otherwise throw an exception
                    // to simplify debugging.
                    if (hit == NULL)
                        {
                            streamlog_out(ERROR) << "something went wrong, got empty hit " << i << "from the vector" << std::endl;
                            throw std::runtime_error(std::string("TrackFinderRectangularProcessor::doRegression hit is NULL "));
                        }

                    // add the variables to the sums
                    sumX += hit->getPosition()[0];
                    sumY += hit->getPosition()[1];
                    sumXY += hit->getPosition()[0] * hit->getPosition()[1];
                    sumZ += hit->getPosition()[2];
                    sumXZ += hit->getPosition()[0] * hit->getPosition()[2];
                    sumXsqare += hit->getPosition()[0] * hit->getPosition()[0];
                }

            // Now that the sums are calculated we can compute the
            // track parameters
            straightline.slope1     = (sumX * sumY - n * sumXY) / (sumX * sumX - n * sumXsqare);
            straightline.intersept1 = (sumX * sumXY - sumXsqare * sumY) / (sumX * sumX - n * sumXsqare);
            straightline.slope2     = (sumZ * sumX - n * sumXZ) / (sumX * sumX - n * sumXsqare);
            straightline.intersept2 = (sumX * sumXZ - sumXsqare * sumZ) / (sumX * sumX - n * sumXsqare);

        }// if  n>=2

    return straightline;
}

}// namespace marlintpc
