#ifdef USE_TRACK_FITTER_KALMAN
#include "TPCHitMap.h"
#include "TPCHitDomain.h"
#include <EVENT/LCCollection.h>
#include <marlin/Global.h>
#include <gear/TPCModule.h>
#include <kaltest/TVMeasLayer.h>
#include <streamlog/streamlog.h>
#include <Exceptions.h> 

using namespace gear;
using namespace std;
using namespace EVENT;
using namespace marlin;
using namespace kaldet;

namespace marlintpc
{

TPCHitMap::TPCHitMap(gear::TPCParameters const * tpcParameters, 
		             GearTPCKalDetector const * det,
					 std::string modulesNotUsed) 
  : _tpcParameters(tpcParameters), _nHits(0), _kalDetector(det)
{
	_modulesNotUsedSet.clear();
	
    stringstream ssin(modulesNotUsed);

	int i = 0;
	ssin >> i;

	if(i>0)
	{
		_modulesNotUsedSet.insert(i);
		while (ssin.good()){ ssin >> i; _modulesNotUsedSet.insert(i); }
	}

}


TPCHitMap::~TPCHitMap()
{}

void TPCHitMap::removeHits(vector<ModuleRowHit>& hitVector)
{
    streamlog_out(DEBUG3) << "In RemoveHits()" << endl;
    for (vector<ModuleRowHit>::iterator mrhIter = hitVector.begin(); mrhIter != hitVector.end(); ++mrhIter)
    {
      removeHit(*mrhIter);
    }

    hitVector.clear();
}

void TPCHitMap::removeHit(ModuleRowHit mrHit)
{
    HitContainer::iterator moduleRowIter = _hitContainer.find( ModuleRow(mrHit.module, mrHit.row) );
    
	if (moduleRowIter==_hitContainer.end())
    {
      streamlog_out(ERROR) << "HitMap::removeHit(), unknown ModuleRowHit in hitContainer"
		                   << ", module: " << mrHit.module << ", row: " << mrHit.row << endl;

	  throw EVENT::Exception("Unknown ModuleRowHit");
    }

    set< MarlinTPCHit > & moduleRowsHits = moduleRowIter->second;

    streamlog_out(DEBUG3) << "before removing hit, the size of module " <<  mrHit.module 
		                  << " row " <<  mrHit.row << " is: " << moduleRowsHits.size() << endl;

	MarlinTPCHit hit(mrHit.tpcHit);

	streamlog_out(DEBUG3) << "a hit is removed from the map, module: " << mrHit.module 
		                  << ", row: " << mrHit.row << ", (x, y, z) = " 
						  << mrHit.getHitPosition()[0] << ", "
						  << mrHit.getHitPosition()[1] << ", "
						  << mrHit.getHitPosition()[2] << endl;

	moduleRowsHits.erase( hit );

    streamlog_out(DEBUG3) << "after removing hit, the size is " <<  moduleRowsHits.size() << endl;

    --_nHits;
     
    if (moduleRowsHits.empty())
    {
       int module = mrHit.module;
	   int row    = mrHit.row;

	   streamlog_out(DEBUG3) << "module " <<  mrHit.module << ", row " <<  mrHit.row 
		                    << " is empty, erase it." << endl;
	  
	   _hitContainer.erase( moduleRowIter );

	   int layerIndex = getLayerIndex(module, row);
	   _indexModuleRow[layerIndex].erase( ModuleRow(module, row) );

	   if(_indexModuleRow[layerIndex].size()==0)
	   {
		   _indexModuleRow.erase(layerIndex);
	   }
    }	
}


void TPCHitMap::clear()
{
	_hitContainer.clear();
	_indexModuleRow.clear();
	_nHits = 0;
}

bool TPCHitMap::isEmpty() const
{
    return _hitContainer.empty();
}


void TPCHitMap::fillEventHits(LCCollection const * inputHits)
{
    //read input hits, and put data into hitSet
    int hitNum = inputHits->getNumberOfElements();
    streamlog_out(DEBUG3) << "Hit number=" << hitNum << endl;

	clear();

    for (int iHit = 0; iHit < hitNum; iHit++)
    { 
		TrackerHit* inputHit = dynamic_cast<TrackerHit *>(inputHits->getElementAt(iHit));
        const double* pos = inputHit->getPosition();

        streamlog_out(DEBUG3) << "hit" << iHit << ": pos=(" << pos[0] << "," << pos[1] << "," << pos[2] << ")" << endl;

		//FIXME::
		//for beam test data, the constructed hit maybe negtive in z direction
		//it will cause error in KalTest
		if(pos[2]<0) 
		{
			streamlog_out(DEBUG3) << "TPCHitMap: Rejecting this hit, because of its negtive z component" << endl; 
			continue;
		}

        const TPCModule & tpcmodule = _tpcParameters->getNearestModule(pos[0], pos[1]);
        int moduleID = tpcmodule.getModuleID();

	    if(!tpcmodule.isInsideModule(pos[0], pos[1]))
	    {
			streamlog_out(DEBUG3) << "TPCHitMap: Rejecting hit at (" << pos[0] << "," << pos[1] << "," << pos[2] 
				                    << ") because it is outside module "<< moduleID << endl;
			continue;
	    }

		//check whether this module is used
		if(_modulesNotUsedSet.find(moduleID)!=_modulesNotUsedSet.end())
		{
			//if this module is not used in track reconstruction
			cout << "hit invalid" << endl;
			continue;
		}

        int pad = tpcmodule.getNearestPad(pos[0], pos[1]);
	    
		if(!tpcmodule.isInsidePad(pos[0], pos[1], pad))
	    {
	      streamlog_out(DEBUG3) << "TPCHitMap: Rejecting hit at (" << pos[0] << "," << pos[1] << "," << pos[2] 
		             		      << ") because it is outside any pad on module "<< moduleID << endl;
	      continue;
	    }

		int row = tpcmodule.getRowNumber(pad);
        
		//get layer index
		_hitContainer[ModuleRow(moduleID, row)].insert( MarlinTPCHit(inputHit, false) );

		int layerIndex = getLayerIndex(moduleID, row);
		_indexModuleRow[layerIndex].insert( ModuleRow(moduleID, row) );

		streamlog_out(DEBUG3) << "moduleID=" << moduleID << ", row=" << row 
			                  << ", layerIndex=" << layerIndex << endl;

		++_nHits;
    }
}

int TPCHitMap::getLayerIndex(int module, int row)
{
	if(_kalDetector)
	{ 
		return _kalDetector->GetMeasLayer(module, row)->GetIndex();
    }
	else
	{
		//FIXME:: if there is only one module, it is OK
	    return row;
	}
}

ModuleRowHit TPCHitMap::getFirstHit(int mode)
{
    if(mode)//get the first hit of track by layer index 
	{ 
	    //the contents in _indexModuleRow are arranged by the (layerindex, module, row)
		
		if( _indexModuleRow.begin()!=_indexModuleRow.end() )
		{
			ModuleRow moduleRow = *( (_indexModuleRow.begin()->second).begin() );
			set< MarlinTPCHit >& moduleRowsHitSet=_hitContainer.find(moduleRow)->second;

			return ModuleRowHit( moduleRow.first, moduleRow.second, moduleRowsHitSet.begin()->hit );
		}
		else
		{
			return ModuleRowHit();
		}
	}
	else//when mode is 0, get the first hit by z position
	{
		double zMax = 0.;
		MarlinTPCHit hitWithZmax;
		ModuleRow moduleRow;

		//get the first hit which has z maximum
		for(HitContainer::iterator moduleRowIter = _hitContainer.begin();
		moduleRowIter != _hitContainer.end(); ++moduleRowIter)
		{
			set< MarlinTPCHit >& hitSet = moduleRowIter->second;

			//the hit in the set has been sorted as z coordinate
			const MarlinTPCHit& aHit = *(hitSet.rbegin());
			double zOfHit = fabs(aHit.hit->getPosition()[2]);

			if(zOfHit>zMax)
			{
				zMax = zOfHit;
				hitWithZmax = aHit;
				moduleRow = moduleRowIter->first;
			}
		
		}

		return ModuleRowHit(moduleRow.first, moduleRow.second, hitWithZmax);
	}
}

list<ModuleRowHit>  TPCHitMap::getHitsInDomain(
			     set< ModuleRow > const & moduleRowsInDomain)
{
  //this function returns valid hits(which is not used in the current searching track) in the hit map

  // the return value
  list<ModuleRowHit> hitsInDomain;

  streamlog_out(DEBUG3) << "Get hits in a domain..." << endl;

  for(set< ModuleRow >::iterator moduleRowsIter = moduleRowsInDomain.begin();
      moduleRowsIter!=moduleRowsInDomain.end(); ++moduleRowsIter)
  {
    int module = moduleRowsIter->first;
    int row    = moduleRowsIter->second;

	streamlog_out(DEBUG3) << "module: " << module << ", row: " << row << endl;
    
    HitContainer::iterator containerIter = _hitContainer.find( ModuleRow(module, row) );
	if(containerIter==_hitContainer.end())continue;

    set< MarlinTPCHit >& hitSet = containerIter->second;
    
    for(HitSetIter hitSetIter = hitSet.begin(); hitSetIter != hitSet.end(); ++hitSetIter)
    { 
	  if( !hitSetIter->isUsed )
	  {
		streamlog_out(DEBUG3) << "this hit is not used before" << endl;
        hitsInDomain.push_back( ModuleRowHit(module, row, hitSetIter->hit) );
	  }
	  else
	  {
		streamlog_out(DEBUG3) << "this hit is used before" << endl;
	  }
    }
  }

  streamlog_out(DEBUG3) << "Found " << hitsInDomain.size() << " hits in this domain." << endl;

  return hitsInDomain;
}

ModuleRowHit TPCHitMap::getHitInDomain(set< ModuleRow > const & moduleRowsInDomain, 
										ModuleRowHit referenceHit)
{
  //this function also returns valid hits in the hit map
  //and the hits are listed by the distance to the reference hit

  //all hits in the domain
  list<ModuleRowHit> hitsInDomain = getHitsInDomain(moduleRowsInDomain);

  int domainSize = hitsInDomain.size();
  
  ModuleRowHit minDistanceHit; 
  
  //pick up the hit which the shortest disrance with the reference.
  for(int n = 0; n < domainSize; n++)
  {
	  double minDistance = numeric_limits<double>::max();
	  minDistanceHit = *hitsInDomain.begin();

	  for(list<ModuleRowHit>::iterator iter = hitsInDomain.begin(); 
			  iter != hitsInDomain.end(); iter++)
	  {
		  double distance = getHitsDistance(*iter, referenceHit); 

		  if( distance < minDistance)
		  {
			  minDistance = distance;
			  minDistanceHit = *iter;
		  }
	  }
  }

  return minDistanceHit;
}

size_t TPCHitMap::size() const
{
  return _nHits;
}

list< ModuleRow >  TPCHitMap::getActiveModuleRows()
{
  list< ModuleRow > activeModuleRowsList;

  for (HitContainer::iterator moduleRowIter = _hitContainer.begin();
       moduleRowIter != _hitContainer.end(); ++moduleRowIter)
  {
    activeModuleRowsList.push_back( moduleRowIter->first );
  }

  return activeModuleRowsList;
}

void TPCHitMap::setHitUsed(const ModuleRowHit& mrHit, bool used) 
{
    HitContainer::iterator moduleRowIter = _hitContainer.find( ModuleRow(mrHit.module, mrHit.row) );
    
	if (moduleRowIter==_hitContainer.end())
    {
      streamlog_out(ERROR) << "HitMap:: unknown ModuleRowHit in hitContainer"
		                   << ", module: " << mrHit.module << ", row: " << mrHit.row << endl;

	  throw EVENT::Exception("Unknown ModuleRowHit");
    }

    set< MarlinTPCHit > & moduleRowsHits = moduleRowIter->second;

    set< MarlinTPCHit >::iterator iter =
		   moduleRowsHits.find( MarlinTPCHit( mrHit.getTrackerHit() ) );

	if(iter==moduleRowsHits.end())
	{ 
		streamlog_out(ERROR) << "HitMap:: unknown pair< EVENT::TrackerHit*, bool> in hitContainer" << endl;
		throw EVENT::Exception("Unknown hit");
	}
	else
	{ 
		const_cast<MarlinTPCHit &>(*iter).isUsed = used;
	}
}

double TPCHitMap::getHitsDistance(ModuleRowHit& hit1, ModuleRowHit& hit2)
{

    const double *pos1 = hit1.getHitPosition();
    const double *pos2 = hit2.getHitPosition();
 
    double delta12 = sqrt( (pos1[0] - pos2[0])*(pos1[0] - pos2[0]) +
                           (pos1[1] - pos2[1])*(pos1[1] - pos2[1]) +
                           (pos1[2] - pos2[2])*(pos1[2] - pos2[2]) );
 
    return delta12;
}

set< MarlinTPCHit > const & TPCHitMap::getHitsInModuleRow(int module, int row)
{
  HitContainer::iterator moduleRowIter = _hitContainer.find( ModuleRow(module, row) );
  if (moduleRowIter==_hitContainer.end())
  {
    streamlog_out(ERROR) << "No hits in module " << module << ", row " << row << endl;
    throw EVENT::DataNotAvailableException("module/row not found");
  }

  // return the set which belongs to this row
  return moduleRowIter->second;
}

}// namespace marlintpc
#endif //USE_TRACK_FITTER_KALMAN

