//// LCIO
#include "EVENT/LCIO.h"
#include "EVENT/LCCollection.h"
#include "EVENT/TrackerHit.h"
#include <IMPL/LCCollectionVec.h>
#include "IMPL/TrackImpl.h"
#include <IMPL/LCFlagImpl.h>

#include <iostream>
#include <iomanip>
#include <cmath>
#include <algorithm>

// Marlin
#include "marlin/Global.h"
#include "TrackFitterSimpleMinimizerProcessor.h"
#include "TrackFitterBase.h"

// Aida
#ifdef MARLIN_USE_AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>

//marlin
#include <marlin/ConditionsProcessor.h>


#endif

using namespace marlin;
using namespace lcio;

namespace marlintpc
{

TrackFitterSimpleMinimizerProcessor aTrackFitterSimpleMinimizerProcessor;

TrackFitterSimpleMinimizerProcessor::TrackFitterSimpleMinimizerProcessor()
        : Processor("TrackFitterSimpleMinimizerProcessor"),
	  _rejectOutliers(false), _trackCollectionParameters(0)
{

    _description = "Fit tracks using the TrackFitterSimpleMinimizer.";

    registerInputCollection(LCIO::TRACK,
                            "InputSeedTracks",
                            "The the name of the input collection of track candidates (default: TPCSeedTracks)",
                            _input_seed_tracks_collection_name,
                            std::string("TPCSeedTracks"));

    registerOutputCollection(LCIO::TRACK,
                             "OutputTracks",
                             "The name of the output collection with the fitted tracks(default: TPCTracks)",
                             _output_tracks_collection_name,
                             std::string("TPCTracks"));

    registerProcessorParameter("WriteOutputToStorage" ,
                               " If true the output collection is written to file (default: true)",
                               _writeOutput,
                               true);

    registerProcessorParameter("MaxIterations" ,
                               "Maximum number of iterations for outlier rejection (default: 10)",
                               _maxIterations,
                               10);

    registerOptionalParameter("CutValue" ,
                               "Cut outliers wich are _cutValue*RMS away from the fitted track, optional: Setting this value turns on this feature.",
                               _cutValue,
			      float(5.));

}


TrackFitterSimpleMinimizerProcessor::~TrackFitterSimpleMinimizerProcessor()
{
    delete _trackCollectionParameters;
}


void TrackFitterSimpleMinimizerProcessor::init()
{

    printParameters();

    // initialise the collection parameters
    _trackCollectionParameters = new IMPL::LCParametersImpl();

    // add processor parameters if any
    if (parameterSet("CutValue"))
    {
      _rejectOutliers=true;
    }
}


void TrackFitterSimpleMinimizerProcessor::processRunHeader(lcio::LCRunHeader *run)
{

    run->parameters().setValue(_processorName + "_revision", "$Rev:$");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
                }
        }

}

void TrackFitterSimpleMinimizerProcessor::processEvent(lcio::LCEvent *evt)
{

    streamlog_out(DEBUG4) << "processEvent" << std::endl;

    IMPL::LCCollectionVec* seedTracks = 0;
    try
        {
            // the tracker hit collection
            seedTracks =
                dynamic_cast<IMPL::LCCollectionVec*>
                (evt->getCollection(_input_seed_tracks_collection_name));
        }
    catch (DataNotAvailableException &)
        {
            streamlog_out(MESSAGE1) << "No seed tracks collection \"" << _input_seed_tracks_collection_name
            << "\" in event " << evt->getEventNumber() << std::endl;
            return;
        }

    // a collection to hold the new track
    IMPL::LCCollectionVec* outputTrackCollection = new IMPL::LCCollectionVec(EVENT::LCIO::TRACK);

    if (!_writeOutput)
    {
      outputTrackCollection->setTransient(true);
    }

    // Set flag that the hits are stored with the track.
    // Make sure the hit collection is not transient.
    LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(LCIO::TRBIT_HITS) ;
    outputTrackCollection->setFlag(trkFlag.getFlag());

    // set parameters for the output collection
    StringVec integerKeys;
    _trackCollectionParameters->getIntKeys(integerKeys);
    for (StringVec::iterator it = integerKeys.begin();
            it < integerKeys.end(); it ++)
        {
            outputTrackCollection->parameters().setValue(*it,
                                                         _trackCollectionParameters->getIntVal(*it));
        }

    StringVec floatKeys;
    _trackCollectionParameters->getFloatKeys(floatKeys);
    for (StringVec::iterator it = floatKeys.begin();
            it < floatKeys.end(); it ++)
        {
            outputTrackCollection->parameters().setValue(*it,
                                                         _trackCollectionParameters->getFloatVal(*it));
        }

    StringVec stringKeys;
    _trackCollectionParameters->getStringKeys(stringKeys);
    for (StringVec::iterator it = stringKeys.begin();
            it < stringKeys.end(); it ++)
        {
            outputTrackCollection->parameters().setValue(*it,
                                                         _trackCollectionParameters->getStringVal(*it));
        }

    streamlog_out(DEBUG3) << "requesting fitter" << std::endl;

    // request a fitter from the factory
    // a pointer to this event is passed so conditions data can be taken into
    // account
    TrackFitterBase *fitter = TrackFitterFactory::getTrackFitter(TrackFitterBase::FitterTypes::SIMPLEMINIMIZER,
                                                                 _trackCollectionParameters);

    streamlog_out(DEBUG3) << "fitter is at " << fitter << " fitter type "
    << fitter->getFitterType() << std::endl;

    //loop all tracks
    for (IMPL::LCCollectionVec::iterator trackIter = seedTracks->begin();
            trackIter < seedTracks->end() ; trackIter++)
        {
            EVENT::Track *inputTrack = dynamic_cast<EVENT::Track *>(*trackIter);

            if (inputTrack == 0)
                {
                    streamlog_out(WARNING2) << "bad_cast: Track candidates collection"
                    << _input_seed_tracks_collection_name
                    << " does not contain EVENT::Tracks!" << std::endl;
                    throw DataNotAvailableException("No track candidates found.");
                }


            streamlog_out(DEBUG3) << "fitting track" << std::endl;

            IMPL::TrackImpl* fittedTrack = fitter->fitTrack(inputTrack);

	    std::vector<double> currentMeanAndRMS = calculateMeanAndRMS( fittedTrack, fitter );

	    if (_rejectOutliers)
	    {
	      unsigned int iteration = 0;

	      for ( ; iteration < static_cast<unsigned int>(_maxIterations); ++iteration )
	      {
		// create a track with only the hits which are inside the cut
		IMPL::TrackImpl *newTrack = new IMPL::TrackImpl;
		newTrack->setD0( fittedTrack->getD0() );
		newTrack->setPhi( fittedTrack->getPhi() );
		newTrack->setTanLambda( fittedTrack->getTanLambda() );
		newTrack->setZ0( fittedTrack->getZ0() );
		newTrack->setOmega( fittedTrack->getOmega() );

		double cutXY = _cutValue*currentMeanAndRMS[2];
		double cutZ  = _cutValue*currentMeanAndRMS[3];

      		// add the hits which are inside the cut. Use the input track to take in hits
		// which might have been discarded in a previous iteration
		for (TrackerHitVec::const_iterator hitIter = inputTrack->getTrackerHits().begin() ;
		     hitIter < inputTrack->getTrackerHits().end() ; ++hitIter)
		{
		  EVENT::DoubleVec residuals = fitter->calculateResiduals(*hitIter,
									  fittedTrack);

		  if (fabs(currentMeanAndRMS[0] - residuals[0]) < cutXY 
		      && fabs(currentMeanAndRMS[1] - residuals[1]) < cutZ )
		  {
		    newTrack->addHit( *hitIter );
		  }
		}

//		if (newTrack->getTrackerHits().size() == fittedTrack->getTrackerHits().size() )
//		{
//		  // the loop has converged, we can stop here. No need to refit the track;
//		  delete newTrack;
//		  break;
//		}
		
		IMPL::TrackImpl *newFittedTrack = fitter->fitTrack(newTrack);
		delete newTrack; // we don't need the newTrack any more. It was just a container for the hits.
		newTrack=0;
		
		std::vector<double> newMeanAndRMS = calculateMeanAndRMS( newFittedTrack,
									 fitter );

		std::cout << "Iteration " << iteration
			  << ": meanXY is " <<newMeanAndRMS[0]
			  << ": rmsXY is "  <<newMeanAndRMS[2]
			  << ", meanZ is "  <<newMeanAndRMS[1]
			  << ", rmsZ is "   <<newMeanAndRMS[3]
			  << ", nHits is "  << newFittedTrack->getTrackerHits().size() 
			  << std::endl;

		// if one rms is growing, the other is going down: use the product
		// of the ratios. If the product is smaller than 1 the new setting is better on
		// average
		if ( newMeanAndRMS[2]/currentMeanAndRMS[2] 
		     * newMeanAndRMS[3]/currentMeanAndRMS[3] < 1. )
		{
		  
		  // prepare for the next iteration
		  delete fittedTrack;
		  fittedTrack = newFittedTrack;
		  currentMeanAndRMS=newMeanAndRMS;
		}
		else
		{
		  break;
		}
	      }
	      
	      if (iteration == static_cast<unsigned int>(_maxIterations))
	      {
		streamlog_out(MESSAGE0) << "cutting outliers did not converge within "
				      << _maxIterations << " iterations." << std::endl;
	      }
	      else
	      {
		std::cout << "cuting converged after " << iteration-1 
				      << " iterations." << std::endl;
	      }
	    }

	    // Now that we have the final fitted track we turn on the TRBIT_HITS flag
	    // so tracker hits are written with the track.
            fittedTrack->setTypeBit(LCIO::TRBIT_HITS);

            streamlog_out(DEBUG3) << "fitting done" << std::endl;

            //determine dEdx:
            //===============
            //FIXME: This currently calculates the mean charge per hit,
            // not the dE/dx because it is difficult to calculate the actual
            // track length in a general formula

            TrackerHitVec const & hitVec = fittedTrack->getTrackerHits();

            //      double phi =  outputTrack->getPhi();
            //      double d0 =  outputTrack->getD0();
            double Q_sum = 0;

            for (TrackerHitVec::const_iterator hit = hitVec.begin();
                    hit < hitVec.end(); hit++)
                {
                    Q_sum += (*hit)->getEDep();
                }

            fittedTrack->setdEdx(Q_sum / hitVec.size());

            outputTrackCollection->push_back(fittedTrack);

        } // loop tracks

    evt->addCollection(outputTrackCollection, _output_tracks_collection_name);

}// processEvent


Processor* TrackFitterSimpleMinimizerProcessor::newProcessor()
{

    return new TrackFitterSimpleMinimizerProcessor();
}


void TrackFitterSimpleMinimizerProcessor::check(lcio::LCEvent *)
{
}


void TrackFitterSimpleMinimizerProcessor::end()
{


}

std::vector<double> TrackFitterSimpleMinimizerProcessor::calculateMeanAndRMS( 
						EVENT::Track *track,
						TrackFitterBase *fitter)
{
  double sumResXY=0;
  double sumResSquareXY=0;
  double sumResZ=0;
  double sumResSquareZ=0;
  for (TrackerHitVec::const_iterator hitIter = track->getTrackerHits().begin() ;
       hitIter < track->getTrackerHits().end() ; ++hitIter)
  {
    EVENT::DoubleVec residuals = fitter->calculateResiduals(*hitIter,
							    track);
    sumResXY       += residuals[0];
    sumResSquareXY += residuals[0]*residuals[0];
    sumResZ        += residuals[1];
    sumResSquareZ  += residuals[1]*residuals[1];
  }
  
  size_t nHits=track->getTrackerHits().size();
  double rmsXY = sqrt(sumResSquareXY/nHits -  sumResXY/nHits * sumResXY/nHits);
  double rmsZ = sqrt(sumResSquareZ/nHits -  sumResZ/nHits * sumResZ/nHits);

  std::vector<double> retval(4);
  retval[0]=sumResXY/nHits;
  retval[1]=sumResZ/nHits;
  retval[2]=rmsXY;
  retval[3]=rmsZ;

  return retval;
}
  

} // namespace marlintpc
