
// STL
#include <vector>
#include <map>
#include <algorithm>
#include <iomanip>
#include <cassert>

// LCIO
#include "EVENT/LCIO.h"
#include "EVENT/LCCollection.h"
#include "EVENT/TrackerHit.h"
#include "IMPL/LCCollectionVec.h"
#include "IMPL/TrackImpl.h"
#include <IMPL/LCFlagImpl.h>
#include <IMPL/TrackerPulseImpl.h>

// GEAR
#include "gear/GEAR.h"
#include "gear/TPCParameters.h"
#include "gear/PadRowLayout2D.h"
#include "gearxml/GearXML.h"

// Marlin
#include "marlin/Global.h"
#include "TrackMergerProcessor.h"
#include "TrackConnection.h"

using namespace marlin;
using namespace lcio;

namespace marlintpc
{

TrackMergerProcessor aTrackMergerProcessor;

TrackMergerProcessor::TrackMergerProcessor() : Processor("TrackMergerProcessor")
{

    _description = "Merging track candidates for TrackFitterKalmanProcessor for MultiModule";

    registerInputCollection(LCIO::TRACK,
                            "InputTrackCandidates",
                            "The the name of the input collection of track candidates (default: TPCTrackCandidates)",
                            _input_track_candidates_collection_name,
                            std::string("TPCTrackCandidates"));

    registerOutputCollection(LCIO::TRACK,
                             "OutputMergeTracks",
                             "The name of the collection of the merged tracks (default: TPCMergedTracks)",
                             _output_merged_tracks_collection_name,
                             std::string("TPCMergedTracks"));

    registerProcessorParameter("SetConeAngle",
                               "The value of cosine for track connection (default: 0.94)",
                               _coneAngle,
                               double(0.94));

    registerProcessorParameter("SetOutputTransient" ,
                               "if not 0 the output hits collection is set transient (default: 0)"  ,
                               _outputIsTransient,
                               int(0));

}


TrackMergerProcessor::~TrackMergerProcessor()
{
}


void TrackMergerProcessor::init()
{

    // FIXME!!! This if statement will become obsolete / should be removed
    //          once the new logging mechanism is implemented in printParameters()
    if (streamlog::out.write< streamlog::MESSAGE3 >())
        {
            printParameters();
        }
}


void TrackMergerProcessor::processRunHeader(lcio::LCRunHeader *run)
{

    run->parameters().setValue(_processorName + "_revision", "$Rev: 1 $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
                }
        }
}


void TrackMergerProcessor::processEvent(lcio::LCEvent *evt)
{

    streamlog_out(DEBUG4) << "TrackMergerProcessor" << std::endl;

    IMPL::LCCollectionVec* trackMergeCandidates = 0;
    try
        {
            // the tracker hit collection
            trackMergeCandidates = dynamic_cast<IMPL::LCCollectionVec*>(evt->getCollection(_input_track_candidates_collection_name));
        }
    catch (DataNotAvailableException &)
        {
            streamlog_out(MESSAGE3) << "No seed tracks collection \"" << _input_track_candidates_collection_name
            << "\" in event " << evt->getEventNumber() << std::endl;
            return;
        }
    catch (std::bad_cast)
        {
            streamlog_out(WARNING1) << "Could not convert collection \"" << _input_track_candidates_collection_name
            << "\" to LCCollectionVec, skipping event " << evt->getEventNumber() << std::endl;
            return;
        }

    // a collection to hold the new track
    IMPL::LCCollectionVec* mergedTrackCollection = new IMPL::LCCollectionVec(EVENT::LCIO::TRACK);

    LCFlagImpl trkFlag(0);
    trkFlag.setBit(LCIO::TRBIT_HITS);
    mergedTrackCollection->setFlag(trkFlag.getFlag());

    _tracksByModuleID = new std::multimap<int, IMPL::TrackImpl*>;

    //map the track by cell number
    streamlog_out(DEBUG) << "Track mapping start" << std::endl;
    for (IMPL::LCCollectionVec::iterator trackIter = trackMergeCandidates->begin();
            trackIter < trackMergeCandidates->end() ; trackIter++)
        {
            IMPL::TrackImpl *seedTrack;
            try
                {
                    seedTrack = dynamic_cast<IMPL::TrackImpl *>(*trackIter);
                }
            catch (std::bad_cast)
                {
                    streamlog_out(WARNING1) << "bad cast. Convert to seedTrack, skipping " << std::endl;
                    return;
                }

            seedTrack->ext<TrackConnection::UsedFlag>() = 0; // reset UsedFlag

            TrackerHitVec::const_iterator hitIter = seedTrack->getTrackerHits().begin();

            TrackerPulse *pulse;
            try
                {
                    pulse = dynamic_cast<TrackerPulse *>((* hitIter)->getRawHits()[0]);
                }
            catch (std::bad_cast)
                {
                    streamlog_out(WARNING1) << "bad cast. Convert to pulse, skipping " << std::endl;
                    return;
                }
            streamlog_out(DEBUG0) << " pulse converting is OK!! " << std::endl;

            int ModuleID = pulse->getCellID1();
            //streamlog_out(MESSAGE) << " ModuleID = " << ModuleID << std::endl;
            _tracksByModuleID->insert(std::pair<int, IMPL::TrackImpl*>(ModuleID, seedTrack));
#if 0
            //layer and hitposition display
            int PadIndex = pulse->getCellID0();
            const double * hitPos = (*hitIter)->getPosition();
            streamlog_out(MESSAGE) << " PadIndex " << PadIndex << " at ( " << hitPos[0]
            << ", " << hitPos[1] << ", " << hitPos[2] << ") " << std::endl;
#endif
        }

#if 0
    //mapping check
    gear::TPCParameters const &theTPCParameters = marlin::Global::GEAR->getTPCParameters();
    int nmodules = theTPCParameters.getNModules();
    //streamlog_out(MESSAGE) << "# of Modules is "<< nmodules << std::endl;
    for (int ModuleID = 0; ModuleID < nmodules; ModuleID++)
        {
            std::pair< std::multimap<int, IMPL::TrackImpl*>::iterator, std::multimap<int, IMPL::TrackImpl*>::iterator> mapcheck;
            mapcheck = _tracksByModuleID->equal_range(ModuleID);
            streamlog_out(MESSAGE) << "Module " << ModuleID << " has " << distance(mapcheck.first, mapcheck.second) << " Tracks " << std::endl;
        }
#endif

#if 0
    //hitPosition check
    for (int ModuleID = 0; ModuleID < nmodules; ModuleID++)
        {
            std::multimap<int, IMPL::TrackImpl*>::iterator mapcheck;
            mapcheck = _tracksByModuleID->find(ModuleID);
            if (mapcheck == _tracksByModuleID->end())
                {
                    continue;
                }
            IMPL::TrackImpl *seedTrack = mapcheck->second;

            for (TrackerHitVec::const_iterator hitIter = seedTrack->getTrackerHits().begin();
                    hitIter < seedTrack->getTrackerHits().end(); hitIter++)
                {
                    TrackerPulse *pulse;
                    try
                        {
                            pulse = dynamic_cast<TrackerPulse *>((* hitIter)->getRawHits()[0]);
                        }
                    catch (std::bad_cast)
                        {
                            streamlog_out(WARNING1) << "bad cast. Convert to pulse, skipping " << std::endl;
                            return;
                        }
                    streamlog_out(DEBUG0) << " pulse converting is OK!! " << std::endl;

                    //streamlog_out(MESSAGE) << " ModuleID = " << ModuleID << std::endl;
                    streamlog_out(MESSAGE) << " ModuleID = " << pulse->getCellID1() << std::endl;
                    const double * hitPos = (*hitIter)->getPosition();
                    streamlog_out(MESSAGE) << " Hit Position at ( " << hitPos[0]
                    << ", " << hitPos[1] << ", " << hitPos[2] << ") " << std::endl;
                }
        }
#endif

    //create track connection object
    TrackConnection trackConnection;
    int trackNumber = 0;
    int numMergedTracks = 0;
    for (IMPL::LCCollectionVec::iterator trackIter = trackMergeCandidates->begin();
            trackIter < trackMergeCandidates->end() ; trackIter++)
        {
            IMPL::TrackImpl *seedTrack;
            try
                {
                    seedTrack = dynamic_cast<IMPL::TrackImpl *>(*trackIter);
                }
            catch (std::bad_cast)
                {
                    streamlog_out(WARNING1) << "bad cast. Convert to seedTrack, skipping " << std::endl;
                    return;
                }
            //continue if seedTrack already used
            if (seedTrack->ext<TrackConnection::UsedFlag>())
                {
                    streamlog_out(DEBUG) << "Track # " << trackNumber << " is already used." << std::endl;
                    trackNumber++;
                    continue;
                }
            IMPL::TrackImpl *mergedTrack = new IMPL::TrackImpl(*seedTrack); //To be renewd in mergeTracks

            trackConnection.setConeAngle(_coneAngle);
            trackConnection.mergeTracks(mergedTrack, *_tracksByModuleID);
            seedTrack->ext<TrackConnection::UsedFlag>() = 1; // set UsedFlag
            streamlog_out(DEBUG) << "Track # " << trackNumber << " set UsedFlag." << std::endl;
            trackNumber++;
            streamlog_out(DEBUG3) << "MergedTrack has " << (int)distance(mergedTrack->getTrackerHits().begin(), mergedTrack->getTrackerHits().end())
            << " Hits." << std::endl;

            mergedTrackCollection->push_back(mergedTrack);
            numMergedTracks++;
        }
    streamlog_out(MESSAGE4) << "# of Tracks after merging: " << numMergedTracks << std::endl;

    // now add the collection to the event!
    try
        {
            evt->addCollection(mergedTrackCollection, _output_merged_tracks_collection_name);
        }
    catch (EVENT::EventException e)
        {
            streamlog_out(WARNING) << e.what() << "Collection " << _output_merged_tracks_collection_name << " is being deleted." << std::endl;
            std::vector<EVENT::LCObject*>::iterator iter = mergedTrackCollection->begin();
            for (; iter != mergedTrackCollection->end(); iter++)
                {
                    if (*iter)
                        {
                            delete(*iter);
                            (*iter) = NULL;
                        }
                }
            mergedTrackCollection->clear();
            delete mergedTrackCollection;
        }
}// processEvent


Processor* TrackMergerProcessor::newProcessor()
{

    return new TrackMergerProcessor();
}


void TrackMergerProcessor::check(lcio::LCEvent * /*evt*/)
{
}


void TrackMergerProcessor::end()
{
}

} // namespace marlintpc
