
// MarlinTPC
#include "ElectronCloudDrifterProcessor.h"
#include "TPCConditions.h"
#include "ElectronCloud.h"
#include "GlobalFieldProcessor.h"

// CLHEP
#include <CLHEP/Random/RandGauss.h>
#include <CLHEP/Matrix/Vector.h>
#include <CLHEP/Vector/ThreeVector.h>

// LCIO
#include <EVENT/LCIO.h>
#include <EVENT/LCCollection.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/SimTrackerHitImpl.h>

// GEAR
#include <gear/BField.h>
#include <gear/GearMgr.h>
#include <gearimpl/Vector3D.h>

// Marlin
#include <marlin/Global.h>
#include <marlin/Exceptions.h>

// GSL
#include <gsl/gsl_errno.h>
#include <gsl/gsl_odeiv.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#ifdef MARLIN_USE_AIDA
//AIDA
#include <marlin/AIDAProcessor.h>
#include <AIDA/AIDA.h>
#include <AIDA/ITupleFactory.h>

//ROOT
#include <TBranch.h>
#endif


#include <cassert>

using namespace std;
using namespace EVENT;
using namespace IMPL;
using namespace marlin;
using namespace CLHEP;
using namespace tpcconddata;
using namespace gear;

namespace marlintpc {

// static variables
/** the ODE solving functions from the GSL
*/
gsl_odeiv_step *ElectronCloudDrifterProcessor::stepper = 0;

gsl_odeiv_control *ElectronCloudDrifterProcessor::step_control = 0;

gsl_odeiv_evolve *ElectronCloudDrifterProcessor::evolver = 0;

// create one to register it with Marlin
ElectronCloudDrifterProcessor anElectronCloudDrifterProcessor;


ElectronCloudDrifterProcessor::ElectronCloudDrifterProcessor():
  Processor("ElectronCloudDrifterProcessor"),
	_input_collection_name("TPCElectronClouds"),
	_tpc_conditions_collection_name(TPCConditions::getDefaultColName()),
	_tpc_section_name(""),
	_primary_drift_section(false),
	_omega_tau(10.),
	_drift_velocity(40.),
	_longitudinal_diffusion(200.),
	_transverse_diffusion(50.),
	_tpc_outer_radius(1500.),
	_tpc_inner_radius(300.),
	_section_outer_z(100.),
	_section_inner_z(10.),
	_disable_magnetic_field(false)
  {

	registerInputCollection(
		LCIO::LCGENERICOBJECT,
		"InputTPCElectronClouds",
		"Name of the input electron cloud collection.",
		_input_collection_name,
		string("TPCElectronClouds") );

	registerOptionalParameter(
		"TPCConditionsName",
		"The name to use for the tpc conditions collection.",
		_tpc_conditions_collection_name,
		TPCConditions::getDefaultColName());

	registerOptionalParameter(
		"TPCSectionName",
		"The name of this section of the TPC.",
		_tpc_section_name,
		string(""));

	registerOptionalParameter(
		"PrimaryDriftSection?",
		"Set as true if this drift section is the main drift section of the TPC.",
		_primary_drift_section,
		false);

	registerOptionalParameter(
    "DisableMagneticField",
    "Set true if you want to disable the magnetic field in the drifting process.",
    _disable_magnetic_field,
    _disable_magnetic_field);

	registerProcessorParameter(
		"OmegaTau",
		"The gas parameterization for this section of the TPC",
		_omega_tau,
		10. );

	registerProcessorParameter(
		"DriftVelocity",
		"The parameterized drift velocity in this section of the TPC (mm/us)",
		_drift_velocity,
		40. );

	registerProcessorParameter(
		"LongitudinalDiffusion",
		"The parameterized longitudinal diffusion in this section of the TPC (um/sqrt(cm))",
		_longitudinal_diffusion,
		200. );

	registerProcessorParameter(
		"TransverseDiffusion",
		"The parameterized transverse diffusion in this section of the TPC (um/sqrt(cm))",
		_transverse_diffusion,
		50. );

	registerProcessorParameter(
		"TPCOuterRadius",
		"The outer radius of the TPC (mm)",
		_tpc_outer_radius,
		1500.);

	registerProcessorParameter(
		"TPCInnerRadius",
		"The inner radius of the TPC (mm)",
		_tpc_inner_radius,
		300.);

	registerProcessorParameter(
		"SectionOuterZ",
		"The absolute value of the z coordinate of the outermost position of this section of the TPC (mm)",
		_section_outer_z,
		100.);

	registerProcessorParameter(
		"SectionInnerZ",
		"The absolute value of the z coordinate of the innermost position of this section of the TPC (mm)",
		_section_inner_z,
		10.);

	if( !stepper ) {

		stepper = gsl_odeiv_step_alloc (gsl_odeiv_step_rkf45, 4);
	}

	if( !step_control ) {

		step_control = gsl_odeiv_control_y_new(1.E-03, 0.0);
	}

	if( !evolver ) {

		evolver = gsl_odeiv_evolve_alloc(4);
	}
}


ElectronCloudDrifterProcessor::~ElectronCloudDrifterProcessor() {

	if( evolver ) {

		gsl_odeiv_evolve_free(evolver);
	}

	if( step_control ) {

		gsl_odeiv_control_free(step_control);
	}

	if( stepper ) {

		gsl_odeiv_step_free(stepper);
	}
}


void ElectronCloudDrifterProcessor::processRunHeader(lcio::LCRunHeader* run) {

	// store the processor parameters in the event
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {

        if ( ! i->second->isOptional() || i->second->valueSet() ) {

            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }
}


void ElectronCloudDrifterProcessor::processEvent(lcio::LCEvent * evt ) {

	// grab the collection of electron clouds
	LCCollection *electron_cloud_collection = NULL;
	try {

		electron_cloud_collection = evt->getCollection( _input_collection_name );
	}
	catch (DataNotAvailableException &e) {

		// the collection of electron clouds with that name wasn't in the LCIO event file, skip simulating this event
		streamlog_out( WARNING ) << e.what() << endl;
		throw marlin::SkipEventException( this );
	}

#ifdef MARLIN_USE_AIDA
	/**
	 *  Initializations needs to be here since the number of branches depends on the number of clouds per event.
	 *  \remarks It is assumed that the number of clouds per event is constant!
	 */
  if(_tDrift.size() == 0){
    try {
      AIDAProcessor::tree(this);
      for(int photodot = 0; photodot < electron_cloud_collection->getNumberOfElements(); photodot++){
        driftInfo info;
        TTree* t_tmp = new TTree((TString("driftInfo")+(long)photodot),"Detailed information about the electron drift");
        t_tmp->Branch("x","std::vector<double>",&info.position[0]);
        t_tmp->Branch("y","std::vector<double>",&info.position[1]);
        t_tmp->Branch("z","std::vector<double>",&info.position[2]);
        t_tmp->Branch("vx","std::vector<double>",&info.velocity[0]);
        t_tmp->Branch("vy","std::vector<double>",&info.velocity[1]);
        t_tmp->Branch("vz","std::vector<double>",&info.velocity[2]);
        t_tmp->Branch("Ex","std::vector<double>",&info.Ex);
        t_tmp->Branch("Ey","std::vector<double>",&info.Ey);
        t_tmp->Branch("Ez","std::vector<double>",&info.Ez);
        t_tmp->Branch("Bz","std::vector<double>",&info.Bz);
        t_tmp->Branch("Br","std::vector<double>",&info.Br);
        _tDrift.push_back(t_tmp);
      }

    } catch (lcio::Exception &e) {
      streamlog_out(DEBUG) << "No AIDA processor found. Will not produce a root file..." << endl;
      streamlog_out(DEBUG)<< e.what() << endl;
    }
  }
#endif
	streamlog_out( MESSAGE0 ) << "Drifting " << electron_cloud_collection->getNumberOfElements() << " electron clouds through TPC section " << _tpc_section_name << "\n";
	streamlog_out( DEBUG4 ) << "Gas Parameters; drift velocity: " << _drift_velocity << "mm/us, long diffusion: " << _longitudinal_diffusion << "sqrt(mm), trans diffusion: " << _transverse_diffusion << "sqrt(mm), omega*tau: " << _omega_tau << endl;

	// if this is the primary drift section of the TPC then add the parameters to the TPC conditions class
	if( _primary_drift_section ) {

		TPCConditions* tpc_conditions = NULL;

		try {
			LCCollection *tpc_conditions_collection = evt->getCollection( _tpc_conditions_collection_name );
			tpc_conditions = dynamic_cast<TPCConditions*>( tpc_conditions_collection->getElementAt(0) );
			assert(tpc_conditions);
		}
		catch (DataNotAvailableException &e) {

			LCCollectionVec *tpc_conditions_collection = new LCCollectionVec( LCIO::LCGENERICOBJECT );
			tpc_conditions_collection->setTransient(false);
			tpc_conditions_collection->setSubset(false);
			tpc_conditions = new TPCConditions();
			tpc_conditions_collection->addElement( tpc_conditions );
			evt->addCollection( tpc_conditions_collection, _tpc_conditions_collection_name );
		}

		tpc_conditions->setDriftVelocity( _drift_velocity );

		// tpcconddata uses sqrt(mm) for the diffusion constants... 1sqrt(mm) = 10^-3.5 um/sqrt(cm)
		tpc_conditions->setLongDiffusionCoef( _longitudinal_diffusion );
		tpc_conditions->setTransDiffusionCoef( _transverse_diffusion );
	}
	
	// now loop through the clouds, transporting them to the end
	for( _current_cloud = 0; _current_cloud < electron_cloud_collection->getNumberOfElements(); _current_cloud++ ) {

		ElectronCloud *current_cloud = dynamic_cast<ElectronCloud*>( electron_cloud_collection->getElementAt( _current_cloud ) );
		assert(current_cloud);

		// initialize the ode state
		double state[4];
		state[0] = current_cloud->get_x_position();
		state[1] = current_cloud->get_y_position();
		state[2] = current_cloud->get_z_position();
		state[3] = current_cloud->get_time();

		// now transport the cloud to the end
		drift_through_cylinder( state, _tpc_inner_radius, _tpc_outer_radius, _section_inner_z, _section_outer_z, _drift_velocity, _omega_tau, _disable_magnetic_field, this );

		const double z_drift_distance = fabs(this->_section_outer_z - current_cloud->get_z_position());

		// now smear the cloud position a bit
		const double sxy = sqrt(z_drift_distance) * this->_transverse_diffusion;
		const double st = sqrt(z_drift_distance) * this->_longitudinal_diffusion;

		streamlog_out( DEBUG3 ) << "Drifted cloud #" << _current_cloud + 1 << " from (x,y,z,t): ("
		    << current_cloud->get_x_position() << "mm, "
		    << current_cloud->get_y_position() << "mm, "
		    << current_cloud->get_z_position() << "mm, "
		    << current_cloud->get_time() << ") to (x,y,z,t): ("
		    << state[0] << "mm, "
		    << state[1] << "mm, "
		    << state[2] << "mm, "
		    << state[3] << "us)\n";

		current_cloud->set_x_position( state[0] );
		current_cloud->set_y_position( state[1] );
		current_cloud->set_z_position( state[2] );
		current_cloud->set_transverse_stdev(  sqrt( (current_cloud->get_transverse_stdev()*current_cloud->get_transverse_stdev()) + (sxy*sxy) ) );
		current_cloud->set_longitudinal_stdev(  sqrt( (current_cloud->get_longitudinal_stdev()*current_cloud->get_longitudinal_stdev()) + (st*st) ) );
		current_cloud->set_time( state[3] );

		streamlog_out( DEBUG1 ) << "Cloud trans stdev: " << current_cloud->get_transverse_stdev() << "mm, long stddev: " << current_cloud->get_longitudinal_stdev() << "us\n";
	}

}


void ElectronCloudDrifterProcessor::init() {

	// convert the transverse diffusion constants into sqrt(mm)
	// convert the longitudinal diffusion into us/sqrt(mm)
	// tpcconddata uses sqrt(mm) for the diffusion constants... 1sqrt(mm) = 10^-3.5 um/sqrt(cm)
	const double conversion = pow(10., -3.5);
	_longitudinal_diffusion *= conversion/_drift_velocity;
	_transverse_diffusion *= conversion;

 // usually a good idea to
	printParameters();
}


void ElectronCloudDrifterProcessor::check(lcio::LCEvent *  /*evt*/ ) {

}


void ElectronCloudDrifterProcessor::end() {


}


int ElectronCloudDrifterProcessor::ion_drift_derivative( double  /*t*/, const double y[], double f[], void *params ) {

	const Hep3Vector position ( y[0], y[1], y[2] );
  DriftParameters &drift_parameters = *( (DriftParameters*)(params) );

	// calculate B and E
	Hep3Vector tmp;
	if(drift_parameters._disable_magnetic_field){
	  tmp = Hep3Vector(0,0,0);
	} else {
	  Vector3D tmp_B = marlin::Global::GEAR->getBField().at(Vector3D( y[0], y[1], y[2] ));
    tmp = Hep3Vector(tmp_B.x(),tmp_B.y(),tmp_B.z()).unit();
	}
	const Hep3Vector B = tmp.unit();
	const Hep3Vector E = GlobalFieldProcessor::get_instance().get_global_electric_field( position ).unit();

	// use references to make the formula easier to read
	const double &ot = drift_parameters._omega_tau;
	const double &ot_2 = drift_parameters._omega_tau_squared;

	const double prefix = drift_parameters._drift_velocity / (1. + ot_2);

	// remember, E and B are already unit vectors
	const double e_dot_b = E.dot(B);					// unit(E) . unit(B)
	const CLHEP::Hep3Vector e_cross_b = E.cross(B);		// unit(E) X unit(B)

	const CLHEP::Hep3Vector velocity = prefix * ( E + (ot*e_cross_b) + (ot_2*e_dot_b*B) );

	streamlog_out( DEBUG5 ) << "At: " << position << ", E: " << E << ", B: " << B << ", Velocity: " << velocity << endl;
	streamlog_out( DEBUG5 ) << "ot*(E X B): " << (ot*e_cross_b) << ", (ot)^2*(E.B)B: " << (ot_2*e_dot_b*B) << endl;
  // trace drifted clouds

  if(drift_parameters._info){
    drift_parameters._info->position[0]->push_back(y[0]);
    drift_parameters._info->position[1]->push_back(y[1]);
    drift_parameters._info->position[2]->push_back(y[2]);
    drift_parameters._info->velocity[0]->push_back(velocity.x());
    drift_parameters._info->velocity[1]->push_back(velocity.y());
    drift_parameters._info->velocity[2]->push_back(velocity.z());
    drift_parameters._info->Ex->push_back(E.x());
    drift_parameters._info->Ey->push_back(E.y());
    drift_parameters._info->Ez->push_back(E.z());
    drift_parameters._info->Bz->push_back(B.z());
    drift_parameters._info->Br->push_back(B.rho());
  }

	f[0] = velocity.x();
	f[1] = velocity.y();
	f[2] = velocity.z();
	f[3] = 1.;

	return GSL_SUCCESS;
}


int ElectronCloudDrifterProcessor::drift_through_cylinder( double (&state)[4], const double  /*min_radius*/, const double  /*max_radius*/, const double  /*min_z*/, const double max_z, const double drift_velocity, const double omega_tau, const bool disable_magnetic_field, marlintpc::ElectronCloudDrifterProcessor* processor ) {
  driftInfo* info = 0;
  TTree* tDrift = 0;
#ifdef MARLIN_USE_AIDA
  if((unsigned int)processor->_current_cloud < processor->_tDrift.size())
    tDrift = processor->_tDrift.at(processor->_current_cloud);
  if(tDrift){
    info = new driftInfo();
    info->position[0] = new vector<double>();
    info->position[1] = new vector<double>();
    info->position[2] = new vector<double>();
    info->velocity[0] = new vector<double>();
    info->velocity[1] = new vector<double>();
    info->velocity[2] = new vector<double>();
    info->Ex = new vector<double>();
    info->Ey = new vector<double>();
    info->Ez = new vector<double>();
    info->Bz = new vector<double>();
    info->Br = new vector<double>();
    TBranch* tmp;
    tmp = tDrift->GetBranch("x");
    if(tmp) tmp->SetAddress(&info->position[0]);
    tmp = tDrift->GetBranch("y");
    if(tmp) tmp->SetAddress(&info->position[1]);
    tmp = tDrift->GetBranch("z");
    if(tmp) tmp->SetAddress(&info->position[2]);
    tmp = tDrift->GetBranch("vx");
    if(tmp) tmp->SetAddress(&info->velocity[0]);
    tmp = tDrift->GetBranch("vy");
    if(tmp) tmp->SetAddress(&info->velocity[1]);
    tmp = tDrift->GetBranch("vz");
    if(tmp) tmp->SetAddress(&info->velocity[2]);
    tmp = tDrift->GetBranch("Ex");
    if(tmp) tmp->SetAddress(&info->Ex);
    tmp = tDrift->GetBranch("Ey");
    if(tmp) tmp->SetAddress(&info->Ey);
    tmp = tDrift->GetBranch("Ez");
    if(tmp) tmp->SetAddress(&info->Ez);
    tmp = tDrift->GetBranch("Bz");
    if(tmp) tmp->SetAddress(&info->Bz);
    tmp = tDrift->GetBranch("Br");
    if(tmp) tmp->SetAddress(&info->Br);
  }
#endif

	DriftParameters drift_parameters;

	drift_parameters._drift_velocity = drift_velocity;
	drift_parameters._omega_tau = omega_tau;
	drift_parameters._omega_tau_squared = omega_tau*omega_tau;
	drift_parameters._disable_magnetic_field = disable_magnetic_field;
	drift_parameters._info = info;

	// reset the solvers
	gsl_odeiv_step_reset(stepper);
	// This function initializes the control function c with the parameters eps_abs (absolute error), eps_rel (relative error), a_y (scaling factor for y) and a_dydt (scaling factor for derivatives)
	// If eps_abs is too small wierd effects can happen!
	gsl_odeiv_control_init(step_control, 1.E-06, 0., 1., 0.);
	gsl_odeiv_evolve_reset(evolver);

	gsl_odeiv_system system = {ion_drift_derivative, 0, 4, &drift_parameters};

	double step_size = 1.;

	double time = state[3];

	// drift the cloud further towards the electronics
	while( fabs(state[2]) < max_z ) {
	  // time + 1000  is the maximum allowed time step
		if( gsl_odeiv_evolve_apply(evolver, step_control, stepper, &system, &time, time + 1000., &step_size, state) != GSL_SUCCESS ) {

			break;
		}
	}

	// now move the cloud backwards to hit the end of this section
	double derivative[4];

	ion_drift_derivative( 0., state, derivative, (void*)&drift_parameters );
	const double back_distance = state[2] - max_z;

	// if the last step in z was dz we have to go back realtive to dz: (back_distance)/dz
	// this factor is also used to scale the other steps, eg x -= (back_distance)/dz * dx, y -= (back_distance)/dz * dy, z -= (back_distance)/dz * dz
	state[0] -= (back_distance / derivative[2]) * derivative[0];
	state[1] -= (back_distance / derivative[2]) * derivative[1];
	state[2] -= back_distance;
	state[3] -= (back_distance / derivative[2]) * derivative[3];

	if(tDrift){
	  // add final poition
   info->position[0]->push_back(state[0]);
   info->position[1]->push_back(state[1]);
   info->position[2]->push_back(state[2]);
	}

#ifdef MARLIN_USE_AIDA
	if(tDrift) tDrift->Fill();
	delete info;
#endif
	return GSL_SUCCESS;
}


}
