#ifndef PHOTOELECTRICPULSEPROCESSOR_H
#define PHOTOELECTRICPULSEPROCESSOR_H 1

#include <string>
#include <vector>

#include "marlin/Processor.h"
#include "PhotoelectricXMLHelper.h"

namespace EVENT {

	class LCRunHeader;
	class LCEvent;
}

namespace marlintpc {

/** PhotoelectricPulseProcessor : This processor takes a file containing positions of photosensitive dots and lines and simulates them being hit by a pulse of photons
 * 
 *    @author J. Abernathy (UVic)<br>
 */
	class PhotoelectricPulseProcessor : public marlin::Processor {

public:
	/**
	 * constructor
	 */
	PhotoelectricPulseProcessor();

	/** destructor
	 */
	virtual ~PhotoelectricPulseProcessor();

	/** used by Marlin to create a new processor
	 */
	virtual Processor*  newProcessor() { return new PhotoelectricPulseProcessor() ; }

	/** Called at the begin of the job before anything is read.
	 * Use to initialize the processor, e.g. book histograms.
	 */
	virtual void init() ;

	/** Called for every run.
	 */
	virtual void processRunHeader(lcio::LCRunHeader* run ) ;

	/** Called for every event - the working horse.
	 */
	virtual void processEvent(lcio::LCEvent * evt ) ;

	/** Called after the event has been processed
	 */
	virtual void check(lcio::LCEvent * evt ) ;

	/** Called after data processing for clean up.
	 */
	virtual void end() ;


private:

	/** the name of the file containing the positions of the photodots and photolines
	 */
	std::string _photogeometry_filename;

	/** the name of the output LCGenericObject collection from this processor containing electron clouds
	 */
	std::string _output_collection_name;

	/** the name of the output collection which will contain the SimTrackerHits
	 */
	std::string _output_simtrackerhit_collection_name;

	/** the mean number of electrons to create per photodot
	*/
	double _mean_electrons_per_photodot;

	/** the initial z position of the electron clouds
	*/
	double _initial_z_position;

	/** the collection of photodots (assume they will stay the same for the entire run)
	 */
	std::vector<std::vector<Photodot>::const_iterator> _photodots;

	/** the collection of photolines (assume they will stay the same for the entire run)
	 */
	std::vector<std::vector<Photoline>::const_iterator> _photolines;

	/** The xml handler which has all the information about the photodots.
	 */
	PhotoelectricXMLFile* _xmlFile;

	/** this number corresponds to the module number which is considered for the simulation
	 *  Only these photodots that are counterpart to the considered module are used.
	 *  Module numbers: 
	 *    -top: left 1, right 2
	 *    -middle: left 3, middle 4, right 5 
	 *    -bottom: left 6, right 7
	 *  All modulesw: -1
	 */
	int _module;
	int _transient;///<Sets output collection to transient when set to another value than 0.
	bool _doPhotoLines;///<Enables the simulations of photo lines
	bool _doPhotoDots;///<Enables the simulations of photo dots
	bool _disableSmearing;///<Disables the uniform distribution of electrons on the photodot with diameter 2mm
};

} // namespace marlintpc

#endif // PHOTOELECTRICPULSEPROCESSOR_H
