// MarlinTPC
#include "ElectronCloud.h"
#include "PhotoelectricPulseProcessor.h"

// Marlin
#include <marlin/Exceptions.h>

// LCIO
#include <EVENT/LCIO.h>
#include <EVENT/LCCollection.h>
#include <EVENT/MCParticle.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/SimTrackerHitImpl.h>

// GEAR
#include <gear/TPCModule.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>
#include <gearxml/tinyxml.h>

// CLHEP
#include <CLHEP/Random/RandPoisson.h>
#include <CLHEP/Random/RandFlat.h>
#include <CLHEP/Vector/TwoVector.h>

// stl
#include <iostream>
#include <fstream>
#include <math.h>

using namespace std;
using namespace EVENT;
using namespace IMPL;
using namespace marlin;
using namespace CLHEP;
using namespace gear;

namespace marlintpc {

// create one to register it with Marlin
PhotoelectricPulseProcessor aPhotoelectricPulseProcessor;


PhotoelectricPulseProcessor::PhotoelectricPulseProcessor():
	Processor("PhotoelectricPulseProcessor"),
	_photogeometry_filename("photo_geometry.txt"),
	_output_collection_name ( "TPCElectronClouds" ),
	_output_simtrackerhit_collection_name( "TPCSimPhotodots" ),
	_mean_electrons_per_photodot(100.),
	_doPhotoLines( false ),
	_doPhotoDots( true ),
	_disableSmearing( false ){

	registerOutputCollection(
		LCIO::LCGENERICOBJECT,
		"OutputTPCElectronClouds",
		"Name of the output LCGeneric collection full of electron clouds",
		_output_collection_name,
		string("TPCElectronClouds") );

	registerOutputCollection(
		LCIO::SIMTRACKERHIT,
		"OutputSimPhotodots",
		"Name of the output SimTrackerHit collection containing the true centroids of the photodots.",
		_output_simtrackerhit_collection_name,
		string("TPCSimPhotodots") );

	registerOptionalParameter(
		"GeometryFilename",
		"The name of the file containing the positions of the photodots and photolines.",
		_photogeometry_filename,
		string("photo_geometry.txt") );

	registerOptionalParameter(
		"MeanElectronsPerPhotodot",
		"The mean number of electrons to create per photodot (Poisson distribution)",
		_mean_electrons_per_photodot,
		100.);

	registerOptionalParameter(
		"InitialZPosition",
		"The initial z position of the electron clouds.",
		_initial_z_position,
		0.);
	registerOptionalParameter(
		"Module",
		"Give a certain module number to consider only photodots from this module.",
		_module,
		-1);
	registerOptionalParameter(
	   "doPhotoLines",
	   "If true electrons from photo lines will be used.",
	   _doPhotoLines,
	   false);
	registerOptionalParameter(
     "doPhotoDots",
     "If true electrons from photo lines will be used.",
     _doPhotoDots,
     true);

  registerProcessorParameter("Transient",
      "switch off writing of output collection (derfault: 0)", _transient,
      (int) 0);
  registerProcessorParameter("disablePositionSmearing",
      "Switch off the uniform distribution of electrons in the photodot with diameter 2mm",
      _disableSmearing,
      false);
}


PhotoelectricPulseProcessor::~PhotoelectricPulseProcessor() {}


void PhotoelectricPulseProcessor::processRunHeader(lcio::LCRunHeader* run) {

	// store the processor parameters in the event
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {

        if ( ! i->second->isOptional() || i->second->valueSet() ) {

            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }
}


void PhotoelectricPulseProcessor::processEvent(lcio::LCEvent * evt ) {

	// create a collection to store the electron clouds (don't write this collection at the end)
	LCCollectionVec *electroncloud_collection = new LCCollectionVec( LCIO::LCGENERICOBJECT );
	if (_transient)
	  electroncloud_collection->setTransient(true);

	// create a collection to store the SimTrackerHits
	LCCollectionVec *simtrackerhit_collection = new LCCollectionVec( LCIO::SIMTRACKERHIT );
	if (_transient)
	  simtrackerhit_collection->setTransient(true);

	// keep track of the number of clouds produced for output
	unsigned int number_of_electron_clouds = 0;

	streamlog_out(MESSAGE) << "Size of the photodot vector: " << _photodots.size() << endl;

	if(_doPhotoDots) {
    // now loop through the photodots creating clouds of electrons
	  for ( vector<vector<Photodot>::const_iterator>::iterator i_photodot = _photodots.begin(); i_photodot != _photodots.end(); i_photodot++ ) {
      const int number_of_electrons = RandPoisson::shoot(_mean_electrons_per_photodot);

      if( number_of_electrons <= 0 ) {

        continue;
      }

      // create an electron cloud
      ElectronCloud *new_cloud = new ElectronCloud();

      // set up the deviation in the position of the electrons in the cloud
      if(_disableSmearing){
        new_cloud->set_transverse_stdev( (*i_photodot)->_diameter/4. ); // mm
      } else {
        // do not add something here since the position itself is smeared
        new_cloud->set_transverse_stdev( 0 );
      }
      new_cloud->set_longitudinal_stdev(0.01); // us

      new_cloud->set_cloud_id( number_of_electron_clouds++ );

      // set up the mean position of the electrons in the cloud
      if(_disableSmearing){
        new_cloud->set_x_position((*i_photodot)->_position.x() );
        new_cloud->set_y_position( (*i_photodot)->_position.y() );
      } else {
        // create a random point in a circle with radius 2mm corresponding to the photodot
        double r = 100;
        double dx,dy;
        while (r > 2){
          dx = RandFlat::shoot(-2,2);
          dy = RandFlat::shoot(-2,2);
          r = sqrt(dx*dx + dy*dy);
        }
        new_cloud->set_x_position((*i_photodot)->_position.x() + dx);
        new_cloud->set_y_position( (*i_photodot)->_position.y() + dy);
      }
      new_cloud->set_z_position( _initial_z_position );

      // set up the other parameters
      new_cloud->set_time( 0. );
      new_cloud->set_electron_count( number_of_electrons );

      streamlog_out( DEBUG ) << "Produced a cloud with " << new_cloud->get_electron_count()
                             << " electrons at \t" << new_cloud->get_x_position()
                             << "mm, \t" << new_cloud->get_y_position()
                             << "mm\t time:" << new_cloud->get_time() << "\n";
      // now add the cloud to the collection
      electroncloud_collection->addElement( new_cloud );

      // create a SimTrackerHit which holds the true centroid of the photodot
      SimTrackerHitImpl *simtrackerhit = new SimTrackerHitImpl();

      double position[3];
      position[0] = (*i_photodot)->_position.x();
      position[1] = (*i_photodot)->_position.y();
      position[2] = _initial_z_position;
      simtrackerhit->setPosition( position );

      simtrackerhit_collection->addElement( simtrackerhit );
    }
	}
	if(_doPhotoLines) {
    streamlog_out ( DEBUG ) << "Producing photoelectrons along the photolines." << endl;

    // now loop through the lines making clouds of electrons
    for ( vector<vector<Photoline>::const_iterator>::iterator i_photoline = _photolines.begin(); i_photoline != _photolines.end(); i_photoline++ ) {

      const Hep2Vector start ( (*i_photoline)->_start.x(), (*i_photoline)->_start.y() );
      const Hep2Vector end ( (*i_photoline)->_end.x(), (*i_photoline)->_end.y() );

      const float length = ( start - end ).mag();
      const float width = (*i_photoline)->_width;

      // assuming a photoelectron density of ~16 e / mm^2 calculate the linear density (along the length)
      //const float linear_density = 16. * width;

      // try to place the clouds once every 3 mm along the length
      const int n_length_divisions = static_cast<int>(ceil ( length / 2. ));
      const Hep2Vector d_length = ( length / n_length_divisions ) * ( end - start ).unit();

      Hep2Vector current_position = start;
      for ( int i_length = 0; i_length <= n_length_divisions; i_length++ ) {

        const int number_of_electrons = RandPoisson::shoot(_mean_electrons_per_photodot);

        if( number_of_electrons <= 0 ) {

          continue;
        }

        // create an electron cloud
        ElectronCloud *new_cloud = new ElectronCloud();

        // set up the deviation in the position of the electrons in the cloud
        new_cloud->set_transverse_stdev( width / 4. );
        new_cloud->set_longitudinal_stdev(0.01);

        new_cloud->set_cloud_id( number_of_electron_clouds++ );

        // set up the mean position of the electrons in the cloud
        new_cloud->set_x_position( current_position.x() );
        new_cloud->set_y_position( current_position.y() );
        new_cloud->set_z_position( _initial_z_position );

        streamlog_out( DEBUG ) << "Produced a cloud with " << number_of_electrons << " electrons at \t" << new_cloud->get_x_position() << " mm, \t" << new_cloud->get_y_position() << " mm.\n";

        // set up the other parameters
        new_cloud->set_time( 0. );
        new_cloud->set_electron_count( number_of_electrons );

        // now add the cloud to the collection
        electroncloud_collection->addElement( new_cloud );

        current_position += d_length;
      }
    }
	}

	streamlog_out( MESSAGE ) << "Produced " << number_of_electron_clouds << " electron clouds.\n";

	// now add the collection to the event so other processors may access it
	try {

		evt->addCollection( electroncloud_collection, _output_collection_name );
		evt->addCollection( simtrackerhit_collection, _output_simtrackerhit_collection_name );
	}
	catch( ... ) {

		throw SkipEventException( this );
	}
	
}


void PhotoelectricPulseProcessor::init() {

	// usually a good idea to
	printParameters();

	_xmlFile = new PhotoelectricXMLFile( _photogeometry_filename.c_str() );
	_photodots = _xmlFile->GetPhotodots();
	_photolines = _xmlFile->GetPhotolines();

	if(_photodots.size() == 0 && _photolines.size() == 0){
		streamlog_out( ERROR ) << "Could not find a photodot or a photoline!" << endl;
	}

	streamlog_out(MESSAGE) << "Loaded information of " << _photodots.size() << " photodots and " << _photolines.size() << " photolines." << endl;
}


void PhotoelectricPulseProcessor::check(lcio::LCEvent *  /*evt*/ ) { 

}


void PhotoelectricPulseProcessor::end() { 
  delete _xmlFile;
}

}
