/*
 * TAltroConverter.hh
 *
 *  Created on: Jul 6, 2012
 *      Author: caiazza
 */

#ifndef TALTROCONVERTER_HH_
#define TALTROCONVERTER_HH_

#include <iostream>
#include <fstream>
#include <utility>
#include <string>
#include <vector>

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "IO/LCWriter.h"

//LCCD
#include "lccd.h"
#include "lccd/DBInterface.hh"

#include "TIAltroBlockFactory.hh"
#include "AltroConverterExceptions.hh"

#include "GenericADCElectronicsParameters.h"
#include "ExtendedAltroParameters.h"

namespace marlintpc
{

class TAltroConverter
{
public:
    /**
     * Default constructor.
     * The class will need further initialization of the input and output at least to work correctly.
     */
    TAltroConverter();

    /**
     * Standard usage constructor.
     * Initialize input and output references
     * @param input
     * @param output
     */
    TAltroConverter(std::vector<std::ifstream*>& input, lcio::LCWriter* output);

    /**
     * Copy constructor
     * @param other
     */
    TAltroConverter(const TAltroConverter& other);

    /**
     * Destructor
     */
    virtual ~TAltroConverter();

    /**
     * Assignment operator
     * @param other
     * @return reference to the resulting object to chain the copying
     */
    TAltroConverter& operator=(const TAltroConverter& other);

    /**
     * Sets the input streams to be converted.
     * The vector must contained the ordered list of filestreams and they should be already opened.
     * The sanity check on the filestream is left to the user.
     * @param input Vector containing the opened filestreams of the files to be read and converted
     */
    virtual void SetInput(std::vector<std::ifstream*>& input);

    /**
     * Sets a valid lcio output.
     * The LCWriter should be already initialized.
     * @param output The
     */
    virtual void SetOutput(lcio::LCWriter* output);

    /**
     * Retrieves a collection of GenericADCElectronicsParameters to be fed to file or database
     * @return Collection of GenericADCElectronicsParameters
     */
    virtual lcio::LCCollection* GetGenericElectronicsPar() const;

    /**
     * Sets the name to be used as collection name when storing the Generic Electronics Parameter Collection to DB
     * @param ColName Collection Name
     */
    virtual void SetGenericElectronicsParColName(std::string ColName = tpcconddata::GenericADCElectronicsParameters::getDefaultColName());

    /**
     * Retrieves a collection of LCGenericObject containing ALTRO specific information to be fed to the DB
     * @return Collection of LCGenericObjects
     */
    virtual lcio::LCCollection* GetExtendedElectronicsPar() const;

    /**
     * Sets the name to be used as collection name when storing the Extended Electronics Parameter Collection to DB
     * @param ColName Collection Name
     */
    virtual void SetExtendedElectronicParColName(std::string ColName = tpcconddata::ExtendedAltroParameters::getDefaultColName());

    /**
     * Feeds the condition data contained in the raw data files to the database
     * @param Database lccd Interface to the database
     * @param CollectionName Name of the collection to store in the database
     */
    virtual void ToDB(lccd::DBInterface* GenericInfo, lccd::DBInterface* ExtendedInfo) const;

    /**
     * Retrieves the timestamp of the data run start
     * @return 64 bit integer with the unix timestamp of the data start
     */
    virtual lccd::LCCDTimeStamp RunStart() const;

    /**
     * Retrieves the timestamp of the data run end
     * @return 64 bit integer with the unix timestamp of the data end
     */
    virtual lccd::LCCDTimeStamp RunEnd() const;

    /**
     * Check that the data stream is not corrupted
     * @return Result of the check. If true everything is ok
     */
    virtual bool CheckIntegrity() const noexcept(false);
    /**
     * Main function to execute the decoding of the file stream
     */
    virtual void Decode();

    /**
     * Prints the data information on an output stream provided by the user
     * @param os The output stream where to print the data
     */
    virtual void Print(std::ostream& os) const;

    /**
     * Gets the version of the data encoding as a pair of integers: daq version and dataformat
     * @return a pair of integers representing the encoding format
     */
    virtual std::pair<int, int> DataVersion() const;

    /**
     * Returns the version of the altro converter as a pair of integer: version, subversion
     * @return a pair of integer representing the version and subversion of the converter
     */
    virtual std::pair<int, int> ConverterVersion() const;

    /**
     * Brings back the converter to the same status as it was created by the default constructor.
     */
    virtual void Reset();

    /**
     * Set the detector name for the LCIO RunHeader and Event
     * @param Detector name
     */
    virtual void SetDetectorName(std::string Name);

    /**
     * Set the Run Description for the LCIO Run Header
     * @param Run Description
     */
    virtual void SetRunDescription(std::string Description);

    /**
     * Sets the name for the input collection
     * @param Name
     */
    virtual void SetDataCollectionName(std::string Name);

    /**
     * Overrides the value that is stored in the raw data for the zero suppression threshold.
     * This is necessary because there is no way to retrieve this information from the data
     * for DAQ version up to 4.5 and this value must be provide manually.
     * A negative value for this data member will trigger the usage of the data stored in the
     * file for version 4.6+ and will set the ZS Threshold to 0 for previous versions.
     * @param Threshold
     */
    virtual void SetZSThresholdOverride(int Threshold = -1);

private:
    /**
     * The inputfilestreams already opened and sorted from the first to the last file
     */
    std::vector<std::ifstream*> _input;

    /**
     * The LCIO file where to write the output
     */
    lcio::LCWriter* _output;

    /**
     * The name of the collection containing the generic electronics information
     */
    std::string _GenericParColName;

    /**
     * The name of the collection containing the extended electronics information
     */
    std::string _ExtendedParColName;

    /**
     * Timestamp of the start time of the run
     */
    lccd::LCCDTimeStamp _runStart;

    /**
     * Timestamp of the end time of the run
     */
    lccd::LCCDTimeStamp _runEnd;

    /**
     * A pair of integers representing the encoding format: DAQ Version, DataFormat
     */
    std::pair<int, int> _DataVersion;

    /**
     * Pointer to the Factory used to produce the actual data blocks to decode
     */
    TIAltroBlockFactory* _Factory;

    /**
     * Detector name for the event and run header
     */
    std::string _DetectorName;

    /**
     * Run description for the run header
     */
    std::string _RunDescription;

    /**
     * Storage name of the data collection
     */
    std::string _DataColName;

    /**
     * Collection of general electronics parameters (e.g. GenericADCElectronicsParameters)
     */
    lcio::LCCollection* _GenericElectronicsPar;

    /**
     * Collection of extended electronics inormation (e.g. ExtendedAltroParameters)
     */
    lcio::LCCollection* _ExtendedElectronicsPar;

    int _ZSThresholdOverride;
};

}//End namespace


#endif /* TALTROCONVERTER_HH_ */
