/*
 * TISORBlock.hh
 *
 *  Created on: Jul 9, 2012
 *      Author: caiazza
 */

#ifndef TISORBLOCK_HH_
#define TISORBLOCK_HH_

#include "TIAltroBlock.hh"

//LCIO
#include "lcio.h"
#include "EVENT/LCRunHeader.h"
#include "UTIL/LCTime.h"

namespace marlintpc
{
/**
 * Interface for an object representing the Start of Run Block of the Altro data format.
 * The implementations of this class will be version dependent
 */
class TISORBlock : public TIAltroBlock
{
public:
    /**
     * Virtual destructor to enable polymorphic behaviours
     */
    virtual ~TISORBlock(){}

    /**
     * Creates an LCIO Run Header based on the information contained in the SOR Block.
     * The additional informations that are stored in the SOR block produced by different
     * DAQ version will be stored in the Run Header parameters
     * @return A newly created LCRunHeader
     */
    virtual lcio::LCRunHeader* CreateRunHeader() const = 0;

    /**
     * Decodes the run number information
     * @returnThe run number
     */
    virtual uint32_t GetRunNumber() const = 0;

    /**
     * Decodes the timestamp of the start of the run and encodes it in an LCTime object
     * @return An LCTime object representing the run start timestamp
     */
    virtual lcio::LCTime GetRunStart() const = 0;

};// End interface

/**
 * Implementation header of the SOR Block for the daq version 4.2 to 4.5
 */
class TSORBlock_V42: public TISORBlock
{
public:
    TSORBlock_V42(std::vector<uint32_t>& Content);

    virtual ~TSORBlock_V42();

    /**
     * Get the block lenght in number of 32 bit words
     * @return number of 32 bit words in the block
     */
     virtual uint32_t GetLength() const;

    /**
     * Verifies that the content of the block matches the expected content of the type of block we build.
     * @return True if the content of the block matches the expectation, False otherwise
     */
     virtual bool CheckIntegrity() const;

    /**
     * Returns a constant reference to the content of the block
     * @return vector of 32 bit data words
     */
     virtual const std::vector<uint32_t>& Content() const;

    /**
     * Returns a data word that identifies the block type
     * @return 32 bit data word identifying the block
     */
     virtual uint32_t GetIdentifier() const;

    /**
     * Creates an LCIO Run Header based on the information contained in the SOR Block.
     * The additional informations that are stored in the SOR block produced by different
     * DAQ version will be stored in the Run Header parameters
     * @return A newly created LCRunHeader
     */
     virtual lcio::LCRunHeader* CreateRunHeader() const;

    /**
     * Decodes the run number information
     * @returnThe run number
     */
     virtual uint32_t GetRunNumber() const;

    /**
     * Decodes the timestamp of the start of the run and encodes it in an LCTime object
     * @return An LCTime object representing the run start timestamp
     */
     virtual lcio::LCTime GetRunStart() const;

     /**
      * Allows to set additional parameters to the RunHeader.
      * @param RunHeader
      */
     virtual void SetAdditionalParameters(lcio::LCRunHeader* RunHeader);
    /**
     * Allows to set the timestamp of the start of the run if different from what
     * is done in constructor
     */
    virtual void SetRunStart(lcio::LCTime time);

private:
    std::vector<uint32_t> _content;

    uint32_t _runNumber;

    lcio::LCTime _start;

    uint16_t _PCA16DAC;

    enum gainSettings {MV12 = 0, MV15 = 1, MV19 = 2, MV27 = 3};
    enum shaperSettings {NS30 = 7, NS60 = 3, NS90 = 1, NS120 = 0};

    struct PCA16Registers
    {
        bool _preampEnable;
        gainSettings _gain;
        shaperSettings _shaper;
        bool _shutdown;
        bool _polarity;
    } _registers;

    struct RunSettings
    {
        bool _zeroSuppression;
        bool _pedestalSubtraction;
    } _runSets;

    uint32_t _EUDAQRunNumber;
};//End implementation v4.2-4.5

/**
 * Implementation of a class decoding the SOR block for daq version 4.6.
 * In this version there are only some additional fields added so it's possible to simply
 * extend the previous version
 */
class TSORBlock_V46: public TSORBlock_V42
{
public:
    TSORBlock_V46(std::vector<uint32_t>& Content);

    virtual ~TSORBlock_V46();

    virtual bool CheckIntegrity() const;

    virtual lcio::LCRunHeader* CreateRunHeader() const;

private:
    enum TriggerSource{DATA = 1, DBOX = 2, DATA_DBOX=3};

    struct RunSettingsExt
    {
        short _overthreshold;
        short _threshold;
        short _postsamples;
        short _presamples;
        short _pretrigger;
        TriggerSource _triggerSource;
    }_runSetsExt;
}; //end implementation V4.6

/**
 * Implementation of a class decoding the SOR block for daq version 4.7 and higher.
 * In this version the timeformat has changed to unix time to allow consistency between event time stamp and 
 * times in the run header.
 */
class TSORBlock_V47: public TSORBlock_V46
{
public:
    TSORBlock_V47(std::vector<uint32_t>& Content);
  
    virtual ~TSORBlock_V47();

}; //end implementation V4.7

}


#endif /* TISORBLOCK_HH_ */
