#ifndef FIELDMAPREADER_H_
#define FIELDMAPREADER_H_

// stl
#include <fstream>
#include <string>

// LCIO
#include "EVENT/LCCollection.h"
#include "IMPL/LCCollectionVec.h"


namespace marlintpc {

/** FieldReader : A reader for field maps.
 *
 *   Takes a text file and produces an LCCollection from it.
 *   The text file has to provide the type of the field map:
 *   	"FieldType <FieldMapType>"
 *   Any number of Int, Float and String parameters for the
 *   LCCollection can be set:
 *     "<type> <parameterName> <value>"
 *    The actual field map starts after: "### Start Field Map ###"
 *    The values are read depending on the type of the
 *    field map.
 *
 *    @author Christian Grefe <br>
 */

class FieldMapReader {

public:
	/** Returns the default instance.
	 */
	static FieldMapReader* READER();

	/** destructor
	 */
	virtual ~FieldMapReader();

	/** reads the field map at the given path and returns it as an LCCollection*.
	 *   returns 0 if the input file is not a valid field map.
	 */
	virtual EVENT::LCCollection* read_field_map( const std::string );

	/** sets level of verbose output
	 */
	void set_verbose( const int );

protected:

	/** Reads all lines after "### Start Field Map ###".
	 *   Expects four floats in each line: "<r> <z> <Br> <Bz>"
	 *   Adds one element to the LCCollection per line.
	 */
	virtual void read_Cylindrical2DMagneticField();

	/** Reads all lines after "### Start Field Map ###".
	 *   Expects six floats in each line: "<r> <phi> <z> <Br> <Bphi> <Bz>"
	 *   Adds one element to the LCCollection per line.
	 */
	virtual void read_Cylindrical3DMagneticField();

	/** Reads a string, extracts a parameter and adds it to the parameters in the LCCollection.
	 *   The string has to follow the pattern: "<type> <parameterName> <value>".
	 *   Known types are "String", "Int" and "Float". Only the first value is taken, except for "String"
	 *   which takes everything after the parameter name.
	 *   "FieldType <FieldMapType>" sets the type of the field map. It has to be defined before
	 *   "### Start Field Map ###". Only the last definition is taken.
	 */
	virtual void read_parameter( const std::string );

private:
	/** constructor
	 *   only one instance of FieldReader is needed
	 */
	FieldMapReader();

	static FieldMapReader* defaultInstance;
	IMPL::LCCollectionVec* _field_map;
	std::ifstream* _file;
	std::string _field_type;
	int _verbose;
};

} // namespace marlintpc

#endif /* FIELDMAPREADER_H_ */
