#include "BFieldMapListener.h"

//C++
#include <utility>
#include <map>
#include <iostream>
#include <string.h>
#include <exception>
#include <sstream>

//lccd
#include <lccd.h>
#include <lccd/LCConditionsMgr.hh>
#include <lccd_exceptions.h>

//lcio
#include "lcio.h"
#include "EVENT/LCIO.h"
#include "UTIL/LCFixedObject.h"
#include "UTIL/LCTime.h"

//MarlinTPC
#include "TPCCondDataExeptions.h"

namespace marlintpc
{

    BFieldMapListener::BFieldMapListener(std::string collectionName, double cX, double cY, double cZ, double crotX, double crotY) :  _collectionName(collectionName)
    {
        _condHandler = NULL;
        _thisEmptyCollection = new LCCollectionVec(LCIO::LCGENERICOBJECT);
        _thisEmptyCollection->parameters().setValue("CollectionName", "EmptyCollectionBFieldMapListener");

        // try to register with the Marlin conditions processor
        if(!(marlin::ConditionsProcessor::registerChangeListener(this, collectionName)))
            {
                std::stringstream message;
                message << " BFieldMapListener couldn't register collection " << collectionName << " with the ConditionsProcessor.";
                throw lccd::LCCDException(message.str());
            }
        _theFieldMap = new Cylindrical2DMagneticFieldMap();
        _theFieldMap->init(cX, cY, cZ, crotX, crotY);
    }



    BFieldMapListener::~BFieldMapListener()
    {
        delete _thisEmptyCollection;
    }



    void BFieldMapListener::conditionsChanged(EVENT::LCCollection* col)
    {
        streamlog_out(DEBUG) << "BFieldMapListener::conditionsChanged()" << std::endl;

        streamlog_out(MESSAGE) << "BFieldMapListener::conditionsChanged(): new values available."  << std::endl;
        streamlog_out(DEBUG) << "BFieldMapListener::conditionsChanged(): CollectionName: " << col->getParameters().getStringVal("CollectionName") << std::endl;
        _theFieldMap->updateMagneticField(col);
    }



    void BFieldMapListener::registeredWithHandler(lccd::IConditionsHandler* condHandler)
    {
        _condHandler = condHandler;
        streamlog_out(MESSAGE) << "BFieldMapListener::registeredWithHandler(): registered: " << condHandler->name() << std::endl;
        streamlog_out(DEBUG) << "BFieldMapListener::registeredWithHandler(): register default collection:" << std::endl;
        // try to get the default collection
        try
            {
                LCCollection* col = condHandler->defaultCollection();

                if(! col)            // it will be null if none has so far been registered. So let's register ours
                    {
                        condHandler->registerDefaultCollection(_thisEmptyCollection);
                        streamlog_out(DEBUG) << "BFieldMapListener::registeredWithHandler(): default collection registered; address: " << _thisEmptyCollection << std::endl;
                        // and put in the map for this handler
                        _handlerDefaultCollectionMap[_thisEmptyCollection] = condHandler;
                    }
                //then the default handler was already registered, that's odd ... ;)
                else if(col == _thisEmptyCollection)
                    {
                        streamlog_out(DEBUG) << "BFieldMapListener::registeredWithHandler(): default collection is already set to thisEmptyCollection:" << std::endl;
                    }
                else                 // somebody has got there before us, let's see if we like the default
                    {
                        // if it has the same CollectionName, we can use it, no problem ...
                        if(col->getParameters().getStringVal("CollectionName") == "EmptyCollectionBFieldMapListener")
                            {
                                streamlog_out(DEBUG) << "BFieldMapListener::registeredWithHandler(): I like your default ;)" << std::endl;
                                _handlerDefaultCollectionMap[col] = condHandler;
                            }
                        // if not, an exception is thrown because things should be handled in a uniform standardized way
                        else
                            {
                                streamlog_out(ERROR) << "BFieldMapListener::registeredWithHandler(): a non-standard empty default collection has already been registered with the conditions handler" << std::endl;
                                throw lccd::InconsistencyException("BFieldMapListener::registeredWithHandler() found a default collection that differs from the standard one has already been registered with the handler");
                            }
                    }
            }
        // catch MemberNotImplemented exception, which happens, if a simple file is used as input
        catch(lccd::MemberNotImplementedException)
            {
                streamlog_out(MESSAGE) << "BFieldMapListener::registeredWithHandler: Could not register default collection. Seems simple file input is used." << std::endl;
            }
    }



    void BFieldMapListener::deRegisteredWithHandler(lccd::IConditionsHandler* condHandler)
    {
        streamlog_out(WARNING) << "BFieldMapListener::deRegisteredWithHandler() deregistered with:" << condHandler->name() << std::endl;
        _condHandler = NULL;
    }

}
