
// LCIO
#include <EVENT/LCIO.h>

// Marlin
#include <marlin/ConditionsProcessor.h>
#include <marlin/Exceptions.h>

// MarlinTPC
#include "Field.h"
#include "RelativePosition.h"
#include "GlobalFieldProcessor.h"
#ifdef USE_CGA
#include "CGAElectricField.h"
#include "CGAMagneticField.h"
#endif

#include <math.h>
#include <cassert>

using namespace EVENT;
using namespace lccd;
using namespace CLHEP;
using namespace std;
using namespace marlin;

namespace marlintpc {

// create one to register it with Marlin
GlobalFieldProcessor aGlobalFieldProcessor;


GlobalFieldProcessor *GlobalFieldProcessor::_instance = 0;


GlobalFieldProcessor::GlobalFieldProcessor():
	Processor("GlobalFieldProcessor"),
	_use_gear_bfield( false ) {

	#ifdef USE_CGA
	_use_cga_fields = false;
	#endif
		
	registerOptionalParameter(
		"UseGEARBField?",
		"Controls the source of the magnetic field. If this is set to true then it overrides any other magnetic field information.",
		_use_gear_bfield,
		_use_gear_bfield );

	#ifdef USE_CGA
	registerOptionalParameter(
		"UseCGAFields?",
		"Controls the source of the electromagnetic field data. If this is set to true then it attempts to use CGA (from Mokka). It overrides any other field information.",
		_use_cga_fields,
		_use_cga_fields );
	#endif // USE_CGA

	registerOptionalParameter(
		"MagneticFields",
		"An array of pairs of strings of the format: (field type, lcio collection) defining magnetic fields.",
		_input_magnetic_fields,
		std::vector< std::string >() );

	registerOptionalParameter(
		"ElectricFields",
		"An array of pairs of strings of the format: (field type, lcio collection) defining electric fields.",
		_input_electric_fields,
		std::vector< std::string >() );

	registerOptionalParameter(
		"MainCrossingAngle",
		"The main crossing angle of the ILC (mrads).",
		_main_crossing_angle,
		0. );

	registerOptionalParameter(
		"CoilPositionCollection",
		"Sets the lcio collection which holds the coil position",
		_coil_position_collection,
		std::string(""));
}


GlobalFieldProcessor::~GlobalFieldProcessor() {}


Processor *GlobalFieldProcessor::newProcessor() {

	return (&get_instance());
}


void GlobalFieldProcessor::init() {

	// FIXME determine if the steering file is consistant with the simulation
	if( parameterSet("MainCrossingAngle") ) {

		// (assumed that the Mokka parameter was stored with the lcio or GEAR file but I don't see it)
	}

	#ifdef USE_CGA
	// do a sanity check with CGA and create the fields
	if ( _use_cga_fields ) {

		if ( _use_gear_bfield ) {

			streamlog_out( ERROR ) << "The GlobalFieldProcessor has been instructed to use field information from CGA _AND_ GEAR. This is currently not supported so execution is aborted!";
			throw StopProcessingException( this );
		}

		if ( parameterSet("MagneticFields") ) {

			streamlog_out( WARNING ) << "The GlobalFieldProcessor has been instructed to use the magnetic field information supplied by CGA in addition to fields supplied by LCIO. Was this intentional? The LCIO fields are being ignored.";
		}

		if ( parameterSet("ElectricFields") ) {

			streamlog_out( WARNING ) << "The GlobalFieldProcessor has been instructed to use the electric field information supplied by CGA in addition to fields supplied by LCIO. Was this intentional? The LCIO fields are being ignored.";
		}

		return;
	}
	#endif // USE_CGA

	if( _use_gear_bfield ) {

		if( parameterSet("MagneticFields") ) {

			streamlog_out( WARNING ) << "The GlobalFieldProcessor has been instructed to use the magnetic field information supplied by GEAR in addition to fields supplied by LCIO. Was this intentional?";
		}

		Field *active_field = _field_by_typename["GEARMagneticField"]->new_field();

		_active_magnetic_fields.push_back( active_field );

		streamlog_out( MESSAGE4 ) << "Activating GEARMagneticField";
	}

	// parse the processor parameters and register the active magnetic fields
	if ( ! _use_gear_bfield ) {
		
		for( vector<string>::const_iterator input = _input_magnetic_fields.begin(); input != _input_magnetic_fields.end(); input++ ) {

			const string field_typename = (*input++);
			const string collection_name = (*input);

			// create the field
			if( _field_by_typename.find(field_typename) == _field_by_typename.end() ) {

				streamlog_out( WARNING ) << "Field of type: " << field_typename << " has not been registered with GlobalFieldProcessor." << endl;
				continue;
			}

			Field *active_field = _field_by_typename[field_typename]->new_field();

			_active_magnetic_fields.push_back( active_field );

			// now create the object which is responsible for telling the field when it has changed
			FieldNotifier *field_notifier = new FieldNotifier( *active_field );

			if( !ConditionsProcessor::registerChangeListener( field_notifier, collection_name ) ) {

				delete active_field;
				_active_magnetic_fields.pop_back();

				delete field_notifier;

				streamlog_out( WARNING ) << "Could not register collection " << collection_name << " with the LCCD conditions processor for the field " << field_typename << ". It has been disabled." << endl;
			}
			else {

				streamlog_out( MESSAGE4 ) << "Activating " << field_typename << " using collection " << collection_name << '\n';
				_field_notifiers.push_back( field_notifier );
			}
		}
	}

	// parse the processor parameters and register the active electric fields
	for( vector<string>::const_iterator input = _input_electric_fields.begin(); input != _input_electric_fields.end(); input++ ) {

		const string field_typename = (*input++);
		const string collection_name = (*input);

		// create the field
		if( _field_by_typename.find(field_typename) == _field_by_typename.end() ) {

			streamlog_out( WARNING ) << "Field of type: " << field_typename << " has not been registered with GlobalFieldProcessor." << endl;
			continue;
		}

		Field *active_field = _field_by_typename[field_typename]->new_field();

		_active_electric_fields.push_back( active_field );

		// now create the object which is responsible for telling the field when it has changed
		FieldNotifier *field_notifier = new FieldNotifier( *active_field );

		if( !ConditionsProcessor::registerChangeListener( field_notifier, collection_name ) ) {

			delete active_field;
			_active_electric_fields.pop_back();

			delete field_notifier;

			streamlog_out( WARNING ) << "Could not register collection " << collection_name << " with the LCCD conditions processor for the field " << field_typename << ". It has been disabled." << endl;
		}
		else {

			streamlog_out( MESSAGE4 ) << "Activating " << field_typename << " using collection " << collection_name << '\n';
			_field_notifiers.push_back( field_notifier );
		}
	}

	// register the coil position collection
	if( _coil_position_collection != "" ) {

		_coil_position_notifier = new CoilPositionNotifier( _active_magnetic_fields );

		if ( !ConditionsProcessor::registerChangeListener( _coil_position_notifier, _coil_position_collection ) ) {

			delete _coil_position_notifier;

			streamlog_out( WARNING ) << "Could not register collection " << _coil_position_collection << " with the LCCD conditions processor." << endl;
		}
		else {

			streamlog_out( MESSAGE4 ) << "Setting coil position using collection " << _coil_position_collection << endl;
		}
	}

}


void GlobalFieldProcessor::processRunHeader(lcio::LCRunHeader*  /*run*/ ) {

	#ifdef USE_CGA
	// initialization of CGA needs to be done with something which has access to the run header... that is why it is done here and not in the regular init() function
	if ( _use_cga_fields ) {

		CGAElectricField *cga_electric_field = new CGAElectricField( run );
		_active_electric_fields.push_back( cga_electric_field );

		CGAMagneticField *cga_magnetic_field = new CGAMagneticField( run );
		_active_magnetic_fields.push_back( cga_magnetic_field );
	}
	#endif // USE_CGA
}


void GlobalFieldProcessor::processEvent(lcio::LCEvent *  /*evt*/ ) {}


void GlobalFieldProcessor::check(lcio::LCEvent *  /*evt*/ ) {}


void GlobalFieldProcessor::end() {

	for( std::vector<FieldNotifier*>::iterator field_notifier = _field_notifiers.begin(); field_notifier != _field_notifiers.end(); field_notifier++ ) {

		if( *field_notifier ) {

			delete *field_notifier;
		}
	}

	for( std::vector<Field*>::iterator field = _active_magnetic_fields.begin(); field != _active_magnetic_fields.end(); field++ ) {

		if( *field ) {

			delete *field;
		}
	}

	for( std::vector<Field*>::iterator field = _active_electric_fields.begin(); field != _active_electric_fields.end(); field++ ) {

		if( *field ) {

			delete *field;
		}
	}

//	if ( _coil_position_notifier ) {
//
//		delete _coil_position_notifier;
//	}

}


/** registers a field with the field manager
*/
bool GlobalFieldProcessor::register_field( Field* field ) {

	const string field_typename = field->get_typename();

	if( _field_by_typename.find(field_typename) == _field_by_typename.end() ) {

		_field_by_typename[field_typename] = field;
	}

	return true;
}


/** Returns the magnetic field (in T) at a given point and time in the detector
*/
CLHEP::HepVector GlobalFieldProcessor::get_global_magnetic_field( const CLHEP::HepVector& x ) const {

	// convert to Hep3Vector and get the field value
	CLHEP::Hep3Vector field_vec = get_global_magnetic_field( CLHEP::Hep3Vector(x[0], x[1], x[2]) );

	// convert the field value back into a HepVector
	CLHEP::HepVector total_field = CLHEP::HepVector(3);
	total_field[0] = field_vec.x();
	total_field[1] = field_vec.y();
	total_field[2] = field_vec.z();

	return total_field;
}

/** Returns the magnetic field (in T) at a given point and time in the detector
*/
CLHEP::Hep3Vector GlobalFieldProcessor::get_global_magnetic_field( const CLHEP::Hep3Vector& x ) const {

	Hep3Vector total_field = Hep3Vector();

	for( std::vector<Field*>::const_iterator field = _active_magnetic_fields.begin(); field != _active_magnetic_fields.end(); field++ ) {

		total_field += (*field)->get_field_value( x );
	}

	return total_field;
}

/** Returns the combined symmetric field errors (in T) of measured magnetic field maps
 */
CLHEP::Hep3Vector GlobalFieldProcessor::get_global_magnetic_field_error( ) const {

	double errX = 0.;
	double errY = 0.;
	double errZ = 0.;

	Hep3Vector total_field_error_sq = Hep3Vector();

	for ( std::vector<Field*>::const_iterator field = _active_magnetic_fields.begin(); field != _active_magnetic_fields.end(); field++) {

		Hep3Vector error = (*field)->get_field_error();

		errX += pow(error.x(),2);
		errY += pow(error.y(),2);
		errZ += pow(error.z(),2);
	}

	return Hep3Vector( sqrt(errX), sqrt(errY), sqrt(errZ) );
}


/** Returns the electric field (in V) at a given point and time in the detector
*/
CLHEP::HepVector GlobalFieldProcessor::get_global_electric_field( const CLHEP::HepVector& x ) const {

	// convert to Hep3Vector and get the field value
	CLHEP::Hep3Vector field_vec = get_global_electric_field( CLHEP::Hep3Vector(x[0], x[1], x[2]) );

	// convert the field value back into a HepVector
	CLHEP::HepVector total_field = CLHEP::HepVector(3);
	total_field[0] = field_vec.x();
	total_field[1] = field_vec.y();
	total_field[2] = field_vec.z();

	return total_field;
}

/** Returns the electric field (in V) at a given point and time in the detector
*/
CLHEP::Hep3Vector GlobalFieldProcessor::get_global_electric_field( const CLHEP::Hep3Vector& x ) const {

	Hep3Vector total_field = Hep3Vector();
	
	for( std::vector<Field*>::const_iterator field = _active_electric_fields.begin(); field != _active_electric_fields.end(); field++ ) {

		total_field += (*field)->get_field_value( x );
	}

	return total_field;
}

/** Returns the combined symmetric field errors (in T) of measured electric field maps
 */
CLHEP::Hep3Vector GlobalFieldProcessor::get_global_electric_field_error( ) const {

	double errX = 0.;
	double errY = 0.;
	double errZ = 0.;

	Hep3Vector total_field_error_sq = Hep3Vector();

	for ( std::vector<Field*>::const_iterator field = _active_electric_fields.begin(); field != _active_electric_fields.end(); field++) {

		Hep3Vector error = (*field)->get_field_error();

		errX += pow(error.x(),2);
		errY += pow(error.y(),2);
		errZ += pow(error.z(),2);
	}

	return Hep3Vector( sqrt(errX), sqrt(errY), sqrt(errZ) );
}


/** returns the main crossing angle of the ILC model
*/
double GlobalFieldProcessor::get_main_crossing_angle() const {

	return _main_crossing_angle;
}


/** return the instance of the global field processor
	*/
GlobalFieldProcessor &GlobalFieldProcessor::get_instance() {

	if( !_instance ) {

		_instance = new GlobalFieldProcessor();
	}

	return (*_instance);
}


GlobalFieldProcessor::FieldNotifier::FieldNotifier( Field& field ):
	_field(field) {

}


/** from IConditionsChangeListener. this function is called whenever a condition has changed in this collection
*/
void GlobalFieldProcessor::FieldNotifier::conditionsChanged( lcio::LCCollection* col ) {

	_field.construct( col );
}


GlobalFieldProcessor::CoilPositionNotifier::CoilPositionNotifier( std::vector<Field*> active_magnetic_fields ) {

	for( std::vector<Field*>::const_iterator field = active_magnetic_fields.begin(); field != active_magnetic_fields.end(); field++) {

		// only use fields of type MagneticField
		MagneticField* magField = dynamic_cast<MagneticField*> (*field);
		if (magField != 0) _magnetic_fields.push_back(magField);
	}
}


/** from IConditionsChangeListener. this function is called whenever a condition has changed in this collection
*/
void GlobalFieldProcessor::CoilPositionNotifier::conditionsChanged( lcio::LCCollection* col ) {

	RelativePosition* coilPosition = new RelativePosition( col->getElementAt(0) );

	for ( std::vector<MagneticField*>::iterator field = _magnetic_fields.begin(); field != _magnetic_fields.end(); field++) {

		(*field)->set_translation( coilPosition->get_position() );
		(*field)->set_rotation( coilPosition->get_x_rotation(), coilPosition->get_y_rotation() );
	}
}


}	// namespace marlintpc
