// STL
#include <vector>
#include <algorithm>
#include <cmath>
#include <iomanip>
#include <cassert>

// LCIO
#include "EVENT/LCIO.h"
#include "EVENT/LCCollection.h"
#include "EVENT/SimTrackerHit.h"
#include "EVENT/TrackerHit.h"
#include "EVENT/TrackerData.h"
#include "EVENT/TrackerPulse.h"
#include "IMPL/LCCollectionVec.h"
#include "IMPL/TrackImpl.h"

// GEAR
#include <gear/GEAR.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>
#include <gearimpl/Vectors.h>
#include <gear/TPCModule.h>

// CLHEP
#include <CLHEP/Matrix/Matrix.h>
#include <CLHEP/Vector/LorentzVector.h>
#include <CLHEP/Vector/TwoVector.h>
#include <CLHEP/Geometry/Transform3D.h>

// Marlin
#include <marlin/Global.h>
#include <HepRepXMLWriter.h>
#include <HepRepOutputProcessor.h>

//ROOT
#include <TMath.h>

using namespace std;
using namespace lcio;
using namespace marlin;

HepRepOutputProcessor aHepRepOutputProcessor;

HepRepOutputProcessor::HepRepOutputProcessor() :
        Processor ("HepRepOutputProcessor")
{
    //_input_tracker_hits_collection_name("TPCHits"),
    //_output_seed_track_collection_name("TPCSeedTracks") {

    _description = "Finds a seed track in the input TrackerHit collection!";

    registerOptionalParameter ("SimTrackerHitCollections",
                               "A list of SimTrackerHit collections to draw",
                               _namesOfSimTrackerHitCollections,
                               _namesOfSimTrackerHitCollections);

    registerOptionalParameter ("TrackerDataCollections",
                               "A list of TrackerPulse collections to draw",
                               _namesOfTrackerDataCollections,
                               _namesOfTrackerDataCollections);

    registerOptionalParameter ("TrackerHitCollections",
                               "A list of TrackerHit collections to draw.",
                               _namesOfTrackerHitCollections,
                               _namesOfTrackerHitCollections);

    registerOptionalParameter ("TrackerPulseCollections",
                               "A list of TrackerPulse collections to draw",
                               _namesOfTrackerPulseCollections,
                               _namesOfTrackerPulseCollections);
//                               std::string (""));

    registerOptionalParameter ("TrackCollections",
                               "A list of Track collections to draw",
                               _namesOfTrackCollections,
                               _namesOfTrackCollections);

    registerOptionalParameter ("EventNumbersToDraw",
                               "A list of event numbers to display",
                               _eventsToDisplay,
                               _eventsToDisplay);
    
    registerProcessorParameter ("OutputFileName",
                               "Name of the HepRep output file.",
                               _nameOfOutputFile,
                               std::string ("HeprepOutput.heprep"));
	
	registerProcessorParameter ("PulseColorScheme" ,
	                            "0: RAINBOW; 1: REDBLUE; 2: TOPOGRAPHIC; 3: GREYSCALE  (default: 0)"  ,
	                            _pulseColorScheme,
	                            int (0));

    registerProcessorParameter ("DrawPadPlane",
                               "Flag whether to draw the pad plane",
                               _drawPadPlane,
				bool(true) );
}


HepRepOutputProcessor::~HepRepOutputProcessor()
{


}


void HepRepOutputProcessor::init()
{

    printParameters();

	stringstream name;

    _hepRepWriter.open (_nameOfOutputFile.c_str());

	// Draw the axis cross at the origin
	
	_hepRepWriter.addType ("CoordinateSystem", 0);
	_hepRepWriter.addInstance();

	
	_hepRepWriter.addType ("Axis", 0);
	_hepRepWriter.addAttValue ("DrawAs", "Line");
    _hepRepWriter.addAttValue ("LineWidth", "1");

	_hepRepWriter.addInstance();
	
	_hepRepWriter.addPrimitive();
	_hepRepWriter.addAttValue ("Axis", "X");
	_hepRepWriter.addPoint (0 , 0 , 0);
	_hepRepWriter.addPoint (10, 0 , 0);

	_hepRepWriter.addPrimitive();
	_hepRepWriter.addAttValue ("Axis", "Y");
	_hepRepWriter.addPoint (0 , 0 , 0);
	_hepRepWriter.addPoint (0, 10 , 0);

	_hepRepWriter.addPrimitive();
	_hepRepWriter.addAttValue ("Axis", "Z");
	_hepRepWriter.addPoint (0 , 0 , 0);
	_hepRepWriter.addPoint (0, 0 , 10);


	// Draw axis labels movable in space
	_hepRepWriter.addType ("Letter", 0);
	_hepRepWriter.addAttValue ("DrawAs", "Line");

	_hepRepWriter.addInstance();
	
	_hepRepWriter.addPrimitive();
	_hepRepWriter.addAttValue ("Letter", "X");
	_hepRepWriter.addPoint (10 , 1 , 0);
	_hepRepWriter.addPoint (8, 4 , 0);
	_hepRepWriter.addPrimitive();
	_hepRepWriter.addPoint (10 , 4 , 0);
	_hepRepWriter.addPoint (8, 1 , 0);

	_hepRepWriter.addPrimitive();
	_hepRepWriter.addAttValue ("Letter", "Y");
	_hepRepWriter.addPoint (2 , 7 , 0);
	_hepRepWriter.addPoint (2, 8.5 , 0);
	_hepRepWriter.addPoint (1, 10 , 0);
	_hepRepWriter.addPrimitive();
	_hepRepWriter.addPoint (2 , 8.5 , 0);
	_hepRepWriter.addPoint (3, 10 , 0);

	_hepRepWriter.addPrimitive();
	_hepRepWriter.addAttValue ("Letter", "Z");
	_hepRepWriter.addPoint (0 , 4 , 10);
	_hepRepWriter.addPoint (0, 4 , 8);
	_hepRepWriter.addPoint (0, 1 , 10);
	_hepRepWriter.addPoint (0, 1 , 8);
	
	// Draw origin
	_hepRepWriter.addType ("Origin", 0);
        _hepRepWriter.addAttValue ("DrawAs", "Point");
	_hepRepWriter.addInstance();
	_hepRepWriter.addPrimitive();
    _hepRepWriter.addAttValue ("MarkName", "Circle");
    _hepRepWriter.addAttValue ("MarkSize", 10.);
	_hepRepWriter.addPoint (0 , 0 , 0);


    try
    {

        const gear::TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();

        const double max_drift_length = tpc_parameters.getMaxDriftLength();

        // DRAW THE TPC

        // try and get some extra parameters from MOKKA_GEAR
        const double tpc_outer_radius = tpc_parameters.getDoubleVal ("TPC_outer_radius");  // mm

        // try to get an additional offset of the TPC cylinder
        double cylinder_centre_c0 = 0;// mm or r
        double cylinder_centre_c1 = 0;// mm or phi

        try
        {
            cylinder_centre_c0 = tpc_parameters.getDoubleVal ("TPC_cylinder_centre_c0");  // mm or r
        }
        catch (...) {}
        try
        {
            cylinder_centre_c1 = tpc_parameters.getDoubleVal ("TPC_cylinder_centre_c1");  // mm or phi
        }
        catch (...) {}

        double  cylinder_centre_x, cylinder_centre_y;

        switch (tpc_parameters.getCoordinateType())
        {
            case gear::PadRowLayout2D::POLAR :
                cylinder_centre_x = cylinder_centre_c0 * cos (cylinder_centre_c1);
                cylinder_centre_y = cylinder_centre_c0 * sin (cylinder_centre_c1);
                break;
            case gear::PadRowLayout2D::CARTESIAN :
                cylinder_centre_x = cylinder_centre_c0;
                cylinder_centre_y = cylinder_centre_c1;
                break;
            default:
                throw Exception ("HepRepOutputProcessor: unknown geometry type in TPC layout");
        }

        _hepRepWriter.addType ("TPC_OuterCylinder", 0);
        _hepRepWriter.addAttValue ("DrawAs", "Cylinder");
        _hepRepWriter.addAttValue ("Faces", 20);

        _hepRepWriter.addInstance();
        _hepRepWriter.addAttValue ("Radius1", tpc_outer_radius);  //PETER
        _hepRepWriter.addAttValue ("Radius2", tpc_outer_radius);
        _hepRepWriter.addPrimitive();
        //  _hepRepWriter.addPoint( cylinder_centre_x, cylinder_centre_y, -1* max_drift_length );
        // assume we have a half TPC in prototype case
        _hepRepWriter.addPoint (cylinder_centre_x, cylinder_centre_y, 0.);
        _hepRepWriter.addPoint (cylinder_centre_x, cylinder_centre_y,  max_drift_length);

        const double tpc_inner_radius = tpc_parameters.getDoubleVal ("TPC_inner_radius");  // mm

        // if there is a inner field cage add the seocnd half of the outer field cage
        _hepRepWriter.addType ("TPC_OuterCylinder2", 0);
        _hepRepWriter.addAttValue ("DrawAs", "Cylinder");
        _hepRepWriter.addAttValue ("Faces", 20);

        _hepRepWriter.addInstance();
        _hepRepWriter.addAttValue ("Radius", tpc_outer_radius);
        _hepRepWriter.addPrimitive();
        //  _hepRepWriter.addPoint( cylinder_centre_x, cylinder_centre_y, -1* max_drift_length );
        // assume we have a half TPC in prototype case
        _hepRepWriter.addPoint (cylinder_centre_x, cylinder_centre_y, -1.* max_drift_length);
        _hepRepWriter.addPoint (cylinder_centre_x, cylinder_centre_y,  0.);

        _hepRepWriter.addType ("TPC_InnerCylinder", 0);
        _hepRepWriter.addAttValue ("DrawAs", "Cylinder");
        _hepRepWriter.addAttValue ("Faces", 20);

        _hepRepWriter.addInstance();
        _hepRepWriter.addAttValue ("Radius1", tpc_inner_radius); //PETER
        _hepRepWriter.addAttValue ("Radius2", tpc_inner_radius);
        _hepRepWriter.addPrimitive();
        _hepRepWriter.addPoint (cylinder_centre_x, cylinder_centre_y, -1 * max_drift_length);
        _hepRepWriter.addPoint (cylinder_centre_x, cylinder_centre_y,  max_drift_length);
    
		

	}
    catch (...)
    {

    }
    try
    {
        // DRAW THE PAD PLANE ...

        // For convenience shorten the definition of the module vector
        typedef const std::vector<gear::TPCModule*> GearModuleVector;

        // Some preparation ...
        const gear::TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();

        GearModuleVector &modules = tpc_parameters.getModules();

	if (_drawPadPlane)
	{
	  // Iterate through all modules of the GEAR file ...
	  for (GearModuleVector::const_iterator module = modules.begin();
	       module < modules.end(); module++)
	    {
	      // Prepare HepRep entry for module and its pad plane ...
	      int moduleID = (**module).getModuleID();
	      const int COUNTING_OFFSET = 1;
	      
	      ostringstream moduleName;
	      moduleName << "Module" << moduleID + COUNTING_OFFSET;
	      
	      ostringstream padPlaneName;
	      padPlaneName << "PadPlane" << moduleID + COUNTING_OFFSET;
	      
	      // HepRep entry for module
	      _hepRepWriter.addType (moduleName.str().c_str(), 0);
	      _hepRepWriter.addInstance();
	      
	      // Hep Rep Entry for pad plane consisting of its pads ...
	      _hepRepWriter.addType (padPlaneName.str().c_str(), 1);
	      _hepRepWriter.addInstance();
	      
	      _hepRepWriter.addType ("Pads", 2);
	      _hepRepWriter.addAttValue ("DrawAs", "Polygon");
	      _hepRepWriter.addAttValue ("LineWidth", "1");
	      _hepRepWriter.addAttValue ("Fill", "False");
	      _hepRepWriter.addAttValue ("FillStyle", "NonZero");
	      _hepRepWriter.addAttValue ("ModuleID", moduleID);
	      
	      
	      // Test for correct pad shape ...
	      if ( (**module).getPadShape() != gear::PadRowLayout2D::RECTANGLE)
	      {
		throw Exception ("HepRepOutputProcessor only supports RECTANGULAR shaped pads at the moment.");
	      }
	      
	      const int numberOfRowsInModule = (**module).getNRows();
	      
	      // Iterate through the pad plane by row and pad ...
	      for (int rowNumber = 0; rowNumber < numberOfRowsInModule; rowNumber++)
		{
		  typedef const std::vector<int> GearPadVector;
		  GearPadVector &padsInRow = (**module).getPadsInRow (rowNumber);
		  
		  for (GearPadVector::const_iterator itPad = padsInRow.begin(); itPad < padsInRow.end(); itPad++)
		    {
		      const int padNumber = (**module).getPadNumber (*itPad);
		      
		      _hepRepWriter.addInstance();
		      _hepRepWriter.addAttValue ("Row", rowNumber);
		      _hepRepWriter.addAttValue ("Pad", padNumber);
		      
		      const int padIndex = (**module).getPadIndex (rowNumber, padNumber);
		      
		      _hepRepWriter.addAttValue ("PadIndex", padIndex);
		      
		      // Same procedure is used for drawing hits, so dedicated function makes sense.
		      drawHepRepPad (moduleID, padIndex);
		      //if(moduleID == 1) return;
		  	}
		}
	      //TODO: add sensitive volumes of modules here.
	    }
	}// if (_drawPadfPlane)
    }
    catch (DataNotAvailableException &e)
    {

        streamlog_out (WARNING) << e.what() << std::endl;
    }
}


void HepRepOutputProcessor::processRunHeader (LCRunHeader *run)
{

	// Revision logging ...
	run->parameters().setValue (_processorName + "_revision", "$Rev$");

	// Parameter logging ...
	for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
	{
		if (! i->second->isOptional() || i->second->valueSet())
		{
			run->parameters().setValue (_processorName + "_" + i->second->name(), i->second->value());
		}
	}
	
    _hepRepWriter.addType ("Run", 0);

    _hepRepWriter.addAttDef ("Run Number", "The run number", "Physics", "");

    _hepRepWriter.addInstance();
	


}


void HepRepOutputProcessor::processEvent (LCEvent *evt)
{

    if (_eventsToDisplay.size() > 0)
    {

        if (find (_eventsToDisplay.begin(), _eventsToDisplay.end(), evt->getEventNumber()) == _eventsToDisplay.end())
        {
            return;
        }
    }

	//Save each event to a unique layer in the single heprep output file
	stringstream eventName;

	eventName << "Event_" << evt->getEventNumber();

    _hepRepWriter.addType (eventName.str().c_str(), 0);

    _hepRepWriter.addInstance();

    try
    {

        /* DRAW THE HIT COLLECTIONS */


        std::vector< std::string >::const_iterator hit_collection_name = _namesOfTrackerHitCollections.begin();

        _hepRepWriter.addType ("TPC_TrackerHit", 1);
        _hepRepWriter.addAttValue ("Color", 1., 0.7, 0.);
        _hepRepWriter.addAttValue ("DrawAs", "Point");
        _hepRepWriter.addAttValue ("MarkName", "Circle");  //This is where you can change the hit shape. --PETER
        _hepRepWriter.addAttValue ("MarkSize", 4.);

        for (; hit_collection_name != _namesOfTrackerHitCollections.end(); hit_collection_name++)
        {

            EVENT::LCCollection* tracker_hits = evt->getCollection (*hit_collection_name);

            streamlog_out (MESSAGE3) << "Drawing TrackerHit collection " << *hit_collection_name << ".\n";

            for (int hit_counter = 0; hit_counter < tracker_hits->getNumberOfElements(); hit_counter++)
            {

                TrackerHit* current_hit = dynamic_cast<TrackerHit*> (tracker_hits->getElementAt (hit_counter));

                assert (current_hit);

                const double* position = current_hit->getPosition();
				
				if(position[0] == position[0] && position[1] == position[1] && position[2] == position[2])
				{
					_hepRepWriter.addInstance();
					_hepRepWriter.addPrimitive();
					_hepRepWriter.addPoint (position[0], position[1], position[2]);
					streamlog_out(DEBUG1) << "x = " << position[1] << endl;
				}
            }
        }


    }
    catch (DataNotAvailableException &e)
    {

        streamlog_out (WARNING) << e.what() << std::endl;
    }


    try
    {

        /* DRAW THE SIMTRACKERHIT COLLECTIONS */ //Uncomment if you have simulated hits to draw.
/*
        std::vector< std::string >::const_iterator sim_hit_collection_name = _namesOfSimTrackerHitCollections.begin();

        _hepRepWriter.addType ("TPC_SimTrackerHit", 0);
        _hepRepWriter.addAttDef ("dEdx", "Energy Desposit", "Physics", "");
        _hepRepWriter.addAttValue ("Color", 0., 1., 0.);
        _hepRepWriter.addAttValue ("DrawAs", "Point");
        _hepRepWriter.addAttValue ("MarkName", "Circle");
        _hepRepWriter.addAttValue ("MarkSize", 2.);

        for (; sim_hit_collection_name != _namesOfSimTrackerHitCollections.end(); sim_hit_collection_name++)
        {

            EVENT::LCCollection* tracker_hits = evt->getCollection (*sim_hit_collection_name);

            streamlog_out (MESSAGE3) << "Drawing SimTrackerHit collection " << *sim_hit_collection_name << ".\n";

            for (int hit_counter = 0; hit_counter < tracker_hits->getNumberOfElements(); hit_counter++)
            {

                SimTrackerHit* current_hit = dynamic_cast<SimTrackerHit*> (tracker_hits->getElementAt (hit_counter));

                assert (current_hit);

                const double* position = current_hit->getPosition();

                streamlog_out (DEBUG2) << "\tHit: " << hit_counter << " (x,y,z): " << position[0] << ", " << position[1] << ", " << position[2] << endl;

                _hepRepWriter.addInstance();
                _hepRepWriter.addPrimitive();
                _hepRepWriter.addPoint (position[0], position[1], position[2]);
            }
        }
*/
    }
    catch (DataNotAvailableException &e)
    {

        streamlog_out (WARNING) << e.what() << std::endl;
    }
	_hepRepWriter.addType ("TPC_TrackerPulse", 1);
	//_hepRepWriter.addAttDef ("TrackerPulse", "Pulse on pad", "Physics", "charge");
	_hepRepWriter.addAttDef ("PulseCharge", "Charge of pulse", "Physics", "");
	_hepRepWriter.addAttDef ("RowID", "Row # of pulse", "Physics", "");
	_hepRepWriter.addAttDef ("ColID", "Col # of pulse", "Physics", "");
	_hepRepWriter.addAttValue ("DrawAs", "Polygon");
	_hepRepWriter.addAttValue ("Fill", "True");
	_hepRepWriter.addAttValue ("FillStyle", "NonZero");

	const gear::TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();

	/* DRAW THE PULSE COLLECTIONS */

	for(std::vector<std::string>::const_iterator colIt = _namesOfTrackerPulseCollections.begin();
			colIt != _namesOfTrackerPulseCollections.end(); ++colIt)
	{

		try
		{

			EVENT::LCCollection* tracker_pulses = evt->getCollection (*colIt);


			streamlog_out (MESSAGE3) << "Representing " << tracker_pulses->getNumberOfElements() << " " << tracker_pulses->getTypeName() << " from collection " << *colIt << ".\n";


			// determine the largest charge on the pads
			float maxCharge = 0.;
			float minCharge = 0.;
			for (int pulse_counter = 0; pulse_counter < tracker_pulses->getNumberOfElements(); pulse_counter++)
			{

				TrackerPulse* current_pulse = dynamic_cast<TrackerPulse*> (tracker_pulses->getElementAt (pulse_counter));

				assert (current_pulse);

				minCharge = std::min (current_pulse->getCharge(), minCharge);
				maxCharge = std::max (current_pulse->getCharge(), maxCharge);
			}


			for (int pulse_counter = 0; pulse_counter < tracker_pulses->getNumberOfElements(); pulse_counter++)
			{

				TrackerPulse* current_pulse = dynamic_cast<TrackerPulse*> (tracker_pulses->getElementAt (pulse_counter));

				assert (current_pulse);


				_hepRepWriter.addInstance();

				const int padIndex = current_pulse->getCellID0();
				const int moduleID =  current_pulse->getCellID1();

				const gear::TPCModule& module =  tpc_parameters.getModule (moduleID);

				const int rowID = module.getRowNumber(padIndex);
				const int colID = module.getPadNumber(padIndex);

				const double padCharge = current_pulse->getCharge();
				const int pulseQuality = current_pulse->getQuality();

				_hepRepWriter.addAttValue ("PulseCharge", padCharge);
				_hepRepWriter.addAttValue ("RowID", rowID);
				_hepRepWriter.addAttValue ("ColID", colID);

				setHepRepPadColor (padCharge, minCharge, maxCharge, _pulseColorScheme, pulseQuality);
				drawHepRepPad (moduleID, padIndex);

			}
		}

		catch (DataNotAvailableException &e)
		{

			streamlog_out (WARNING) << e.what() << std::endl;
		}
	}


	/* DRAW THE TRACK COLLECTIONS */

	for(std::vector<std::string>::const_iterator nameIt = _namesOfTrackCollections.begin();
			nameIt != _namesOfTrackCollections.end(); ++nameIt)
	{

		EVENT::LCCollection* tracker_tracks = NULL;

		try
		{


			tracker_tracks = evt->getCollection (*nameIt);

		}

		catch (DataNotAvailableException &e)
		{

			streamlog_out (WARNING) << e.what() << std::endl;
			continue;
		}

		_hepRepWriter.addType (nameIt->c_str(), 1); //TEST
		_hepRepWriter.addAttValue ("DrawAs", "Line");
		_hepRepWriter.addAttValue ("LineColor", 1., 0.647, 0.);
		_hepRepWriter.addAttDef ("d0", "Spatial Distance from Reference Point", "Physics", "");
		_hepRepWriter.addAttDef ("phi", "Angle at Reference Point", "Physics", "");
		_hepRepWriter.addAttDef ("omega", "Curvature", "Physics", "");
		_hepRepWriter.addAttDef ("z0", "Distance in Z from Reference Point", "Physics", "");
		_hepRepWriter.addAttDef ("tanLambda", "TanLambda", "Physics", "");

		streamlog_out (MESSAGE3) << "Drawing collection " << *nameIt << ".\n";

		for (int track_counter = 0; track_counter < tracker_tracks->getNumberOfElements(); track_counter++)
		{

			//TMP
			streamlog_out (MESSAGE3) << "Drawing track " << track_counter << " of " << tracker_tracks->getNumberOfElements() << endl;

			Track* current_track = dynamic_cast<Track*> (tracker_tracks->getElementAt (track_counter));

			assert (current_track);

			const double d0 = current_track->getD0();
			const double phi0 = current_track->getPhi();
			const double z0 = current_track->getZ0();
			const double omega = current_track->getOmega();
			const double tanLambda = current_track->getTanLambda();
			const float * refPt = current_track->getReferencePoint(); //Peter

			const CLHEP::Hep3Vector n_pca (-1 * sin (phi0), cos (phi0));
			const CLHEP::Hep3Vector ref (refPt[0], refPt[1]);
			const CLHEP::Hep3Vector pca = ref + n_pca * d0;

			_hepRepWriter.addInstance();
			_hepRepWriter.addAttValue ("d0", d0);
			_hepRepWriter.addAttValue ("phi", phi0);
			_hepRepWriter.addAttValue ("omega", omega);
			_hepRepWriter.addAttValue ("z0", z0);
			_hepRepWriter.addAttValue ("tanLambda", tanLambda);

			_hepRepWriter.addPrimitive(); //Trying to get this to draw. --PETER

			// draw 5m of track
			if (fabs (omega) >= 0.0000001)
			{
				//				const double xC = refPt[0] + (1./omega - d0)*sin(phi0);
				//				const double yC = refPt[1] - (1./omega - d0)*cos(phi0);

				const CLHEP::Hep3Vector helix_centre = ref - (1. / omega - d0) * n_pca;
				//const double angular_distance = 5000 * omega;

				//			  for (double phi = 0.; phi <= angular_distance; phi += (M_PI / 180))
				for (double phi = -3*M_PI/4.; phi <= M_PI/4.; phi += (M_PI / 720.)/32.) //This ensures a good track will be plotted regardless of the omega. --Peter
				{
					//				  const double angle = (omega < 0) ? phi0 + phi : phi0 - phi;

					//				  double x = cos(angle)/fabs(omega) + xC; //Trying a direct approach.
					//				  double y = sin(angle)/fabs(omega) + yC; //Trying a direct approach.
					//double z = refPt[2]; //There should be more here, but I know it's at the end plate, so I'm not writing it.
					//				  double z = z0 + refPt[2] + (angle + phi0) * tanLambda / fabs(omega); 

					const double angle = (omega < 0) ? phi0 + phi : phi0 - phi;
					//				  const CLHEP::Hep3Vector n1 (-1 * sin (phi + phi0), cos (angle));
					//				  const CLHEP::Hep3Vector p1 = helix_centre + (1 / omega - d0) * n1;
					const CLHEP::Hep3Vector n1 (-1 * sin (angle), cos (angle));
					const CLHEP::Hep3Vector p1 = helix_centre + (1./omega) * n1;

					_hepRepWriter.addPoint (p1.x(), p1.y(), refPt[2] + z0 + (angle - phi0)*tanLambda/omega);

					//				  _hepRepWriter.addPoint (x, y, z);
				}
			}
			else
			{

				_hepRepWriter.addPoint (pca.x(), pca.y(), z0);
				_hepRepWriter.addPoint (pca.x() + (5000 * cos (phi0)), pca.y() + (5000*sin (phi0)), z0 + 5000*tanLambda);
			}


		}

		for (int track_counter = 0; track_counter < tracker_tracks->getNumberOfElements(); track_counter++)
		{

			//Temp. Drawing the pulses associated with a track.

			std::stringstream hitPulseName;
			hitPulseName << "TPC_HitPulse_" << *nameIt << "_[" << track_counter << "]";

			_hepRepWriter.addType (hitPulseName.str().c_str(), 1);
			_hepRepWriter.addAttDef ("TrackerHitPulse", "Pulse on pad associated with a hit used in the track", "Physics", "charge");
			_hepRepWriter.addAttValue ("DrawAs", "Polygon");
			_hepRepWriter.addAttValue ("Fill", "True");
			_hepRepWriter.addAttValue ("FillStyle", "NonZero");

			//TMP
			streamlog_out (MESSAGE3) << "Drawing track Pulses " << track_counter << " of " << tracker_tracks->getNumberOfElements() << endl;

			Track* current_track = dynamic_cast<Track*> (tracker_tracks->getElementAt (track_counter));

			assert (current_track);

			// determine the largest charge on the pads
			float maxCharge = 0.;
			float minCharge = 0.;

			const EVENT::TrackerHitVec& inputHits = current_track->getTrackerHits();

			for(TrackerHitVec::const_iterator hitIt = inputHits.begin(); hitIt != inputHits.end(); ++hitIt)
			{
				//Get the pulses in the hit
				LCObjectVec hitPulses = (*hitIt)->getRawHits();

				for(LCObjectVec::iterator pulseIt = hitPulses.begin(); pulseIt != hitPulses.end(); ++pulseIt)
				{

					TrackerPulse* current_pulse = dynamic_cast<TrackerPulse*> (*pulseIt);

					assert (current_pulse);

					minCharge = std::min (current_pulse->getCharge(), minCharge);
					maxCharge = std::max (current_pulse->getCharge(), maxCharge);
				}
				for(LCObjectVec::iterator pulseIt = hitPulses.begin(); pulseIt != hitPulses.end(); ++pulseIt)
				{

					TrackerPulse* current_pulse = dynamic_cast<TrackerPulse*> (*pulseIt);

					assert (current_pulse);


					_hepRepWriter.addInstance();

					const int padIndex = current_pulse->getCellID0();
					const int moduleID =  current_pulse->getCellID1();

					const double padCharge = current_pulse->getCharge();
					const int pulseQuality = current_pulse->getQuality();

					_hepRepWriter.addAttValue ("Primary_Electrons", padCharge);


					setHepRepPadColor (padCharge, minCharge, maxCharge, _pulseColorScheme, pulseQuality);
					drawHepRepPad (moduleID, padIndex);

				}
			}
		}
	}

}


Processor* HepRepOutputProcessor::newProcessor()
{

    return new HepRepOutputProcessor();
}


void HepRepOutputProcessor::check (LCEvent * /*evt*/)
{


}


void HepRepOutputProcessor::end()
{

    _hepRepWriter.close();
}



void HepRepOutputProcessor::drawHepRepPad (int moduleID, int padIndex)
{
    const gear::TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();
    const double max_drift_length = tpc_parameters.getMaxDriftLength();
    const gear::TPCModule& module =  tpc_parameters.getModule (moduleID);
    const gear::PadRowLayout2D & padLayout = module.getLocalPadLayout();

    if (module.getPadShape() != gear::PadRowLayout2D::RECTANGLE)
    {

        throw Exception ("HepRepOutputProcessor only supports RECTANGULAR shaped pads at the moment.");
    }

    //const gear::Vector2D padCentre = module.getPadCenter (padIndex);
    const gear::Vector2D padCentreLocal = padLayout.getPadCenter (padIndex);

    const double halfPadHeight = padLayout.getPadHeight (padIndex) / 2.;

    const double padWidth = padLayout.getPadWidth (padIndex);
    double halfPadWidth;

    switch (padLayout.getCoordinateType())
    {
        case gear::PadRowLayout2D::CARTESIAN:
        {
            halfPadWidth = padWidth / 2.;
        }
        break;

        case gear::PadRowLayout2D::POLAR:
        {

            halfPadWidth = padWidth * padCentreLocal[0] / 2.;
        }
        break;

        default:
            throw Exception ("HepRepOutputProcessor: unknown geometry type in module layout");
    }

    HepGeom::Vector3D<double> point_1 (-halfPadWidth, -halfPadHeight, max_drift_length);
    HepGeom::Vector3D<double> point_2 (-halfPadWidth,  halfPadHeight, max_drift_length);
    HepGeom::Vector3D<double> point_3 (halfPadWidth,  halfPadHeight, max_drift_length);
    HepGeom::Vector3D<double> point_4 (halfPadWidth, -halfPadHeight, max_drift_length);

    // move to correct positon on module pad plane
    switch (padLayout.getCoordinateType())
    {

        case gear::PadRowLayout2D::CARTESIAN:
        {

            HepGeom::Vector3D<double> translate (padCentreLocal[0], padCentreLocal[1], 0);
            point_1 += translate;
            point_2 += translate;
            point_3 += translate;
            point_4 += translate;
        }
        break;

        case gear::PadRowLayout2D::POLAR:
        {

            //HepGeom::Transform3D transform( HepGeom::Rotate3D(pad_centre[1], HepGeom::Vector3D(0., 0., 1.)), HepGeom::Vector3D(0., pad_centre[0], 0.) );
            //??? why have 90 degrees offset?

            const HepGeom::Vector3D<double> translate (0., padCentreLocal[0], 0.);
            const HepGeom::Rotate3D rotate (padCentreLocal[1] - M_PI / 2 , HepGeom::Vector3D<double> (0., 0., 1.));

            point_1 += translate;
            point_2 += translate;
            point_3 += translate;
            point_4 += translate;

            point_1 =  rotate * point_1;
            point_2 =  rotate * point_2;
            point_3 =  rotate * point_3;
            point_4 =  rotate * point_4;


        }
        break;

        default:
            throw Exception ("HepRepOutputProcessor: unknown geometry type in module layout");
    }

    // apply rotation of the module
    const HepGeom::Rotate3D rotate_module (module.getAngle(), HepGeom::Vector3D<double> (0., 0., 1.));
    point_1 = rotate_module * point_1;
    point_2 = rotate_module * point_2;
    point_3 = rotate_module * point_3;
    point_4 = rotate_module * point_4;

    // now shift to module offset*/

    switch (tpc_parameters.getCoordinateType())
    {
        case gear::PadRowLayout2D::CARTESIAN:
        {
            HepGeom::Vector3D<double> translate (module.getOffset() [0], module.getOffset() [1], 0);
            point_1 += translate;
            point_2 += translate;
            point_3 += translate;
            point_4 += translate;
      }
        break;

        case gear::PadRowLayout2D::POLAR:
        {
            const HepGeom::Vector3D<double> translate (module.getOffset() [0] * cos (module.getOffset() [1]) , module.getOffset() [0] * sin (module.getOffset() [1]), 0.);

            point_1 += translate;
            point_2 += translate;
            point_3 += translate;
            point_4 += translate;
        }
        break;

        default:
            throw Exception ("HepRepOutputProcessor: unknown coordinate type in module layout");
    }

    // setup the pad as upright at the origin
    // set the points
    _hepRepWriter.addPrimitive();
    _hepRepWriter.addPoint (point_1[0], point_1[1], point_1[2]);
    _hepRepWriter.addPoint (point_2[0], point_2[1], point_2[2]);
    _hepRepWriter.addPoint (point_3[0], point_3[1], point_3[2]);
    _hepRepWriter.addPoint (point_4[0], point_4[1], point_4[2]);
}



void HepRepOutputProcessor::setHepRepPadColor (double padCharge, float minCharge, float maxCharge, int colorScheme, int  /*pulseQuality*/)
{
/*	// If pulse Quality is bad just plot a gray blob, later one can colorcode the kind of error.
	if(pulseQuality != 0)
	{
		double red = 125;
		double green = 125;
		double blue = 125;

		streamlog_out (DEBUG3) << "Color (RGB) is " << red << "\t" << green << "\t" << blue << std::endl ;

		_hepRepWriter.addAttValue ("FillColor", red, green, blue);
        _hepRepWriter.addAttValue ("Color", red, green, blue);

		return;
	}
*/	//TODO: Doesn't work yet, but didn't have a look into pulse quality and how t's filled


	// apply the chosen color scheme
	switch (colorScheme)
    {
        case REDBLUE:
        {
			double red = 0;
			double green = 0;
			double blue = 0;
			
			if (padCharge > 0) // if == 0 its not a pulse, is it?
            {
                red = padCharge / maxCharge;
				green = 1.0 -red;
				blue = 1.0 -red;
            }
            
			if (padCharge < 0)
            {
                blue = padCharge / minCharge;
				green = 1.0 -blue;
				red = 1.0 -blue;
            }
			
			streamlog_out (DEBUG3) << "Color (red) is " << red << "\t" << green << "\t" <<  blue << std::endl ;

			_hepRepWriter.addAttValue ("FillColor", red, green, blue);
			_hepRepWriter.addAttValue ("LineColor", red, green, blue);
			_hepRepWriter.addAttValue ("Color", red, green, blue);
        }
        break;

        case RAINBOW:
        {
			double red = 0;
            double green = 0;
            double blue = 0;
			
			const double cyanWeak = 0.8; // weakens the brightness of cyan a little

            double frac = (padCharge - minCharge) / (maxCharge - minCharge);
            
			// orange -> red
            if (frac <= 1)
            {
                red = 1;
                green = (1 - 5 * (frac - 0.8)) * 1 * 0.5;
                blue = 0;
            }
            // yellow -> orange
            if (frac < 0.8)
            {
                red = 1;
                green = (2 - 5 * (frac - 0.6)) * 1 * 0.5 ;
                blue = 0;
            }
            // green -> yellow
            if (frac < 0.6)
            {
                red = 5 * (frac - 0.4) * 1;
                green = 1;
                blue = 0;
            }
            // cyan -> green
            if (frac < 0.4)
            {
                red = 0;
                green = ( (1 - cyanWeak) * 5 * (frac - 0.2) + cyanWeak) * 1;
                blue = (1 - 5 * (frac - 0.2)) * 1 * cyanWeak;
            }
            //violet -> cyan
            if (frac < 0.2)
            {
                red = (1 - 5 * frac) * 1 * 0.5 * cyanWeak;
                green = (5 * (frac)) * 1 * cyanWeak;
                blue = 1 * cyanWeak;
	
			}

            streamlog_out (DEBUG3) << "Color (RGB) is: " << red << "\t" << green << "\t" <<  blue << std::endl ;
            
			_hepRepWriter.addAttValue ("FillColor", red, green, blue);
// 			_hepRepWriter.addAttValue ("LineColor", red, green, blue);
//           _hepRepWriter.addAttValue ("Color", red, green, blue);
        }
        break;

        case TOPOGRAPHIC:
        {
			double red = 0;
            double green = 0;
            double blue = 0;
			
			const double cyanWeak = 0.8; // weakens the brightness of cyan a little
			const double oneThird = 1/3.0;
			const double twoThirds = 2 / 3.0;
            double frac = (padCharge - minCharge) / (maxCharge - minCharge);

			if(padCharge > 0) // if == 0 its not a pulse, is it?
			{
				frac = (padCharge - 0) / (maxCharge - 0);

				// orange -> red
				if (frac <= 1)
				{
					red = 1;
					green = (1 - 3 * (frac - twoThirds)) * 1 * 0.5;
					blue = 0;
				}
				// yellow -> orange
				if (frac < twoThirds)
				{
					red = 1;
					green = (2 - 3 * (frac - oneThird)) * 1 * 0.5 ;
					blue = 0;
				}
				// green -> yellow
				if (frac < oneThird)
				{
					red = 3 * (frac) * 1;
					green = 1;
					blue = 0;
				}
			}
			else
			{
				frac = (padCharge - minCharge) / (0 - minCharge);

				red = (1 - frac) * 1 * 0.5 * cyanWeak;
				green = (frac) * 1 * cyanWeak;
				blue = 1 * cyanWeak;
			}

            streamlog_out (DEBUG3) << "Color (RGB) is: " << red << "\t" << green << "\t" <<  blue << std::endl ;
            _hepRepWriter.addAttValue ("FillColor", red, green, blue);
 		//	_hepRepWriter.addAttValue ("LineColor", red, green, blue);
           _hepRepWriter.addAttValue ("Color", red, green, blue);
        }
        break;
		
		case GREYSCALE:
		{
			double red = 0;
            double green = 0;
            double blue = 0;
			const double MIN_GREY_VALUE = 0.2;
			const double MAX_GREY_VALUE = 0.9;
			const double SCALE_FACTOR = MAX_GREY_VALUE - MIN_GREY_VALUE;
	

            double frac = (padCharge - minCharge) / (maxCharge - minCharge);
			
			red = SCALE_FACTOR * frac + MIN_GREY_VALUE;
			green = SCALE_FACTOR * frac + MIN_GREY_VALUE;
			blue = SCALE_FACTOR * frac + MIN_GREY_VALUE;

            _hepRepWriter.addAttValue ("FillColor", red, green, blue);
			_hepRepWriter.addAttValue ("LineColor", red, green, blue);
            _hepRepWriter.addAttValue ("Color", red, green, blue);
		}
		break;

		case REDBLUESCALE:
		{
		}
		break;

		case COLORBRIGHTNESS:
		{
		}
		break;

		case COLORWHITE:
		{
		}
		break;

        default:
            throw Exception ("HepRepOutputProcessor: Unknown color scheme!");
    }
}

