//LCIO
#include "lcio.h"
#include "IO/LCWriter.h"
#include "EVENT/LCIO.h"
#include "IMPL/LCEventImpl.h"
#include "IMPL/LCGenericObjectImpl.h"
#include "IMPL/LCCollectionVec.h"
#include "IMPL/LCTOOLS.h"
#include "UTIL/LCTime.h"
//LCCD
#include "lccd.h"
#include "lccd/DBInterface.hh"
//GEAR
#include "gear/TPCParameters.h"
#include "gear/TPCModule.h"
#include "gearxml/GearXML.h"
#include "gear/GearMgr.h"
#include "gear/GEAR.h"
//TPCCONDDATA
#include "ADCChannelMapping.h"

//C++
#include <ctype.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>

void printUsage()
{
    std::cout<<"Usage:"<<std::endl;
    std::cout<<"To write data to a lcio simple file:"<<std::endl;
    std::cout<<"  MappingGenerator mapping.txt gearfile.xml [outfile.slcio]" <<std::endl;
    std::cout<<"  (outfile name is optional, default is TPCChannelMapping.slcio)"<< std::endl;
    std::cout<<"To write data to a database:"<<std::endl;
    std::cout<<"  MappingGenerator mapping.txt gearfile.xml connectionString folder startLCIOTimeStamp endLCIOTimeStamp"<<std::endl;
    std::cout<<"Formats:"<<std::endl;
    std::cout<<" Text file: hwGroup hwChannel moduleID row padNrInRow" << std::endl;
    std::cout<<" Connection String: hostname:database:login:password[:port]"<<std::endl;
    std::cout<<" LCIO time stamp: nano seconds since 1.1.1970 00:00:00"<<std::endl;
}



int main(int argc, char **argv)
{
    bool writeSimpleFile=false;
    bool writeToDatabase=false;
    std::string outFileName;
    lcio::LCWriter* lcWrt=NULL;
    lcio::LCEventImpl*  evt=NULL;
    if ( argc == 3 || argc == 4 )
    {
        writeSimpleFile=true;
        
    }
    else if ( argc == 7 )
    {
        writeToDatabase=true;
    }
    else
    {
        printUsage();
        return(1);
    }
    
    //DEBUG
    for (int i=0; i< argc ; ++i)
    {
        std::cout<<"DEBUG: argument ["<<i<<"]: "<<argv[i]<<std::endl;
    } 
    
    // read pad layout from gear
    gear::GearXML gearXML( argv[2] ) ;
    gear::GearMgr* gearMgr = gearXML.createGearMgr() ;
    gear::TPCParameters const &tpcParams=gearMgr->getTPCParameters();

    if ( writeSimpleFile )
    {
        // get outfile name
        if (argc==4)
            outFileName=argv[3];
        else
            outFileName="TPCChannelMapping.slcio";
        // open outfile
        lcWrt = lcio::LCFactory::getInstance()->createLCWriter() ;
        lcWrt->open( outFileName, lcio::LCIO::WRITE_NEW ) ;
        // create new lcio event
        evt = new lcio::LCEventImpl() ;
    }
    // create the collection we want to store the data in
    lcio::LCCollectionVec* mappingCol = new lcio::LCCollectionVec( lcio::LCIO::LCGENERICOBJECT );

    // open the input file
    std::filebuf infile;
    infile.open(argv[1],std::ios::in);
    std::istream inputstream(&infile);

    int linenumber = -1;
    // read all lines in the file
    while (!inputstream.eof() )
    {
        // read the next line from the file
        std::string line;
        std::getline(inputstream, line);
        if (inputstream.eof())   //eof is reached
        {
            if (!line.empty())
            {
                std::cout << "Ignoring non-empty line at end of file: " << line << std::endl;
            }
            break;
        }
        // the line as a stream so ew can stream into the variables
        std::stringstream linestream(line);
        linenumber++;
        // check if the next line is a comment
        if (linestream.peek() == '#')
        {
            std::cout << "DEBUG: skipping comment line " << linenumber << " : "  << line <<  std::endl;
            continue;
        }
        int hw_group, hw_channel, moduleID, row, padnr_in_row;
        // stream the data into the variables
        linestream >>  hw_group >> hw_channel >> moduleID >> row >>  padnr_in_row;
        //std::cout << "DEBUG:  reading line " << linenumber << " : "<< line << std::endl;
        // test if everything went fine
        if (linestream.fail())
        {
            std::cout << "error reading line " << linenumber << " : "<< line << std::endl;
            continue;
        }
        //std::cout << "DEBUG: retrieving module "<<moduleID << std::std::endl;
        gear::TPCModule const & module = tpcParams.getModule( moduleID );

        //std::cout << "DEBUG: got module with id "<< module.getModuleID() << " at "<< &module << std::endl;
        int pad_index = module.getPadIndex(row, padnr_in_row);
        tpcconddata::ADCChannelMapping * channelmapping =
                new  tpcconddata::ADCChannelMapping(hw_channel,  pad_index , -1, moduleID, hw_group);
        mappingCol->addElement( channelmapping );
    }                            //  while (!inputstream.eof() )

    if ( writeToDatabase )
    {
        std::string connectionString=argv[3];
        std::string folder=argv[4];
        lcio::long64 startTimeStamp;
        std::stringstream startstr(argv[5]);
        startstr >> startTimeStamp;
        lcio::long64 endTimeStamp;
        std::stringstream endstr(argv[6]);
        endstr >> endTimeStamp;
        lccd::DBInterface* dbConnection = new lccd::DBInterface(connectionString, folder, true );
        dbConnection->storeCollection( startTimeStamp, endTimeStamp, mappingCol, "TPCChannelMapping");
        delete dbConnection;
    }

    if ( writeSimpleFile )
    {
        // add the mapping collection to the event
        evt->addCollection( mappingCol , "TPCChannelMapping" ) ;
        // write the event to the file
        lcWrt->writeEvent( evt ) ;
        // dump the event to the screen
        lcio::LCTOOLS::dumpEvent( evt ) ;
        lcWrt->close() ;
        // ------------ IMPORTANT ------------- !
        // we created the event so we need to delete it ...
        delete evt ;
        // -------------------------------------
    }
    // close the input text file
    infile.close();
    return 0;
}
