#ifndef TESTFIELDSETTINGPROCESSOR_H
#define TESTFIELDSETTINGPROCESSOR_H

#include <marlin/Processor.h>
#include <lcio.h>
#include <string>

// MARLINTPC
#include "FieldSettingListener.h"
#include "FieldSetting.h"

namespace AIDA
{
  class IHistogram1D;
  class ITuple;
}

namespace marlintpc {
/**  Processor to test the field setting listener. Serves also as an example on how to use access FieldSettings
 * 
 * @author Klaus Zenker, DESY
 *
 * \section(Terminal output)
 * If either TPCPulses or TrackerRawData are given as input, all field information are printed.
 * This is done every time these data change for the given events.
 * \section(AIDA output)
 * If an ADIA processor is available there are two tuples that are filled.
 * \subsection(fieldData tuple)
 * The fieldData tuple includes all information retrieved from the database (filled for both kinds of input collection).
 * Every time the data in the data base for the processed data changes a new entry is added.
 * \subsection(driftVelocity tuple)
 * \pre{TPCPulses}
 * The driftVelocity tuple includes the drift velocity and corresponding drift field (filled only if TPCPulses are available). This is filled every time the conditions change.
 * For constant conditions the histogram including the pulse time is filled. When the conditions change the mean is calculated and the drift velocity is calculated.
 *
 */

class TestFieldSettingProcessor: public marlin::Processor {
public:

  virtual Processor* newProcessor() {
    return new TestFieldSettingProcessor;
  }

  TestFieldSettingProcessor();

  virtual void init();

  virtual void processRunHeader(lcio::LCRunHeader* run);

  virtual void processEvent(lcio::LCEvent* evt);

  virtual void check(lcio::LCEvent* evt);

  virtual void end();

protected:
  std::string _inputColNameRawData; ///< Name of the raw data input collection
  std::string _inputColNamePules; ///< Name of the pulse data input collection
  std::string _inputFieldSettingCollectionName; ///name of the FieldSetting collection

private:

  FieldSettingListener* _fieldSettingListener;
  AIDA::ITuple *_fieldData; ///< If AIDA is available field data is stored to a tree
  AIDA::ITuple *_driftVelocity; ///< If AIDA is available and TPCPulses are available the drift velocity is calculated
  AIDA::IHistogram1D *_tmpPulseTime; ///< Histogram with pulse times for constant field settings
  tpcconddata::FieldSetting* _currentSettings; ///< The last field data retrieved from the listener.
  double _driftLength; ///< The expected drift length needed to calculate the drift velocity
  double _maxDriftLength; ///< The drift length of the TPC needed to calculate the drift field given the cathode/anode voltage
  double _startTime; ///< The beginning of the histogram for pulse times and also a cut on the pulse times
  double _endTime; ///< The beginning of the histogram for pulse times and also a cut on the pulse times
  double _lastDriftField; ///< Store the last drift field in order to store this information together with the drift velocity
  double _pulseCharge; ///<If greater 0 it will applied to all pulses as minimum pulse charge cut
};
} // namespace marlintpc
#endif // TESTFIELDSETTINGPROCESSOR_H
