#ifndef TimePixCalibParamReaderProcessor_h
#define TimePixCalibParamReaderProcessor_h 1

#include "marlin/DataSourceProcessor.h"
#include <fstream>

namespace marlin{
  
  /** Reads TimePix calibration parameters stored in a .txt file\n
   *  Writes data to TrackerDataCollection named CalibrationParameters
   *  
   *
   *  <h4>Input - Prerequisites</h4>
   *  Calibration parameters file.
   *
   *  <h4>Output</h4>
   *  \li A runheader
   *  \li Collection with parameters in a LCGenericObject
   *
   * @param FileListName   Text file with list of input calibretion parameters files
   * \attention{TimePix file names must be absolute pathes!}
   * @param OutputCollectionName Name of the output collection, defaut \c CalibrationParameters
   * @param DetectorName Detector name for run header, default "TPC"
   * @param Description  Data descripton for run header, default "no description given"
   * @param SubDetectorName Sub-detector name for run header, default "TimePix"
   * @param ChipID Interer ID number of the chip 
   * @param NRow Number of rows on the chip, default 256
   * @param NColumn Number of columns on the chip, default 256
   * @param NParameters Number of calibration parameters per chip, default 8
   * @param ChipsPerBoardX Number of chips in a row, default 1
   * @param ChipsPerBoardY Number of chips in a column, default 1
   * @param SetOutputTransient If not 0 the output collection is set transient
   * @param TimePixChipID Identification name of chip
   *
   * @author R.Ulman , Bonn
   */
  
  class TimePixCalibParamReaderProcessor : public DataSourceProcessor {
    
  public:

   TimePixCalibParamReaderProcessor() ;

    TimePixCalibParamReaderProcessor*  newProcessor() ;


    /** Creates an event with a collection from the calibration parameters input file,
     *  where it gets 8 parameters for each chip (4-ToT, 2-THL, 2-TimeWalk),
     *  and calls all active processors' processEvent() and processRunHeader Method.
     *
     */
    void readDataSource( int numEvents ) ;    
    
    void init() ;
    void end() ;

     
  protected:
    
    std::string _outputColName;

    std::string _CalibParamFileName ;
    std::ifstream _infile;
    int _chipID;
    int _runNr;
    int _evtNr;
    int _nRow;
    int _nColumn;
    int _nParameters;
    int _chipsPerBoardX ;
    int _chipsPerBoardY ;
    std::string _detName;
    std::string _dataDescription;
    std::string _subDetName;
    std::string _timePixChipID;
    
    /** give the status which is set to the transient flag of the output collection
     * it is an int instead of a boolean, because the processor parameter cannot be a boolean
     */
    bool _outputIsTransient;
    
  };

 
} // end namespace marlin 

#endif
