#ifndef TimePixReaderProcessor_h
#define TimePixReaderProcessor_h 1

#include "marlin/DataSourceProcessor.h"
#include <fstream>

namespace marlin{
  
  /** Reads TimePix raw data: matrix of 256 x 256 values stored in a .dat file\n
   *  Writes data to TrackerRawDataCollection named TimePixRawData
   *  Currelty only work for one chip
   *  
   *
   *  <h4>Input - Prerequisites</h4>
   *  TimePix data files.
   *
   *  <h4>Output</h4>
   *  \li A runheader
   *  \li Events with TimePixRawData (lcio::TrackerRawData)
   *  \li First event contains TimePixModeMap collection
   *
   * @param FileListName   Text file with list of input TimePix files
   * \attention{TimePix file names must be absolute pathes!}
   * @param OutputCollectionName Name of the output collection, defaut \c TimePixRawData
   * @param ModeMapCollectionName Name of the output mode map collection, default \c TimePixMap
   * @param DetectorName Detector name for run header, default "TPC"
   * @param Description  Data descripton for run header, default "no description given"
   * @param SubDetectorName Sub-detector name for run header, default "TimePix"
   * @param ChipID Interer ID number of the chip (lcio::TrackerRawData::CellID0), 
   *   should be 0 as it's only one chip in the current version
   * @param Mode Operation mode of the chip: TOT, TIME, MM1 (mixed mode with CellID0 =  TIME), MM2 (mixed mode with CellID0 =  TOT)
   * @param NRow Number of rows on the chip, default 256
   * @param NColumn Number of columns on the chip, default 256
   * @param MaxRecordNumber Maximum number of events 100000
   * @param SetOutputTransient If not 0 the output collection is set transient
   *
   * @author Simone Zimmermann(University of Bonn) 
   */
  
  class TimePixReaderProcessor : public DataSourceProcessor {
    
  public:

   TimePixReaderProcessor() ;

    virtual TimePixReaderProcessor*  newProcessor() ;


    /** Creates events with TrackerRawData collections from the TimePix raw data input files and
     *  calls all active processors' processEvent() and processRunHeader Method.
     *
     */
    virtual void readDataSource( int numEvents ) ;
    
    
    virtual void init() ;
    virtual void end() ;

     
  protected:
    
    std::string _outputColName;

    bool _isFirstEvent;
    std::string _listName ;
    std::ifstream _infile;
    int _chipID;
    int _runNr;
    int _evtNr;
    int _nRow;
    int _nColumn;
    int _maxRecordN;
    std::string _detName;
    std::string _dataDescription;
    std::string _subDetName;
    
/** give the status which is set to the transient flag of the output collection
 * it is an int instead of a boolean, because the processor parameter cannot be a boolean
 */
    int _outputIsTransient;
    
  };

 
} // end namespace marlin 

#endif
