#include "ClockSyncProcessor.h"
#include <iostream>
#include <string>
#include <numeric>
#include <cassert>


// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/TrackerHit.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/LCFlagImpl.h>
#include <Exceptions.h>
#include <streamlog/streamlog.h>
// the common flag word definitions in MarlinTPC
#include "FlagwordDefinitions.h"

// GEAR
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>
#include <gearimpl/GlobalPadIndex.h>

// MarlinTPC
#include "ADCChannelMapping.h"
#include "TPCCondDataExeptions.h"
#include "TPCConditions.h"
#include "ChannelMappingListener.h"
#include "LCObjectCopier.h"

//MARLIN
#include <marlin/Global.h>

using namespace lcio;
using namespace marlin;
using namespace std;


namespace marlintpc
{

ClockSyncProcessor aClockSyncProcessor;


ClockSyncProcessor::ClockSyncProcessor() : Processor("ClockSyncProcessor")
{
	// the processor description
	_description = "ClockSyncProcessor is intended to correct the synchronization between the different RCUs.";

	/* register steering parameters: name, description, class-variable, default value

    for example:
	 */

	registerInputCollection(    LCIO::TRACKERHIT,
			"InputCollectionName" ,
			"Name of the input TrackerHit collection, default: TPCHit"  ,
			_inputColName ,
			string("TPCHits")) ;

	registerOutputCollection(   LCIO::TRACKERHIT,
			"OutputCollectionName" ,
			"Name of the output TrackerHit collection, default: TPCHitsSync",
			_outputCollectionName ,
			string("TPCHitsSync")) ;

	registerProcessorParameter("ChannelMappingCollectionName" ,
			"Name of the collection containing the ChannelMappings",
			_channelMappingColName,
			tpcconddata::ADCChannelMapping::getDefaultColName());

	registerOptionalParameter("InputTPCConditions",
			"Optional: Name of the input collection containing the TPC conditions data",
			_inputTPCConditionsCollectionName,
			tpcconddata::TPCConditions::getDefaultColName()) ;

	registerProcessorParameter("WriteOutputToStorage" ,
			"if true, the output hits collection is written to storage (default: true)"  ,
			_outputHitsPersistent,
			true);

	registerProcessorParameter( "RCUOutofSync" ,
			"which of the three RCUs is out of sync? If all RCUs are in Sync use -1. (default: -1)",
			_rcuOutOfSync,
			int(-1));

	registerProcessorParameter( "reduceTime" ,
			"How much should one reduce a too large Hit Time? (is in general negative)(default:0)",
			_reduceTime,
			double(0));

	registerProcessorParameter( "delayTime" ,
			"How much should one delay a too small Hit time? (is in general positive) (default:0)",
			_delayTime,
			double(0));


	registerProcessorParameter( "cut" ,
			"Threshold to calculate whether one should add or subtract the time.",
			_cut,
			double(0));

	registerOptionalParameter("VDriftOverride" ,
			"Optional: Set drift velocity in case there is no conditions data in mm/mu s"  ,
			_driftVelocityOverride ,
			float(0.)) ;


}


void ClockSyncProcessor::init()
{ 
	streamlog_out(DEBUG) << " init() called  " << endl;

	// you might want to list the used parameters by invoking
	printParameters();
	try{
		_channelMappingListener = new ChannelMappingListener(_channelMappingColName);
	}
	catch (tpcconddata::CondDataNotAvailableException &e)
	{
		streamlog_out(WARNING) << " ChannelMappingCollecection can not be found! End program as this is needed for this processor. " << endl;;
		return;
	}
	if(parameterSet("VDriftOverride"))
	{
		streamlog_out(WARNING) << "the drift velocity is taken from the steering file. please change" << endl;
		_driftVelocity=_driftVelocityOverride;
	}
	else // CAUTION! this is not only true for this value, but currently it's the only one in use
	{
		try
		{
			_tpcConditionsListener = new TPCConditionsListener(_inputTPCConditionsCollectionName);
		}
		catch(lccd::LCCDException &)
		{
			streamlog_out(ERROR) << "Collection named \"" << _inputTPCConditionsCollectionName << "\" with TPCConditions is not available." << std::endl;
			streamlog_out(ERROR) << "Use the ConditionsProcessor to provide it, or set the VDriftOverride parameter if this is not possible." << std::endl;
			throw;
		}
	}

}


void ClockSyncProcessor::processRunHeader(LCRunHeader* run)
{
	  run->parameters().setValue(_processorName + "_revision", "$Rev: 3786 $");
	    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
	      {
	        if (! i->second->isOptional() || i->second->valueSet())
	          {
	            run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
	          }
	      }

} 


void ClockSyncProcessor::processEvent(LCEvent* evt)
{ 

	const gear::TPCParameters& tpcParameters = Global::GEAR->getTPCParameters();

	 if (!parameterSet("DriftVelocityOverride"))
	                {
	                    if (_tpcConditionsListener->noConditionsDataAvailable())
	                    {
	                        std::stringstream out;
	                        out << _processorName << "::processEvent: No TPCCondition data available"
	                                << std::endl;
	                        throw tpcconddata::CondDataNotAvailableException(out.str());
	                    }assert(_tpcConditionsListener!=0);

	                    streamlog_out(DEBUG)
	                                << "The Drift Velocity was not set by the user. Attempting to recover it from the database"
	                                << std::endl;

	                    assert(_tpcConditionsListener->getTPCConditions() !=0);
	                    _driftVelocity = _tpcConditionsListener->getDriftVelocity();
	                    streamlog_out(DEBUG) << "Drift velocity set to " << _driftVelocity
	                                << "[mum/ns] using the condition data from the database" << std::endl;
	                }

	// use an LCCollectionVec instead of LCCollection so we can use iterators later
	LCCollectionVec* inputColVec = 0;
	// do a try / catch in case there is an empty event
	try
	{
		inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColName ) );
	}
	catch (DataNotAvailableException &e)
	{
		// no input data, just return and process the next event
		return;
	}

	LCCollectionVec* correctedHitCollection = new LCCollectionVec(LCIO::TRACKERHIT);
	correctedHitCollection->setTransient(!_outputHitsPersistent);
	// set the flags that cellID1 should be stored
	lcio::LCFlagImpl trkFlag(0);
	trkFlag.setBit(lcio::LCIO::RTHBIT_ID1);
	correctedHitCollection->setFlag(trkFlag.getFlag());

	int refRow=7;
	int refRCU;
	if (_rcuOutOfSync==0)
		refRCU=1;
	else
		refRCU=0;

	// loop the data collection and do something
	std::vector<double>* refTimes = new std::vector<double>;
	std::vector<double>* checkTimes = new std::vector<double>;

	for( LCCollectionVec::iterator inputObjectIter = inputColVec->begin() ;
			inputObjectIter < inputColVec->end() ;
			inputObjectIter++ )
	{
		// dynamic cast the input object (LCObject) to what it really is e.g.
		TrackerHit * inputHit = dynamic_cast<TrackerHit *>( *inputObjectIter );
		gear::GlobalPadIndex nearestPadIndex = tpcParameters.getNearestPad( inputHit->getPosition()[0],
				inputHit->getPosition()[1]);
		const gear::TPCModule& module = tpcParameters.getModule( nearestPadIndex.getModuleID() );

		int hitRow=module.getRowNumber(nearestPadIndex.getPadIndex());

		streamlog_out(DEBUG4) << "nearestPadIndex.getPadIndex()=" <<nearestPadIndex.getPadIndex() << ", nearestPadIndex.getModuleID=" <<nearestPadIndex.getModuleID() << endl;
		std::pair<int, int> hitHardwareIDPair=_channelMappingListener->getHardwareMapping(std::make_pair(nearestPadIndex.getPadIndex(),nearestPadIndex.getModuleID()));

		if(hitHardwareIDPair.second==refRCU && hitRow==refRow)
		{
			refTimes->push_back(inputHit->getTime());
		}
		if(hitHardwareIDPair.second==_rcuOutOfSync && hitRow>refRow && hitRow<=(refRow+6))
		{
			checkTimes->push_back(inputHit->getTime());
		}

	}

	std::vector<double>* diffTimes = new std::vector<double>;;
	for(unsigned int i=0; i<refTimes->size();i++)
	{
		for(unsigned int j=0; j<checkTimes->size();j++)
		{
			diffTimes->push_back(refTimes->at(i)-checkTimes->at(j));
		}
	}

	double mean = std::accumulate(diffTimes->begin(), diffTimes->end(), 0.0) / diffTimes->size();
	for( LCCollectionVec::iterator inputObjectIter = inputColVec->begin() ;
			inputObjectIter < inputColVec->end() ;
			inputObjectIter++ )
	{
		// dynamic cast the input object (LCObject) to what it really is e.g.
		TrackerHit * inputHit = dynamic_cast<TrackerHit *>( *inputObjectIter );

		gear::GlobalPadIndex nearestPadIndex = tpcParameters.getNearestPad( inputHit->getPosition()[0],
				inputHit->getPosition()[1]);
		std::pair<int, int> hitHardwareIDPair=_channelMappingListener->getHardwareMapping(std::make_pair(nearestPadIndex.getPadIndex(),nearestPadIndex.getModuleID()));

		TrackerHitImpl* outputHit= new TrackerHitImpl();
		outputHit= LCObjectCopier::copy(inputHit);
		if(hitHardwareIDPair.second==_rcuOutOfSync)
		{
			outputHit->setTime(outputHit->getTime()+((mean>_cut) ? _delayTime:_reduceTime)-_cut);
			double position[3] = {inputHit->getPosition()[0],inputHit->getPosition()[1],Global::GEAR->getTPCParameters().getMaxDriftLength()-_driftVelocity * (outputHit->getTime()) / 1000.};
			outputHit->setPosition(position);
		}

		correctedHitCollection->addElement(outputHit);
	}

	evt->addCollection(correctedHitCollection, _outputCollectionName);
}


void ClockSyncProcessor::check(LCEvent*  /*evt*/)
{ 
	// check what is needed here
}


void ClockSyncProcessor::end()
{ 
	// do something at the end of the processing, e.g. cleaning up
}
} // close namespace brackets
