#include "OctoboardReaderProcessor.h"
#include "TimePixPixelMode.h"

using namespace lcio;
using namespace tpcconddata;
using namespace std;

namespace marlin{



  OctoboardReaderProcessor aOctoboardReaderProcessor ;
  
  
  OctoboardReaderProcessor::OctoboardReaderProcessor() : DataSourceProcessor("OctoboardReaderProcessor") {
    
    _description = std::string("Reads TimePix raw data files as input and creates LCIO events with TrackerRawData collections.")
      + std::string(" Make sure to not specify any LCIOInputFiles in the steering in order to read TimePix raw data files.") ;

  
  registerOutputCollection( LCIO::TRACKERRAWDATA,
			    "OutputCollectionName" , 
			    "Name of the output collection"  ,
			    _outputColName,
			    std::string("TimePixRawData") ) ;
  
  registerProcessorParameter( "FileListName" , 
			      "input file with list of TimePix files"  ,
			      _listName ,
			      std::string("TimePixData.txt") ) ;

  registerProcessorParameter( "DetectorName" , 
			      "Name of the detector"  ,
			      _detName ,
			      std::string("TPC" ) );
  
  registerProcessorParameter( "Description" , 
			      "Description of data"  ,
			      _dataDescription ,
			      std::string("no description given" ) ) ;
  
  registerProcessorParameter( "SubDetectorName" , 
			      "Name of the subdetector"  ,
			      _subDetName ,
			      std::string("TimePix" ) );
  
  /*registerProcessorParameter( "ChipID" ,
			      "ID number of the chip "  ,
			      _chipID,
			      (int) 0 ) ;*/
  
  registerProcessorParameter( "NRow" , 
			      "Number of rows on the chip "  ,
			      _nRow,
			      (int) 256 ) ; 
  
  registerProcessorParameter( "NColumn" , 
			      "Number of columns on the chip "  ,
			      _nColumn,
			      (int) 256 ) ; 
  
  registerProcessorParameter( "MaxRecordNumber" , 
			      "_maxRecordN "  ,
			      _maxRecordN,
			      (int) 100000 ) ; 

  registerProcessorParameter( "SetOutputTransient" , 
			      "if not 0 the output collection is set transient"  ,
			      _outputIsTransient,
			      0);
  registerProcessorParameter( "TOTTIME" ,
			      "1=TOT, 0=TIME"  ,
			      _totTime,
			      (int) 0);

  registerProcessorParameter( "RunNumber","Set RunNumber (default: 0)",_runNr,(int) 0);
  
  }
  
  OctoboardReaderProcessor*  OctoboardReaderProcessor::newProcessor() {
    return new OctoboardReaderProcessor ;
  }
  

  void OctoboardReaderProcessor::init()
  {
	  printParameters() ;

	  //same declaration stuff
	  /*run->parameters().setValue(_processorName + "_revision", "$Rev$");

	  //for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
	  //{
	  //  if (! i->second->isOptional() || i->second->valueSet())
	  //  {
	  //	  run->parameters().setValue(_processorName + "_" + i->second->name(),
	  //			  i->second->value());
	  //  }
	  //}*/

	  //to use: svn propset svn:keywords Rev myprocessor.cc

	  _isFirstEvent = true;
	  _evtNr=0;

	  //read in .dat files
	  try { _infile.open( _listName.c_str() ); }
	  catch (IOException& e )
	  {
		  std::cout<< e.what() <<std::endl;
	  }
	  if ( !_infile.good() )
	  {
		  _infile.close();
		  streamlog_out(ERROR) << "Input file "<< _listName << "not found" <<std::endl;
	  }
  }


  void OctoboardReaderProcessor::readDataSource( int  /*numEvents*/ )
  {
	  streamlog_out(MESSAGE1) << "readDataSource" <<std::endl;
	  int _numberOfChips=8;
	  int datapos = 0;

	  while ( _evtNr < _maxRecordN )
	  {
		  streamlog_out(DEBUG0)<< _evtNr<<" < "<<_maxRecordN<<std::endl;

		  if (_evtNr%1000==0)
		  {
			  streamlog_out(MESSAGE3)<< "Event: "<<_evtNr<<std::endl;
		  }
		  else
		  {
			  if (_evtNr%100==0) streamlog_out(MESSAGE2)<< "Event: "<<_evtNr<<std::endl;
		  }

		  //read in .dat files
		  std::string datafile;

		  _infile >> datafile;
		  if ( !_infile.good() ) break;

		  // streamlog_out(DEBUG0)<<_evtNr<<" . ";

		  // create new lcio event
		  LCEventImpl*  evt = new LCEventImpl() ;
		  evt->setRunNumber(  _runNr   ) ;
		  evt->setEventNumber( _evtNr ) ;
		  evt->setDetectorName( _detName ) ;

		  // we want to store a collection of trackerRawData
		  LCCollectionVec* trackerRawDataVec = new LCCollectionVec( LCIO::TRACKERRAWDATA )  ;

		  // set the flags that cellID1 should be stored
		  lcio::LCFlagImpl trkFlag(0) ;
		  trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
		  trackerRawDataVec->setFlag( trkFlag.getFlag() );

		  // set the transient flag as definded by the processor parameter
		  trackerRawDataVec->setTransient( (_outputIsTransient != 0) );

		  std::ifstream eventfile;

		  for(int chiploop=0;chiploop<_numberOfChips;chiploop++)
		  {
			  bool cFile = false; //flag to rember that we have to close the file
			  streamlog_out(DEBUG3)<<"for begin "<< chiploop << std::endl;
			  datapos=datafile.find(".")-11;

			  //create filename
			  stringstream sdatapos;
			  int chipNum=chiploop+1;
			  streamlog_out(DEBUG3)<<"for mid "<< chiploop << std::endl;
			  sdatapos << chipNum;
			  datafile.replace(datapos,1,sdatapos.str());

			  //open file
			  eventfile.open( datafile.c_str() );

			  if (!eventfile.good())
			  {
				  streamlog_out(ERROR)<<"data file " << datafile.c_str() << " not .good()"<<std::endl;
				  eventfile.close();
				  //	  continue;
			  }
			  else{
				  streamlog_out(DEBUG0)<<"file found " << datafile << std::endl;
				  cFile = true; //set flaf to close file
			  }

			  //create lcio objects
			  TrackerRawDataImpl* rawData = new TrackerRawDataImpl;

			  rawData->setCellID0( 0 );
			  rawData->setCellID1( chiploop );

			  EVENT::ShortVec ADCValues;

			  //get data from file
			  while (1)
			  {

				  short data;
				  eventfile >> data;
				  if (!eventfile.good()) break;
				  //if (_totTime == 0 && (data >560 || data < 292)) data=0;
				//  if (_totTime == 1 && (data <0 || data > 700)) data=0;
				  ADCValues.push_back( data );
			  }

			  rawData->setADCValues(ADCValues);

			  // add the  rawData to the collection
			  trackerRawDataVec->push_back(rawData);
			  streamlog_out(DEBUG3)<<"for end "<< chiploop << std::endl;

			  //close file
			  if(cFile) eventfile.close();
		  }
		  // add the rawData collection to the event
		  evt->addCollection( trackerRawDataVec , _outputColName ) ;


		  if ( _isFirstEvent ) {

			  //create RunHeader
			  streamlog_out(DEBUG0)<<"create RunHeader"<<std::endl;

			  LCRunHeaderImpl* runHdr = new LCRunHeaderImpl ;

			  runHdr->setRunNumber( _runNr) ;
			  runHdr->setDetectorName( _detName ) ;
			  runHdr->setDescription( _dataDescription ) ;
			  runHdr->addActiveSubdetector( _subDetName ) ;

			  ProcessorMgr::instance()->processRunHeader( runHdr ) ;
			  delete runHdr;

			  _isFirstEvent = false;
		  }

		  // dump the event to the screen
		  streamlog_out(DEBUG0)<<"in readDataSource"<<std::endl;

		  if( streamlog::out.write< streamlog::DEBUG2 >() )
			  LCTOOLS::dumpEvent( evt ) ;

		  ProcessorMgr::instance()->processEvent( evt ) ;


		  // ------------ IMPORTANT ------------- !
		  // we created the event so we need to delete it ...
		  delete evt ;
		  // -------------------------------------

		  _evtNr++;

	  }

	  streamlog_out(DEBUG0)<<"\nend of readDataSource"<<std::endl;

  }



  void OctoboardReaderProcessor::end() {



  }



}
