#include "TimePixCalibParamReaderProcessor.h"
#include "TimePixCalibParam.h"

#include "marlin/ProcessorMgr.h"

#include "IO/LCWriter.h"
#include "IMPL/LCEventImpl.h"
#include "IMPL/LCRunHeaderImpl.h"
#include "IMPL/TrackerDataImpl.h"
#include "IMPL/LCCollectionVec.h"
#include <IMPL/LCFlagImpl.h>

//#include "UTIL/LCStdHepRdr.h"
#include "UTIL/LCTOOLS.h"

using namespace lcio;
using namespace tpcconddata;

namespace marlin{


  TimePixCalibParamReaderProcessor aTimePixCalibParamReaderProcessor ;
  
  
  TimePixCalibParamReaderProcessor::TimePixCalibParamReaderProcessor() : DataSourceProcessor("TimePixCalibParamReaderProcessor") {
    
    _description = std::string("Reads Calibration parameters data as input and creates a LCIO file with a LCGenericObject.")
      + std::string(" Make sure to not specify any LCIOInputFiles in the steering file in order to read the calibration parameters.") ;

  
  registerOutputCollection( LCIO::LCGENERICOBJECT,
			    "OutputCollectionName" , 
			    "Name of the output collection"  ,
			    _outputColName,
			    std::string( "CalibrationParameters" ) ) ;
  
  registerProcessorParameter( "CalibParamFileName" , 
			      "input file with calibration parameters, one chip after the other"  ,
			      _CalibParamFileName ,
			      std::string( "CalibrationParameters.txt" ) ) ;

  registerProcessorParameter( "DetectorName" , 
			      "Name of the detector"  ,
			      _detName ,
			      std::string( "TPC" ) );
  
  registerProcessorParameter( "Description" , 
			      "Description of data"  ,
			      _dataDescription ,
			      std::string("Calibration parameters" ) ) ;
  
  registerProcessorParameter( "SubDetectorName" , 
			      "Name of the subdetector"  ,
			      _subDetName ,
			      std::string( "TimePix" ) );
  
  registerProcessorParameter( "ChipID" , 
			      "ID number of the chip "  ,
			      _chipID,
			      0 ) ;
  
  registerProcessorParameter( "NRow" , 
			      "Number of rows on the chip "  ,
			      _nRow,
			      256 ) ; 
  
  registerProcessorParameter( "NColumn" , 
			      "Number of columns on the chip "  ,
			      _nColumn,
			      256 ) ; 

  registerProcessorParameter( "NParameters" , 
			      "Number of calibration parameters per Chip"  ,
			      _nParameters,
			      8 ) ;

  registerProcessorParameter( "ChipsPerBoardX" , 
			      "Number of chips in a row "  ,
			      _chipsPerBoardX ,
			      1 ) ;
  
  registerProcessorParameter( "ChipsPerBoardY" , 
			      "Number of chops in a column "  ,
			      _chipsPerBoardY ,
			      1 ) ;

  registerProcessorParameter( "SetOutputTransient" , 
			      "if not 0 the output collection is set transient"  ,
			      _outputIsTransient,
			      false );

  registerProcessorParameter( "TimePixChipID" , 
			      "Identification name of chip"  ,
			      _timePixChipID ,
			      std::string( "H09 - W0009" ) );
  
  }
  
  TimePixCalibParamReaderProcessor*  TimePixCalibParamReaderProcessor::newProcessor() { 
    return new TimePixCalibParamReaderProcessor ;
  }
  
  void TimePixCalibParamReaderProcessor::init() {    
    printParameters() ;    
    
    _isFirstEvent = true; 
    _runNr=0;
    _evtNr=0;

    //read in data files  
    try { _infile.open( _CalibParamFileName.c_str() ); }
    catch ( IOException& e ){
      streamlog_out(WARNING0)<< e.what() <<std::endl;
      }
    if ( !_infile.good() ) {
      _infile.close();
      streamlog_out(ERROR) << "***********************************" << std::endl;
      streamlog_out(ERROR) << " *** " << _CalibParamFileName.c_str() <<" not good ***"<<std::endl;
      streamlog_out(ERROR) << "***********************************" << std::endl;
    }

  }


  void TimePixCalibParamReaderProcessor::readDataSource( int  /*numEvents*/ ) {
 
    streamlog_out(MESSAGE1)<<"readDataSource"<<std::endl;

    
    if ( _infile.good() ) {
      
      streamlog_out(DEBUG0)<<_evtNr<<std::endl;

      // create new lcio event
      LCEventImpl*  evt = new LCEventImpl() ;
      evt->setRunNumber( _runNr ) ;
      evt->setEventNumber( _evtNr ) ;
      evt->setDetectorName( _detName ) ;
      
      
      // want to store a collection of LCGenericObject
      LCCollectionVec* calibParameters = new LCCollectionVec( LCIO::LCGENERICOBJECT )  ;

      // set the flags that cellID1 should be stored
      lcio::LCFlagImpl trkFlag(0) ;
      trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
      calibParameters->setFlag( trkFlag.getFlag() );

      // set the transient flag as defined by the processor parameter
      calibParameters->setTransient( _outputIsTransient != 0 );
    
      // number of chips on the board
      int nChips = _chipsPerBoardX*_chipsPerBoardY ;
      int paramNr = 0 ;
      
      for ( int chip=0; chip<nChips; chip++ ){	
	for ( int pa = 0; pa < _nParameters; pa++ ){   //loop over parameters
	  
	  float data;
	  // parameter index for a quad board
	  paramNr = chip*_nParameters + pa ;  
	  
	  _infile >> data; 
	  if ( !_infile.good() ) break;
	  
	  streamlog_out(MESSAGE4)<<"Parameter " << paramNr << " is: " << data << std::endl;
	  // read in chipID, parameter index, and parameter value
	  CalibParam* CP = new CalibParam( chip, paramNr , data) ;
	  	  
	  // add the Data to the collection
	  calibParameters->addElement( CP );
 
	}   
      }   

      // chip indentification name
      calibParameters->parameters().setValue(_timePixChipID, _evtNr);
   
      // add the Data collection to the event
      evt->addCollection( calibParameters , _outputColName ) ;

	
      // create RunHeader
      streamlog_out(DEBUG0)<<"create RunHeader"<<std::endl;
      
      LCRunHeaderImpl* runHdr = new LCRunHeaderImpl ; 
      
      runHdr->setRunNumber( _runNr ) ;
      runHdr->setDetectorName( _detName ) ;
      runHdr->setDescription( _dataDescription ) ;
      runHdr->addActiveSubdetector( _subDetName ) ;
      
      ProcessorMgr::instance()->processRunHeader( runHdr ) ;
      delete runHdr;
	
    
      // dump the event to the screen 
      streamlog_out(DEBUG0)<<"in readDataSource"<<std::endl;

      if( streamlog::out.write< streamlog::DEBUG2 >() )
	  LCTOOLS::dumpEvent( evt ) ;

      ProcessorMgr::instance()->processEvent( evt ) ;     
	
      
      // we created the event so we need to delete it 
      delete evt ;


      streamlog_out(DEBUG0)<<"\nend of readDataSource"<<std::endl;
    }
  }
  


  void TimePixCalibParamReaderProcessor::end() {
    // nothing done here
  }


}
