#include "TimePixMaskCreaterProcessor.h"

namespace marlin{

TimePixMaskCreaterProcessor aTimePixMaskCreaterProcessor ;

TimePixMaskCreaterProcessor::TimePixMaskCreaterProcessor() : DataSourceProcessor("TimePixMaskCreaterProcessor"),
		_noisyPixels(1,0),
		_deadPixels(1,0),
		_noisyColumns(1,0),
		_deadColumns(1,0),
		_noisyRows(1,0),
		_deadRows(1,0)
{

	_description = "Creates 1 LCIO event with LCGenericObjectCollection containing the dead or noisy pixel "
			" Make sure to not specify any LCIOInputFiles in the steering in order to read the occupancy histogram.";


	registerOutputCollection( LCIO::LCGENERICOBJECT,
			"ModeMapCollectionName" ,
			"Name of the output mode map collection"  ,
			_statusMapColName,
			std::string("TimePixStatusMap") ) ;

	registerProcessorParameter( "DetectorName" ,
			"Name of the detector"  ,
			_detName ,
			std::string("TPC" ) );

	registerProcessorParameter( "Description" ,
			"Description of data"  ,
			_description ,
			std::string("no description given" ) ) ;

	registerProcessorParameter( "SubDetectorName" ,
			"Name of the subdetector"  ,
			_subDetName ,
			std::string("TimePix" ) );

	registerProcessorParameter( "ChipID" ,
			"ID number of the chip "  ,
			_chipID,
			(int) 0 ) ;

	registerProcessorParameter( "NColumn" ,
			"Number of columns on the chip "  ,
			_nColumn,
			(int) 256 ) ;

	registerProcessorParameter( "NRow" ,
			"Number of rows on the chip "  ,
			_nRow,
			(int) 256 ) ;

	registerProcessorParameter( "SetOutputTransient" ,
			"if not 0 the output collection is set transient"  ,
			_outputIsTransient,
			0);

	registerOptionalParameter( "NoisyPixels",
			"CellID0s of pixels to be marked as noisy",
			_noisyPixels,
			_noisyPixels);

	registerOptionalParameter( "DeadPixels",
			"CellID0s of pixels to be marked as dead",
			_deadPixels,
			_deadPixels);

	registerOptionalParameter( "NoisyColumns",
			"CellID0s of Columns to be marked as noisy",
			_noisyColumns,
			_noisyColumns);

	registerOptionalParameter( "DeadColumns",
			"CellID0s of Columns to be marked as dead",
			_deadColumns,
			_deadColumns);

	registerOptionalParameter( "NoisyRows",
			"CellID0s of Rows to be marked as noisy",
			_noisyRows,
			_noisyRows);

	registerOptionalParameter( "DeadRows",
			"CellID0s of Rows to be marked as dead",
			_deadRows,
			_deadRows);

}

TimePixMaskCreaterProcessor*  TimePixMaskCreaterProcessor::newProcessor() {
	return new TimePixMaskCreaterProcessor ;
}

void TimePixMaskCreaterProcessor::init() {
	printParameters() ;

	_isFirstEvent = true;
	_runNr=0;
	_evtNr=0;

}


void TimePixMaskCreaterProcessor::readDataSource( int  /*numEvents*/ ) {

	// create new lcio event
	lcio::LCEventImpl*  evt = new lcio::LCEventImpl() ;
	evt->setRunNumber(  _runNr   ) ;
	evt->setEventNumber( _evtNr ) ;
	evt->setDetectorName( _detName ) ;

	streamlog_out(DEBUG0)<<"Created lcio::LCEvent evt"<<std::endl;

	// we want to store a collection of LCGenericObject
	lcio::LCCollectionVec* statusMapVec = new lcio::LCCollectionVec( LCIO::LCGENERICOBJECT )  ;

	streamlog_out(DEBUG0)<<"Created lcio::LCCollectionVec statusMapVec"<<std::endl;

	// set the transient flag as definded by the processor parameter
	statusMapVec->setTransient( (_outputIsTransient != 0) );

	std::map<int,int> statusMap;

	if(parameterSet("DeadPixels"))
	{
		streamlog_out(DEBUG0) << _deadPixels.size() << " pixels were defined as dead, their CellID0s are:" << std::endl;

		for(std::vector<int>::iterator iter = _deadPixels.begin(); iter < _deadPixels.end(); iter++)
		{
			streamlog_out(DEBUG0) << *iter << std::endl;
			statusMap[*iter] = tpcconddata::TimePixPixelMode::DEAD;
		}
	}

	if(parameterSet("DeadColumns"))
	{
		streamlog_out(DEBUG0) << _deadColumns.size() << " columns were defined as dead, their numbers are:" << std::endl;

		for(std::vector<int>::iterator iter = _deadColumns.begin(); iter < _deadColumns.end(); iter++)
		{
			streamlog_out(DEBUG0) << *iter << std::endl;
			for(int i = 0; i < _nRow; i++)
			{
				statusMap[ (i*_nColumn) + *iter] = tpcconddata::TimePixPixelMode::DEAD;
			}
		}
	}

	if(parameterSet("DeadRows"))
	{
		streamlog_out(DEBUG0) << _deadRows.size() << " rows were defined as Dead, their numbers are:" << std::endl;

		for(std::vector<int>::iterator iter = _deadRows.begin(); iter < _deadRows.end(); iter++)
		{
			streamlog_out(DEBUG0) << *iter << std::endl;
			for(int i = 0; i < _nColumn; i++)
			{
				statusMap[ (*iter * _nColumn) + i ] = tpcconddata::TimePixPixelMode::DEAD;
			}
		}
	}

	if(parameterSet("NoisyPixels"))
	{
		streamlog_out(DEBUG0) << _noisyPixels.size() << " pixels were defined as noisy, their CellID0s are:" << std::endl;

		for(std::vector<int>::iterator iter = _noisyPixels.begin(); iter < _noisyPixels.end(); iter++)
		{
			streamlog_out(DEBUG0) << *iter << std::endl;
			statusMap[*iter] = tpcconddata::TimePixPixelMode::NOISY;
		}
	}

	if(parameterSet("NoisyColumns"))
	{
		streamlog_out(DEBUG0) << _noisyColumns.size() << " columns were defined as noisy, their numbers are:" << std::endl;

		for(std::vector<int>::iterator iter = _noisyColumns.begin(); iter < _noisyColumns.end(); iter++)
		{
			streamlog_out(DEBUG0) << *iter << std::endl;
			for(int i = 0; i < _nRow; i++)
			{
				statusMap[ (i*_nColumn) + *iter] = tpcconddata::TimePixPixelMode::NOISY;
			}
		}
	}

	if(parameterSet("NoisyRows"))
	{
		streamlog_out(DEBUG0) << _noisyRows.size() << " rows were defined as noisy, their numbers are:" << std::endl;

		for(std::vector<int>::iterator iter = _noisyRows.begin(); iter < _noisyRows.end(); iter++)
		{
			streamlog_out(DEBUG0) << *iter << std::endl;
			for(int i = 0; i < _nColumn; i++)
			{
				statusMap[ (*iter * _nColumn) + i ] = tpcconddata::TimePixPixelMode::NOISY;
			}
		}
	}

	streamlog_out(DEBUG0) << statusMap.size() << " pixels were marked as dead or noisy" << std::endl;

	streamlog_out(DEBUG0) << "Adding dead and noisy pixels to statusMapVec" << std::endl;

	for(std::map<int,int>::iterator iter = statusMap.begin(); iter != statusMap.end(); iter++)
	{
		statusMapVec->addElement( new tpcconddata::TimePixPixelMode(_chipID, iter->first, tpcconddata::TimePixPixelMode::ONEHIT, iter->second));
	}

	streamlog_out(DEBUG0) << "Adding statusMapVec to event" << std::endl;

	evt->addCollection( statusMapVec, _statusMapColName );

	//create RunHeader
	streamlog_out(DEBUG0)<<"Create lcio::LCRunHeader runHdr" << std::endl;

	lcio::LCRunHeaderImpl* runHdr = new lcio::LCRunHeaderImpl ;

	runHdr->setRunNumber( _runNr) ;
	runHdr->setDetectorName( _detName ) ;
	runHdr->setDescription( _description ) ;
	runHdr->addActiveSubdetector( _subDetName ) ;

	streamlog_out(DEBUG0)<<"Process runHdr" << std::endl;

	ProcessorMgr::instance()->processRunHeader( runHdr ) ;
	delete runHdr;

	// dump the event to the screen
	if( streamlog::out.write< streamlog::DEBUG2 >() )
		LCTOOLS::dumpEvent( evt ) ;

	streamlog_out(DEBUG0)<<"Process evt" << std::endl;

	ProcessorMgr::instance()->processEvent( evt ) ;


	// ------------ IMPORTANT ------------- !
	// we created the event so we need to delete it ...
	delete evt ;
	// -------------------------------------

	streamlog_out(DEBUG0)<<"Deleted evt" << std::endl;

}



void TimePixMaskCreaterProcessor::end() {



}


}
