#include "TimePixSimpleAlternativeReaderProcessor.h"
#include "TimePixPixelMode.h"

#include "marlin/ProcessorMgr.h"

#include "IO/LCWriter.h"
#include "IMPL/LCEventImpl.h"
#include "IMPL/LCRunHeaderImpl.h"
#include "IMPL/TrackerRawDataImpl.h"
#include "IMPL/LCCollectionVec.h"
#include <IMPL/LCFlagImpl.h>

#include "UTIL/LCTOOLS.h"

using namespace lcio;
using namespace tpcconddata;

namespace marlin{



TimePixSimpleAlternativeReaderProcessor aTimePixSimpleAlternativeReaderProcessor ;


TimePixSimpleAlternativeReaderProcessor::TimePixSimpleAlternativeReaderProcessor() : DataSourceProcessor("TimePixSimpleAlternativeReaderProcessor") {

	_description = std::string("Reads TimePix data files (format: x y value) as input and creates LCIO events with TrackerRawData collections.")
	+ std::string(" Make sure to not specify any LCIOInputFiles in the steering in order to read TimePix raw data files.")
	+ std::string(" Warning: Although input files are zero suppressed, output collections are NOT!");


	registerOutputCollection( LCIO::TRACKERRAWDATA,
			"OutputCollectionName",
			"Name of the output collection",
			_outputColName,
			std::string("TimePixRawData") );

	registerProcessorParameter( "FileListName",
			"input file with list of TimePix files",
			_listName,
			std::string("TimePixData.txt") );

	registerProcessorParameter( "DetectorName",
			"Name of the detector",
			_detName,
			std::string("TPC" ) );

	registerProcessorParameter( "Description",
			"Description of data",
			_dataDescription,
			std::string("no description given" ) );

	registerProcessorParameter( "SubDetectorName",
			"Name of the subdetector",
			_subDetName,
			std::string("TimePix" ) );

	registerProcessorParameter( "ChipID",
			"ID number of the chip",
			_chipID,
			(int) 0 );

	registerProcessorParameter( "NRow",
			"Number of rows on the chip ",
			_nRow,
			(int) 256 );

	registerProcessorParameter( "NColumn",
			"Number of columns on the chip ",
			_nColumn,
			(int) 256 );

	registerProcessorParameter( "MaxRecordNumber",
			"_maxRecordN",
			_maxRecordN,
			(int) 100000 );

	registerProcessorParameter( "SetOutputTransient",
			"if not 0 the output collection is set transient",
			_outputIsTransient,
			0);

	registerProcessorParameter( "RunNumber",
			"Set RunNumber (default: 0)",
			_runNr,
			(int) 0);

}

TimePixSimpleAlternativeReaderProcessor*  TimePixSimpleAlternativeReaderProcessor::newProcessor() {
	return new TimePixSimpleAlternativeReaderProcessor ;
}



void TimePixSimpleAlternativeReaderProcessor::init() {
	printParameters() ;

	_isFirstEvent = true;
	_evtNr=0;

	//read in .dat files
	try { _infile.open( _listName.c_str() ); }
	catch (IOException& e ){
		std::cout<< e.what() <<std::endl;
	}
	if ( !_infile.good() ) {
		_infile.close();
		streamlog_out(ERROR)<<"!"<<_listName<<".good()"<<std::endl;
	}

}



void TimePixSimpleAlternativeReaderProcessor::readDataSource( int  /*numEvents*/ ) {

	streamlog_out(MESSAGE1)<<"readDataSource"<<std::endl;

	while ( _evtNr < _maxRecordN ) {

		streamlog_out(DEBUG0)<< _evtNr<<" < "<<_maxRecordN<<std::endl;

		if (_evtNr%1000==0)
		{
			streamlog_out(MESSAGE3)<< "Event: "<<_evtNr<<std::endl;
		}
		else
		{
			if (_evtNr%100==0)
				streamlog_out(MESSAGE2)<< "Event: "<<_evtNr<<std::endl;
		}

		//read in data files
		char datafile[7200];

		_infile >> datafile;

		if ( !_infile.good() ) break;

		// streamlog_out(DEBUG0)<<_evtNr<<" . ";

		// create new lcio event
		LCEventImpl*  evt = new LCEventImpl() ;
		evt->setRunNumber(  _runNr   ) ;
		evt->setEventNumber( _evtNr ) ;
		evt->setDetectorName( _detName ) ;


		// we want to store a collection of trackerRawData
		LCCollectionVec* trackerRawDataVec = new LCCollectionVec( LCIO::TRACKERRAWDATA )  ;

		// set the flags that cellID1 should be stored
		lcio::LCFlagImpl trkFlag(0) ;
		trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
		trackerRawDataVec->setFlag( trkFlag.getFlag() );

		// set the transient flag as definded by the processor parameter
		trackerRawDataVec->setTransient( (_outputIsTransient != 0) );

		std::ifstream eventfile;
		eventfile.open( datafile );

		if (!eventfile.good()) {
			std::cout<<"!"<<datafile<<".good()"<<std::endl;
			eventfile.close();
			continue;
		}

		EVENT::ShortVec ADCValues;

		int cellid0(0);

		std::vector<int> cellID0Vec;

		short data;
		int pixX;
		int pixY;

		while (1) {

			//load position and ADCValue of one hit pixel
			eventfile >> pixX;
			eventfile >> pixY;
			eventfile >> data;

			//calculate cellID0 - cellID0=0 corresponds to (x,y)=(0,0)
			cellid0 = (pixY  * _nColumn) + pixX;

			cellID0Vec.push_back( cellid0 );
			ADCValues.push_back( data );

			//break if no more data is available
			if (!eventfile.good()) break;
		}

		EVENT::ShortVec chip;

		for(int iHit = 0; iHit < _nColumn*_nRow; iHit++){

			chip.push_back(0);
		}

		for(int iHit = 0; iHit < static_cast<int>(cellID0Vec.size()); iHit++){

			chip[cellID0Vec[iHit]] = ADCValues[iHit];
		}

		//create new TrackerRawData
		TrackerRawDataImpl* rawData = new TrackerRawDataImpl;

		//assign cellID0
		rawData->setCellID0( 0 );
		rawData->setCellID1( _chipID );

		//assign ADC counts
		rawData->setADCValues(chip);
		trackerRawDataVec->push_back(rawData);

		// add the rawData collection to the event
		evt->addCollection( trackerRawDataVec , _outputColName ) ;

		if ( _isFirstEvent ) {

			//create RunHeader
			streamlog_out(DEBUG0)<<"create RunHeader"<<std::endl;

			LCRunHeaderImpl* runHdr = new LCRunHeaderImpl ;

			runHdr->setRunNumber( _runNr) ;
			runHdr->setDetectorName( _detName ) ;
			runHdr->setDescription( _dataDescription ) ;
			runHdr->addActiveSubdetector( _subDetName ) ;

			ProcessorMgr::instance()->processRunHeader( runHdr ) ;
			delete runHdr;

			_isFirstEvent = false;
		}

		// dump the event to the screen
		streamlog_out(DEBUG0)<<"in readDataSource"<<std::endl;

		if( streamlog::out.write< streamlog::DEBUG2 >() )
			LCTOOLS::dumpEvent( evt ) ;

		ProcessorMgr::instance()->processEvent( evt ) ;


		// ------------ IMPORTANT ------------- !
		// we created the event so we need to delete it ...
		delete evt ;
		// -------------------------------------

		_evtNr++;

	}//while ( _evtNr < _maxRecordN )

	streamlog_out(DEBUG0)<<"\nend of readDataSource"<<std::endl;

}//readDataSource



void TimePixSimpleAlternativeReaderProcessor::end() {
}

} //namespace marlin
