#!/usr/bin/python

## \package createProcessor
# A very simple script to generate a processor skeleton for Marlin.
# \author: Ch. Rosemann, DESY
# \author: M. Killenberg, CERN
# \author: O. Schaefer, DESY

import os, sys, getopt
from sys import argv

usage  ="%s needs at least one argument, a processor name.\n" % argv[0]
usage +="By this a skeleton processor will be generated as .h and .cc file.\n"
usage +="Use either [--help/-h] to see this message.\n"
usage +="Optional is [-p=/--path=], which requires a path as an additional argument.\n"
usage +="The files will be placed into the /include and the /src directory relative to this path, respectively (if the directories exist).\n"
usage +="The --aida option inserts template code how to register and fill "
usage +="AIDA histograms using the AIDAProcessor.\n"
usage +="\n"
usage +="Examples: \n"
usage +="   %s MyTestProcessor\n" % argv[0]
usage +="   %s --path=justTesting --aida MyTestProcessor" % argv[0]

headerPath="./"
sourcePath="./"

## get the options, store any other command line argument
try:
    opts, args = getopt.getopt(sys.argv[1:], "hp:", ["path=","help","aida"])
except getopt.GetoptError, ex:
    print usage
    sys.exit(1)

## set the actual processor name, first test for bad input
if len(args) != 1:
    print usage
    sys.exit(1)

try:
    name=args[0]
except:
    print usage
    sys.exit(1)

useAida="false"

## look through the options
for opt, arg in opts:
    if opt in ("--path","-p"):
        relPath = arg
        headerPath = relPath+"/include/"
        sourcePath = relPath+"/src/"
    if opt in ("--help","-h"):
        print usage
        sys.exit(1)
    if opt == "--aida":
        useAida = "true"


## assemble the names of the output files and test if the paths are valid AND the files don't yet exist
headerFileName=headerPath+name+".h"
sourceFileName=sourcePath+name+".cc"

fileList = [headerFileName, sourceFileName]

for f in fileList:
    if (os.path.exists(f)):
        print "%s already exists, I refuse to overwrite. Please clear before." %f
        sys.exit(1)

try:
    headerFile = open(headerFileName, 'w')
    sourceFile = open(sourceFileName, 'w')
except IOError:
    print "The files can not be written, most likely the directory doesn't exist or you forgot to specify a processor name."
    print "Try invoking %s --help for further information" % argv[0]
    sys.exit(1)

headerText  = "#ifndef %s_H\n" % name.upper()
headerText += "#define %s_H\n" % name.upper()
headerText += "\n"
headerText += "#include <marlin/Processor.h>\n"
headerText += "#include <lcio.h>\n"
headerText += "#include <string>\n"
headerText += "\n"

if useAida == "true":
    headerText += "// In case you want AIDA histograms\n"
    headerText += "namespace AIDA\n"
    headerText += "{\n"
    headerText += "        class IHistogram1D;\n"
    headerText += "}\n"

headerText += "\n"
headerText += "namespace marlintpc\n"
headerText += "{\n"
headerText += "///  Short description of processor.\n"
headerText += "\n"
headerText += "/**\n" 
headerText += " * Lengthy and detailed description of processor ...\n"
headerText += " * @author Fn. Lastname, Affiliation\n"
headerText += " *\n"
headerText += " * \par Credits:\n"
headerText += " * The processor skeleton was generated by the script \c createProcessor.py\n"
headerText += " *\n"
headerText += " */\n"
headerText += "\n"
headerText += "class %s : public marlin::Processor\n" % name
headerText += "{\n"
headerText += " public:\n"
headerText += "  \n"
headerText += "  virtual Processor*  newProcessor() { return new %s; }\n" % name
headerText += "  \n"
headerText += "  %s();\n"  % name
headerText += "  \n"
headerText += "  virtual void init();\n"
headerText += "  \n"
headerText += "  virtual void processRunHeader(lcio::LCRunHeader* run);\n"
headerText += "  \n"
headerText += "  virtual void processEvent(lcio::LCEvent* evt);\n"
headerText += "  \n"
headerText += "  virtual void check(lcio::LCEvent* evt);\n"
headerText += "  \n"
headerText += "  virtual void end();\n"
headerText += "  \n"
headerText += "  \n"
headerText += " protected:\n"
headerText += "  /* the place for protected and private member data and functions */\n"
headerText += "  std::string _inputColName; ///< Name of the input collection\n"

if useAida == "true":
    headerText += "\n"
    headerText += "  AIDA::IHistogram1D * _myAidaHistogram; ///< Pointer to the histogram"
    headerText += "\n"
    headerText += "  int _histoNBins; ///< Number of bins in the histogram\n"
    headerText += "  float _histoMin; ///< Minimum of the histogram\n"
    headerText += "  float _histoMax; ///< Maximum of the histogram\n"

headerText += "\n"
headerText += "};\n"
headerText += "} // namespace marlintpc\n"
headerText += "#endif // %s_H\n" % name.upper()

headerFile.write(headerText)
headerFile.close()

sourceText  = "#include \"%s.h\"\n" % name
sourceText += "#include <iostream>\n"
sourceText += "#include <string>\n"
sourceText += "\n"
sourceText += "// include what you need from lcio\n"
sourceText += "#include <EVENT/LCEvent.h>\n"
sourceText += "#include <IMPL/LCCollectionVec.h>\n"
sourceText += "#include <EVENT/TrackerData.h>\n"
sourceText += "#include <Exceptions.h>\n"
sourceText += "\n"
sourceText += "using namespace lcio;\n"
sourceText += "using namespace marlin;\n"
sourceText += "using namespace std;\n"
sourceText += "\n"

if useAida == "true":
    sourceText += "// Use AIDA histograms\n"
    sourceText += "#include <AIDA/AIDA.h>\n"
    sourceText += "#include <marlin/AIDAProcessor.h>\n"
    sourceText += "#include <AIDA/IHistogramFactory.h>\n"
    sourceText += "#include <AIDA/IHistogram1D.h> // or IHistogram2D\n"    

sourceText += "\n"
sourceText += "namespace marlintpc\n"
sourceText += "{\n"
sourceText += "\n"
sourceText += "%s a%s;\n" % (name,name)
sourceText += "\n"
sourceText += "\n"
sourceText += "%s::%s() : Processor(\"%s\")\n" % (name, name, name)
sourceText += "{\n"
sourceText += "  // the processor description\n"
sourceText += "  _description = \"%s is intended to ...\";\n" % name
sourceText += "  \n"
sourceText += "  /* register steering parameters: name, description, class-variable, default value\n"
sourceText += "\n"
sourceText += "    for example:\n"
sourceText += "  */\n"
sourceText += "\n"
sourceText += "        registerInputCollection(    LCIO::TRACKERDATA,\n"
sourceText += "            \"InputCollectionName\" ,\n"
sourceText += "            \"Name of the input TrackerData collection, default: TPCData\"  ,\n"
sourceText += "            _inputColName ,\n"
sourceText += "            string(\"TPCData\")) ;\n"
sourceText += "\n"
sourceText += "  /*\n"
sourceText += "        registerOutputCollection(   LCIO::TRACKERDATA,\n"
sourceText += "          \"OutputCollectionName\" ,\n"
sourceText += "          string(\"Name of the output TrackerData collection\")+\n"
sourceText += "          string(\"only the data related to a pulse will be stored\")+\n"
sourceText += "          string(\, default: TPCDataInPulses),\n"
sourceText += "          _outputCollectionName ,\n"
sourceText += "          string(\"TPCDataInPulses\")) ;\n"
sourceText += "\n"
sourceText += "        // It's nive to have the default value in the comment\n"
sourceText += "        registerProcessorParameter( \"MinPulsesPerHit\" ,\n"
sourceText += "	       \"Minimum number of pulses for a valid hit, default: 3\",\n"
sourceText += "	       _minPulsesPerHit,\n"
sourceText += "	       int(3));\n"
sourceText += "\n"
sourceText += "        // Optional parameters should only be used when set\n"
sourceText += "        // in the steering file.\n"
sourceText += "        // The default value is there only for technical reasons,\n"
sourceText += "        // it is meaningless.\n"
sourceText += "        //\n"
sourceText += "        // If properly implemented the cut is not applied when the parameter is not set:\n"
sourceText += "        //   if (parameterSet(\"ChageCut\")) { //apply the cut }\n"
sourceText += "        //\n"
sourceText += "        registerOptionalParameter( \"ChargeCut\" ,\n"
sourceText += "            \"If the charge is below this value the data will be ignored\",\n"
sourceText += "            _chargeCut,\n"
sourceText += "           float(17.8));\n"
sourceText += "  */\n"

if useAida == "true":
    sourceText += "\n"
    sourceText += "  registerProcessorParameter( \"HistoNBins\" ,\n"
    sourceText += "			      \"Number of bins in the histogram, default: 100\",\n"
    sourceText += "			      _histoNBins,\n"
    sourceText += "			      int(100));\n"
    sourceText += "\n"
    sourceText += "  registerProcessorParameter( \"HistoMin\",\n"
    sourceText += "			      \"Minimum of histogram, default: 0.\",\n"
    sourceText += "			      _histoMin,\n"
    sourceText += "			      float(0.));\n"
    sourceText += "\n"
    sourceText += "  registerProcessorParameter( \"HistoMax\" ,\n"
    sourceText += "			      \"Maximum of histogram, default: 0.\",\n"
    sourceText += "			      _histoMax,\n"
    sourceText += "			      float(250.));\n"
    sourceText += "\n"

sourceText += "}\n"
sourceText += "\n"
sourceText += "\n"
sourceText += "void %s::init()\n" % name
sourceText += "{ \n"
sourceText += "  streamlog_out(DEBUG) << \" init() called  \" << endl;\n"
sourceText += "  \n"
sourceText += "  // you might want to list the used parameters by invoking\n"
sourceText += "  printParameters();\n"
sourceText += "\n"

if useAida == "true":
    sourceText += "_myAidaHistogram =  AIDAProcessor::histogramFactory(this)->\n"
    sourceText += "                    createHistogram1D( \"myAidaHistogram\",\n"
    sourceText += "                                       \"Title for myAidaHistogram\",\n"
    sourceText += "                                       _histoNBins, _histoMin , _histoMax);\n"
    sourceText += "\n"

sourceText += "}\n"
sourceText += "\n"
sourceText += "\n"
sourceText += "void %s::processRunHeader(LCRunHeader* run)\n" % name
sourceText += "{\n"
sourceText += "  // do something for each run\n"
sourceText += "} \n"
sourceText += "\n"
sourceText += "\n"
sourceText += "void %s::processEvent(LCEvent* evt)\n" % name
sourceText += "{ \n"
sourceText += "  // do something for each event, usually access the input collection:\n"
sourceText += "\n"
sourceText += "  // use an LCCollectionVec instead of LCCollection so we can use iterators later\n"
sourceText += "  LCCollectionVec* inputColVec = 0;\n"
sourceText += "\n"  
sourceText += "  // do a try / catch in case there is an empty event\n"
sourceText += "  try\n"
sourceText += "  {\n"
sourceText += "    inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColName ) );\n"
sourceText += "  }\n"
sourceText += "  catch (DataNotAvailableException &e)\n" 
sourceText += "  {\n"
sourceText += "    // no input data, just return and process the next event\n"
sourceText += "    return;\n"
sourceText += "  }\n"
sourceText += "\n"  
sourceText += "  // loop the data collection and do something \n"
sourceText += "  for( LCCollectionVec::iterator inputObjectIter = inputColVec->begin() ;\n" 
sourceText += "       inputObjectIter < inputColVec->end() ;\n"
sourceText += "       inputObjectIter++ )\n"
sourceText += "  {\n"
sourceText += "    // dynamic cast the input object (LCObject) to what it really is e.g.\n"
sourceText += "    TrackerData * myTrackerData = dynamic_cast<TrackerData *>( *inputObjectIter );\n"
sourceText += "\n"
sourceText += "    // for objects derrived from LCGenericObject or LCFixedObject (like conditions data)\n"
sourceText += "    // you have to use special constructor\n"
sourceText += "    // TPCConditions myTPCConditions( dynamic_cast<LCGenericObject*>( *inputObjectIter ) );\n"
sourceText += "    \n"
sourceText += "    //now do something with the data ...\n"

if useAida == "true":
    sourceText += "    _myAidaHistogram->fill( myTrackerData->getTime() );\n"

sourceText += "  }\n"
sourceText += "}\n"
sourceText += "\n"
sourceText += "\n"
sourceText += "void %s::check(LCEvent* evt)\n" % name
sourceText += "{ \n"
sourceText += "  // check what is needed here \n"
sourceText += "}\n"
sourceText += "\n"
sourceText += "\n"
sourceText += "void %s::end()\n" % name
sourceText += "{ \n"
sourceText += "  // do something at the end of the processing, e.g. cleaning up\n"
sourceText += "}\n"
sourceText += "} // close namespace brackets"

sourceFile.write(sourceText)
sourceFile.close()

print "Created:\n"
print "\t%s\n" % headerFileName
print "\t%s\n" % sourceFileName
print "Please remember that you may need to update your CMakeList."
