/* ChannelCorrection.cc
 * $Id$ 
 */
#include <bitset>
#include <sstream>
#include <iostream>
#include <string.h>

#include "ChannelCorrection.h"
#include "TPCCondDataExeptions.h"

namespace tpcconddata{

    std::string ChannelCorrection::getRevision()
    {
        return std::string("$Rev: 5162 $");
    }
    
    
    std::string ChannelCorrection::getDefaultColName()
    {
        return std::string("TPCChannelCorrection");
    }
    
        
    ChannelCorrection::ChannelCorrection(int padID, int moduleID, int channelID, int readoutGroup, int quality, float calibrationValue, float timeOffset)
    {
        setGearID(padID, moduleID);
        setHardwareID(channelID, readoutGroup);
        setQuality(quality);
        setCalibrationFactor(calibrationValue);
        setTimeOffset(timeOffset);
    }
    
    
    ChannelCorrection::ChannelCorrection(int padID, int moduleID, int channelID, int readoutGroup)
    {
        setGearID(padID, moduleID);
        setHardwareID(channelID, readoutGroup);
        setQuality(0);
        setCalibrationFactor(1.0);
        setTimeOffset(0.0);
    }
    
    
    ChannelCorrection::~ChannelCorrection() { /* no op*/  }
    
    
    std::pair<int,int> ChannelCorrection::getGearID() const
    {
        if ( getIntVal(0)!=-1 && getIntVal(1)!=-1 )
        {
            return std::make_pair(getIntVal(0), getIntVal(1));
        }
        else
        {
            std::stringstream message;
            message << "TPCCondData::ChannelCorrection::getGearID(): padID and/or moduleID not set (out of range!)";
            throw tpcconddata::CondDataNotAvailableException(message.str());
        }
    } 
    
    
    
    std::pair<int,int> ChannelCorrection::getHardwareID() const
    {
        if ( getIntVal(2)!=-1 && getIntVal(3)!=-1 )
        {
            return std::make_pair(getIntVal(2), getIntVal(3));
        }
        else
        {
            std::stringstream message;
            message << "TPCCondData::ChannelCorrection::getHardwareID(): channelID and/or readoutGroup not set (out of range!)";
            throw tpcconddata::CondDataNotAvailableException(message.str());
        }
    } 
    
     
    
    int ChannelCorrection::getQuality() const
    {
        return getIntVal(4);
        
    }
    
    
    
    bool ChannelCorrection::isDead()
    {
        return (getIntVal(4)>>0) & 1;
    }
    
    
    
    bool ChannelCorrection::isNoisy()
    {
        return (getIntVal(4)>>1) & 1;
    }
    
    
    
    bool ChannelCorrection::isAddonBitASet()
    {
        return (getIntVal(4)>>2) & 1;
    }
    
    
    
    bool ChannelCorrection::isAddonBitBSet()
    {
        return (getIntVal(4)>>3) & 1;
    }
    
    
    
    float ChannelCorrection::getCalibrationFactor() const
    {
        return getFloatVal( 0 );
    } 
    
    
    float ChannelCorrection::getTimeOffset() const
    {
        return getFloatVal( 1 );
    } 


    void ChannelCorrection::setCalibrationFactor(float calibrationValue)
    {
        _obj->setFloatVal( 0, calibrationValue );
    }
    
    
    void ChannelCorrection::setTimeOffset(float timeOffset)
    {
        _obj->setFloatVal( 1, timeOffset );
    }
    

    
    void ChannelCorrection::setQuality(int quality)
    {
        if ( quality > 15 || quality < 0)
        {
            std::stringstream message;
            message << "TPCCondData::ChannelCorrection::setQuality(int): Integer to set quality bit must be between 0 and 15, but is: "<<quality<<" (out of range!)";
            throw tpcconddata::ConditionsObjectException(message.str());
        }
        _qualityBit=std::bitset<4>(quality);
        _obj->setIntVal( 4, quality );
    }
    
    
    
    void ChannelCorrection::setIsDead(bool toggle)
    {
        
        _qualityBit.set(0, toggle);
        _obj->setIntVal( 4, static_cast<int>(_qualityBit.to_ulong()) );
    }
    
    
    
    void ChannelCorrection::setIsNoisy(bool toggle)
    {
        
        _qualityBit.set(1, toggle);
        _obj->setIntVal( 4, static_cast<int>(_qualityBit.to_ulong()) );
    }
    
    
    
    void ChannelCorrection::setAddonBitA(bool toggle)
    {
        
        _qualityBit.set(2, toggle);
        _obj->setIntVal( 4, static_cast<int>(_qualityBit.to_ulong()) );
    }
    
    
    
    void ChannelCorrection::setAddonBitB(bool toggle)
    {
        
        _qualityBit.set(3, toggle);
        _obj->setIntVal( 4, static_cast<int>(_qualityBit.to_ulong()) );
    }
    
    
    
    void ChannelCorrection::print(  std::ostream& os ) const
    {
        std::streamsize PrecisionBefore = os.precision();
        std::streamsize WidthBefore = os.width();
        os.precision(3);  os.setf(std::ios_base::fixed);
        os << "ChannelCorrection: pad:" << getGearID().first << "/module:" << getGearID().second << 
                "//channel:" << getHardwareID().first << "/readoutGroup:" << getHardwareID().second << 
                " quality:" << getQuality()<< std::fixed; 
        os << " calibrationValue: " << getCalibrationFactor() << " timeOffset: " << getTimeOffset() << std::endl;
        os.unsetf( std::ios_base::fixed );
        os.precision( PrecisionBefore );
        os.width( WidthBefore );
    }
    

    
    void ChannelCorrection::setGearID(int padID, int moduleID)
    { 
        _obj->setIntVal( 0, padID );
        _obj->setIntVal( 1, moduleID );
    }
    
    
    
    void ChannelCorrection::setHardwareID(int channelID, int readoutGroup)
    { 
        _obj->setIntVal( 2, channelID );
        _obj->setIntVal( 3, readoutGroup );
    }
    
    
    std::ostream &operator<<(std::ostream &os, const ChannelCorrection &p) {
	p.print(os);
	return os;
    }

}
