
#include "TPCDistortionMeasurement.h"
#include <Exceptions.h>

namespace tpcconddata {

std::string TPCDistortionMeasurement::getRevision() {
  return std::string("$Rev: 2568 $");
}

std::string TPCDistortionMeasurement::getDefaultColName() {
  return std::string("TPCDistortionMeasurement");
}

TPCDistortionMeasurement::TPCDistortionMeasurement(CLHEP::Hep3Vector* truePosition, CLHEP::Hep3Vector* reconstructedPosition, int dot_id) :
    _createObj(true) // remember that we have to delete the object in the destructor
{
  // create a fixed size generic object with 5 ints
  _myObj = new IMPL::LCGenericObjectImpl(1, 0, 6);

  // set the 4 values
  _myObj->setDoubleVal(0, truePosition->getX());
  _myObj->setDoubleVal(1, truePosition->getY());
  _myObj->setDoubleVal(2, truePosition->getZ());
  _myObj->setDoubleVal(3, reconstructedPosition->getX());
  _myObj->setDoubleVal(4, reconstructedPosition->getY());
  _myObj->setDoubleVal(5, reconstructedPosition->getZ());
  _myObj->setIntVal(0, dot_id);
}

TPCDistortionMeasurement::TPCDistortionMeasurement(EVENT::LCObject* obj) :
    _createObj(false) // we did not create the object, so we must not delete it in the d'tor
{
  // initialise the object pointer
  _myObj = dynamic_cast<IMPL::LCGenericObjectImpl*>(obj);

  if (_myObj == 0)         // the cast did not succeed
      {
    throw lcio::Exception("Cannot create TPCDistortionMeasurement from sth."
        " that is not LCGenericObjectImpl");
  }

  // check the size of the generic object
  // there have to be no ints, no floats and 6 doubles
  if ((_myObj->getNInt() != 1)
      || (_myObj->getNFloat() != 0) || (_myObj->getNDouble() != 6)) {
    throw lcio::Exception(
        "TPCDistortionMeasurement(LCObject* obj): Wrong number of elements in object");
  }

}

TPCDistortionMeasurement::~TPCDistortionMeasurement() {
  // delete the LCGenericObjectImpl, but only if it was created in the constructor
  if (_createObj)
    delete (_myObj);
}

void TPCDistortionMeasurement::print(std::ostream& os) const {
  os << "TPCDistortionMeasurement: true photodot position (" << getTruePosition().getX()
      << "," <<  getTruePosition().getY()
      << "," <<  getTruePosition().getZ() << std::endl;
  os << "TPCDistortionMeasurement: reconstructed photodot position (" << getReconstructedPosition().getX()
      << "," <<  getReconstructedPosition().getY()
      << "," <<  getReconstructedPosition().getZ() << std::endl;

}

CLHEP::Hep3Vector TPCDistortionMeasurement::getTruePosition() const{
  CLHEP::Hep3Vector v;
  v.set(_myObj->getDoubleVal(0),_myObj->getDoubleVal(1),_myObj->getDoubleVal(2));
  return v;
}

CLHEP::Hep3Vector TPCDistortionMeasurement::getReconstructedPosition() const{
  CLHEP::Hep3Vector v;
  v.set(_myObj->getDoubleVal(3),_myObj->getDoubleVal(4),_myObj->getDoubleVal(5));
  return v;
}


int TPCDistortionMeasurement::getDotID() const {
  return _myObj->getIntVal(0);
}

std::ostream &operator<<(std::ostream &os, const TPCDistortionMeasurement &acm) {
  acm.print(os);
  return os;
}

}
 //namespace
