#include "lcio.h"

#include "IO/LCWriter.h"
#include "EVENT/LCIO.h"

#include "IMPL/LCEventImpl.h" 
#include "IMPL/LCRunHeaderImpl.h" 
#include "IMPL/LCCollectionVec.h"
#include "IMPL/LCTOOLS.h"
#include "IMPL/MCParticleImpl.h"

#include <iostream>
#include <sstream>
#include <string>
#include <cstdio>
#include <typeinfo>
#include <cmath>
#include <cstdlib>

#include <CLHEP/Random/RandFlat.h>

using namespace std ;
using namespace lcio ;

void printUsage()
{
  cout << "generateMuons energy phi lambda nEvents [outfile.slcio]" <<endl<<endl;
  cout <<"* Generates muons coming from the vertex flying in direction  phi in r/phi plane with inclination lambda" << endl;
  cout <<"* Use 0..2pi to define a fixed phi angle. Negative phi gives random angle." << endl;
  cout <<"* Lambda range is -pi .. pi "<< std::endl;
    cout <<"* Outfile name is optional, " << " default is MCMuons.slcio"<< endl;
    cout << "* Energy is in GeV" << endl;
}

int main(int argc, char **argv)
{

  if (argc!=5 && argc!=6 )
  {
    printUsage();
    return(1);
  }

  
   // get outfile name
  string outFileName;
  if (argc==6)
    outFileName=argv[5];
  else
    outFileName="MCMuons.slcio";

  double energy = std::atof( argv[1] );
  double phi_in =  std::atof( argv[2] );
  double lambda =  std::atof( argv[3] );
  unsigned int nEvents = std::atoi( argv[4] );

  static const double m = 0.1056583668; // the muon mass in GeV

  if (energy < m)
  {
    cout << "Error, energy cannot be smaller than mass of the muom ( "
	 << m << " GeV)!" << endl;
    return -1;
  }

  // open outfile			     
  try{
    // create sio writer
    LCWriter* lcWrt = LCFactory::getInstance()->createLCWriter() ;
    
    try{  lcWrt->open( outFileName, LCIO::WRITE_NEW ) ;
    }
    catch( IOException& e ){    
      cout << e.what() << endl ;
      return 0 ;
    }
      
      // create runheader and write it
      // the information should come from a file
      // or be given interactively in a future version
//      LCRunHeaderImpl* runHdr = new LCRunHeaderImpl ; 
      int runNr=0;
//      runHdr->setRunNumber( runNr) ;
//      
//      lcWrt->writeRunHeader( runHdr ) ;
  
      
      // loop the events, generate 100,000 events
      for (unsigned int eventNr=0; eventNr < nEvents; eventNr++)
      {
	// preparation: calculate the momentum
	  double p[3];
	  double phi;
	  if (phi_in < 0)
	  {
	    phi = CLHEP::RandFlat::shoot( 2*M_PI );
	  }
	  else 
	  {
	    phi = phi_in;
	  }

	  p[0]=sqrt( energy*energy - m*m ) * cos(phi) *cos(lambda);
	  p[1]=sqrt( energy*energy - m*m ) * sin(phi) *cos(lambda);
	  p[2]=sqrt( energy*energy - m*m ) * sin(lambda);


       // now start the real work...
	// create new lcio event
	LCEventImpl*  evt = new LCEventImpl() ;
	
	evt->setRunNumber(  runNr   ) ;
	evt->setEventNumber( eventNr ) ;
      
	// we want to store a collection of trackerRawData
	LCCollectionVec* mcMuonsVec = new LCCollectionVec( LCIO::MCPARTICLE )  ;
	
	MCParticleImpl *muon =  new MCParticleImpl;
	muon->setPDG(13); // it's a muon
	muon->setCharge(-1); // muons have chage -1
	muon->setMass( m ); // the muon mass
	muon->setGeneratorStatus(1); // a stable particle in the detector
	muon->setMomentum( p ); // the the corresponding momentum 

	// add the muon
	mcMuonsVec->addElement( muon );

	// add the rawData collection to the event
	evt->addCollection( mcMuonsVec , "MCMuons" ) ;

	// write the event to the file
	lcWrt->writeEvent( evt ) ;
	
	// dump the event to the screen 
	LCTOOLS::dumpEvent( evt ) ;

	// ------------ IMPORTANT ------------- !
	// we created the event so we need to delete it ...
	delete evt ;
	// -------------------------------------

      } // for eventNr

      //      delete runHdr ;
      
      lcWrt->close() ;
      
  } catch( std::exception& ex){
    
    cout << " an excpetion occured: " << endl ;
    cout << "  " << ex.what() << endl ;
    return 1 ;
  }
     
return 0;

}
