#include "TrackFitterFactory.h"
#include "TrackFitterBase.h"

// lcio
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/TrackImpl.h>
#include <lcio.h>

// clhep
#include <CLHEP/Vector/ThreeVector.h>

// default libraries
#include <iostream>
#include <cmath>

using namespace lcio;
using namespace marlintpc;
using namespace CLHEP;

// just a container struct to store the track parameters conveniently in a vector
struct TrackParameters
{
  double omega, d0, phi0, z0, tanLambda;
  
  // convenienve constructor
  TrackParameters (double _omega, double _d0, double _phi0, double _z0, double _tanLambda) :
    omega(_omega),
    d0(_d0),
    phi0(_phi0),
    z0(_z0),
    tanLambda(_tanLambda)
  {}    
};


// note: This structure is anything but safe.
// It's just a type for the return value of the calculateHitsPhiEqualsZero function.
struct HitsResiduals
{
  std::vector<Hep3Vector> hitsOnHelix;
  std::vector<std::pair<double, double> > realResiduals;
  
  HitsResiduals(size_t _nHits) : hitsOnHelix(_nHits), realResiduals(_nHits){}
};

// the calculation for straight tracks
HitsResiduals calculateHitsStraight(TrackParameters const & trackParameters)
{
  // We stick to the same trick as for the helix:
  // Calculate for phi0=0, the rotate.
  // For a straight line this means hits on a horizontal line at y=d0.
  // We use steps of d0, shifted by d0 / 2
  
  size_t nHits=8;
  HitsResiduals hitsResiduals(nHits);
  for (unsigned int i = 0; i < nHits ; i++)
  {
    Hep3Vector & theHit = hitsResiduals.hitsOnHelix[i];
    
    // n is running from [-nHits/2 .. nHits/2)
    int n = static_cast<int>(i) - nHits/2;

    // make alternating negative and positive offsets with different values.
    // With 8 hits there is a positive and a negative residual in each quadrant
    hitsResiduals.realResiduals[i].first  = (i%2==0 ? - 1 : 2);
    // also make positive and negative offsets for z
    hitsResiduals.realResiduals[i].second = (n % 4 == 3 ? 1.2 : -1.7 );

    theHit[0] = n*trackParameters.d0 + trackParameters.d0/2.;
    theHit[1] = trackParameters.d0 + hitsResiduals.realResiduals[i].first;
    theHit[2] = trackParameters.z0 + theHit[0] * trackParameters.tanLambda 
                +  hitsResiduals.realResiduals[i].second;
   

    // finally rotate the hit by phi0, which is convenient with Hep3Vector
    theHit.rotateZ( trackParameters.phi0 );
  }

  return hitsResiduals;
}

// Calculation for a helix
// Calculate eight hits on a full cirle, spaced by pi / 4. They have an offset of
// pi/8 wtr the axis from the helix centre to the origin, which runs through the PCA.
// At the PCA the parameter s (track length on the helix projection) is 0 (and angle s/radius)
HitsResiduals calculateHits(TrackParameters const & trackParameters)
{
  // The following code is written for real helix. If omega=0 it will break.
  // Treat straight tracks differently.
  if (trackParameters.omega==0) return calculateHitsStraight(trackParameters);
  

  // First Create a helix phi=0. This is easy to calculate with the following parameters:
  // Center at (0 , d0 - 1./omega)
  // Radius = 1/omega
  // phi = 0
  // tanLambda
  // z0
  
  double xCentre = 0;
  double yCentre =  trackParameters.d0 - 1./trackParameters.omega;
  double radius = fabs(1./trackParameters.omega);
  double signOmega = (trackParameters.omega > 0 ? 1. : -1.);

  size_t nHits=8;
  HitsResiduals hitsResiduals(nHits);
  for (unsigned int i = 0; i < nHits ; i++)
  {
    Hep3Vector & theHit = hitsResiduals.hitsOnHelix[i];
    
    // n is running from [-nHits/2 .. nHits/2)
    int n = static_cast<int>(i) - nHits/2;

    // make alternating negative and positive offsets with different values.
    // With 8 hits there is a positive and a negative residual in each quadrant
    hitsResiduals.realResiduals[i].first  = (i%2==0 ? - 1 : 2);
    // also make positive and negative offsets for z
    hitsResiduals.realResiduals[i].second = (n % 4 == 3 ? 1.2 : -1.7 );

    // with the change of curvature a negativ offset is either further away or closer
    // to the centre of the helix, but a negative offset is always left of the track,
    // looking in positive s direction. No matter if the centre is left or right of the
    // track. So we have to multiply the residual with the sign of omega here.
    double thisRadius = radius  - hitsResiduals.realResiduals[i].first * signOmega;

    // the angle on the helix wrt the PCA (i.e. s/radius)
    double s_angle=M_PI/nHits + n*M_PI/nHits*2;
    
    theHit[0] = xCentre + thisRadius * sin( s_angle);
    theHit[1] = yCentre + signOmega* thisRadius * cos( s_angle );

    // r*angle = s , with s=0 at the pci, s is the track length on the projection circle
    // We use the nominal radius for s because we have to be at the right angle
    theHit[2] = trackParameters.z0 + radius * s_angle*trackParameters.tanLambda 
                +  hitsResiduals.realResiduals[i].second;

    // finally rotate the hit by phi0, which is convenient with Hep3Vector
    theHit.rotateZ( trackParameters.phi0 );
  }

  return hitsResiduals;
}

int main()
{
  // Create helices. Systematically check 3 different radii, both signs of curvature , curvature = 0,
  // sign of d0, d0 = 0 , phi0 in all quadrants, both signs of z0, z0=0, both signs of tanLambda,
  // tanLambda = 0
  std::vector<TrackParameters> trackParametersVec;

  // let's just test all combinations
  std::vector<double> omegas;
  omegas.push_back(-1./660);
  omegas.push_back(-1./1000);
  omegas.push_back(-1./2500);
  omegas.push_back(0.);
  omegas.push_back(1./2500);
  omegas.push_back(1./1000);
  omegas.push_back(1./660);
  
  std::vector<double> d0s;
  d0s.push_back(-380.);
  d0s.push_back(0.);
  d0s.push_back(380.);

  std::vector<double> z0s;
  z0s.push_back(-275.);
  z0s.push_back(0.);
  z0s.push_back(275.);
 
  std::vector<double> phi0s;
  // put in 8 values at some angles between -pi and pi
  for (double angle = -2.73; angle <= M_PI ; angle += 0.78 )
  {
    phi0s.push_back(angle);
  }
  // now put in exactly 0, +- pi/2, +- pi
  phi0s.push_back(0.);
  phi0s.push_back(-M_PI_2);
  phi0s.push_back(M_PI_2);
  phi0s.push_back(-M_PI);
  phi0s.push_back(M_PI);
 

  std::vector<double> tanLambdas;
  tanLambdas.push_back(-0.1);
  tanLambdas.push_back(0.);
  tanLambdas.push_back(0.1);

  // now loop them all. This gives 2457 combinations.
  for (std::vector<double>::const_iterator omegaIter = omegas.begin();
       omegaIter != omegas.end(); ++omegaIter)
    for (std::vector<double>::const_iterator d0Iter = d0s.begin();
	 d0Iter != d0s.end(); ++d0Iter)
      for (std::vector<double>::const_iterator phi0Iter = phi0s.begin();
	   phi0Iter != phi0s.end(); ++phi0Iter)
	for (std::vector<double>::const_iterator z0Iter = z0s.begin();
	     z0Iter != z0s.end(); ++z0Iter)
	  for (std::vector<double>::const_iterator tanLambdaIter = tanLambdas.begin();
	       tanLambdaIter != tanLambdas.end(); ++tanLambdaIter)
	    trackParametersVec.push_back( TrackParameters(*omegaIter,
							  *d0Iter,
							  *phi0Iter,
							  *z0Iter,
							  *tanLambdaIter));


  int nTestsFailed = 0;

  for (std::vector<TrackParameters>::const_iterator trackParamsIter = trackParametersVec.begin();
       trackParamsIter != trackParametersVec.end(); ++trackParamsIter)
  {
    // create an lcio track
    TrackImpl *curvedTrack = new TrackImpl;

    curvedTrack->setTanLambda( trackParamsIter->tanLambda );
    curvedTrack->setZ0( trackParamsIter->z0 );
    curvedTrack->setPhi( trackParamsIter->phi0 );
    curvedTrack->setD0(  trackParamsIter->d0 );
    curvedTrack->setOmega( trackParamsIter->omega  );

    HitsResiduals hitsResiduals = calculateHits( *trackParamsIter );


    // debug: let's dump the hits
    //  for (unsigned int i = 0; i < hitsOnHelix.size() ; i++)
    //  {
    //    Hep3Vector & theHit = hitsOnHelix[i];
    //    std::cout << theHit[0] << "\t" << theHit[1] << "\t" << theHit[2] << std::endl;
    //  }
    
    for (unsigned int i = 0; i < hitsResiduals.hitsOnHelix.size() ; i++)
    {
      Hep3Vector & theHit = hitsResiduals.hitsOnHelix[i];
      TrackerHitImpl  * newLcioHit = new TrackerHitImpl;
      
      double pos[3];
      pos[0]= theHit[0];
      pos[1]= theHit[1];
      pos[2]= theHit[2];
      
      newLcioHit->setPosition( pos );
      
      curvedTrack->addHit( newLcioHit );      
    }


    // test the residuals
    TrackFitterBase *trackFitter = TrackFitterFactory::getTrackFitter( TrackFitterBase::FitterTypes::SIMPLEMINIMIZER, 0 );

    int nSubTestsFailed = 0;


    //    std::cout << "#Curved track"<< std::endl;
    for (unsigned int i = 0; i < hitsResiduals.hitsOnHelix.size() ; i++)
    {
      TrackerHit * hit =  curvedTrack->getTrackerHits()[i];

      EVENT::DoubleVec calculatedResiduals = trackFitter->calculateResiduals(hit, curvedTrack);
      
      double const * pos = hit->getPosition();
      // debug
//      std::cout << "Hit : "<< pos[0] << " ; "
//		<< pos[1] << " ; "
//		<< pos[2] << " ; Residuals: "
//		<< calculatedResiduals[0] << " ; " << calculatedResiduals[1]  << std::endl;

      // get the reference residual
      std::pair<double, double> &realResiduals = hitsResiduals.realResiduals[i];

      // FIXME: The 1e-3 is one micron. This is way too bad, but the numerical precision with
      // 1e-4 is not good enough and gives errors. 1e-6 (one nanometer) would be nice, 1e-5 ok.
      if ( ( fabs(realResiduals.first - calculatedResiduals[0]) > 1e-3 ) ||
	   ( fabs(realResiduals.second - calculatedResiduals[1]) > 1e-3 )||
	   (!std::isnormal(calculatedResiduals[0])) ||
	   (!std::isnormal(calculatedResiduals[1])) 
	   )
      {
	++nSubTestsFailed;
	std::cout << "#Error in Hit : "<< pos[0] << " ; "
		  << pos[1] << " ; "
		  << pos[2] << " ; calcuatedResiduals: "
		  << calculatedResiduals[0] << " ; " << calculatedResiduals[1] 
		  << " ; realResiduals " << realResiduals.first << " ; "
		  << realResiduals.second 
		  << " ;  residual differences are ("
		  << realResiduals.first - calculatedResiduals[0]
		  << " , " <<  realResiduals.second - calculatedResiduals[1]
		  << ")" << std::endl;
	std::cout << "GNUPLOT " << pos[0] << "\t"
		  << pos[1] << "\t"
		  << pos[2] << std::endl;
     }
    }

    // clean up the lcio stuff
    for (unsigned int i = 0; i < hitsResiduals.hitsOnHelix.size() ; i++)
    {
      delete curvedTrack->getTrackerHits()[i];
    }
    delete curvedTrack;


    if (nSubTestsFailed != 0)
    {
      std::cout << nSubTestsFailed << " Hits had errors on track with parameters "
		<< "omega="<<trackParamsIter->omega
		<< " ; d0="<<trackParamsIter->d0
		<< " ; phi0="<<trackParamsIter->phi0
		<< " ; z0="<<trackParamsIter->z0
		<< " ; tanLambda=" <<trackParamsIter->tanLambda
		<<std::endl; 
	++nTestsFailed;
    }
    
  }// for trackParamsIter


  std::cout << nTestsFailed <<  " out of " << trackParametersVec.size() 
	    << " tracks had errors." << std:: endl;
  // all return codes are %256 
  return (nTestsFailed>255 ? 255 : nTestsFailed);
}
