#include <gearxml/GearXML.h>
#include <gear/GearMgr.h>
#include <gear/GEAR.h>

#include <gearimpl/TPCParametersImpl.h>
#include <gearimpl/FixedPadSizeDiskLayout.h>
#include <gearimpl/TPCModuleImpl.h>

#include <marlin/Global.h>

#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <EVENT/LCIO.h>

#include <kaltest/TKalDetCradle.h>
#include <kaldet/GearTPCKalDetector.h>
#include <kaldet/EXTPCKalDetector.h>
#include <iostream>
#include <cmath>

#include "TPCHitDomain.h"

using namespace gear;
using namespace marlintpc;
using namespace IMPL;
using namespace EVENT;
using namespace std;
using namespace kaldet;

void printModuleRows(std::set< std::pair<int, int> > moduleRows)
{
   for(std::set< std::pair<int,int> >::iterator iter = moduleRows.begin(); iter != moduleRows.end(); ++iter)
   {
      cout << "module: " << iter->first << "   row: " << iter->second << endl;
   }
}

/// This function loops all rows in testModule, calculates the positon at the global angle (straight line) and 
/// checks the calculated domain. All rows have to be the same, and all modules in modulesInDomain have to be found.
/// Returns the number of failed tests and the total number of tests.
std::pair<unsigned int, unsigned int> testPositionsInModule(gear::TPCModule const& testModule, std::set<int> modulesInDomain, TPCHitDomain & hitDomain, double angle = 0)
{
  unsigned int nTestsFailed=0;
  for (int row = 0; row < testModule.getNRows(); ++row)
  {
    // we use pad 0 to determine the radius of this row
    Vector2D localPad0Center = testModule.getLocalPadLayout().getPadCenter( testModule.getPadIndex(row, 0) );
    double r = localPad0Center[0];
    double localPhi = angle - testModule.getAngle();

	cout << "rotation angle of the module is " << testModule.getAngle() << endl;
    

    // global coordinates at phi = angle
    Vector2D globalGearCoordinates = testModule.localToGlobal( r, localPhi );
	if(!testModule.isInsideModule(globalGearCoordinates[0], globalGearCoordinates[1] ))
	{
		cout << "the point for getting a domain is not inside the module!!!" << endl;
		continue;
	}

    // FIXME: thransfor the global Gear coordinates to cartesian, or at least do a check
    
    TVector3 padPosition(globalGearCoordinates[0], globalGearCoordinates[1], 0.);
    std::set< std::pair<int, int> > moduleRows = hitDomain.getModuleRows(padPosition);

	cout << endl << "Domain size is " << moduleRows.size() << endl;

    std::set< int> modulesFound;

    bool thisTestFailed = false;
    for(std::set< pair<int, int> >::iterator moduleRowIter = moduleRows.begin(); 
	    moduleRowIter != moduleRows.end(); ++moduleRowIter)
    {
      int thisModule = moduleRowIter->first;
      int thisRow    = moduleRowIter->second;

      if (thisRow != row)
      {
		 cout << "Row " << thisRow << " does not match." << endl;
	     thisTestFailed=true;
	     break;
      }

      if ( modulesInDomain.find(thisModule) == modulesInDomain.end() ) // the module is not in the list of reference modules
      {
		 cout << "module " << thisModule << " is not found!" << endl;
	     thisTestFailed=true;
	     break;
      }

      modulesFound.insert(thisModule);
    }

    // Now test that modulesFound and modulesInDomain have the same size. As we checked that all entries in modulesFound are also in modulesInDomain we are sure
    // that we found all modules we should have.
    if ( modulesFound.size() != modulesInDomain.size() )
    {
      cout << "modulesFound size is " << modulesFound.size() 
		   << ", but modulesInDomain size is " << modulesInDomain.size() << endl;

      thisTestFailed=true;
    }

    // now the referenceModuleRows and the moduleRows retuned from the hitDomain should be the same
    if (thisTestFailed)
    {
      ++nTestsFailed;
      std::cout << "Test failed!\n"
		<< "moduleRows from hitDomain:\n";
      printModuleRows( moduleRows );
      std::cout << "should have been row " << row << " in modules ";

      for (std::set<int>::iterator iter =  modulesInDomain.begin(); iter != modulesInDomain.end(); ++iter) 
	  {
		  std::cout << " " << *iter;
	  }
      std::cout << std::endl;

    }  
  }

  return pair<int, int>(nTestsFailed, testModule.getNRows());
}

int main(int  /*argc*/, char ** /*argv*/)
{
  // we hardcode the gear file name because we hard coded against it's content.
  std::string const gearFilename("gear_LP_TPC_GEM_7module.xml");

  // read pad layout from gear
  GearMgr* gearMgr;
  try{
    GearXML gearXML(gearFilename);

    gearMgr = gearXML.createGearMgr();
  }
  catch(gear::ParseException &)
  {
    std::cout << "Error: Gear file \"" << gearFilename << "\" not found." << std::endl;
    return -1;
  }

  GearTPCKalDetector gearTPCKalDetector = GearTPCKalDetector(*gearMgr);

  TKalDetCradle kalDetCradle;
  kalDetCradle.Install(gearTPCKalDetector);
  kalDetCradle.Close();

  TPCParameters const& tpcParameters = gearMgr->getTPCParameters();
  TPCHitDomain tpcHitDomain(&kalDetCradle, &gearTPCKalDetector, &(gearMgr->getTPCParameters()));

  unsigned int nTestsFailed=0;
  unsigned int nTestsTotal=0;

  // a module set
  std::set<int> modulesInDomain;

  modulesInDomain.insert(0);
  modulesInDomain.insert(1);	


  // test module 0, module 0 and 1 are on the same kalman layer	   
  std::pair<unsigned int, unsigned int> testResult = 
				testPositionsInModule(tpcParameters.getModule(1), modulesInDomain, tpcHitDomain, 5.39/180*M_PI);		  
  nTestsFailed+=testResult.first;
  nTestsTotal+=testResult.second;

  modulesInDomain.clear();
  modulesInDomain.insert(2);
  modulesInDomain.insert(3);
  modulesInDomain.insert(4);

  // test module 2
  testResult = testPositionsInModule(tpcParameters.getModule(2), modulesInDomain, tpcHitDomain, -8.50/180*M_PI);		  
  nTestsFailed+=testResult.first;
  nTestsTotal+=testResult.second;

  // test module 4
  testResult = testPositionsInModule(tpcParameters.getModule(4), modulesInDomain, tpcHitDomain, 8.27/180*M_PI);		  
  nTestsFailed+=testResult.first;
  nTestsTotal+=testResult.second;

  modulesInDomain.clear();
  modulesInDomain.insert(5);
  modulesInDomain.insert(6);

  // test module 5
  testResult = testPositionsInModule(tpcParameters.getModule(5), modulesInDomain, tpcHitDomain, -5./180*M_PI);		  
  nTestsFailed+=testResult.first;
  nTestsTotal+=testResult.second;

  cout << "Tested: " << nTestsTotal << ",  failed: " << nTestsFailed << endl;

  return nTestsFailed;
}
