#include <gearimpl/TPCParametersImpl.h>
#include <gearimpl/FixedPadSizeDiskLayout.h>
#include <gearimpl/TPCModuleImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <EVENT/LCIO.h>
#include <iostream>

#include "TPCHitMap.h"

using namespace gear;
using namespace marlintpc;
using namespace IMPL;
using namespace EVENT;
using namespace std;

LCCollectionVec hitCollection(LCIO::TRACKERHIT);

void testRowForOneHit( int row, LCObject const * hit,
		       std::set< MarlinTPCHit > const & hitsInRow,
		       unsigned int & nTestsTotal, unsigned int & nTestsFailed)
{
  // there should only be one hit with index 1
  ++nTestsTotal;
  if (hitsInRow.size() != 1)
  {
    cout << "Row " << row << " has " << hitsInRow.size() << " hits but should have 1" 
	 << endl;
    ++nTestsFailed;    
    return;
  }
  
  ++nTestsTotal;
  // be now that there is one hit, so it's safe to assume that begin() is valid and
  // may be dereferenced
  if ( hitsInRow.begin()->hit != hit)
  {
    cout << "The remaining hit in row 0 does not point hit index 1."
	 << endl;
    ++nTestsFailed;    
  }	
}

void testRowForTwoHits( int row, LCObject const * hit1,  LCObject const * hit2,
		       std::set< MarlinTPCHit > const & hitsInRow,
		       unsigned int & nTestsTotal, unsigned int & nTestsFailed)
{
  // there should only be one hit with index 1
  ++nTestsTotal;
  if (hitsInRow.size() != 2)
  {
    cout << "Row " << row << " has " << hitsInRow.size() << " hits but should have 2" 
	 << endl;
    ++nTestsFailed;    
    return;
  }

  // we are allowed to use begin and increase the iterator once because we tested the size  
  std::set< MarlinTPCHit >::iterator hitIter = hitsInRow.begin();

  ++nTestsTotal;
  if ( hitIter->hit != hit1)
  {
    cout << "The first hit in row " << row << " does not point the correct hit."
	 << endl;
    ++nTestsFailed;    
  }	
  
  ++hitIter;
  ++nTestsTotal;
  if ( hitIter->hit != hit2)
  {
    cout << "The first hit in row " << row << " does not point the correct hit."
	 << endl;
    ++nTestsFailed;    
  }
}

void testDomain(std::list<ModuleRowHit>& hitList, std::list<int> hitIndexList, unsigned int row,
		        unsigned int & nTestsTotal, unsigned int & nTestsFailed)
{
	//test hit number
	++nTestsTotal;

	if(hitList.size()!=hitIndexList.size())
	{
		cout << "Test failed: hit number is not correct in row " << row 
			 << ", size: " << hitList.size() << ", " << hitIndexList.size() << endl;

		++nTestsFailed;
		return;
	}

	//if hit number is correct, then test the hit content
	unsigned int hitsize = hitIndexList.size();

	for(unsigned int n = 0; n < hitsize; n++)
	{ 
		++nTestsTotal;

		int index = hitIndexList.front();
		ModuleRowHit hit = hitList.front();

		if( !(hit.module==0&&(unsigned int)hit.row==row&&hit.getTrackerHit()==hitCollection[2*row + index]) ) 
		{  
			cout << "Test failed: hit content is not correct in row " << row << endl;
			++nTestsFailed;
		}

		hitList.remove(hit);
		hitIndexList.remove(index);
	}
}


int main()
{
  // initialise the TPCParameters with 600 mm drift length
  // and kartesian coordinates
  TPCParametersImpl tpcParameters(600, PadRowLayout2D::CARTESIAN);
  
  // 10 cm rows with 1 cm wide pads from 10 cm to 1 m radius
  PadRowLayout2D *padLayout = new FixedPadSizeDiskLayout( 100 /*rMin*/, 1000 /*rMax*/,
							  100 /*padHeight*/, 10 /*PadWidth*/);
  
  int moduleID = 0;
  TPCModule *module = new TPCModuleImpl(moduleID, padLayout,  
					tpcParameters.getCoordinateType(), 40. /*readoutFrequency*/);
  tpcParameters.addModule(module);

  TPCHitMap hitMap(&tpcParameters, 0);

  // create two hits for each pad row


  for (int row = 0; row < padLayout->getNRows() ; ++row)
  {
    //FIXME: The interface does not quarantee that there is a pad 0 in the row
    double r = padLayout->getPadCenter( padLayout->getPadIndex(row, 0) )[0];
    
    // just create two straight line. one at phi = 0.1 rad, the other one at phi = -0.2 rad
    double phi = 0.1;
    gear::Vector2D globalPositionVector = module->localToGlobal(r, phi);

    double position[3];
    position[0]=globalPositionVector[0];
    position[1]=globalPositionVector[1];
    position[2]=1000.0-row;

    TrackerHitImpl * hit = new TrackerHitImpl;
    hit->setPosition(position);
    hitCollection.addElement(hit);

    phi = -0.1;
    globalPositionVector = module->localToGlobal(r, phi);

    position[0]=globalPositionVector[0];
    position[1]=globalPositionVector[1];
    position[2]=1000.0-(2*row+1);

    hit = new TrackerHitImpl;
    hit->setPosition(position);
    hitCollection.addElement(hit);
  }

  // start the testing and keep track how many tests failed.
  unsigned int nTestsFailed=0;
  unsigned int nTestsTotal=0;

  // the map should be empty
  ++nTestsTotal;
  if (!hitMap.isEmpty())
  {
    cout << "hit map is not empty, but should be" << endl;
    ++nTestsFailed;
  }
  
  hitMap.fillEventHits(&hitCollection);

  size_t nHitsInMap = hitCollection.size();
  
   // the map should be not empty
  ++nTestsTotal;
  if (hitMap.isEmpty())
  {
    cout << "hit map is empty, but should not be" << endl;
    ++nTestsFailed;
  }

 // Test: check the size
  ++nTestsTotal;
  if ( hitMap.size() != nHitsInMap )
  {
    cout << "Test failed: hitMap.size() is " << hitMap.size()
	 << " but should be " <<  nHitsInMap << std::endl;
    ++nTestsFailed;
  }

  ++nTestsTotal;
  if (hitMap.getActiveModuleRows().size() != static_cast<unsigned int>(padLayout->getNRows()))
  {
    cout << "Test failed: there are " << hitMap.getActiveModuleRows().size()
	 << " active moduleRows, but should be " <<  padLayout->getNRows() << std::endl;
    ++nTestsFailed;    
  }

  //test the content in domain
  for(int row = 0; row < padLayout->getNRows() ; ++row)
  {
     std::set< std::pair<int, int> >  moduleRowsInDomain;
	 moduleRowsInDomain.insert( pair<int, int>(0, row) );
     std::list<ModuleRowHit>  hitList = hitMap.getHitsInDomain(moduleRowsInDomain);

	 const unsigned int HITSIZE = 2; 
	 ++nTestsTotal;
	 if(hitList.size()!=HITSIZE)
	 {
		 cout << "Test failed:  the hit number in this domain, a module row, is not 2." << endl;
		 ++nTestsFailed;
	 }
	 else
	 {

	     for(unsigned int n = 0; n < HITSIZE; n++)
		 {
			++nTestsTotal;
			ModuleRowHit hit = hitList.back();
			if( !(hit.module==0&&hit.row==row&&hit.getTrackerHit() ) ) 
			{
				cout << "Test failed:  the index of modulerow is not consistent with the that in a ModuleRowHit, " 
					 << hit.module << "  " << hit.row << "  "
					 << endl;

				++nTestsFailed;
			}

		    ++nTestsTotal;
			if(hit.getTrackerHit()!=hitCollection[2*row+n])
			{
				cout << "Test failed:  the address of real hit is not consistent with the that in a ModuleRowHit, " 
					 << hit.getTrackerHit() << endl;

				++nTestsFailed;
			}

			hitList.remove(hit);
		 }

	 }
  }

  // let's remove some of the hits:
  // the first hit in row 2 (hit index 4)
  // the second it in row 5 (hit index 11)
  ModuleRowHit hit4(moduleID, 2,  dynamic_cast<TrackerHit*>(hitCollection[4]));
  hitMap.removeHit(hit4);
  --nHitsInMap;

  ModuleRowHit hit11(moduleID, 5, dynamic_cast<TrackerHit*>(hitCollection[11]));
  hitMap.removeHit(hit11);
  --nHitsInMap;

  // retest size and size of active moduleRows
  ++nTestsTotal;
  if ( hitMap.size() != nHitsInMap )
  {
    cout << "Test failed: hitMap.size() is " << hitMap.size()
	 << " but should be " <<  nHitsInMap << std::endl;
    ++nTestsFailed;
  }

  // there should still be all rows active
  ++nTestsTotal;
  if (hitMap.getActiveModuleRows().size() != static_cast<unsigned int>(padLayout->getNRows()))
  {
    cout << "Test failed: there are " << hitMap.getActiveModuleRows().size()
	 << " active moduleRows, but should be " <<  padLayout->getNRows() << std::endl;
    ++nTestsFailed;    
  }
  
  // test the first hit in the map
  ++nTestsTotal;
  ModuleRowHit firstHit = hitMap.getFirstHit();
  if ( (firstHit.module != moduleID) ||
       (firstHit.row != 0 ) ||
       (firstHit.getTrackerHit() != hitCollection[0]) )
  {
    cout << "Test failed: first hit is (" << firstHit.module << ", " << firstHit.row
	 << ", " << firstHit.getTrackerHit() <<") but should be ("
	 << moduleID << ", 0, " <<  hitCollection[0] << ")" << endl;
    ++nTestsFailed;    
  }

  // now let's remove some more hits from a list:
  // the very first hit (hit index 0)
  // the first hit in row 7 (hit index 14)
  // the first hit in row 5 (hit index 10, should change the total number of active module rows)
  vector<ModuleRowHit> removeMeList;
  removeMeList.push_back( ModuleRowHit(moduleID, 0,  dynamic_cast<TrackerHit*>(hitCollection[0])) );
  removeMeList.push_back( ModuleRowHit(moduleID, 7,  dynamic_cast<TrackerHit*>(hitCollection[14])) );
  removeMeList.push_back( ModuleRowHit(moduleID, 5,  dynamic_cast<TrackerHit*>(hitCollection[10])) );

  nHitsInMap -=  removeMeList.size();
  hitMap.removeHits(removeMeList);

  // retest the size
  ++nTestsTotal;
  if ( removeMeList.size() != 0 )
  {
    cout << "Test failed: removeMeList should be empty now." << std::endl;
    ++nTestsFailed;
  }


  // retest the size
  ++nTestsTotal;
  if ( hitMap.size() != nHitsInMap )
  {
    cout << "Test failed: hitMap.size() is " << hitMap.size()
	 << " but should be " <<  nHitsInMap << std::endl;
    ++nTestsFailed;
  }

  // do a full test of all the remaining hits
  std::list< std::pair<int, int> > activeModuleRows = hitMap.getActiveModuleRows();
  
  ++nTestsTotal;
  if (activeModuleRows.size() != static_cast<unsigned int>(padLayout->getNRows())-1 )
  {
    cout << "Test failed: there are " << activeModuleRows.size()
	 << " active moduleRows, but should be " <<  padLayout->getNRows()-1 << std::endl;
    ++nTestsFailed;    
  }

  // test all the hits
  for (std::list< std::pair<int, int> >::iterator moduleRowIter = activeModuleRows.begin();
       moduleRowIter != activeModuleRows.end(); ++moduleRowIter)
  {
    int thisModuleID = moduleRowIter->first;
    int thisRow = moduleRowIter->second;

    std::set< MarlinTPCHit > const & hitsInRow = 
      hitMap.getHitsInModuleRow(thisModuleID,  thisRow);

    // ok, this is hard coded, as are the hits which we removed
    switch (thisRow)
    {
      case 0: 
	// there should only be one hit with index 1
	testRowForOneHit( thisRow, hitCollection[1], hitsInRow, nTestsTotal, nTestsFailed);
	break;

      case 1: 
	testRowForTwoHits( thisRow, hitCollection[3], hitCollection[2],
			   hitsInRow, nTestsTotal, nTestsFailed);
	break;

      case 2: 
	testRowForOneHit( thisRow, hitCollection[5], hitsInRow, nTestsTotal, nTestsFailed);
	break;

      case 3: 
	testRowForTwoHits( thisRow, hitCollection[7], hitCollection[6],
			   hitsInRow, nTestsTotal, nTestsFailed);
	break;

      case 4: 
	testRowForTwoHits( thisRow, hitCollection[9], hitCollection[8],
			   hitsInRow, nTestsTotal, nTestsFailed);
	break;

      case 5:
	++nTestsTotal;
	cout << "Test failed: row 5 is still in the list of active module rows, but should not"
	     << endl;
	++nTestsFailed;    
	break;

      case 6: 
	testRowForTwoHits( thisRow, hitCollection[13], hitCollection[12],
			   hitsInRow, nTestsTotal, nTestsFailed);
	break;

      case 7: 
	testRowForOneHit( thisRow, hitCollection[15], hitsInRow, nTestsTotal, nTestsFailed);
	break;	

      case 8: 
	testRowForTwoHits( thisRow, hitCollection[17], hitCollection[16],
			   hitsInRow, nTestsTotal, nTestsFailed);
	break;

      case 9: 
	testRowForTwoHits( thisRow, hitCollection[19], hitCollection[18],
			   hitsInRow, nTestsTotal, nTestsFailed);
	break;

      default:
        ++nTestsTotal;
	cout << "Unexpected row number " << thisRow
	     << endl;
	++nTestsFailed;    
    }// switch
  }// for moduleRowIter


  // retest the first hit in the map, now should be hit index 1
  ++nTestsTotal;
  ModuleRowHit newFirstHit = hitMap.getFirstHit();
  if ( (newFirstHit.module != moduleID) ||
       (newFirstHit.row != 0 ) ||
       (newFirstHit.getTrackerHit() != hitCollection[1]) )
  {
    cout << "Test failed: first hit is (" << newFirstHit.module << ", " << newFirstHit.row
	 << ", " << newFirstHit.getTrackerHit() <<") but should be ("
	 << moduleID << ", 0, " <<  hitCollection[1] << ")" << endl;
    ++nTestsFailed;    
  }

  //
  //retest getHitsInDomain
  //
  for(int row = 0; row < padLayout->getNRows() ; ++row)
  {
	  std::set< std::pair<int, int> >  moduleRowsInDomain;
	  moduleRowsInDomain.insert( pair<int, int>(0, row) );
      std::list<ModuleRowHit>  hitList = hitMap.getHitsInDomain(moduleRowsInDomain);

	  std::list<int> hitIndexList;
	  switch(row)
	  {
		  case 0:
		  //hit 1 
	      hitIndexList.push_back(1);
		  break;
	       
		  case 1:
		  //hit 2,3 
	      hitIndexList.push_back(1);
	      hitIndexList.push_back(0);
		  break;

		  case 2:
		  //hit 5 
	      hitIndexList.push_back(1);
		  break;

		  case 3:
		  //hit 6,7
	      hitIndexList.push_back(1);
	      hitIndexList.push_back(0);
		  break;

		  case 4:
		  //hit 8,9
	      hitIndexList.push_back(1);
	      hitIndexList.push_back(0);
		  break;

		  case 5:
		  //hit 
		  break;

		  case 6:
		  //hit 12, 13
		  hitIndexList.push_back(1);
		  hitIndexList.push_back(0);
		  break;

		  case 7:
		  //hit 15
		  hitIndexList.push_back(1);
		  break;

		  case 8:
		  //hit 16,17
		  hitIndexList.push_back(1);
		  hitIndexList.push_back(0);
		  break;
	  }	  

	  testDomain(hitList, hitIndexList, row, nTestsTotal, nTestsFailed);
  }

  // now clear the list and test that it is empty and size is 0
  hitMap.clear();

  ++nTestsTotal;
  if (!hitMap.isEmpty())
  {
    cout << "hitMap is not empty, but should be" << endl;
    ++nTestsFailed;    
  }

  ++nTestsTotal;
  if (hitMap.size()!=0)
  {
    cout << "hitMap size is not 0, but should be" << endl;
    ++nTestsFailed;    
  }

  cout << nTestsFailed << " out of "<<nTestsTotal <<" tests failed." <<endl;
  return nTestsFailed;
}
